/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include "accountdlg.h"

#include <qiconview.h>
#include <qpushbutton.h>

#include <iaccount.h>
#include <intl.h>
#include <isubsystem.h>
#include "kiltdown.h"
#include "localized.h"
#include "main.h"
#include "selectme.h"
#include "uiicons.h"

/**
 *=------------------------------------------------------------------------=
 *                          Local Classes
 *=------------------------------------------------------------------------=
 * these two classes help us store a little extra data.
 */
class NewAccountIconItem : public QIconViewItem { 
  public:
    NewAccountIconItem(QIconView *a, QIconViewItem *b, const QString &c, const QPixmap &d)
        : QIconViewItem(a, b, c, d) { }
};

class AccountIconItem : public QIconViewItem {

  public:
    AccountIconItem(QIconView *a, QIconViewItem *b, const QString & c, const QPixmap &d)
        : QIconViewItem(a, b, c, d) { }
    IAccount *account;
};

/**
 *=------------------------------------------------------------------------=
 * AccountsDialog::AccountsDialog
 *=------------------------------------------------------------------------=
 * creates and initializes everything necessary for the dialog.
 */
AccountsDialog::AccountsDialog()
: QDialog(0,0, true, WStyle_NoBorder)
{

    iconView = NULL;
    addAccount = NULL;
    closeButton = NULL;
}


/**
 *=------------------------------------------------------------------------=
 * AccountsDialog::~AccountsDialog
 *=------------------------------------------------------------------------=
 * cleans up!
 */
AccountsDialog::~AccountsDialog()
{
    if (iconView) delete iconView;
    if (closeButton) delete closeButton;
}

/**
 *=------------------------------------------------------------------------=
 * AccountsDialog::addAccounts
 *=------------------------------------------------------------------------=
 * adds all the accounts to the icon view.
 *
 * Parameters:
 *      LinkList<IAccount *> &          - [in]  accounts to add.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountsDialog::addAccounts
(
    LinkList<IAccount *> & accounts
)
{
    QIconViewItem *root;
    AccountIconItem *item;
    const char *pstr;
    QPixmap *pixmap;
    IAccount *a;
    ERRCODE er;

    /**
     * First, add the "Add new Account" icon.
     */
    pstr = intlLoadString(ID_ADD_NEW_ACCOUNT);
    if (!pstr) return E_OUTOFMEMORY;
    root = new NewAccountIconItem(this->iconView, NULL, pstr, QPixmap(accountslgIcon));
    localFree(pstr);
    if (!root) {
        K_FAIL("How come I can't create the root icon ???\n");
        return E_OUTOFMEMORY;
    }
    addAccount = root;

    /**
     * Whiz through the accounts getting their name and pixmap.
     */
    er = accounts.firstElement(&a);
    while (er == S_OK) {
        er = a->getDisplayName(&pstr);
        RETURN_ON_FAILURE(er);

        er = a->getPixmap(LargePixmap, pixmap);
        if (FAILED(er)) {
            localFree(pstr);
            return er;
        }

        item = new AccountIconItem(this->iconView, root, pstr, *pixmap);
        item->account = a;
        delete pixmap;
        localFree(pstr);

        er = accounts.nextElement(&a);
    }
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * AccountsDialog::addNewAccount
 *=------------------------------------------------------------------------=
 * shows a dialog to let you choose an account type and then brings up the 
 * appropriate UI to create a new account.
 */
void AccountsDialog::addNewAccount()
{
    LinkList<unsigned int> list;
    const char *pstr, *pstr2;
    ISubsystem *s = NULL;
    SelectMeDialog smd;
    unsigned int ui;
    ERRCODE er;
    int ret;

    /**
     * initialize the dialog
     */
    pstr = intlLoadString(ID_ACCOUNTTYPE);
    pstr2 = intlLoadString(ID_ACCOUNTTYPE_TEXT);
    if (!pstr || !pstr2) {
        if (pstr) localFree(pstr);
        if (pstr2) localFree(pstr2);
        errorShowError(E_OUTOFMEMORY);
        return;
    }

    smd.initializeUI(pstr, pstr2);
    localFree(pstr);
    localFree(pstr2);

    /**
     * Add a list of possible account types.
     */
    er = subsystems.firstElement(&s);
    while (er == S_OK) {
        s->getDisplayName(&pstr);
        if (pstr) {
            smd.addItem(pstr, (unsigned int)s);
            localFree(pstr);
        }
        er = subsystems.nextElement(&s);
    }

    /**
     * Show the dialog
     */
    ret = smd.exec();

    /**
     * If they chose something, figure out what it is.
     */
    if (ret == QDialog::Accepted) {
        
        ui = 0;
        smd.getSelectedItems(list);
        list.firstElement(&ui);
        if ((ISubsystem *)ui) ((ISubsystem *)ui)->createNewAccount();

        LinkList<IAccount *> accts;
        localMainWidget->getAccounts(accts);

        iconView->clear();
        addAccounts(accts);
    }
}


/**
 *=------------------------------------------------------------------------=
 * AccountsDialog::closeButtonPushed
 *=------------------------------------------------------------------------=
 */
void AccountsDialog::closeButtonPushed()
{
    this->accept();
}

/**
 *=------------------------------------------------------------------------=
 * AccountsDialog::iconDoubleClicked
 *=------------------------------------------------------------------------=
 * an icon was double clicked.  figure out what to do with it.
 *
 * Parameters:
 *      QIconViewIcon *                 - [in]  item clicked
 */
void AccountsDialog::iconDoubleClicked
(
    QIconViewItem *item
)
{
    NewAccountIconItem *naii;
    AccountIconItem *aii;

    if ((naii = dynamic_cast<NewAccountIconItem *>(item)) != NULL) {
        addNewAccount();
    } else if ((aii = dynamic_cast<AccountIconItem *>(item)) != NULL) {
        aii->account->showAccountProperties(this);
    }
}


/**
 *=------------------------------------------------------------------------=
 * AccountsDialog::initialize
 *=------------------------------------------------------------------------=
 * sets up da UI.
 *
 * Parameters:
 *      LinkList<ISubsystem *> &        - [in]  possible subsystems.
 *      LinkList<IAccount *> &          - [in]  list of accounts.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountsDialog::initialize
(
    LinkList<ISubsystem *> & subs,
    LinkList<IAccount *> & accounts
)
{
    const char *pstr;
    ERRCODE er;

    /**
     * This is a deep copy.
     */
    this->subsystems = subs;

    /**
     * Set our title
     */
    pstr = intlLoadString(ID_ACCOUNTS_DIALOG);
    if (!pstr) return E_OUTOFMEMORY;
    this->setCaption(pstr);
    localFree(pstr);

    /**
     * Next, create the icon view !!!
     */
    this->iconView = new QIconView(this);
    if (!this->iconView) return E_OUTOFMEMORY;
    iconView->setAutoArrange(true);
    iconView->setResizeMode(QIconView::Adjust);
    iconView->setGridX(75);
    iconView->setGridY(75);

    /**
     * Create the close button
     */
    this->closeButton = new QPushButton(this);
    if (!this->closeButton) return E_OUTOFMEMORY;
    pstr = intlLoadString(ID_FILE_CLOSE_TEXT);
    if (!pstr) return E_OUTOFMEMORY;
    closeButton->setText(pstr);
    closeButton->setDefault(true);
    localFree(pstr);

    /**
     * Set up the accounts.
     */
    er = addAccounts(accounts);
    RETURN_ON_FAILURE(er);

    /**
     * Set our size.
     */
    this->resize(400, 300);

    /**
     * finally, hook up a coupla events
     */
    QObject::connect(closeButton, SIGNAL(clicked()), this, SLOT(closeButtonPushed()));
    QObject::connect(iconView, SIGNAL(doubleClicked(QIconViewItem *)), this, SLOT(iconDoubleClicked(QIconViewItem *)));

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * AccountsDialog::resizeEvent
 *=------------------------------------------------------------------------=
 * dialog is changing size.  figure things out.
 *
 * Parameters:
 *      QResizeEvent *              - [in]  details
 */
void AccountsDialog::resizeEvent
(
    QResizeEvent *event
)
{
    QSize sz = this->closeButton->sizeHint();
    QSize size = event->size();

    this->iconView->setGeometry(0, 0, size.width(), size.height() - (sz.height() + 8) );
    this->closeButton->move(size.width() - sz.width() - 20, size.height() - sz.height() - 3);
}


