/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <sanity.h>

#include <ctype.h>
#include <string.h>

#include "address.h"


/**
 *=------------------------------------------------------------------------=
 * addressExtractEmailAddress
 *=------------------------------------------------------------------------=
 * given a string that is potentially a full-blown address of the form:
 *  "User Name" <user@domain.com>
 * extract the 'user@domain.com' portion of it and return it to the caller.
 *
 * Parameters:
 *      const char *                - [in]  address to decode
 *      const char *&               - [out] the email portion of it.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE addressExtractEmailAddress
(
    const char *addr,
    const char *&email
)
{
    char buf[MAX_EADDRESS_LEN];
    int x = 0;
    int y = 0;
    bool inaddr = false;

    while (addr[x] != '\0') {

        switch (addr[x]) {
            case '"':
            case '\'':
            case '\r':
            case '\n':
            case ' ':
            case '\t':
            case '\v':
                break;

            case '<':
                if (inaddr) fprintf(stderr, "malformed email address %s\n", addr);
                inaddr = true;
                y = 0;
                break;

            case '>':
                if (!inaddr) fprintf(stderr, "malformed email address %s\n", addr);
                buf[y] = '\0';
                goto Done;

            default:
                buf[y++] = addr[x];
                break;


        }

        x++;
    }

    /**
     * Okay, if we're here, then we think we had an email address with no <>'s
     * in it, and just want to return that now.
     */
  Done:
    buf[y] = '\0';
    if (buf[0] == '\0') strncpy(buf, "<unknown-recipient>", sizeof(buf));
    email = copyString(buf);
    return (email) ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * AddressExtractNameParts
 *=------------------------------------------------------------------------=
 * given a string representinga whole name, extract the first and last name
 * from it.
 *
 * Parameters:
 *      const char *                - [in]  whole name.
 *      const char * &              - [out] first name.
 *      const char * &              - [out] last name.
 */
ERRCODE addressExtractNameParts
(
    const char *wholeName,
    const char * &firstName,
    const char * &lastName
)
{
    char buf[256];
    int x, y;

    if (!wholeName) return E_PARAM;
    firstName = lastName = NULL;

    /**
     * Whiz through looking for punctuation.  If there isn't any, then we'll
     * just assume white space separates the names.  If there's only one name,
     * it's a first name.  We hav a limit of 256 characters on either name.
     */
    x = 0;
    y = 0;
    while (isspace(wholeName[x]) && wholeName[x] != '\0') x++;
    
    while (wholeName[x] != '\0' && wholeName[x] != ',' && wholeName[x] != ';') {
        buf[y++] = wholeName[x++];
        if ((unsigned int)y >= (sizeof(buf) / sizeof(char)) - 1) break;
    }

    /**
     * If they didn't have a last name/first name separator, then see if there
     * is a space that would separate the names.
     */
    if (wholeName[x] == '\0') {
        int z = y;
        buf[y] = '\0';
        y = 0;
        while (buf[y] != '\0' && !isspace(buf[y])) y++;
        buf[y] = '\0';
        firstName = copyString(buf);
        if (!firstName) return E_OUTOFMEMORY;
        if (y != z) {
            lastName = copyString(buf + y + 1);
            if (!lastName) {
                localFree(firstName);
                return E_OUTOFMEMORY;
            }
        }
        return S_OK;
    }

    /**
     * Now, skip over the , or ;, skip over any space, and everything else
     * is the first name.
     */
    buf[y] = '\0';
    lastName = copyString(buf);
    if (wholeName[x] == ',' || wholeName[x] == ';') x++;
    while (isspace(wholeName[x]) && wholeName[x] != '\0') x++;
    y = 0;
    while (wholeName[x] != '\0') {
        buf[y++] = wholeName[x++];
        if ((unsigned int)y >= (sizeof(buf) / sizeof(char)) - 1) break;
    }

    buf[y] = '\0';
    firstName = copyString(buf);

    if (!firstName || !lastName) {
        if (firstName) localFree(firstName);
        if (lastName) localFree(lastName);
        return E_OUTOFMEMORY;
    }
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * addressExtractUserName
 *=------------------------------------------------------------------------=
 * given a possibly full email address, such as "joe Blow" <joe@blow.com>,
 * extract the user name "Joe Blow" (sans quotes)
 *
 * Parameters:
 *      const char *                - [in]  extract my name.
 *      char *                      - [in]  where to put result.
 *      int                         - [in]  size of buf.
 */
void addressExtractUserName
(
    const char *address,
    char *buf,
    int sizeofBuf
)
{
    bool inquotes;
    int x = 0;
    int y = 0;

    /**
     * First, skip over any whitespace.
     */
    while (isspace(address[x])) x++;

    /**
     * Now, if the first char is a <, then it's part of the address, and
     * we'll just extract that.  Otherwise, it's the name.
     */
    if (address[x] == '<') {
        x++;
        while (y < sizeofBuf && address[x] != '\0') {
            
            if (address[x] == '>') break;
            if (address[x] != '\r' && address[x] != '\n')
                buf[y++] = address[x++];
            else
                x++;
        }
    } else {

        /**
         * Extract the user name now, making sure to be able to handle quotes
         * and escaped quotes.
         */
        if (address[x] == '"') {
            inquotes = true;
            x++;
        } else {
            inquotes = false;
        }

        /**
         * trim off leading whitespace
         */
        while (isspace(address[x])) x++;

        /**
         * Whiz through copying things over.  If we're a quoted name, then
         * non-escaped quotes terminate us.  Otherwise, a < character will
         * definitely end us.
         */
        while (y < sizeofBuf && address[x] != '\0') {

            if (inquotes && x && address[x] == '"' && address[x - 1] != '\\')
                break;

            if (!inquotes && address[x] == '<')
                break;

            if (address[x] != '\r' && address[x] != '\n')
                buf[y++] = address[x++];
            else
                x++;
        }
    }

    /**
     * Tie it off, strip trailing whitespace,  and call it a day.
     */
    if (y == sizeofBuf) y--;
    buf[y] = '\0';
    y--;
    while (isspace(buf[y])) y--;
    buf[++y] = '\0';
}

/**
 *=------------------------------------------------------------------------=
 * addressExtractUserNameCopy
 *=------------------------------------------------------------------------=
 * given a possibly full email address, such as "joe Blow" <joe@blow.com>,
 * extract the user name "Joe Blow" (sans quotes) and return a copy of the
 * result.
 *
 * Parameters:
 *      const char *                - [in]  extract my name.
 *      const char * &              - [out] the new name
 *
 * Output:
 *      ERRCODE
 */
ERRCODE addressExtractUserNameCopy
(
    const char *address,
    const char * &name
)
{
    char buf[MAX_EADDRESS_LEN];

    addressExtractUserName(address, buf, sizeof(buf));

    name = copyString(buf);
    return name ? S_OK : E_OUTOFMEMORY;
}


/**
 *=------------------------------------------------------------------------=
 * addressGenerateFullName
 *=------------------------------------------------------------------------=
 * given a user first anem and last name, generate a full name string for
 * them and put it in the given buffer.
 *
 * Parameters:
 *      AddressBookItem *           - [in]  item with the fino.
 *      char *                      - [in]  buf
 *      unsigned int                - [in]  sizeof buf
 */
void addressGenerateFullName
(
    const char *first,
    const char *last,
    char *buf,
    unsigned int sizeofBuf
)
{
    if (first && last) {
        snprintf(buf, sizeofBuf, "%s %s", first, last);
    } else if (first) {
        strncpy(buf, first, sizeofBuf);
        buf[sizeofBuf - 1] = '\0';
    } else {
        strncpy(buf, last, sizeofBuf);
        buf[sizeofBuf - 1] = '\0';
    }
}

/**
 *=------------------------------------------------------------------------=
 * addressGenerateSmtp
 *=------------------------------------------------------------------------=
 * given a name and email address, generate the SMTP header version of this.
 *
 * i.e.   bobo the clown and bobo@theclown.com 
 *         become ...
 *       "bob the clown" <bobo@theclown.com>
 *
 * Parameters:
 *      QString &                       - [in]  user name.
 *      const char *                    - [in]  email address
 *      char *buf                       - [in]  where to put result.
 *      unsigned int                    - [in]  sizeof of resulting buf.
 *
 * Notes:
 *      UNDONE: marcw, 3.15.2001 -- we need to be able to encode addresss
 *      and the like that have more than US-ASCII
 */
void addressGenerateSmtp
(
    const QString &name,
    const char *email,
    char *buf,
    unsigned int sizeofBuf
)
{
    snprintf(buf, sizeofBuf, "\"%s\" <%s>", (const char *)name, email);
}



/**
 *=------------------------------------------------------------------------=
 * addressGenerateSmtpCopy
 *=------------------------------------------------------------------------=
 * given a name and email address, generate the SMTP header version of this.
 *
 * i.e.   bobo the clown and bobo@theclown.com 
 *         become ...
 *       "bob the clown" <bobo@theclown.com>
 *
 * Parameters:
 *      QString &                       - [in]  user name.
 *      const char *                    - [in]  email address
 *      const char * &                  - [out] where to put it.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE addressGenerateSmtpCopy
(
    const QString &name,
    const char *email,
    const char * & result
)
{
    char buf[MAX_EADDRESS_LEN];

    addressGenerateSmtp(name, email, buf, sizeof(buf));

    result = copyString(buf);

    return result ? S_OK : E_OUTOFMEMORY;
}



