/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>
#include <errno.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qlistview.h>
#include <sys/param.h>


#include "about.dlg.h"
#include "addrbookentry.h"
#include "address.h"
#include "addressbook.h"
#include "kwindowbase.h"
#include "kiltdown.h"
#include "localized.h"
#include "main.h"
#include "uimenu.h"
#include "uitoolbar.h"
#include "userpref.h"


const char *defaultAddressBookName = "addresses.csv";
const char *versionInfo = "1.ADDRESS";
const char *versionHeader = "1.ADDRESS\n";

/**
 * Notes/Comments field is this big.
 */
static const int MAX_ENTRY_SIZE = 512;
static const int MAX_ADDRLINE_LEN = 8192;

/**
 * Passed in to the constructor for KWindowBase to save us from having to keep
 * track of EVERYTHING
 */
KWBVariables addressBookVars = {

    &prefAddressBookWindowLeft,
    &prefAddressBookWindowTop,
    &prefAddressBookWindowWidth,
    &prefAddressBookWindowHeight,
    &prefAddressBookToolBarVisible,
    &prefAddressBookToolBarUsesBig,
    &prefAddressBookToolBarShowLabels,
    &prefAddressBookToolBarLocation
};

/**
 * This is our overloaded version of the QListViewItem.  It lets us store some
 * extra goo.
 */
class ABListViewItem : public QListViewItem {

  public:
    ABListViewItem( AddressBookItem *item, QListView * parent, QString label1, 
                    QString label2 = QString::null, 
                    QString label3 = QString::null, 
                    QString label4 = QString::null, 
                    QString label5 = QString::null, 
                    QString label6 = QString::null, 
                    QString label7 = QString::null,
                    QString label8 = QString::null)
         : QListViewItem(parent, label1, label2, label3, label4, label5, label6,
                         label7, label8) {
        this->item = item;
    }

    AddressBookItem *item;
};


#define CK(ptr)                if (!(ptr)) return E_OUTOFMEMORY;
#define LEFT_MARGIN            2
#define CELL_PAD               3


/**
 *=------------------------------------------------------------------------=
 * AddressBook::AddressBook
 *=------------------------------------------------------------------------=
 * initializes a blank isntance of our addressbook class. you really need
 * to call initialize() which in turn calls KWindowBase::initializeUI() to
 * make anything work properly, however!!!
 */
AddressBook::AddressBook()
: KWindowBase(addressBookVars, 0)
{
    DEBUG_TEST_START;

    listView = NULL;
    searchEdit = NULL;
    searchLabel = NULL;
}


/**
 *=------------------------------------------------------------------------=
 * AddressBook::~AddressBook
 *=------------------------------------------------------------------------=
 * time to take out the trash !!!
 */
AddressBook::~AddressBook()
{
    prefAddressBookNameFieldWidth = listView->columnWidth(0);
    prefAddressBookEmailFieldWidth = listView->columnWidth(1);
    prefAddressBookAddrFieldWidth = listView->columnWidth(2);
    prefAddressBookPhoneFieldWidth = listView->columnWidth(3);

    if (listView) delete listView;
    if (searchEdit) delete searchEdit;
    if (searchLabel) delete searchLabel;

    if (localMainWidget) localMainWidget->addressBookClosed();

    AddressBookItem *i;
    ERRCODE er;

    er = addresses.firstElement(&i);
    while (er == S_OK) {
        nukeAddressBookItem(i);
        er = addresses.nextElement(&i);
    }

    DEBUG_TEST_STOP;

}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::addNewEntryFromAddress
 *=------------------------------------------------------------------------=
 * given an SMTP email address, go and
 *
 * 1. make sure we don't already have it.
 * 2. add it to our list.
 * 3. add it to the UI, if appropriate.
 * 4. save it to our db.
 *
 * Parameters:
 *      const char *                    - [in]  smtp email address
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AddressBook::addNewEntryFromAddress
(
    const char *fullAddress
)
{
    AddressBookItem *item = NULL;
    AddressBookItem *itemFind;
    const char *pstrAddress = NULL;
    const char *pstrFirst = NULL;
    const char *pstrName = NULL;
    const char *pstrLast = NULL;
    ERRCODE er;


    /**
     * First, extract the parts we need to extract.
     */
    er = addressExtractEmailAddress(fullAddress, pstrAddress);
    GOTO_ON_FAILURE(er, Error);
    er = addressExtractUserNameCopy(fullAddress, pstrName);
    GOTO_ON_FAILURE(er, Error);
    er = addressExtractNameParts(pstrName, pstrFirst, pstrLast);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Now, go and see if we already have this entry?
     */
    er = findAddress(pstrAddress, itemFind);
    if (er != S_FALSE) {
        /**
         * If the address exists, show an error.
         */
        if (er == S_OK) {
            char buf[512];
            er = E_ADDRESS_EXISTS;
            addressGenerateFullName(itemFind->firstName, itemFind->lastName,
                                    buf, sizeof(buf));
            errorSaveRichErrorInfo(pstrAddress, buf);
        }
        goto Error;
    }

    /**
     * Next, create the empty entry to hold all the info.
     */    
    item = (AddressBookItem *)localAlloc(sizeof(AddressBookItem));
    if (!item) goto OOM;
    memset(item, 0, sizeof(AddressBookItem));

    /**
     * Now, save out the info we have already.
     */
    item->emailAddress = pstrAddress;
    item->firstName = pstrFirst;
    item->lastName = pstrLast;

    /**
     * UNDONE: marcw, 2001-4-19 -- show the address properties here so the
     * user can change them !!!
     */

    /**
     * Now add the entry to the UI and to the database.
     */
    addUIEntry(item);
    er = writeEntry(item);
    GOTO_ON_FAILURE(er, Error);

    /**
     * be-de-be-de-be-de that's all folks!
     */
    localFree(pstrName);    
    return S_OK;

  OOM:
    er = E_OUTOFMEMORY;
  Error:
    if (item) nukeAddressBookItem(item);
    if (pstrAddress) localFree(pstrAddress);
    if (pstrFirst) localFree(pstrFirst);
    if (pstrName) localFree(pstrName);
    if (pstrLast) localFree(pstrLast);
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::addUIEntry
 *=------------------------------------------------------------------------=
 * add the given entry to the UI.
 *
 * Parameters:
 *      AddressBookItem *               - [in]  add me please.
 */
void AddressBook::addUIEntry
(
    AddressBookItem *item
)
{
    char buf[512];

    addressGenerateFullName(item->firstName, item->lastName, buf, sizeof(buf));
    new ABListViewItem(item, listView, buf, item->emailAddress);
    addresses.addElement(item);
}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::createBlankDB
 *=------------------------------------------------------------------------=
 * createsa  blank addressbook file.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AddressBook::createBlankDB
(
    const char *path
)
{
    FILE *db;

    db = fopen(path, "a");
    if (!db) {
        errorSaveRichErrorInfo(strerror(errno), path);
        return E_FILE_ERR;
    }

    if (fputs(versionHeader, db) == EOF) {
        errorSaveRichErrorInfo(strerror(errno), path);
        return E_FILE_ERR;
    }

    fclose(db);
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * AddressBook::editCut
 *=------------------------------------------------------------------------=
 */
void AddressBook::editCut()
{

}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::editCopy
 *=------------------------------------------------------------------------=
 */
void AddressBook::editCopy()
{

}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::editPaste
 *=------------------------------------------------------------------------=
 */
void AddressBook::editPaste()
{

}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::editSelectAll
 *=------------------------------------------------------------------------=
 */
void AddressBook::editSelectAll()
{

}


/**
 *=------------------------------------------------------------------------=
 * AddressBook::fileClose
 *=------------------------------------------------------------------------=
 */
void AddressBook::fileClose()
{

}


/**
 *=------------------------------------------------------------------------=
 * AddressBook::fileDelete
 *=------------------------------------------------------------------------=
 */
void AddressBook::fileDelete()
{

}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::fileNew
 *=------------------------------------------------------------------------=
 */
void AddressBook::fileNew()
{
    AddressBookEntry abe(this, "entry", true, 0);
    AddressBookItem *abi;
    ERRCODE er;
    int ret;

    /**
     * Run the dialog, get the item, and then return.
     */
    ret = abe.exec();
    if (ret == QDialog::Accepted) {

        abe.getItem(abi);
        if (!abi) return;

        /**
         * Now add and save the entry !!
         */
        addUIEntry(abi);
        er = writeEntry(abi);
        if (FAILED(er)) {
            errorShowError(er);
            return;
        }
    }
}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::filePrint
 *=------------------------------------------------------------------------=
 */
void AddressBook::filePrint()
{

}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::fileProperties
 *=------------------------------------------------------------------------=
 */
void AddressBook::fileProperties()
{

}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::findAddress
 *=------------------------------------------------------------------------=
 * tries to find the given email address in our list of items.  Returns
 * S_OK if we found it and returns the appropriate item.  Otherwise returns
 * S_FALSE or an error.
 *
 * Parameters:
 *      const char *                    - [in]  find me -- email address
 *      AddressBookItem * &             - [out] found item, or NULL.
 *
 * Output:
 *      ERRCODE                         - S_OK, S_FALSE, or ERROR
 */
ERRCODE AddressBook::findAddress
(
    const char * findMe,
    AddressBookItem *&found
)
{
    AddressBookItem *item;
    QListViewItem *i;    

    found = NULL;

    /**
     * whiz through the list looking for this item.
     */
    i = this->listView->firstChild();
    while (i != NULL) {

        item = ((ABListViewItem *)i)->item;

        /** 
         * IF the addresses match, then by jove, we've found it!
         */
        if (!strcasecmp(findMe, item->emailAddress)) {
            found = item;
            return S_OK;
        }

        i = i->nextSibling();
    }

    /**
     * Didn't find it.
     */
    return S_FALSE;
}


/**
 *=------------------------------------------------------------------------=
 * AddressBook::getAddressListRef
 *=------------------------------------------------------------------------=
 * returns the list of address information -- should be used right away
 * as it is only a reference.
 *
 * Parameters:
 *      LinkList<AddressBookItem *>     - [in]  where to put info.
 */
void AddressBook::getAddressListRef
(
    LinkList<AddressBookItem *> &list
)
{
    list = this->addresses;
}


/**
 *=------------------------------------------------------------------------=
 * AddressBook::helpHelp
 *=------------------------------------------------------------------------=
 */
void AddressBook::helpHelp()
{

}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::helpAbout
 *=------------------------------------------------------------------------=
 */
void AddressBook::helpAbout()
{
    AboutDialog ad(this, "ad", true, 0);
    ad.exec();
}



/**
 *=------------------------------------------------------------------------=
 * AddressBook::initialize
 *=------------------------------------------------------------------------=
 * initializes our window and sets up some basic goo.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AddressBook::initialize()
{
    ERRCODE er;

    /**
     * Initialize our toolbars and menus
     */
    er = this->initializeUI(this, addressBookMenu, addressBookToolBarData);
    RETURN_ON_FAILURE(er);

    this->setCaption(intlLoadStringX(ID_ADDRESSBOOK));

    /**
     * Okay, now we need to go and start creating our dudes.
     */
    searchLabel = new QLabel(this);
    CK(searchLabel);
    searchLabel->setText(intlLoadStringX(ID_AB_SEARCH_FOR));

    searchEdit = new QLineEdit(this);
    CK(searchEdit);

    /**
     * set up the list view.  A little more complicated.
     */
    listView = new QListView(this);
    CK(listView);

    listView->addColumn(intlLoadStringX(ID_AB_FIELD_NAME), prefAddressBookNameFieldWidth);
    listView->addColumn(intlLoadStringX(ID_AB_FIELD_EMAIL), prefAddressBookEmailFieldWidth);
    listView->addColumn(intlLoadStringX(ID_AB_FIELD_ADDR), prefAddressBookAddrFieldWidth);
    listView->addColumn(intlLoadStringX(ID_AB_FIELD_PHONE), prefAddressBookPhoneFieldWidth);

    listView->setSorting(0, true);

    /**
     * Finally, load in the database.
     */
    return loadAddressDB();
}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::layoutComponents
 *=------------------------------------------------------------------------=
 * positions everything properly within this window.  Not all that hard
 * actually ...
 */
void AddressBook::layoutComponents()
{
    int ctlHeight;
    QRect rect;
    int top;
    int i;

    /**
     * Compute some size information
     */
    ctlHeight = searchLabel->font().pixelSize() + 2 * CELL_PAD;  // 2 x 4 for over/under-hang
    getClientRect(rect);
    top = rect.top() + 2 *CELL_PAD;

    /**
     * Now go and position the controls.
     */
    i = searchLabel->sizeHint().width();
    searchLabel->setGeometry(LEFT_MARGIN, top, i, ctlHeight);
    searchEdit->setGeometry(i + 3 * CELL_PAD, top - 2, 250, ctlHeight +  CELL_PAD);
    top += ctlHeight + CELL_PAD;

    listView->setGeometry(LEFT_MARGIN, top, rect.width() - 2 * LEFT_MARGIN, rect.height() - (top - rect.top() + CELL_PAD));
}


/**
 *=------------------------------------------------------------------------=
 * AddressBook::loadAddressDB
 *=------------------------------------------------------------------------=
 * loads in the Address Database.  The location is a subsystem preference.
 *
 * Output:
 *    ERRCODE
 */
ERRCODE AddressBook::loadAddressDB()
{
    AddressBookItem *newItem;
    char readBuffer[MAX_ADDRLINE_LEN];
    char path[MAXPATHLEN];
    FILE *addressbook;
    ERRCODE er;

    /**
     * Get the path to the deeley-bop.
     */
    if (prefAddressBookPath) {
        strncpy(path, prefAddressBookPath, sizeof(path));
        path[sizeof(path) - 1] = '\0';
    } else {
        Kiltdown::getKiltdownDirectory(path, sizeof(path));
        strcat(path, defaultAddressBookName);
    }

    /**
     * Open it.
     */
    addressbook = fopen(path, "r");
    if (!addressbook && errno != ENOENT) {
        errorSaveRichErrorInfo(strerror(errno), path);
        return E_FILE_ERR;
    } else if (!addressbook) {
        
        return createBlankDB(path);
    }

    /**
     * First line should be version.
     */
    if (!fgets(readBuffer, sizeof(readBuffer), addressbook)) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_FILE_ERR; goto Error;
    }

    /**
     * Make sure it starts with a version header we like.
     */
    if (strncasecmp(readBuffer, versionInfo, sizeof(versionInfo) - 1)) {
        errorSaveRichErrorInfo(path);
        er = E_ADDR_VERSION; goto Error;
    }

    /**
     * Now start reading in the lines, one at a time, and parsing out the 
     * info.
     */
    while (!feof(addressbook)) {
        /**
         * Read the line.
         */
        readBuffer[0] = '\0';
        if (!fgets(readBuffer, sizeof(readBuffer), addressbook)
            && !feof(addressbook)) {
            errorSaveRichErrorInfo(strerror(errno));
            er = E_FILE_ERR; goto Error;
        }

        /** 
         * Parse it.
         */
        if (readBuffer[0] != '\0') {
            er = parseEntry(readBuffer, newItem);
            GOTO_ON_FAILURE(er, Error);
            addUIEntry(newItem);
        }
    }

    er = S_OK;

  Error:
    fclose(addressbook);
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::nukeAddressBookItem
 *=------------------------------------------------------------------------=
 * clears out and frees an addressbookitem.
 *
 * Parameters:
 *      AddressBookItem *           - [in]  clear and free me.
 */
void AddressBook::nukeAddressBookItem
(
    AddressBookItem *item
)
{
    if (item->firstName) localFree(item->firstName);
    if (item->lastName) localFree(item->lastName);
    if (item->emailAddress) localFree(item->emailAddress);

    localFree(item);
}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::parseEntry
 *=------------------------------------------------------------------------=
 * given a line from the input file, go and parse in the entry
 *
 * Parameters:
 *      const char *                - [in]  line from file.
 *      AddressBookItem *&          - [out] new localAlloc'd entry.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AddressBook::parseEntry
(
    const char *line,
    AddressBookItem * &newEntry
)
{
    char entry[MAX_ENTRY_SIZE];
    int item = 0;
    int y = 0;
    int x = 0;

    newEntry = (AddressBookItem *)localAlloc(sizeof(AddressBookItem));
    if (!newEntry) return E_OUTOFMEMORY;
    memset(newEntry, 0, sizeof(AddressBookItem));

    /**
     * Whiz through, looking for commas separating things.  If a comma has
     * a \ before it, we ignore it.
     */
    while (isspace(line[x])) x++;
    while (1) {

        if (line[x] == '\0' || (line[x] == ',' && x && line[x - 1] != '\\')) {
            if (y) {
                entry[y] = '\0';
                switch (item) {
                    case 0:
                        if (y) newEntry->firstName = copyString(entry);
                        break;
                    case 1:
                        if (y) newEntry->lastName = copyString(entry);
                        break;
                    case 2:
                        if (y) newEntry->emailAddress = copyString(entry);
                        break;
                }
            }

            item++;
            y = 0;
            x++;
            while (isspace(line[x])) x++;
            if (line[x] == '\0')
                break;
            else
                continue;
        }

        if (line[x] != '\r' && line[x] != '\n')
            entry[y++] = line[x++];
        else
            x++;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * AddressBook::showAddressEntry
 *=------------------------------------------------------------------------=
 * given a full SMTP spec email address, show a dialog with properties for
 * it.  If it's an address in our book, go and show that one, of course.
 *
 * Parameters:
 *      const char *                    - [in]  smtp spec address.
 *      QWidget *                       - [in]  parent widget.
 */
void AddressBook::showAddressEntry
(
    const char *address,
    QWidget *parent
)
{
    AddressBookEntry abe(address, parent, "abe", true, 0);

    // UNDONE: marcw, 2001-4-18 -- look for the address in the db

    abe.exec();
}
 

/**
 *=------------------------------------------------------------------------=
 * AddressBook::writeEntry
 *=------------------------------------------------------------------------=
 * writes the given address to our file.
 *
 * Parameters:
 *      AddressBookItem *           - [in]  item to write out.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AddressBook::writeEntry
(
    AddressBookItem *item
)
{
    char buf[MAX_ADDRLINE_LEN];
    char path[MAXPATHLEN];
    FILE *db;

    /**
     * Get the path to the deeley-bop.
     */
    if (prefAddressBookPath) {
        strncpy(path, prefAddressBookPath, sizeof(path));
        path[sizeof(path) - 1] = '\0';
    } else {
        Kiltdown::getKiltdownDirectory(path, sizeof(path));
        strcat(path, defaultAddressBookName);
    }

    /**
     * Open the DB file for appends!
     */
    db = fopen(path, "a");
    if (!db) {
        errorSaveRichErrorInfo(strerror(errno), path);
        return E_FILE_ERR;
    }

    /** 
     * Generate the string.
     */
    snprintf(buf, sizeof(buf), "%s,%s,%s\n", 
             item->firstName ? item->firstName : "",
             item->lastName ? item->lastName : "",
             item->emailAddress ? item->emailAddress : "");

    /**
     * Write it out
     */
    if (fputs(buf, db) == EOF) {
        errorSaveRichErrorInfo(strerror(errno), path);
        fclose(db);
        return E_FILE_ERR;
    }

    /**
     * Finito!
     */
    fclose(db);
    return S_OK;
}




