/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <qpixmap.h>
#include <sys/stat.h>
#include <unistd.h>

#include "attachment.h"
#include "filemapping.h"



/** 
 * This class defines an attachment to a message or article.  It can encode
 * itself or decode itself from base64 and uuencode formats.  Sometimes it
 * will store itself as binary, others text, in the interest of efficiency.
 */

/**
 *=------------------------------------------------------------------------=
 * Attachment::Attachment
 *=------------------------------------------------------------------------=
 * Creates a new attachment object from the given file.  we'll guess what
 * the MIMETYPE is.
 *
 * Parameters:
 *      const char *            - [in]  the name (usually filename) of the
 *                                  attachment.
 */
Attachment::Attachment
(
    const char *fileName
)
{
    this->fullPath = copyString(fileName);
    m_cref = 1;

    data = NULL;
    size = -1;

    chooseMimeTypeAndEncoding(fileName, this->type, this->encoding);

    DEBUG_TEST_START;
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::Attachment
 *=------------------------------------------------------------------------=
 * Creates a new attachment object with the given values.
 *
 * Parameters:
 *      const char *            - [in]  the name (usually filename) of the
 *                                  attachment.
 *      BYTE *                  - [in]  the data for the attachment.  We 
 *                                  assume ownership of this pointer.
 *      int                     - [in]  size of the data.
 *      MimeEncoding            - [in]  what type of encoding to use.
 *      MIMETYPE                - [in]  what type of attachment this is.
 */
Attachment::Attachment
(
    const char *fileName,
    BYTE *data,
    int size,
    MimeEncoding enc,
    MIMETYPE type
)
{
    this->fullPath = copyString(fileName);
    this->encoding = enc;
    this->type = type;
    this->data = data;
    this->size = size;

    m_cref = 1;

    DEBUG_TEST_START;
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::~Attachment
 *=------------------------------------------------------------------------=
 * Time to take out the trash!
 */
Attachment::~Attachment()
{
    if (fullPath) localFree(fullPath);
    if (data) localFree(data);


    DEBUG_TEST_STOP;
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::chooseIcon
 *=------------------------------------------------------------------------=
 * Returns an icon to display for this attachment.
 *
 * Output:
 *      QPixmap *           - you must 'delete' me.  NULL ==> E_OUTOFMEMORY
 */
QPixmap *Attachment::chooseIcon()
{
    return chooseIconForFile(this->fullPath);
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::getBinaryDataRef
 *=------------------------------------------------------------------------=
 * returns a pointer (a reference pointer, NOT a copy) to teh binary data
 * for this attachment.
 *
 * Parameters:
 *      BYTE **             - [out] pointer to a reference to the data. DO NOT
 *                              EVEN THINK OF FREEING ME!!!!
 *
 * Output:
 *      ERRCODE
 */
ERRCODE Attachment::getBinaryDataRef
(
    BYTE **pdata
)
{
    ERRCODE er;
    int i, res;
    FILE *f = NULL;

    if (!pdata) return E_PARAM;
    
    /**
     * if we've already got it, this is ez.
     */
    if (this->data) {
        *pdata = data;
        return S_OK;
    }

    /**
     * otherwise, we have to load it in from disk.
     */
    K_ASSERT(fullPath, "I need a filename to load this dude in !!!");
    er = getBinarySize(i);
    RETURN_ON_FAILURE(er);
    this->size = i;

    this->data = (BYTE *)localAlloc(i);
    if (!this->data) return E_OUTOFMEMORY;
    
    f = fopen(fullPath, "r");
    if (!f) {
        K_FAIL("Couldn't open attachment for reading??");
        er = E_FILE_ERR;
        goto Error;
    }

    res = fread(this->data, sizeof(char), i, f);
    if (!res) {
        if (ferror(f)) {
            er = E_FILE_ERR;
            goto Error;
        }
    }

    *pdata = this->data;
    er = S_OK;

  Error:
    if (f) fclose(f);
    if (FAILED(er)) {
        if (this->data) localFree(this->data);
        this->data = NULL;
        *pdata = NULL;
    }
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::getBinarySize
 *=------------------------------------------------------------------------=
 * returns the size of this object, in decoded form (i.e. this doesn't return
 * the size of the encoded form of this object.
 * 
 * Parameters:
 *      int &                       - [out] the size in bytes
 *
 * Output:
 *      ERRCODE
 */
ERRCODE Attachment::getBinarySize
(
    int &psize
)
{
    struct stat info;

    /**
     * If we already have the attachment loaded, this is EZ.
     */
    if (size != -1) {
        psize = size;
        return S_OK;
    }

    /**
     * Grab the info from the filesystem.
     */
    if (lstat(this->fullPath, &info)) return E_FILE_ERR;

    psize = info.st_size;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * Attachment::getEncoding
 *=------------------------------------------------------------------------=
 * returns the encoding that is used for this attachment.
 *
 * Parameters:
 *      MimeEncoding &            - [out] our encoding.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE Attachment::getEncoding
(
    MimeEncoding &enc
)
{
    enc = this->encoding;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * Attachment::getFileName
 *=------------------------------------------------------------------------=
 * returns the base filename for this attachment.
 *
 * Parameters:
 *      const char **               - [out] the base file name.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE Attachment::getFileName
(
    const char **val
)
{
    QString str(this->fullPath);
    int idx = str.findRev('/');
    str = str.mid(idx + 1);


/**
 * UNDONE: marcw, 2001-5-7: this is definitely a huge bug.  figure out
 * what's going on here !!!
 */
K_ASSERT((const char *)str != NULL, "Attachment has no name !!! AUGH!!!\n");
if ((const char *)str == NULL) {
    str = "Attachment";
}

    *val = copyString((const char *)str);
    return (*val) ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::getFileNameLen
 *=------------------------------------------------------------------------=
 * returns the length of the attachment's file name.
 *
 * Parameters:
 *      int *               - [out] the length
 *
 * Output:
 *      ERRCODE
 */
ERRCODE Attachment::getFileNameLen
(
    int *len
)
{
    QString str(this->fullPath);
    int idx = str.findRev('/');
    str = str.mid(idx + 1);

    *len = strlen((const char *)str);
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * Attachment::getFullPath
 *=------------------------------------------------------------------------=
 * returns the full available path for this attachment.
 *
 * Parameters:
 *      const char **           - [out] a COPY of our full path
 *
 * Output:
 *      ERRCODE
 */
ERRCODE Attachment::getFullPath
(
    const char **name
)
{
    if (!name) return E_PARAM;

    *name = copyString(this->fullPath);
    return (*name) ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::getMimeType
 *=------------------------------------------------------------------------=
 * returns the MIME type for this attachment.
 *
 * Parameters:
 *      MIMETYPE &              - [out]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE Attachment::getMimeType
(
    MIMETYPE &mimeType
)
{
    mimeType = this->type;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::getPrintableSize
 *=------------------------------------------------------------------------=
 * returns the size of this widget in a printable format.  Very exciting.
 *
 * Output:
 *      char *              - pointer to a buffer holding the size.  Sub-
 *                              sequent calls to this will clobber old vals.
 */
char * Attachment::getPrintableSize()
{
    struct stat info;
    int len;

    /**
     * If we couldn't get the info, then there's not much to do here.
     */
    if (lstat(this->fullPath, &info)) return "";

    sizeBuf[0] = '(';
    sizeBuf[1] = '\0';
    printByteCount(sizeBuf + 1, sizeof(sizeBuf) - 2, info.st_size);
    len = strlen(sizeBuf);
    sizeBuf[len++] = ')';
    sizeBuf[len] = '\0';
    
    return sizeBuf;
}

/**
 *=------------------------------------------------------------------------=
 * Attachment::saveToFile
 *=------------------------------------------------------------------------=
 * saves this attachment to the given file.  Doesn't matter if the file is
 * different from our 'official' name.
 *
 * Parameters:
 *      const char *                - [in]  where to save me to.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE Attachment::saveToFile
(
    const char *filename
)
{
    FILE *f;
    BYTE *d;
    int s, x;


    if (!filename) return E_PARAM;

    f = fopen(filename, "w");
    if (!f) {
        perror("crap  ");
         return E_FILE_ERR;
    }

    d = data ? data : (BYTE *)"";
    s = data ? size : 1;

    x = fwrite(d, sizeof(BYTE), s, f);
    fclose(f);
    if (!x) {
        if (ferror(f))
            return E_FILE_ERR;
    }

    return S_OK;
    
}


