/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <db.h>
#include <errno.h>

#include <lerror.h>
#include <llist.h>


/**
 * This is a list of all the databases in our system.
 */
LinkList<DB *> databaseList;

/**
 *=------------------------------------------------------------------------=
 * dbmgrCloseDatabase
 *=------------------------------------------------------------------------=
 * we're done with the database -- close it and be done with it !!!
 *
 * Parameters:
 *      DB *                        - [in]  close me !!!
 *
 * Output:
 *      ERRCODE
 */
ERRCODE dbmgrCloseDatabase
(
    DB * database
)
{
    ERRCODE er;

    if (!database) return E_PARAM;

    /**
     * Remove it from our list.
     */
    er = databaseList.removeElement(database);
    if (er != S_OK) {
        K_FAIL("Can't find database in our caching table.  Very bad !!!\n");
        return E_INTERNAL;
    }

    /**
     * Close it.
     */
    database->close(database, 0);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * dbmgrOpenDatabase
 *=------------------------------------------------------------------------=
 * opens a database and adds it to our list of databases.
 *
 * Parameters:
 *      DB * &                      - [out] new db.
 *      const char *                - [in]  path
 *      bool                        - [in]  create if necessary?
 *      int                         - [in]  permissions.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE dbmgrOpenDatabase
(
    DB * & database,
    const char *path,
    bool createIfNecessary,
    int mode
)
{
    int flags, result;
    ERRCODE er;

    /**
     * Now create/open the database.
     */
    result = db_create(&database, NULL, 0);
    if (result != 0) { goto Failure; }

    flags = createIfNecessary ? DB_CREATE : 0;
    result = database->open(database, path, NULL, DB_BTREE, flags, mode);
    if (result != 0) { goto Failure; }

    /** 
     * add it to our list of databases.
     */
    er = databaseList.addElement(database);
    if (FAILED(er)) { result = ENOMEM; goto Failure; }
    return S_OK;

  Failure:
    errorSaveRichErrorInfo(db_strerror(result));
    if (database) database->close(database, 0);
    database = NULL;
    return E_DB_ERR;
}

/**
 *=------------------------------------------------------------------------=
 * dbmgrSyncDatabasesOnDeath
 *=------------------------------------------------------------------------=
 * closes up the all the databases so that no data is lost on crash.
 *
 * Note:
 *      THIS ROUTINE INVALIDATES ALL EXISTING "DB *" POINTERS.  VERY CREEPY.
 */
void dbmgrSyncDatabasesOnDeath()
{
    ERRCODE er;
    DB *db;

    er = databaseList.firstElement(&db);
    while (er == S_OK) {
        db->close(db, 0);
        databaseList.removeElement(db);
        er = databaseList.firstElement(&db);
    }

    // that's it!
}


