/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <sanity.h>

#include <qpixmap.h>

#include "filemapping.h"
#include "uiicons.h"


#define STRINGS_ARE_EQUAL(s1, s2)   (s1[0] == s2[0] && !strcasecmp(s1, s2))

/**
 * This file contains some routines to help set up mappings between files, 
 * icons, and the programs that are used to execute them.  It will attempt
 * to work with KDE/Gnome, or whatever other window systems are available
 * on the given computer.
 */
struct FileMapping {

    const char *name;
    const char **icon;
    MimeMediaType mediaType;
    MimeMediaSubType subType;
    MimeEncoding encoding;
};


FileMapping baseMappings[] = {

    { "jpg", imageIcon, Image, Jpeg, Base64 },
    { "jpeg", imageIcon, Image, Jpeg, Base64 },
    { "jpe", imageIcon, Image, Jpeg, Base64 },
    { "gif", imageIcon, Image, Gif, Base64 },
    { "png", imageIcon, Application, OctetStream, Base64 },
    { "bmp", imageIcon, Application, OctetStream, Base64 },
    { "pcx", imageIcon, Application, OctetStream, Base64 },
    { "xpm", imageIcon, Application, OctetStream, Base64 },
    { "zip", zipIcon, Application, OctetStream, Base64 },
    { "gz", gzipIcon, Application, OctetStream, Base64 },
    { "Z", gzipIcon, Application, OctetStream, Base64 },
    { "tar", tarIcon, Application, OctetStream, Base64 },
    { "exe", exeIcon, Application, OctetStream, Base64 },
    { "txt", textfileIcon, Text, Plain, QuotedPrintable },
    { "c", csourceIcon, Text, Plain, QuotedPrintable },
    { "cc", csourceIcon, Text, Plain, QuotedPrintable },
    { "h", csourceIcon, Text, Plain, QuotedPrintable },
    { "cpp", csourceIcon, Text, Plain, QuotedPrintable },
    { "cxx", csourceIcon, Text, Plain, QuotedPrintable },
    { "hxx", csourceIcon, Text, Plain, QuotedPrintable },
    { "java", jsourceIcon, Text, Plain, QuotedPrintable },
    { "jav", jsourceIcon, Text, Plain, QuotedPrintable },
    { "mpg", movieIcon, Video, Mpeg, Base64 },
    { "mpeg", movieIcon, Video, Mpeg, Base64 },
    { "mpe", movieIcon, Video, Mpeg, Base64 },
    { "html", htmlfileIcon, Text, Html, QuotedPrintable },
    { "htm", htmlfileIcon, Text, Html, QuotedPrintable }
};


/**
 *=------------------------------------------------------------------------=
 * chooseIconForFile
 *=------------------------------------------------------------------------=
 * given a filename, figure out what icon to associate with it.
 *
 * Parameters:
 *      const char *                - [in]  filename.
 *
 * Output:
 *      QPixmap *                   - NULL implies E_OUTOFMEMORY;
 */
QPixmap *chooseIconForFile
(
    const char *fileName
)
{
    QString file(fileName);
    const char *pstr;
    int idx;

    /**
     * Look for a . to see if there's an extension on the file.
     */
    idx = file.findRev('.');

    /**
     * If no extension, then just give them a default icon.
     */
    if (idx < 0) return new QPixmap(genericfileIcon);

    /**
     * Otherwise, start looking for the various extensions we know about.
     */
    file = file.mid(idx + 1);
    pstr = (const char *)file;

    for (unsigned int x = 0; x < sizeof(baseMappings) / sizeof(FileMapping); x++) {

        if (STRINGS_ARE_EQUAL(pstr, baseMappings[x].name))
            return new QPixmap(baseMappings[x].icon);
    }

    return new QPixmap(genericfileIcon);
}

/**
 *=------------------------------------------------------------------------=
 * chooseMimeTypeAndEncoding
 *=------------------------------------------------------------------------=
 * given a file path, try and figure out a good MIME type and encoding for
 * the given file.
 * 
 * Parameters:
 *      const char *                    - [in]  path to guess for
 *      MIMETYPE &                      - [out] the suggested type.
 *      MimeEncoding &                  - [out] the suggested encoding.
 */
void chooseMimeTypeAndEncoding
(
    const char *fileName,
    MIMETYPE &type,
    MimeEncoding &encoding
)
{
    QString file(fileName);
    const char *pstr;
    int idx;

    /**
     * some safe/default values, first
     */
    type.setType(Application, OctetStream);
    encoding = Base64;

    /**
     * Look for a . to see if there's an extension on the file.
     */
    idx = file.findRev('.');

    /**
     * If no extension, then that's all we can do.
     */
    if (idx < 0) return;

    /**
     * Otherwise, start looking for the various extensions we know about.
     */
    file = file.mid(idx + 1);
    pstr = (const char *)file;


    for (unsigned int x = 0; x < sizeof(baseMappings) / sizeof(FileMapping); x++) {

        if (STRINGS_ARE_EQUAL(pstr, baseMappings[x].name)) {
            type.setType(baseMappings[x].mediaType, baseMappings[x].subType);
            encoding = baseMappings[x].encoding;
            return;
        }

    }

    // just return with the default values we've got.
}

/**
 *=------------------------------------------------------------------------=
 * getCommandLineForAttachment
 *=------------------------------------------------------------------------=
 * given a "file" with a (hopefully) specified MIME type, try and figure out
 * what command line to use for it, or bring up a dialog asking the user what
 * program to associate with it.
 *
 * Please note that the output string will be of the form:
 *
 *          <<programname>> <<arguments>> %s <<possibly-something-else>>
 * i.e.
 *          /usr/local/bin/graphicsView %s
 *
 * The output is designed to used as a basic string for snprintf or some
 * such function.
 *
 * Parameters:
 *      MIMETYPE &                  - [in]  if we have it, the type
 *      const char *                - [in]  the name of the file.
 *      const char * &              - [out] the command line to run.
 *
 * Output:
 *      ERRCODE
 */
#if 0
ERRCODE getCommandLineForAttachment
(
    MIMETYPE &
    const char *filename,
    const char * &commandLine
)
{
    return E_NOTIMPL;
}
#endif // 0
