/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "fileutil.h"
#include <lerror.h>

/**
 *=------------------------------------------------------------------------=
 *                          local prototypes
 *=------------------------------------------------------------------------=
 */
ERRCODE removeFile(const char *);
ERRCODE removeDirectory(const char *);


/**
 *=------------------------------------------------------------------------=
 * fileCopyFile
 *=------------------------------------------------------------------------=
 * copies the given file to the given file or directory.
 *
 * Parameters:
 *      const char *                - [in]  source path.
 *      const char *                - [in]  dest path.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE fileCopyFile
(
    const char *sourcePath,
    const char *destPath
)
{
    FILE *source = NULL, *dest = NULL;
    char buf[4096];
    size_t result;
    ERRCODE er;
    size_t len;

    /**
     * Open the source for read only.
     */
    source = fopen(sourcePath, "r");
    if (!source) {
        errorSaveRichErrorInfo(sourcePath, strerror(errno));
        er = E_FILE_ERR; goto Error;
    }

    /** 
     * Open the destination for writing.
     */
    dest = fopen(destPath, "w+");
    if (!dest) {
        errorSaveRichErrorInfo(destPath, strerror(errno));
        er = E_FILE_ERR; goto Error;
    }

    /**
     * Now do some readin' and writing'.
     */
    while (!feof(source)) {

        result = fread(buf, sizeof(char), sizeof(buf) / sizeof(char), source);
        if ( result != sizeof(buf)) {
            if (ferror(source)) {
                errorSaveRichErrorInfo(sourcePath, strerror(errno));
                er = E_FILE_ERR; goto Error;
            }
        }
        len = fwrite(buf, sizeof(char), result, dest);
        if (len != result) {
            if (ferror(dest)) {
                errorSaveRichErrorInfo(destPath, strerror(errno));
                er = E_FILE_ERR; goto Error;
            }
        }

    }

    er = S_OK;

  Error:
    if (source) fclose(source);
    if (dest) fclose(dest);
    if (FAILED(er)) unlink(destPath);

    return er;
}


/**
 *=------------------------------------------------------------------------=
 * fileRmfr
 *=------------------------------------------------------------------------=
 * remove a file or directories and all the crap underneath it.  There is no
 * "kill this damn thing" API, unforunately.
 *
 * Parameters:
 *      const char *                - [in]  DIE DIE DIE !!!
 *
 * Output:
 *      ERRCODE
 */
ERRCODE fileRmfr
(
    const char *removeMe
)
{
    struct stat info;
    int res;

    /** 
     * Get some info about this file.
     */
    res = stat(removeMe, &info);
    if (res < 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    if (S_ISDIR(info.st_mode)) 
        return removeDirectory(removeMe);
    else
        return removeFile(removeMe);
}


/**
 *=------------------------------------------------------------------------=
 * removeDirectory
 *=------------------------------------------------------------------------=
 * given a directory, go and iterate over everything in it, recursively 
 * yanking them.
 *
 * Parameters:
 *      const char *                - [in]  directory to destroy.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE removeDirectory
(
    const char *directory
)
{
    char buf[MAXPATHLEN];
    DIR *directoryCursor;
    struct dirent *entry;
    struct stat info;
    ERRCODE er;
    int len;
    int res;

    /**
     * Open a cursor on the directory.
     */
    directoryCursor = opendir(directory);
    if (!directory) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    strncpy(buf, directory, sizeof(buf));
    buf[sizeof(buf)] = '\0';
    len = strlen(buf);
    buf[len++] = '/';
    buf[len] = '\0';

    /**
     * Iterate over the directory and check out what we've got!!!
     */
    while ((entry = readdir(directoryCursor)) != NULL) {

        /**
         * Skip over "." and ".."
         */
        if (!strncasecmp(entry->d_name, ".", 1) 
            || !strncasecmp(entry->d_name, "..", 2))
            continue;

        /**
         * Append to our path
         */
        strncat(buf, entry->d_name, sizeof(buf)); 
        buf[sizeof(buf)] = '\0';

        /**
         * Figure out what type of deeley-bop it is.
         */
        res = stat(buf, &info);
        if (res < 0) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_FILE_ERR;
        }

        /**
         * Is it a file??
         */
        if (S_ISREG(info.st_mode) || S_ISLNK(info.st_mode)) {
            er = removeFile(buf);
        } else if (S_ISDIR(info.st_mode)) {
            er = removeDirectory(buf);
        } else {
            er = S_OK;
        }
        if (FAILED(er)) {
            closedir(directoryCursor);
            return er;
        }

        /**
         * Tie off the buffer again.
         */
        buf[len] = '\0';
    }

    /**
     * We're done nuking everytning.  close our cursor.
     */
    closedir(directoryCursor);

    /**
     * Now destroy this directory.
     */
    res = rmdir(directory);
    if (res < 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    /**
     * That's all folks!!
     */
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * removeFile
 *=------------------------------------------------------------------------=
 * nukes a file.  Fweeee!
 *
 * Parameters:
 *      const char *                - [in]  nuke me.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE removeFile
(
    const char *removeMe
)
{
    int res;

    /**
     * Unlink the file and be done with it
     */
    res = unlink(removeMe);
    if (res < 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    return S_OK;
}



/**
 *=------------------------------------------------------------------------=
 *                              FileReader
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * FileReader::FileReader
 *=------------------------------------------------------------------------=
 * sets up an instance of our reader.
 *
 * Parameters:
 *      FILE *                  - [in]  file handle.
 *      const char *            - [in]  full path to the file.
 *      int                     - [in]  max length of a line.
 */
FileReader::FileReader
(
    FILE *file,
    const char *fullPath,
    int maxLineLen
)
{
    K_ASSERT(file && fullPath, "Bogus FileReader parameters");
    this->file = file;
    this->path = copyString(fullPath);
    this->readBuf = NULL;
    this->maxLineLen = maxLineLen;
}

/**
 *=------------------------------------------------------------------------=
 * FileReader::~FileReader
 *=------------------------------------------------------------------------=
 * wake up ... time to die.
 */
FileReader::~FileReader()
{
    if (this->file) fclose(this->file);
    if (this->path) localFree(this->path);
    if (this->readBuf) localFree(this->readBuf);
}


/**
 *=------------------------------------------------------------------------=
 * FileReader::getReadBuf
 *=------------------------------------------------------------------------=
 * returns the read buffer that should be used for input.
 *
 * Output:
 *      char *                  - the read buffer that should be used or
 *                                NULL in OOM situations.
 */
char *FileReader::getReadBuf()
{
    if (!this->readBuf)
        this->readBuf = (char *)localAlloc(maxLineLen);
    return this->readBuf;
}


/**
 *=------------------------------------------------------------------------=
 * FileReader::readLine                                           [IReader]
 *=------------------------------------------------------------------------=
 * reads in a line of text, making sure it ends with \r\n
 *
 * Parameters:
 *      void *                  - [in]  where to put resulting line.
 *      unsigned int            - [in]  size of buf
 *      int &                   - [out] actual number of chars read.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FileReader::readLine
(
    void *vbuf,
    unsigned int sizeofBuf,
    int &cread
)
{
    char *res, *read, *buf;
    int len;

    buf = (char *)vbuf;

    /**
     * Get the buffer into which we'll read a line.
     */
    read = getReadBuf();
    if (!read) return E_OUTOFMEMORY;

    /**
     * Read in the line
     */
    res = fgets(read, maxLineLen, this->file);
    if (!res) {
        if (feof(this->file)) {
            buf[0] = '\0';
            return S_OK;
        } else if (ferror(this->file)) {
            buf[0] = '\0';
            errorSaveRichErrorInfo(this->path, strerror(errno));
            return E_FILE_ERR;
        }
    }

    /**
     * Copy it into the user's buf.
     */
    len = strlen(read);
    strncpy(buf, read, sizeofBuf);
    if (len > 1 && (buf[len - 1] != '\n'  || buf[len - 2] != '\r')) {
        if (buf[len - 1] == '\n' || buf[len - 1] == '\r') {
            buf[len - 1] = '\r';
            buf[len] = '\n';
            buf[len++] = '\0';
        } else {
            buf[len++] = '\r';
            buf[len++] = '\n';
            buf[len++] = '\0';
        }
    }
    
    cread = len;
    return S_OK;
}


