/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
 
#include <sanity.h>

#include <errno.h>
#include <string.h>
#include <sys/param.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

#include "attachment.h"
#include "dbmgr.h"
#include "folderdb.h"
#include "mailfolder.h"
#include "mailmessage.h"
#include "main.h"

/**
 * This folder class provides the implementation of a mail folder object. 
 * These are where we store the messages for a folder in an given account.
 * Each folder is made up of two files:  one will contain the header for
 * the messages, and the other will contain the data.  The folders database
 * files are keyed based on a time stamp value.  All message objects will
 * have one of these keys to help us identify the object.
 */
 
/**
 * NOTE:  This class is a 'friend' class of the MailMessage class.  I've never
 * actually used this feature before, but it seems to be a good idea to give
 * this class access to the inner workings of the MailMessage class to save it
 * the hassle of having to set everything through accessor methods.
 */


/**
 * This is the LATEST version of the header and body message format.  Older 
 * versions are taken into consideration in our persistence code.
 */
const int HEADER_VERSION_STAMP = 2;
const int BODY_VERSION_STAMP = 4;

const int MESSAGE_HEADER_CHUNK = 8192;
const int MAX_HEADER_ENTRY_LEN = 8192;


/**
 *=------------------------------------------------------------------------=
 * These are the various extensions and file names we use for the dbs.
 */
static const char *headerSuffix = "_h.db";
static const char *bodiesSuffix = "_b.db";

static const char *compactHeaders = "__headers";
static const char *compactBodies  = "__bodies";
static const char *renamedOldExt  = ".dying";


/**
 *=------------------------------------------------------------------------=
 * These are the old MIME types that we used to understand/support.
 * NOTE: THIS ENUMERATION IS **OBSOLETE**, AND IS KEPT HERE STRICTLY TO
 * SUPPORT LOADING OLD MESSAGES.  NEW CODE SHOULD USE MimeMediaType AND 
 * MimeSubType.
 *=------------------------------------------------------------------------=
 * DO NOT ADD TO ME!!!
 */
enum OldMimeTypeEnum {

    UnknownType = -1,
    TextPlain = 0,
    MultipartMixed,
    MultipartAlternative,
    TextHtml,
    ApplicationOctetStream,
    ImageJpeg,
    ImageGif,
    VideoMpeg,
    MSTnef,
    MultipartRelated,
    ApplicationXMSDownload,
    MessageRFC822,
    VideoXMSAsf
};

struct OldMimeTypeMap {
    OldMimeTypeEnum oldType;
    MimeMediaType mediaType;
    MimeMediaSubType subType;

} oldMimeTypeMap [] = { 

    { TextPlain, Text, Plain },
    { MultipartMixed, Multipart, Mixed },
    { MultipartAlternative, Multipart, Alternative },
    { TextHtml, Text, Html },
    { ApplicationOctetStream, Application, OctetStream },
    { ImageJpeg, Image, Jpeg },
    { ImageGif, Image, Gif },
    { VideoMpeg, Video, Mpeg },
    { MSTnef, Application, MsTnef },
    { MultipartRelated, Multipart, Related },
    { ApplicationXMSDownload, Application, XMSDownload },
    { MessageRFC822, Message, RFC822 },
    { VideoXMSAsf, Video, XMSAsf },
    { UnknownType, UnknownMediaType, UnknownSubType }
};

/**
 *=------------------------------------------------------------------------=
 * This Table is used to load and save in the properties for a message in the
 * Database.  It contains the type of the property store, as well as what
 * version it first appeared in, and the member function on the MailMessage
 * object that should be called to get or set the properties.  This table uses
 * C++ Member function pointers.  If you do not understand these, PLEASE be 
 * sure to go and learn about them first.  They're pretty cool, but a little
 * tricky.  The best treatment I have seen of them is Section 30 of the C++
 * faq-lite.
 *
 * Note that ANSI C/C++ doesn't allow for initialization of anything other
 * than the first member of a union, and it doesn't appear to work converting
 * member function pointers to void * and back, so i had to do this table in
 * kind of a yucky way.  blech.
 */
enum DBPersistenceType {
    PT_STRING = 0,
    PT_STRINGLIST,
    PT_DATE,
    PT_BOOLEAN,
    PT_INTEGER,
    PT_FINITO
};



struct DBPersistenceTable {

    enum DBPersistenceType type;
    int version;
    void (MailMessage::*getFunction)();
    void (MailMessage::*setFunction)();

} persistenceTable [] = {

    // sender
    { PT_STRINGLIST, 1, {(void (MailMessage::*)())&MailMessage::getSenders}, {(void (MailMessage::*)())&MailMessage::setSenders} },
    // realsender
    { PT_STRING, 1, {(void (MailMessage::*)())&MailMessage::getRealSender}, {(void (MailMessage::*)())&MailMessage::setRealSender} },
    // recipient(s)
    { PT_STRINGLIST, 1, {(void (MailMessage::*)())&MailMessage::getRecipients}, {(void (MailMessage::*)())&MailMessage::setRecipients} },
    // replyto
    { PT_STRING, 2, {(void (MailMessage::*)())&MailMessage::getReplyTo}, {(void (MailMessage::*)())&MailMessage::setReplyTo} },
    // returnpath
    { PT_STRING, 2, {(void (MailMessage::*)())&MailMessage::getReturnPath}, {(void (MailMessage::*)())&MailMessage::setReturnPath} },
    // cc(s)
    { PT_STRINGLIST, 1, {(void (MailMessage::*)())&MailMessage::getCCs}, {(void (MailMessage::*)())&MailMessage::setCCs} },
    // bcc(s)
    { PT_STRINGLIST, 1, {(void (MailMessage::*)())&MailMessage::getBCCs}, {(void (MailMessage::*)())&MailMessage::setBCCs} },
    // subject
    { PT_STRING, 1, {(void (MailMessage::*)())&MailMessage::getSubject}, {(void (MailMessage::*)())&MailMessage::setSubject} },
    // sent
    { PT_DATE, 1, {(void (MailMessage::*)())(ERRCODE (MailMessage::*)(time_t *))&MailMessage::getSent}, {(void (MailMessage::*)())&MailMessage::setSent}  },
    // received
    { PT_DATE, 1, {(void (MailMessage::*)())(ERRCODE (MailMessage::*)(time_t *))&MailMessage::getReceived}, {(void (MailMessage::*)())&MailMessage::setReceived} },
    // has attachment(s)
    { PT_BOOLEAN, 1, {(void (MailMessage::*)())&MailMessage::getHasAttachments}, {(void (MailMessage::*)())&MailMessage::setHasAttachments} },
    // priority
    { PT_INTEGER, 1, {(void (MailMessage::*)())&MailMessage::getPriority}, {(void (MailMessage::*)())&MailMessage::setPriority} },
    // size
    { PT_INTEGER, 1, {(void (MailMessage::*)())&MailMessage::getSize}, {(void (MailMessage::*)())&MailMessage::setSize} },
    // status
    { PT_INTEGER, 1, {(void (MailMessage::*)())&MailMessage::getStatus}, {(void (MailMessage::*)())&MailMessage::setStatus} },
    // extra headers
    { PT_STRINGLIST, 1, {(void (MailMessage::*)())&MailMessage::getExtraHeaders}, {(void (MailMessage::*)())&MailMessage::setExtraHeaders} },

    // that's it!
    { PT_FINITO, -1, NULL, NULL }
};



/**
 *=------------------------------------------------------------------------=
 * FolderDB::FolderDB
 *=------------------------------------------------------------------------=
 * Initalizes a few things.  Should never be called directory, but instead
 * users should be calling openFolder()
 */
FolderDB::FolderDB()
{
    headers = bodies = NULL;
    messagesLoaded = false;
    parent = NULL;

    DEBUG_TEST_START;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::~FolderDB
 *=------------------------------------------------------------------------=
 * Cleans up!!!
 */
FolderDB::~FolderDB()
{
    MailMessage *msg;
    ERRCODE er;

    close();

    if (headerPath != NULL) localFree(headerPath);
    if (bodyPath != NULL) localFree(bodyPath);


    /**
     * Go and clean up all the message objects we have.
     */
    er = messageList.firstElement(&msg);
    while (er == S_OK) {

        messageList.removeElement(msg);
        msg->release();
        er = messageList.firstElement(&msg);
    }

    parent = NULL;

    DEBUG_TEST_STOP;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::addMessage
 *=------------------------------------------------------------------------=
 * Adds the given message to the database.  Basically, we'll extract the 
 * header information and the body, and save out the relevant information to
 * the various databases.
 *
 * Parameters:
 *      MailMessage *           - [in]  the message to save out.
 *
 * Output:
 *      ERRCODE
 *
 * Notes:
 */
ERRCODE FolderDB::addMessage
(
    MailMessage *message
)
{
    struct timeval tv;
    ERRCODE er;
    time_t t;

    if (!message)  return E_PARAM;

    /**
     * Make sure the body for the message is loaded, and that we've got all
     * the other messages in this folder.
     */
    message->ensureBodyLoaded();
    er = loadAllMessages();
    RETURN_ON_FAILURE(er);

    /**
     * get the timestamp for this message.
     */
    if (!message->tv.tv_sec)
        gettimeofday(&tv, NULL);
    else
        tv = message->tv;

    /**
     * Set up the sent time, if it hasn't already been set.
     */
    er = message->getSent(&t);
    RETURN_ON_FAILURE(er);
    if (t == (time_t)-1) {
        t = time(NULL);
        message->setSent(t);
    }

    /**
     * likewise, set up the received time if it hasn't been set.
     */
    t = (time_t)-1;
    er = message->getReceived(&t);
    RETURN_ON_FAILURE(er);
    if (t == (time_t)-1) {
        t = time(NULL);
        message->setReceived(t);
    }

    /**
     * write out the header!!!
     */
    er = writeMessageHeader(message, tv);
    RETURN_ON_FAILURE(er);

    /**
     * then the body!
     */
    er = writeMessageBody(message, tv);
    RETURN_ON_FAILURE(er);

    /**
     * Finally, set up a few things about the message.
     */
    message->hostFolder = this->parent;
    message->tv = tv;
    if (messagesLoaded)  {
        messageList.addElement(message);
        message->addRef();
    }
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::close
 *=------------------------------------------------------------------------=
 * shuts everything down and makes sure that the database are properly 
 * written out.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::close()
{
    /**
     * UNDONE: marcw, 2001-5-12:  error handling here!!!
     */
    if (headers) dbmgrCloseDatabase(headers);
    if (bodies) dbmgrCloseDatabase(bodies);
    headers = bodies = NULL;

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::compact
 *=------------------------------------------------------------------------=
 * compacts this database...
 *
 * We are going to:
 *
 *      i.   create a 'lock' file to prepare for disaster recovery.
 *      ii.  create a copy database in a temporary location.
 *      iii. rename the existing database to a new file name.
 *      iv.  move the temporary copies to be the new dbs.
 *      v.   delete the old renamed dbs.
 *      vi.  remove the lock file.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::compact()
{
    ERRCODE er;

    /**
     * Create the disaster recovery lock.  On startup, we'll look for this
     * and do our best to restore what's going on.
     */
    er = createDisasterLock();
    RETURN_ON_FAILURE(er);

    /**
     * Copy over the database to a new location.
     */
    er = compactCopyDatabases();
    if (!FAILED(er)) {

        /**
         * Close down the databases
         */
        er = close();
        if (!FAILED(er)) {

            /**
             * Rename the current databases to get them out of the way.
             */
            er = compactRenameDatabases();
            if (!FAILED(er)) {

                /**
                 * Make the temp copy databases the main databases now.
                 */
                er = compactReplaceDatabases();
                if (!FAILED(er)) {

                    /**
                     * Finally, delete the old database files 
                     */
                    compactDeleteOldDatabases();
                } else {

                    /**
                     * Restore the renamed dbs
                     */
                    compactRestoreDatabases();
                    compactDeleteCopies();
                }
            } else {

                /**
                 * remove the temp ones and abort.
                 */
                compactDeleteCopies();
            }
        } else {

            /**
             * Delete the copies and abort.
             */
            compactDeleteCopies();
        }
    } else {
        // we'll just return the error code and abort.
    }

    openDatabases();
    removeDisasterLock();

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::compactCopyDatabases
 *=------------------------------------------------------------------------=
 * copies the databases we have here into two temporary databases that
 * will ONLY have the messages that we're adding to it ....
 *
 * Output:
 *      ERRCODE.
 *
 * Notes:
 *      we always generate dbs of the same name since they are supposed to
 *  be SUPER short lived.
 */
ERRCODE FolderDB::compactCopyDatabases()
{
    char path[MAXPATHLEN];
    char *lastSlash;
    DB *headerT;
    ERRCODE er;
    DB *bodyT;


    /**
     * Create the path for the NEW headers and NEW bodies DB.
     */
    strncpy(path, this->headerPath, sizeof(path));
    lastSlash = strrchr(path, '/');
    if (lastSlash != NULL)
        lastSlash++;
    else
        lastSlash = path;

    strcpy(lastSlash, compactHeaders);

    /**
     * Okay, now go and open this database
     */
    er = dbmgrOpenDatabase(headerT, path, true, 0600);
    RETURN_ON_FAILURE(er);

    /**
     * Now create the bodies DB path and try to open it.
     */
    strcpy(lastSlash, compactBodies);
    er = dbmgrOpenDatabase(bodyT, path, true, 0600);
    if (FAILED(er)) {

        strcpy(lastSlash, compactHeaders);
        unlink(path);
        return er;
    }

    /**
     * Create, the new databases are created.  Go and copy over all the data
     */
    er = copyDatabase(this->headers, headerT);
    if (!FAILED(er)) {
        er = copyDatabase(this->bodies, bodyT);
    }

    /**
     * We're done, regardless of success, so close them out!
     */
    dbmgrCloseDatabase(headerT);
    dbmgrCloseDatabase(bodyT);

    /**
     * If we failed, then remove the dbs, since we're going to abort.
     */
    if (FAILED(er)) {
        strcpy(lastSlash, compactHeaders);
        unlink(path);
        strcpy(lastSlash, compactBodies);
        unlink(path);
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::compactDeleteCopies
 *=------------------------------------------------------------------------=
 * Deletes the copied databases from the system -- this happens when there
 * is an error and we need to recover.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::compactDeleteCopies()
{
    char path[MAXPATHLEN];
    char *lastSlash;

    /**
     * Create the path for the NEW headers and NEW bodies DB.
     */
    strncpy(path, this->headerPath, sizeof(path));
    lastSlash = strrchr(path, '/');
    if (lastSlash != NULL)
        lastSlash++;
    else
        lastSlash = path;

    /**
     * First nuke the __headers file
     */
    strcpy(lastSlash, compactHeaders);
    unlink(path);

    /**
     * Now the __bodies file.
     */
    strcpy(lastSlash, compactBodies);
    unlink(path);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::compactDeleteOldDatabases
 *=------------------------------------------------------------------------=
 * The databases have been successfully copied and compacted -- this goes 
 * and clears out the old databases that are no longer needed.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::compactDeleteOldDatabases ()
{
    char path[MAXPATHLEN];
    int len;

    /**
     * First, we'll do the headers.
     */
    len = strlen(this->headerPath);
    strncpy(path, this->headerPath, sizeof(path));
    strncpy(path + len, renamedOldExt, sizeof(path) - len - 1);
    unlink(path);

    /**
     * Now we'll do the bodies.
     */
    len = strlen(this->bodyPath);
    strncpy(path, this->bodyPath, sizeof(path));
    strncpy(path + len, renamedOldExt, sizeof(path) - len - 1);
    unlink(path);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::compactRenameDatabases
 *=------------------------------------------------------------------------=
 * renames our current live dbs to be backup ones so we can slide the
 * compacted ones in and have them be the real dbs.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::compactRenameDatabases()
{
    char path[MAXPATHLEN];
    int len;
    int x;

    /**
     * First, we'll do the headers.
     */
    len = strlen(this->headerPath);
    strncpy(path, this->headerPath, sizeof(path));
    strncpy(path + len, renamedOldExt, sizeof(path) - len - 1);

    /**
     * Rename it.
     */
    x = rename(this->headerPath, path);
    if (x != 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    /**
     * Now we'll do the bodies.
     */
    len = strlen(this->bodyPath);
    strncpy(path, this->bodyPath, sizeof(path));
    strncpy(path + len, renamedOldExt, sizeof(path) - len - 1);

    /**
     * Rename it.
     */
    x = rename(this->bodyPath, path);
    if (x != 0) {

        /**
         * Restore the headers too !!
         */
        len = strlen(this->headerPath);
        strncpy(path, this->headerPath, sizeof(path));
        strncpy(path + len, renamedOldExt, sizeof(path) - len - 1);
        rename(path, this->headerPath);

        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::compactReplaceDatabases
 *=------------------------------------------------------------------------=
 * renames our temporary compacted dbs to be the new main dbs.  This means
 * the dbs are fully compacted, and we only have to do cleanup from here.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::compactReplaceDatabases()
{
    char path[MAXPATHLEN];
    char *lastSlash;
    int x;

    /**
     * Create the path for the NEW headers and NEW bodies DB.
     */
    strncpy(path, this->headerPath, sizeof(path));
    lastSlash = strrchr(path, '/');
    if (lastSlash != NULL)
        lastSlash++;
    else
        lastSlash = path;

    strcpy(lastSlash, compactHeaders);

    /**
     * Okay, now go and open this database
     */
    x = rename(path, this->headerPath);
    if (x != 0) {

        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    /**
     * Now create the bodies DB path and try to open it.
     */
    strcpy(lastSlash, compactBodies);
    x = rename(path, this->bodyPath);
    if (x != 0) {
        errorSaveRichErrorInfo(strerror(errno));

        strcpy(lastSlash, compactHeaders);
        rename(this->headerPath, path);

        return E_FILE_ERR;

    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::compactRestoreDatabases
 *=------------------------------------------------------------------------=
 * restores the renamed old database files back to their original 
 * function.  This is for error conditions only.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::compactRestoreDatabases()
{
    char path[MAXPATHLEN];
    int len;
    int x;

    /**
     * First, we'll do the headers.
     */
    len = strlen(this->headerPath);
    strncpy(path, this->headerPath, sizeof(path));
    strncpy(path + len, renamedOldExt, sizeof(path) - len - 1);

    /**
     * Rename it.
     */
    x = rename(path, this->headerPath);
    if (x != 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    /**
     * Now we'll do the bodies.
     */
    len = strlen(this->bodyPath);
    strncpy(path, this->bodyPath, sizeof(path));
    strncpy(path + len, renamedOldExt, sizeof(path) - len - 1);

    /**
     * Rename it.
     */
    x = rename(path, this->bodyPath);
    if (x != 0) {

        /**
         * Restore the headers too !!
         */
        len = strlen(this->headerPath);
        strncpy(path, this->headerPath, sizeof(path));
        strncpy(path + len, renamedOldExt, sizeof(path) - len - 1);
        rename(path, this->headerPath);

        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::copyDatabase
 *=------------------------------------------------------------------------=
 * copies DB source to DB dest.  Very complex.
 *
 * Parameters:
 *      DB *                    - [in]  source
 *      DB *                    - [in]  dest
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::copyDatabase
(
    DB *sourceDB,
    DB *destDB
)
{
    DBC *cursor = NULL;
    DBT key, data;
    ERRCODE er;
    int x;

    /**
     * get a cursor for the header database.
     */
    x = sourceDB->cursor(sourceDB, NULL, &cursor, 0);
    if (x != 0) {
        K_FAIL("Couldn't get Cursor for source DB!!!!");
        errorSaveRichErrorInfo(db_strerror(x));
        return E_DB_ERR;
    }

    /**
     * set up the data to get the first row !!!
     */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    x = cursor->c_get(cursor, &key, &data, DB_FIRST);
    if (x == DB_NOTFOUND) {
        er = S_OK;
        goto Error;
    }

    if (x) {
        K_FAIL("Unable to get a row in the database");
        errorSaveRichErrorInfo(db_strerror(x));
        er = E_DB_ERR;
        goto Error;
    }

    /**
     * Great, now just copy it to the output database !!!!
     */
    K_ASSERT(key.size == sizeof(struct timeval), "What the -- Bogus key in DB!!");

    x = destDB->put(destDB, NULL, &key, &data, DB_NOOVERWRITE);
    if (x) {
        K_FAIL("Unable to write a row in the database");
        errorSaveRichErrorInfo(db_strerror(x));
        er = E_DB_ERR;
        goto Error;
    }

    /**
     * Now loop through the rest!!!
     */
    while (x != DB_NOTFOUND) {

        if (x) break;

        /**
         * get the next row.
         */
        memset(&key, 0, sizeof(key));
        memset(&data, 0, sizeof(data));
        x = cursor->c_get(cursor, &key, &data, DB_NEXT);
        if (!x) {

            K_ASSERT(key.size == sizeof(struct timeval), "What the -- Bogus key in DB!!");

            /**
             * IF we successfully got the row, then go and write it out
             * to the destination db now ...
             */
            x = destDB->put(destDB, NULL, &key, &data, DB_NOOVERWRITE);
            if (x) {
                K_FAIL("Unable to write a row in the database");
                errorSaveRichErrorInfo(db_strerror(x));
                er = E_DB_ERR;
                goto Error;
            }
        }
    }

    if (x && x != DB_NOTFOUND) {
        errorSaveRichErrorInfo(db_strerror(x));
        er = E_DB_ERR;
        goto Error;
    }

    er = S_OK;

  Error:
    if (cursor) cursor->c_close(cursor);
    return er;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::createDisasterLock
 *=------------------------------------------------------------------------=
 * creates a lock for disaster recovery when compacting databases.  this is
 * actually pretty important.
 *
 * Output:
 *      ERRCODE             - failure means the lock wasn't created.
 */
ERRCODE FolderDB::createDisasterLock()
{
    return parent->requestCompactDisasterLock();
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::findMessage
 *=------------------------------------------------------------------------=
 * tries to find teh given message.
 *
 * Parameters:
 *      MailMessage *               - [in]  find me !!!
 *
 * Output:
 *      ERRCODE                     - S_OK found it
 *                                    S_FALSE nope.
 */
ERRCODE FolderDB::findMessage
(
    MailMessage *message
)
{
    MailMessage *m;
    ERRCODE er;

    er = messageList.firstElement(&m);
    while (er == S_OK) {
        if (m == message) return S_OK;

        er = messageList.nextElement(&m);
    }

    return S_FALSE;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::flush
 *=------------------------------------------------------------------------=
 * Makes sure that the data is written out to the db files.  We'll probably
 * call this on a semi-regular basis to make sure that we don't lose anything
 * in the event of power outage or crash.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::flush()
{
    /** 
     * UNDONE: MarcW, 10.8.2000 -- error checking here !!!!!  Where to find
     * out about this???
     */
    if (headers)
        (headers->sync)(headers, 0);
    if (bodies)
        (bodies->sync)(bodies, 0);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::generatePaths
 *=------------------------------------------------------------------------=
 * generate the paths for the two databsae files ...
 *
 * Parameters:
 *      const char *            - [in]  directory they will reside in.
 *      const char *            - [in]  base name for the files.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::generatePaths
(
    const char *dir,
    const char *baseName
)
{
    char path[MAXPATHLEN];
    int len2;
    int len;

    /**
     * Make sure they didn't give us a ridiculously huge path.
     */
    len = strlen(dir);
    len2 = strlen(baseName);
    if ((len + len2 + 6) > MAXPATHLEN) return E_PARAM;

    strcpy(path, dir);
    if (path[len - 1] != '/') {
        path[len] = '/';
        len++;
    }
    strcpy(path + len, baseName);

    /**
     * this points to the end of the string.
     */
    len = len + len2;

    /**
     * Okay, path now contains the full baseName.
     * 1. create the header db path
     */
    strcpy(path + len, headerSuffix);
    this->headerPath = copyString(path);
    if (headerPath == NULL) return E_OUTOFMEMORY;

    /**
     * 2. create the body path.
     */
    strcpy(path + len, bodiesSuffix);
    this->bodyPath = copyString(path);
    if (bodyPath == NULL) return E_OUTOFMEMORY;

    /**
     * Tutti va bene!!
     */
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::getMessageBody
 *=------------------------------------------------------------------------=
 * loads in the body for this message and sets everything in it up correctly.
 *
 * Parameters:
 *      MailMessage *           - [in]  message whose body we need to load.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::getMessageBody
(
    MailMessage *msg
)
{
    const char *buf, *pstr, *pstr2;
    int len, count, size, enc;
    MIMETYPE type;
    unsigned int index, tp;
    BYTE *attachData;
    int cbodytypes;
    Attachment *a;
    DBT key, data;
    int version;
    ERRCODE er;
    int x;

    if (!msg) return E_PARAM;
    K_ASSERT(bodies, "I can't load a message without any bodies");

    /**
     * Go and look for the row
     */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    key.size = sizeof(struct timeval);
    key.data = &msg->tv;
    x = bodies->get(bodies, NULL, &key, &data, 0);
    if (x) {
        fprintf(stderr, "Failed to read from bodies DB %x %s\n", x, db_strerror(x));
        errorSaveRichErrorInfo(db_strerror(x));
        return E_DB_ERR;
    }
    buf = (char *)data.data;

    /**
     * First thing is the version number.
     */
    index = 0;
    memcpy(&version, buf + index, sizeof(version));
    index += sizeof(version);

    /**
     * As of now, there is only one version supported.
     */
    if (version > BODY_VERSION_STAMP) {
        K_FAIL("Bogus Version Stamp in Database!!!!!!");
        return E_DB_VERSION;
    }

    /**
     * Next thing is count of body MIME types.
     */
    memcpy(&cbodytypes, buf + index, sizeof(cbodytypes));
    index += sizeof(cbodytypes);

    /**
     * Now, for each body, read in:
     *  - type
     *  - charset size
     *  - charset
     *  - size
     *  - body
     */
    for (int abc = 0; abc < cbodytypes; abc++) {

        pstr2 = NULL;

        /**
         * Type
         */
        memcpy(&tp, buf + index, sizeof(tp));
        index += sizeof(tp);
        if (version <= 3) {
            int zyx = 0;
            while (oldMimeTypeMap[zyx].oldType != UnknownType) {
                if (oldMimeTypeMap[zyx].oldType == (OldMimeTypeEnum)tp) {
                    type.setType(oldMimeTypeMap[zyx].mediaType, oldMimeTypeMap[zyx].subType);
                    break;
                }
                zyx++;
            }
            if (oldMimeTypeMap[zyx].oldType == UnknownType)
                type.setType(UnknownMediaType, UnknownSubType);

        } else {
            type.setWholeThing(tp);
        }

        /**
         * We added this in version 3 of the persistence format.
         */
        if (version > 2) {
            memcpy(&len, buf + index, sizeof(len));
            index += sizeof(len);
            if (len) {
                pstr2 = copyString(buf + index);
                if (!pstr2) { return E_OUTOFMEMORY; }
            }
        }

        /**
         * Body size
         */
        memcpy(&len, buf + index, sizeof(len));
        index += sizeof(len);

        /** 
         * Now copy out the body -- we wrote it out with a trailing 0
         * so this is easy.
         */
        pstr = copyString(buf + index);
        if (!pstr) { localFree(pstr2); return E_OUTOFMEMORY; }
        index += len;

        /**
         * Set it.  NULL charset is ok
         */
        er = msg->setBodyByType(type, pstr, pstr2);
        if (pstr2) localFree(pstr2);
        localFree(pstr);
        RETURN_ON_FAILURE(er);
    }

    /**
     * Now read in the count of attachments.
     */
    memcpy(&count, buf + index, sizeof(count));
    index += sizeof(count);

    /**
     * now for each attachment, read in:
     *  - size
     *  - encoding
     *  - type
     *  - filename size
     *             data
     *  - data
     */
    for (x = 0; x < count; x++) {
        
        memcpy(&size, buf + index, sizeof(size));
        index += sizeof(size);
        memcpy(&enc, buf + index, sizeof(enc));
        index += sizeof(enc);
        memcpy(&tp, buf + index, sizeof(tp));
        index += sizeof(tp);
        if (version <= 3) {

            int zyx = 0;
            while (oldMimeTypeMap[zyx].oldType != UnknownType) {
                if (oldMimeTypeMap[zyx].oldType == (OldMimeTypeEnum)tp) {
                    type.setType(oldMimeTypeMap[zyx].mediaType, oldMimeTypeMap[zyx].subType);
                    break;
                }
                zyx++;
            }
            if (oldMimeTypeMap[zyx].oldType == UnknownType)
                type.setType(UnknownMediaType, UnknownSubType);

        } else {
            type.setWholeThing(tp);
        }

        attachData = NULL;

        memcpy(&len, buf + index, sizeof(len));
        index += sizeof(len);
        pstr = copyString(buf + index);
        index += len + 1;
        if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
        
        if (size) {
            attachData = (BYTE *)localAlloc(size);
            if (!attachData) { er = E_OUTOFMEMORY; localFree(pstr); goto Error; }
            memcpy(attachData, buf + index, size);
            index += size;
        }

        /**
         * Create the attachment.
         */
        a = new Attachment(pstr, attachData, size, (MimeEncoding)enc, type);
        localFree(pstr);
        if (!a) { er = E_OUTOFMEMORY; localFree(attachData); goto Error; }

        /**
         * Add it!
         */
        er = msg->addAttachment(a);
        a->release();
        GOTO_ON_FAILURE(er, Error);
    }

    er = S_OK;

  Error:
    return er;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::getMessageHeaders
 *=------------------------------------------------------------------------=
 * goes and loads in all the headers and returns the mail message objects
 * in a big array.
 *
 * Parameters:
 *      MailMessage ***         - [out] we return a pointer to an array of 
 *                                  pointers to msg objects.
 *      int *                   - [out] number of them.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::getMessageHeaders
(
    MailMessage ***pmessages,
    int *cmessages
)
{
    MailMessage *msg;
    ERRCODE er;
    int c;

    if (!pmessages || !cmessages) return E_PARAM;
    *pmessages = NULL;
    *cmessages = 0;

    /**
     * make sure we've loaded in all the messages.
     */
    er = loadAllMessages();
    RETURN_ON_FAILURE(er);

    /**
     * Figure out how many messages we've got (if none, this is E-Zed.)
     */
    c = messageList.countElements();
    if (c == 0) return S_OK;

    /** 
     * put 'em in an array
     */
    *pmessages = (MailMessage **)localAlloc(sizeof(MailMessage *) * c);
    if (!*pmessages) return E_OUTOFMEMORY;
    *cmessages = c;

    memset(*pmessages, 0, sizeof(MailMessage *) * c);
    c = 0;

    er = messageList.firstElement(&msg);
    while (er == S_OK) {
        (*pmessages)[c++] = msg;

        er = messageList.nextElement(&msg);
    }
    
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::getMessages
 *=------------------------------------------------------------------------=
 * returns a copy of the list of messages in this folder.
 *
 * Parameters:
 *      LinkList<MailMessage *> & msgs
 *
 * Output:
 *      ERRCODE
 *
 * NOTE:  THE MESSAGES RETURNED HAVE THEIR REF COUNT INCREASED BY ONE, SO
 *      IN CASE THEY ARE DELETED OR REMOVED, THERE WON'T BE A BAD/DANGLING
 *      POINTER.  CALLEE *MUST* release() THESE POINTERS!!!!
 */
ERRCODE FolderDB::getMessages
(
    LinkList<MailMessage *> &msgs
)
{
    MailMessage *m;
    ERRCODE er;

    er = loadAllMessages();
    RETURN_ON_FAILURE(er);

    er = messageList.firstElement(&m);
    while (er == S_OK) {

        m->addRef();
        msgs.addElement(m);

        er = messageList.nextElement(&m);
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::hasMessages
 *=------------------------------------------------------------------------=
 * indicates whether this folder has any messages in it or not.
 *
 * Output:
 *      ERRCODE                 - S_OK -- yes, S_FALSE -- no.
 */
ERRCODE FolderDB::hasMessages()
{ 
    ERRCODE er = loadAllMessages();
    RETURN_ON_FAILURE(er);
    return messageList.isEmpty() ? S_FALSE : S_OK; 
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::loadAllMessages
 *=------------------------------------------------------------------------=
 * Loads in the message headers and then puts them in a list.  Fweeee!
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::loadAllMessages()
{
    DBC *cursor = NULL;
    struct timeval tv;
    MailMessage *msg;
    DBT key, data;
    ERRCODE er;
    int x;

    /**
     * If we've already loaded them all in, then we're done.
     */
    if (messagesLoaded) return S_OK;

    K_ASSERT(headers && bodies, "um ... I should have databases by now ...");

    /**
     * get a cursor for the header database.
     */
    x = headers->cursor(headers, NULL, &cursor, 0);
    if (x != 0) {
        K_FAIL("Couldn't get Cursor for headers DB!!!!");
        errorSaveRichErrorInfo(db_strerror(x));
        return E_DB_ERR;
    }

    /**
     * Set up the key and things, and get the first row.
     */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    x = cursor->c_get(cursor, &key, &data, DB_FIRST);
    if (x == DB_NOTFOUND) {
        messagesLoaded = true;
        er = S_OK;
        goto Error;
    }
    if (x) {
        K_FAIL("Unable to get a row in the database");
        errorSaveRichErrorInfo(db_strerror(x));
        er = E_DB_ERR;
        goto Error;
    }

    /**
     * Get the message object representing this entry in the DB and add it
     * to our list of messages.
     */
    K_ASSERT(key.size == sizeof(struct timeval), "What the -- Bogus key in DB!!");
    memcpy(&tv, key.data, sizeof(struct timeval));
    er = parseMessageHeader(tv, (const char *)data.data, &msg);
    GOTO_ON_FAILURE(er, Error);
    messageList.addElement(msg);

    /**
     * Now loop through the rest!!!
     */
    while (x != DB_NOTFOUND) {

        if (x) break;

        memset(&key, 0, sizeof(key));
        memset(&data, 0, sizeof(data));
        x = cursor->c_get(cursor, &key, &data, DB_NEXT);
        if (!x) {

            /**
             * Get the message object for this entry in the DB and add it to 
             * oust list of messages.
             */
            K_ASSERT(key.size == sizeof(struct timeval), "What the -- Bogus key in DB!!");
            memcpy(&tv, key.data, sizeof(struct timeval));
            er = parseMessageHeader(tv, (const char *)data.data, &msg);
            GOTO_ON_FAILURE(er, Error);
            messageList.addElement(msg);
        }
    }

    if (x && x != DB_NOTFOUND) {
        errorSaveRichErrorInfo(db_strerror(x));
        er = E_DB_ERR;
        goto Error;
    }

    /**
     * Great!!! we're done, and we'll flag the messages as all being loaded.
     */
    messagesLoaded = true;
    er = S_OK;

  Error:
      if (cursor) cursor->c_close(cursor);
    if (FAILED(er)) {
        ERRCODE er2;
        er2 = messageList.firstElement(&msg);
        while (er2 == S_OK) {
            messageList.removeElement(msg);
            if (msg) msg->release();
            er2 = messageList.firstElement(&msg);
        }
    }

    return er;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::lockForOutput
 *=------------------------------------------------------------------------=
 * prepares the folder to have all of its messages sent.  Basically, we just
 * need to move any UNSENT messages into the SENDING state.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::lockForOutput()
{
    ReadStatus ms;
    MailMessage *m;
    ERRCODE er;

    er = loadAllMessages();
    RETURN_ON_FAILURE(er);

    /**
     * Whiz through the list converting UNSENT messages to SENDING.
     */
    er = messageList.firstElement(&m);
    while (er == S_OK) {
        m->getStatus(ms);

        if (ms == MSG_UNSENT)
            m->setStatus(MSG_SENDING);
        
        er = messageList.nextElement(&m);
    }

    return (FAILED(er)) ? er : S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::compactReopenDatabases
 *=------------------------------------------------------------------------=
 * opens the database files for this account.  Very exciting!!!
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::openDatabases()
{
    ERRCODE er;

    if (headers != NULL || bodies != NULL) {
        K_FAIL("We've already got databases open !!!! ARGH!");
        return E_INTERNAL;
    }

    /**
     * Open the headers database.
     */
    er = dbmgrOpenDatabase(headers, headerPath, false, 0600);
    RETURN_ON_FAILURE(er);

    /**
     * 2. create the bodies db.
     */
    er = dbmgrOpenDatabase(bodies, bodyPath, false, 0600);
    RETURN_ON_FAILURE(er);

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::openFolder
 *=------------------------------------------------------------------------=
 * Creates/Opens the given folder as a new folder and returns the appropriate
 * object.  
 *
 * Parameters:
 *      MailFolder *            - [in]  parent folder.
 *      const char *            - [in]  filename given to this folder.
 *      const char *            - [in]  directory in which to create this folder
 *      bool                    - [in]  indicates whether we should create or
 *                                      just open
 *      FolderDB **               - [out] the new folder object
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::openFolder
(
    MailFolder *parent,
    const char *fileName,
    const char *directory,
    bool create,
    FolderDB **pfolder
)
{
    ERRCODE er;

    if (!directory || !fileName || !pfolder) return E_PARAM;

    /**
     * Create the new folder object.
     */
    *pfolder = new FolderDB();
    if (!(*pfolder)) return E_OUTOFMEMORY;

    (*pfolder)->parent = parent;

    /**
     * Tell the folder to figure out its paths.
     */
    er = (*pfolder)->generatePaths(directory, fileName);
    GOTO_ON_FAILURE(er, CleanUp);

    /**
     * Open the headers database.
     */
    er = dbmgrOpenDatabase((*pfolder)->headers, (*pfolder)->headerPath, create, 0600);
    GOTO_ON_FAILURE(er, CleanUp);

    /**
     * 2. create the bodies db.
     */
    er = dbmgrOpenDatabase((*pfolder)->bodies, (*pfolder)->bodyPath, create, 0600);
    GOTO_ON_FAILURE(er, CleanUp);

    /**
     * Sweet!! We've got the db's created.
     */
    er = S_OK;

  CleanUp:
    if (FAILED(er)) (*pfolder)->close();
    return er;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::removeFilesForPrefix
 *=------------------------------------------------------------------------=
 * nukes the files associated with the given database name.   This is 
 * typically done for temporary databases ..
 *
 * Parameters:
 *      const char *            - [in]  base filename.
 *      const char *            - [in]  directory
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::removeFilesForPrefix
(
    const char *fileName,
    const char *directory
)
{
    char path[MAXPATHLEN];
    int len, len2;

    /**
     * Set up the path so we can add on the extension for the header and bodies
     * db files.  We are going to duplicate this code in openFolder and 
     * because I don't want to create a new routine to slow this down
     * any ...
     */
    len = strlen(directory);
    len2 = strlen(fileName);
    if ((len + len2 + 6) > MAXPATHLEN) return E_PARAM;

    strcpy(path, directory);
    if (path[len - 1] != '/') {
        path[len] = '/';
        len++;
    }
    strcpy(path + len, fileName);

    // this points to the end of the string.
    //
    len = len + len2;

    /**
     * Okay, path now contains the full filename.
     * 1. create the header db.
     */
    strcpy(path + len, headerSuffix);

    /**
     * Now destroy the headers db.
     */
    unlink(path);    

    /**
     * Bodies
     */
    strcpy(path + len, bodiesSuffix);
    unlink(path);
    
    return S_OK;
}



/**
 *=------------------------------------------------------------------------=
 * FolderDB::parseMessageHeader
 *=------------------------------------------------------------------------=
 * Given the contents of a message header loaded in from the database, go
 * and get a MailMessage object for it. 
 *
 * Parameters:
 *      struct timeval          - [in]  key to this message.
 *      const char *            - [in]  the header to parse out.
 *      MailMessage **          - [out] the new message object for this hdr
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::parseMessageHeader
(
    struct timeval tv,
    const char *header,
    MailMessage **pmessage
)
{
    LinkList<const char *> list;
    char buf[MAX_HEADER_ENTRY_LEN];
    int index, len, count;
    const char *pstr;
    int version;
    time_t dval;
    ERRCODE er;
    bool bval;
    int ival;
    int x;

    if (!header || !pmessage) return E_PARAM;

    /**
     * Create the new message and copy over it's key
     */
    *pmessage = new MailMessage();
    if (!(*pmessage)) return E_OUTOFMEMORY;
    (*pmessage)->tv = tv;
    (*pmessage)->hostFolder = parent;

    /**
     * First, get the version number!!!
     */
    index = 0;
    memcpy(&version, header + index, sizeof(version));
    index += sizeof(version);

    x = 0;
    while (persistenceTable[x].type != PT_FINITO) {

        /**
         * if It's only for a newer version, then skip it.
         */
        if (persistenceTable[x].version > version) {
            x++;
            continue;
        }

        switch (persistenceTable[x].type) {

            /**
             * Strings have a length, and then the string (including
             * a null terminator).
             */
            case PT_STRING:
                memcpy(&len, header + index, sizeof(len));
                index += sizeof(len);
                if (len) {
                    memcpy(buf, header + index, len + 1);
                    index += len + 1;
                    er = ((*(*pmessage)).*((ERRCODE (MailMessage::*)(const char *))persistenceTable[x].setFunction))(buf);
                    GOTO_ON_FAILURE(er, Error);
                }
                break;

            /**
             * Stringlists have a count of strings, followed by a sequence
             * of string lengths and then the strings themselves, which are
             * all null terminated.
             */
            case PT_STRINGLIST:
                memcpy(&count, header + index, sizeof(count));
                index += sizeof(count);

                /**
                 * Now whiz through and load in each one
                 */
                for (int xx = 0; xx < count; xx++) {
                    memcpy(&len, header + index, sizeof(len));
                    index += sizeof(len);
                    memcpy(buf, header + index, len + 1);
                    index += len + 1;
                    pstr = copyString(buf);
                    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
                    list.addElement(pstr);
                }

                er = ((*(*pmessage)).*((ERRCODE (MailMessage::*)(LinkList<const char *>&))persistenceTable[x].setFunction))(list);
                list.emptyList();
                GOTO_ON_FAILURE(er, Error);
                break;

            case PT_DATE:
                memcpy(&len, header + index, sizeof(len));
                index += sizeof(len);
                memcpy(&dval, header + index, len);
                index += len;
                er = ((*(*pmessage)).*((ERRCODE (MailMessage::*)(time_t))persistenceTable[x].setFunction))(dval);
                GOTO_ON_FAILURE(er, Error);
                break;

            case PT_INTEGER:
                memcpy(&len, header + index, sizeof(len));
                index += sizeof(len);
                memcpy(&ival, header + index, len);
                index += len;
                er = ((*(*pmessage)).*((ERRCODE (MailMessage::*)(int))persistenceTable[x].setFunction))(ival);
                GOTO_ON_FAILURE(er, Error);
                break;

            case PT_BOOLEAN:
                memcpy(&len, header + index, sizeof(len));
                index += sizeof(len);
                memcpy(&bval, header + index, len);
                index += len;
                er = ((*(*pmessage)).*((ERRCODE (MailMessage::*)(bool))persistenceTable[x].setFunction))(bval);
                GOTO_ON_FAILURE(er, Error);
                break;

            default:
                K_FAIL("Bogus Property type in persistence table.  very BAD!!!");
                er = E_INTERNAL;
                goto Error;
                break;
        }

        x++;
    }

    er = S_OK;

  Error:
    if (FAILED(er)) {
        (*pmessage)->release();
        nukeStringList(list);
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::removeDisasterLock
 *=------------------------------------------------------------------------=
 * removes the compact database disaster lock, which means others can 
 * continue operating.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::removeDisasterLock()
{
    return parent->releaseCompactDisasterLock();
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::removeMessage
 *=------------------------------------------------------------------------=
 * removes the given message from this folder.  It would, of course, be
 * very bad if the message didn't exist.
 *
 * Parameters:
 *      MailMessage *           - [in]  remove me please
 *
 * Output:
 *      ERRCODE                 - E_NOTFOUND if not there.
 */
ERRCODE FolderDB::removeMessage
(
    MailMessage *removemePlease
)
{
    MailMessage *m;
    ERRCODE er;
    DBT key;
    int x;

    if (!bodies || !headers) return E_INTERNAL;
    if (!removemePlease) return E_PARAM;

    er = loadAllMessages();
    RETURN_ON_FAILURE(er);

    /**
     * find it first!!!
     */
    er = messageList.firstElement(&m);
    while (er == S_OK) {
        if (m == removemePlease) break;

        er = messageList.nextElement(&m);
    }
    if (FAILED(er) || er == S_NODATA) {
        K_FAIL("Asked to delete a message we don'ts gots");
        return E_NOTFOUND;
    }

    /**
     * now begin the actual work of removing it
     */
    messageList.removeElement(m);
    memset(&key, 0, sizeof(key));
    key.size = sizeof(struct timeval);
    key.data = &m->tv;

    x = headers->del(headers, NULL, &key, 0);
    if (x != 0) {
        fprintf(stderr, "Couldn't remove message from headers db '%s'\n", db_strerror(x));
        errorSaveRichErrorInfo(db_strerror(x));
        return E_DB_ERR;
    }
    x = bodies->del(bodies, NULL, &key, 0);
    if (x != 0) {
        fprintf(stderr, "Couldn't remove message from bodies db '%s'\n", db_strerror(x));
        errorSaveRichErrorInfo(db_strerror(x));
        return E_DB_ERR;
    }

    m->release();
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::unloadAllMessages
 *=------------------------------------------------------------------------=
 * frees up the memory that the messages are consuming, in an attempt to
 * save memory.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::unloadAllMessages()
{
    MailMessage *msg;
    ERRCODE er;

    /**
     * IF we haven't loaded any messages, this is easy !!
     */
    if (!messagesLoaded) return S_OK;

    /**
     * Otherwise, whiz through them, releasing them all, and then clear out
     * the list.
     */
    er = messageList.firstElement(&msg);
    while (er == S_OK) {

        messageList.removeElement(msg);
        msg->release();
        er = messageList.firstElement(&msg);
    }
    
    messagesLoaded = false;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * FolderDB::updateHeader
 *=------------------------------------------------------------------------=
 * updates the header for the given message.  We have to delete the old header
 * from the headers db and then write out the new.  it's a bit pedantic, but
 * it works.
 *
 * Parameters:
 *      MailMessage *               - [in]  message to update.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::updateHeader
(
    MailMessage *message
)
{
    DBT key;
    int x;

    if (!message) return E_PARAM;

    /**
     * First thing we have to do is delete the header from the appropriate
     * db
     */
    K_ASSERT(headers, "how the HELL do I not have this now???");
    memset(&key, 0, sizeof(key));

    key.size = sizeof(struct timeval);
    key.data = &message->tv;

    x = headers->del(headers, NULL, &key, 0);
    if (x) {
        K_FAIL("Couldn't delete header from database.  Yikes!!!");
        fprintf(stderr, "Database Delete failed with %s\n", db_strerror(x));
        errorSaveRichErrorInfo(db_strerror(x));
        return E_DB_ERR;
    }

    return writeMessageHeader(message, message->tv);
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::writeMessageBody
 *=------------------------------------------------------------------------=
 * writes out the message body, being sure to handle all things such as
 * different text formats and attachments, etc ...
 *
 * Parameters:
 *      MailMessage *               - [in]  message we're writing out
 *      struct timeval &            - [in]  the key for this message.
 *
 * Output:
 *      ERRCODE
 *
 * Notes:
 */
ERRCODE FolderDB::writeMessageBody
(
    MailMessage *message,
    struct timeval & tv
)
{
    unsigned int totalSize, index, numattached, tp;
    const char *body = NULL, *pstr, *charset;
    LinkList<Attachment *> attachments;
    MIMETYPE *types = NULL, type;
    int *bodySizes = NULL;
    Attachment *attach;
    MimeEncoding enc;
    BYTE *attachData;
    char *buf = NULL;
    int sz, len;
    int cbodytypes;
    DBT key, data;
    ERRCODE er;
    int x;

    struct timeval xx = tv;

    /**
     *=-------------------------------------------------------------=
     * Message body format is:
     *
     * 0. Version #
     * 1. Count of MIME Bodies
     * 2. MIME Bodies
     *      a. type
     *      b. size
     *      c. charset
     *          I.  length
     *          II. charset
     *      c. body
     * 3. Count of Attachments.
     * 4. Attachments:
     *      a. see below for details of this.
     *=-------------------------------------------------------------=
     */
    totalSize = 0;

    /**
     * we need to compute the total body size and allocate a buffer for it.
     * we want to compute it all up front, because otherwise this is going to
     * be huge!!!
     *
     * First, space for version and count of bodies
     */
    totalSize += 2 *sizeof(int);
    
    er = message->getBodyFormats(&types, &cbodytypes);
    RETURN_ON_FAILURE(er);

    K_ASSERT(cbodytypes >= 1, "There must be at least ONE body for a msg\n");

    bodySizes = (int *)localAlloc(sizeof(int) * cbodytypes);
    if (!bodySizes) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * Compute the size for each body.
     */
    for (int abc = 0; abc < cbodytypes; abc++) {
        er = message->getBodyRefByType(types[abc], &body, &charset);
        GOTO_ON_FAILURE(er, Error);

        bodySizes[abc] = strlen(body) + 1;

        /**
         * For total size, add room for: body type, body size,
         * charset size, charset, and body
         */
        totalSize += 3 * sizeof(int);
        totalSize += bodySizes[abc];

        if (charset) totalSize += strlen(charset);
    }

    numattached = 0;
    er = message->getAttachments(attachments);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Add number of attachments.
     */
    totalSize += sizeof(int);

    /**
     * And for each attachment, get the binary size, as well as the file
     * name length.
     */
    er = attachments.firstElement(&attach);
    while (er == S_OK ) {
        er = attach->getBinarySize(sz);
        GOTO_ON_FAILURE(er, Error);
        totalSize += (sz + (3 * sizeof(int)));    // 1 for size, 1 for encoding, 1 for type

        er = attach->getFileNameLen(&len);
        GOTO_ON_FAILURE(er, Error);
        totalSize += len + 1 + sizeof(int);

        numattached++;
        er = attachments.nextElement(&attach);
    }

    /**
     * Great, we've got the total size now.  Allocate a buffer for it.
     */
    buf = (char *)localAlloc(totalSize);
    if (!buf) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * Okay, now write out the body.
     * 0. version number.
     */
    index = 0;
    len = BODY_VERSION_STAMP;
    memcpy(buf + index, &len, sizeof(len));
    index += sizeof(len);

    /**
     * 1. count of bodies.
     */
    memcpy(buf + index, &cbodytypes, sizeof(cbodytypes));
    index += sizeof(cbodytypes);

    /**
     * 2. MIME bodies.
     */
    for (int abc = 0; abc < cbodytypes; abc++) {
        er = message->getBodyRefByType(types[abc], &body, &charset);
        GOTO_ON_FAILURE(er, Error);
        
        /**
         * First, the type.
         */
        tp = types[abc].getWholeThing();
        memcpy(buf + index, &tp, sizeof(tp));
        index += sizeof(tp);

        /**
         * next the charset size and charset itself.
         */
        len = charset ? strlen(charset) : 0;
        memcpy(buf + index, &len, sizeof(len));
        index += sizeof(len);
        if (charset) memcpy(buf + index, charset, len + 1);

        /**
         * Next the size of the body.
         */
        memcpy(buf + index, &(bodySizes[abc]), sizeof(int));
        index += sizeof(int);

        /**
         * Next, the actual body itself.
         */
        memcpy(buf + index, body, bodySizes[abc]);
        index += bodySizes[abc];
    }

    /**
     * 3. Number of attachments.
     */
    memcpy(buf + index, &numattached, sizeof(numattached));
    index += sizeof(numattached);

    /**
     * For each attachment:
     * 4. size
     * 5. encoding
     * 6. MIME type
     * 7a.sizeof filename
     * 7b.filename
     * 8. attachment body in decoded format.
     */
    er = attachments.firstElement(&attach);
    while (er == S_OK) {
        er = attach->getBinarySize(sz);
        GOTO_ON_FAILURE(er, Error);

        memcpy(buf + index, &sz, sizeof(sz));
        index += sizeof(sz);

        er = attach->getEncoding(enc);
        GOTO_ON_FAILURE(er, Error);
        
        memcpy(buf + index, &enc, sizeof(enc));
        index += sizeof(enc);

        er = attach->getMimeType(type);
        GOTO_ON_FAILURE(er, Error);

        tp = type.getWholeThing();
        memcpy(buf + index, &tp, sizeof(tp));
        index += sizeof(tp);

        er = attach->getFileName(&pstr);
        GOTO_ON_FAILURE(er, Error);

        len = strlen(pstr);
        memcpy(buf + index, &len, sizeof(len));
        index += sizeof(len);

        memcpy(buf + index, pstr, len + 1);
        index += len + 1;
        localFree(pstr);

        er = attach->getBinaryDataRef(&attachData);
        GOTO_ON_FAILURE(er, Error);

        memcpy(buf + index, attachData, sz);
        index += sz;

        er = attachments.nextElement(&attach);
    }

    /**
     * Write out the base body for the message now.
     */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    key.data = &xx;
    key.size = sizeof(xx);

    data.data = (void *)buf;
    data.size = totalSize;

    x = bodies->put(bodies, NULL, &key, &data, DB_NOOVERWRITE);
    if (x) {    
        K_FAIL("There was an error attempting to write to the bodies db!!!!");
        fprintf(stderr, "DB Error: %s\n", db_strerror(x));
        errorSaveRichErrorInfo(db_strerror(x));
        er = E_DB_ERR;
        goto Error;
    }

    /**
     * Fweeeee!!!!!!!!!!
     */
    er = S_OK;

  Error:
    if (types) localFree(types);
    if (bodySizes) localFree(bodySizes);
    if (buf) localFree(buf);
    if (FAILED(er)) {
        /**
         * Crap -- try to remove the header too, so we don't have inconsistent 
         * data.
         */
        memset(&key, 0, sizeof(key));
        key.size = sizeof(xx);
        key.data = &xx;
        headers->del(headers, NULL, &key, 0);
    }
    
    return er;
}


/**
 *=------------------------------------------------------------------------=
 * FolderDB::writeMessageHeader
 *=------------------------------------------------------------------------=
 * writes out the given message to the headers db.  Makes sure that all
 * the information we need is saved out.  The body and attachments are 
 * ignored, of course.
 *
 * Parameters:
 *      MailMessage *               - [in]  write my header out.
 *      struct timeval &            - [in]  the timestamp for this message.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderDB::writeMessageHeader
(
    MailMessage *message,
    struct timeval &tv
)
{
    unsigned int bufSize, index, len, count;
    LinkList<const char *> list;
    const char *sval;
    DBT key, data;
    ERRCODE er;
    time_t dval;
    char *buf;
    int ival;
    bool bval;
    int x, z;

    if (!headers || !bodies) return E_INTERNAL;
    if (!message) return E_PARAM;

    /**
     * First thing to do is set up the buf so that it has some room.
     */
    buf = (char *)localAlloc(MESSAGE_HEADER_CHUNK);
    if (!buf) return E_OUTOFMEMORY;
    index = 0;
    bufSize = MESSAGE_HEADER_CHUNK;

    /**
     * The first thing we write out is the version number.
     */
    ival = HEADER_VERSION_STAMP;
    memcpy(buf + index, &ival, sizeof(ival));
    index += sizeof(ival);

    /**
     * Now, start whizzing through our array, and writing out whatever we's got.
     */
    x = 0;
    while (persistenceTable[x].type != PT_FINITO) {

        switch (persistenceTable[x].type) {

            /**
             * For Strings, we write out:
             *  - strlen (withOUT the trailing \0)
             *  - string (WITH trailing \0)
             */
            case PT_STRING:
                er = ((*message).*((ERRCODE (MailMessage::*)(const char **))persistenceTable[x].getFunction))(&sval);
                GOTO_ON_FAILURE(er, Error);
                len = strlen(sval ? sval : "");

                if (index + len + sizeof(len) + 1 >= bufSize) {
                    char *buf2 = (char *)localReAlloc(buf, bufSize + MESSAGE_HEADER_CHUNK);
                    if (!buf2) { er = E_OUTOFMEMORY; goto Error; }
                    buf = buf2;
                    bufSize += MESSAGE_HEADER_CHUNK;
                }

                memcpy(buf + index, &len, sizeof(len));
                index += sizeof(len);
                if (len > 0) {
                    memcpy(buf + index, sval, len + 1);
                    index += len + 1;
                }
                if (sval) localFree(sval);
                break;

            /**
             * For String Lists we write out:
             *  - count of strings.
             *  - for each string:
             *      - strlen(withOUT the trailing \0)
             *      - string (WITH trailing \0)
             */
            case PT_STRINGLIST:
                er = ((*message).*((ERRCODE (MailMessage::*)(LinkList<const char *> &))persistenceTable[x].getFunction))(list);
                GOTO_ON_FAILURE(er, Error);

                count = list.countElements();
                if (index + sizeof(int) >= bufSize) {
                    char *buf2 = (char *)localReAlloc(buf, bufSize + MESSAGE_HEADER_CHUNK);
                    if (!buf2) { er = E_OUTOFMEMORY; goto Error; }
                    buf = buf2;
                    bufSize += MESSAGE_HEADER_CHUNK;
                }
                memcpy(buf + index, &count, sizeof(count));
                index += sizeof(count);

                /**
                 * Now whiz through the list writing out elements one at
                 * a time.
                 */
                er = list.firstElement(&sval);
                z = 0;
                while (er == S_OK) {

                    len = strlen(sval);

                    if (index + len + sizeof(len) + 1 >= bufSize) {
                        char *buf2 = (char *)localReAlloc(buf, bufSize + MESSAGE_HEADER_CHUNK);
                        if (!buf2) { er = E_OUTOFMEMORY; goto Error; }
                        buf = buf2;
                        bufSize += MESSAGE_HEADER_CHUNK;
                    }

                    memcpy(buf + index, &len, sizeof(len));
                    index += sizeof(len);
                    if (len > 0) {
                        memcpy(buf + index, sval, len + 1);
                        index += len + 1;
                    }
                    er = list.nextElement(&sval);
                }
                list.emptyList();
                break;

            /**
             * For dates we write out:
             *  - size of date
             *  - date
             */
            case PT_DATE:
                er = ((*message).*((ERRCODE (MailMessage::*)(time_t *))persistenceTable[x].getFunction))(&dval);
                GOTO_ON_FAILURE(er, Error);

                if (index + sizeof(int) +sizeof(time_t) >= bufSize) {
                    char *buf2 = (char *)localReAlloc(buf, bufSize + MESSAGE_HEADER_CHUNK);
                    if (!buf2) { er = E_OUTOFMEMORY; goto Error; }
                    buf = buf2;
                    bufSize += MESSAGE_HEADER_CHUNK;
                }

                len = sizeof(dval);
                memcpy(buf + index, &len, sizeof(len));
                index += sizeof(len);
                memcpy(buf + index, &dval, sizeof(dval));
                index += sizeof(dval);
                break;

            /**
             * For Integers we write out:
             *  - sizeof int
             *  - int
             */
            case PT_INTEGER:
                er = ((*message).*((ERRCODE (MailMessage::*)(int *))persistenceTable[x].getFunction))(&ival);
                GOTO_ON_FAILURE(er, Error);

                if (index + 2 * sizeof(int) >= bufSize) {
                    char *buf2 = (char *)localReAlloc(buf, bufSize + MESSAGE_HEADER_CHUNK);
                    if (!buf2) { er = E_OUTOFMEMORY; goto Error; }
                    buf = buf2;
                    bufSize += MESSAGE_HEADER_CHUNK;
                }

                len = sizeof(ival);
                memcpy(buf + index, &len, sizeof(len));
                index += sizeof(len);
                memcpy(buf + index, &ival, sizeof(ival));
                index += sizeof(ival);
                break;

            /**
             * For booleans we write out:
             *  - sizeof bool
             *  - bool value
             */
            case PT_BOOLEAN:
                er = ((*message).*((ERRCODE (MailMessage::*)(bool *))persistenceTable[x].getFunction))(&bval);
                GOTO_ON_FAILURE(er, Error);

                if (index + sizeof(int) + sizeof(bool) >= bufSize) {
                    char *buf2 = (char *)localReAlloc(buf, bufSize + MESSAGE_HEADER_CHUNK);
                    if (!buf2) { er = E_OUTOFMEMORY; goto Error; }
                    buf = buf2;
                    bufSize += MESSAGE_HEADER_CHUNK;
                }

                len = sizeof(bval);
                memcpy(buf + index, &len, sizeof(len));
                index += sizeof(len);
                memcpy(buf + index, &bval, sizeof(bval));
                index += sizeof(bval);
                break;


            default:
                K_FAIL("There is a bogus type in the persistence table.  This is hugely bad");
                er = E_INTERNAL;
                goto Error;
                break;
        }
    
        x++;
    }

    /**
     * OKay, the buffer is all set up -- we need to go and actually write the 
     * data to the database now!!!!
     */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    key.data = &tv;
    key.size = sizeof(tv);
    data.data = buf;
    data.size = index + 1;

    /**
     * Write out the data.
     */
    x = headers->put(headers, NULL, &key, &data, DB_NOOVERWRITE);
    if (x) {
        fprintf(stderr, "There was an error attempting to write to the headers db: \"%s\"\n", db_strerror(x));
        errorSaveRichErrorInfo(db_strerror(x));
        er = E_DB_ERR;
        goto Error;
    }

    er = S_OK;

  Error:
    if (buf) localFree(buf);
    return er;
}



