/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _FOLDERDB_H_
#define _FOLDERDB_H_

#include <db.h>

#include <llist.h>

class MailFolder;
class MailMessage;
class PropertiesFile;
struct timeval;

/**
 * This class encapsulates a mail folder database file.  It is used to load 
 * and save messages.  It maintains a list of MailMessages representing the
 * messages for which it is responsible.
 * 
 * Each folder 'file' is, in fact, two files.  I decided to have one that
 * represents the headers only, so headers could be loaded in quickly, and
 * then one for the bodies, since those aren't required as often, and vary
 * in size quite a bit.   The key and unifying factor for the two files is
 * a timeval struct timestamp, accurate to the nearest micro-sec.  There are
 * never duplicates.
 */
class FolderDB {

    DEBUG_TEST_DECL;

  public:
    /**
     * Opens the folder with the given name.  Returns failure on error -- i.e.
     * we do not create it if it exists -- the only way this would happen is
     * if the user deleted the file.
     * Returns the opened folder object.
     */
    static ERRCODE openFolder(MailFolder *parent,
                              const char *fileName,
                              const char *dir,
                              bool createIfNecessary,
                              FolderDB **);

    /**
     * Closes down this folder to make sure that everything is saved/written
     * out.
     */
    ERRCODE close();

    /**
     * Compact Me.
     */
    ERRCODE compact();

    /**
     * Makes sure that everything is up to date
     */
    ERRCODE flush();

    /**
     * Indicates whether or not we have any messages in this folder.
     */
    ERRCODE hasMessages();

    /**
     * Adds the given message (Both header and body) to the folder.
     */
    ERRCODE addMessage(MailMessage *addMe);

    /** 
     * Tries to find the given message in this folder.  If it's there,
     * we return S_OK, else S_FALSE.
     */
    ERRCODE findMessage(MailMessage *);

    /**
     * Retrieves the body of the given message.
     */
    ERRCODE getMessageBody(MailMessage *imMissingSomething);

    /**
     * Returns an array with all the messages in them -- we've only loaded
     * the headers for these messages though.  We return an array of pointers
     * to objects. (hence the ***).
     */
    ERRCODE getMessageHeaders(MailMessage ***, int *);

    /**
     * Prepares the folder for uploading messages.
     */
    ERRCODE lockForOutput();

    /**
     * Removes the given message from this database and releases any holds we
     * have on it.
     */
    ERRCODE removeMessage(MailMessage *);

    /**
     * Updates the header for the given message.  most typically called when
     * a message's status changed.
     */
    ERRCODE updateHeader(MailMessage *message);

    /**
     * Clears out the memory for the messages.  makes loading in new folders
     * perhaps slightly more expensive, but definitely saves us some memory
     * room.
     */
    ERRCODE unloadAllMessages();

    /**
     * Returns a copy of the list of messages in this folder.
     */
    ERRCODE getMessages(LinkList<MailMessage *> &);

    /**
     * Removes the files for the given database name.  There are two files
     * created for each base name, so this function knows how to make those
     * and kill 'em.
     */
    static ERRCODE removeFilesForPrefix(const char *file, const char *dir);

    ~FolderDB();

  private:
    LinkList<MailMessage *> messageList;
    bool messagesLoaded;

    /**
     * These are the actual Berkeley DB objects.
     */
    DB *headers, *bodies;

    /**
     * This is our parent folder.
     */
    MailFolder *parent;

    /**
     * These are the paths to the two folder files.
     */
    const char *headerPath;
    const char *bodyPath;


    /**
     =-----------------------------------------------------------------------=
     * Private routines.
     */
    FolderDB();

    /**
     * Loads in the messages (headers only) and stores them in our list of
     * messages.
     */
    ERRCODE loadAllMessages();

    /**
     * Given a message header, parse out the details and return a message
     * object for it.
     */
    ERRCODE parseMessageHeader(struct timeval, const char *, MailMessage **);

    ERRCODE writeMessageBody(MailMessage *, struct timeval &);
    ERRCODE writeMessageHeader(MailMessage *, struct timeval &);

    /**
     * Generate the paths for the db files ..
     */
    ERRCODE generatePaths(const char *dir, const char *baseName);    

    /**
     * Compacting operations.  See FolderDB::compact for more details
     * on how all this works -- it's critical to be super robust here !!!
     */
    ERRCODE compactCopyDatabases();
    ERRCODE compactRenameDatabases();
    ERRCODE compactReplaceDatabases();
    ERRCODE compactRestoreDatabases();
    ERRCODE compactDeleteOldDatabases();
    ERRCODE compactDeleteCopies();

    ERRCODE createDisasterLock();
    ERRCODE removeDisasterLock();
    ERRCODE openDatabases();

    /**
     * Cute little routine to copy a database.
     */
    static ERRCODE copyDatabase(DB *source, DB *dest);

};



#endif // _FOLDERDB_H_



