/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>


#include "address.h"
#include "foldersort.h"
#include <kmaillist.h>
#include "mailcolumns.h"
#include "mailmessage.h"


/**
 *=------------------------------------------------------------------------=
 * Local function prototypes
 *=------------------------------------------------------------------------=
 */
int compareAddresses(const char *, const char *);
int compareSubjects(const char *, const char *);
const char *stripPrefixes(const char *);

/**
 *=------------------------------------------------------------------------=
 * Local Data
 *=------------------------------------------------------------------------=
 */

/**
 * UNDONE: marcw, 2001-4-4 -- we need a way to add other prefixes we know
 *       about to this list!!!
 */
const char *subjectPrefixes [] = {
    "re",
    "fw",
    "aw",
    "ot"
};



/**
 *=------------------------------------------------------------------------=
 * messageSortDate
 *=------------------------------------------------------------------------=
 * given the two items that are being sorted by date, compare them.
 *
 * Parameters:
 *      int                 - [in]  column to sort on.
 *      unsigned int        - [in]  header itemData
 *      KMailListItem *     - [in]  item 1 to compare
 *      KMailListItem *     - [in]  item 2 to compare
 *
 * Output:
 *      int                 - < 0 if item 1 is greater.
 *                              0 if they're the same
 *                            > 0 if item 2 is greater
 */
int messageSortDate
(
    int col,
    unsigned int itemData,
    KMailListItem *item1,
    KMailListItem *item2
)
{
    time_t t1, t2;

    switch ((MailColumn)itemData) {
        case MC_SENT:
            ((MailMessage *)item1->itemData)->getSent(&t1);
            ((MailMessage *)item2->itemData)->getSent(&t2);
            break;

        /**
         * The default case is here for secondary sorts:  basically, when a
         * primary sort on a field such as Subject: returns two items that
         * are the same, the code just comes here and asks us to figure out
         * which message is older, and says that's the lesser of the two.
         */
        case MC_RECEIVED:
        default:
            ((MailMessage *)item1->itemData)->getReceived(&t1);
            ((MailMessage *)item2->itemData)->getReceived(&t2);
            break;
    }

    if (t1 > t2)
        return 1;
    else if (t1 == t2)
        return 0;
    else 
        return -1;
}


/**
 *=------------------------------------------------------------------------=
 * messageSortPixmap
 *=------------------------------------------------------------------------=

BOGUS PLACE HOLDER -- GET RID OF ME!!!

*/

int messageSortPixmap
(
    int col,
    unsigned int itemData,
    KMailListItem *item1,
    KMailListItem *item2
)
{
    return 0;
}



/**
 *=------------------------------------------------------------------------=
 * messageSortString
 *=------------------------------------------------------------------------=
 * sorts strings in a mail list item.
 *
 * Parameters:
 *      int                 - [in]  column to sort on.
 *      unsigned int        - [in]  header itemData
 *      KMailListItem *     - [in]  item 1 to compare
 *      KMailListItem *     - [in]  item 2 to compare
 *
 * Output:
 *      int                 - < 0 if item 1 is greater.
 *                              0 if they're the same
 *                            > 0 if item 2 is greater
 */
int messageSortString
(
    int col,
    unsigned int itemData,
    KMailListItem *item1,
    KMailListItem *item2
)
{
    int result;

    K_ASSERT(item1 && item2, "Bogus items to compare !! Crap");
    if (!item1 || !item2) return 0;

    int x = (item1->columns[col].text != NULL) << 1;
    x |= item2->columns[col].text != NULL;

    switch (x) {
        case 0:
            return 0;
        case 1:
            return -1;
        case 2:
            return 1;
        case 3: 
            switch ((MailColumn)itemData) {

                case MC_SUBJECT:
                    result = compareSubjects(item1->columns[col].text, item2->columns[col].text);
                    if (result == 0) return messageSortDate(col, itemData, item1, item2);
                    return result;

                case MC_FROM:
                case MC_TO:
                    result = compareAddresses(item1->columns[col].text, item2->columns[col].text);
                    if (result == 0) return messageSortDate(col, itemData, item1, item2);
                    return result;

                default:
                    return strcasecmp(item1->columns[col].text, item2->columns[col].text);
            }
            break;
    }

    // dead code
    return 0;
}

/**
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 *                          Local functions
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * compareAddresses
 *=------------------------------------------------------------------------=
 * compares two email addresses, making sure to compare just the name parts
 * and not any quotes, etc ..
 *
 * Parameters:
 *      const char *                - [in]  str 1
 *      const char *                - [in]  str 2
 *
 * Output:
 *      int                         - result from strncasecmp
 */
int compareAddresses
(
    const char *addr1,
    const char *addr2
)
{
    char buf1[64], buf2[64];

#if K_DEBUG
    if (!addr1 || !addr2) fprintf(stderr, "ARGH! missing an address\n");
#endif // K_DEBUG

    /**
     * First thing we need to do is go and get the user name.
     */
    addressExtractUserName(addr1, buf1, sizeof(buf1));
    addressExtractUserName(addr2, buf2, sizeof(buf2));

    /**
     * Now just use some lovely existing APIs for the rest.
     */
    return strcasecmp(buf1, buf2);

}


/**
 *=------------------------------------------------------------------------=
 * compareSubjects
 *=------------------------------------------------------------------------=
 * given two strings that are, in fact, message subjects, compare them,
 * making sure to remove Re, Fw, AW, etc ...
 *
 * Parameters:
 *      const char *                - [in]  subject 1
 *      const char *                - [in]  subject 2
 *
 * Output:
 *      int                         - > 0 if 1 > 2, 0 if 1 == 2, or < 0 if
 *                                      1 < 2
 */
int compareSubjects
(
    const char *subject1,
    const char *subject2
)
{
    const char *pstr1, *pstr2;

#if K_DEBUG
    if (!subject1 || !subject2) fprintf(stderr, "ARGH! missing a subject\n");
#endif // K_DEBUG

    /**
     * First thing we need to do is go and remove the prefixes.
     */
    pstr1 = stripPrefixes(subject1);
    pstr2 = stripPrefixes(subject2);

    /**
     * Now just use some lovely existing APIs for the rest.
     */
    return strcasecmp(pstr1, pstr2);
}







/**
 *=------------------------------------------------------------------------=
 * stripPrefixes
 *=------------------------------------------------------------------------=
 * given a string, return a pointer without any mail message prefixes, such as
 *
 * Re
 * Fw
 * AW
 *
 * etc.  please note that the original string isn't modified -- we just return
 * a pointer into it.
 *
 * Parameters:
 *      const char *            - [in]  strip me.
 *
 * Output:
 *      const char *
 */
const char *stripPrefixes
(
    const char *stripMe
)
{
    unsigned int y;
    int x = 0;
    int len;

    /**
     * As long as the string starts with a prefix, increment x to skip over
     * it.  This lets us change Re: Re: Re: MOo to just MOo.
     */
    while (stripMe[x] != '\0') {

        /**
         * Skip over any space that starts the subject string.
         */
        while (isspace(stripMe[x])) x++;

        /**
         * If the string starts with one of the prefixes, then we are very
         * happy!
         */
        for (y = 0; y < sizeof(subjectPrefixes) / sizeof(const char *); y++) {
            len = strlen(subjectPrefixes[y]);

            /**
             * Not only does the subject have to start with the prefix, 
             * it also has to have nothing after it or a ':'.  Otherwise,
             * we might interpret "Subject:RedHat" as Re !!
             */
            if (!strncasecmp(stripMe + x, subjectPrefixes[y], len)
                && (isspace(stripMe[x + len])
                    || stripMe[x + len] == ':'))
                break;
        }

        /**
         * IF there were no matches, then we're done.  Otherwise, skip over 
         * any : chars and white space and see if there are more.
         */
        if (y == sizeof(subjectPrefixes) / sizeof(const char *)) {
            break;
        } else {
            x += len;
            while (isspace(stripMe[x])) x++;
            if (stripMe[x] == ':') x++;
            while (isspace(stripMe[x])) x++;
        }
    }
     
    /**
     * and that should do it !!!
     */
    return stripMe + x;
}

