/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>
#include <qdragobject.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qpopupmenu.h>
#include <string.h>

#include "folderview.h"
#include "inputbox.h"
#include <intl.h>
#include <isubsystem.h>
#include "kiltdown.h"
#include "localized.h"
#include "main.h"
#include "uiicons.h"
#include "userpref.h"


/**
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 * This is the implementation of our folderview widget.  It's very exciting.
 * It shows all the folders available for all the accounts, in a very
 * convenient collapsable listview/treeview format.
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * This class is used to verify folder names.
 *=------------------------------------------------------------------------=
 */
class AccountNameChooser : public InputBox {

  public:
    AccountNameChooser(FolderView *a) : InputBox() { folderView = a; }

  protected:
    void accept();

  private:
    FolderView *folderView;
};


/**
 *=------------------------------------------------------------------------=
 * We have to subclass QListView to supply some support for drag and drop
 *=------------------------------------------------------------------------=
 */
class FolderListView : public QListView {

  public:
    FolderListView(QWidget *parent, const char *name = 0)
        : QListView(parent, name) {
        inDrag = false;
    }

    void dragEnterEvent(QDragEnterEvent *);
    void dragMoveEvent(QDragMoveEvent *);
    void dropEvent(QDropEvent *);
    void dragLeaveEvent(QDragLeaveEvent *);

    inline bool dragging() { return inDrag; }

  private:
    QListViewItem *beforeDrag;
    bool inDrag;
};




/**
 * This is how long the selection has to stay on a folder before it is 
 * selected -- this is so that keyboard navigators can quickly skip over
 * folders
 */
#define SELECTION_TIMEOUT             400
#define FOLDERTIMER_TIMEOUT           100


/**
 *=------------------------------------------------------------------------=
 * FolderView::FolderView
 *=------------------------------------------------------------------------=
 * Constructor.  Doesn't do much.
 *
 * Parameters:
 *      QWidget *           - [in]  our parent object.
 */
FolderView::FolderView(QWidget *parent) : QWidget(parent)
{
    folderListView  = NULL;

    /**
     * Note:  the Qt documentation claims we never have to unallocate this.
     */
    timer = NULL;
    folderTimer = NULL;

    firstSelect = true;
    selectedFolder = NULL;

    accountContext = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::~FolderView
 *=------------------------------------------------------------------------=
 */
FolderView::~FolderView()
{
    // UNDONE: marcw, 9.27.2000 -- i really don't understand the lifetime
    // issues in Qt -- do I clean up the list view, or does it do it 
    // automatically for me ???

    if (accountContext) delete accountContext;

}

/**
 *=------------------------------------------------------------------------=
 * FolderView::accountMenuItemSelected
 *=------------------------------------------------------------------------=
 * an item in the context menu for an account was clicked.
 *
 * Parameters:
 *      int                         - [in]  id of item clicked.
 */
void FolderView::accountMenuItemSelected
(
    int id
)
{
    IAccount *account;

    switch (id) {
        case ID_ACCOUNT_DELETE:
            this->deleteCurrentAccount();
            break;

        case ID_MAKE_DEFAULT_ACCT:
            if (prefDefaultAccount) localFree(prefDefaultAccount);
            account = getCurrentAccount();
            if (account) account->getDisplayName(&prefDefaultAccount);
            break;

        case ID_NEW_FOLDER:
            account = getCurrentAccount();
            if (account) account->newFolder();
            break;

        case ID_RENAME:
            renameCurrentAccount();
            break;
    }
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::addAccount
 *=------------------------------------------------------------------------=
 * Adds the given account to our folderview.  we create a site for it, and
 * get the data we need.
 *
 * Parameters:
 *      IAccount *              - [in]  the new account.
 * 
 * Output:
 *      ERRCODE
 */
ERRCODE FolderView::addAccount
(
    IAccount *addMe
)
{
    AccountManager *mgr = NULL;
    FolderListViewItem *acct;
    IFolder **folders = NULL;
    const char *name = NULL;
    ERRCODE er;
    int count;

    /**
     * Get the printable name for the account and add it.
     */
    er = addMe->getDisplayName(&name);
    RETURN_ON_FAILURE(er);

    acct = new FolderListViewItem(NULL, folderListView, QString(name), 0);
    if (!acct) {
        er = E_OUTOFMEMORY; goto Error;
    }

    /**
     * Create a site and set up the relationship.
     */
    mgr = new AccountManager(addMe, this, acct);
    if (!mgr) {
        er = E_OUTOFMEMORY;
        goto Error;
    }

    addMe->setAccountHost(mgr);

    /**
     * extract the folder names and add them to our treeview
     */
    er = addMe->getFolderCount(&count);
    GOTO_ON_FAILURE(er, Error);

    /**
     * get the list of folders.
     */
    er = addMe->getFolderList(&folders);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Now, for each folder, get it's name and then add it to our list,
     * saving out the IFolder pointer so we can identify it later. Don't 
     * forget to site each object.
     */
    for (int x = 0; x < count; x++) {

        er = addFolder(folders[x], acct);
        GOTO_ON_FAILURE(er, Error);
    }

    folderListView->setOpen(acct, !isCollapsedAccount(name));
    accountList.addElement(mgr);
    er = S_OK;
    
  Error:
    if (name) localFree(name);
    if (folders != NULL) localFree(folders);
    if (FAILED(er)) delete mgr;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::addFolder
 *=------------------------------------------------------------------------=
 * adds the given folder to our UI and stuff.
 *
 * Parameters:
 *      IFolder *                   - [in]  add me.
 *      FolderListViewItem *        - [in]  my parent.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderView::addFolder
(
    IFolder *addMe,
    FolderListViewItem *parent
)
{
    FolderListViewItem *folder;
    QPixmap *icon = NULL;
    const char * name;
    ERRCODE er;

    /**
     * Site the thingie
     */
    addMe->setFolderSite(this);

    /**
     * get its name -- we'll get the unread count later !!!
     */
    er = addMe->getDisplayName(&name);
    RETURN_ON_FAILURE(er);


    /**
     * Get its icon
     */
    er = addMe->getIcon(&icon);
    RETURN_ON_FAILURE(er);

    folder = new FolderListViewItem(addMe, parent, name, -1);
    localFree(name);
    if (!folder) {
        if (icon) delete icon;
        return E_OUTOFMEMORY;
    }
    folder->setPixmap(0, *icon);

    /** 
     * Finally, start the timer to get the count of unread messages
     */
    getFolderTimer()->start(FOLDERTIMER_TIMEOUT, false);

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * FolderView::closing                                        [IFolderSite]
 *=------------------------------------------------------------------------=
 * the given folder is going away.  remove it from our list of goo.
 *
 * Parameters:
 *      IFolder *               - [in]  remove me.
 */
void FolderView::closing
(
    IFolder *folder
)
{
    FolderListViewItem *item;

    findItemForFolder(folder, item);
    if (!item) {
        K_FAIL("Asked to remove a folder that doesn't exist ... ??\n");
        return;
    }

    if (selectedFolder == folder)
        selectedFolder = NULL;
    delete item;
}


/**
 *=------------------------------------------------------------------------=
 * FolderView::deleteCurrentAccount
 *=------------------------------------------------------------------------=
 * deletes the current account.  Obvious makes sure you REALLY REALLY want
 * to do this.
 */
void FolderView::deleteCurrentAccount()
{
    const char *title, *text;
    ISubsystem *sub;
    IAccount *account;
    int ret;

    title = intlLoadString(ID_ARE_YOU_SURE);
    text = intlLoadString(ID_DELETE_ACCOUNT);
    if (!title || !text) {
        if (title) localFree(title);
        if (text) localFree(text);
        return;
    }

    /**
     * Ask them if they want to remove it from the UI.
     */
    ret = QMessageBox::critical(localMainWidget, title, text, QMessageBox::Yes, QMessageBox::No);
    localFree(title);
    localFree(text);

    if (ret != QMessageBox::Yes) return;

    title = intlLoadString(ID_ARE_YOU_SURE);
    text = intlLoadString(ID_DELETE_ACCT_FILES);
    if (!title || !text) {
        if (title) localFree(title);
        if (text) localFree(text);
        return;
    }
    ret = QMessageBox::critical(localMainWidget, title, text, QMessageBox::Yes, QMessageBox::No);
    localFree(title);
    localFree(text);

    /**
     * Okay, go and remove it from the UI.
     */
    account = getCurrentAccount();
    K_ASSERT(account, "How can I Delete an account that doesn't exist??\n");
    if (!account) return;
    account->getHostSubsystem(sub);
    K_ASSERT(sub, "How come an account doesn't have a parent subsystem?\n");
    if (!sub) return;

    sub->removeAccount(account, ret == QMessageBox::Yes);
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::deleteInCurrentFolder
 *=------------------------------------------------------------------------=
 * tells the currently selected folder that the user wants to delete the
 * 'current' stuff.  It is up the folder/account to determine what exactly
 * this is.  
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderView::deleteInCurrentFolder()
{
    IAccount *a = getCurrentAccount();
    if (!a) return S_FALSE;

    return a->deleteSelectedInFolder(selectedFolder);
}


/**
 *=------------------------------------------------------------------------=
 * FolderView::findAccountForItem
 *=------------------------------------------------------------------------=
 * given a list view item, find the account manager (and therefore account)
 * associated with it.
 *
 * Parameters:
 *      FolderListViewItem *        - [in]  item to find.
 *      AccountManager **           - [out] the account manager associated
 *
 * Output:
 *      ERRCODE                     - E_NOTFOUND is bad.
 *
 * Notes:
 *      we search all top level items first, and then look for children to see
 *      if they match if we can't find one the other way.
 */
ERRCODE FolderView::findAccountForItem
(
    FolderListViewItem *item,
    AccountManager **pmgr
)
{
    FolderListViewItem *parent;
    AccountManager *mgr;
    const char *pstr;
    ERRCODE er;


    if (!item || !pmgr) return E_PARAM;
    *pmgr = NULL;

    /**
     * Zip up the chain until we have a parent of null.
     */
    parent = item;
    while (parent->parent() != NULL) {
        QListViewItem *p = parent->parent();
        parent = p ? (FolderListViewItem *)p : NULL;
    }

    /** 
     * Get the name, and match it to an account.
     */
    er = accountList.firstElement(&mgr);
    while (er == S_OK) {
        er = mgr->account->getDisplayName(&pstr);
        RETURN_ON_FAILURE(er);

        if (!strcasecmp(pstr, (const char *)parent->text(0))) {
            localFree(pstr);
            *pmgr = mgr;
            return S_OK;
        }
        localFree(pstr);
        er = accountList.nextElement(&mgr);
    }

    return E_NOTFOUND;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::findItemForFolder
 *=------------------------------------------------------------------------=
 * given the poitner to a folder, find the item for it.
 *
 * Parameters:
 *      IFolder *               - [in]  find me.
 *      FolderListViewItem * &  - [out] what you found.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderView::findItemForFolder
(
    IFolder *findMe,
    FolderListViewItem * &found
)
{
    FolderListViewItem *item, *child = NULL;
    bool gotIt = false;

    /**
     * Whip through the accounts.
     */
    item = (FolderListViewItem *)folderListView->firstChild();
    while (item != NULL && !gotIt)  {
        
        /**
         * Whiz through the child folders and find the right child.
         */
        child = (FolderListViewItem *)item->firstChild();
        while (child != NULL) {
            if (child->getFolder() == findMe) {
                gotIt = true;
                break;
            }
            child = (FolderListViewItem *)child->nextSibling();
        }

        item = (FolderListViewItem *)item->nextSibling();
    }
    
    found = child;

    return found == NULL ? E_NOTFOUND : S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::folderNameChanged                              [IFolderSite]
 *=------------------------------------------------------------------------=
 * the name of the given folder has changed.  update our ui.
 *
 * Parameters:
 *      IFolder *           - [in] 
 *      const char *        - [in]  new name
 */
void FolderView::folderNameChanged
(
    IFolder *folder,
    const char *newName
)
{
    FolderListViewItem *item;
    ERRCODE er;

    er = findItemForFolder(folder, item);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }

    item->setText(0, newName);
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::folderSelectionTimeout
 *=------------------------------------------------------------------------=
 * find the next folder who's unread count we haven't computed and compute
 * that now.
 */
void FolderView::folderSelectionTimeout()
{
    FolderListViewItem *item, *child = NULL;
    bool gotIt = false;
    IFolder *folder;
    int count;

    /**
     * Whip through the accounts.
     */
    item = (FolderListViewItem *)folderListView->firstChild();
    while (item != NULL && !gotIt)  {
        
        /**
         * Whiz through the child folders and find the right child.
         */
        child = (FolderListViewItem *)item->firstChild();
        while (child != NULL) {
            if (child->getUnreadCount() == -1) {
                gotIt = true;
                break;
            }
            child = (FolderListViewItem *)child->nextSibling();
        }

        item = (FolderListViewItem *)item->nextSibling();
    }

    /**
     * exit and kill the timer.
     */
    if (!gotIt) {
        folderTimer->stop();
        return;
    }

    /**
     * got one, get it's unread count !!
     */
    folder = child->getFolder();
    K_ASSERT(folder, "I shouldn't do this for something without a folder!\n");
    if (!folder) return;
    folder->getUnreadCount(&count);
    
    child->setUnreadCount(count);
    child->repaint();    
}


/**
 *=------------------------------------------------------------------------=
 * FolderView::forwardSelected
 *=------------------------------------------------------------------------=
 * tells the currently active folder to forward.
 */
void FolderView::forwardSelected()
{
    if (selectedFolder && selectedFolder->supportsReplyForward())
        selectedFolder->forwardSelected();
}
 
/**
 *=------------------------------------------------------------------------=
 * FolderView::getAccountPopup
 *=------------------------------------------------------------------------=
 * returns the popup for the current account item!!
 *
 * Parameters:
 *      FolderListViewItem *        - [in]  item we need popup for.
 *
 * Output:
 *      QPopupMenu *
 */
QPopupMenu *FolderView::getAccountPopup
(
    FolderListViewItem *item
)
{
    QPopupMenu *popup;
    const char *pstr;

    popup = new QPopupMenu();
    if (!popup) return NULL;
    popup->setCheckable(true);

    pstr = intlLoadString(ID_RENAME);
    if (!pstr) goto OOM;
    popup->insertItem(pstr, this, SLOT(accountMenuItemSelected(int)), 0, ID_RENAME);
    localFree(pstr);

    pstr = intlLoadString(ID_SEARCH);
    if (!pstr) goto OOM;
    popup->insertItem(pstr, this, SLOT(accountMenuItemSelected(int)), 0, ID_SEARCH);
    localFree(pstr);

    popup->insertSeparator();

    pstr = intlLoadString(ID_ACCOUNT_DELETE);
    if (!pstr) goto OOM;
    popup->insertItem(pstr, this, SLOT(accountMenuItemSelected(int)), 0, ID_ACCOUNT_DELETE);
    localFree(pstr);

    pstr = intlLoadString(ID_MAKE_DEFAULT_ACCT);
    if (!pstr) goto OOM;
    popup->insertItem(pstr, this, SLOT(accountMenuItemSelected(int)), 0, ID_MAKE_DEFAULT_ACCT);
    localFree(pstr);
    if (!strncasecmp(item->text(0), prefDefaultAccount, strlen(prefDefaultAccount))) 
        popup->setItemChecked(ID_MAKE_DEFAULT_ACCT, true);

    popup->insertSeparator();

    pstr = intlLoadString(ID_NEW_FOLDER);
    if (!pstr) goto OOM;
    popup->insertItem(pstr, this, SLOT(accountMenuItemSelected(int)), 0, ID_NEW_FOLDER);
    localFree(pstr);

    return popup;

  OOM:
    delete popup;
    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::getAccounts
 *=------------------------------------------------------------------------=
 * returns a list of accounts.
 *
 * Parameters:
 *      LinkList<IAccount *> &          - [out] new list
 */
void FolderView::getAccounts
(
    LinkList<IAccount *> &accounts
)
{
    AccountManager *am;
    ERRCODE er;

    er = accountList.firstElement(&am);
    while (er == S_OK) {
        accounts.addElement(am->account);
        er = accountList.nextElement(&am);
    }
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::getCollapsedAccounts
 *=------------------------------------------------------------------------=
 * returns a list of accounts that are currently collapsed.
 *
 * Output:
 *      const char *                - [out] you must free me.
 */
const char *FolderView::getCollapsedAccounts()
{
    QListViewItem *item = folderListView->firstChild();
    char buf[4096];
    int x = 0;

    while (item) {
        if (!item->isOpen()) {
            strncpy(buf + x, (const char *)item->text(0), sizeof(buf) - x);
            x = strlen(buf);
            if (x < (int)sizeof(buf) - 1) {
                buf[x++] = ':';
                buf[x++] = ':';
                buf[x] = '\0';
            } else
                break;
        }
        item = item->nextSibling();
    }

    /**
     * None collapsed?  e-zed.
     */
    if (x == 0) return NULL;

    return copyString(buf);
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::getFolderPopup
 *=------------------------------------------------------------------------=
 * returns the popup for a folder.
 *
 * Parameters:
 *      FolderListViewItem *                - [in]
 *
 * Output:
 *      QPopupMenu *
 */
QPopupMenu *FolderView::getFolderPopup
(
    FolderListViewItem *item
)
{
    AccountManager *am = NULL;
    QPopupMenu *popup;

    findAccountForItem(item, &am);
    if (!am) {
        fprintf(stderr, "i've got an item i don't have an account for !!!\n");
        return NULL;
    }

    am->account->getFolderPopup(item->getFolder(), popup);
    return popup;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::getFolderTimer
 *=------------------------------------------------------------------------=
 * returns our little timer dude that we will use to track selection in the
 * UI
 *
 * Output:
 *      QTimer *
 */
QTimer *FolderView::getFolderTimer()
{
    if (folderTimer) return folderTimer;

    folderTimer = new QTimer(this);
    if (!folderTimer) return NULL;

    QObject::connect(folderTimer, SIGNAL(timeout()), this, SLOT(folderSelectionTimeout()));
    return folderTimer;
}



/**
 *=------------------------------------------------------------------------=
 * FolderView::getHostInfo
 *=------------------------------------------------------------------------=
 * returns pointers for the mail list and the preview pane edit window.
 *
 * Parameters:
 *      KMailList **            - [out] the mail list pointer.
 *      QMultiLineEdit **       - [out] the preview pane edit box
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderView::getHostInfo
(
    KMailList **pmaillist,
    QMultiLineEdit **pedit
)
{
    if (!pmaillist || !pedit) return E_PARAM;
    K_ASSERT(messageMailList && messagePreview, "These shouldn't ever be NULL!!!");
    *pmaillist = messageMailList;
    *pedit = messagePreview;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * FolderView::getTimer
 *=------------------------------------------------------------------------=
 * returns our little timer dude that we will use to track selection in the
 * UI
 *
 * Output:
 *      QTimer *
 */
QTimer *FolderView::getTimer()
{
    if (timer) return timer;

    timer = new QTimer(this);
    if (!timer) return NULL;

    QObject::connect(timer, SIGNAL(timeout()), this, SLOT(selectionTimeout()));
    return timer;
}


/**
 *=------------------------------------------------------------------------=
 * FolderView::initUI
 *=------------------------------------------------------------------------=
 * initialize our UI and create our list view that we'll use to show our
 * accounts and folders.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderView::initUI()
{
    const char *str;

    /**
     * create our tree view
     */
    folderListView = new FolderListView(this);
    if (!folderListView) return E_OUTOFMEMORY;
    folderListView->setAcceptDrops(true);
    folderListView->setGeometry(0, 0, 100, 100);
    folderListView->setRootIsDecorated(true);
    QObject::connect(folderListView, SIGNAL(selectionChanged(QListViewItem *)), 
                     this, SLOT(selectionChanged(QListViewItem *)));
    QObject::connect(folderListView, SIGNAL(rightButtonClicked(QListViewItem *, const QPoint &, int)),
                     this, SLOT(rightButtonClicked(QListViewItem *, const QPoint &, int)));

    /**
     * set up the title in the tree view -- we'll add the data later when 
     * we get the account information.
     */
    str = intlLoadString(ID_FOLDERVIEW_TITLE);
    folderListView->addColumn(QString(str), -1); // -1 == maximum width
    localFree(str);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::isCollapsedAccount
 *=------------------------------------------------------------------------=
 * indicates whether the account should be collapsed or not.
 *
 * Parameters:
 *      const char *            - [in]  should I be collapsed?
 *
 * Output:
 *      bool                    - true == yes
 */
bool FolderView::isCollapsedAccount
(
    const char *acct
)
{
    int x  = 0;

    if (!prefFolderViewCollapsed) return false;

    while (prefFolderViewCollapsed[x] != '\0') {

        if (!strncasecmp(prefFolderViewCollapsed + x, acct, strlen(acct))) {
            return true;
        } else {

            while (prefFolderViewCollapsed[x] != '\0' 
                   && !(prefFolderViewCollapsed[x] == ':'
                       && prefFolderViewCollapsed[x + 1] == ':')) {
                x++;
            }
            if (prefFolderViewCollapsed[x] == '\0') break;
            x += 2;
        }
    }

    return false;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::mailto
 *=------------------------------------------------------------------------=
 * tells the default account to send mail to the given person.
 *
 * Parameters:
 *      const char *                - [in]  send mail to me!!!
 */
void FolderView::mailto
(
    const char *mailto
)
{
    IAccount *acct;
    ERRCODE er;
     
    /**
     * First, get the default account.
     */
    er = selectAccount(prefDefaultAccount);
    GOTO_ON_FAILURE(er, Error);
    acct = getCurrentAccount();
    if (!acct) {
        return;
    }

    if (acct->isMailAccount()) {
        er = acct->mailto(mailto);
        GOTO_ON_FAILURE(er, Error);
    }

    return;

  Error:
    errorShowError(er);
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::removeAccount
 *=------------------------------------------------------------------------=
 * removes the given account from our UI.
 *
 * Parametesr:
 *      AccountManager *                - [in]
 */
void FolderView::removeAccount
(
    AccountManager *am
)
{
    ERRCODE er;

    /**
     * remove the account from the list of accounts.
     */
    er = accountList.removeElement(am);
    if (er != S_OK) {
        K_FAIL("Asked to remove an account we don't know about !!!!\n");
        return;
    }

    /**
     * Remove it from the UI and clean up!
     */
    this->folderListView->takeItem(am->thisAccount);
    delete am->thisAccount;
    delete am;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::renameCurrentAccount
 *=------------------------------------------------------------------------=
 * get the current account, put up a dialog asking for a new name, and set
 * said account to said name.
 */
void FolderView::renameCurrentAccount()
{
    AccountNameChooser ib(this);
    const char *title, *text;
    const char *pstr;
    IAccount *acct;
    ERRCODE er;
    int ret;

    /**
     * Show a dialog asking for a new account name.
     */
    title = intlLoadString(ID_ACCOUNT_NAME);
    text = intlLoadString(ID_NEW_ACCOUNT_NAME);
    if (!title || !text) {
        if (title) localFree(title);
        if (text) localFree(text);
        return ;
    }

    er = ib.initialize(title, text);
    localFree(title);
    localFree(text);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }

    /**
     * Run da dialog -- it will make sure the name isn't a duplicate
     * or some such thing.
     */
    ret = ib.exec();
    if (ret != QDialog::Accepted) return;
    
    /**
     * Change the name.
     */
    ib.getInput(pstr);
    if (pstr) {
        acct = getCurrentAccount();
        if (acct) acct->renameAccount(pstr);
        localFree(pstr);
    }
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::replySelected
 *=------------------------------------------------------------------------=
 * tells the currently selected folder to reply to the selected messages.
 */
void FolderView::replySelected()
{
    if (selectedFolder && selectedFolder->supportsReplyForward())
        selectedFolder->replySelected(false);
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::replyAllSelected
 *=------------------------------------------------------------------------=
 * tells the currently selected folder to reply to the selected messages.
 */
void FolderView::replyAllSelected()
{
    if (selectedFolder && selectedFolder->supportsReplyForward())
        selectedFolder->replySelected(true);
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::resizeEvent
 *=------------------------------------------------------------------------=
 * we override the resize event here so we can resize our children as 
 * appropriate.  We then resume with normal processing of this.
 *
 * Parameters:
 *      QResizeEvent *          - [in] the event with the information.
 */
void FolderView::resizeEvent
(
    QResizeEvent *event
)
{
    if (event != NULL) {
        folderListView->resize(event->size());
        folderListView->setColumnWidth(0, event->size().width() - 4);
    }
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::rightButtonClicked
 *=------------------------------------------------------------------------=
 * the right button was clicked.  Let's see if there's an item underneat it,
 * and if appropriate, show a context menu.
 *
 * Parameters:
 *      QListViewItem *             - [in]  item
 *      const QPoint &              - [in]  location.
 *      int                         - [in]  column.
 */
void FolderView::rightButtonClicked
(
    QListViewItem *clicked,
    const QPoint &where,
    int column
)
{
    FolderListViewItem *item;
    QPopupMenu *menu;

    if (!clicked) return;

    item = dynamic_cast<FolderListViewItem *>(clicked);
    if (!item) {
        fprintf(stderr, "Somebody gave me an item that isn't a FolderListViewItem !!!\n");
        return;
    }

    if (item->getFolder()) {
        menu = getFolderPopup(item);
    } else {
        menu = getAccountPopup(item);
    }

    if (menu) menu->exec(where);
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::selectAccount
 *=------------------------------------------------------------------------=
 * selects the given account, asking it for what folder to select.
 *
 * Parameters:
 *      const char *            - [in]  folder to select.
 *      bool                    - [in]  OPTIONAL force selection now!!
 *
 * Output:
 *      ERRCODE
 */
ERRCODE FolderView::selectAccount
(
    const char *name,
    bool forceSelection
)
{
    const char *pstr, *pstr2;
    QListViewItem *item;
    AccountManager *mgr;
    ERRCODE er;

    if (!name) return E_PARAM;
    if (!folderListView) return E_INTERNAL;

    /**
     * Whip through the accounts looking for this one.
     */
    item = folderListView->firstChild();
    while (item != NULL) {
        if (!strcasecmp((const char *)item->text(0), name))
            break;
        item = item->nextSibling();
    }
    
    if (item == NULL) return S_NOTFOUND;


    item->setOpen(true);

    /**
     * Found it -- select it's default folder.
     */
    er = accountList.firstElement(&mgr);
    while (er == S_OK) {
        er = mgr->account->getDisplayName(&pstr);
        RETURN_ON_FAILURE(er);
        if (!strcasecmp(pstr, name)) {
            localFree(pstr);
            mgr->account->getDefaultFolder(&pstr2);
            break;
        }
        localFree(pstr);
        er = accountList.nextElement(&mgr);
    }

    if (pstr2 != NULL) {
        item = item->firstChild();
        while (item != NULL) {
            if (!strcasecmp(item->text(0), pstr2)) {
                folderListView->setSelected(item, true);
                break;
            }
            item = item->nextSibling();
        }
        localFree(pstr2);
    }

    if (forceSelection) selectionChanged(item);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::selectionChanged
 *=------------------------------------------------------------------------=
 * this method is called whenever an item in the list view is selected.
 *
 * Parameters:
 *      QListViewItem *         - [in]  the newly selected item.
 */
void FolderView::selectionChanged
(
    QListViewItem *newItem
)
{
    AccountManager *mgr;
    QTimer *time;
    ERRCODE er;

    /**
     * IF we're dragging, then selection will change all over the place!
     */
    if (folderListView->dragging()) return;

    er = findAccountForItem((FolderListViewItem *)newItem, &mgr);
    if (!FAILED(er)) {
        currentAccount = mgr;
    }

    /**
     * Now, set a timer, and after SELECTION_TIMEOUT seconds, we'll go and
     * actually select that folder -- we don't do it right away, in case 
     * they're just passing through this folder on the way up one ...
     * Any existing timers are cancelled.
     */
    if (firstSelect) {
        selectionTimeout();
        firstSelect = false;
    } else {
        time = getTimer();
        if (time) time->start(SELECTION_TIMEOUT, TRUE);
    }
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::selectionTiemout
 *=------------------------------------------------------------------------=
 * we have time out appropriately since the user last selected an item in
 * the list view.
 */
void FolderView::selectionTimeout()
{
    FolderListViewItem *sel;
    QListViewItem *item;

    if (currentAccount != NULL) {
        item = folderListView->selectedItem();
        if (item)
            sel = (FolderListViewItem *)item;
        else
            return;

        /**
         * Don't do anything if the new folder is the same as the current
         * one.
         */
        if (sel->getFolder() == selectedFolder) return;

        /**
         * Notify the old one that it's going to lose the focus and give the
         * focus to the new one!!
         */
        if (selectedFolder) selectedFolder->loseFocus();
        selectedFolder = sel->getFolder();

        if (selectedFolder)  {
            selectedFolder->gainFocus();
            localMainWidget->enableReplyForward(selectedFolder->supportsReplyForward());
        } else {
            messageMailList->nukeAll();
            localMainWidget->enableReplyForward(false);
        }
    }
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::setPeerInfo
 *=------------------------------------------------------------------------=
 * Sets the peer mail list and preview controls
 *
 * Parameters:
 *      KMailList *             - [in]  the new mail list.
 *      QMultiLineEdit *        - [in]  the preview edit box
 */
void FolderView::setPeerInfo
(
    KMailList *newMailList,
    QMultiLineEdit *newMessagePreview
)
{
    this->messageMailList = newMailList;
    this->messagePreview = newMessagePreview;
}

/**
 *=------------------------------------------------------------------------=
 * FolderView::terminate
 *=------------------------------------------------------------------------=
 * we be going now.
 */
void FolderView::terminate()
{
    if (prefFolderViewCollapsed) localFree(prefFolderViewCollapsed);
    prefFolderViewCollapsed = getCollapsedAccounts();
}


/**
 *=------------------------------------------------------------------------=
 * FolderView::unreadMessageCountChange                       [IFolderSite]
 *=------------------------------------------------------------------------=
 * the number of unread messages in a given folder has changed.  We need to
 * go and update that now.
 *
 * Parameters:
 *      IFolder *               - [in]  the folder whose number has changed.
 *      int                     - [in]  the new number of unread messages.
 */
void FolderView::unreadMessageCountChange
(
    IFolder *folder,
    int cunread
)
{
    FolderListViewItem *item;

    findItemForFolder(folder, item);

    if (item == NULL ) {
        K_FAIL("No item for this folder !!!! \n");
        return;
    }

    item->setUnreadCount(cunread);
    item->repaint();
}

/**
 *=------------------------------------------------------------------------=
 *                              AccountManager                             *
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * AccountManager::AccountManager
 *=------------------------------------------------------------------------=
 * Constructor
 *
 * Parameters:
 *      IAccount *                  - [in]  this acct.
 *      FolderView *                - [in]  host folderview.
 *      FolderListViewItem *        - [in]  item in the list view.

 */
AccountManager::AccountManager
(
    IAccount *acct,
    FolderView *view,
    FolderListViewItem *parent
)
{
    thisAccount = parent;
    folderView =view;
    account = acct;
}

/**
 *=------------------------------------------------------------------------=
 * AccountManager::~AccountManager
 *=------------------------------------------------------------------------=
 * destructor.  very very exciting.
 */
AccountManager::~AccountManager()
{

}

/**
 *=------------------------------------------------------------------------=
 * AccountManager::accountNameChanged                        [IAccountHost]
 *=------------------------------------------------------------------------=
 * for some reason or other, the name of this account has changed.  update
 * all our UI now ...
 *
 * Parameters:
 *      const char *                - [in]  new account name.
 */
void AccountManager::accountNameChanged
(
    const char *newName
)
{
    this->thisAccount->setText(0, newName);
}

/**
 *=------------------------------------------------------------------------=
 * AccountManager::addFolder                                 [IAccountHost]
 *=------------------------------------------------------------------------=
 * the client wants to add a folder.  Go do so now.
 *
 * Parameters:
 *      IFolder *               - [in]  add me please!!!*
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountManager::addFolder
(
    IFolder *addMe
)
{
    return this->folderView->addFolder(addMe, thisAccount);
}

/**
 *=------------------------------------------------------------------------=
 * AccountManager::closing                                   [IAccountHost]
 *=------------------------------------------------------------------------=
 * the account is going away.  clean up it's ui.
 */
void AccountManager::closing()
{
    this->folderView->removeAccount(this);
}

/**
 *=------------------------------------------------------------------------=
 * AccountManager::getCurrentFolder                          [IAccountHost]
 *=------------------------------------------------------------------------=
 * returns the currently selected folder
 *
 * Output:
 *      IFolder *
 */
IFolder *AccountManager::getCurrentFolder()
{
    FolderListViewItem *item;
    QListViewItem *i;

    i = this->folderView->folderListView->selectedItem();
    if (!i) return NULL;

    item = dynamic_cast<FolderListViewItem *>(i);
    if (!item) return NULL;

    return item->getFolder();
}

/**
 *=------------------------------------------------------------------------=
 * AccountManager::setStatusText                             [IAccountHost]
 *=------------------------------------------------------------------------=
 * lets the account set some status bar text
 *
 * Parameters:
 *      const char *                - [in]
 */
void AccountManager::setStatusText
(
    const char *text
)
{
    localMainWidget->setStatusBarText(text);
}




/**
 *=------------------------------------------------------------------------=
 *                          FolderListViewItem                             *
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * FolderListViewItem::paintCell
 *=------------------------------------------------------------------------=
 * paints a cell -- we need to re-implement this because we do bold items,
 * and will update the count of unread messages, etc ...
 *
 * Parameters:
 *      QPainter *                  - [in]  how to paint.
 *      const QColorGroup &         - [in]  colors to paint with
 *      int                         - [in]  which column to paint.
 *      int                         - [in]  width of the column
 *      int                         - [in]  how to align contents
 */
void FolderListViewItem::paintCell
(
    QPainter *p,
    const QColorGroup & cg,
    int column,
    int width,
    int align
)
{
    QFont fontSave, ft;
    const char *pstr;
    QPen penSave;
    char buf[96];

    /**
     * If there are no unread messages, then we're not going to do anything
     * here at all !!!
     */
    if (unreadCount == 0 || unreadCount == -1) {
        QListViewItem::paintCell(p, cg, column, width, align);
        return;
    }

    K_ASSERT(this->ifolder, "I most certainly should have a folder !!!!\n");

    /**
     * Otherwise, get to work.  First, background:
     */
    if (isSelected()) {

        p->fillRect(0, 0, width , height(), QBrush(cg.highlight()));
        penSave = p->pen();
        p->setPen(QPen(cg.highlightedText()));

    } else {

        p->fillRect(0, 0, width, height(), QBrush(cg.base()));

    }

    /**
     * Get and print the icon ...
     */
    p->drawPixmap(1, 1, *pixmap(0));

    /**
     * Generate the text for the entry
     */
    ifolder->getDisplayName(&pstr);
    snprintf(buf, sizeof(buf), "%s (%d)", pstr, this->unreadCount);
    localFree(pstr);

    /**
     * Output the text with the correct font.
     */
    ft = p->font();
    fontSave = ft;
    ft.setBold(true);
    p->setFont(ft);

    p->drawText(pixmap(0)->width() + 1, 1, width - pixmap(0)->width() - 1, height(),
                Qt::AlignLeft, buf);


    /**
     * Clean Up !!!
     */
    if (isSelected()) {
        p->setPen(penSave);
    }
    p->setFont(fontSave);
}






/**
 *=------------------------------------------------------------------------=
 *                          AccountNameChooser                              *
 *=------------------------------------------------------------------------=
 */

/**
 *=------------------------------------------------------------------------=
 * AccountNameChooser::accept
 *=------------------------------------------------------------------------=
 * verifies that the given name is a valid one, and doesn't represent the 
 * name of an already existing folder...
 */
void AccountNameChooser::accept()
{
    const char *pstr = (this->edit != NULL) ? (const char *)edit->text() : NULL;
    LinkList<IAccount *> accounts;
    const char *aname;
    bool ok = false;
    IAccount *a;
    ERRCODE er;
    int x = 0;

    folderView->getAccounts(accounts);

    /** 
     * IF there's no string, or the string doesn't have a letter or number
     * in it, then it's not okay.  It's also bad if there's a folder with
     * the same name already
     */
    if (!pstr) {
        ok = false;
    } else {

        x = 0;
        while (pstr[x] != '\0') {
            if (isalnum(pstr[x])) {
                ok = true;
                break;
            }
            x++;
        }

        /**
         * Now, make sure we don't have this folder already ...
         */
        if (ok) {
            er = accounts.firstElement(&a);
            while (er == S_OK) {
                a->getDisplayName(&aname);
                if (aname && !strcasecmp(aname, pstr)) {
                    ok = false;
                    localFree(aname);
                    break;
                }
                localFree(aname);

                er = accounts.nextElement(&a);
            }
        }
    }

    /**
     * IF it's not okay, show a dialog and abort.
     */
    if (!ok) {
        errorShowError(E_INVALID_NAME);
        return;
    }
    InputBox::accept();
}





/**
 *=------------------------------------------------------------------------=
 *                          FolderListView
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * FolderListView::dragEnterEvent
 *=------------------------------------------------------------------------=
 * determines whether we accept the incoming type or not.
 *
 * Parameters:
 *      QDragEnterEvent *               - [in]  info
 */
void FolderListView::dragEnterEvent
(
    QDragEnterEvent *event
)
{
    beforeDrag = selectedItem();
    inDrag = true;
    event->accept(true);
}

/**
 *=------------------------------------------------------------------------=
 * FolderListView::dragLeaveEvent
 *=------------------------------------------------------------------------=
 * they left without dropping.  Reset the selected item.
 *
 * Parameters:
 *      QDragLeaveEvent *               - [in]  details.
 */
void FolderListView::dragLeaveEvent
(
    QDragLeaveEvent *event
)
{
    inDrag = false;
    setSelected(beforeDrag, true);
    beforeDrag = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * FolderListView::dragMoveEvent
 *=------------------------------------------------------------------------=
 * the user is moving around in a drag-drop operation.  Figure out if the 
 * folder underneath the current item will take it.
 *
 * Parameters:
 *      QDragMoveEvent *                - [in]  details.
 */
void FolderListView::dragMoveEvent
(
    QDragMoveEvent *event
)
{
    FolderListViewItem *item;
    QListViewItem *i;
    QWidget *vp;
    QPoint p;

    /**
     * Figure out what item is underneath the current drag.
     */
    vp = viewport();    
    p.setX(event->pos().x() - vp->x());
    p.setY(event->pos().y() - vp->y());
    i = itemAt(p);
    if (!i) return;
    item = dynamic_cast<FolderListViewItem *>(i);
    if (!item) {
        K_FAIL("I've got a listview item that i didn't create !!!\n");
        return;
    }

    /**
     * If the item is a folder, see if it accepts the given drop type,
     * otherwise, we don't accept it.
     */
    if (!item->getFolder()) {
        event->ignore();
        return;
    } else {
        if (item->getFolder()->acceptsFormat(event->format()) == S_OK) {
            setSelected(item, true);
            event->accept();
        } else {
            event->ignore();
        }
    }
}

/**
 *=------------------------------------------------------------------------=
 * FolderListView::dropEvent
 *=------------------------------------------------------------------------=
 * accept the drop and figure out what needs to happen now !!!
 *
 * Parameters:
 *      QDropEvent *                    - [in]  info
 */
void FolderListView::dropEvent
(
    QDropEvent *event
)
{
    FolderListViewItem *item;
    QListViewItem *i;
    QWidget *vp;
    QPoint p;

    /**
     * Figure out what item is underneath the current drag.
     */
    vp = viewport();    
    p.setX(event->pos().x() - vp->x());
    p.setY(event->pos().y() - vp->y());
    i = itemAt(p);
    if (!i) return;
    item = dynamic_cast<FolderListViewItem *>(i);
    if (!item) {
        K_FAIL("I've got a listview item that i didn't create !!!\n");
        return;
    }

    /**
     * If the item is a folder, see if it accepts the given drop type,
     * otherwise, we don't accept it.
     */
    if (!item->getFolder()) return;

    item->getFolder()->acceptData(event->encodedData(event->format()).data(),
                                  event->encodedData(event->format()).size());

    inDrag = false;
    setSelected(beforeDrag, true);
    beforeDrag = NULL;
}


