/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
n *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _FOLDERVIEW_H_
#define _FOLDERVIEW_H_

#include <sanity.h>

#include <qevent.h>
#include <qlistview.h>
#include <qtimer.h>

#include <iaccount.h>
#include <ifolder.h>
#include <llist.h>

class FolderView;
class QPopupMenu;


/**
 * We need to implement our own version of this for some drag and drop
 * purposes.
 */
class FolderListView;

/**
 *=------------------------------------------------------------------------=
 * We need to add some custom data to this item.
 *=------------------------------------------------------------------------=
 */
class FolderListViewItem : public QListViewItem {

  public:
    FolderListViewItem(IFolder *f, QListView *parent, QString text,
                       int unread)
        : QListViewItem(parent, text), ifolder(f) { unreadCount = unread; }
    FolderListViewItem(IFolder *f, QListViewItem *parent, QString text,
                       int unread)
        : QListViewItem(parent, text), ifolder(f) { unreadCount = unread; }

    /**
     * Returns the IFolder interface that this item is holding on to.
     */
    IFolder *getFolder() {
        return ifolder;
    }

    int  getUnreadCount() { return this->unreadCount; }
    void setUnreadCount(int uc) { this->unreadCount = uc; }

  protected:
    virtual void paintCell(QPainter *p, const QColorGroup &cg, int column,
                           int width, int align);

  private:
    IFolder *ifolder;
    int unreadCount;
};


/**
 *=------------------------------------------------------------------------=
 * This class is a simple little site object for hosting accounts. Each 
 * account needs it's own site object, so we'll maintain a little list of
 * these
 *=------------------------------------------------------------------------=
 */
class AccountManager : public IAccountHost {

  public:
    /**
     * Constructor
     */
    AccountManager(IAccount *, FolderView *, FolderListViewItem *);

    /**
     * IAccountHost methods
     */
    virtual void    accountNameChanged(const char *);
    virtual ERRCODE addFolder(IFolder *);
    virtual void    closing();
    virtual IFolder *getCurrentFolder();
    virtual void    setStatusText(const char *);

    virtual ~AccountManager();

    /**
     * Our account.
     */
    IAccount *account;
    FolderListViewItem *thisAccount;

  private:
    FolderView *folderView;
};


/**
 *=------------------------------------------------------------------------=
 *                               FolderView
 *=------------------------------------------------------------------------=
 * This defines our folderview widget.  It's very exciting.  It shows all
 * the folders available for all the accounts, in a very convenient
 * collapsable listview/treeview format.
 *=------------------------------------------------------------------------=
 */
class FolderView : public QWidget, public IFolderSite {

    /**
     * All objects using Qt extensions such as slots and signals need this.
     */
    Q_OBJECT
    friend AccountManager;

  public:
    FolderView(QWidget *);
    ~FolderView();

    /**
     * Initialize our UI and make sure we're coo'
     */
    ERRCODE initUI();

    /**
     * We're going away.  make peace with the gos.
     */
    void terminate();

    /**
     * Adds an account to our folder view.
     */
    ERRCODE addAccount(IAccount *);

    /**
     * Deletes the 'selected' items in the currently selected folder.
     */
    ERRCODE deleteInCurrentFolder();

    /**
     * returns our current account.
     */
    inline IAccount *getCurrentAccount() {
        return currentAccount ? currentAccount->account : NULL;
    }

    /**
     * Returns the currently active folder.
     */
    inline IFolder *getCurrentFolder() {
        return selectedFolder;
    }

    /**
     * Deletes the current account.  Dialogs galore will pop up confirming
     * this !!!
     */
    void deleteCurrentAccount();

    /**
     * send mail to the given person .... basically -- use the default
     * account and tell them to send mail to the given person.
     */
    void mailto(const char *mailto);

    /**
     * Puts up a dialog and renames the current account.  Fweee!
     */
    void renameCurrentAccount();

    /** 
     * Selects the given account.
     */
    ERRCODE selectAccount(const char *accountName, bool force = false);

    /**
     * Sets the mail list accounts should work with.
     */
    virtual void setPeerInfo(KMailList *, QMultiLineEdit *);

    /**
     * IFolderSite methods 
     */
    virtual ERRCODE getHostInfo(KMailList **, QMultiLineEdit **);
    virtual void unreadMessageCountChange(IFolder *, int);
    virtual void folderNameChanged(IFolder *, const char *);
    virtual void closing(IFolder *);

    /**
     * Returns a list of accounts.
     */
    virtual void getAccounts(LinkList<IAccount *> & accounts);

    /**
     * toolbar interaction.
     */
    void forwardSelected();
    void replySelected();
    void replyAllSelected();


  protected:
    /**
     * We need to override the resize event so we can set our children's sizes
     * correctly
     */
    virtual void resizeEvent(QResizeEvent *);

  private:
    LinkList<AccountManager *> accountList;
    AccountManager *currentAccount;
    QMultiLineEdit *messagePreview;
    QTimer *timer, *folderTimer;
    QPopupMenu *accountContext;
    KMailList *messageMailList;
    FolderListView *folderListView;
    IFolder *selectedFolder;
    bool firstSelect;

    /**
     * Finds the account associated with this list item.
     */
    ERRCODE findAccountForItem(FolderListViewItem *item, AccountManager **);

    /**
     * Finds the FolderListViewItem for the given IFolder pointer.
     */
    ERRCODE findItemForFolder(IFolder *findMe, FolderListViewItem * &item);

    /**
     * Adds the given folder to the UI.
     */
    ERRCODE addFolder(IFolder *, FolderListViewItem *parent);

    /**
     * Remove the given account from our UI
     */
    void removeAccount(AccountManager *);

    /**
     * For remembering collapsed folders.
     */
    const char *getCollapsedAccounts();
    bool isCollapsedAccount(const char *);

    /**
     * Popup menu APIs
     */
    QPopupMenu *getAccountPopup(FolderListViewItem *);
    QPopupMenu *getFolderPopup(FolderListViewItem *);

    QTimer *getTimer();
    QTimer *getFolderTimer();

  private slots:
    void selectionChanged(QListViewItem *newItem);
    void selectionTimeout();
    void folderSelectionTimeout();
    void rightButtonClicked(QListViewItem *, const QPoint &, int);
    void accountMenuItemSelected(int);
};


#endif // _FOLDERVIEW_H_



