/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <sanity.h>


#include <qlineedit.h>
#include <qhbox.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qvbox.h>

#include "inputbox.h"
#include <intl.h>
#include "localized.h"


/**
 *=------------------------------------------------------------------------=
 *                          InputBox                                       *
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * InputBox::InputBox
 *=------------------------------------------------------------------------=
 * initializes our dialog, but not the UI stuff ..
 */
InputBox::InputBox()
: QDialog(0,0, true, WStyle_NoBorder)
{
    edit = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * InputBox::~InputBox
 *=------------------------------------------------------------------------=
 * cleans up our dialog, but not the UI stuff ..
 */
InputBox::~InputBox()
{
    if (edit) delete edit;
//    if (vbox) delete vbox;
    if (hbox) delete hbox;
//    if (ok) delete ok;
//    if (cancel) delete cancel;
}

/**
 *=------------------------------------------------------------------------=
 * InputBox::getInput
 *=------------------------------------------------------------------------=
 * returns what the user typed in.
 *
 * Parameters:
 *      const char * &          - [out]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE InputBox::getInput
(
    const char * & pstr
)
{
    if (!edit) return E_PARAM;

    const char *p = edit->text();

    pstr = copyString(p);
    return pstr ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * InputBox::initialize
 *=------------------------------------------------------------------------=
 * initializes our UI
 *
 * Parameters:
 *      const char *            - [in]  title
 *      const char *            - [in]  text
 *
 * Output:
 *      ERRCODE
 */
ERRCODE InputBox::initialize
(
    const char *title,
    const char *text
)
{
    const char *pstr;

    /**
     * Create a vbox to flow things vertically.
     */
    vbox = new QVBox(this);
    if (!vbox) return E_OUTOFMEMORY;
    vbox->setSpacing(5);
    vbox->setMargin(5);

    /**
     * Stuff in a label and an edit box
     */
    label = new QLabel(vbox);
    if (!label) return E_OUTOFMEMORY;
    label->setText(text);
    label->setAlignment(AlignCenter);
    label->adjustSize();

    edit = new QLineEdit(vbox);
    if (!edit) return E_OUTOFMEMORY;

    /**
     * Now create a horizontally flowed box to hold the two command buttons.
     */
    hbox = new QHBox(vbox);
    if (!hbox) return E_OUTOFMEMORY;
    hbox->setSpacing(15);
    hbox->setMargin(5);

    ok = new QPushButton(hbox);
    pstr = intlLoadString(ID_OK_TEXT);
    if (!pstr) return E_OUTOFMEMORY;
    ok->setText(pstr);
    ok->setDefault(true);
    localFree(pstr);

    cancel = new QPushButton(hbox);
    pstr = intlLoadString(ID_CANCEL_TEXT);
    if (!pstr) return E_OUTOFMEMORY;
    cancel->setText(pstr);
    localFree(pstr);

    /**
     * Hook up signals
     */
    QObject::connect(ok, SIGNAL(clicked()), this, SLOT(accept()));
    QObject::connect(cancel, SIGNAL(clicked()), this, SLOT(reject()));

    /**
     * finally, set up some of our own properties, some size, and focus
     */
    this->setCaption(title);
    vbox->setGeometry(0, 0, label->width() + 10, 120);
    this->setBaseSize(QSize(300, 100));
    this->edit->setFocus();

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * InputBox::resizeEvent
 *=------------------------------------------------------------------------=
 * we need to know when our size has changed so we can resize the components
 * as appropriate
 *
 * Parameters:
 *      QResizeEvent *              - [in] our new size.
 */
void InputBox::resizeEvent
(
    QResizeEvent *event
)
{
    if (!event) return;
    QSize size = event->size();

    vbox->setGeometry(0, 0, size.width(), size.height());
}



