/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <errno.h>
#include <fcntl.h>
#include <qthread.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/types.h>
#include <unistd.h>

#include "instance.h"
#include "kiltdown.h"
#include "main.h"



static const char *lockDir = ".kiltdownLock/";
static const char *fifoName = ".kiltdownFifo";

static const char *justShow = "justShow";
static const char *showMailto = "showMailTo:";
static const char *finishedData = "xxxx123123";


/**
 *=------------------------------------------------------------------------=
 *                              BUG BUG BUG
 *=------------------------------------------------------------------------=
 * FreeBSD appears to have a bug/problem by which open(2) will block on a FIFO
 * and block all other threads in the process.  This is, of course, extremely
 * bad.  We thus do some spinning and other crazy things to try and get it to
 * not lock up the whole system.  On other systems, we do not try to do this.
 * However, on those systems, we will do a couple of usleep() calls to let
 * other threads close things down before we try to read from them again, etc.
 *
 * NOTE: This might very well be a NetBSD/OpenBSD problem as well !!!!!
 */
#ifdef BROKEN_FIFO_OPEN

#define BROKENFIFO_READ_SLEEP         500000            // half a second.
#define CLOSE_WRITE_SLEEP             600000            // READ_SLEEP + .1 sec

#else  // BROKEN_FIFO_OPEN

#define CLOSE_WRITE_SLEEP             300000            // .1s on all other OS

#endif // BROKEN_FIFO_OPEN



/**
 *=------------------------------------------------------------------------=
 *                           Local Prototypes
 *=------------------------------------------------------------------------=
 */
ERRCODE createInstanceFifo();
void    sendFifoData(const char *);

/**
 *=------------------------------------------------------------------------=
 *                             KFifoThread
 *=------------------------------------------------------------------------=
 */
class KFifoThread : public QThread {
  protected:
    void run();
};


/**
 * Unfortunately, we have to keep one of these running so we can block on
 * open(2) calls and listen for other instances ...
 */
KFifoThread instanceThread;



/**
 *=------------------------------------------------------------------------=
 * instanceCheckInstance
 *=------------------------------------------------------------------------=
 * we do not want to have more than one instance of this program running
 * at a time.  This routine looks for the FIFO that we set up and writes
 * to it if it exists, or goes and creates it.
 *
 * Parameters:
 *      const char *                - [in]  person to mail to if we're not
 *                                      just launching the whole big thin.
 *
 * Note:
 *      O_EXCL is broken for NFS filesystems, so we can't use it.  Instead,
 *      we'll create a directory, which definitely will not succeed if it
 *      already exists, and then just put the named pipe (FIFO) we'll use
 *      for communications in that directory.
 */
void instanceCheckInstance
(
    const char *mailto
)
{
    char buf[MAXPATHLEN];
    ERRCODE er;
    int res;

    /**
     * First, construct the filename for the lockdir
     */
    er = Kiltdown::getKiltdownDirectory(buf, sizeof(buf));
    GOTO_ON_FAILURE(er, Error);
    strncat(buf, lockDir, sizeof(buf));

    /**
     * Look for the lock.
     */
    res = mkdir(buf, 0700);

    /**
     * If it already exists, then we'll try and open the FIFO.
     */
    if (res < 0 && errno == EEXIST) {
        sendFifoData(mailto);
    } else if (res < 0) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_FILE_ERR;
        goto Error;
    }

    /**
     * we're the first process, go and create the fifo, fork on it,
     * and then return.
     */
    er = createInstanceFifo();

  Error:
    if (FAILED(er)) {
        errorShowError(er);
        exit(-1);
    }
}

/**
 *=------------------------------------------------------------------------=
 * instanceClearInstance
 *=------------------------------------------------------------------------=
 * if we've got a thread that's just sitting there waiting for an open, then
 * we need to shut it down.
 */
void instanceClearInstance()
{
    char buf[MAXPATHLEN];
    ERRCODE er;
    int fifo;
    int res;

    /**
     * get the fifo path.
     */
    er = Kiltdown::getKiltdownDirectory(buf, sizeof(buf));
    if (FAILED(er)) return;
    strncat(buf, lockDir, sizeof(buf));
    strncat(buf, fifoName, sizeof(buf));

    /**
     * Open the FIFO for writing.
     */
    fifo = open(buf, O_WRONLY);
    if (fifo == -1) {
        perror("open(2) on a FIFO: ");
        fprintf(stderr, "Fatal Error:  please report a bug\n");
        return;
    }

    /**
     * write out our termination sequence.
     */
    res = write(fifo, finishedData, strlen(finishedData) + 1);
    if (res == -1) {
        perror("write(2) on a FIFO");
        fprintf(stderr, "Fatal Error:  please report a bug\n");
        return;
    }

    /**
     * We need to wait here until the other thread gets a chance to read
     * the data.  Otherwise, for some reason, the data can be lost if we
     * close the pipe too quickly ... ???
     */
    close(fifo);
    usleep(CLOSE_WRITE_SLEEP);
    instanceNukeFifo();
}

/**
 *=------------------------------------------------------------------------=
 * instanceNukeFifo
 *=------------------------------------------------------------------------=
 * clears out the fifo file -- it's important that we get rid of it asap.
 */
void instanceNukeFifo()
{
    char buf[MAXPATHLEN];
    ERRCODE er;
    int sv;

    /**
     * Get the FIFO name.
     */
    er = Kiltdown::getKiltdownDirectory(buf, sizeof(buf));
    if (FAILED(er)) return;
    strncat(buf, lockDir, sizeof(buf));
    sv = strlen(buf);
    strncat(buf, fifoName, sizeof(buf));

    /**
     * Remove it and the directory.
     */
    unlink(buf);
    buf[sv] = '\0';
    rmdir(buf);

}






/**
 *=------------------------------------------------------------------------=
 *                           Helper routines
 *=------------------------------------------------------------------------=
 */

/**
 *=------------------------------------------------------------------------=
 * createInstanceFifo
 *=------------------------------------------------------------------------=
 * Creates the FIFO that we'll use to listen for other instances coming
 * along to us.  We will actually spawn a thread that will then sit there
 * and do blocking open()s on the FIFO listening for other instances ...
 */
ERRCODE createInstanceFifo()
{
    char buf[MAXPATHLEN];
    ERRCODE er;
    int res;

    /**
     * First thing to do is go and create the FIFO.
     */
    er = Kiltdown::getKiltdownDirectory(buf, sizeof(buf));
    RETURN_ON_FAILURE(er);
    strncat(buf, lockDir, sizeof(buf));
    strncat(buf, fifoName, sizeof(buf));

    res = mkfifo(buf, 0700);
    K_ASSERT(res == 0 || (res == -1 && errno == EEXIST), "How on earth does this exist yet ???\n");
    if (res < 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    /**
     * Okay, got the fifo.  Now we'll fork a thread to go and block on open(2)
     * calls on this FIFO.
     */
    instanceThread.start();
    return S_OK;
}
 
/**
 *=------------------------------------------------------------------------=
 * sendFifoData
 *=------------------------------------------------------------------------=
 * sends data to the fifo telling them to show the main window or open a
 * new message for us.
 *
 * Parameters:
 *      const char *                - [in]  mailto string.
 */
void sendFifoData
(
    const char * mailto
)
{
    char buf[MAXPATHLEN];
    ERRCODE er;
    int fifo;
    int res;

    /**
     * First thing to do is go and create the FIFO.
     */
    er = Kiltdown::getKiltdownDirectory(buf, sizeof(buf));
    if (FAILED(er)) return;
    strncat(buf, lockDir, sizeof(buf));
    strncat(buf, fifoName, sizeof(buf));

    /**
     * Open the FIFO for writing.
     */
    fifo = open(buf, O_WRONLY);
    if (fifo == -1) {
        perror("open(2) on a FIFO: ");
        fprintf(stderr, "Fatal Error:  please report a bug\n");
        return;
    }

    /**
     * write out our termination sequence.
     */
    if (!mailto) {
        res = write(fifo, justShow, strlen(justShow) + 1);
        if (res == -1) {
            perror("write(2) on a FIFO");
            fprintf(stderr, "Fatal Error:  please report a bug\n");
        }
    } else {
        snprintf(buf, sizeof(buf), "%s%s", showMailto, mailto);
        res = write(fifo, buf, strlen(buf) + 1);
        if (res == -1) {
            perror("write(2) on a FIFO");
            fprintf(stderr, "Fatal Error:  please report a bug\n");
        }
    }

    close(fifo);
    exit(-1);
}




/**
 *=------------------------------------------------------------------------=
 *                              KFifoThread
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * KFifoThread::run
 *=------------------------------------------------------------------------=
 * runs in a separate thread, and loops, waiting for processes to open the
 * pipe and send us some info.
 */
void KFifoThread::run()
{
    char buf[MAXPATHLEN];
    char fifoBuf[128];
    ERRCODE er;
    int fifo;
    int res;
    int sv;

    /**
     * Get the FIFO name.
     */
    er = Kiltdown::getKiltdownDirectory(buf, sizeof(buf));
    if (FAILED(er)) return;
    strncat(buf, lockDir, sizeof(buf));
    sv = strlen(buf);
    strncat(buf, fifoName, sizeof(buf));

    /**
     * Loop, trying to open the FIFO.
     */
    while (1) {

#ifndef BROKEN_FIFO_OPEN

        fifo = open(buf, O_RDONLY);

#else // BROKEN_FIFO_OPEN
        
        /**
         * Please see the comment at the top of the file whey BROKEN_FIFO_OPEN
         * is defined -- basically, FreeBSD will block the whole dang
         * process here unless we specify O_NONBLOCK.  This causes us to have
         * to do read(2)s in a loop until somebody opens the other end of the
         * pipe.  We'll do a usleep() in the loop so we don't burn off CPU
         * cycles like mad !!!!
         */
        fifo = open(buf, O_RDONLY | O_NONBLOCK);
usleep(CLOSE_WRITE_SLEEP);
#endif // BROKEN_FIFO_OPEN

        if (fifo == -1) {
            perror("open(2) on a FIFO:");
            fprintf(stderr, "Fatal Error:  Can't continue.  Please file a bug\n");
            break;
        }

        /**
         * Hey! we're alive.  Go and read from the pipe.
         */
#ifndef BROKEN_FIFO_OPEN

        res = read(fifo, fifoBuf, sizeof(fifoBuf));

#else // BROKEN_FIFO_OPEN

        do {
            errno = 0;
            res = read(fifo, fifoBuf, sizeof(fifoBuf) - 1);
            if (res == 0 && errno == 0) {
                /**
                 * Sleep a bit and try again.
                 */
                usleep(BROKENFIFO_READ_SLEEP);
            }
        } while (res == 0);

#endif // BROKEN_FIFO_OPEN

        if (res == -1) {
            perror("read(2) on a FIFO:");
            fprintf(stderr, "Fatal error: Can't continue:  Please file a bug\n");
            break;
        }

        /**
         * If they asked us to terminate, then do so now.
         */
        if (!strcasecmp(fifoBuf, finishedData)) {
            close(fifo);
            unlink(buf);
            buf[sv] = '\0';
            rmdir(buf);
            break;
        } else if (!strcasecmp(fifoBuf, justShow)) {
            QThread::postEvent(localMainWidget, new KInstanceEvent(NULL));
        } else if (!strncasecmp(fifoBuf, showMailto, strlen(showMailto))) {
            const char *p = copyString(fifoBuf + strlen(showMailto));
            QThread::postEvent(localMainWidget, new KInstanceEvent(p));
        }

        usleep(CLOSE_WRITE_SLEEP);
        close(fifo);
    }

}
 

