/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include "kdispatch.h"

#include <qpushbutton.h>

#include <iaccount.h>
#include "main.h"

/**
 * This is the other thread.  There is only ever one of these.   This is good.
 */
KDispatch downloadThread;


/**
 *=------------------------------------------------------------------------=
 * KDispatch::KDispatch
 *=------------------------------------------------------------------------=
 * Initializes a bunch of good.
 */
KDispatch::KDispatch()
{
    areWeRunning = false;
    threadInfo = NULL;
    threadInfoCount = 0;
}

/**
 *=------------------------------------------------------------------------=
 * KDispatch::~KDispatch
 *=------------------------------------------------------------------------=
 * there should actually be nothing here to do.
 */
KDispatch::~KDispatch()
{
    K_ASSERT(threadInfoCount == 0, "All of these should be gone by now!!!\n");
}


/**
 *=------------------------------------------------------------------------=
 * KDispatch::downloading
 *=------------------------------------------------------------------------=
 * Are we already downloading yet?
 *
 * Output:
xc *      bool                - true == yes.
 */
bool KDispatch::downloading()
{
    bool b;

    threadRunning.lock();
    b = areWeRunning;
    threadRunning.unlock();
    return b;
}

/**
 *=------------------------------------------------------------------------=
 * KDispatch::beginDownload
 *=------------------------------------------------------------------------=
 * gathers all the information and starts the download running.
 *
 * Paramters:
 *      int                     - [in]  count of items to work on.
 *      AccountThreadInfo *     - [in]  array of information to work with.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KDispatch::beginDownload
(
    int count,
    AccountThreadInfo *info
)
{
    K_ASSERT(threadInfo == NULL, "I should'nt have any of this yet !!!!\n");

    threadInfo = info;
    threadInfoCount = count;

    K_ASSERT(!areWeRunning, "how on earth are we already running ???\n");

    threadRunning.lock();
    areWeRunning = true;
    threadRunning.unlock();
    QThread::start();
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * KDispatch::run
 *=------------------------------------------------------------------------=
 * this is in a different thread.  Isn't this exciting!!!??!?!??
 */
void KDispatch::run()
{
    bool hadError = false;
    char buf[128];
    ERRCODE *ers;

    /**
     * Param checking.
     */
    K_ASSERT(threadInfo && threadInfoCount, "I need info to do work!!!!\n");
    if (!threadInfoCount) return;

    /**
     * Allocate an array of errcodes so we can track which accounts had problems
     * and which didn't!!
     */
    ers = (ERRCODE *)localAlloc(sizeof(ERRCODE) * threadInfoCount);
    if (!ers) {
        errorShowError(E_OUTOFMEMORY);
        return;
    }

    /**
     * Okay, for each account, go and tell it to do its work.
     */
    for (int x = 0; x < threadInfoCount; x++) {
        ers[x] = threadInfo[x].downloadFn(threadInfo[x].account, threadInfo[x].parameters);
    }

    /**
     * Now print out the results.
     */
    fprintf(stderr, "=---------------------------------=\n");
    for (int x = 0; x < threadInfoCount; x++) {
        const char *pstr = NULL;
        threadInfo[x].account->getDisplayName(&pstr);
        snprintf(buf, sizeof(buf), "Account '%s' downloaded with code: %x\n", pstr, ers[x]);
        if (pstr) localFree(pstr);
        if (FAILED(ers[x])) hadError = true;
        fprintf(stderr, buf);
    }
    fprintf(stderr, "=---------------------------------=\n");

    /**
     * IF everything went well, tell the host to make the download dialog go
     * away.
     */
    if (!hadError) 
        QThread::postEvent((QObject *)localMainWidget, new DispatchCompleteEvent());

    delete threadInfo;
    threadInfo = NULL;
    threadInfoCount = 0;

    /**
     * Clean up, clear the flag, and exit.
     */
    if (ers) localFree(ers);
    threadRunning.lock();
    areWeRunning = false;
    threadRunning.unlock();
}


/**
 *=------------------------------------------------------------------------=
 *                          DownloadMessageEvent
 *=------------------------------------------------------------------------=
 */

/**
 *=------------------------------------------------------------------------=
 * DownloadMessageEvent::DownloadMessageEvent
 *=------------------------------------------------------------------------=
 * creates a new dle object.
 *
 * Parameters:
 *      DownloadMessageType                 - [in]  duh
 *      const char *                        - [in]  message, if applicable
 *      int                                 - [in]  count of something, if
 *                                              applicable
 */
DownloadMessageEvent::DownloadMessageEvent
(
    DownloadMessageType t,
    const char *pstr,
    int c
)
: QEvent(User) 
{
    this->type = t; 
    if (pstr) {
        strncpy(msg, pstr, sizeof(msg));
        msg[255] = '\0';
    } else {
        msg[0] = '\0';
    }
    this->count = c;  
}




