/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#ifndef _KDISPATCH_H_
#define _KDISPATCH_H_

#include <sanity.h>

#include <qevent.h>
#include <qthread.h>

/**
 *=------------------------------------------------------------------------=
 * this header contains all the classes and routines that will be used in the
 * dispatch thread, which goes and does all the work of actually downloading
 * the mail, etc ...
 *=------------------------------------------------------------------------=
 */
 
class KDispatch;
class IAccount;


/**
 *=------------------------------------------------------------------------=
 * DispatchEvents are used to send your own account some information on the
 * main GUI thread.
 *=------------------------------------------------------------------------=
 */
class DispatchEvent : public QEvent {
  public:
    DispatchEvent(IAccount *a, void *d) : QEvent(User) {
        account = a; data = d;
    }

    IAccount *account;
    void *data;
};

/**
 *=------------------------------------------------------------------------=
 * Tells the main program we're done.
 *=------------------------------------------------------------------------=
 */
class DispatchCompleteEvent : public QEvent {
  public:
    DispatchCompleteEvent() : QEvent(User) { }
};

/**
 *=------------------------------------------------------------------------=
 * these events are all used to signal the download window or the main GUI
 * window that something is going on.  They are then taken and displayed as
 * the main UI sees fit !!
 *=------------------------------------------------------------------------=
 */
enum DownloadMessageType {
    InitAccount = 0,
    SetCount,
    NextMessage,
    SetDisplayText,
    AddMessage
};

class DownloadMessageEvent : public QEvent {

  public:
    DownloadMessageEvent(DownloadMessageType t, const char *pstr, int c);

    DownloadMessageType type;
    char msg[256];
    int count;
};

class ShowPasswordDialogEvent : public QEvent {
  public:
    ShowPasswordDialogEvent(IAccount *a, const char *acctIn, const char *name) 
        : QEvent(User) {
        this->acct = a; account = acctIn;  userName = name; }

    IAccount *acct;
    const char *account;
    const char *userName;
};


/**
 *=------------------------------------------------------------------------=
 * this is the thread that handles downloads!!
 *=------------------------------------------------------------------------=
 */
extern KDispatch downloadThread;


/**
 * Each account will specify a function and a pointer with the parameter(s)
 * that function expects.  They are passed directly to the function.
 */
struct AccountThreadInfo {

    ERRCODE (*downloadFn)(IAccount *, void *);
    IAccount *account;
    void *parameters;
};



/**
 *=------------------------------------------------------------------------=
 * Qt uses a scheme like java -- the "run()" method on a class inheriting 
 * from a thread class will do the work in another thread.  We will use
 * this class to define said thread.
 *=------------------------------------------------------------------------=
 */
class KDispatch : public QThread {

  public:
    /**
     * Once all the informatin has been gathered for all the 
     * accounts, this is the function that is called to dispatch the
     * thread and start the downloading.
     */
    ERRCODE beginDownload(int count, AccountThreadInfo *);

    /**
     * Are we already downloading?
     */
    bool downloading();

    KDispatch();
    ~KDispatch();


    void terminate() { }

  protected:
    void run();

  private:
    AccountThreadInfo *threadInfo;
    int threadInfoCount;

    bool areWeRunning;
    QMutex threadRunning;

};



#endif // _KDISPATCH_H_

