/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <kmaillist.h>

#include <qdrawutil.h>
#include <qpainter.h>
#include <stdio.h>
#include <string.h>

/**
 * handy little constants.
 */
const int PAD = 2;
const int ROWINCREMENT = 50;
const int MINCOLUMNWIDTH = 5;

#define WIDTH_FROM_PCT(pc, ttl)        ((int)(((pc) / 100.0) * (double)(ttl)))
#define DBL_PCT_FROM_WIDTH(wd, ttl)    (((double)(wd) / (double)(ttl)) * 100.0)

#define GET_QT_ALIGN(al)    ((al) == CA_LEFT ? Qt::AlignLeft : ((al) == CA_RIGHT ? Qt::AlignRight : Qt::AlignCenter))

#define min(a, b)    ((a) < (b) ? (a) : (b))
#define max(a, b)    ((a) >= (b) ? (a) : (b))



/**
 *=------------------------------------------------------------------------=
 * KMailList::KMailList
 *=------------------------------------------------------------------------=
 * creates a new widget.
 */
KMailList::KMailList
(
    QWidget *parent,
    const char *name
) : QWidget(parent, name)
{
    /**
     * Rows & Columns
     */
    headers = NULL;
    cheaders = 0;

    front = end = NULL;
    crows = 0;
    crowTop = 0;
    rowHeight = -1;
    viewableRows = 0;

    /**
     * Selection management
     */
    focusSelect = 0;
    shiftAnchor = 0;

    /**
     * mouse control and tracking.
     */
    mouseMode = MMNormal;

    /**
     * Painting and updating.
     */
    _doubleBuffer = new QPixmap(2, 2);
    _backgroundBrush = NULL;

    /**
     * Scrollbar
     */
    scrollBar = NULL;

    /**
     * general GUI Goo
     */
    setBackgroundMode(NoBackground);
    setMouseTracking(true);
    setFocusPolicy(StrongFocus);

    DEBUG_TEST_START;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::~KMailList
 *=------------------------------------------------------------------------=
 * Clean up da house.
 */
KMailList::~KMailList()
{
    nukeAll();

    if (_doubleBuffer) delete _doubleBuffer;
    if (_backgroundBrush) delete _backgroundBrush;

    DEBUG_TEST_STOP;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::addRow
 *=------------------------------------------------------------------------=
 * Adds the given row to the control.  There should be the same number of 
 * columns of data as there are columns in the control.
 *
 * Parameters:
 *      KMailListItem *     - [in]  the data for the row
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KMailList::addRow
(
    KMailListItem *rowData
)
{
    KInternalItem *newRow;

    if (!rowData) return E_PARAM;
    if (rowData->ccolumns != cheaders) return E_PARAM;
    if (cheaders == 0) return E_PARAM;

    /**
     * Okay, we need to go and allocate a new item for this row.
     */
    newRow = (KInternalItem *)localAlloc(sizeof(KInternalItem));
    if (!newRow) return E_OUTOFMEMORY;
    memset(newRow, 0, sizeof(KInternalItem));

    newRow->ccolumns = rowData->ccolumns;
    newRow->boldEntry = rowData->boldEntry;
    newRow->itemData = rowData->itemData;
    newRow->columns = rowData->columns;

    return addRowInternal(newRow);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::addRowInternal
 *=------------------------------------------------------------------------=
 * adds the item to the kmaillist.
 *
 * Parameters:
 *      KInternalItem *             - [in]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KMailList::addRowInternal
(
    KInternalItem *newRow
)
{
    KInternalItem *iter;
    bool done = false;
    int idx = 0;
    ERRCODE er;

    /**
     * Now we need to insert it into the list, in sorted fashion.
     */
    if (front == NULL) {
        front = end = newRow;
    } else {
        iter = front;

        while (iter != NULL && !done) {
            int cmp = (headers[primarySort].sortfn)(primarySort, headers[primarySort].itemData, newRow, iter);
            cmp = (headers[primarySort].sort == CS_PRIMARY_DOWN)
                    ? cmp
                    : -cmp;
            if (cmp < 0) {
                /**
                 * It's less -- insert.
                 */
                newRow->next = iter;
                if (iter->prev) {
                    iter->prev->next = newRow;
                    newRow->prev = iter->prev;
                }
                else {
                    front = newRow;
                    newRow->next = iter;
                }
                iter->prev = newRow;
                if (idx <= focusSelect) focusSelect++;
                break;
            } else if (cmp > 0) {
                /**
                 * It's greater than -- continue searching.
                 */
                iter = iter->next;
                idx++;
            } else {
                /**
                 * they're the same -- we need to sort on the secondary col
                 */
                while (iter != NULL) {

                    cmp = (headers[primarySort].sortfn)(primarySort, headers[primarySort].itemData, newRow, iter);
                    cmp = (headers[primarySort].sort == CS_PRIMARY_DOWN)
                            ? cmp
                            : -cmp;
                    int cmp2 = (headers[secondarySort].sortfn)(secondarySort, headers[secondarySort].itemData, newRow, iter);
                    if (cmp < 0 || cmp2 < 0) {
                        newRow->next = iter;
                        if (iter->prev) {
                            iter->prev->next = newRow;
                            newRow->prev = iter->prev;
                        } else {
                            front = newRow;
                            newRow->next = iter;
                        }
                        iter->prev = newRow;
                        done = true;
                        if (idx <= focusSelect) focusSelect++;
                        break;
                    } else {
                        iter = iter->next;
                        idx++;
                    }
                } // while

            }

        } // while

        /**
         * If we ran off the end of the list, insert at the end!!!
         */
        if (iter == NULL) {
            end->next = newRow;
            newRow->prev = end;
            end = newRow;
        }
    }

    er = S_OK;

    crows++;
    adjustScrollBar();
    this->forceUpdate();
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::adjustScrollBar
 *=------------------------------------------------------------------------=
 * figures out if we need to show or hide our scrollbar.
 *
 * Parameters:
 *      int             - [in] the new height of the control, or -1 if we
 *                          don't know it yet.
 */
void KMailList::adjustScrollBar
(
    int newHeight
)
{
    int h = (newHeight == -1) 
            ? this->height() - getRowHeight()
            : newHeight - getRowHeight();

    /**
     * If we've computed a row height, then go and figure out how many
     * rows we can display now.
     */
    this->viewableRows = h / getRowHeight();
    if (crowTop >= (crows - viewableRows)) {
        crowTop = crows - viewableRows;
        if (crowTop < 0) crowTop = 0;
    }

    /**
     * If there are more rows than viewable rows, then we need to add a 
     * scrollbar here now.
     */
    if (viewableRows < crows) {
        QScrollBar *sb = this->getScrollBar();
        if (sb) {
            sb->setGeometry(this->width() - sb->sizeHint().width(), 0,
                            sb->sizeHint().width(), this->height());
            sb->show();

            sb->setRange(0, crows - viewableRows);
            sb->setPageStep(this->viewableRows);
        }
    } else {
        crowTop = 0;
        if (this->scrollBar) this->scrollBar->hide();
        this->forceUpdate();
    }
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::cleanUpHeaders
 *=------------------------------------------------------------------------=
 * nukes our header information.
 */
void KMailList::cleanUpHeaders()
{
    for (int x = 0; x < cheaders; x++) {
        if (headers[x].text) localFree(headers[x].text);
        if (headers[x].pixmap) delete headers[x].pixmap;
    }

    if (headers) localFree(headers);
    headers = NULL;
    cheaders = 0;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::cleanUpRows
 *=------------------------------------------------------------------------=
 * nukes all the data we have for rows and anything inside it.
 */
void KMailList::cleanUpRows()
{
    KInternalItem *item, *next;

    item = front;
    while (item != NULL) {
        next = item->next;
        clearColumnInfo(item->columns);
        localFree(item);
        item = next;
    }

    focusSelect = 0;
    front = end = NULL;
    crows = 0;
    primarySort = secondarySort = -1;
    primaryUp = false;
    crowTop = 0;

    if (scrollBar) scrollBar->width();
    if (scrollBar) scrollBar->setValue(0);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::clearColumnInfo
 *=------------------------------------------------------------------------=
 * routine that cleans up the columns in a row.
 *
 * Parameters:
 *      KMailListColumn *               - [in]  columns to clear.
 */
void KMailList::clearColumnInfo
(
    KMailListColumn *clearme
)
{
    if (clearme) {
        for (int x = 0; x < cheaders; x++) {
            if (clearme[x].text) localFree(clearme[x].text);
        }
        localFree(clearme);
    }
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::columnSortingChanged
 *=------------------------------------------------------------------------=
 * this happens when the user has clicked on a column, and wishes to sort
 * differently.  We must first figure out how to resort, and then go 
 * actually do it.
 */
void KMailList::columnSortingChanged()
{
    KInternalItem *ourFront, *item, *focus = NULL;
    int crowCache;

    K_ASSERT(activeColumn >= 0 && activeColumn < cheaders, "Bogus active col!");
    /**
     * First, figure out what column is being sorted, and then begin resorting
     */
    switch (headers[activeColumn].sort) {
        case CS_PRIMARY_UP:
            headers[activeColumn].sort = CS_PRIMARY_DOWN;
            primarySort = activeColumn;
            break;
        case CS_PRIMARY_DOWN:
            headers[activeColumn].sort = CS_PRIMARY_UP;
            primarySort = activeColumn;
            break;

        case CS_SECONDARY:
        case CS_NONE:
            /**
             * Clear out the other column sorting info
             */
            for (int x = 0; x < activeColumn; x++) headers[x].sort = CS_NONE;
            for (int x = activeColumn + 1; x < cheaders; x++) 
                headers[x].sort = CS_NONE;
            
            /** 
             * Guess on a good secondary sort.
             */
            if (activeColumn > 0 && headers[activeColumn - 1].text) {
                secondarySort = activeColumn - 1;
                headers[secondarySort].sort = CS_SECONDARY;
            } else if (activeColumn < cheaders - 1 && headers[activeColumn + 1].text) {
                secondarySort = activeColumn + 1;
                headers[secondarySort].sort = CS_SECONDARY;
            } else if (activeColumn > 0) {
                secondarySort = activeColumn - 1;
                headers[secondarySort].sort = CS_SECONDARY;
            } else if (activeColumn < cheaders - 1) {
                secondarySort = activeColumn + 1;
                headers[secondarySort].sort = CS_SECONDARY;
            }

            /**
             * Great, now go and set this to be  primary down.
             */
            headers[activeColumn].sort = CS_PRIMARY_DOWN;
            primarySort = activeColumn;
            break;
    }


    /**
     * Resort all the rows now.  We're going to 'pull' them all out of the control,
     * and just go add them all again.
     */
    setUpdatesEnabled(false);
    getScrollBar()->setRange(0, 0);
    if (focusSelect >= 0) focus = itemAtIndex(focusSelect);
    crowCache = crows;
    ourFront = front;
    shiftAnchor = 0;
    crowTop = 0;
    crows = 0;

    front = end = NULL;
    
    /**
     * Now whiz through all the rows and re-add them.  Hope this works!
     */
    while (ourFront != NULL) {
        item = ourFront;
        ourFront = ourFront->next;
        if (ourFront) ourFront->prev = NULL;
        item->next = NULL;
        item->prev = NULL;

        addRowInternal(item);
    }

    K_ASSERT(crows == crowCache, "Should have the same number !!!\n");
    if (focus) selectRowFromItemData(focus->itemData);

    /**
     * Finally, repaint!!
     */
    setUpdatesEnabled(true);
    forceUpdate();

    emitColumnsChanged();
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::copyColumnInfo
 *=------------------------------------------------------------------------=
 * given some column information for a row, copy it over.  We factored this
 * out as we do it in a few places ...
 *
 * Parameters:
 *      KMailListColumn *           - [in]  array of columns to copy
 *      KMailListColumn **          - [out] new copied ones.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KMailList::copyColumnInfo
(
    KMailListColumn *copyme,
    KMailListColumn **pcopy
)
{
    ERRCODE er;
    int x;

    *pcopy = (KMailListColumn *)localAlloc(sizeof(KMailListColumn) * cheaders);
    if (!(*pcopy)) return E_OUTOFMEMORY;
    memset(*pcopy, 0, sizeof(KMailListColumn) * cheaders);

    for (x = 0;  x < cheaders; x++) {
        if (copyme[x].text) {
            (*pcopy)[x].text = copyString(copyme[x].text);
            if (!(*pcopy)[x].text) {
                er = E_OUTOFMEMORY;
                goto Error;
            }
        }
        if (copyme[x].pixmap) {
            (*pcopy)[x].pixmap = new QPixmap(*(copyme[x].pixmap));
            if (!(*pcopy)[x].pixmap) {
                er = E_OUTOFMEMORY;
                goto Error;
            }
        }
        (*pcopy)[x].alignment = copyme[x].alignment;
    }

    er = S_OK;

  Error:
    if (FAILED(er)) {
        clearColumnInfo(*pcopy);
        *pcopy = NULL;
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::drawableWidth
 *=------------------------------------------------------------------------=
 * returns the width of the drawable portion of the control.  This is the 
 * regular width if there is no scrollbar, or the width minus the scrollbar's
 * width if there is one!
 *
 * Output:
 *      int
 */
int KMailList::drawableWidth()
{
    int w = this->width();

    if (scrollBar && scrollBar->isVisible())
        w -= scrollBar->width();

    return w;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::drawContentsInBuffer
 *=------------------------------------------------------------------------=
 * the control has changed in some way, shape or form, so we need to redraw
 * now.  We can't always do this right away, as I've found that some events
 * crash when you try to get painters and the likes within them.
 */
void KMailList::drawContentsInBuffer()
{
    QPainter bufferPainter;
    QColorGroup cg = colorGroup();

    bufferPainter.begin(_doubleBuffer);

    _doubleBuffer->fill(cg.base());
    
    /**
     * Draw the headers in the buffer first.
     */ 
    drawHeadersInBuffer(bufferPainter, cg);

    /**
     * Now go and draw the actual rows.
     */
    drawRowsInBuffer(bufferPainter, cg);

    /**
     * Finally, if we are resizing headers, then draw an XOR Line to show 
     * that.
     */
    if (mouseMode == MMColumnResize) {

        bufferPainter.setPen(gray);
        bufferPainter.setRasterOp(XorROP);
        bufferPainter.drawLine(resizeX, 0, resizeX, this->height());
    }

    bufferPainter.end();

    /**
     * Finally, reset this flag so we don't waste a tonne of time on this.
     */
    mustUpdateBuffer = false;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::drawHeadersInBuffer
 *=------------------------------------------------------------------------=
 * Draws the header portion of the control.
 *
 * Parameters:
 *      QPainter &          - [in] the painter with which we are working.
 *      QColorGroup &       - [in] where we figure out what colours to use.
 */
void KMailList::drawHeadersInBuffer
(
    QPainter &painter,
    QColorGroup &cg
)
{
    int w = this->drawableWidth() - PAD;
    int colwidth, colleft = PAD;
    int x;

    /**
     * First thing we need to do is figure out how much space the fixed width
     * columns are taking up.  The variable width columns split the rest.
     */
    w -= getFixedWidth();
    if (w < 0) w = 0;

    /**
     * great, now whiz through the columns and paint them.
     */
    for (x = 0; x < cheaders; x++) {
        
        /**
         * Compute real column width;
         */
        if (headers[x].type == CW_FIXED)
            colwidth = headers[x].width;
        else {
            colwidth = WIDTH_FROM_PCT(headers[x].pct, w);
        }

        headers[x].widthCache = colwidth;

        /**
         * Draw the button.
         */
        if (!_backgroundBrush) _backgroundBrush = new QBrush(cg.color(QColorGroup::Button));
        if (!_backgroundBrush) return;

        style().drawBevelButton(&painter, colleft, 0, colwidth, getRowHeight(), 
                                cg, headers[x].pressed, _backgroundBrush);

        /**
         * Next, draw the text or the image.
         */
        if (headers[x].text == NULL) {
            /**
             * If we've got a pixmap, then draw it centred.
             */
            if (headers[x].pixmap) {
                int a = headers[x].pixmap->width();
                int l = colleft + ((colwidth - 2*PAD) - a) / 2 + PAD;
                painter.drawPixmap(l, PAD, *headers[x].pixmap);
            }
        } else {
            painter.drawText(colleft + PAD, PAD, colwidth - PAD * 4, getRowHeight() - PAD,
                             Qt::AlignLeft, headers[x].text);
        }

        /**
         * Finally, if appropriate, draw the triangle showing sorting.
         */
        if (x == primarySort) {

            if (headers[x].stringCache == -1) {
                if (headers[x].text)
                    headers[x].stringCache = painter.fontMetrics().width(headers[x].text);
                else
                    headers[x].stringCache = headers[x].pixmap->width();
            }

            /**
             * Draw the sorting triangle depending on whether they are down or up.
             */
            int triRight = colleft + headers[x].widthCache - 4 * PAD;
            if (headers[x].widthCache >= headers[x].stringCache + 6 * PAD + 6) {

                QPen pen = painter.pen();

                if (headers[x].sort == CS_PRIMARY_DOWN) {
                    painter.setPen(cg.light());
                    painter.drawLine(triRight - 6, 12, triRight, 12);
                    painter.drawLine(triRight - 3, 7, triRight, 12);
                    painter.setPen(cg.dark()); 
                    painter.drawLine(triRight - 6, 12, triRight - 3, 7);
                } else { // up
                    painter.setPen(cg.mid()); 
                    painter.drawLine(triRight - 6, 7, triRight, 7);
                    painter.drawLine(triRight - 6, 7, triRight - 3, 12);
                    painter.setPen(cg.light());
                    painter.drawLine(triRight - 3, 12, triRight, 7);
                }
                painter.setPen(pen);
            }
        }
        colleft += colwidth;
    }
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::drawRowsInBuffer
 *=------------------------------------------------------------------------=
 * draws the rows in the buffer, using correct viewport and all that jazz.
 * Also updates the selection and all that jazz!!!
 *
 * Parameters:
 *      QPainter &              - [in]  what to paint with.
 *      QColorGroup &           - [in]  information on colours
 */
void KMailList::drawRowsInBuffer
(
    QPainter &painter,
    QColorGroup &cg
)
{
    KInternalItem *currentItem;
    int itemsPainted;
    int a = 0, b = 0;
    QFont font;
    int sw, sh;
    int left;

    /**
     * if there are no rows, this e-z.
     */
    if (crows == 0) return;

    font = painter.font();

    /**
     * Now draw the rows, one at a time.
     */
    for (currentItem = itemAtIndex(crowTop), itemsPainted = 0;
         currentItem && itemsPainted < (viewableRows + 1);
         currentItem = currentItem->next, itemsPainted++) {

        left = 0;

        /**
         * If the row is selected, then we need to draw a filled rect and
         * change the font colour
         */
        if (currentItem->selected == true) {
            painter.fillRect(0, itemsPainted * getRowHeight() + getRowHeight(), this->drawableWidth(), getRowHeight(),
                             cg.highlight());
            painter.setPen(white);
        } else {
            painter.setPen(cg.text());
        }

        /**
         * Now whiz through and draw each header one at a time.
         */
        for (int y = 0; y < cheaders; y++) {

            /**
             * draw it bold if necessary!
             */
            if (currentItem->boldEntry == true) {
                font.setBold(true);
                painter.setFont(font);
            } else {
                font.setBold(false);
                painter.setFont(font);
            }

            /**
             * Do we have any text?
             */
            if (currentItem->columns[y].text != NULL) {
                
                painter.drawText(left + PAD, itemsPainted * getRowHeight() + getRowHeight() + PAD, 
                                 headers[y].widthCache - 2 * PAD,
                                 getRowHeight(),
                                 GET_QT_ALIGN(currentItem->columns[y].alignment),
                                 currentItem->columns[y].text);
            }

            /**
             * do we have a pixmap??
             */
            if (currentItem->columns[y].pixmap != NULL) {

                switch (currentItem->columns[y].alignment) {
                    case CA_LEFT:
                        a = PAD;
                        b = 0;
                        break;
                    case CA_CENTRE:
                        a = (currentItem->columns[y].pixmap->width() > (headers[y].widthCache - 2 * PAD))
                            ? PAD
                            : (currentItem->columns[y].pixmap->width() - 2 * PAD) / 2 + PAD;
                        b = (currentItem->columns[y].pixmap->height() > getRowHeight() - 2 * PAD)
                            ? 0
                            : (currentItem->columns[y].pixmap->height() - 2 * PAD) / 2 + PAD;
                        break;

                    case CA_RIGHT:
                        a = (currentItem->columns[y].pixmap->width() > (headers[y].widthCache - 2 * PAD))
                            ? PAD
                            : headers[y].widthCache - PAD - currentItem->columns[y].pixmap->width();
                        b = (currentItem->columns[y].pixmap->height() > getRowHeight() - 2 * PAD)
                            ? 0
                            : getRowHeight() - PAD - currentItem->columns[y].pixmap->width();
                        break;
                }
                sw = (currentItem->columns[y].pixmap->width() > (headers[y].widthCache - 2 * PAD))
                     ? headers[y].widthCache - 2 * PAD
                     : currentItem->columns[y].pixmap->width();
                sh = (currentItem->columns[y].pixmap->height() > getRowHeight() - 2 *PAD)
                     ? getRowHeight() - 2 * PAD
                     : currentItem->columns[y].pixmap->height();

                a += left;
                b += itemsPainted * getRowHeight() + getRowHeight();
                /**
                 * Phew!  Finally draw that muthah!!!
                 */
                painter.drawPixmap(a, b, *(currentItem->columns[y].pixmap), 0, 0, sw, sh);
            }


            left += headers[y].widthCache;

        } // for (y < cheaders)


    } // for (itemsPainted < viewableRows + 1)

    /**
     * If we've got focus, and we have some rows, then draw a selection rect 
     * around the last selected one.
     */
    if (this->hasFocus()
        && crows > 0 
        && focusSelect >= crowTop
        && focusSelect < (crowTop + viewableRows + 1)) {

        
        painter.drawWinFocusRect(0, (focusSelect - (crowTop)) * getRowHeight() + getRowHeight(),
                                 this->drawableWidth(), getRowHeight());
    }

}


/**
 *=------------------------------------------------------------------------=
 * KMailList::emitColumnsChanged
 *=------------------------------------------------------------------------=
 * the columns have changed.  we need to tell our hosts about it.
 */
void KMailList::emitColumnsChanged()
{
    KMailListHeader *hdr = (KMailListHeader *)localAlloc(sizeof(KMailListHeader) * cheaders);
    if (!hdr) return;

    for (int x = 0; x < cheaders; x++) {
        hdr[x].text = headers[x].text;
        hdr[x].pixmap = headers[x].pixmap;
        hdr[x].width = headers[x].width;
        hdr[x].pct = headers[x].pct;
        hdr[x].type = headers[x].type;
        hdr[x].itemData = headers[x].itemData;
        hdr[x].sort = headers[x].sort;
    }

    emit columnHeadersChanged(cheaders, hdr);
    localFree(hdr);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::emitDragItemStarted
 *=------------------------------------------------------------------------=
 * the user has done some UI stuff that could qualify as starting a drag. If
 * anybody is actually interested in doing so, drag it now.
 */
void KMailList::emitDragItemStarted()
{
    emit dragItemStarted();
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::emitItemLaunched
 *=------------------------------------------------------------------------=
 * an item has been selected.  Notify anybody who wants to know that
 */
void KMailList::emitItemLaunched()
{
    KMailListItem **array;
    int citems;
    ERRCODE er;

    er = internalGenerateEventInfo(array, citems);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }

    /**
     * The event object will clean up the array when it gets destroyed (on
     * leaving scope)
     */
    KMailListEvent kme(focusSelect, citems, array);
    emit itemLaunched(&kme);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::emitItemSelected
 *=------------------------------------------------------------------------=
 * an item has been selected in the UI -- it now has the focus rect.  Fire
 * an event telling whomever could possibly care about this.
 */
void KMailList::emitItemSelected()
{
    KMailListItem **array = NULL;
    ERRCODE er;
    int citems;

    er = internalGenerateEventInfo(array, citems);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }

    /**
     * The event object will clean up the array when it gets destroyed (on
     * leaving scope)
     */
    KMailListEvent kme(focusSelect, citems, array);
    emit itemSelected(&kme);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::emitKeyPressed
 *=------------------------------------------------------------------------=
 * The user has pressed a key -- just pass the event on.  Callees should note
 * that they're only being given keys that the control itself doesn't care
 * about.  Others are swallowed entirely, and you need to do serious event
 * filtering if you want to prevent those !!
 *
 * Parameters:
 *      QKeyEvent *                 - [in]  details, details, details ...
 */
void KMailList::emitKeyPressed
(
    QKeyEvent *event
)
{
    emit keyPressed(event);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::emitRightButtonPressed
 *=------------------------------------------------------------------------=
 * the user has clicked the right button in the control.  go and figure
 * out what items are selected and pass this information on.
 *
 * Parameters:
 *      QMouseEvent *               - [in]  details to emit.
 */
void KMailList::emitRightButtonPressed
(
    QMouseEvent *event
)
{
    emit rightButtonPressed(event);
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::endKey
 *=------------------------------------------------------------------------=
 * the user has pressed the END key -- cope.
 *
 * Parameters:
 *      ButtonState                 - [in]
 */
void KMailList::endKey
(
    ButtonState state
)
{
    KInternalItem *item;
    int x;

    /**
     * Set everything up at the end, and exit if there are no rows
     */
    crowTop = crows - viewableRows;
    if (crowTop < 0) crowTop = 0;
    focusSelect = crows ? crows - 1 : crows;
    if (!crows) return;
    setScrollValue(crowTop);

    /**
     * Figure out what sort of selection we need to do.
     */
    if (state & ShiftButton) {

        x = 0;
        item = front;
        while (x < shiftAnchor) {
            if (!item) break;

            item->selected = false;
            item = item->next;
            x++;
        }

        while (item) {
            item->selected = true;
            item = item->next;
        }

        emitItemSelected();

    } else if (state & ControlButton) {
    
        // do nothing -- we're just changing the focused row

    } else {
        selectRowFromItemData(end->itemData);
    }

    this->forceUpdate();
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::focusInEvent
 *=------------------------------------------------------------------------=
 * We've gained the focus.  Deal with it.
 *
 * Parameters:
 *      QFocusEvent *               - [in]  da event.
 */
void KMailList::focusInEvent
(
    QFocusEvent *event
)
{
    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::focusOutEvent
 *=------------------------------------------------------------------------=
 * We've lost the focus.  Deal with it.
 *
 * Parameters:
 *      QFocusEvent *               - [in]  da event.
 */
void KMailList::focusOutEvent
(
    QFocusEvent *event
)
{
    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::getColumnLeft
 *=------------------------------------------------------------------------=
 * returns the left-most x-coordinate of this column.
 *
 * Parameters:
 *      int                     - [in]  column we want to size.
 *
 * Output:
 *      int
 */
int KMailList::getColumnLeft
(
    int column
)
{
    int w = 0;

    K_ASSERT((column >= 0) && (column < cheaders), "D'oh!  Bogus column number");

    for (int x = 0; x < column; x++) {
        w += headers[x].widthCache;
    }

    return w;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::getFixedWidth
 *=------------------------------------------------------------------------=
 * returns the sum of the widths of the columns of type CW_FIXED.
 *
 * Output:
 *      int                     - zero if there are no fixed ones.
 */
int KMailList::getFixedWidth()
{
    int sum = 0;

    if (cheaders == 0) return 0;

    for (int x = 0; x < cheaders; x++) {
        if (headers[x].type == CW_FIXED) {
            sum += headers[x].width;
        }
    }

    return sum;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::getMinRightWidth
 *=------------------------------------------------------------------------=
 * returns the minimum width that must be to the right of this column.
 * Basically, this is the sum of all fixed columns to the right of this one
 * plus a minimum of 5 pixels per resizable column ...
 *
 * Parameters:
 *      int                     - [in]  column to see to the right of.
 *
 * Output:
 *      int
 */
int KMailList::getMinRightWidth
(
    int column
)
{
    int w = 0;

    for (int x = column + 1; x < cheaders; x++) {

        if (headers[x].type == CW_FIXED)
            w += headers[x].width;
        else
            w += MINCOLUMNWIDTH;
    }

    return w;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::getRowHeight
 *=------------------------------------------------------------------------=
 * computes the height of a row in the control.
 *
 * Output:
 *      int
 */
int KMailList::getRowHeight()
{
    QPainter painter;
    
    /**
     * If we haven't computed it yet, do so now.
     */
    if (rowHeight == -1) {

        painter.begin(_doubleBuffer);

        rowHeight = painter.fontMetrics().height() + 5;
        painter.end();
    }

    K_ASSERT(rowHeight != -1, "This is bad !!!! RowHeight is bogus");
    return rowHeight;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::getScrollBar
 *=------------------------------------------------------------------------=
 * returns our scrollbar object.  Creates it if it hasn't already been
 * created.
 *
 * Output:
 *      QScrollBar *
 */
QScrollBar *KMailList::getScrollBar()
{
    if (!this->scrollBar) {
        this->scrollBar = new QScrollBar(this);
        QObject::connect(this->scrollBar, SIGNAL(valueChanged(int)),
                         this, SLOT(scrollValueChanged(int)));
    }

    K_ASSERT(this->scrollBar, "Out of memory situation!!???");

    return this->scrollBar;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::getSelectedCount
 *=------------------------------------------------------------------------=
 * returns the number of items that are selected.
 *
 * Output:
 *      int
 */
int KMailList::getSelectedCount()
{
    KInternalItem *iter = front;
    int c = 0;

    while (iter) {
        if (iter->selected) c++;
        iter = iter->next;
    }

    return c;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::getSelectedRows
 *=------------------------------------------------------------------------=
 * returns the item datas for the rows that are selected in the view 
 * currently.
 *
 * Parameters:
 *      LinkList<unsigned int> &        - [in]  where to put the item datas
 *                                          for the selected rows.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KMailList::getSelectedRows
(
    LinkList<unsigned int> &list
)
{
    KInternalItem *iter = front;
    ERRCODE er;

    while (iter) {
        if (iter->selected) {
            er = list.addElement(iter->itemData);
            RETURN_ON_FAILURE(er);
        }
        iter = iter->next;
    }


    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::homeKey
 *=------------------------------------------------------------------------=
 * the HOME key has been pressed.  Figure it out.
 *
 * Parameters:
 *      ButtonState                 - [in]  CTRL/SHIFT/ALT button state
 */
void KMailList::homeKey
(
    ButtonState state
)
{
    KInternalItem *item;
    int x;

    /**
     * Set up at the top, and then exit if there are no rows.
     */
    setScrollValue(0);
    crowTop = 0;
    focusSelect = 0;
    if (!crows) return;

    /**
     * Figure out what sort of selecting we need to do.
     */
    if (state & ShiftButton) {
        x = 0;
        item = front;
        while (x <= shiftAnchor) {
            if (!item) break;

            item->selected = true;
            item = item->next;
            x++;
        }

        while (item) {
            item->selected = false;
            item = item->next;
        }

        emitItemSelected();

    } else if (state & ControlButton) {
    
        // do nothing -- we're just changing the focused row

    } else {
        selectRowFromItemData(front->itemData);
    }

    this->forceUpdate();
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::internalGenerateEventInfo
 *=------------------------------------------------------------------------=
 * Generate the event information based on what is currently selected.
 *
 * Parameters:
 *      KMailListItem ** &          - [out] array of selected items.
 *      int &                       - [out] count of items.
 *
 * Output:
 *      ERRCODE                     - S_OK or E_OUTOFMEMORY
 */
ERRCODE KMailList::internalGenerateEventInfo
(
    KMailListItem ** &array,
    int & citems
)
{
    KInternalItem *iter;
    int c;

    /**
     * first, get the number of selected items.
     */
    citems = getSelectedCount();

    /**
     * Now create an array of the items
     */
    if (citems) {
        array = (KMailListItem **)localAlloc(sizeof(KMailListItem *) * citems);
        if (!array) {
            return E_OUTOFMEMORY;
        }

        /**
         * Now copy over the rows.
         */
        c = 0;
        iter = front;
        while (iter != NULL) {
            if (c == citems) break;
            if (iter->selected)
                array[c++] = iter;

            iter = iter->next;
        }

        K_ASSERT(c == citems, "How the hell could this happen?");
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::itemAtIndex
 *=------------------------------------------------------------------------=
 * returns the item at the given index.
 *
 * Parameters:
 *      int             - [in]  index to look for item.
 *
 * Output:
 *      KInternalItem * - the item.
 */
KMailList::KInternalItem *KMailList::itemAtIndex
(
    int index
)
{
    KInternalItem *iter = front;
    int c = 0;

    if (index == -1) return NULL;

    while (iter) {
         if (c == index) return iter;
        c++;
        iter = iter->next;
    }

    return NULL;
}



/**
 *=------------------------------------------------------------------------=
 * KMailList::keyPressEvent
 *=------------------------------------------------------------------------=
 * The user has pressed a key.  We will process it now!!!
 *
 * Parameters:
 *      QKeyEvent *             - [in]  details
 */
void KMailList::keyPressEvent
(
    QKeyEvent *event
)
{
    ButtonState state;

    /**
     * If there's no event, or we don't have any rows, then this is easy!
     */
    if (!event) return;
    if (!crows) {
        event->ignore();
        return;
    }

    state = event->state();

    /**
     * see if it's an arrow key.
     */
    switch (event->key()) {
        case Key_Down:
            nextElement(state);
            break;

        case Key_Up:
            previousElement(state);
            break;

        case Key_Enter:
        case Key_Return:
            if (crows && itemAtIndex(focusSelect))
                emitItemLaunched();
            break;

        case Key_Space:
            if (crows) {
                itemAtIndex(focusSelect)->selected = true;
                this->forceUpdate();
                emitItemSelected();
            }
            break;

        case Key_Left:
        case Key_Right:
            break;

        case Key_Comma:
            if (state & ControlButton) previousElement(state);
            break;

        case Key_Period:
            if (state & ControlButton) nextElement(state);
            break;
            
        case Key_Home:
            homeKey(state);
            break;

        case Key_End:
            endKey(state);
            break;

        case Key_PageDown:
            pageDown(state);
            break;

        case Key_PageUp:
            pageUp(state);
            break;

        default:
            emitKeyPressed(event);
            break;
    }
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::leftMousePress
 *=------------------------------------------------------------------------=
 * the user has clicked the left mouse button ... figure out how to change
 * selection now ...
 *
 * Parameters:
 *      QMouseEvent *                   - [in]  details
 */
void KMailList::leftButtonPress
(
    QMouseEvent *event
)
{
    KInternalItem *item;
    ButtonState bs;
    int x, y, w;

    x = event->x();
    y = event->y();
    w = 0;

    if (y < getRowHeight()) {
        /**
         * If they're in a variable width column header, then we are
         * going to start a column resize.
         */
        for (int z = 0; z < cheaders; z++) {
            if ((z != 0) && (x > (w - PAD))
                && (x < (w + PAD + 1))
                && (headers[z - 1].type != CW_FIXED)) {

                this->mouseMode = MMColumnResize;
                activeColumn = z - 1;
                resizeX = x;
                this->forceUpdate();
                break;
            } else {
                w += headers[z].widthCache;
            }
        }

        /**
         * If they didn't press BETWEEN two headers, then that probably
         * means they pressed ON a header.
         */
        if (this->mouseMode != MMColumnResize) {
            w = 0;
            for (int z = 0; z < cheaders; z++) {

                if (w + headers[z].widthCache > x) {
                    headers[z].pressedX = x;
                    headers[z].pressed = true;
                    this->mouseMode = MMColumnDrag;
                    this->activeColumn = z;
                    this->forceUpdate();
                    break;
                } else
                    w += headers[z].widthCache;
            }
        }

    } else {

        bs = event->state();
        rememberPress = false;

        /**
         * If they're over an item, then we have work to do!
         */
        int r = (y - getRowHeight()) / getRowHeight();
        if (crowTop + r < crows) {
            /**
             * If they pushed select, unselect everything except all the rows
             * between focusselect and us, which will now be selected.
             */
            if (bs & ShiftButton) {
                x = 0;
                item = front;
                while (x < min(shiftAnchor, crowTop + r)) {
                    if (!item) break;

                    item->selected = false;
                    item = item->next;
                    x++;
                }

                while (x <= max(shiftAnchor, crowTop + r)) {
                    if (!item) break;

                    item->selected = true;
                    item = item->next;
                    x++;
                }

                while (item) {
                    item->selected = false;
                    item = item->next;
                }

                focusSelect = crowTop + r;
                emitItemSelected();

            } else if (bs & ControlButton) {

                // do nothing until mouse release.

            } else {
                /**
                 * This has the advantage of clearing out all existing 
                 * selections, which is what we want !!!
                 */
                item = itemAtIndex(crowTop + r);
                rememberPress = item->selected;
                if (!item->selected)
                    selectRowFromItemData(itemAtIndex(crowTop + r)->itemData);
            }

            this->forceUpdate();
        }
    }

}


/**
 *=------------------------------------------------------------------------=
 * KMailList::mouseDoubleClickEvent
 *=------------------------------------------------------------------------=
 * The user has double clicked.  If the mouse is over a selected item, then 
 * show that item, and show multiple items if it's multi-selected.  yikes!
 *
 * Parameters:
 *      QMouseEvent *           - [in]  details, details, details.
 */
void KMailList::mouseDoubleClickEvent
(
    QMouseEvent *event
)
{
    if (!event) return;

    int x, y, rowIndex;

    x = event->x();
    y = event->y();

    /**
     * See if there is an item underneath this location.
     */
    if (y > getRowHeight()) {

        rowIndex = (y - getRowHeight()) / getRowHeight();

        if (crows && (crowTop + rowIndex) < crows)
            if (itemAtIndex(crowTop + rowIndex)->selected)
                emitItemLaunched();
    }
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::mouseMoveEvent
 *=------------------------------------------------------------------------=
 * Indicates the the mouse has moved.  We need to support column resizing,
 * selection of items, and who knows what else here  ...
 *
 * Parameters:
 *      QMouseEvent *           - [in]  details, details, details.
 */
void KMailList::mouseMoveEvent
(
    QMouseEvent *event
)
{
    int thiswidth;
    int x, y;
    bool b;

    if (!event) return;

    thiswidth = this->drawableWidth() - 1;
    x = event->x();
    y = event->y();

    if (x > thiswidth) x = thiswidth;

    switch (mouseMode) {
        case MMNormal:
            {
            int w = 0;

            /**
             * If the left button is pressed, then we need to go and start a
             * drag drop operation.
             */
            if (event->state() & LeftButton) {
                emitDragItemStarted();
                break;
            }

            /**
             * If the event is within our header region, then see if we need 
             * to set up a vertical splitter cursor now.  You can only resize
             * non-fixed columns
             */
            if (y < getRowHeight()) {
                for (int z = 0; z < cheaders; z++) {
                    if ((z != 0)
                        && (x > (w - PAD) )
                        && (x < (w + PAD + 1))
                        && (headers[z - 1].type != CW_FIXED)) {

                        this->setCursor(SplitVCursor);
                        break;
                    } else {
                        this->setCursor(ArrowCursor);
                        w += headers[z].widthCache;
                    }
                }
            } else {
                this->setCursor(ArrowCursor);
            }
            }
            break;

        /**
         * We need to be able to column re-sorting and eventually column
         * drags here.
         */
        case MMColumnDrag:
            b = headers[activeColumn].pressed;

            if (x <= (headers[activeColumn].pressedX + 3)
                && x >= (headers[activeColumn].pressedX - 3))
                headers[activeColumn].pressed = true;
            else 
                headers[activeColumn].pressed = false;
            if (b != headers[activeColumn].pressed)
                this->forceUpdate();
            break;

        /** 
         * move the line for adjusting the column's width as appropriate.  Do
         * not let them blow over any boundaries.
         */
        case MMColumnResize:
            {
            int left = getColumnLeft(activeColumn);
            if (x < (left + 5))
                x = left + 5;

            int pad = getMinRightWidth(activeColumn);
            if (x > (thiswidth - pad))
                x = thiswidth - pad;

            resizeX = x;
            this->forceUpdate();
            }
            break;

        default:
            K_FAIL("Unknown Internal State for the mouse.  No es bueno!!!!");
            break;
    }
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::mousePressEvent
 *=------------------------------------------------------------------------=
 * the user has clicked on the mouse.  They can use this to resize columns,
 * sort columns, select items, and who knows what else.
 *
 * Parameters:
 *      QMouseEvent *           - [in]  details, details, details.
 */
void KMailList::mousePressEvent
(
    QMouseEvent *event
)
{
    ButtonState bs;

    if (!event) return;

    /**
     * Is it a right button event?
     */
    bs = event->button();
    if ((bs & RightButton) != 0 && (bs & LeftButton) == 0) {
        leftButtonPress(event);
        rightButtonPress(event);
    } else if ((bs & LeftButton) != 0 && (bs & RightButton) == 0) {
        leftButtonPress(event);
    }

}

/**
 *=------------------------------------------------------------------------=
 * KMailList::mouseReleaseEvent
 *=------------------------------------------------------------------------=
 * the user has let go of the mouse.  If they are resizing a column, then 
 * update the positions, etc.
 *
 * Parameters:
 *      QMouseEvent *           - [in]  details, details, details.
 */
void KMailList::mouseReleaseEvent
(
    QMouseEvent *event
)
{
    int x, y, w, thiswidth;
    KInternalItem *item;
    bool b;

    if (!event) return;

    thiswidth = this->drawableWidth();
    x = event->x();
    y = event->y();
    w = 0;

    switch (this->mouseMode) {

        /**
         * OKay, they're done resizing -- we need to compute where all the 
         * columns lie now.
         */
        case MMColumnDrag:
            K_ASSERT(headers[activeColumn].pressedX != -1, "I have no idea how this works any more!");
            if (x <= headers[activeColumn].pressedX + 3
                || x >= headers[activeColumn].pressedX - 3) {
                columnSortingChanged();
            }
            b = headers[activeColumn].pressed;
            headers[activeColumn].pressed = false;
            headers[activeColumn].pressedX = -1;
            if (b) this->forceUpdate();
            break;

        case MMColumnResize:
            resizeColumnHeader(x, y);
            break;

        case MMNormal:
            /**
             * If they're over an item, then we have work to do!
             */
            w = (y - getRowHeight()) / getRowHeight();
            if (crowTop + w < crows) {

                item = itemAtIndex(crowTop + w);

                if (event->state() & ControlButton) {

                    /**
                     * Toggle the selection on this item, and that's it.
                     */
                    item->selected = !item->selected;
                    focusSelect = crowTop + w;
                    shiftAnchor = focusSelect;
                    emitItemSelected();
                    this->forceUpdate();
                } else if (rememberPress) {

                    if (item->selected)
                        selectRowFromItemData(item->itemData);
                }
                rememberPress = false;
            }
            break;

        default:
            K_FAIL("Unknown Internal State for the mouse.  No es bueno!!!!");
            break;
    }

    this->mouseMode = MMNormal;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::nextElement
 *=------------------------------------------------------------------------=
 * go to the next element in the list.  if there isn't one, then stay, fido,
 * stay!!!
 *
 * Parameters:
 *      ButtonState             - [in]  state of control and shift keys
 */
void KMailList::nextElement
(
    ButtonState bs
)
{
    KInternalItem *item;
    int x;

    /**
     * If there are no more elements, then this is easy!
     */
    if (focusSelect >= (crows - 1)) return;
        
    focusSelect++;

    if (bs & ShiftButton) {
        x = 0;
        item = front;
        while (x < min(shiftAnchor, focusSelect)) {
            if (!item) break;

            item->selected = false;
            item = item->next;
            x++;
        }

        while (x <= max(shiftAnchor, focusSelect)) {
            if (!item) break;

            item->selected = true;
            item = item->next;
            x++;
        }

        while (item) {
            item->selected = false;
            item = item->next;
        }

        emitItemSelected();

    } else if (bs & ControlButton) {

        /**
         * Just move the focus rect -- we don't actually select an item here.
         */
        shiftAnchor = focusSelect;

    } else {

        /**
         * Go and select the next item, and unselect everybody else !!!!
         */
        selectRowFromItemData(itemAtIndex(focusSelect)->itemData);
    }

    /**
     * make sure it's visible
     */
    if (focusSelect - (viewableRows - 1) > crowTop) {
        crowTop++;
        setScrollValue(crowTop);
    }

    /**
     * Finally, repaint it all
     */
    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::nextItem
 *=------------------------------------------------------------------------=
 * given the item data for an item in the list, get the next item and return
 * it.
 *
 * Parameters:
 *      unsigned int                - [in]  item data for item we are 
 *                                      indexing off.
 *
 * Output:
 *      KMailListItem *             - NULL if there isn't a next.
 */
KMailListItem *KMailList::nextItem
(
    unsigned int itemData
)
{
    KInternalItem *iter = front;

    /**
     * Whiz through looking.
     */
    while (iter) {
        if (iter->itemData == itemData) return iter->next;
        iter = iter->next;
    }

    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::nukeAll
 *=------------------------------------------------------------------------=
 * wipes out all the data and the column headers currently set.
 */
void KMailList::nukeAll()
{
    /**
     * YOU MUST clean up rows before headers, as the rows need to know how
     * many columns there are !!!!
     */
    cleanUpRows();
    cleanUpHeaders();
    if (this->scrollBar) this->scrollBar->hide();
    this->forceUpdate();
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::pageDown
 *=------------------------------------------------------------------------=
 * the user has hit the PageDown key.   We need to figure out if there are
 * items to scroll, multiple items to select, etc ...
 *
 * Parameters:
 *      ButtonState                     - [in]  state of Shift/CTRL/etc.
 */
void KMailList::pageDown
(
    ButtonState state
)
{
    KInternalItem *item;
    int x;

    /**
     * No rows??? this is easy!!
     */
    if (!crows) return;

    /**
     * First, we need to update the currently selected row, making sure
     * we don't drift off the end of the list.
     */
    if (focusSelect < crowTop + viewableRows - 1)  {
        focusSelect = crowTop + viewableRows - 1;
        if (focusSelect >= crows) focusSelect = crows - 1;
        if (crowTop + viewableRows - 1 > crows) crowTop = crows - viewableRows + 1;
    } else {
        crowTop += viewableRows;
        focusSelect += viewableRows;
        if (focusSelect >= crows) focusSelect = crows - 1;
        if (crowTop >= crows) crowTop = crows - viewableRows;
        if (crowTop < 0) crowTop = 0;
        if (crowTop + viewableRows - 1 > crows) crowTop = crows - viewableRows + 1;
    }

    /**
     * Next figure out if there's any wacky selecting changes that we
     * need to do.
     */
    if (state & ShiftButton) {

        x = 0;
        item = front;
        while (x < min(shiftAnchor, focusSelect)) {
            if (!item) break;

            item->selected = false;
            item = item->next;
            x++;
        }

        while (x <= max(shiftAnchor, focusSelect)) {
            if (!item) break;

            item->selected = true;
            item = item->next;
            x++;
        }

        while (item) {
            item->selected = false;
            item = item->next;
        }

        emitItemSelected();

    } else if (state & ControlButton) {
        
        // do nothing -- we DON'T want to change the focusSelected item's
        // selected value here.

    } else {
        selectRowFromItemData(itemAtIndex(focusSelect)->itemData);
    }

    this->forceUpdate();
    setScrollValue(crowTop);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::pageUp
 *=------------------------------------------------------------------------=
 * the user has hit the PageUp key.   We need to figure out if there are
 * items to scroll, multiple items to select, etc ...
 *
 * Parameters:
 *      ButtonState                     - [in]  state of Shift/CTRL/etc.
 */
void KMailList::pageUp
(
    ButtonState state
)
{
    KInternalItem *item;
    int x;

    /**
     * No rows??? this is easy!!
     */
    if (!crows) return;

    /**
     * First, we need to adjust the currently focused row, making sure we
     * don't jump off the top of the list.
     */
    if (focusSelect > crowTop)
        focusSelect = crowTop;
    else {
        crowTop -= viewableRows;
        focusSelect -= viewableRows;
        if (crowTop < 0) {
            crowTop = 0;
            focusSelect = 0;
        }
    }

    /**
     * Next, figure out what sort of selection manipulation we need to do:
     */
    if (state & ShiftButton) {

        x = 0;
        item = front;
        while (x < min(shiftAnchor, focusSelect)) {
            if (!item) break;

            item->selected = false;
            item = item->next;
            x++;
        }

        while (x <= max(shiftAnchor, focusSelect)) {
            if (!item) break;

            item->selected = true;
            item = item->next;
            x++;
        }

        while (item) {
            item->selected = false;
            item = item->next;
        }

        emitItemSelected();

    } else if (state & ControlButton) {

        // do nothing -- we want to move the row without changing it's 
        // "selected" value.

    } else {
        selectRowFromItemData(itemAtIndex(focusSelect)->itemData);
    }

    this->forceUpdate();
    setScrollValue(crowTop);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::paintEvent
 *=------------------------------------------------------------------------=
 * paints our control.  We will most certainly use double buffering in this
 * control for better perf.
 *
 * Parameters:
 *      QPaintEvent *           - [in]  the event.
 */
void KMailList::paintEvent
(
    QPaintEvent *event
)
{
    if (mustUpdateBuffer)
        drawContentsInBuffer();

    bitBlt(this, 0, 0, _doubleBuffer);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::previousElement
 *=------------------------------------------------------------------------=
 * Selects the previous row in the list, or does nothing if we're already
 * at the first one.
 *
 * Parameters:
 *      ButtonState             - [in]  state of control and shift keys
 */
void KMailList::previousElement
(
    ButtonState bs
)
{
    KInternalItem *item;
    int x;

    /**
     * If there are no previous elements, then this is easy!
     */
    if (focusSelect <= 0) return;
        
    focusSelect--;

    if (bs & ShiftButton) {
        x = 0;
        item = front;
        while (x < min(shiftAnchor, focusSelect)) {
            if (!item) break;

            item->selected = false;
            item = item->next;
            x++;
        }

        while (x <= max(shiftAnchor, focusSelect)) {
            if (!item) break;

            item->selected = true;
            item = item->next;
            x++;
        }

        while (item) {
            item->selected = false;
            item = item->next;
        }

        emitItemSelected();

    } else if (bs & ControlButton) {

        /**
         * Just move the focus rect -- we don't actually select an item here.
         */
        shiftAnchor = focusSelect;

    } else {

        /**
         * Go and select the next item, and unselect everybody else !!!!
         */
        selectRowFromItemData(itemAtIndex(focusSelect)->itemData);
    }

    /**
     * make sure it's visible
     */
    if (focusSelect < crowTop) {
        crowTop = focusSelect;
        setScrollValue(crowTop);
    }

    /**
     * Finally, repaint it all
     */
    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::previousItem
 *=------------------------------------------------------------------------=
 * given the item data for an item in the list, get the previous item and 
 * return it.
 *
 * Parameters:
 *      unsigned int                - [in]  item data for item we are 
 *                                      indexing off.
 *
 * Output:
 *      KMailListItem *             - NULL if there isn't a previous.
 */
KMailListItem *KMailList::previousItem
(
    unsigned int itemData
)
{
    KInternalItem *iter = front;

    /**
     * Whiz through looking.
     */
    while (iter) {
        if (iter->itemData == itemData) return iter->prev;
        iter = iter->next;
    }

    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::removeRows
 *=------------------------------------------------------------------------=
 * nukes the given Row(s) from the list of rows.  We do some fancy stuff to
 * try and make this more efficient in the case where the number of rows
 * is greater than a number we pull out of our butt, basically.
 *
 * Parameters:
 *      int                     - [in]  count of rows to yank.
 *      unsigned int *          - [in]  array of item data to id the rows
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KMailList::removeRows
(
    int count,
    unsigned int *itemDataArray
)
{
    KInternalItem *iter, *killme, *focus;
    int deletedSoFar = 0;
    int currentIndex = 0;

    iter = front;

    focus = itemAtIndex(focusSelect);

    /**
     * whiz through the list.  at each item, see if it is any of the items
     * we want to delete.  we also want to be careful to decrement
     * focusSelect if we are deleting rows before it -- otherwise you end
     * up with a weird effect where it seems to "Jump down" by the number
     * of rows you deleted before it.
     */
    while (iter) {

        for (int y = 0; y < count; y++) {
            if (iter->itemData == itemDataArray[y]) {
                killme = iter;
                if (iter->prev)
                    iter->prev->next = iter->next;
                else
                    front = iter->next;
                if (iter->next) 
                    iter->next->prev = iter->prev;
                else
                    end = iter->prev;
                clearColumnInfo(killme->columns);
                localFree(killme);
                deletedSoFar++;
                crows--;
                if (currentIndex < focusSelect) {
                    focusSelect--;
                    currentIndex--;
                }
                break;
            }
        }

        if (deletedSoFar == count) break;
        iter = iter->next;
        currentIndex++;
    }

    /**
     * Update the selection correctly, etc ...
     */
    if (focusSelect >= crows) focusSelect = crows - 1;
    if (focusSelect == -1) focusSelect = 0;
    iter = itemAtIndex(focusSelect);
    if (iter) iter->selected = true;
    shiftAnchor = focusSelect;

    if (crowTop + viewableRows >= crows) {
        crowTop = crows - viewableRows;
        if (crowTop < 0) crowTop = 0;
        setScrollValue(crowTop);
    }

    this->forceUpdate();
    adjustScrollBar();

    /** 
     * If the focused item changed, fire an event.
     */
    if (iter != focus) emitItemSelected();

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::resizeColumnHeader
 *=------------------------------------------------------------------------=
 * the user is done resizing a column header.  go and make the change 
 * effective now.
 *
 * Parameters:
 *      int                         - [in]  x coord of where they ended up.
 *      int                         - [in]  y coord of where they ended up.
 */ 
void KMailList::resizeColumnHeader
(
    int x,
    int y
)
{
    int thiswidth = this->drawableWidth();

    /**
     * Figure out where the mouse actually ended up.
     */
    int left = getColumnLeft(activeColumn);
    if (x < (left + 5))
        x = left + 5;

    int pad = getMinRightWidth(activeColumn);
    if (x > (thiswidth - pad))
        x = thiswidth - pad;

    resizeX = x;

    /**
     * Set our column's new width.
     */
    int widthDiff = x - (left + headers[activeColumn].widthCache);
    double d = headers[activeColumn].pct;

    headers[activeColumn].widthCache += widthDiff;
    headers[activeColumn].pct = DBL_PCT_FROM_WIDTH(x - left, thiswidth - getFixedWidth());

    /**
     * Now, share the extra space we created.
     */
    int cvarsRight = 0;
    for (int x = activeColumn + 1; x < cheaders; x++) {
        if (headers[x].type == CW_PERCENTAGE)
            cvarsRight ++;
    }

    /**
     * now, split up our excess/shortfall!
     */
    d = d - headers[activeColumn].pct;

    for (int x = activeColumn + 1; x < cheaders; x++) {
        if (headers[x].type == CW_PERCENTAGE)
            headers[x].pct += d / (double)cvarsRight;
    }

    /**
     * Finally, reset the cursor, mode, and repaint ourself.
     */
    this->setCursor(ArrowCursor);
    this->mouseMode = MMNormal;
    emitColumnsChanged();
    this->forceUpdate();
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::resizeEvent
 *=------------------------------------------------------------------------=
 * resizes our control as appropriate.
 * 
 * Parameters:
 *      QResizeEvent *                  - [in]
 */
void KMailList::resizeEvent
(
    QResizeEvent *event
)
{
    adjustScrollBar(event ? event->size().height() : -1);

    if (event)
        _doubleBuffer->resize(event->size());
    else
        _doubleBuffer->resize(this->size());

    mustUpdateBuffer = true;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::rightButtonPress
 *=------------------------------------------------------------------------=
 * the user has clicked the 'right' button.  go and tell the host that this
 * has happened, in case they want to show a popup menu or sumthin'.
 *
 * Parameters:
 *      QMouseEvent *               - [in]  details, detail, details...
 */
void KMailList::rightButtonPress
(
    QMouseEvent *event
)
{
    emitRightButtonPressed(event);
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::scrollValueChanged
 *=------------------------------------------------------------------------=
 * the scrollbar has moved.  fweee!
 *
 * Parameters:
 *      int                     - [in]  the new value of the scrollbar.
 */
void KMailList::scrollValueChanged
(
    int value
)
{
    crowTop = value;
    if (crowTop >= (crows - viewableRows)) {
        crowTop = (crows - viewableRows);
        if (crowTop < 0) crowTop = 0;
    }
    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::selectFirst
 *=------------------------------------------------------------------------=
 * selects the first item in the list.
 */
void KMailList::selectFirst()
{
//    homeKey((Qt::ButtonState)0);
    if (crows) selectRowFromItemData(front->itemData);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::selectRowFromItemData
 *=------------------------------------------------------------------------=
 * given item data identifying a row, go and mark that row as selected.  All
 * other selected rows are unselected ...
 *
 * Parameters:
 *      unsigned int                - [in]  select me.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KMailList::selectRowFromItemData
(
    unsigned int itemData
)
{
    KInternalItem *iter = front;
    int x = 0;

    /**
     * Whiz through the list looking for this item, and clearing out any
     * existing selections ...
     */
    while (iter) {
        if (iter->selected) iter->selected = false;
        if (iter->itemData == itemData) break;
        iter = iter->next;
        x++;
    }
    
    if (iter) {
        focusSelect = x;
        itemAtIndex(x)->selected = true;

        /**
         * Clear out the rest of the selections
         */
        iter = iter->next;
        while (iter) {
            if (iter->selected) iter->selected = false;
            iter = iter->next;            
        }
    } else {
        focusSelect = crows;
        iter = itemAtIndex(focusSelect);
        if (iter) iter->selected = true;
    }

    /**
     * make sure it's visible
     */
    if (focusSelect < crowTop) {
        crowTop = focusSelect;
    } else if (focusSelect > (crowTop + viewableRows)) {
        crowTop = focusSelect - viewableRows + 1;
    }

    if (crowTop + viewableRows >= crows) {
        crowTop = crows - viewableRows;
        if (crowTop < 0) crowTop = 0;
    }
    setScrollValue(crowTop);
    shiftAnchor = focusSelect;
    this->forceUpdate();
    this->emitItemSelected();

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::selectLast
 *=------------------------------------------------------------------------=
 * selects the last item in the list.
 */
void KMailList::selectLast()
{
//    endKey((Qt::ButtonState)0);
    if (crows) selectRowFromItemData(end->itemData);
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::setColumnHeaders
 *=------------------------------------------------------------------------=
 * Sets the headers for this control.  If there already are headers, they
 * are removed, and if there is data, it is remapped to the new headers. If
 * there are more columns of data than headers, the only the first columns of
 * data are shown.
 *
 * Parameters:
 *      KMailListHeader *           - [in]  array of headers.
 *      int                         - [in]  number of headers.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KMailList::setColumnHeaders
(
    KMailListHeader *newHeaders,
    int count
)
{
    ERRCODE er;
    int x = 0;

    if (cheaders && !newHeaders) return E_PARAM;

    /**
     * Delete any existing headers first.
     */
    if (cheaders && headers)
        cleanUpHeaders();

    if (!count) {
        this->forceUpdate();
        return S_OK;
    }

    /**
     * Great, now copy over the new information.
     */
    headers = (KInternalHeader *)localAlloc(sizeof(KInternalHeader) * count);
    if (!headers) return E_OUTOFMEMORY;

    for (x = 0; x < count; x++) {
        if (newHeaders[x].text) {
            headers[x].text = copyString(newHeaders[x].text);
            if (!headers[x].text) {
                er = E_OUTOFMEMORY;
                goto Error;
            }
        } else {
            headers[x].text = NULL;
        }
        if (newHeaders[x].pixmap) {
            headers[x].pixmap = new QPixmap(*newHeaders[x].pixmap);
            if (!headers[x].pixmap) {
                er = E_OUTOFMEMORY;
                goto Error;
            }
        } else {
            headers[x].pixmap = NULL;
        }

        headers[x].width = newHeaders[x].width;
        headers[x].pct = newHeaders[x].pct;
        headers[x].type = newHeaders[x].type;
        headers[x].itemData = newHeaders[x].itemData;
        headers[x].sort = newHeaders[x].sort;
        headers[x].widthCache = -1;
        headers[x].stringCache = -1;
        headers[x].pressedX = -1;
        headers[x].pressed = false;
        headers[x].sortfn = newHeaders[x].sortfn;
        if (headers[x].sort == CS_PRIMARY_DOWN ||
            headers[x].sort == CS_PRIMARY_UP)
            primarySort = x;
        else if (headers[x].sort == CS_SECONDARY)
            secondarySort = x;
    }

    if (primarySort == -1) {
        K_FAIL("No primary sort column specified !!!!");
        primarySort = 0;
    }
    primaryUp = headers[x].sort == CS_PRIMARY_UP;

    cheaders = count;
    this->forceUpdate();
    er = S_OK;

  Error:
    if (FAILED(er)) {
        for (int y = 0; y < x; y++) {
            if (headers[x].text) localFree(headers[x].text);
            if (headers[x].pixmap) delete headers[x].pixmap;
        }
        headers = NULL;
        cheaders = 0;
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::setScrollValue
 *=------------------------------------------------------------------------=
 * updates the value of the scroll bar, making sure that doing so doesn't
 * upset the crowTop and all that jazz.
 *
 * Parameters:
 *      int                         - [in]  new value
 */
void KMailList::setScrollValue
(
    int value
)
{
    this->getScrollBar()->setValue(value);
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::updateRowFromItemData
 *=------------------------------------------------------------------------=
 * Updates the row identified by the item data given.  Note that we haven't
 * designed this control to handle the case where there might be duplicate
 * item datas!
 *
 * Parameters:
 *      unsigned int                - [in]  item data to find row
 *      KMailListItem *             - [in]  new data for that row.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE KMailList::updateRowFromItemData
(
    unsigned int itemData,
    KMailListItem *newRow
)
{
    KInternalItem *iter;
    bool sel;

    if (!newRow) return E_PARAM;

    iter = front;
    while (iter) {
        if (iter->itemData == itemData) {
            sel = iter->selected;

            clearColumnInfo(iter->columns);

            iter->ccolumns = newRow->ccolumns;
            iter->selected = sel;
            iter->boldEntry = newRow->boldEntry;
            iter->itemData = newRow->itemData;
            iter->columns = newRow->columns;
            this->forceUpdate();
            return S_OK;
        }
        iter = iter->next;
    }

    K_FAIL("Somebody asked for an item that doesn't exist !!!");
    return E_NOTFOUND;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::wheelEvent
 *=------------------------------------------------------------------------=
 * the mouse wheel has been spun.  Fweeeeeeeeeeeeeeeeeeeeeee--haaaaa!
 *
 * Parameters:
 *      QWheelEvent *
 */
void KMailList::wheelEvent
(
    QWheelEvent *event
)
{
    fprintf(stderr, "I Really did get a wheel event!!! Wow!!\n");
}



/**
 *=------------------------------------------------------------------------=
 *                            KMailListEvent                               +
 *=------------------------------------------------------------------------=
 */ 





/**
 *=------------------------------------------------------------------------=
 * KMailListEvent::KMailListEvent
 *=------------------------------------------------------------------------=
 * creates a new event object.
 *
 * Parameters:
 *      int                 - [in]  index of the selected row when the event
 *                              occurred
 *      int                 - [in]  number of items in this event.
 *      KMailListItem **    - [in]  array of pointers to items that are 
 *                              affected by this event.  We will take owner-
 *                              ship and free using localFree()
 */
KMailListEvent::KMailListEvent
(
    int indx,
    int citems,
    KMailListItem **rgitems
)
{
    itemIndex = indx;
    countOfItems = citems;
    itemValues = rgitems;
}

/**
 *=------------------------------------------------------------------------=
 * KMailListEvent::~KMailListEvent
 *=------------------------------------------------------------------------=
 * Cleans up!
 */
KMailListEvent::~KMailListEvent()
{
    if (itemValues) localFree(itemValues);
}

/**
 *=------------------------------------------------------------------------=
 * KMailListEvent::items
 *=------------------------------------------------------------------------=
 * returns an array with the items in it.  This is a new allocated array,
 * so CALLER frees.
 *
 * Output:
 *      KMailListItem **        - new array with pointers to affected items.
 */
KMailListItem **KMailListEvent::items()
{
    KMailListItem **rg;

    if (!itemValues) return NULL;

    rg = (KMailListItem **)localAlloc(sizeof(KMailListItem *) * countOfItems);
    if (!rg) return NULL;

    memcpy(rg, itemValues, sizeof(KMailListItem *) * countOfItems);
    return rg;
}


