/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <kmaillist.h>


/**
 * this file implements the iterator that is used by the KMailList to control
 * sorting of columns, etc ....
 */

const int ITEM_CHUNK_SIZE = 50;


/**
 *=------------------------------------------------------------------------=
 * KMailListSorter::KMailListSorter
 *=------------------------------------------------------------------------=
 * creates a new iterator object for our mail list.
 */
KMailList::KMailListSorter::KMailListSorter()
{
    primarySort = -1;
    secondarySort = -1;
    sortedItems = NULL;
    count = 0;
}

/**
 *=------------------------------------------------------------------------=
 * KMailListSorter::~KMailListSorter
 *=------------------------------------------------------------------------=
 * creates a new iterator object for our mail list.
 */
KMailList::KMailListSorter::~KMailListSorter()
{
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::KMailListSorter::cleanup
 *=------------------------------------------------------------------------=
 * cleans up all the crap we might have.
 */
void KMailList::KMailListSorter::cleanup()
{
    while (sortedItems) {
        IteratorNode *n = sortedItems->next;
        delete sortedItems;
        sortedItems = n;
    }
    sortedItems  = NULL;
    count = 0;

    primarySort = secondarySort = -1;
    primaryUp = false;
}
 
/**
 *=------------------------------------------------------------------------=
 * KMailList::KMailListSorter::getSortedArray
 *=------------------------------------------------------------------------=
 * returns the items in this list in an array fashion, sorted.
 *
 * Output:
 *      KInternalItem **            - sorted array.
 */
KMailList::KInternalItem **KMailList::KMailListSorter::getSortedArray()
{
    KMailList::KInternalItem **items = (KInternalItem **)localAlloc(sizeof(KInternalItem *) * count);
    IteratorNode *n = sortedItems;

    if (!items) return NULL;

    /**
     * Copy over the pointers.
     */
    if (primaryUp) {
        for (int x = count - 1; x >= 0; x--) {

            K_ASSERT(n, "Count is greater than the actual count of items in the list!");
            items[x] = n->item;
            n = n->next;
        }
    } else {
        for (int x = 0; x < count; x++) {

            K_ASSERT(n, "Count is greater than the actual count of items in the list!");
            items[x] = n->item;
            n = n->next;
        }
    }

    return items;
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::KMailListSorter::itemAdded
 *=------------------------------------------------------------------------=
 * we've added an item to the list.  add it to the sorted list correctly
 * and proceed.
 *
 * Parameters:
 *      KInternalItem *         - [in]  the item added.
 */
void KMailList::KMailListSorter::itemAdded
(
    KMailList::KInternalItem *item
)
{
    IteratorNode *nd, *iter, *prev;

    if (!item) return;

    /**
     * create a node for it.
     */
    nd = new IteratorNode();
    if (!nd) return;

    nd->item = item;
    nd->next = NULL;

    /**
     * Search for the appropriate place to put this item.
     */
    if (sortedItems == NULL) {
        sortedItems = nd;
    } else {

        iter = sortedItems;
        prev = NULL;
        while (iter != NULL) {

            int x = strcasecmp(item->columns[primarySort].text, iter->item->columns[primarySort].text);
            if (x < 0) {
                nd->next = iter;
                if (prev)
                    prev->next = nd;
                else
                    sortedItems = nd;
                break;
            } else if (x == 0) {
                IteratorNode *iter2, *prev2;
                iter2 = iter;
                prev2 = prev;

                while (iter2 != NULL) {

                    int y = strcasecmp(item->columns[secondarySort].text, iter2->item->columns[secondarySort].text);
                    if (y < 0 || y == 0) {
                        nd->next = iter2;
                        if (prev2)
                            prev2->next = nd;
                        else
                            sortedItems = nd;
                        break;
                    } else {
                        prev2 = iter2;
                        iter2 = iter2->next;
                    }
                } // while
                
                if (iter2 == NULL) {
                    if (prev2)
                        prev2->next = nd;
                    else
                        sortedItems = nd;
                }
                break;

            } else {
                prev = iter;
                iter = iter->next;
            }
        } // while

        if (iter == NULL) {
            if (prev)
                prev->next = nd;
            else
                sortedItems = nd;
        }
    } // if

    count++;
}

/**
 *=------------------------------------------------------------------------=
 * KMailList::KMailListSorter::itemRemoved
 *=------------------------------------------------------------------------=
 * the given item has been removed from the list.  Go and yank it now.
 *
 * Parameters:
 *      KMailList::KInternalItem *  - [in]  item removed.
 */
void KMailList::KMailListSorter::itemRemoved
(
    KMailList::KInternalItem *item
)
{
    IteratorNode *n, *prev;

    K_ASSERT(item, "There should alwawys be an item here!!!!");
    K_ASSERT(sortedItems, "this should never happen!!!");

    if (!sortedItems) return;

    if (sortedItems->item == item) {
        n = sortedItems->next;
        delete sortedItems;
        sortedItems = n;
        count--;
    } else {
        n = sortedItems->next;
        prev = sortedItems;
        while (n) {
            if (n->item == item) {
                prev->next = n->next;
                delete n;
                break;
            }
            prev = n;
            n = n->next;
        }
        K_ASSERT(n != NULL, "Couldn't find the item to remove -- very bad !!");
        count--;
    }
}


/**
 *=------------------------------------------------------------------------=
 * KMailList::KMailListSorter::newList
 *=------------------------------------------------------------------------=
 * We've been given new list information.  Find out what the sorting columns
 * are and go and set that up now.  Clean up any other goo we might have.
 *
 * Parameters:
 *      int                     - [in]  primary sorting column
 *      int                     - [in]  secondary sorting column
 *      bool                    - [in]  true if the list is sorted up.
 */
void KMailList::KMailListSorter::newList
(
    int primary,
    int secondary,
    bool primaryup
)
{
    this->cleanup();

    primaryUp = primaryup;
    primarySort = primary;
    secondarySort = secondary;
}


/**
 *=------------------------------------------------------------------------=
 * KItemDataIterator::KItemDataIterator
 *=------------------------------------------------------------------------=
 * Creates a new iterator.  This is private, and should only be called from
 * the KMailList
 */
KItemDataIterator::KItemDataIterator()
{
    DEBUG_TEST_START;
    m_cref = 1;
}

/**
 *=------------------------------------------------------------------------=
 * KItemDataIterator::~KItemDataIterator
 *=------------------------------------------------------------------------=
 * destroys us.  Should ONLY be called by KItemDataIterator::release
 */
KItemDataIterator::~KItemDataIterator()
{
    DEBUG_TEST_STOP;
    K_ASSERT(m_cref == 0, "Being destroyed, but still have refs!!!");
}

/**
 *=------------------------------------------------------------------------=
 * KItemDataIterator::addRef
 *=------------------------------------------------------------------------=
 * adds a ref to our lifetime.
 *
 * Output:
 *      int             - new ref count 
 */
int KItemDataIterator::addRef()
{
    return ++m_cref;
}


/**
 *=------------------------------------------------------------------------=
 * KItemDataIterator::next
 *=------------------------------------------------------------------------=
 * given an item, returns the next one.
 *
 * Parameters:
 *      unsigned int                - [in]  item to search for.
 *      unsigned int *              - [out] the next item, or S_FALSE as ret
 *                                      if there isn't one !!
 *
 * Output:
 *      ERRCODE                     - S_FALSE if there is no next.
 */
ERRCODE KItemDataIterator::next
(
    unsigned int findMe,
    unsigned int *out
)
{
    if (!out) return E_PARAM;

    *out = 0xbaadf00d;

    for (int x = 0; x < citems; x++) {
        if (items[x] == findMe) {
            if (x == (citems -1)) {
                return S_FALSE;
            } else {
                *out = items[x + 1];
                return S_OK;
            }
        }
    }

    return E_NODATA;
}


/**
 *=------------------------------------------------------------------------=
 * KItemDataIterator::previous
 *=------------------------------------------------------------------------=
 * returns the previous item in the iterator, or S_FALSE if there isn't one.
 *
 * Paramters:
 *      unsigned int            - [in]  find me.
 *      unsigned int *          - [out] i'm the previous.
 *
 * Output:
 *      ERRCODE                 - S_FALSE if is no prev, E_NODATA if it's not
 *                                  there at all.
 */
ERRCODE KItemDataIterator::previous
(
    unsigned int findMe,
    unsigned int *out
)
{
    if (!out) return E_PARAM;

    *out = 0xbaadf00d;

    for (int x = 0; x < citems; x++) {
        if (items[x] == findMe) {
            if (x == 0) {
                return S_FALSE;
            } else {
                *out = items[x - 1];
                return S_OK;
            }
        }
    }

    return E_NODATA;
}


/**
 *=------------------------------------------------------------------------=
 * KItemDataIterator::release
 *=------------------------------------------------------------------------=
 * Release a ref on our lifetime.  if it hits zero, then we destroy ourselves.
 *
 * Output:
 *      int                 - new ref count
 */
int KItemDataIterator::release()
{
    int c = --m_cref;
    if (c <= 0) delete this;
    return c;
}



