/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <qthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>



#ifdef K_DEBUG

QMutex lallocMutex;

/*
 * this is the struct we will use to track allocations
 */
typedef struct TrackNode {

    char *fileName;
    int   lineNumber;
    int   byteCount;
    void *pointer;
    struct TrackNode *pnext;

} TRACKNODE;


typedef struct Tracker {

    int refCount;
    TRACKNODE *node;

} TRACKER;

/**
 * We'll create a hash table of allocations to try and speed things up some
 */
#define HASHVALUE 256
#define HASHPTR(ptr)    ((((unsigned int)(ptr) & 0x3c) >> 2) & 0xf)


TRACKER globalTracker[HASHVALUE];


/**
 * local function prototypes
 */
TRACKNODE *addAllocation(char *, int, int, void *);
BOOL        removeAllocation(void *);
BOOL        haveAnyAllocations(void);
void        dumpAllAllocations();
#endif // K_DEBUG



/**
 *=------------------------------------------------------------------------=
 * ensureBufferSize
 *=------------------------------------------------------------------------=
 * makes sure the buffer is big enough to accomodate the given size.
 *
 * Parameters:
 *      char **                 - [in/out]  buffer.
 *      int *                   - [in/out]  max buffer size.
 *      int                     - [in]      required size.
 *      int                     - [in]      size to increment by
 *
 * Output:
 *      bool                    - false == OOM
 */
bool ensureBufferSize
(
    char **pbuf,
    int *psize,
    int required,
    int incr
)
{
    char *save = *pbuf;

    if (required > *psize) {

        if (*pbuf) {
            *pbuf = (char *)localReAlloc(*pbuf, *psize + incr);
        } else {
            *pbuf = (char *)localAlloc(incr);
        }
        if (!(*pbuf)) {
            *pbuf = save;
            return false;
        }
        *psize += incr;
    }

    return true;
}


#if K_DEBUG
/**
 *=------------------------------------------------------------------------=
 * lallocInit
 *=------------------------------------------------------------------------=
 * initializes our allocation routines.  Basically just NULLs out our hash
 * table
 *
 * Output:
 *    BOOL                    - FALSE means we couldn't initailize it
 */
BOOL lallocInit
(
    void
)
{
    memset(globalTracker, 0, sizeof(globalTracker));

    return TRUE;
}
        

/**
 *=------------------------------------------------------------------------=
 * lallocTerm
 *=------------------------------------------------------------------------=
 * terminates our memory allocation code, checking for any unfreed pointers.
 * If there are some, we dump them to stderr
 */
void lallocTerm
(
    void
)
{
    if (haveAnyAllocations()) {

        fprintf(stderr, "\n**** UNFREED ALLOCATIONS EXIST !! DUMPING NOW****\n");
        dumpAllAllocations();
    }
}


/**
 *=------------------------------------------------------------------------=
 * localAllocImpl
 *=------------------------------------------------------------------------=
 * performs the given allocation and tracks it in our global hash table
 *
 * Parameters:
 *    char *                - [in]  filename where allocation ocurred
 *    int                   - [in]  line number in said file of alloc
 *    int                   - [in]  # bytes allocated
 *
 * Output:
 *    void *                - pointer to allocated memory!
 */
void *localAllocImpl
(
    char  *fileName,
    int    lineNumber,
    size_t byteCount
)
{
    lallocMutex.lock();
    void *ptr = malloc(byteCount);

    if (!ptr) {
        lallocMutex.unlock();
        return NULL;
    }

    // now add it to our tracking hash table.
    //
    if (!addAllocation(fileName, lineNumber, byteCount, ptr)) {
        free(ptr);
        lallocMutex.unlock();
        return NULL;
    }

    lallocMutex.unlock();
    return ptr;
}

/**
 *=------------------------------------------------------------------------=
 * localFreeImpl
 *=------------------------------------------------------------------------=
 * Frees the given allocation, and also removes it from our allocation
 * hash table.
 *
 * Parameters:
 *    const char *           - [in] filename where they tried this.
 *    int                    - [in] line number
 *    void *                 - [in] allocation to free up
 */
void localFreeImpl
(
    const char *filename,
    int line,
    void *ptr
)
{
    lallocMutex.lock();
    if (!removeAllocation(ptr)) {
        fprintf(stderr, " *** %s:%d: Unable to free allocation %x *** \n",
                filename, line, (unsigned int) ptr);
    }
    lallocMutex.unlock();
}

/**
 *=------------------------------------------------------------------------=
 * localReAllocImpl
 *=------------------------------------------------------------------------=
 * Reallocates the given allocation, adn then copies over the existing data
 * in the new allocation.  returns NULL if a new allocation can't be made.
 * IF THE NEW ALLOCATION IS SMALLER THAN THE OLDER ONE, DATA IS LOST!!!!!
 *
 * Parameters:
 *      const char *                - [in]  filename.
 *      int                         - [in]  line number.
 *      void *                      - [in]  poitner to realloc
 *      size_t                      - [in]  new size for pointer.
 *
 * Output:
 *      void *
 */
void *localReAllocImpl
(
    char *filename,
    int line,
    void *ptr,
    size_t size
)
{
    lallocMutex.lock();
    if (!removeAllocation(ptr)) {
        fprintf(stderr, " *** %s:%d: Unable to free allocation %x *** \n",
                filename, line, (unsigned int) ptr);
    }

    void *pnew = realloc(ptr, size);
    if (pnew) {
        if (!addAllocation(filename, line, size, pnew))
            fprintf(stderr, "Couldn't add allocation to hash table !!!\n");
        lallocMutex.unlock();
        return pnew;
    }

    /**
     * Allocation failed, add old allocation back and return it.
     */
    if (!addAllocation(filename, line, size, ptr))
        fprintf(stderr, "Couldn't add allocation to hash table !!!\n");

    lallocMutex.unlock();
    return ptr;
}


/**
 *=------------------------------------------------------------------------=
 * addAllocation
 *=------------------------------------------------------------------------=
 * Adds the given allocation to the currently open allocations.  We store
 * out information such as file name, line number, byte count, and pointer
 *
 * Parameters:
 *    char *                - [in]  filename from where allocation was made
 *    int                   - [in]  line number in said file of alloc
 *    int                   - [in]  # bytes allocated
 *    void *                - [in]  actual allocation
 *
 * Output:
 *    TRACKNODE *          - [out] unit created, or NULL if OOM
 */
TRACKNODE * addAllocation
(
    char *fileName,
    int   lineNumber,
    int   byteCount,
    void *allocation
)
{
    TRACKNODE *ptrack;
    char *p;
    int i;

    // arg checking
    //
    if (!fileName || !allocation || (byteCount <= 0) || (lineNumber <= 0))
        return NULL;

    K_ASSERT(globalTracker[HASHPTR(allocation)].refCount >= 0, "Bogus Ref Count !!!");

    // allocate a new unit
    //
    ptrack = (TRACKNODE *)malloc(sizeof(TRACKNODE));
    if (!ptrack) return NULL;

    // copy over the string filename
    //
    i = strlen(fileName);
    i++; // for '\0'

    p = (char *)malloc(i);
    if (!p) {
        free(ptrack);
        return NULL;
    }
    strncpy(p, fileName, i);

    // copy over the rest of the data
    //
    ptrack->fileName = p;
    ptrack->lineNumber = lineNumber;
    ptrack->byteCount = byteCount;
    ptrack->pointer = allocation;

    // set this in the array
    //
    ptrack->pnext = globalTracker[HASHPTR(allocation)].node;
    globalTracker[HASHPTR(allocation)].node = ptrack;
    globalTracker[HASHPTR(allocation)].refCount++;

    return ptrack;
}

/**
 *=------------------------------------------------------------------------=
 * removeAllocation
 *=------------------------------------------------------------------------=
 * removes the given allocation from our global tracking hash table
 *
 * Parameters:
 *    void *                    - [in]  allocation to remove
 *
 * Output:
 *    BOOL                      - TRUE if successful, FALSE if it couldn't
 *                                be found.
 */
BOOL removeAllocation
(
    void *removeMe
)
{
    TRACKNODE *p, *plast;

    if (!removeMe) return FALSE;

    p = globalTracker[HASHPTR(removeMe)].node;
    plast = NULL;
    
    // whip through all the items in this hash value, looking for this
    // pointer.  keep track of the previous pointer so we can skip the one
    // we're removing
    //
    while (p) {

        if (p->pointer == removeMe)
            break;

        plast = p;
        p = p->pnext;
    }

    // if we found one, remove it, clean it up, and set the pointers
    //
    if (p) {

        if (plast)
            plast->pnext = p->pnext;
        else
            globalTracker[HASHPTR(removeMe)].node = p->pnext;

        free(p->fileName);
        free(p);
        globalTracker[HASHPTR(removeMe)].refCount--;
        K_ASSERT(globalTracker[HASHPTR(removeMe)].refCount >= 0, "Bogus Ref Count on Delete");

        return TRUE;
    }

    // couldn't find it!!
    //
    return FALSE;
}

/**
 *=------------------------------------------------------------------------=
 * haveAnyAllocations
 *=------------------------------------------------------------------------=
 * returns a value indicating whether or not there are any outstanding
 * allocations
 *
 * Output:
 *    BOOL                    - TRUE if there are any outstanding allocs
 */
BOOL haveAnyAllocations
(
    void
)
{
    int x;

    for (x = 0; x < HASHVALUE; x++) {
        if (globalTracker[x].node != NULL)
            return TRUE;
    }

    return FALSE;
}

/**
 *=------------------------------------------------------------------------=
 * dumpAllAllocations
 *=------------------------------------------------------------------------=
 * goes through and prints out all of the outstanding allocations
 */
void dumpAllAllocations
(
    void
)
{
    TRACKNODE *p;
    int x;

    fputs("\n", stderr);

    for (x = 0; x < HASHVALUE; x++) {

        if (globalTracker[x].node != NULL) {
            p = globalTracker[x].node;
    
            while (p) {
                printf("%s (line %d): %d bytes\n", p->fileName, p->lineNumber,
                       p->byteCount);
                p = p->pnext;
            }
        }
    }
}


#endif // K_DEBUG


