/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>


#include <pthread.h>
#include <qdialog.h>
#include <qhbox.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qthread.h>
#include <qvbox.h>

#include "error.dlg.h"
#include <intl.h>
#include "kiltdown.h"
#include <lerror.h>
#include <llist.h>
#include "localized.h"
#include "main.h"


#if K_DEBUG
#include <qmessagebox.h>
#include "kiltdown.h"
#include "main.h"
#endif 

#define DEFINE_ERRORS_HERE        1
static const char *whoops = "(whoops)";
#define REAL_STR(p)        ((p) ? p : whoops)


/**
 *=------------------------------------------------------------------------=
 * Error module mutex
 *=------------------------------------------------------------------------=
 * All error routines need to be mutex'd so that only one occurs at a time
 * [we will have different threads trying to report an error.
 */
QMutex errorMutex;

/**
 *=------------------------------------------------------------------------=
 * This is a list of arrays for each thread in the application.  Each thread
 * is allowed to set up rich error information, up to five values.
 */
struct ThreadRichErrorInfo {
#ifndef WIN32
    pthread_t threadId;
#else
    DWORD threadId;
#endif // WIN32
    const char **richErrorParams;
};

LinkList<ThreadRichErrorInfo *> richErrorInfo;


/**
 *=------------------------------------------------------------------------=
 *                         The Error Table
 *=------------------------------------------------------------------------=
 */
struct ErrorTable {
    ERRCODE er;
    STRINGID baseMsg;
    STRINGID msgEx;
    int cRichSlots;
} errorTable [] = {
    #include "errortable.h"

    { 0, 0, 0, 0}
};


/**
 *=------------------------------------------------------------------------=
 *                          Local routines
 *=------------------------------------------------------------------------=
 */
void errorIdFromErrCode(ERRCODE, int &, int &, int &);
const char **richErrorParamsForThread();
void cleanRichErrorInfo();


/**
 *=------------------------------------------------------------------------=
 * errorConstructFullErrorString
 *=------------------------------------------------------------------------=
 * given the errcode and rich error information, fill in the substitutions
 * and return the full error string.
 *
 * Parameters:
 *      ERRCODE                 - [in]
 *      const char * &          - [out] newly allocated full error string.
 */
void errorConstructFullErrorString
(
    ERRCODE er,
    const char * &pstrOut
)
{
    int base = 0 , ex = 0, ct = 0;
    const char **rg;
    const char *pstr;
    char buf[512];

    pstrOut = NULL;

    rg = richErrorParamsForThread();
    if (!rg) {
        fprintf(stderr, "huge memory failure in Error module.  Can't show errs!!");
        return;
    }

    /**
     * First, get the strings.
     */
    errorIdFromErrCode(er, base, ex, ct);
    pstr = intlLoadString(base);
    if (!pstr) {
        fprintf(stderr, "catastrophic memory failure.  Can't show errrors!!!");
        return;
    }

    /**
     * Now fill in da blanks.
     */
    switch (ct) {
        case 0:
            strncpy(buf, pstr, sizeof(buf));
            buf[sizeof(buf) - 1] = '\0';
            break;

        case 1:
            snprintf(buf, sizeof(buf), pstr, REAL_STR(rg[0]));
            break;

        case 2:
            snprintf(buf, sizeof(buf), pstr, REAL_STR(rg[0]),
                                             REAL_STR(rg[1]));
            break;

        case 3:
            snprintf(buf, sizeof(buf), pstr, REAL_STR(rg[0]),
                                             REAL_STR(rg[1]),
                                             REAL_STR(rg[2]));
            break;

        case 4:
            snprintf(buf, sizeof(buf), pstr, REAL_STR(rg[0]),
                                             REAL_STR(rg[1]),
                                             REAL_STR(rg[2]),
                                             REAL_STR(rg[3]));
            break;

        case 5:
            snprintf(buf, sizeof(buf), pstr, REAL_STR(rg[0]),
                                             REAL_STR(rg[1]),
                                             REAL_STR(rg[2]),
                                             REAL_STR(rg[3]),
                                             REAL_STR(rg[4]));
            break;

        default:
            fprintf(stderr, "very bad number of rich error strings. can't do errors");
            return;
    }

    localFree(pstr);
    pstrOut = copyString(buf);
}

#ifdef K_DEBUG
/**
 *=------------------------------------------------------------------------=
 * errorFailedAssertion
 *=------------------------------------------------------------------------=
 * show a dialog for a failed assertion.
 *
 * Parameters:
 *      const char *            - [in]  failed assertion.
 *      const char *            - [in]  file name
 *      const char *            - [in]  message.
 *      int                     - [in]  line number in file.
 */
void errorFailedAssertion
(
    const char *assertion,
    const char *filename,
    const char *message,
    int lineNumber
)
{
    char buf[384];

    snprintf(buf, sizeof(buf), "\n\n\n*********************************\n\nASSERTION FAILED: %s\nFILE: %s, \
line %d\n\nmessage: %s\n*********************************\n\n\n", assertion, filename, lineNumber, message);

    fputs(buf, stderr);

//    QMessageBox::critical(localMainWidget, "FAILED ASSERTION", buf, QMessageBox::Ok, 0, 0);
}
#endif // K_DEBUG

/**
 *=------------------------------------------------------------------------=
 * clearRichErrorInfo
 *=------------------------------------------------------------------------=
 * nukes the error info.
 */
void clearRichErrorInfo()
{
    ThreadRichErrorInfo *info;
    ERRCODE er;

    /**
     * Whiz through the list clearing out the info
     */
    er = richErrorInfo.firstElement(&info);
    while (er == S_OK) {
        for (int x = 0; x < 5; x++)
            if (info->richErrorParams[x]) localFree(info->richErrorParams[x]);
        localFree(info->richErrorParams);
        
        richErrorInfo.removeElement(info);
        localFree(info);
        er = richErrorInfo.firstElement(&info);
    }
}


/**
 *=------------------------------------------------------------------------=
 * errorIdFromErrCode
 *=------------------------------------------------------------------------=
 * given an ERRCODe, get the IDs for the error message.
 *
 * Parameters:
 *      ERRCODE                 - [in]  error code
 *      int &                   - [out] base error message STRINGID
 *      int &                   - [out] extended error message STRINGID
 *      int &                   - [out] # rich error slots
 */   
void errorIdFromErrCode
(
    ERRCODE er,
    int &base,
    int &ex,
    int &ct
)
{
    int x = 0;

    while (errorTable[x].er != 0) {
        if (errorTable[x].er == er) {
            base = errorTable[x].baseMsg;
            ex = errorTable[x].msgEx;
            ct = errorTable[x].cRichSlots;
            return;
        }
        x++;
    }

    K_FAIL("ARGH!  Bogus ErrorCode not in errortable.h !!!!!");
}

/**
 *=------------------------------------------------------------------------=
 * errorSaveRichErrorInfo
 *=------------------------------------------------------------------------=
 * saves out the rich error information for a given error.  Up to five
 * items may be saved.  Currently we do not support localized reordering.
 * if it becomes a chronic problem, then we will change this.
 *
 * Parameters:
 *      const char *                - [in]  msg1
 *      const char *                - [in]  msg2 [optional]
 *      const char *                - [in]  msg3 [optional]
 *      const char *                - [in]  msg4 [optional]
 *      const char *                - [in]  msg5 [optional]
 */
void errorSaveRichErrorInfo
(
    const char *msg1,
    const char *msg2,
    const char *msg3,
    const char *msg4,
    const char *msg5
)
{
    clearRichErrorInfo();
    const char **ptr = richErrorParamsForThread();

    if (msg1) ptr[0] = copyString(msg1);
    if (msg2) ptr[1] = copyString(msg2);
    if (msg3) ptr[2] = copyString(msg3);
    if (msg4) ptr[3] = copyString(msg4);
    if (msg5) ptr[4] = copyString(msg5);
}


/**
 *=------------------------------------------------------------------------=
 * errorShowError
 *=------------------------------------------------------------------------=
 * Shows the given error as a message.
 *
 * Parameter:
 *      ERRCODE             - [in]  error to show.
 */
void errorShowError
(
    ERRCODE er
)
{
    const char **richError;
    const char * pstr;
    ErrorDialog ed(localMainWidget, "error", true, 0);

    ed.cancelB->setFocus();

    richError = richErrorParamsForThread();
    if (!richError) return;

    errorConstructFullErrorString(er, pstr);
    if (pstr == NULL) {
        fprintf(stderr, "Catastrophic failure:  Unable to show error!!!\n");
        return;
    }

    ed.mainLabelL->setText(pstr);
    localFree(pstr);
    ed.exec();
}

/**
 *=------------------------------------------------------------------------=
 * lerrorInit
 *=------------------------------------------------------------------------=
 * initializes the error handling code.
 */
void lerrorInit()
{
}


/**
 *=------------------------------------------------------------------------=
 * lerrorTerm
 *=------------------------------------------------------------------------=
 * terminates the error handling code and cleans up.
 */
void lerrorTerm()
{
    clearRichErrorInfo();
}



/**
 *=------------------------------------------------------------------------=
 * richErrorParamsForThread
 *=------------------------------------------------------------------------=
 * returns the rich error parameter array for this thread.  They can be set
 * or read here.
 *
 * Output:
 *      const char **               - a 1x5 array of const char 8 ptrs.
 */
const char **richErrorParamsForThread()
{
    ThreadRichErrorInfo *info;
#ifndef WIN32
    pthread_t currentId;
#else
    DWORD currentId;
#endif // WIN32
    ERRCODE er;

    errorMutex.lock();

    /**
     * Get the current id -- this is how it is keyed in the list.
     */
#ifndef WIN32
    currentId = pthread_self();
#else
    currentId = GetCurrentThreadId();
#endif // WIN32

    /**
     * now whiz through the list looking for the threadid to match this one.
     */
    er = richErrorInfo.firstElement(&info);
    while (er == S_OK) {
        if (info->threadId == currentId)  {
            errorMutex.unlock();
            return info->richErrorParams;
        }
        er = richErrorInfo.nextElement(&info);
    }

    /**
     * It wasn't in the list.  go and add a new one.
     */
    info = (ThreadRichErrorInfo *)localAlloc(sizeof(ThreadRichErrorInfo));
    if (!info) return NULL;
    info->threadId = currentId;
    info->richErrorParams = (const char **)localAlloc(sizeof(const char *) * 5);
    memset(info->richErrorParams, 0, sizeof(const char *) * 5);
    if (!info->richErrorParams) return NULL;
    richErrorInfo.addElement(info);


    errorMutex.unlock();
    return info->richErrorParams;
}


