/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>
#include <errno.h>
#include <pthread.h>
#include <qmessagebox.h>
#include <qthread.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "deencode.h"
#include "folderdb.h"
#include "inputbox.h"
#include <intl.h>
#include "kiltdown.h"
#include "localized.h"
#include "maileditwindow.h"
#include "mailaccount.h"
#include "mailfolder.h"
#include "mailprops.h"
#include "main.h"
#include "newacctwiz.h"
#include <properties.h>
#include "submail.h"
#include "uiicons.h"


/**
 *=------------------------------------------------------------------------=
 *  This class implements the Mail Account processing stuff.  Please note  *
 * that some of the code for this class is in "maildownload.cc".  That is  *
 * all the code for uploading and downloading of messages.                 *
 *=------------------------------------------------------------------------=
 */


/**
 * local prototypes
 */
ERRCODE stripNonAlnum(char *, const char *, int);


/**
 * These are lock files for various people who want them.
 */
const char *lockFileNames[2] = {

    ".incomingLock",
    ".compactLock"
};



/**
 *=------------------------------------------------------------------------=
 * Config file name and property names.
 *=------------------------------------------------------------------------=
 */
const char *folderConfigFile = "account-config";
const char *folderCountProp = "FolderCount";
const char *folderNameProp = "Folder.%d.Name";
const char *folderFileNameProp = "Folder.%d.fileName";

const char *userNameProp = "userName";
const char *userOrganizationProp = "userOrganization";
const char *userEmailProp = "userEmail";
const char *userReplyProp = "userReply";
const char *incomingPopProp = "incomingPop";
const char *spoolLocationProp = "spoolLocation";
const char *rememberPopPasswordProp = "rememberPopPasswordProp";
const char *popRemoveFromServerProp = "popRemoveFromServer";
const char *popNoDuplicatesProp = "popNoDuplicates";
const char *popUserNameProp = "popUserName";
const char *popPasswordProp = "popPassword";
const char *outgoingSMTPProp = "outgoingSMTP";
const char *smtpNeedsAuthProp = "smtpNeedsAuth";
const char *usePopForSmtpProp = "usePopForSmtp";
const char *useInsecureSmtpProp = "useInsecureSmtp";
const char *useSecureSmtpProp = "useSecureSmtp";
const char *popSecurityTypeProp = "popSecurityType";
const char *smtpUserNameProp = "smtpUserName";
const char *smtpPasswordProp = "smtpPassword";
const char *rememberSmtpPasswordProp = "rememberSmtpPassword";
const char *smtpPortProp = "smtpPort";
const char *pop3PortProp = "pop3Port";
const char *securePop3PortProp = "securePop3Port";
const char *signatureFileProp = "signatureFile";
const char *serverTimeoutProp = "serverTimeout";

/**
 *=------------------------------------------------------------------------=
 *                          Local Structs and Classes                      *
 *=------------------------------------------------------------------------=
 */



/**
 * This class is used to verify folder names.
 */
class FolderNameChooser : public InputBox {

  public:
    FolderNameChooser(MailAccount *a) : InputBox() { account = a; }

  protected:
    void accept();

  private:
    MailAccount *account;
};

/**
 *=------------------------------------------------------------------------=
 *                              MailAccount                                *
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * MailAccount::MailAccount
 *=------------------------------------------------------------------------=
 * Private little constructor that should never be called.  You should 
 * always use createNewAccount() or openExistingAccount() instead.
 */
MailAccount::MailAccount()
{
    host = NULL;
    parent = NULL;
    accountName = NULL;
    inbox = outbox = deletedItems = sentItems = NULL;

    /**
     * Per account info
     */
    accountName = NULL;
    accountDirectory = NULL;
    sigFile = NULL;

    DEBUG_TEST_START;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::~MailAccount
 *=------------------------------------------------------------------------=
 * Destructor that does any appropriate cleanup.
 */
MailAccount::~MailAccount()
{
    MailFolder *pfolder;
    ERRCODE er ;

    er = folderList.firstElement(&pfolder);
    while (er == S_OK) {
        delete pfolder;
        er = folderList.nextElement(&pfolder);
    }

    /**
     * Clean up our account info stuff.
     */
    if (accountName) localFree(accountName);
    if (accountDirectory) localFree(accountDirectory);
    cleanAccountInfo(&accountInfo);

    DEBUG_TEST_STOP;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::addFolder
 *=------------------------------------------------------------------------=
 * adds a folder with the given name.
 *
 * Parameters:
 *      const char *                    - [in]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::addFolder
(
    const char *name
)
{
    PropertiesFile *props = NULL;
    char folderHolder[96];
    MailFolder *folder;
    bool done = false;
    const char *pstr;
    char buf[128];
    ERRCODE er;
    int len;

    /**
     * First, get a suitable name.
     */
    er = stripNonAlnum(buf, name, sizeof(buf));
    RETURN_ON_FAILURE(er);

    /**
     * Now make sure this isn't an existing file name.
     */
    while (!done) {
        len = strlen(buf);
        er = folderList.firstElement(&folder);
        while (er == S_OK) {

            pstr = folder->getBaseName();
            if (!strncasecmp(pstr, buf, len)) {
                localFree(pstr);
                buf[len++] = 'A';
                buf[len] = '\0';
                break;
            }

            localFree(pstr);
            er = folderList.nextElement(&folder);
            if (er == S_NODATA) done = true;
        }
    }

    /**
     * We need our config file.
     */
    er = getAccountProperties(this->accountDirectory, props);
    GOTO_ON_FAILURE(er, Error);

    /** 
     * Next, create the silly folder and then go and open it.
     */
    er = MailFolder::createFolder(name, buf, this->accountDirectory, NormalFolder,
                                  &folder);
    GOTO_ON_FAILURE(er, Error);
    folder->close();
    delete folder;
    folder = NULL;
    er = MailFolder::openFolder(this, name, buf, this->accountDirectory, props,
                                &folder);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Next, add it to the list of folders.
     */
    er = folderList.addElement(folder);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Update the count of folders.
     */
    props->writeIntProperty(folderCountProp, folderList.countElements());

    /**
     * write out the two properties for this folder.
     */
    sprintf(folderHolder, folderNameProp, folderList.countElements() - 1);
    props->writeStringProperty(folderHolder, name);
    sprintf(folderHolder, folderFileNameProp, folderList.countElements() - 1);
    props->writeStringProperty(folderHolder, buf);

    /**
     * Finally, add it to the UI.
     */
    host->addFolder((IFolder *)folder);

    er = S_OK;

  Error:
    if (FAILED(er)) delete folder;
    if (props) {
        props->close();
        delete props;
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::cleanAccountInfo
 *=------------------------------------------------------------------------=
 * nukes all the dynamic pointers in this struct.
 *
 * Parameters:
 *      MailAccountInfo *               - [in]  nuke me please.
 */
void MailAccount::cleanAccountInfo
(
    MailAccountInfo *inf
)
{
    if (!inf) return;
    if (inf->userName) localFree(inf->userName);
    if (inf->userOrganization) localFree(inf->userOrganization);
    if (inf->userEmail) localFree(inf->userEmail);
    if (inf->incomingPop) localFree(inf->incomingPop);
    if (inf->spoolLocation) localFree(inf->spoolLocation);
    if (inf->popUserName) localFree(inf->popUserName);
    if (inf->popPassword) localFree(inf->popPassword);
    if (inf->outgoingSMTP) localFree(inf->outgoingSMTP);
    if (inf->smtpUserName) localFree(inf->smtpUserName);
    if (inf->smtpPassword) localFree(inf->smtpPassword);
    if (inf->signatureFile) localFree(inf->signatureFile);
    memset(inf, 0, sizeof(MailAccountInfo));
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::copyAccountInfo
 *=------------------------------------------------------------------------=
 * copies an account info struct.
 *
 * Parameters;
 *      MailAccountInfo *           - [in]  desination 
 *      MailAccountInfo *           - [in]  source.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::copyAccountInfo
(
    MailAccountInfo *dest,
    MailAccountInfo *src
)
{
    if (src->userName) {
        dest->userName = copyString(src->userName);
        if (!dest->userName) { goto OOM; }
    }
    if (src->userOrganization) {
        dest->userOrganization = copyString(src->userOrganization);
        if (!dest->userOrganization) { goto OOM; }
    }
    if (src->userEmail) {
        dest->userEmail = copyString(src->userEmail);
        if (!dest->userEmail) { goto OOM; }
    }
    if (src->userReply) {
        dest->userReply = copyString(src->userReply);
        if (!dest->userReply) { goto OOM; }
    }
    if (src->incomingPop) {
        dest->incomingPop = copyString(src->incomingPop);
        if (!dest->incomingPop) { goto OOM; }
    }
    if (src->spoolLocation) {
        dest->spoolLocation = copyString(src->spoolLocation);
        if (!dest->spoolLocation) { goto OOM; }
    }
    if (src->popUserName) {
        dest->popUserName = copyString(src->popUserName);
        if (!dest->popUserName) { goto OOM; }
    }
    if (src->popPassword) {
        dest->popPassword = copyString(src->popPassword);
        if (!dest->popPassword) { goto OOM; }
    }
    if (src->outgoingSMTP) {
        dest->outgoingSMTP = copyString(src->outgoingSMTP);
        if (!dest->outgoingSMTP) { goto OOM; }
    }
    if (src->smtpUserName) {
        dest->smtpUserName = copyString(src->smtpUserName);
        if (!dest->smtpUserName) { goto OOM; }
    }
    if (src->smtpPassword) {
        dest->smtpPassword = copyString(src->smtpPassword);
        if (!dest->smtpPassword) { goto OOM; }
    }
    if (src->signatureFile) {
        dest->signatureFile = copyString(src->signatureFile);
        if (!dest->signatureFile) { goto OOM; }
    }
    dest->rememberPopPassword = src->rememberPopPassword;
    dest->popRemoveFromServer = src->popRemoveFromServer;
    dest->popNoDuplicates = src->popNoDuplicates;
    dest->smtpNeedsAuth = src->smtpNeedsAuth;
    dest->usePopForSmtp = src->usePopForSmtp;
    dest->useSecureSmtp = src->useSecureSmtp;
    dest->popSecurityType = src->popSecurityType;
    dest->rememberSmtpPassword = src->rememberSmtpPassword;
    dest->smtpPort = src->smtpPort;
    dest->pop3Port = src->pop3Port;
    dest->securePop3Port = src->securePop3Port;
    dest->serverTimeout = src->serverTimeout;
    return S_OK;

  OOM:
    if (dest->userName) localFree(dest->userName);
    if (dest->userOrganization) localFree(dest->userOrganization);
    if (dest->userEmail) localFree(dest->userEmail);
    if (dest->userReply) localFree(dest->userReply);
    if (dest->incomingPop) localFree(dest->incomingPop);
    if (dest->spoolLocation) localFree(dest->spoolLocation);
    if (dest->popUserName) localFree(dest->popUserName);
    if (dest->popPassword) localFree(dest->popPassword);
    if (dest->outgoingSMTP) localFree(dest->outgoingSMTP);
    if (dest->signatureFile) localFree(dest->signatureFile);
    if (dest->smtpUserName) localFree(dest->smtpUserName);
    if (dest->smtpPassword) localFree(dest->smtpPassword);
    return E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::createNewAccount
 *=------------------------------------------------------------------------=
 * show the create account wizard and create the account selected.
 *
 * Parameters:
 *      MailSubsystem *                 - [in]  our parent subsystem
 *      const char **                   - [out] the name of the new account.
 *      const char **                   - [out] the subdir of the account.
 *      MailAccount **                  - [out] where to put the new acct.
 *
 * Output:
 *      ERRCODE                         - S_FALSE if the user elected not
 *                                          to create one, S_OK, or error
 *                                          elsewise.
 */
ERRCODE MailAccount::createNewAccount
(
    MailSubsystem *parent,
    const char **paccountName,
    const char **paccountDirectory,
    MailAccount **pnewaccount
)
{
    MailAccountInfo info;
    char newFolderName[64], folderHolder[96];
    STRINGID baseFolders [] = { 
        ID_FOLDER_INBOX,
        ID_FOLDER_OUTBOX,
        ID_FOLDER_SENTITEMS,
        ID_FOLDER_DELETED
    };
    MailFolderType baseFoldersType[] = {        // this is synchronized with 'baseFolders'
        InboxFolder,
        OutboxFolder,
        SentItemsFolder,
        DeletedItemsFolder
    };
    PropertiesFile *accountConfig;
    MailFolder *newFolder;
    struct stat fileInfo;
    char buf[MAXPATHLEN];
    const char *pstr;
    ERRCODE er;
    int res;

    /**
     * Run the wizard and get the information that we need to create the
     * new account.
     */
    er = runCreateAccountWizard(paccountName, &info);
    if (FAILED(er) || er == S_FALSE) return er;

    /**
     * Generate the account directory and save that out;
     */
    er = getDefaultAccountDir(buf, sizeof(buf), *paccountName);
    GOTO_ON_FAILURE(er, Error);
    *paccountDirectory = copyString(buf);
    if (!(*paccountDirectory)) {
        er = E_OUTOFMEMORY;
        goto Error;
    }

    /**
     * Now, see if the directory already exists, and if so warn them, otherwise
     * make it.
     */
    res = stat(buf, &fileInfo);
    if (res == -1) {
        res = mkdir(buf, 0700);
        if (res == -1) {
            er = E_FILEERR;
            goto Error;
        }
    }

    /**
     * Get the properties file for this account.
     */
    er = getAccountProperties(*paccountDirectory, accountConfig);
    GOTO_ON_FAILURE(er, Error);

    /**
     * write out the account properties.
     */
    er = writeBaseConfigProperties(accountConfig, info);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Great.  Now go and create the basic folders they need.
     * 1. Inbox
     */
    for (unsigned int x = 0; x < sizeof(baseFolders) / sizeof(int); x++) {

        pstr = intlLoadString(baseFolders[x]);
        if (!pstr) {
            er = E_OUTOFMEMORY;
            goto Error;
        }

        er = stripNonAlnum(newFolderName, pstr, sizeof(newFolderName));
        er = MailFolder::createFolder(pstr, newFolderName, *paccountDirectory,
                                      baseFoldersType[x], &newFolder);
        if (FAILED(er)) {
            localFree(pstr);
            goto Error;
        }

        sprintf(folderHolder, folderNameProp, x);
        accountConfig->writeStringProperty(folderHolder, pstr);
        sprintf(folderHolder, folderFileNameProp, x);
        er = accountConfig->writeStringProperty(folderHolder, newFolderName);

        /**
         * For each new folder, we need to save it out, so that the
         * type is correctly saved out.
         */
        newFolder->saveConfiguration(accountConfig);

        localFree(pstr);
        newFolder->close();
        delete newFolder;
        newFolder = NULL;

        GOTO_ON_FAILURE(er, Error);
    }

    
    /** 
     * Write out the configuration for the account.
     */
    accountConfig->writeIntProperty(folderCountProp, sizeof(baseFolders) / sizeof(int));
    accountConfig->close();
    delete accountConfig;

    er = S_OK;

  Error:
    /**
     * Clean up.
     */
    if (info.userName) localFree(info.userName);
    if (info.userOrganization) localFree(info.userOrganization);
    if (info.userEmail) localFree(info.userEmail);
    if (info.incomingPop) localFree(info.incomingPop);
    if (info.spoolLocation) localFree(info.spoolLocation);
    if (info.popUserName) localFree(info.popUserName);
    if (info.popPassword) localFree(info.popPassword);
    if (info.outgoingSMTP) localFree(info.outgoingSMTP);

    /**
     * Okay, now that it's been created, just go and open it like normal. Yes
     * this is a little slower than just keeping the stuff I created here open
     * but I really don't want two code paths for the same thing.
     */
    return (FAILED(er)) ? er : openExistingAccount(parent, *paccountName,
                                                   *paccountDirectory,
                                                   pnewaccount);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::deleteFolder
 *=------------------------------------------------------------------------=
 * removes the given folder from the account.
 *
 * Parameters:
 *      MailFolder *            - [in]  make me go away.
 */
void MailAccount::deleteFolder
(
    MailFolder *folder
)
{
    const char *title, *text;
    ERRCODE er;
    int ret;

    title = intlLoadString(ID_ARE_YOU_SURE);
    text = intlLoadString(ID_DELETE_ACCOUNT);
    if (!title || !text) {
        if (title) localFree(title);
        if (text) localFree(text);
        return;
    }

    /**
     * Ask them if they want to remove it from the UI.
     */
    ret = QMessageBox::critical((QWidget *)localMainWidget, title, text, QMessageBox::Yes, QMessageBox::No);
    localFree(title);
    localFree(text);

    if (ret != QMessageBox::Yes) return;

    title = intlLoadString(ID_ARE_YOU_SURE);
    text = intlLoadString(ID_DELETE_ACCT_FILES);
    if (!title || !text) {
        if (title) localFree(title);
        if (text) localFree(text);
        return;
    }

    ret = QMessageBox::critical((QWidget *)localMainWidget, title, text, QMessageBox::Yes, QMessageBox::No);
    localFree(title);
    localFree(text);

    
    /**
     * Okay, we're here -- remove the stuff!!!
     */
    folder->dieDieDIE();
    er = folderList.removeElement(folder);
    if (er != S_OK) {
        K_FAIL("I couldn't remove a folder ... how come??");
        return;
    }

    /**
     * Update the list of accounts in our config file.
     */
    text = folder->getBaseName();
    er = writeFolderList(text);
    localFree(text);
    if (FAILED(er)) {
        errorShowError(er);
    }

    /**
     * Should we delete it??
     */
    if (ret == QMessageBox::Yes) {
        const char *pstr;
        pstr = folder->getBaseName();
        FolderDB::removeFilesForPrefix(pstr, this->accountDirectory);
        if (pstr) localFree(pstr);
    }

    delete folder;
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::deleteSelectedInFolder                           [IAccount]
 *=------------------------------------------------------------------------=
 * deletes the "selected" items in the given folder.  The folder will always
 * be the currently selected folder.
 *
 * Parameters:
 *      IFolder *               - [in]  folder in which to remove items.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::deleteSelectedInFolder
(
    IFolder *folder
)
{
    MailMessage **items = NULL;
    MailFolder *f;
    ERRCODE er;
    int count;

    K_ASSERT(folder, "Somebody's asking me to delete in a NULL folder!!!!");
    if (!folder) return E_PARAM;

    f = dynamic_cast<MailFolder *>(folder);
    if (!f) {
        K_FAIL("Somebody gave me a folder that is not a MailFolder");
        return E_INTERNAL;
    }

    /**
     * get the messages to be deleted
     */
    er = f->getSelectedMessages(&count, &items);
    RETURN_ON_FAILURE(er);

    /**
     * If the folder isn't the deleted items folder, move the items to the
     * deleted items folder.
     */
    if (f != getDeletedItems()) {
        for (int x = 0; x < count; x++) {
            er = getDeletedItems()->addMessage(items[x]);
            if (FAILED(er)) {
                localFree(items);
                return er;
            }
        }
    }

    /**
     * finally, remove the messages from whatever folder is appropriate.
     */
    f->deleteMessages(count, items, true);
    if (items) localFree(items);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::dieDieDIE
 *=------------------------------------------------------------------------=
 * we're going away.  clean up.
 */
void MailAccount::dieDieDIE()
{
    // UNDONE: marcw, 2.21.2001 -- we need to orphan messages that are
    // open again this account right now.  we crash right now.
    host->closing();
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::editNew                                          [IAccount]
 *=------------------------------------------------------------------------=
 * Lets the user edit a new message.  We will show our cute little mail 
 * editor window, and we'll associated a mailmessage with it.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::editNew()
{
    MailEditWindow *wnd;
    ERRCODE er;

    /**
     * We need to associate a mail message with this window so that it knows
     * where to send the message, etc ...
     */
    er = MailEditWindow::composeNew(this, &wnd);
    RETURN_ON_FAILURE(er);

    wnd->show();
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::folderMenuItemSelected
 *=------------------------------------------------------------------------=
 * an item in the folder context menu has been selected
 *
 * Parameters:
 *      int                             - [in]  id
 */
void MailAccount::folderMenuItemSelected
(
    int id
)
{
    IFolder *folder;
    MailFolder *f;
    bool b = false;

    switch (id) {
        /**
         * remove all the messages in this deleted items folder.
         */
        case ID_EMPTY_DELETED:
            f = getDeletedItems();
            if (!f) {
                fprintf(stderr, "Can't find deleted items!!!!\n");
                return;
            }

            parent->getDontConfirmDIDelete(b);
            f->removeAllMessages(!b);
            break;

        case ID_NEW_FOLDER:
            newFolder();
            break;

        case ID_SEARCH:
            fprintf(stderr, "search!\n");
            break;

        case ID_FOLDER_DELETE:
        case ID_RENAME:
            folder = host->getCurrentFolder();
            if (!folder) {
                K_FAIL("Got a folder event, but nobody caused it??\n");
                break;
            }
            f = dynamic_cast<MailFolder *>(folder);
            if (!f) {
                K_FAIL("What kind of folder is this ???\n");
                break;
            }

            if (id == ID_FOLDER_DELETE)
                deleteFolder(f);
            else
                renameFolder(f);
            break;
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getDefaultAccountDir
 *=------------------------------------------------------------------------=
 * Returns the default location for account directories.
 *
 * Parameters:
 *      char *                  - [in]  where to put path.
 *      int                     - [in]  max size of said buffer
 *      const char *            - [in]  account name.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getDefaultAccountDir
(
    char *buf,
    int max,
    const char *accountName
)
{
    ERRCODE er;
    int len;

    /**
     * Get the base directory.
     */
    er = Kiltdown::getKiltdownDirectory(buf, max);
    len = strlen(buf);

    /**
     * If they're calling this routine, then we haven't already computed a 
     * directory for this account, and need to do so now.
     */
    er = stripNonAlnum(buf + len, accountName, max - len);
    RETURN_ON_FAILURE(er);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getAccountInfo
 *=------------------------------------------------------------------------=
 * Loads, from the configuration file, the information about this account.
 *
 * Parameters:
 *      PropertiesFile *            - [in]  where to load from.
 *      MailAccountInfo *           - [in]  struct in which to put results.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getAccountInfo
(
    PropertiesFile *props,
    MailAccountInfo *info
)
{
    const char *jumble = NULL;
    ERRCODE er;

    er = props->getStringProperty(userNameProp, &info->userName);
    GOTO_ON_FAILURE(er, Error);

    er = props->getStringProperty(userOrganizationProp, &info->userOrganization);
    GOTO_ON_FAILURE(er, Error);

    er = props->getStringProperty(userEmailProp, &info->userEmail);
    GOTO_ON_FAILURE(er, Error);

    er = props->getStringProperty(userReplyProp, &info->userReply);
    GOTO_ON_FAILURE(er, Error);

    er = props->getStringProperty(incomingPopProp, &info->incomingPop);
    GOTO_ON_FAILURE(er, Error);

    er = props->getStringProperty(spoolLocationProp, &info->spoolLocation);
    GOTO_ON_FAILURE(er, Error);

    er = props->getBooleanProperty(rememberPopPasswordProp, &info->rememberPopPassword);
    GOTO_ON_FAILURE(er, Error);

    er = props->getBooleanProperty(popRemoveFromServerProp, &info->popRemoveFromServer);
    GOTO_ON_FAILURE(er, Error);

    er = props->getBooleanProperty(popNoDuplicatesProp, &info->popNoDuplicates);
    GOTO_ON_FAILURE(er, Error);

    er = props->getStringProperty(popUserNameProp, &info->popUserName);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Load in the POP3 password and quickly unscramble it.
     */
    er = props->getStringProperty(popPasswordProp, &jumble);
    GOTO_ON_FAILURE(er, Error);
    if (jumble && *jumble) info->popPassword = unjumbleString(jumble);
    if (jumble) localFree(jumble);

    er = props->getStringProperty(outgoingSMTPProp, &info->outgoingSMTP);
    GOTO_ON_FAILURE(er, Error);

    er = props->getBooleanProperty(smtpNeedsAuthProp, &info->smtpNeedsAuth);
    GOTO_ON_FAILURE(er, Error);

    er = props->getBooleanProperty(usePopForSmtpProp, &info->usePopForSmtp);
    GOTO_ON_FAILURE(er, Error);

    er = props->getBooleanProperty(useSecureSmtpProp, &info->useSecureSmtp);
    GOTO_ON_FAILURE(er, Error);

    er = props->getIntProperty(popSecurityTypeProp, (int *)&info->popSecurityType);
    GOTO_ON_FAILURE(er, Error);

    er = props->getStringProperty(smtpUserNameProp, &info->smtpUserName);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Load in the SMTP password and quickly unscramble it.
     */
    jumble = NULL;
    er = props->getStringProperty(smtpPasswordProp, &jumble);
    GOTO_ON_FAILURE(er, Error);
    if (jumble && *jumble) info->smtpPassword = unjumbleString(jumble);
    if (jumble) localFree(jumble);

    er = props->getBooleanProperty(rememberSmtpPasswordProp, &info->rememberSmtpPassword);
    GOTO_ON_FAILURE(er, Error);

    er = props->getIntProperty(smtpPortProp, &info->smtpPort);
    GOTO_ON_FAILURE(er, Error);

    er = props->getIntProperty(pop3PortProp, &info->pop3Port);
    GOTO_ON_FAILURE(er, Error);

    er = props->getIntProperty(securePop3PortProp, &info->securePop3Port);
    GOTO_ON_FAILURE(er, Error);

    er = props->getStringProperty(signatureFileProp, &info->signatureFile);
    GOTO_ON_FAILURE(er, Error);

    er = props->getIntProperty(serverTimeoutProp, &info->serverTimeout);
    GOTO_ON_FAILURE(er, Error);

    er = S_OK;

  Error:
    if (FAILED(er)) {
        cleanAccountInfo(info);
    }
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getAccountProperties
 *=------------------------------------------------------------------------=
 * given the directory for an account, open the properties file that should
 * be associated with said account.
 *
 * Parameters:
 *      const char *                - [in]  directory
 *      PropertiesFile *&           - [out] the propertiesfile for it.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getAccountProperties
(
    const char *acctDir,
    PropertiesFile * &props
)
{
    char buf[MAXPATHLEN];
    int len;

    /**
     * Get the account directory so we can open the config file.
     */
    strncpy(buf, acctDir, sizeof(buf));
    buf[sizeof(buf) -1] = '\0';
    len = strlen(buf);
    if (buf[len - 1] != '/') {
        buf[len++] = '/';
        buf[len] = '\0';
    }
    strcat(buf, folderConfigFile);

    /**
     * Okay, go and load in the config file and see what we've got.
     */
    return PropertiesFile::openPropertiesFile(buf, &props, true);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getCanCompactFolders
 *=------------------------------------------------------------------------=
 * returns true if folders can be compacted ...
 *
 * Output:
 *      bool
 */
bool MailAccount::getCanCompactFolders()
{
    return true;
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::getCloseOnReply
 *=------------------------------------------------------------------------=
 * returns a preference global to all mail accounts.
 *
 * Parameters:
 *      bool &                  - [out] the value of the preference.
 *
 * Output:
 *      void
 */
void MailAccount::getCloseOnReply
(
    bool & b
)
{
    parent->getCloseOnReply(b);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getComposeFont
 *=------------------------------------------------------------------------=
 * returns the font used for Composing mail.
 *
 * Parameters:
 *      const char * &              - [out] you must free me.
 *      int &                       - [out]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getComposeFont
(
    const char * & family,
    int & size
)
{
    return parent->getComposeFont(family, size);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getDefaultFolder                                 [IAccount]
 *=------------------------------------------------------------------------=
 * returns the name of the default folder in this account. CALLER frees
 * resulting string.
 *
 * Parameters:
 *      const char **           - [out]  the name of the default folder.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getDefaultFolder
(
    const char **pfolder
)
{
    MailFolder *f;

    f = getFolderByType(InboxFolder);
    K_ASSERT(f, "account with no inbox -- AUGH!");
    if (!f) return E_INTERNAL;

    return f->getDisplayName(pfolder);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getDeletedItems
 *=------------------------------------------------------------------------=
 * returns the deleted Items folder.  Returns NULL if it can't be found, 
 * which is a catastrophic failure.
 *
 * Output:
 *      MailFolder *
 */
MailFolder * MailAccount::getDeletedItems()
{
    return getFolderByType(DeletedItemsFolder);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getDisplayName                                   [IAccount]
 *=------------------------------------------------------------------------=
 * returns the display name for this account.  very difficult. CALLER 
 * assumes responsibility for freeing returned string.
 *
 * Parameters:
 *      const char **           - [out] the display name for this account.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getDisplayName
(
    const char **pname
)
{
    if (!pname) return E_PARAM;

    K_ASSERT(accountName, "We're being asked for account name we don't have");
    *pname = copyString(accountName);
    return (*pname) ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getFolderByName
 *=------------------------------------------------------------------------=
 * returns the folder object for the given string, or NULL if no such
 * folder exists.
 * 
 * Parameters:
 *      const char *                - [in]  folder name
 *
 * Output:
 *      MailFolder *
 */
MailFolder *MailAccount::getFolderByName
(
    const char *name
)
{
    MailFolder *folder;
    ERRCODE er;

    K_ASSERT(name, "I need a name here dumbass!!!\n");

    if (!name) return NULL;

    er = folderList.firstElement(&folder);
    while (er == S_OK) {
        const char *pstr;        
        if (FAILED(folder->getDisplayName(&pstr)))
            return NULL;
        if (pstr[0] == name[0] && !strcasecmp(name, pstr)) {
            localFree(pstr);
            return folder;
        }
        localFree(pstr);

        er = folderList.nextElement(&folder);
    }

    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getFolderCount                                   [IAccount]
 *=------------------------------------------------------------------------=
 * Returns the number of folders that we have in this account.
 *
 * Parameters:
 *      int *                   - [out] the number of folders that we gots.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getFolderCount
(
    int *pcount
)
{
    if (!pcount) return E_PARAM;
    *pcount = folderList.countElements();
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getFolderList                                    [IAccount]
 *=------------------------------------------------------------------------=
 * Returns an array of pointers to IFolder objects representing the folders
 * in this account. CALLER must free array using localFree()
 *
 * Parameters:
 *      IFolder ***             - [out] where to put an array of pointers.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getFolderList
(
    IFolder ***plist
)
{
    MailFolder *f;
    ERRCODE er;
    int count;
    int i = 0;

    if (!plist) return E_PARAM;

    count = folderList.countElements();

    /**
     * If there are no folders, this is easy!!
     */
    if (!count) {
        *plist = NULL;
        return S_OK;
    }

    *plist = (IFolder **)localAlloc(sizeof(IFolder *) * count);
    if (!(*plist)) return E_OUTOFMEMORY;

    /**
     * Loop through getting the pointers.
     */
    er = folderList.firstElement(&f);
    while (er == S_OK) {
        (*plist)[i++] = f;

        er = folderList.nextElement(&f);
    }

    if (FAILED(er)) 
        localFree(*plist);
    else
        er = S_OK;

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getFolderPopup                                   [IAccount]
 *=------------------------------------------------------------------------=
 * return the popup for the given folder.
 *
 * Parameters:
 *      IFolder *               - [in]  folder
 *      QPopupMenu * &          - [out] new popup menu.  CALLER FREES using
 *                                  'delete'
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getFolderPopup
(
    IFolder *folder,
    QPopupMenu * & popup
)
{
    MailFolderType type;
    const char *pstr;
    MailFolder *f;

    f = dynamic_cast<MailFolder *>(folder);
    if (!f) {
        K_FAIL("Given a very bogus folder item !!!!");
        return E_INTERNAL;
    }

    type = f->getType();

    popup = new QPopupMenu();
    if (!popup) return E_OUTOFMEMORY;

    pstr = intlLoadString(ID_RENAME);
    if (!pstr) goto OOM;
    popup->insertItem(pstr, this, SLOT(folderMenuItemSelected(int)), 0, ID_RENAME);
    localFree(pstr);

    if (type != NormalFolder) popup->setItemEnabled(ID_RENAME, false);

    pstr = intlLoadString(ID_SEARCH);
    if (!pstr) goto OOM;
    popup->insertItem(pstr, this, SLOT(folderMenuItemSelected(int)), 0, ID_SEARCH);
    localFree(pstr);

    /**
     * If it's a normal folder, then "Delete Folder".  If it's a deleted items
     * folder, then "Empty Deleted Items".  Otherwise, nothing.
     */
    if (type == DeletedItemsFolder) {
        
        popup->insertSeparator();
        pstr = intlLoadString(ID_EMPTY_DELETED);
        if (!pstr) goto OOM;
        popup->insertItem(pstr, this, SLOT(folderMenuItemSelected(int)), 0, ID_EMPTY_DELETED);
        localFree(pstr);

    } else if (type == NormalFolder) {

        popup->insertSeparator();
        pstr = intlLoadString(ID_FOLDER_DELETE);
        if (!pstr) goto OOM;
        popup->insertItem(pstr, this, SLOT(folderMenuItemSelected(int)), 0, ID_FOLDER_DELETE);
        localFree(pstr);
    }

    popup->insertSeparator();
    pstr = intlLoadString(ID_NEW_FOLDER);
    if (!pstr) goto OOM;
    popup->insertItem(pstr, this, SLOT(folderMenuItemSelected(int)), 0, ID_NEW_FOLDER);
    localFree(pstr);

    return S_OK;

  OOM:
    delete popup;
    return E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getForwardPrefix
 *=------------------------------------------------------------------------=
 * returns the prefix to use for the included text when forwarding mail.
 *
 * Parameters:
 *      const char *&               - [out] prefix each line with this.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getForwardPrefix
(
    const char *&prefix
)
{
    return parent->getForwardPrefix(prefix);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getHostSubsystem                                 [IAccount]
 *=------------------------------------------------------------------------=
 * returns our subsystem
 *
 * Parameters:
 *      ISubsystem *&           - [out]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getHostSubsystem
(
    ISubsystem * &sub
)
{
    sub = this->parent;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getIgnoreAskOnDelete
 *=------------------------------------------------------------------------=
 * if true, then we NEVER ask if the user wants to delete some messages.
 *
 * Parameters:
 *      bool &                      - [out] the value of this.
 */
void MailAccount::getIgnoreAskOnDelete
(
    bool &askOnDelete
)
{
    parent->getIgnoreAskOnDelete(askOnDelete);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getInbox
 *=------------------------------------------------------------------------=
 * returns the Inbox folder.  Returns NULL if it can't be found, 
 * which is a catastrophic failure.
 *
 * Output:
 *      MailFolder *
 */
MailFolder * MailAccount::getInbox()
{
    return getFolderByType(InboxFolder);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getFolderByType
 *=------------------------------------------------------------------------=
 * given a type of folder, return a MailFolder for it.  If the type has more
 * than one match (NormalFolder), NULL is returned;
 *
 * Parameters:
 *      int                         - [in]  match me.
 *
 * Output:
 *      MailFolder *                - NULL for a bogus type
 */
MailFolder *MailAccount::getFolderByType
(
    int tt
)
{
    MailFolderType t = (MailFolderType)tt;
    K_ASSERT(t != NormalFolder && t != InvalidFolderType, "Bogus folder type!");

    MailFolder *f;
    ERRCODE er;

    er = folderList.firstElement(&f);
    while (er == S_OK) {
        if (f->getType() == t) {
            return f;
        }
        er = folderList.nextElement(&f);
    }

    /**
     * UNDONE: marcw, 2001.2.5 -- I should be able to deal with this case.
     */
    K_FAIL("There's an account without typed folder.  Very bad !!!");
    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getLineWrapCount
 *=------------------------------------------------------------------------=
 * returns the length of a line after which we must wrap.
 *
 * Parameters:
 *      int &                       - [out]
 */
void MailAccount::getLineWrapCount
(
    int &l
)
{
    parent->getLineWrapCount(l);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getOutbox
 *=------------------------------------------------------------------------=
 * returns the Outbox folder.  Returns NULL if it can't be found, 
 * which is a catastrophic failure.
 *
 * Output:
 *      MailFolder *
 */
MailFolder * MailAccount::getOutbox()
{
    return getFolderByType(OutboxFolder);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getPixmap
 *=------------------------------------------------------------------------=
 * returns a pixmap for this account.
 *
 * Parameters:
 *      AccountPixmapSize           - [in]  size of pixmap to get
 *      QPixmap * &                 - [out] duh.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getPixmap
(
    AccountPixmapSize size,
    QPixmap * &pxm
)
{
    pxm = new QPixmap(helpmelgIcon);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getReadFont
 *=------------------------------------------------------------------------=
 * returns the font used for reading mail.
 *
 * Parameters:
 *      const char * &              - [out] you must free me.
 *      int &                       - [out]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getReadFont
(
    const char * & family,
    int & size
)
{
    return parent->getReadFont(family, size);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getReplyPrefix
 *=------------------------------------------------------------------------=
 * returns the prefix to use for the included text when replying to mail.
 *
 * Parameters:
 *      const char *&               - [out] prefix each line with this.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getReplyPrefix
(
    const char *&prefix
)
{
    return parent->getReplyPrefix(prefix);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getSelectPreviousOnDelete
 *=------------------------------------------------------------------------=
 * indicates what message should next be selected when deleting the current
 * message.
 *
 * Parameters:
 *      bool &                      - [out]
 */
void MailAccount::getSelectPreviousOnDelete
(
    bool &f
)
{
    parent->getSelectPreviousOnDelete(f);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getSenderAddress
 *=------------------------------------------------------------------------=
 * returns the email address associated with this account.
 *
 * Parameters:
 *      const char **           - [out] the address associated with this acct.
 *
 * Output:
 *      ERRCODE
 *
 * UNDONE: marcw, 10.11.2000 -- should i add support for multiple email 
 *      addresses per account ????
 */
ERRCODE MailAccount::getSenderAddress
(
    const char **psender
)
{
    *psender = copyString(accountInfo.userEmail);
    return (*psender) ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getSentItems
 *=------------------------------------------------------------------------=
 * returns the SentItems folder.  Returns NULL if it can't be found, 
 * which is a catastrophic failure.
 *
 * Output:
 *      MailFolder *
 */
MailFolder * MailAccount::getSentItems()
{
    return getFolderByType(SentItemsFolder);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::getSignatureFile
 *=------------------------------------------------------------------------=
 * returns the path to the signature file for the user.
 *
 * Parameters:
 *      const char *&           - [out] the path.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::getSignatureFile
(
    const char *&file
)
{
    file = NULL;
    if (!accountInfo.signatureFile) return S_NODATA;
    file = copyString(accountInfo.signatureFile);
    return file ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::isMailAccount                                    [IAccount]
 *=------------------------------------------------------------------------=
 * this is a mail account?
 */
bool MailAccount::isMailAccount()
{
    return true;
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::isOutboxFolder
 *=------------------------------------------------------------------------=
 * Is the given folder (name) the name of a folder that should get default
 * values like an outbox?
 *
 * Parameters:
 *      const char *            - [in]
 *
 * Output:
 *      bool
 */
bool MailAccount::isOutboxFolder
(
    const char *name
)
{
    const char *pstr, *pstr2;
    bool ret;

    if (!name) return false;

    pstr = intlLoadString(ID_FOLDER_OUTBOX);
    pstr2 = intlLoadString(ID_FOLDER_SENTITEMS);

    if (!pstr && !pstr2)
        ret = false;
    else if (name[0] == pstr[0]
        && !strcasecmp(name, pstr))
        ret = true;
    else if (name[0] == pstr2[0]
             && !strcasecmp(name, pstr2))
        ret = true;
    else
        ret = false;

    if (pstr) localFree(pstr);
    if (pstr2) localFree(pstr2);
    return ret;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::mailto                                           [IAccount]
 *=------------------------------------------------------------------------=
 * sends a message to the given person
 *
 * Parameters:
 *      const char *                - [in]  send to me.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::mailto
(
    const char *mailto
)
{
    MailEditWindow *wnd;
    ERRCODE er;

    er = MailEditWindow::mailto(this, mailto, &wnd);
    RETURN_ON_FAILURE(er);

    wnd->show();
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::newFolder                                        [IAccount]
 *=------------------------------------------------------------------------=
 * Creates a new folder in the account.  First, we ask the user what hte
 * name of it should be.
 * 
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::newFolder()
{
    FolderNameChooser ib(this);
    const char *title, *text;
    const char *pstr;
    ERRCODE er;
    int ret;

    title = intlLoadString(ID_FOLDER_NAME);
    text = intlLoadString(ID_ENTER_FOLDER_NAME);
    if (!title || !text) {
        if (title) localFree(title);
        if (text) localFree(text);
        return E_OUTOFMEMORY;
    }

    er = ib.initialize(title, text);
    localFree(title);
    localFree(text);
    RETURN_ON_FAILURE(er);

    /**
     * Run da dialog -- it will make sure the name isn't a duplicate
     * or some such thing.
     */
    ret = ib.exec();
    if (ret != QDialog::Accepted) return S_FALSE;

    /**
     * Now get the name and create that folder.
     */
    ib.getInput(pstr);

    /**
     * Add it !!! This creates the folder, adds it to the UI, and any internal
     * lists, etc.
     */
    er = addFolder(pstr);
    GOTO_ON_FAILURE(er, Error);

    er = S_OK;

  Error:
    if (pstr) localFree(pstr);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::openExistingAccount
 *=------------------------------------------------------------------------=
 * given a full set of parameters for an existing account, go and open it.
 *
 * Parameters:
 *      MailSubsystem *                 - [in]  our parent
 *      const char *                    - [in]  the account's name.
 *      const char *                    - [in]  it's subdirectory.
 *      MailAccount **                  - [out] where to put the new acct.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::openExistingAccount
(
    MailSubsystem *parent,
    const char *accountName,
    const char *accountDirectory,
    MailAccount **pnewaccount
)
{
    PropertiesFile *accountConfig;
    char nameBuf[MAXPATHLEN];
    MailAccountInfo info;
    const char *pstrName;
    MailFolder *folder;
    ERRCODE er;
    char *ptr;
    int count;

    if (!parent || !accountName || !accountDirectory || !pnewaccount) return E_PARAM;

    /**
     * get the config file associated with this account.
     */
    er = getAccountProperties(accountDirectory, accountConfig);
    RETURN_ON_FAILURE(er);

    /**
     * Create the actual account object.
     */
    *pnewaccount = new MailAccount();
    if (!(*pnewaccount)) {
        er = E_OUTOFMEMORY;
        goto Error;
    }

    /**
     * Set the basic properties for this account: The parent the name, and
     * the directory.
     */
    (*pnewaccount)->parent = parent;
    (*pnewaccount)->accountDirectory = copyString(accountDirectory);
    (*pnewaccount)->accountName = copyString(accountName);
    if (!(*pnewaccount)->accountName || !(*pnewaccount)->accountDirectory) {
        er = E_OUTOFMEMORY;
        goto Error;
    }

    /**
     * Load in the properties for this account now.
     */
    memset(&info, 0, sizeof(MailAccountInfo));
    er = (*pnewaccount)->getAccountInfo(accountConfig, &info);
    GOTO_ON_FAILURE(er, Error);
    (*pnewaccount)->accountInfo = info;

    /**
     * find out how many folders we've got.
     */
    er = accountConfig->getIntProperty(folderCountProp, &count);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Now, for each folder, go and load them in.
     */
    for (int x = 0; x < count; x++) {
        pstrName = NULL; ptr = NULL;
        sprintf(nameBuf, folderNameProp, x);
        er = accountConfig->getStringProperty(nameBuf, &pstrName);
        GOTO_ON_FAILURE(er, Error);
        sprintf(nameBuf, folderFileNameProp, x);
        er = accountConfig->getStringProperty(nameBuf, (const char **)&ptr);
        if (FAILED(er)) {
            localFree(pstrName);
            goto Error;
        }
        
        /**
         * Okay, we've got all the info we need -- go and open the folder!!
         */
        er = MailFolder::openFolder(*pnewaccount, pstrName, ptr, 
                                    accountDirectory, accountConfig, &folder);
        localFree(pstrName);
        localFree(ptr);
        GOTO_ON_FAILURE(er, Error);
        
        /**
         * Add it to our list.
         */
        er = (*pnewaccount)->folderList.addElement(folder);
        if (FAILED(er)) {
            delete folder;
            goto Error;
        }
    }

    er = S_OK;

  Error:
    if (FAILED(er)) {
        if (*pnewaccount)
            delete *pnewaccount;
        *pnewaccount = NULL;
    }

    if (accountConfig) {
        accountConfig->close();
        delete accountConfig;
    }
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::releaseDisasterLock
 *=------------------------------------------------------------------------=
 * releases the disaster recovery lock for the given subsystem.
 *
 * Parameters:
 *      DisasterLockType                - [in]  type of lock working with.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::releaseDisasterLock
(
    DisasterLockType lockType
)
{
    char path[MAXPATHLEN];
    struct stat info;
    int len;
    int ret;

    len = strlen(this->accountDirectory);

    /**
     * We need to build a path to this lock file.
     */
    strncpy(path, this->accountDirectory, sizeof(path) - 1);
    path[len++] = '/';
    path[len] = '\0';

    /**
     * Add on the lock file name.
     */
    strncpy(path + len, lockFileNames[lockType], sizeof(path) - len - 1);
    path[len + strlen(lockFileNames[lockType])] = '\0';

    /**
     * Now see if the file exists or not.
     */
    ret = lstat(path, &info);
    if (ret != 0) {
        errorSaveRichErrorInfo(path);
        return E_NO_LOCKFILE;
    }

    /**
     * Delete the lock!
     */
    unlink(path);
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::renameAccount
 *=------------------------------------------------------------------------=
 * renames the account .
 *
 * Parameters:
 *      const char *                    - [in]  new name
 */
void MailAccount::renameAccount
(
    const char *newName
)
{
    /**
     * Save the new name
     */
    if (this->accountName) localFree(accountName);
    this->accountName = copyString(newName);

    /**
     * Tell the host about it.
     */
    host->accountNameChanged(newName);

    /**
     * Tell the mail subsystem to update it's lists of goo.
     */
    parent->accountNameChanged(this, newName);
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::renameFolder
 *=------------------------------------------------------------------------=
 * renams the given folder
 *
 * Parameters:
 *      MailFolder *                   - [in]
 */
void MailAccount::renameFolder
(
    MailFolder *folder
)
{
    const char *currentName = NULL, *newName = NULL;
    FolderNameChooser ib(this);
    const char *title, *text;
    PropertiesFile *props;
    char folderHolder[96];
    MailFolder *f;
    ERRCODE er;
    int ret;

    if (!folder) return;

    /**
     * first, get the current name of the folder.
     */
    folder->getDisplayName(&currentName);

    /**
     * Now show a dialog getting the new name!
     */
    title = intlLoadString(ID_FOLDER_NAME);
    text = intlLoadString(ID_NEW_FOLDER_NAME);
    if (!title || !text) {
        if (title) localFree(title);
        if (text) localFree(text);
        localFree(currentName);
        return;
    }

    er = ib.initialize(title, text);
    localFree(title);
    localFree(text);
    if (FAILED(er)) {
        localFree(currentName);
        return;
    }
    ret = ib.exec();

    /**
     * IF they accepted, then go and put the new name in the properties file
     * and then tell the folder that it has been renamed as well.
     */
    if (ret == QDialog::Accepted) {
        ib.getInput(newName);

        /**
         * We need our config file.
         */
        er = getAccountProperties(this->accountDirectory, props);
        if (FAILED(er)) {
            localFree(currentName);
            localFree(newName);
            return;
        }

        /**
         * Find this folder.
         */
        ret = 0;
        er = folderList.firstElement(&f);
        while (er == S_OK) {
            if (f == folder) break;
            ret++;
            er = folderList.nextElement(&f);
        }
        if (er != S_OK) {
            K_FAIL("Where did this folder come from??\n");
            localFree(currentName);
            localFree(newName);
            delete props;
            return;
        }
        
        snprintf(folderHolder, sizeof(folderHolder), folderNameProp, ret);
        props->writeStringProperty(folderHolder, newName);
        props->close();
        delete props;

        folder->nameChanged(newName);
        localFree(newName);
    }

    localFree(currentName);
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::requestDisasterLock
 *=------------------------------------------------------------------------=
 * somebody we know about wants a disaster lock.  Try and create it for them
 * now.
 *
 * Parameters:
 *      DisasterLockType            - [in]  who wants the lock.
 *      const char *                - [in]  text to write to the file.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::requestDisasterLock
(
    DisasterLockType lockType,
    const char *text
)
{
    char path[MAXPATHLEN];
    struct stat info;
    FILE *lockFile;
    int len;
    int ret;

    len = strlen(this->accountDirectory);

    /**
     * We need to build a path to this lock file.
     */
    strncpy(path, this->accountDirectory, sizeof(path) - 1);
    path[len++] = '/';
    path[len] = '\0';

    /**
     * Add on the lock file name.
     */
    strncpy(path + len, lockFileNames[lockType], sizeof(path) - len - 1);
    path[len + strlen(lockFileNames[lockType])] = '\0';

fprintf(stderr, "Creating lock file: %s\n", path);

    /**
     * Now see if the file exists or not.
     */
    ret = lstat(path, &info);
    if (ret == 0) {
        errorSaveRichErrorInfo(path);
        return E_LOCKFILE_EXISTS;
    }

    /**
     * Great, it doesn't. Go and create it and write it out!
     */
    lockFile = fopen(path, "w+");
    if (!lockFile) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    /**
     * Now write out the user data.
     */
    ret = fwrite(this->accountName, sizeof(char), strlen(this->accountName),
                 lockFile);
    if (ret == (int)strlen(this->accountName)) {

        ret = fwrite("\n", sizeof(char), 1, lockFile);
        if (ret == 1) {
            len = strlen(text);
            ret = fwrite(text, sizeof(char), strlen(text), lockFile);

        } else {

            ret = -1;
        }

    } else {

        ret = -1;
    }

    fclose(lockFile);

    /**
     * If we failed, remove the lock file.
     */
    if (ret == -1) {
        unlink(path);
        errorSaveRichErrorInfo(strerror(errno));
        return E_FILE_ERR;
    }

    /**
     * All is well!
     */    
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::runCreateAccountWizard
 *=------------------------------------------------------------------------=
 * runs the new account wizard and returns the data that the user specified
 * to create the new account with.
 *
 * Parameters:
 *      const char **               - [out] the account's name.
 *      MailAccountInfo *           - [in]  where to put the account's info.
 *
 * Output:
 *      ERRCODE                     - S_FALSE is returned if they don't create
 *                                      an account.
 */
ERRCODE MailAccount::runCreateAccountWizard
(
    const char **paccountName,
    MailAccountInfo *info
)
{
    NewAccountWizard naw(localMainWidget, "wiz", true, 0);
    ERRCODE er;

    int result = naw.exec();

    /**
     * If an account was fully specified, go and create that now.
     */
    if (result == QDialog::Accepted) {
        er = naw.getAccountInformation(*paccountName, *info);
        RETURN_ON_FAILURE(er);
        er = S_OK;
    } else {
        er = S_FALSE;
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::saveConfiguration
 *=------------------------------------------------------------------------=
 * saves out our current configuration, and that of our folders.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::saveConfiguration()
{
    PropertiesFile *props;
    ERRCODE er;
    MailFolder *f;

    /**
     * get our config file ...
     */
    er = getAccountProperties(this->accountDirectory, props);

    /**
     * Write out the base account properties first!!!
     */
    er = writeAllConfigProperties(props);
    if (FAILED(er)) {
        delete props;
        return er;
    }

    /**
     * Now save out the individual folders, one at a time!!!
     */
    er = folderList.firstElement(&f);
    while (er == S_OK) {
        f->saveConfiguration(props);

        er = folderList.nextElement(&f);
    }

    props->close();
    delete props;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::sendMessage
 *=------------------------------------------------------------------------=
 * put this message in our outbox, which then queues it up for transport
 * later on.  If we are re-editing a message that was already in the outbox,
 * then we need to go and remove the old message and put the new one in
 * instead ...
 * 
 * Parameters:
 *      MailMessage *           - [in]  send me please.
 *      MailMessage *           - [in]  message we are replacing.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::sendMessage
(
    MailMessage *message,
    MailMessage *replacing
)
{
    MailFolder *f;
    ERRCODE er;

    if (!message) return E_PARAM;

    /**
     * We need to find this message -- first look in the outbox, then look in
     * all of the folders.
     */
    if (replacing) {
        f = getOutbox();

        if (f->findMessage(replacing) == S_OK)
            f->removeMessage(replacing);
        else {
            er = folderList.firstElement(&f);
            while (er == S_OK) {
                if (f != getOutbox() && f->findMessage(replacing) == S_OK) {
                    f->removeMessage(replacing);
                    break;
                }
                er = folderList.nextElement(&f);
            }
        }
    }

    /**
     * First, set the status of this message to unsent.
     */
    message->setStatus(MSG_UNSENT);

    /**
     * get our outbox
     */
    MailFolder *outbox = getOutbox();
    if (!outbox) {
        return E_MISSINGFILE;
    }

    /**
     * Add the message to the outbox.  The folder object does all the 
     * crit-section stuff here to make sure that no other threads are
     * also accessing this folder.
     */
    return outbox->addMessage(message);
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::setAccountHost                                    [IAccount]
 *=------------------------------------------------------------------------=
 * sites us.
 *
 * Parameters:
 *      IAccountHost *      - [in]  the new host.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::setAccountHost
(
    IAccountHost *phost
)
{
    K_ASSERT(phost, "Bad phost or view.  I'm the boss.  Need the info!");

    host = phost;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::showAccountProperties                            [IAccount]
 *=------------------------------------------------------------------------=
 * show the properties dialog for this account with the given dialog parent.
 *
 * Parameters:
 *      QWidget *                   - [in]  dialog parent.
 */
void MailAccount::showAccountProperties
(
    QWidget *parent
)
{
    MailAccountInfo updatedInfo;
    PropertiesFile *props;
    const char *newName = NULL;
    ERRCODE er;
    int ret;

    /**
     * Create the dialog and show it.
     */
    MailProperties mpd(accountName, accountDirectory, accountInfo, parent,
                         "props", true, 0);
    ret = mpd.exec();
    if (ret != QDialog::Accepted) return;

    /**
     * Get what was returned, and figure out if we need to rename the
     * account or not.
     */
    memset(&updatedInfo, 0, sizeof(updatedInfo));
    er = mpd.getAccountInformation(newName, updatedInfo);
    GOTO_ON_FAILURE(er, Error);
    if (newName) renameAccount(newName);

    /**
     * Now save it all out !!!
     */
    er = getAccountProperties(this->accountDirectory, props);
    GOTO_ON_FAILURE(er, Error);
    er = writeAllConfigProperties(props);
    GOTO_ON_FAILURE(er, Error);
    props->close();

    /**
     * Everything is ok -- copy over the new account info
     */
    cleanAccountInfo(&accountInfo);
    memcpy(&accountInfo, &updatedInfo, sizeof(updatedInfo));
    memset(&updatedInfo, 0, sizeof(updatedInfo));        

    er = S_OK;

  Error:

    if (props) delete props;
    if (newName) localFree(newName);
    if (FAILED(er)) {
        cleanAccountInfo(&updatedInfo);
        errorShowError(er);
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::terminateAccount                                 [IAccount]
 *=------------------------------------------------------------------------=
 * terminates our interactions within our host.  The IAccountHost pointer
 * is invalid after this routine is called.
 */
void MailAccount::terminateAccount()
{
    host = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::writeAllConfigProperties
 *=------------------------------------------------------------------------=
 * writes out all config properites, not just those used to set up a new
 * account.
 *
 * Parameters:
 *      PropertiesFile *            - [in]  where to write to.
 *
 * Output:
 */
ERRCODE MailAccount::writeAllConfigProperties
(
    PropertiesFile *propertiesFile
)
{
    ERRCODE er;
    
    /**
     * write out the basic ones.
     */     
    er = writeBaseConfigProperties(propertiesFile, this->accountInfo);
    RETURN_ON_FAILURE(er);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::writeBaseConfigProperties
 *=------------------------------------------------------------------------=
 * given a properties file and the informatino about an account, writes
 * out said information to said properties file.  These are the basic properties
 * required to create a new account.
 *
 * Parameters:
 *      PropertiesFile *            - [in]  where to write to.
 *      MailAccountInfo &           - [in]  what to write.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::writeBaseConfigProperties
(
    PropertiesFile *propertiesFile,
    MailAccountInfo &info
)
{
    ERRCODE er;

    /**
     * start writing them out one at a time.  There's no easy way to do this.
     * unfortunately.  [Note:  I suppose I could create an array with offsets
     * and the like, but it seems like an awful hassle ...]
     */
    er = propertiesFile->writeStringProperty(userNameProp, info.userName);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeStringProperty(userOrganizationProp, info.userOrganization);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeStringProperty(userEmailProp, info.userEmail);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeStringProperty(userReplyProp, info.userReply);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeStringProperty(incomingPopProp, info.incomingPop);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeStringProperty(spoolLocationProp, info.spoolLocation);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeStringProperty(popUserNameProp, info.popUserName);
    RETURN_ON_FAILURE(er);
    if (info.rememberPopPassword && info.popPassword) {
        const char *jumble = jumbleString(info.popPassword);
        er = propertiesFile->writeStringProperty(popPasswordProp, jumble);
        localFree(jumble);
        RETURN_ON_FAILURE(er);
    } else {
        er = propertiesFile->writeStringProperty(popPasswordProp, NULL);
    }
    er = propertiesFile->writeStringProperty(outgoingSMTPProp, info.outgoingSMTP);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeBooleanProperty(rememberPopPasswordProp, info.rememberPopPassword);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeBooleanProperty(popRemoveFromServerProp, info.popRemoveFromServer);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeBooleanProperty(popNoDuplicatesProp, info.popNoDuplicates);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeBooleanProperty(smtpNeedsAuthProp, info.smtpNeedsAuth);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeBooleanProperty(usePopForSmtpProp, info.usePopForSmtp);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeBooleanProperty(useSecureSmtpProp, info.useSecureSmtp);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeIntProperty(popSecurityTypeProp, (int)info.popSecurityType);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeStringProperty(smtpUserNameProp, info.smtpUserName);
    RETURN_ON_FAILURE(er);
    if (info.rememberSmtpPassword && info.smtpPassword) {
        const char *jumble = jumbleString(info.smtpPassword);
        er = propertiesFile->writeStringProperty(smtpPasswordProp, jumble);
        localFree(jumble);
        RETURN_ON_FAILURE(er);
    } else {
        er = propertiesFile->writeStringProperty(smtpPasswordProp, NULL);
    }
    er = propertiesFile->writeBooleanProperty(rememberSmtpPasswordProp, info.rememberSmtpPassword);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeIntProperty(smtpPortProp, info.smtpPort);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeIntProperty(pop3PortProp, info.pop3Port);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeIntProperty(securePop3PortProp, info.securePop3Port);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeStringProperty(signatureFileProp, info.signatureFile);
    RETURN_ON_FAILURE(er);
    er = propertiesFile->writeIntProperty(serverTimeoutProp, info.serverTimeout);
    RETURN_ON_FAILURE(er);

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::writeFolderList
 *=------------------------------------------------------------------------=
 * writes out the current list of folders to our config file.
 *
 * Parameters:
 *      const char *                    -[in]   OPTIONAL: clear our any fields
 *                                          i might have.
 */
ERRCODE MailAccount::writeFolderList
(
    const char *deleting
)
{
    const char *pstr1 = NULL, *pstr2 = NULL;
    PropertiesFile *accountConfig;
    char buf[MAX_NAMEVALUE_LEN];
    MailFolder *f;
    ERRCODE er;
    int ct;

    /**
     * Get the properties file for this account.
     */
    er = getAccountProperties(this->accountDirectory, accountConfig);
    GOTO_ON_FAILURE(er, Error);

    /**
     * If we're deleting a folder, then go and clear out its properties.
     */
    if (deleting) {
        
    }

    /**
     * Now write out the count of folders.
     */
    er = accountConfig->writeIntProperty(folderCountProp, folderList.countElements());
    GOTO_ON_FAILURE(er, Error);

    /**
     * Now write out the relevant info for each folder.
     */
    ct = 0;
    er = folderList.firstElement(&f);
    while (er == S_OK) {
        f->getDisplayName(&pstr1);
        pstr2 = f->getBaseName();
        if (!pstr1 || !pstr2) {
            goto Error;
        }
        snprintf(buf, sizeof(buf), folderNameProp, ct);
        er = accountConfig->writeStringProperty(buf, pstr1);
        GOTO_ON_FAILURE(er, Error);
        snprintf(buf, sizeof(buf), folderFileNameProp, ct);
        er = accountConfig->writeStringProperty(buf, pstr2);
        GOTO_ON_FAILURE(er, Error);
        localFree(pstr1);
        localFree(pstr2);

        er = folderList.nextElement(&f);
        ct++;
    }

    GOTO_ON_FAILURE(er, Error);

    /** 
     * Now write out one extra element to clear things out.
     */
    snprintf(buf, sizeof(buf), folderNameProp, ct);
    accountConfig->writeStringProperty(buf, NULL);
    snprintf(buf, sizeof(buf), folderFileNameProp, ct);
    er = accountConfig->writeStringProperty(buf, NULL);

    accountConfig->close();

    er = S_OK;

  Error:
    delete accountConfig;
    return er;
}




/**
 *=------------------------------------------------------------------------=
 *                             Misc Routines                               *
 *=------------------------------------------------------------------------=
 */


/**
 *=------------------------------------------------------------------------=
 * stripNonAlnum
 *=------------------------------------------------------------------------=
 * takes a name and puts it in the given buf, stripping out non alpha
 * numeric characters.  the output will not exceed the number of characters
 * in the third parameter.
 *
 * Parameters:
 *      char *                      - [in]  resulting buffer.
 *      const char *                - [in]  string to strip.
 *      int                         - [in]  max len of output.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE stripNonAlnum
(
    char *buf,
    const char *stripMe,
    int max
)
{
    int i = 0, j = 0;

    if (!buf || !stripMe) return E_PARAM;

    while (stripMe[i] != '\0' && (j < (max - 1))) {
        if (isalnum(stripMe[i]))
            buf[j++] = stripMe[i];
        i++;
    }
    buf[j] = '\0';
    return S_OK;
}
 


/**
 *=------------------------------------------------------------------------=
 *                          FolderNameChooser                              *
 *=------------------------------------------------------------------------=
 */

/**
 *=------------------------------------------------------------------------=
 * FolderNameChooser::accept
 *=------------------------------------------------------------------------=
 * verifies that the given name is a valid one, and doesn't represent the 
 * name of an already existing folder...
 */
void FolderNameChooser::accept()
{
    const char *pstr = (this->edit != NULL) ? (const char *)edit->text() : NULL;
    const char *fname;
    IFolder **folders;
    bool ok = false;
    ERRCODE er;
    int x = 0;
    int c;

    er = account->getFolderList(&folders);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }
    account->getFolderCount(&c);

    /** 
     * IF there's no string, or the string doesn't have a letter or number
     * in it, then it's not okay.  It's also bad if there's a folder with
     * the same name already
     */
    if (!pstr) {
        ok = false;
    } else {

        x = 0;
        while (pstr[x] != '\0') {
            if (isalnum(pstr[x])) {
                ok = true;
                break;
            }
            x++;
        }

        /**
         * Now, make sure we don't have this folder already ...
         */
        if (ok) {
            for (x = 0; x < c; x++) {
                folders[x]->getDisplayName(&fname);

                if (fname && !strcasecmp(fname, pstr)) {
                    ok = false;
                    localFree(fname);
                    break;
                }
                localFree(fname);
            }
        }
    }

    localFree(folders);

    if (!ok) {
        errorShowError(E_INVALID_NAME);
        return;
    }

    InputBox::accept();
}

