/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _MAILACCOUNT_H_

#include <sanity.h>

#include <qpopupmenu.h>

#include <iaccount.h>
#include <llist.h>

/**
 * Forward decl.
 */
class PropertiesFile;
class MailMessage;
class MailColumnInfo;
class MailFolder;
class MailSubsystem;
struct DownloadInfo;

/**
 * These are the types of pop connections that can be made.
 */
enum PopSecurity {

    PopNormal = 0,
    PopOverSSL,
    PopApop
};

/**
 * Various types of disaster recovery locks we can create.
 */
enum DisasterLockType {

    IncomingDatabaseDisaster = 0,
    CompactDatabaseDisaster
};


/**
 * This struct is used to define the contents of a mail account.  It is used
 * to initialize the contents of a mail account.
 */
struct MailAccountInfo {

    const char *userName;
    const char *userOrganization;
    const char *userEmail;
    const char *userReply;
    const char *incomingPop;
    const char *spoolLocation;
    bool rememberPopPassword;
    bool popRemoveFromServer;
    bool popNoDuplicates;
    int  popDeleteAfterDays;
    const char *popUserName;
    const char *popPassword;
    const char *outgoingSMTP;
    bool smtpNeedsAuth;
    bool usePopForSmtp;
    bool useSecureSmtp;
    PopSecurity popSecurityType;
    const char *smtpUserName;
    const char *smtpPassword;
    bool rememberSmtpPassword;
    int smtpPort;
    int pop3Port;
    int securePop3Port;
    const char *signatureFile;
    int serverTimeout;
};


/**
 * this class implements the mail subsystem for our mail program.  It
 * understands basic internet mail messages and MIME extensions as well.
 */

class MailAccount : public QObject, public IAccount {

    Q_OBJECT

    DEBUG_TEST_DECL;

  public:

    /**
     * Runs the new account wizard and creates a new mail account.  The 
     * CALLER must 'delete' the returned MailAccount * object, and 
     * localFree() the returned account name and directory
     */
    static ERRCODE createNewAccount(MailSubsystem *parent,
                                    const char **pname,
                                    const char **pdirectory,
                                    MailAccount **);

    /**
     * Opens the existing account given it's name and directory
     * parameter.  The CALLER must 'delete' the returned MailAccount *
     * object.
     */
    static ERRCODE openExistingAccount(MailSubsystem *parent,
                                       const char *name,
                                       const char *directory,
                                       MailAccount **);

    /**
     * IAccount methods.
     */
    virtual ERRCODE deleteSelectedInFolder(IFolder *);
    virtual ERRCODE downloadProgress(void *);
    virtual ERRCODE editNew();
    virtual bool    getCanCompactFolders();
    virtual ERRCODE getDefaultFolder(const char **);
    virtual ERRCODE getDisplayName(const char **);
    virtual ERRCODE getDownloadInfo(ERRCODE (**pfn)(IAccount *, void *),
                                    void **);
    virtual ERRCODE getFolderCount(int *);
    virtual ERRCODE getFolderList(IFolder ***);
    virtual ERRCODE getFolderPopup(IFolder *, QPopupMenu * &);
    virtual ERRCODE getHostSubsystem(ISubsystem * &);
    virtual ERRCODE getPixmap(AccountPixmapSize, QPixmap *&);
    virtual bool    isMailAccount();
    virtual ERRCODE mailto(const char *);
    virtual ERRCODE newFolder();
    virtual void    passwordDialogResponse(PasswordDialogResponse,
                                           const char *, const char *, bool);
    virtual void    showAccountProperties(QWidget *);
    virtual ERRCODE setAccountHost(IAccountHost *);
    virtual void    terminateAccount();

    /**
     * This is called by the downloading thread to go and actually do the
     * work of downloading the messages.
     */
    static ERRCODE downloadFunction(IAccount *, void *);

    /**
     * This is the email address of the sender of messages in this account.
     */
    ERRCODE getSenderAddress(const char **);

    /**
     * These are preferences global to all mail accounts:
     */
    void    getCloseOnReply(bool &b);
    ERRCODE getComposeFont(const char * & family, int & point);
    ERRCODE getForwardPrefix(const char *&);
    ERRCODE getReadFont(const char *&family, int &point);
    ERRCODE getReplyPrefix(const char *&);
    void    getIgnoreAskOnDelete(bool &f);
    void    getSelectPreviousOnDelete(bool &f);
    void    getLineWrapCount(int &l);
    void    renameAccount(const char * newName);

    /**
     * Preferences unique to this account.
     */
    inline bool hasSignatureFile() { return accountInfo.signatureFile != NULL; }
    ERRCODE getSignatureFile(const char *&);

    /**
     * Saves out our configuration.
     */
    ERRCODE saveConfiguration();

    /**
     * Info about us.
     */
    inline const char *getAccountNameRef() { return this->accountName; }
    inline const char *getAccountDirRef() { return this->accountDirectory; }

    /**
     * Sends the given message
     */
    ERRCODE sendMessage(MailMessage *, MailMessage *replace = NULL);

    /**
     * These methods are used to create and delete disaster locks for objects
     * in this subsystem.
     */
    ERRCODE requestDisasterLock(DisasterLockType, const char *text);
    ERRCODE releaseDisasterLock(DisasterLockType);
    
    /**
     * You're going away. clean up all your UI.
     */
    void dieDieDIE();

    /**
     * returns the given folder.
     */
    MailFolder *getDeletedItems();
    MailFolder *getInbox();
    MailFolder *getOutbox();
    MailFolder *getSentItems();
    MailFolder *getFolderByType(int);

    virtual ~MailAccount();

    /**
     * This is our host.
     */
    IAccountHost *host;

  private:
    /**
     * Use createNew() instead please.
     */
    MailAccount();

    /**
     * Runs the wizard to create a new account.
     */
    static ERRCODE runCreateAccountWizard(const char **, MailAccountInfo *);

    /**
     * Write out the properties to our config file.  It will not
     * close then file when it's done.  This is typically just used for 
     * creating new accounts.
     */
    static ERRCODE writeBaseConfigProperties(PropertiesFile *, MailAccountInfo &);

    /**
     * Save out all the config properties for this account.
     */
    ERRCODE writeAllConfigProperties(PropertiesFile *);

    /**
     * Loads in the account information from our config file.
     */
    ERRCODE getAccountInfo(PropertiesFile *, MailAccountInfo *);

    /**
     * Get the given folder.
     */
    MailFolder *getFolderByName(const char *folder);

    /**
     * indicates whether the given folder is an outbox folder name or not.
     */
    static bool isOutboxFolder(const char *);

    /**
     * copies a MailAccountInfo struct.
     */
    static ERRCODE copyAccountInfo(MailAccountInfo *dest, MailAccountInfo *src);

    /**
     * Gets the directory for the given account name.
     */
    static ERRCODE getDefaultAccountDir(char *, int, const char *name);

    /**
     * Opens the account's config file.
     */
    static ERRCODE getAccountProperties(const char *, PropertiesFile *& props);

    /**
     * Cleans everything up in a MailAccountInfo struct
     */
    static void cleanAccountInfo(MailAccountInfo *nukeMe);

    /**
     * Add a folder with the given name to this account, making sure to
     * update all the configuration files as appropriate.
     */
    ERRCODE addFolder(const char *name);

    /**
     * renames the given folder.
     */
    void renameFolder(MailFolder *);

    /**
     * Deletes the given folder.  Fweee!
     */
    void deleteFolder(MailFolder *);

    /**
     * These are the routines that do the downloading and uploading from
     * downloadFunction()
     */
    static ERRCODE downloadPopMessages(IAccount *, DownloadInfo *);
    static ERRCODE downloadSpoolMessages(IAccount *, DownloadInfo *);
    static ERRCODE uploadMessages(IAccount *, void *);

    /**
     * Updates the list of folders in our account config files.
     */
    ERRCODE writeFolderList(const char *deleted = NULL);

    /**
     * This is our list of folders.
     */
    LinkList<MailFolder *> folderList;

    /**
     * Our parent
     */
    MailSubsystem *parent;

    /**
     * This is all the information about our account.
     */
    const char *accountName;
    const char *accountConfigFile;
    const char *accountDirectory;
    MailAccountInfo accountInfo;

    /**
     * these are some preferences.
     */
    const char *sigFile;

    /**
     *=----------------------------------------------------------------------=
     * Please don't ever use these !!!! always use getInbox()/getOutbox(), etc
     * these will often be null until the above routines are called to set 
     * them. !!!
     *=----------------------------------------------------------------------=
     */
    MailFolder *inbox;
    MailFolder *outbox;
    MailFolder *sentItems;
    MailFolder *deletedItems;

#ifdef K_DEBUG
    const char *test;
#endif // K_DEBUG


  private slots:
    void folderMenuItemSelected(int);
};







#define _MAILACCOUNT_H_
#endif //  _MAILACCOUNT_H_



