/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <unistd.h>

#include "folderdb.h"
#include "kdispatch.h"
#include "localized.h"
#include "mailaccount.h"
#include "mailfolder.h"
#include "mailmessage.h"
#include "mailsmtp.h"
#include "main.h"
#include "pop.h"





const char *incomingdbFile = "incoming";

/**
 *=------------------------------------------------------------------------=
 *                          Local Structs and Classes                      *
 *=------------------------------------------------------------------------=
 */

/**
 * This struct is passed to the download function to get it to do the work
 */
struct DownloadInfo {

    const char *accountName;
    MailAccountInfo info;
    LinkList<MailMessage *> upload;
};



/**
 *=------------------------------------------------------------------------=
 * CrossThreadManager
 *=------------------------------------------------------------------------=
 * This class is a little global queue that is used to hold messages that
 * have been downloaded, but aren't yet saved out in the main application.
 * We save these messages in incoming.db, and then put them in this list. Then
 * we post an event to the main GUI thread, and have it remove the message
 * from this list, add it to the appropriate folder, and then tell this object
 * to remove the message from the incoming.db.  A mutex is used to synchronize
 * the threads.
 *=------------------------------------------------------------------------=
 * Please note that we also use this class to bring up the password dialog
 * on the main GUI thread -- the goal is to have this download thread have 
 * absolutely NO UI whatsoever ...
 */
class CrossThreadManager {

    int x;

  public:
    CrossThreadManager(FolderDB *f) {
        folder = f;
        x = 0;
        userName = NULL;
        userPassword = NULL;
        this->userResponse = NoResponse;
        remember = false;
    }

    void setNamePassword(PasswordDialogResponse resp,
                         const char *name, const char *password, bool r) {

        adderRemover.lock();

        userResponse = resp;
        userName = name;
        userPassword = password;
        remember = r;

        adderRemover.unlock();
    }

    bool getNamePassword(PasswordDialogResponse &response,
                         const char * &name, const char * &password, 
                         bool  &r) {

        adderRemover.lock();

        if (this->userResponse == NoResponse)  {
            adderRemover.unlock();
            return false;
        }

        response = this->userResponse;
        name = userName;
        password = userPassword;
        r = remember;
        userName = userPassword = NULL;

        this->userResponse = NoResponse;
        adderRemover.unlock();
        return true;
    }

    ERRCODE addToDB(MailMessage *m) {

        adderRemover.lock();
        ERRCODE er = folder->addMessage(m);
        folder->flush();
        adderRemover.unlock();
        return er;
    }

    ERRCODE removeFromDB(MailMessage *m) {

        adderRemover.lock();
        ERRCODE er = folder->removeMessage(m);
        folder->flush();
        adderRemover.unlock();
        return er;
    }

    bool hasMessages() {
        bool f;
        adderRemover.lock();
        f = folder->hasMessages() == S_OK;
        adderRemover.unlock();
        return f;
    }

  private:
    FolderDB *folder;
    QMutex adderRemover;

    /**
     * These are the result of showing the password dialog.
     */
    PasswordDialogResponse userResponse;
    int passwordCode;
    const char *userName, *userPassword;
    bool remember;
};


struct CrossThreadInfo {

    bool downloading;
    MailMessage *message;
    CrossThreadManager *far;
};



void updateDownloadStatus(DownloadMessageType t, const char *m);
void updateDownloadStatus(DownloadMessageType t, int count);
void updateDownloadStatus(DownloadMessageType t);

/**
 *=------------------------------------------------------------------------=
 *                    MailAccount Up/Downloading routines                  *
 *=------------------------------------------------------------------------=
 */

/**
 * This is used globally.
 * UNDONE: marcw, 2001.2.2 -- should this be TLS ??/
 */
CrossThreadManager *far = NULL;


/**
 *=------------------------------------------------------------------------=
 * MailAccount::downloadFunction
 *=------------------------------------------------------------------------=
 * it's time to go and send any messages in our outbox, and download any
 * messages to our inbox.
 *
 * Parameters:
 *      IAccount *                  - [in]  the account that's downloading.
 *      void *                      - [in]  the parametesr for this dl
 *
 * Output:
 *      ERRCODE
 *
 * NOTE:  THIS IS ON THE DOWNLOAD THREAD
 */
ERRCODE MailAccount::downloadFunction
(
    IAccount *acct,
    void *params
)
{
    DownloadInfo *info = (DownloadInfo *)params;

    ERRCODE er;

    /**
     * Upload first!!!
     */
    er = uploadMessages(acct, params);
    GOTO_ON_FAILURE(er, Error);

    /**
     * then download as per server type.
     */
    if (info->info.spoolLocation) {
        er = downloadSpoolMessages(acct, info);
    } else {
        er = downloadPopMessages(acct, info);
    }
    GOTO_ON_FAILURE(er, Error);

    er = S_OK;

  Error:
    if (FAILED(er)) {
        const char *perr;

        errorConstructFullErrorString(er, perr);
        if (!perr) fprintf(stderr, "whoooooooooooopsie no error string!!!!\n");

        updateDownloadStatus(AddMessage, perr);
        if (perr) localFree(perr);
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::downloadPopMessages
 *=------------------------------------------------------------------------=
 * does the actual downloading of messages via POP3.  Split this out to be a
 * bit more granular and clear.
 *
 * Parameters:
 *      IAccount *                  - [in]  the account that's downloading.
 *      void *                      - [in]  the parametesr for this dl
 *
 * Output:
 *      ERRCODE
 *
 * NOTE:  THIS IS ON THE DOWNLOAD THREAD
 */
ERRCODE MailAccount::downloadPopMessages
(
    IAccount *acct,
    DownloadInfo *info
)
{
    const char *downloading = NULL;
    FolderDB *incomingdb;
    CrossThreadInfo *cti;
    MailAccount *account;
    const char *pstr;
    int count, size;
    MailMessage *m;
    char buf[256];
    ERRCODE er;

    /**
     * UNDONE: marcw, 12.4.2000 -- timeout should be a configurable property.
     */
    PopConnection popC(60000);  // sixty secs.

    if (!acct || !info) return E_PARAM;

    /**
     * Figure out the real account and the params
     */
    account = dynamic_cast<MailAccount *>(acct);
    if (!account) {
        K_FAIL("Wrong type of account passed in !!!\n");
        return E_INTERNAL;
    }

    /**
     * Set up the UI for this download.
     */
    updateDownloadStatus(InitAccount, account->accountName);

    /**
     * Go and open up an incoming db folder for temporarily receiving
     * messages.
     */
    er = FolderDB::openFolder(NULL, incomingdbFile, account->accountDirectory,
                              true, &incomingdb);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Go and create the object that will manage transactions across
     * threads, etc ..
     */
    K_ASSERT(far == NULL, "this should always be null first !!! \n");
    far = new CrossThreadManager(incomingdb);
    if (!far) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * Now, try to connect and log in.  Keep trying until they succeed
     * or give up.
     */
    do {
        updateDownloadStatus(SetDisplayText, ID_CONNECTING);
        er = popC.connect(info->info.incomingPop, info->info.pop3Port);
        GOTO_ON_FAILURE(er, Error);

        /**
         * Try and log in
         * UNDONE: marcw, 2001-2-06: does POP3 permit blank passwords??
         */
        updateDownloadStatus(SetDisplayText, ID_AUTHORIZING);
        if (info->info.popUserName && info->info.popPassword)
            er = popC.login(info->info.popUserName, info->info.popPassword, PLAIN_TEXT);
        else
            er = E_PASSWORD;

        if (er == E_PASSWORD || er == E_USERNAME) {
            PasswordDialogResponse pdr;
            const char *newname, *newpass;
            bool r;
            QThread::postEvent((QObject *)localMainWidget, 
                               new ShowPasswordDialogEvent((IAccount *)account, 
                                                           account->accountName,
                                                           info->info.popUserName));
            while (!far->getNamePassword(pdr, newname, newpass, r)) {
                usleep(100000);
            }

            /**
             * If the user wants to cancel, do so now !!!
             */
            if (pdr == UserAbort) {
                er = S_FALSE;
                break;
            }

            /**
             * Otherwise, save out the name and password and all that jazz
             */
            if (info->info.popUserName) localFree(info->info.popUserName);
            if (info->info.popPassword) localFree(info->info.popPassword);
            info->info.popUserName = newname;
            info->info.popPassword = newpass;
            info->info.rememberPopPassword = r;
        } else if (FAILED(er)) {
            goto Error;
        }
    } while (er == E_PASSWORD || er == E_USERNAME);

    if (er == S_FALSE) goto Error;
    updateDownloadStatus(SetDisplayText, ID_GATHERINGINFO);
    er = popC.getMessageInfo(&count, &size);
    GOTO_ON_FAILURE(er, Error);

    updateDownloadStatus(SetCount, count);
    downloading = intlLoadString(ID_DOWNLOADING);
    if (!downloading) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * Now whiz through the messages, downloading them, and then sending them
     * on over to the main GUI thread to be added to the appropriate folder.
     */
    for (int x = 0; x < count; x++) {

        snprintf(buf, sizeof(buf), downloading, x + 1, count);
        updateDownloadStatus(SetDisplayText, buf);

        fprintf(stderr, "%d: ----------------\n", x);
        er = popC.downloadMessage(x + 1, info->info.popRemoveFromServer, &m);
        GOTO_ON_FAILURE(er, Error);
        
        updateDownloadStatus(NextMessage);

        /**
         * add it to our incomingdb to keep it safe in case of crash.
         */
        er = far->addToDB(m);
        GOTO_ON_FAILURE(er, Error);

        cti = (CrossThreadInfo *)localAlloc(sizeof(CrossThreadInfo));
        if (!cti) { er = E_OUTOFMEMORY; goto Error; }
        cti->downloading = true;
        cti->far = far;
        cti->message = m;

        QThread::postEvent((QObject *)localMainWidget, new DispatchEvent(acct, (void *)cti));
    }


    pstr = intlLoadString(ID_DLSUCCESS);
    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
    snprintf(buf, sizeof(buf), pstr, account->accountName, count);
    updateDownloadStatus(AddMessage, buf);
    localFree(pstr);

    /**
     * Disconnect cleanly from the server ...
     */
    updateDownloadStatus(SetDisplayText, ID_DISCONNECTING);
    popC.disconnect();

    er = S_OK;

  Error:
    /**
     * Either we've gotten all the messages, or we've had an error and had
     * to abort.  Either way, we need to wait until there are no messages
     * left in the incoming database ...
     */
    if (far) {
        while (1) {
            usleep(100000);
            if (!far->hasMessages()) break;
        }
    }

    /**
     * Make sure this really happened!!!
     */
    if (FAILED(er)) popC.disconnect();

    /**
     * Now close and unlink the database.  We want to be sure that we remove 
     * it when we're done so we don't think there are any messages that 
     * haven't been safely downloaded next time we start.
     */
    incomingdb->close();
    FolderDB::removeFilesForPrefix(incomingdbFile, account->accountDirectory);
    updateDownloadStatus(SetDisplayText, ID_FINISHEDDL);
    if (far) delete far; far = NULL;
    delete incomingdb;

    if (downloading) localFree(downloading);
    cleanAccountInfo(&info->info);
    if (info->accountName) localFree(info->accountName);
    delete info;

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailAccount::downloadSpoolMessages
 *=------------------------------------------------------------------------=
 * downloads the messages that are waiting in a spool file.  Very exciting.
 *
 * Parameters:
 *      IAccount *          - [in]  account with which we're working
 *      DownloadInfo *      - [in]  download information
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailAccount::downloadSpoolMessages
(
    IAccount *acct,
    DownloadInfo *info
)
{
    return E_NOTIMPL;
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::downloadProgress                                 [IAccount]
 *=------------------------------------------------------------------------=
 * a download is in progress.  We are being given a message.  Figure out 
 * where the filters want it to go, add it to the appropriate folder, and
 * then remove it from the incoming db.
 *
 * Parameters:
 *      void *              - [in]  really a CrossThreadInfo struct.
 *
 * Output:
 *      ERRCODE
 *
 * NOTE:  THIS IS ON THE MAIN GUI THREAD
 */
ERRCODE MailAccount::downloadProgress
(
    void *data
)
{
    CrossThreadInfo *cti = (CrossThreadInfo *)data;
    MailFolder *ib;
    ERRCODE er;

    K_ASSERT(cti, "How come I don't have any data yet??\n");
    if (!cti) return E_PARAM;

    /**
     * IF we're downloading, then go and add it to the inbox, and remove it
     * from the incoming db.
     */
    if (cti->downloading) {
        /** 
         * UNDONE: marcw, 12.4.2000 -- add filtering support HERE!!!!
         */
        ib = getInbox();
        er = ib->addMessage(cti->message);
        GOTO_ON_FAILURE(er, Error);
        ib->flush();

        /**
         * now remove it from the incoming db.  this method is thread safe, so
         * we don't have to worry about it here.
         */
        er = cti->far->removeFromDB(cti->message);
        GOTO_ON_FAILURE(er, Error);

    } else {

        /**
         * We've uploaded it -- add it to the Sent Items folder, and remove it
         * from the outbox ...
         */
        MailFolder *si = getSentItems();
        cti->message->setStatus(MSG_READ);
        si->addMessage(cti->message);
        MailFolder *ob = getOutbox();
        ob->removeMessage(cti->message);
        ob->forceRecount();
        si->forceRecount();
    }

    er = S_OK;

  Error:
    cti->message->release();
    localFree(cti);
    return S_OK;
}



/**
 *=------------------------------------------------------------------------=
 * MailAccount::getDownloadInfo                                  [IAccount]
 *=------------------------------------------------------------------------=
 * returns the information need to download messages for this account.  
 *
 * Parameters:
 *      ERRCODE (**pfn)(IAccount *, void *)     - [out] function to call.
 *      void **                                 - [out] params to pass to it.
 *
 * Output:
 *      ERRCODE
 *
 * NOTE:  THIS IS ON THE MAIN GUI THREAD
 */
ERRCODE MailAccount::getDownloadInfo
(
    ERRCODE (**pfn)(IAccount *, void *),
    void **params
)
{
    MailFolder *ob;
    ERRCODE er;

    DownloadInfo *i = new DownloadInfo;
    if (!i) return E_OUTOFMEMORY;
    memset(i, 0, sizeof(DownloadInfo));

    /**
     * Get the information for downloading new messages.
     */
    i->accountName = copyString(this->accountName);
    if (!i->accountName) { er = E_OUTOFMEMORY; goto Error; }

    er = copyAccountInfo(&i->info, &accountInfo);
    if (FAILED(er)) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * next, get the messages for uploading.
     */
    ob = getOutbox();
    er = ob->lockForOutput();
    GOTO_ON_FAILURE(er, Error);

    /**
     * Don't forget that this addref's all the messages.
     */
    er = ob->getMessages(i->upload);
    GOTO_ON_FAILURE(er, Error);

    /**
     * finally set up the params.
     */
    *params = (void *)i;
    *pfn = MailAccount::downloadFunction;
    return S_OK;

  Error:
    if (i->accountName) localFree(i->accountName);
    delete i;
    return er;
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::passwordDialogResponse
 *=------------------------------------------------------------------------=
 * the download thread has asked to see the password dialog.  Here is the
 * response to showing said dialog.
 *
 * Parameters:
 *      PasswordDialogResponse              - [in]  did they enter, or abort?
 *      const char *                        - [in]  new name, if appropriate
 *      const char *                        - [in]  new passwd, if appropriate
 *      bool                                - [in]  remember, if appropriate
 *
 * NOTE:  THIS IS ON THE MAIN GUI THREAD
 */
void MailAccount::passwordDialogResponse
(
    PasswordDialogResponse response,
    const char *name,
    const char *passwd,
    bool remember
)
{
    K_ASSERT(far, "I should definitley have this by now !!!! \n");

    /**
     * I need to save this away in my information too!!!
     */
    if (response == NewNameAndPassword) {
        if (this->accountInfo.popUserName) localFree(this->accountInfo.popUserName);
        if (this->accountInfo.popPassword) localFree(this->accountInfo.popPassword);
        this->accountInfo.popUserName = name ? copyString(name) : name;
        this->accountInfo.popPassword = passwd ? copyString(passwd) : passwd;
        this->accountInfo.rememberPopPassword = remember;
    }

    /**
     * Pass it off to the other thread!
     */
    far->setNamePassword(response, name, passwd, response);
}


/**
 *=------------------------------------------------------------------------=
 * MailAccount::uploadMessages
 *=------------------------------------------------------------------------=
 * go and actually upload the messages to the server.  If there some,
 * of course.
 *
 * Parameters:
 *      IAccount *                  - [in]  the account that's downloading.
 *      void *                      - [in]  the parametesr for this dl
 *
 * Output:
 *      ERRCODE
 *
 * NOTE:  THIS IS ON THE DOWNLOAD THREAD
 */
ERRCODE MailAccount::uploadMessages
(
    IAccount *acct,
    void *params
)
{
    CrossThreadInfo *cti;
    MailAccount *account;
    const char *sending;
    DownloadInfo *info;
    MailMessage *m;
    ReadStatus rs;
    char buf[256];
    ERRCODE er;
    int count;
    int i;

    /**
     * Figure out the parameter information
     */
    if (!acct || !params) return E_PARAM;
    account = dynamic_cast<MailAccount *>(acct);
    if (!account) {
        K_FAIL("Wrong type of account passed in !!!\n");
        return E_INTERNAL;
    }
    info = (DownloadInfo *)params;

    /**
     * Yuck -- first count the number of messages waiting to be sent.
     */
    count = info->upload.countElements();
    er = info->upload.firstElement(&m);
    while (er == S_OK) {
        m->getStatus(rs);
        if (rs != MSG_SENDING) count--;
        er = info->upload.nextElement(&m);
    }

    sending = intlLoadString(ID_SENDINGMSG);
    i = 0;

    updateDownloadStatus(SetCount, count);

    /**
     * Now go and upload any messages that are due to go outbound !!!
     */
    er = info->upload.firstElement(&m);
    while (er == S_OK) {
        
        m->getStatus(rs);
        if (rs == MSG_SENDING) {

            snprintf(buf, sizeof(buf), sending, ++i, count);
            updateDownloadStatus(SetDisplayText, buf);

            /**
             * Upload the message.
             */
            SMTPServerInfo ssi;
            ssi.serverName = info->info.outgoingSMTP;
            ssi.portnum = info->info.smtpPort;
            if (!ssi.serverName) {
                K_FAIL("Need to put a dialog up asking for the server!!!\n");
                ssi.serverName = "localhost";
            }

            int len;
            account->getLineWrapCount(len);
            er = sendSmtpMessage(m, len, &ssi);
            if (FAILED(er)) {
                goto Error;
            }

            updateDownloadStatus(NextMessage);

            cti = (CrossThreadInfo *)localAlloc(sizeof(CrossThreadInfo));
            if (!cti) { er = E_OUTOFMEMORY; goto Error; }
            cti->downloading = false;
            cti->far = NULL;
            cti->message = m;
            m->addRef();
            QThread::postEvent((QObject *)localMainWidget, 
                               new DispatchEvent(acct, (void *)cti));
        }

        er = info->upload.nextElement(&m);
    }

    updateDownloadStatus(SetDisplayText, ID_FINISHEDUL);
    er = S_OK;

  Error:
    er = info->upload.firstElement(&m);
    while (er == S_OK) {
        m->release();
        er = info->upload.nextElement(&m);
    }

    if (sending) localFree(sending);

    return er;
}


/**
 *=------------------------------------------------------------------------=
 *                         Misc Local Routines
 *=------------------------------------------------------------------------=
 */

/**
 *=------------------------------------------------------------------------=
 * updateDownloadStatus
 *=------------------------------------------------------------------------=
 * sends an event to the main GUI thread telling it to update our status
 * with the download window.  These are here just to make things a bit
 * more conveneint and cleaner in the code.
 *
 * Parameters:
 *      DownloadMessageType                 - [in]
 *      const char *                        - [in]  message to send
 */
void updateDownloadStatus
(
    DownloadMessageType type,
    const char *msg
)
{
    QThread::postEvent((QObject *)localMainWidget, 
                       new DownloadMessageEvent(type, msg, 0));
}

/**
 *=------------------------------------------------------------------------=
 * updateDownloadStatus
 *=------------------------------------------------------------------------=
 * sends an event to the main GUI thread telling it to update our status
 * with the download window.  These are here just to make things a bit
 * more conveneint and cleaner in the code.
 *
 * Parameters:
 *      DownloadMessageType                 - [in]
 *      int                                 - [in]  either a string id or a #
 */
void updateDownloadStatus
(
    DownloadMessageType type,
    int ival
)
{
    const char *pstr;

    switch (type) {
        case InitAccount:
        case SetDisplayText:
        case AddMessage:
            pstr = intlLoadString(ival);
            if (!pstr) {
                K_FAIL("Bogus String ID given to updateDownloadStatus\n");
                return;
            }

            updateDownloadStatus(type, pstr);
            localFree(pstr);
            break;
        default:
            QThread::postEvent((QObject *)localMainWidget, 
                               new DownloadMessageEvent(type, NULL, ival));
    }
}

/**
 *=------------------------------------------------------------------------=
 * updateDownloadStatus
 *=------------------------------------------------------------------------=
 * sends an event to the main GUI thread telling it to update our status
 * with the download window.  These are here just to make things a bit
 * more conveneint and cleaner in the code.
 *
 * Parameters:
 *      DownloadMessageType                 - [in]
 */
void updateDownloadStatus
(
    DownloadMessageType type
)
{
    QThread::postEvent((QObject *)localMainWidget, 
                       new DownloadMessageEvent(type, 0, 0));
}

