/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
/**
 * This class is the editor window that you use to compose new mail,
 * replies, or forwards of the above.  It's a top-level window that you can
 * create as many instances of as you'd like ...
 */
#include <sanity.h>

#include <ctype.h>
#include <qclipboard.h>
#include <qfiledialog.h>
#include <qiconview.h>
#include <qlabel.h>
#include <qmessagebox.h>
#include <qsplitter.h>
#include <qstringlist.h>
#include <qtoolbutton.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/stat.h>


#include "about.dlg.h"
#include "addressbook.h"
#include "addrpicker.h"
#include "attachment.h"
#include "maileditwindow.h"
#include "mailfolder.h"
#include <intl.h>
#include "kiltdown.h"
#include <lerror.h>
#include <llist.h>
#include "localized.h"
#include "main.h"
#include "uiicons.h"
#include "uimenu.h"
#include "uitoolbar.h"
#include "userpref.h"

/**
 * Passed in to the constructor for KWindowBase to save us from having to keep
 * track of EVERYTHING
 */
KWBVariables editvars = {

    &prefEditWindowLeft,
    &prefEditWindowTop,
    &prefEditWindowWidth,
    &prefEditWindowHeight,
    &prefEditToolBarVisible,
    &prefEditToolBarUsesBig,
    &prefEditToolBarShowLabels,
    &prefEditToolBarLocation
};

const int MAX_LINE_LEN = 2048;

/**
 * local prototypes
 */
void stripIntoBuf(char *dest, const char *source, int max);
void dumpListIntoEdit(LinkList<const char *> &, const char *, MyMultiLineEdit *, bool addsep = false);

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::MailEditWindow
 *=------------------------------------------------------------------------=
 *  clears out a few pointers, and sets stuff up.  very complex.  Make sure
 * to call initialize before you try to actually do something with this
 * class.
 *
 * Parameters:
 *      MailAccount *               - [in]  the account with which we will
 *                                      send this mesage.
 */
MailEditWindow::MailEditWindow
(
    MailAccount *acct
) 
: KWindowBase(editvars, WDestructiveClose)
{
    K_ASSERT(acct != NULL, "Bogus Message object passed in to Editor!!");

    sourceMessage = NULL;
    account = acct;

    dirty = false;
    editBox = NULL;
    toButton = NULL;
    ccButton = NULL;
    bccButton = NULL;
    subjectLabel = NULL;
    toEdit = ccEdit = bccEdit = NULL;
    subjectEdit = NULL;

    toEditLines = ccEditLines = bccEditLines = -1;

    contentSplitter = NULL;
    attachmentView = NULL;

    priority = NormalPriority;

    updateReadStatus = NoAction;

    DEBUG_TEST_START;
}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::~MailEditWindow
 *=------------------------------------------------------------------------=
 * time to take out the trash ....
 */
MailEditWindow::~MailEditWindow() 
{
    Attachment *a;
    ERRCODE er;

    er = attachments.firstElement(&a);
    while (er == S_OK) {
        attachments.removeElement(a);
        a->release();
        er = attachments.firstElement(&a);
    }

    DEBUG_TEST_STOP;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::aboutToShowEditMenu()
 *=------------------------------------------------------------------------=
 * we are about to show the Edit  menu.  We need to enable and disable items as
 * appropriate.
 */
void MailEditWindow::aboutToShowEditMenu() 
{
    QWidget *wdg = this->focusWidget();
    QLineEdit *e;
    MyMultiLineEdit *me;
    bool enable = false;

    e = dynamic_cast<QLineEdit *>(wdg);
    if (e) {
        enable = e->hasMarkedText();        
    } else {
        me = dynamic_cast<MyMultiLineEdit *>(wdg);
        if (me)
            enable = me->hasMarkedText();
    }

    if (enable) {
        menuBar()->setItemEnabled(ID_EDIT_CUT_TEXT, true);
        menuBar()->setItemEnabled(ID_EDIT_COPY_TEXT, true);
    } else {
        menuBar()->setItemEnabled(ID_EDIT_CUT_TEXT, false);
        menuBar()->setItemEnabled(ID_EDIT_COPY_TEXT, false);
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::aboutToShowFileMenu
 *=------------------------------------------------------------------------=
 * we are about to show the File menu.  We need to enable and disable items as
 * appropriate.
 */
void MailEditWindow::aboutToShowFileMenu() 
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::aboutToShowInsertMenu
 *=------------------------------------------------------------------------=
 * we are about to show the Insert menu.  We need to enable and disable items as
 * appropriate.
 */
void MailEditWindow::aboutToShowInsertMenu()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::aboutToShowMessageMenu
 *=------------------------------------------------------------------------=
 * we are about to show the message menu.  We need to enable and disable items as
 * appropriate.
 */
void MailEditWindow::aboutToShowMessageMenu()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::addAttachmentToDisplay
 *=------------------------------------------------------------------------=
 * given an attachment, adds it to our UI.
 *
 * Parameters:
 *      Attachment *                - [in]  add me.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::addAttachmentToDisplay
(
    Attachment *attach
)
{
    const char *pstr;
    QPixmap *pixmap;
    char buf[128];
    ERRCODE er;

    /**
     * First, pick a pixmap for the attachment.
     */
    pixmap = attach->chooseIcon();
    if (!pixmap) return E_OUTOFMEMORY;

    /**
     * Next, extract the basic file name.
     */
    er = attach->getFileName(&pstr);
    RETURN_ON_FAILURE(er);

    strncpy(buf, pstr, sizeof(buf));
    unsigned int i = strlen(pstr);
    localFree(pstr);
    i = (i >= sizeof(buf)) ? sizeof(buf) - 1 : i;
    buf[i] = '\0';
    strncat(buf, attach->getPrintableSize(), sizeof(buf) - (i + 1));
    
    attachmentView->show();
    new QIconViewItem(attachmentView, buf, *pixmap);
    delete pixmap;

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::addressBook
 *=------------------------------------------------------------------------=
 * asks us to show the address book
 */
void MailEditWindow::addressBook()
{
    AddressBook *b = localMainWidget->getAddressBook();

    if (b) b->show();
}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::askForSave
 *=------------------------------------------------------------------------=
 * Ask the user if they're sure they want to exit, and provide them with a
 * chance to save the changes.
 *
 * Output:
 *      bool            - true if they indicated close, false otherwise.
 */
bool MailEditWindow::askForSave()
{
    const char *title, *text;

    title = intlLoadString(ID_CLOSE_EDIT_TITLE);
    if (!title) return false;
    text = intlLoadString(ID_CLOSE_EDIT_TEXT);
    if (!text) { localFree(title); return false; }

    int ret = QMessageBox::warning(this, title, text, 
                                   QMessageBox::Yes | QMessageBox::Default,
                                   QMessageBox::No,
                                   QMessageBox::Cancel | QMessageBox::Escape);

    localFree(title);
    localFree(text);
    if (ret == QMessageBox::Yes) {
        fileSave();
        return true;
    } else if (ret == QMessageBox::No)
        return true;

    this->setFocus();
    this->setActiveWindow();
    return false;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::attachFile
 *=------------------------------------------------------------------------=
 * the user has indicated that they'd like to attach the given file to the
 * mail message.  go and add it to our list of attached files now !!!
 *
 * Parameters:
 *      const char *                - [in]  full path to attach.
 *
 * Output:
 *      ERRCODE
 *
 * Notes:
 *      Note that we read in the entire file right away, in order to verify
 *      that it exists and can be added to the message no problem!!!
 */
ERRCODE MailEditWindow::attachFile
(
    const char *fileName
)
{
    Attachment *attachment;
    ERRCODE er;

    QString str(fileName);

    if (!fileName) {
        K_FAIL("I should really have a file name!!!");
        return E_PARAM;
    }

    /**
     * Create an attachment object for this file.
     */
    attachment = new Attachment(fileName);
    if (!attachment) return E_OUTOFMEMORY;

    er = addAttachmentToDisplay(attachment);
    if (FAILED(er)) {
        delete attachment;
        return er;
    }

    /**
     * Finally, add it to our list.
     */
    attachments.addElement(attachment);

    dirty = true;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::bccButtonClicked
 *=------------------------------------------------------------------------=
 * the user wants to use the address picker to choose the bccs.
 */
void MailEditWindow::bccButtonClicked()
{
    showAddressPicker(ID_SELECT_BCCS, bccEdit);
}
 

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::ccButtonClicked
 *=------------------------------------------------------------------------=
 * the user wants to use the address picker to choose the ccs.
 */
void MailEditWindow::ccButtonClicked()
{
    showAddressPicker(ID_SELECT_CCS, ccEdit);
}
 


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::checkNames
 *=------------------------------------------------------------------------=
 * checks the names given as input and makes sure they're valid email 
 * addresses, or aliases in our address book that can be converted.
 *
 * Output:
 *      ERRCODE             - S_OK if good, S_FALSE if some are bad, or
 *                              E_ if a fatal error occurred.
 */
ERRCODE MailEditWindow::checkNames()
{
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::closeEvent
 *=------------------------------------------------------------------------=
 * the user has chosen to close the window.  If any changes at all have 
 * been made, then we should ask them if they want to discard them.
 *
 * Parameters:
 *      QCloseEvent *           - [in]  details
 */
void MailEditWindow::closeEvent
(
    QCloseEvent *event
)
{
    if (!event) return;

    if (!dirty || askForSave()) {
        event->accept();
    } else {
        event->ignore();
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::composeNew
 *=------------------------------------------------------------------------=
 * the user wishes to compose a new piece of e-mail.  Just give them a blank
 * window populated with nuthin'.
 *
 * Parameters:
 *      MailAccount *           - [in]  the account in which this operates.
 *      MailMessage **          - [out] the new edit window.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::composeNew
(
    MailAccount *account,
    MailEditWindow **pwindow
)
{
    ERRCODE er;

    if (!pwindow) return E_PARAM;
    
    *pwindow = new MailEditWindow(account);

    er = (*pwindow)->initialize();
    RETURN_ON_FAILURE(er);

    (*pwindow)->editBox->setUndoEnabled(false);

    /**
     * Lastly, insert the user's signature.
     */
    er = (*pwindow)->insertSignature();
    RETURN_ON_FAILURE(er);

    (*pwindow)->editBox->insert("\n");
    (*pwindow)->editBox->setCursorPosition(0, 0);
    (*pwindow)->editBox->setUndoEnabled(true);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::copyOverAttachments
 *=------------------------------------------------------------------------=
 * copies over the attachments to this new window.
 *
 * Parameters:
 *      MailMessage *       - [in]  copy me!!!
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::copyOverAttachments
(
    MailMessage *copyme
)
{
    LinkList<Attachment *> list;
    Attachment *a;
    ERRCODE er;
    bool b;

    /**
     * If there are no attachments to copy over, this is easy!!!
     */
    if (!copyme) return E_PARAM;
    er = copyme->getHasAttachments(&b);
    RETURN_ON_FAILURE(er);
    if (!b) return S_OK;

    /**
     * Now, for each attachment that is in the old message, add it to our
     * list now.
     */
    er = copyme->getAttachments(list);
    RETURN_ON_FAILURE(er);

    er = list.firstElement(&a);
    while (er == S_OK) {
        RETURN_ON_FAILURE(this->attachments.addElement(a))
        addAttachmentToDisplay(a);
        a->addRef();
        er = list.nextElement(&a);
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::createdPopup
 *=------------------------------------------------------------------------=
 * our parent class created a popup menu.  We override this so we can 
 * correctly ssync up on it's aboutToShow() signal.
 *
 * Parameters:
 *      int                 - [in]  id of the popup created.
 *      QPopupMenu *        - [in]  menu created.
 */
void MailEditWindow::createdPopup
(
    int id,
    QPopupMenu *popup
)
{
    switch (id) {
        case ID_FILE_MENU_TEXT:
            QObject::connect(popup, SIGNAL(aboutToShow()),
                             this, SLOT(aboutToShowFileMenu()));
            break;
        case ID_EDIT_MENU_TEXT:
            QObject::connect(popup, SIGNAL(aboutToShow()),
                             this, SLOT(aboutToShowEditMenu()));
            break;
        case ID_INSERT_MENU_TEXT:
            QObject::connect(popup, SIGNAL(aboutToShow()),
                             this, SLOT(aboutToShowInsertMenu()));
            break;
        case ID_MESSAGE_MENU_TEXT:
            QObject::connect(popup, SIGNAL(aboutToShow()),
                             this, SLOT(aboutToShowMessageMenu()));
            break;
        default:
            K_ASSERT(id == ID_HELP_MENU_TEXT, "What other menus are there?");
            break;
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::editCopy
 *=------------------------------------------------------------------------=
 * copies the currently selected text in the given edit control.  once again
 * Qt does all the hard stuff for us.  Fweeeee!
 */
void MailEditWindow::editCopy()
{
    QWidget *wdg = this->focusWidget();
    QLineEdit *e;
    MyMultiLineEdit *me;

    e = dynamic_cast<QLineEdit *>(wdg);
    if (e) {
        e->copy();
    } else {
        me = dynamic_cast<MyMultiLineEdit *>(wdg);
        if (me)
            me->copy();
    }

    // nope -- not an edit, so just skip it.

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::editCut
 *=------------------------------------------------------------------------=
 * cuts the currently selected text in the edit control.
 */
void MailEditWindow::editCut()
{
    QWidget *wdg = this->focusWidget();
    QLineEdit *e;
    MyMultiLineEdit *me;

    e = dynamic_cast<QLineEdit *>(wdg);
    if (e) {
        e->cut();
    } else {
        me = dynamic_cast<MyMultiLineEdit *>(wdg);
        if (me)
            me->cut();
    }

    // nope -- not an edit, so just skip it.
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::editFindMessage
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::editFindMessage()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::editFindPerson
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::editFindPerson()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::editFindText
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::editFindText()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::editPaste
 *=------------------------------------------------------------------------=
 * pastes the current contents of the clipboard into the currently focused
 * edit control.  Fortunately Qt really makes this easy for us!!!!!
 */
void MailEditWindow::editPaste()
{
    QWidget *wdg = this->focusWidget();
    QLineEdit *e;
    MyMultiLineEdit *me;

    e = dynamic_cast<QLineEdit *>(wdg);
    if (e) {
        e->paste();
    } else {
        me = dynamic_cast<MyMultiLineEdit *>(wdg);
        if (me)
            me->paste();
    }

    // nope -- not an edit, so just skip it.
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::editSelectAll
 *=------------------------------------------------------------------------=
 * Selects ALL the text in the currently focused edit control.  Fortunately.
 * Qt Makes this really easy for us.
 */
void MailEditWindow::editSelectAll()
{
    QWidget *wdg = this->focusWidget();
    QLineEdit *e;
    MyMultiLineEdit *me;

    e = dynamic_cast<QLineEdit *>(wdg);
    if (e) {
        e->selectAll();
    } else {
        me = dynamic_cast<MyMultiLineEdit *>(wdg);
        if (me)
            me->selectAll();
    }

    // nope -- not an edit, so just skip it.
}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::fileClose
 *=------------------------------------------------------------------------=
 * Close our window, but only if nothing's happened to it, or we ask and
 * they say they are sure!!!
 */
void MailEditWindow::fileClose()
{
    if (!dirty || askForSave()) {
        dirty = false;
        this->close(true);
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::fileNew
 *=------------------------------------------------------------------------=
 *  Allows the user to edit a new message.  This isn't all that hard, 
 * actually, as we've got all the data here.
 */
void MailEditWindow::fileNew()
{
    ERRCODE er;

    MailEditWindow *wnd;

    er = MailEditWindow::composeNew(this->account, &wnd);
    if (FAILED(er)) { 
        errorShowError(er);
        return;
    }

    wnd->show();
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::filePrint
 *=------------------------------------------------------------------------=
 * print out this message.
 */
void MailEditWindow::filePrint()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::fileSave
 *=------------------------------------------------------------------------=
 * Lets the user save this message to a file somewhere on their computer.
 */
void MailEditWindow::fileSave()
{
    char buf[MAXPATHLEN];
    const char *pstr;
    QString file;
    ERRCODE er;
    char *ptr;
    int len;

    /**
     * We'll just use their home directory as the starting point.
     */
    ptr = getenv("HOME");
    if (ptr)
        strcpy(buf, ptr);
    else
        buf[0] = '\0';

    len = strlen(buf);
    buf[len++] = '/';
    buf[len] = '\0';

    /**
     * If the message has a subject, then just use that as the filename, 
     * stripping out any non alpha numeric characters (or spaces)
     */
    er = getSubject(&pstr);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Figure out if we can just use the subject, or if we need to go and
     * use a generic file name.
     */
    if (pstr && pstr[0]) {
        stripIntoBuf(buf + len, pstr, sizeof(buf) - (len + 1));
    } else {
        if (pstr) localFree(pstr);
        pstr = intlLoadString(ID_SAVE_MESSAGE);
        if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
        strcat(buf, pstr);
    }
    if (pstr) localFree(pstr);

    /**
     * Load in the filter.
     */
    pstr = intlLoadString(ID_MSG_SAVE_FILTER);
    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * Now get the filename
     */
    file = QFileDialog::getSaveFileName(buf, pstr);
    localFree(pstr);

    if (!file.isEmpty()) {
        K_FAIL("UNDONE: 11.1.2000 -- save me now dumbass!!!!");
        fprintf(stderr, "saving file: %s\n", (const char *)file);
    }

    return;

  Error:
    errorShowError(er);
    return;   
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::fileSend
 *=------------------------------------------------------------------------=
 * Takes the contents of this window and sends the message.  We have to do
 * a few checks first, like make sure there is at least one address to send
 * to, and that they are valid.
 */
void MailEditWindow::fileSend()
{
    MailMessage *msg;
    ERRCODE er;

    /** 
     * Get a message object first
     */
    er = populateMailMessage(&msg);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    } else if (er == S_FALSE)
        return;                        // names were bad or something missing

    /**
     * Send it
     */
    er = account->sendMessage(msg,
                              updateReadStatus == ReplacingMessage 
                                  ? sourceMessage : NULL);
    msg->release();
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }

    /**
     * finally, tell the host folder and UI to update themselves about this
     * message's status.
     */
    if (updateReadStatus != NoAction) {
        MailFolder *f = sourceMessage->getParent();
        if (f) f->updateMessageStatus(sourceMessage,
                                      updateReadStatus == MarkAsReplied
                                        ? MSG_REPLIED
                                        : MSG_FORWARDED);
    }

    /**
     * Finally, close this window and call it a day.
     */
    this->dirty = false;
    this->close(true);
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::formatAndIncludeBody
 *=------------------------------------------------------------------------=
 * takes the body from an existing message and reformats it to be suitable
 * for replies or forwards.
 *
 * Parameters:
 *      MailMessage *               - [in]  take my body and format it.
 *      const char *                - [in]  the prefix to apply to each line
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::formatAndIncludeBody
(
    MailMessage *includeMe,
    const char *prefix
)
{
    char buf[1024];
    const char *body;
    MIMETYPE type;
    ERRCODE er;
    int idx = 0, prefixlen;
    int i;

    if (!includeMe) return E_PARAM;

    editBox->setCursorPosition(editBox->numLines() + 1, 0, false);

    /**
     * Includes the text of the original message in this one.
     */
    type.setType(Text, Plain);
    er = includeMe->getBodyRefByType(type, &body, NULL);
    if (FAILED(er) && er == E_NOTFOUND) {
        type.setType(Text, Html);
        er = includeMe->getBodyRefByType(type, &body, NULL);
    }
    RETURN_ON_FAILURE(er);

    /**
     * now, one line at a time, go and start grabbing the text from the
     * original message, prefixing it as appropriate.
     * UNDONE: marcw, 11.2.2000 -- what are the preferences here for formatting
     */
    strcpy(buf, prefix);
    prefixlen = strlen(prefix);

    /**
     * now we have to whiz through the entire message looking for newlines
     * and writing out the prefix as we go.
     */
    i = prefixlen;
    while (body[idx] != '\0') {
        switch (body[idx]) {
            
            case '\n':
                buf[i++] = body[idx];
                buf[i] = '\0';
                editBox->insert(buf);
                i = prefixlen;
                break;
            default:
                buf[i++] = body[idx];
                if (i == sizeof(buf) - 1) {
                    buf[i] = '\0';
                    editBox->insert(buf);
                    i = prefixlen;
                }
                break;
        }
        idx++;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::forward
 *=------------------------------------------------------------------------=
 * takes the given message, extracts the appropriate content, and then 
 * generates a message that would be considered a "forward" of this one. 
 * for Forwards, Attachments are included and forwarded as well.
 *
 * Parameters:
 *      MailMessage *               - [in]  message to forward.
 *      MailAccount *               - [in]  account in which we are operating.
 *      MailEditWindow **           - [out] the new editor.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::forward
(
    MailMessage *forwardMe,
    MailAccount *account,
    MailEditWindow **pwindow
)
{
    const char *subject, *prefix;
    char buf[1024];
    ERRCODE er;
    int len, x;

    if (!forwardMe || !account || !pwindow) return E_PARAM;

    *pwindow = new MailEditWindow(account);
    if(!*pwindow) return E_OUTOFMEMORY;

    /**
     * Do the basics.
     */
    (*pwindow)->updateReadStatus = MarkAsForwarded;
    (*pwindow)->sourceMessage = forwardMe;
    er = (*pwindow)->initialize();
    GOTO_ON_FAILURE(er, Error);

    /**
     * extract the subject.
     */
    er = forwardMe->getSubject(&subject);
    GOTO_ON_FAILURE(er, Error);

    prefix = intlLoadString(ID_FW_TEXT);
    if (!prefix) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * See if the message already starts with "Fw" (or whatever)
     */
    len = strlen(prefix);
    for (x = 0; x < len; x++) {
        if (toupper(subject[x]) != toupper(prefix[x]))
            break;
    }

    if (x != len) {
        snprintf(buf, sizeof(buf), "%s:%s", prefix, subject);
    } else {
        snprintf(buf, sizeof(buf), "%s", subject);
    }
    localFree(prefix);
    localFree(subject);

    (*pwindow)->subjectEdit->setText(buf);
    (*pwindow)->editBox->setUndoEnabled(false);

    /**
     * now go and extract the to/from information and write out the header.
     */
    er = (*pwindow)->writeHeader(forwardMe);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Now format and write out the body.
     */
    er = account->getForwardPrefix(prefix);
    GOTO_ON_FAILURE(er, Error);
    er = (*pwindow)->formatAndIncludeBody(forwardMe, prefix);
    localFree(prefix);
    GOTO_ON_FAILURE(er, Error);

    /**
     * copy over any attachments.
     */
    er = (*pwindow)->copyOverAttachments(forwardMe);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Finally, insert the signature, if appropriate.
     */
    er = (*pwindow)->insertSignature();
    GOTO_ON_FAILURE(er, Error);

    (*pwindow)->editBox->setCursorPosition(0, 0);
    (*pwindow)->editBox->insert("\n\n\n");
    (*pwindow)->editBox->setCursorPosition(0, 0);
    (*pwindow)->toEdit->setFocus();
    (*pwindow)->editBox->setUndoEnabled(true);


    return S_OK;

  Error:
    delete *pwindow;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::getBCCList
 *=------------------------------------------------------------------------=
 * process the info in the BCC: box and returns it in a list.
 * Parameters:
 *      LinkList<const char *> &      - [in]  link list in which to put recips
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::getBCCList
(
    LinkList<const char *> &recips
)
{
    QString qstr = bccEdit->text();
    return parseAddressEdit((const char *)qstr, recips);
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::getCCList
 *=------------------------------------------------------------------------=
 * process the info in the Cc: box and returns it in a list.
 * Parameters:
 *      LinkList<const char *> &      - [in]  link list in which to put recips
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::getCCList
(
    LinkList<const char *> &recips
)
{
    QString qstr = ccEdit->text();
    return parseAddressEdit((const char *)qstr, recips);
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::getRecipientList
 *=------------------------------------------------------------------------=
 * process the information in the To: box and returns it as a list.
 *
 * Parameters:
 *      LinkList<const char *> &      - [in]  link list in which to put recips
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::getRecipientList
(
    LinkList<const char *> &recips
)
{
    QString qstr = toEdit->text();
    return parseAddressEdit((const char *)qstr, recips);
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::getSubject
 *=------------------------------------------------------------------------=
 * returns the subject for this message.
 *
 * Parameters:
 *      const char **           - [out]  the subject.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::getSubject
(
    const char **psubject
)
{
    if (!psubject) return E_PARAM;
    *psubject = NULL;

    if (subjectEdit) {
        const char *str = (const char *)subjectEdit->text();
        if (str) {
            *psubject = copyString(str);
            return *psubject ? S_OK : E_OUTOFMEMORY;
        }
    }
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::getBody
 *=------------------------------------------------------------------------=
 * returns the body of this message in one big text buffer.
 *
 * Parameters:
 *      const char **           - [out] where to put the text.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::getBody
(
    const char **pbody
)
{
    QString qstr;
    if (!pbody) return E_PARAM;
    *pbody = NULL;

    if (editBox) {
        qstr = editBox->text();
        if (qstr.length() > 0) {
            *pbody = copyString(qstr);
            return *pbody ? S_OK : E_OUTOFMEMORY;
        }
    }
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::helpAbout
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::helpAbout()
{
    AboutDialog ad(this, "ad", true, 0);
    ad.exec();
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::helpHelp
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::helpHelp()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::helpHelpMe()
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::helpHelpMe()
{

}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::initialize
 *=------------------------------------------------------------------------=
 * initializes the Qt UI widgets for this class and reads in the initial
 * contents from the given read stream.  
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::initialize()
{
    const char *pstr;
    QFont editFont;
    QFont font;
    ERRCODE er;
    int i;

    /**
     * Initialize our core UI with our base class.
     */
    er = this->initializeUI(this, mailEditWindowMenu, editToolBarData);
    RETURN_ON_FAILURE(er);

    pstr = intlLoadString(ID_NEW_MESSAGE);
    this->setCaption(pstr);
    localFree(pstr);

    this->setIcon(QPixmap(editwindowIcon));

    /**
     * Figure out how to scale the font.
     */
    font = this->font();
    if (prefFontScale > 0) {
        font.setPointSize((int)((double)font.pointSize() * 1.2) + 2);
        this->setFont(font);
    } else if (prefFontScale < 0) {
        font.setPointSize((int)((double)font.pointSize() * 0.75) + 2);
        this->setFont(font);
    }

    editFont.setPointSize(font.pointSize());

    /**
     * We have to have push buttons and edit boxes for our to/cc/subject/etc.
     * UNDONE: marcw, 10.10.2000 -- set the font for the buttons so that there
     * are no sizing issues.
     */
    pstr = intlLoadString(ID_TO_TEXT);
    toButton = new HoverButton(QIconSet(QPixmap(addresstIcon)),
                               QString(pstr), this);
    if (!toButton) return E_OUTOFMEMORY;
    localFree(pstr); pstr = NULL;
    toButton->show();
    toEdit = new MyMultiLineEdit(this, MyMultiLineEdit::TabToNext);
    if (!toEdit) return E_OUTOFMEMORY;
    toEdit->setWordWrap(MyMultiLineEdit::WidgetWidth);
    toEdit->show();
    toEdit->setFont(editFont);

    /**
     * CC Button and edit
     */
    pstr = intlLoadString(ID_CC_TEXT);
    ccButton = new HoverButton(QIconSet(QPixmap(addresstIcon)),
                               QString(pstr), this);
    if (!ccButton) return E_OUTOFMEMORY;
    localFree(pstr);

    ccEdit = new MyMultiLineEdit(this, MyMultiLineEdit::TabToNext);
    if (!ccEdit) return E_OUTOFMEMORY;
    ccEdit->setWordWrap(MyMultiLineEdit::WidgetWidth);
    ccEdit->setFont(editFont);

    /**
     * only show it if we're supposed to!!
     */
    if (prefEditShowCC) {
        ccButton->show();
        ccEdit->show();
    } else {
        ccButton->hide();
        ccEdit->hide();
    }

    /**
     * Bcc button and edit!
     */
    pstr = intlLoadString(ID_BCC_TEXT);
    bccButton = new HoverButton(QIconSet(QPixmap(addresstIcon)),
                                QString(pstr), this);
    if (!bccButton) return E_OUTOFMEMORY;
    localFree(pstr);

    bccEdit = new MyMultiLineEdit(this, MyMultiLineEdit::TabToNext);
    if (!bccEdit) return E_OUTOFMEMORY;
    bccEdit->setWordWrap(MyMultiLineEdit::WidgetWidth);
    bccEdit->setFont(editFont);

    if (prefEditShowBCC) {
        bccButton->show();
        bccEdit->show();
    } else {
        bccButton->hide();
        bccEdit->hide();
    }

    /**
     * Subject label and edit!!
     */
    pstr = intlLoadString(ID_SUBJECT_TEXT);
    subjectLabel = new QLabel(QString(pstr), this);
    if (!subjectLabel) return E_OUTOFMEMORY;
    localFree(pstr);
    subjectLabel->show();
    subjectLabel->setAutoResize(true);
    subjectEdit = new QLineEdit(this);
    if (!subjectEdit) return E_OUTOFMEMORY;
    subjectEdit->show();
    subjectEdit->setFont(editFont);

    /**
     * Priority label
     */
    priorityLabel = new QLabel(this);
    if (!priorityLabel) return E_OUTOFMEMORY;
    priorityLabel->setBackgroundColor(QColor(224, 228,177));
    priorityPixLabel = new QLabel(this);
    if (!priorityPixLabel) return E_OUTOFMEMORY;
    priorityPixLabel->setBackgroundColor(QColor(224, 228,177));

    /**
     * Create the splitter that will hold the main edit window and the
     * attachments icon view
     */
    contentSplitter = new QSplitter(Qt::Vertical, this);
    if (!contentSplitter) return E_OUTOFMEMORY;
    contentSplitter->show();

    /**
     * Create our edit box to display the contents.
     */
    editBox = new MyMultiLineEdit(contentSplitter, MyMultiLineEdit::TabsAreSpaces);
    editBox->setWordWrap(MyMultiLineEdit::WidgetWidth);
    editBox->show();
    account->getComposeFont(pstr, i);
    if (pstr) {
        editBox->setFont(QFont(pstr, i));
        localFree(pstr);
    }

    /**
     * create our attachment viewer
     */
    attachmentView = new QIconView(contentSplitter);
    if (!attachmentView) return E_OUTOFMEMORY;
    attachmentView->setItemTextPos(QIconView::Right);
    attachmentView->setAutoArrange(true);
    attachmentView->setResizeMode(QIconView::Adjust);
    attachmentView->setSelectionMode(QIconView::Extended);
    attachmentView->setMinimumSize(20, 50);
    attachmentView->resize(20, 50);
    attachmentView->hide();

    contentSplitter->setResizeMode(attachmentView, QSplitter::KeepSize);

    /**
     * Hook up some events to make sure our dirty processing is cool.
     */
    QObject::connect(toEdit, SIGNAL(textChanged()),
                     this, SLOT(textChanged()));
    QObject::connect(ccEdit, SIGNAL(textChanged()),
                     this, SLOT(textChanged()));
    QObject::connect(bccEdit, SIGNAL(textChanged()),
                     this, SLOT(textChanged()));
    QObject::connect(subjectEdit, SIGNAL(textChanged(const QString &)),
                     this, SLOT(subjectChanged()));
    QObject::connect(editBox, SIGNAL(textChanged()),
                     this, SLOT(textChanged()));
    QObject::connect(toButton, SIGNAL(clicked()), this, SLOT(toButtonClicked()));
    QObject::connect(ccButton, SIGNAL(clicked()), this, SLOT(ccButtonClicked()));
    QObject::connect(bccButton, SIGNAL(clicked()), this, SLOT(bccButtonClicked()));

    /**
     * finish up with a few final things.
     */
    menuBar()->setItemChecked(ID_MSG_MED_TEXT, true);

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::insertAttachment
 *=------------------------------------------------------------------------=
 * Allow the user to add some attachments to this mail message.  They are
 * added into the attachment area.
 */
void MailEditWindow::insertAttachment()
{
    QString filter, caption;
    QStringList files;
    const char *pstr;
    struct stat info;
    ERRCODE er;

    /**
     * Load in some strings.
     */
    pstr = intlLoadString(ID_INSERT_FILTER);
    if (!pstr) return;
    filter = pstr;
    localFree(pstr);

    pstr = intlLoadString(ID_INSERT_ATTACHMENT);
    if (!pstr) return;
    caption = pstr;
    localFree(pstr);

    /**
     * if there is no directory for this yet, set it to the users's home
     * dir.
     */
    if (!prefAttachmentDir) {
        char *ptr;
        ptr = getenv("HOME");
        if (ptr) {
            prefAttachmentDir = copyString(ptr);
        }
    }

    /**
     * Show the dialog
     */
    files = QFileDialog::getOpenFileNames(filter, prefAttachmentDir, this,
                                          "insert", caption);

    /**
     * Finally, verify the files really exist and then go and actually
     * add the attachments.
     */
    for (uint x = 0; x < files.count(); x++) {

        int ret = lstat((const char *)files[x], &info);
        er = ret ? E_FILE_ERR : S_OK;

        if (er == S_OK) attachFile((const char *)files[x]);
        if (FAILED(er)) {
            errorShowError(er);
            break;
        }
    }

    /**
     * extract the directory they were last in so we can remember
     * it for them.
     */
    if (files.count() > 0) {

        int index = files[0].findRev('/', -1);
        if (index > 0) {
            files[0].truncate(index);
            if (prefAttachmentDir) localFree(prefAttachmentDir);
            prefAttachmentDir = copyString((const char *)files[0]);
        }
    }


}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::insertSignature
 *=------------------------------------------------------------------------=
 * Inserts the user's signature into the message.  We know from the account
 * if they have a signature.
 *
 * Output:
 *      ERRCODE                    - S_FALSE if we didn't insert one.
 */
ERRCODE MailEditWindow::insertSignature()
{
    const char *pstr;
    char buf[1024];
    ERRCODE er;
    FILE *f;

    /**
     * First, see if they even have a signature at all.
     */
    if (!account->hasSignatureFile()) return S_FALSE;

    /**
     * Okay, go and insert it now.
     */
    er = account->getSignatureFile(pstr);
    RETURN_ON_FAILURE(er);

    editBox->setCursorPosition(0, 0, false);
    editBox->insert("\n");

    /**
     * Open it up and load it in, one line at a time.
     */
    f = fopen(pstr, "r");
    localFree(pstr);
    if (!f) return E_FILE_ERR;

    while (!feof(f) && !ferror(f)) {

        if (!fgets(buf, sizeof(buf), f)) break;
        editBox->insert(buf);
    }
    
    fclose(f);

    editBox->insert("\n");
    editBox->setCursorPosition(0, 0, false);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::insertText
 *=------------------------------------------------------------------------=
 * Inserts text from a file that the user specifies.  A dialog is shown 
 * allowing them to choose which file to insert..
 */
void MailEditWindow::insertText()
{
    char readBuf[MAX_LINE_LEN];
    QString filter, caption;
    const char *pstr;
    QString file;
    FILE *f;

    /**
     * Load in some strings.
     */
    pstr = intlLoadString(ID_TEXT_FILTER);
    if (!pstr) return;
    filter = pstr;
    localFree(pstr);

    pstr = intlLoadString(ID_INSERT_TEXT);
    if (!pstr) return;
    caption = pstr;
    localFree(pstr);

    /**
     * if there is no directory for this yet, set it to the users's home
     * dir.
     */
    if (!prefAttachmentDir) {
        char *ptr;
        ptr = getenv("HOME");
        if (ptr) {
            prefAttachmentDir = copyString(ptr);
        }
    }

    /**
     * Show the dialog
     */
    file = QFileDialog::getOpenFileName(prefAttachmentDir, filter, this,
                                        "insert", caption);

    /**
     * Insert the text!!!
     */
    if (!file.isEmpty()) {
        f = fopen((const char *)file, "r");
        if (!f) {
            errorShowError(E_FILE_ERR);
            return;
        }

        editBox->setUpdatesEnabled(false);
        while (!feof(f)) {
            if (!fgets(readBuf, sizeof(readBuf), f)) break;
            editBox->insert(readBuf);
        }
        editBox->setUpdatesEnabled(true);
        editBox->repaint();

        fclose(f);
    }

    /**
     * extract the directory they were last in so we can remember
     * it for them.
     */
    if (!file.isEmpty()) {
        int index = file.findRev('/', -1);
        if (index > 0) {
            file.truncate(index);
            if (prefAttachmentDir) localFree(prefAttachmentDir);
            prefAttachmentDir = copyString((const char *)file);
        }
    }

    dirty = true;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::keyPressEvent
 *=------------------------------------------------------------------------=
 * a key has been pressed.  Wahooo!
 *
 * Parameters:
 *      QKeyEvent *                 - [in]  the event.
 */
void MailEditWindow::keyPressEvent
(
    QKeyEvent *event
)
{
    ButtonState state;
    if (!event) return;

    state = event->state();

#if 0
    /**
     * Handle some handy keyboard accelerators here !!
     */
    switch (event->key()) {

        case Key_N:
            if (state & ControlButton)
                fileNew();
            break;
    }
#endif // 0

    KWindowBase::keyPressEvent(event);
}



/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::layoutComponents
 *=------------------------------------------------------------------------=
 * Lays out all the components in our window, taking into consideration
 * things such as menu and toolbar sizes, etc ...
 *
 * UNDONE: marcw, 10.10.2000 -- get a better way to specify the width and
 * height of buttons!!!
 *
 * UNDONE: marcw, 10.10.2000 -- need to support two lines of input in the 
 * to, cc and bcc edits.
 */
void MailEditWindow::layoutComponents()
{
    int ctlHeight, ctlWidth;
    const char *pstr;
    QRect clientRect;
    int top;

    /**
     * Obtain/Compute some sizes 
     */
    ctlHeight = toButton->font().pixelSize() + 8;  // 2 x 4 for over/under-hang
    ctlWidth = subjectLabel->width();
    getClientRect(clientRect);

    top = clientRect.top() + 3;

    switch (priority) {
        case NormalPriority:
            priorityPixLabel->hide();
            priorityLabel->hide();
            break;

        case HighPriority:
        case LowPriority:
            pstr = intlLoadString(priority == LowPriority
                                  ? ID_LOW_PRIORITY_TEXT 
                                  : ID_HIGH_PRIORITY_TEXT);
            if (pstr) {
                QPixmap p(priority == LowPriority ? lowpriIcon : priorityhdrIcon);
                priorityPixLabel->setPixmap(p);
                priorityPixLabel->setGeometry(4, top, p.width() + 1, ctlHeight);
                priorityLabel->setText(pstr);
                localFree(pstr);
                priorityLabel->setGeometry(p.width() + 5, top,
                                           this->width() - p.width() - 13,
                                           ctlHeight);
                top += ctlHeight + 5;
                priorityPixLabel->show();
                priorityLabel->show();
            } else {
                priorityPixLabel->hide();
                priorityLabel->hide();
            }
            break;
    }

    /**
     * Start setting our childrens' locations.
     */
    toButton->setGeometry(2, top, ctlWidth, ctlHeight);
    if (toEditLines != toEdit->numLines()) {
        toEditLines = toEdit->numLines();
        if (toEditLines > 4) 
            toEditLines = 4;
    }
    toEdit->setGeometry(ctlWidth + 5, top, clientRect.width() - (ctlWidth + 13), ctlHeight * toEditLines);
    top += (ctlHeight * toEditLines + 5);
    if (prefEditShowCC) {
        ccButton->setGeometry(2, top, ctlWidth, ctlHeight);
        if (ccEditLines != ccEdit->numLines()) {
            ccEditLines = ccEdit->numLines();
            if (ccEditLines > 4) 
                ccEditLines = 4;
        }
        ccEdit->setGeometry(ctlWidth + 5, top, clientRect.width() - (ctlWidth + 13), ctlHeight * ccEditLines);
        top += (ctlHeight * ccEditLines + 5);
    }
    if (prefEditShowBCC) {
        bccButton->setGeometry(2, top, ctlWidth, ctlHeight);
        if (bccEditLines != bccEdit->numLines()) {
            bccEditLines = bccEdit->numLines();
            if (bccEditLines > 4) 
                bccEditLines = 4;
        }
        bccEdit->setGeometry(ctlWidth + 5, top, clientRect.width() - (ctlWidth + 13), ctlHeight *bccEditLines);
        top += (ctlHeight * bccEditLines + 5);
    }
    subjectLabel->setGeometry(2, top, ctlWidth, ctlHeight);
    subjectEdit->setGeometry(ctlWidth + 5, top, clientRect.width() - (ctlWidth + 13), ctlHeight);
    top += (ctlHeight + 5);

    contentSplitter->setGeometry(1, top, clientRect.width() - 2, this->height() - top);
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::mailto
 *=------------------------------------------------------------------------=
 * Creates a new mail eidt window in the context of a reply.  Sets up
 * everything as required for a reply.
 *
 * Parameters:
 *      MailAccount *           - [in]  within the context of this account.
 *      const char *            - [in]  to whom is this message being sent
 *      MailEditWindow **       - [out] the new edit window.
 *
 * Output: 
 *      ERRCODE
 */
ERRCODE MailEditWindow::mailto
(
    MailAccount *account,
    const char *mailto,
    MailEditWindow **pwindow
)
{
    ERRCODE er;

    if (!account || !mailto || !pwindow) return E_PARAM;

    *pwindow = new MailEditWindow(account);
    if(!*pwindow) return E_OUTOFMEMORY;

    /**
     * Do the basics.
     */
    er = (*pwindow)->initialize();
    GOTO_ON_FAILURE(er, Error);

    (*pwindow)->toEdit->setText(mailto);
    return S_OK;

  Error:
    delete *pwindow;
    return er;

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::messageAddressBook
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::messageAddressBook()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::messageCheckNames
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::messageCheckNames()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::messageHigh
 *=------------------------------------------------------------------------=
 * The user has changed the priority of the message.  Update the menus and UI
 * now
 */
void MailEditWindow::messageHigh()
{
    priority = HighPriority;

    menuBar()->setItemChecked(ID_MSG_HIGH_TEXT, true);
    menuBar()->setItemChecked(ID_MSG_LOW_TEXT, false);
    menuBar()->setItemChecked(ID_MSG_MED_TEXT, false);

    dirty = true;
    layoutComponents();
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::messageLow
 *=------------------------------------------------------------------------=
 * The user has changed the priority of the message.  Update the menus and UI
 * now
 */
void MailEditWindow::messageLow()
{
    priority = LowPriority;

    menuBar()->setItemChecked(ID_MSG_HIGH_TEXT, false);
    menuBar()->setItemChecked(ID_MSG_LOW_TEXT, true);
    menuBar()->setItemChecked(ID_MSG_MED_TEXT, false);

    dirty = true;
    layoutComponents();
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::messageMedium
 *=------------------------------------------------------------------------=
 * The user has changed the priority of the message.  Update the menus and UI
 * now
 */
void MailEditWindow::messageMedium()
{
    priority = NormalPriority;

    menuBar()->setItemChecked(ID_MSG_HIGH_TEXT, false);
    menuBar()->setItemChecked(ID_MSG_LOW_TEXT, false);
    menuBar()->setItemChecked(ID_MSG_MED_TEXT, true);

    dirty = true;
    layoutComponents();
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::messageCheckRecipients
 *=------------------------------------------------------------------------=
 */
void MailEditWindow::messageRecipients()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::parseAddressEdit
 *=------------------------------------------------------------------------=
 * given the contents of an edit box containing addresses, go and get the
 * list of email addresses in said edit.
 *
 * Parameters:
 *      const char *                - [in]  contents of edit box.
 *      LinkList<const char *> &    - [in]  where to put all the addresses
 *                                      we find.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::parseAddressEdit
(
    const char *text,
    LinkList<const char *> &list
)
{
    char buf[MAX_ADDRESSLEN];
    int start = 0, end = 0;
    const char *pstr;
    ERRCODE er;
    int x;

    if (!text) return E_PARAM;
    if (text[0] == '\0') return S_OK;

    /**
     * Whiz through the list, looking for delimiting characters
     */
    do {
        /**
         * Strip off leading whitespace
         */
        if (isspace(text[start]) && start == end) {
            start++; end++;
        } else if (text[end] == ';' || text[end] == ',' || text[end] == '\0') {
            memcpy(buf, text + start, end - start);
            buf[end - start] = '\0';

            /**
             * Strip off trailing whitespace
             */
            x = end - start;
            while (x && isspace(buf[x]))
                buf[x--] = '\0';

            /** 
             * Copy over the string if there was something left in it.
             */
            if (x) {
                pstr = copyString(buf);
                if (!pstr) {
                    nukeStringList(list);
                    return E_OUTOFMEMORY;
                }
                er = list.addElement(pstr);
                if (FAILED(er)) {
                    nukeStringList(list);
                    return er;
                }
            }
            start = end = end + 1;
        } else {
            end++;
        }
    } while (text[end - 1] != '\0');


    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::populateMailMessage
 *=------------------------------------------------------------------------=
 * given our current contents, create a new mail message set up to match this.
 * This can be used to send (add to Outbox), or it can be saved.
 *
 * Parameters:
 *      MailMessage **          - [out] a new mail message.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::populateMailMessage
(
    MailMessage **pmessage
)
{
    LinkList<const char *> list;
    const char *pstr = NULL;
    MIMETYPE type;
    Attachment *a;
    ERRCODE er;

    /**
     * 1.  check all the names and make sure they're OK.
     */
    er = checkNames();
    GOTO_ON_FAILURE(er, Error);

    /**
     * If there were bogus names, then don't exit at this time -- let the 
     * user fix them. ..
     */
    if (er == S_FALSE) return S_FALSE;

    /**
     * Okay, now go and create a message object and put all the goo in
     * it.
     */
    er = MailMessage::createNew(NULL, pmessage);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Start grabbing information we're going to give to the msg.
     * UNDONE: marcw, 10.11.2000 -- multiple addresses per account??
     */
    er = account->getSenderAddress(&pstr);
    GOTO_ON_FAILURE(er, Error);
    if (pstr) {
        list.addElement(pstr);
        er = (*pmessage)->setSenders(list);
        GOTO_ON_FAILURE(er, Error);
        list.emptyList();
    }

    /**
     * Recipients
     */
    er = getRecipientList(list);
    GOTO_ON_FAILURE(er, Error);
    er = (*pmessage)->setRecipients(list);
    GOTO_ON_FAILURE(er, Error);
    list.emptyList();

    /**
     * CC and BCC lists.
     */
    er = getCCList(list);
    GOTO_ON_FAILURE(er, Error);
    er = (*pmessage)->setCCs(list);
    GOTO_ON_FAILURE(er, Error);
    list.emptyList();
    er = getBCCList(list);
    GOTO_ON_FAILURE(er, Error);
    er = (*pmessage)->setBCCs(list);
    GOTO_ON_FAILURE(er, Error);
    list.emptyList();

    /**
     * Subject
     */
    er = getSubject(&pstr);
    GOTO_ON_FAILURE(er, Error);
    er = (*pmessage)->setSubject(pstr);
    GOTO_ON_FAILURE(er, Error);
    if (pstr) localFree(pstr);

    /**
     * Priority
     */
    er = (*pmessage)->setPriority(this->priority);
    GOTO_ON_FAILURE(er, Error);

    /**
     * The Body!
     */
    er = getBody(&pstr);
    GOTO_ON_FAILURE(er, Error);
    type.setType(Text, Plain);
    er = (*pmessage)->setBodyByType(type, pstr, NULL);
    if (pstr) { localFree(pstr); pstr = NULL; }
    GOTO_ON_FAILURE(er, Error);

    /**
     * Any attachments we might have.
     */
    er = attachments.firstElement(&a);
    while (er == S_OK) {
        ERRCODE er2;

        er2 = (*pmessage)->addAttachment(a);
        GOTO_ON_FAILURE(er2, Error);

        er = attachments.nextElement(&a);
    }
    GOTO_ON_FAILURE(er, Error);

    return S_OK;

  Error:
    if (*pmessage) (*pmessage)->release();
    if (pstr) localFree(pstr);
    return er;     
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::reply
 *=------------------------------------------------------------------------=
 * Creates a new mail eidt window in the context of a reply.  Sets up
 * everything as required for a reply.
 *
 * Parameters:
 *      MailMessage *           - [in]  reply to me.
 *      MailAccount *           - [in]  within the context of this account.
 *      bool                    - [in]  true == reply ALL, false == sender
 *      MailEditWindow **       - [out] the new edit window.
 *
 * Output: 
 *      ERRCODE
 */
ERRCODE MailEditWindow::reply
(
    MailMessage *replyToMe,
    MailAccount *account,
    bool replyAll,
    MailEditWindow **pwindow
)
{
    const char *subject, *prefix = NULL;
    LinkList<const char *> list;
    char buf[1024];
    ERRCODE er;
    int len, x;

    if (!replyToMe || !account || !pwindow) return E_PARAM;

    *pwindow = new MailEditWindow(account);
    if(!*pwindow) return E_OUTOFMEMORY;

    /**
     * Do the basics.
     */
    (*pwindow)->updateReadStatus = MarkAsReplied;
    (*pwindow)->sourceMessage = replyToMe;
    er = (*pwindow)->initialize();
    GOTO_ON_FAILURE(er, Error);

    (*pwindow)->editBox->setUndoEnabled(false);
    (*pwindow)->toEdit->clear();
    (*pwindow)->ccEdit->clear();

    /**
     * Figure out to whom the reply should be sent.
     */
    er = replyToMe->getReplyPath(list, replyAll);
    GOTO_ON_FAILURE(er, Error);
    dumpListIntoEdit(list, "; ", (*pwindow)->toEdit, true);
    list.emptyList();

    /**
     * IF this is a replyAll, then we need to include the recipients and
     * ccs here as well!
     */
    if (replyAll) {
        
        er = replyToMe->getRecipients(list);
        GOTO_ON_FAILURE(er, Error);
        dumpListIntoEdit(list, "; ", (*pwindow)->toEdit, true);
        list.emptyList();

        er = replyToMe->getCCs(list);
        GOTO_ON_FAILURE(er, Error);
        dumpListIntoEdit(list, "; ", (*pwindow)->ccEdit, true);
        list.emptyList();
    }

    /**
     * extract the subject.
     */
    er = replyToMe->getSubject(&subject);
    GOTO_ON_FAILURE(er, Error);

    prefix = intlLoadString(ID_RE_TEXT);
    if (!prefix) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * See if the message already starts with "Re" (or whatever)
     */
    len = strlen(prefix);
    x = 0;
    if (subject) {
        for (x = 0; x < len; x++) {
            if (toupper(subject[x]) != toupper(prefix[x]))
                break;
        }
    }

    if (x != len) {
        snprintf(buf, sizeof(buf), "%s:%s", prefix, subject ? subject : "");
    } else {
        snprintf(buf, sizeof(buf), "%s", subject);
    }
    localFree(prefix);
    if (subject) localFree(subject);

    (*pwindow)->subjectEdit->setText(buf);

    er = account->getReplyPrefix(prefix);
    GOTO_ON_FAILURE(er, Error);

    /**
     * now go and extract the to/from information and write out the header.
     */
    er = (*pwindow)->writeHeader(replyToMe, prefix);
    if (FAILED(er)) {
        localFree(prefix);
        goto Error;
    }

    /**
     * Now format and write out the body.
     */
    er = (*pwindow)->formatAndIncludeBody(replyToMe, prefix);
    localFree(prefix);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Finally, insert the signature, if appropriate.
     */
    er = (*pwindow)->insertSignature();
    GOTO_ON_FAILURE(er, Error);

    (*pwindow)->editBox->setCursorPosition(0, 0);
    (*pwindow)->editBox->insert("\n\n");
    (*pwindow)->editBox->setCursorPosition(0, 0);
    (*pwindow)->toEdit->setFocus();

    (*pwindow)->editBox->setUndoEnabled(true);
    return S_OK;

  Error:
      (*pwindow)->editBox->setUndoEnabled(true);
    delete *pwindow;
    return er;

}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::recomposeNew
 *=------------------------------------------------------------------------=
 * Creates a new editor for re-editing an already existing message ...
 *
 * Parameters:
 *      MailMessage *           - [in]  reply to me.
 *      MailAccount *           - [in]  within the context of this account.
 *      MailEditWindow **       - [out] the new edit window.
 *
 * Output: 
 *      ERRCODE
 */
ERRCODE MailEditWindow::recomposeNew
(
    MailMessage *reeditMe,
    MailAccount *account,
    MailEditWindow **pwindow
)
{
    LinkList<Attachment *> attachments;
    LinkList<const char *> stringList;
    const char *pstr = NULL;
    MessagePriority mp;
    MIMETYPE type;
    Attachment *a;
    ERRCODE er;

    if (!pwindow || !reeditMe || !account) return E_PARAM;
    
    *pwindow = new MailEditWindow(account);

    er = (*pwindow)->initialize();
    RETURN_ON_FAILURE(er);

    /** 
     * Okay, start refilling all the information -- blech.
     */
    reeditMe->getRecipients(stringList);
    dumpListIntoEdit(stringList, "; ", (*pwindow)->toEdit, true);
    stringList.emptyList();

    reeditMe->getCCs(stringList);
    dumpListIntoEdit(stringList, "; ", (*pwindow)->ccEdit, true);
    stringList.emptyList();

    reeditMe->getBCCs(stringList);
    dumpListIntoEdit(stringList, "; ", (*pwindow)->bccEdit, true);
    stringList.emptyList();

    reeditMe->getSubject(&pstr);
    if (pstr) {
        (*pwindow)->subjectEdit->setText(pstr);
        localFree(pstr);
    }
        
    type.setType(Text, Plain);
    reeditMe->getBodyRefByType(type, &pstr);
    if (pstr) {
        (*pwindow)->editBox->setText(pstr);
        pstr = NULL;
    }

    reeditMe->getPriority(mp);
    (*pwindow)->priority = mp;

    /**
     * finally, attachments (blech)
     */
    reeditMe->getAttachments(attachments);
    er = attachments.firstElement(&a);
    while (er == S_OK) {
        a->addRef();
        (*pwindow)->attachments.addElement(a);
        (*pwindow)->addAttachmentToDisplay(a);
        er = attachments.nextElement(&a);
    }

    (*pwindow)->sourceMessage = reeditMe;
    (*pwindow)->updateReadStatus = ReplacingMessage;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::showAddressPicker
 *=------------------------------------------------------------------------=
 * Shows the address picker for the given button and fills in the results
 * to the edit box for that guy too.
 *
 * Parameters:
 *      STRINGID                    - [in]  caption.
 *      MyMultiLineEdit *           - [in]  where to put results.
 *
 * Output:
 *      ERRCODE
 */
void MailEditWindow::showAddressPicker
(
    STRINGID caption,
    MyMultiLineEdit *edit
)
{
    AddressPicker ap(caption, this, "", true, 0);
    LinkList<const char *> tos;
    int ret;

    /**
     * Show the dialog, and if they accepted, dump the selected list into
     * the given edit control.
     */
    ret = ap.exec();

    if (ret == QDialog::Accepted) {
        ap.getAddressList(tos);

        dumpListIntoEdit(tos, "; ", edit, true);
        nukeStringList(tos);
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::subjectChanged
 *=------------------------------------------------------------------------=
 * the subject has changed.  be sure to go and update the caption.
 */
void MailEditWindow::subjectChanged()
{
    this->setCaption(subjectEdit->text());
    dirty = true;
}
 
 
/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::textChanged
 *=------------------------------------------------------------------------=
 * the text in one of the edit boxes has changed.  Mark ourselves as dirty.
 */
void MailEditWindow::textChanged()
{
    dirty = true;

    if (toEdit->numLines() != toEditLines
        || ccEdit->numLines() != ccEditLines
        || bccEdit->numLines() != bccEditLines) {
        
        toEditLines = -1;
        ccEditLines = -1;
        bccEditLines = -1;
        layoutComponents();
        this->repaint();
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::toButtonClicked
 *=------------------------------------------------------------------------=
 * the user wants to use the address picker to choose the recipients.
 */
void MailEditWindow::toButtonClicked()
{
    showAddressPicker(ID_SELECT_RECIPIENTS, this->toEdit);
}
 

/**
 *=------------------------------------------------------------------------=
 * MailEditWindow::writeHeader
 *=------------------------------------------------------------------------=
 * given a mail message that is being replied to or forwarded, write out the
 * basic header in the output window.
 *
 * Parameters:
 *      MailMessage *           - [in]  message to get info from.
 *      const char *            - [in]  prefix for each line.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailEditWindow::writeHeader
(
    MailMessage *useMe,
    const char *prefix
)
{
    LinkList<const char *> list;
    const char *pstr;
    ERRCODE er;

    editBox->setCursorPosition(0, 0, false);
    if (prefix) editBox->insert(prefix);
    editBox->insert("-----------------------------\n");

    /**
     * From:
     */
    pstr = intlLoadString(ID_COLUMN_FROM);
    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }

    if (prefix) editBox->insert(prefix);
    editBox->insert(pstr);
    localFree(pstr);
    editBox->insert("  ");
    
    /**
     * Get the senders and put them in the editbox.
     */
    er = useMe->getSenders(list);
    GOTO_ON_FAILURE(er, Error);
    dumpListIntoEdit(list, ", ", editBox);
    editBox->insert("\n");
    list.emptyList();

    /**
     * To:
     */
    pstr = intlLoadString(ID_COLUMN_TO);
    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
    if (prefix) editBox->insert(prefix);
    editBox->insert(pstr);
    localFree(pstr);
    editBox->insert("  ");

    er = useMe->getRecipients(list);
    dumpListIntoEdit(list, ", ", editBox);
    GOTO_ON_FAILURE(er, Error);
    editBox->insert("\n");
    list.emptyList();

    /**
     * CC:
     */
    er = useMe->getCCs(list);
    GOTO_ON_FAILURE(er, Error);
    if (list.countElements() > 0) {

        pstr = intlLoadString(ID_COLUMN_CC);
        if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
        if (prefix) editBox->insert(prefix);
        editBox->insert(pstr);
        localFree(pstr);
        editBox->insert("  ");

        dumpListIntoEdit(list, ", ", editBox);
        GOTO_ON_FAILURE(er, Error);
        editBox->insert("\n");
        list.emptyList();
        editBox->insert("\n");
        list.emptyList();
    }

    /**
     * Subject:
     */
    pstr = intlLoadString(ID_COLUMN_SUBJECT);
    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }

    if (prefix) editBox->insert(prefix);
    editBox->insert(pstr);
    localFree(pstr);
    editBox->insert("  ");
    
    er = useMe->getSubject(&pstr);
    GOTO_ON_FAILURE(er, Error);

    if (pstr) {
        editBox->insert(pstr);
        localFree(pstr);
    }
    editBox->insert("\n");

    /**
     * Sent
     */
    pstr = intlLoadString(ID_COLUMN_SENT);
    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }

    if (prefix) editBox->insert(prefix);
    editBox->insert(pstr);
    localFree(pstr);
    editBox->insert("  ");
    
    er = useMe->getSent(&pstr);
    GOTO_ON_FAILURE(er, Error);

    editBox->insert(pstr);
    localFree(pstr);

    /**
     * now a coupla blank lines, prefixed if necessary.
     */
    for (int z = 0; z < 3; z++) {
        if (prefix) editBox->insert(prefix);
        editBox->insert("\n");            
    }

    /**
     * that's about it!!!
     */
    er = S_OK;

  Error:
    return er;
}



/**
 *=------------------------------------------------------------------------=
 * stripIntoBuf
 *=------------------------------------------------------------------------=
 * strips the given source of all non alnum/space chars and puts them in 
 * dest.
 *
 * Parameters:
 *      char *                  - [in]  dest buff
 *      const char *            - [in]  source buff
 *      int                     - [in]  max # chars in dest.
 */
void stripIntoBuf
(
    char *dest,
    const char *source,
    int max
)
{
    int x, y;

    y = 0;
    for (x = 0; x < max; x++) {
        if (isalnum(source[x]) || source[x] == ' ')
            dest[y++] = source[x];
        else if (source[x] == '\0') {
            dest[y] = '\0';
            break;
        }
    }
}


/**
 *=------------------------------------------------------------------------=
 * dumpListIntoEdit
 *=------------------------------------------------------------------------=
 * given a list containing strings, dump them into an edit box, separating
 * them with semi-colons.
 *
 * Parameters:
 *      LinkList<const char *> &list        - [in]  list of strings to dump.
 *      const char *                        - [in]  how to separate strings
 *      MyMultiLineEdit *                   - [in]  where to dump 'em.
 *      bool                                - [in]  insert a separator before
 *                                              the text if there is other
 *                                              text in the edit.
 */
void dumpListIntoEdit
(
    LinkList<const char *> &list,
    const char *separator,
    MyMultiLineEdit *editbox,
    bool addsep
)
{
    const char *pstr;
    ERRCODE er;

    editbox->setUndoEnabled(false);
    editbox->end();

    if (addsep && list.countElements() && !editbox->text().isEmpty())
        editbox->insert(separator);

    er = list.firstElement(&pstr);
    if (er == S_OK) {
        editbox->insert(pstr);
        er = list.nextElement(&pstr);
    }

    while (er == S_OK) {
        editbox->insert(separator);
        editbox->insert(pstr);
        er = list.nextElement(&pstr);
    }

    editbox->setUndoEnabled(true);
}


