/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _MAILEDITWINDOW_H_
#define _MAILEDITWINDOW_H_

/**
 * This class is the editor window that you use to compose new mail, news, 
 * replies, or forwards of the above.  It's a top-level window that you can
 * create as many instances of as you'd like ...
 */
#include <sanity.h>

#include <qlabel.h>
#include <qlineedit.h>
#include <qmainwindow.h>
#include <qmenubar.h>
#include <qtoolbar.h>

#include <intl.h>
#include "hoverbutton.h"
#include "kwindowbase.h"
#include <llist.h>
#include "mailmessage.h"

class QSplitter;
class QIconView;
class Attachment;



class MailEditWindow : public KWindowBase {

    /**
     * This has to be here for all Qt objects.
     */
    Q_OBJECT

    DEBUG_TEST_DECL;

  public:

    /**
     * There are three ways to create a new mail editor, depending on what
     * the purpose is.  You can create a new one, reply to a message, or
     * forward a message.
     */
    static ERRCODE composeNew(MailAccount *, MailEditWindow **);
    static ERRCODE forward(MailMessage *, MailAccount *, MailEditWindow **);
    static ERRCODE mailto(MailAccount *, const char *, MailEditWindow **);
    static ERRCODE recomposeNew(MailMessage *, MailAccount *, MailEditWindow **);
    static ERRCODE reply(MailMessage *, MailAccount *, bool replyAll,
                         MailEditWindow **);

    virtual ~MailEditWindow();

  protected:
    /**
     * Lays out all the buttons and all that jazz.
     */
    virtual void layoutComponents();
    virtual void closeEvent(QCloseEvent *);
    virtual void createdPopup(int, QPopupMenu *);
    virtual void keyPressEvent(QKeyEvent *);

  private:
    bool dirty;
    MailAccount *account;
    MailMessage *sourceMessage;
    MessagePriority priority;
    LinkList<Attachment *> attachments;

    /**
     * This is how we will tell the host folder to update the message
     * read status when we send this message.
     */
    enum {
        NoAction = 0,
        ReplacingMessage,
        MarkAsReplied,
        MarkAsForwarded

    } updateReadStatus;

    /**
     * Use composeNew, reply, or forward()
     */
    MailEditWindow(MailAccount *);

    /**
     * UI Widgets
     */
    HoverButton *toButton, *ccButton, *bccButton;
    MyMultiLineEdit *toEdit, *ccEdit, *bccEdit;
    int toEditLines, ccEditLines, bccEditLines;
    QLabel *priorityLabel, *priorityPixLabel;
    QSplitter *contentSplitter;
    QIconView *attachmentView;
    MyMultiLineEdit *editBox;
    QLineEdit *subjectEdit;
    QLabel *subjectLabel;

    /**
     * These routines extract the info from the window and put it in a format
     * we can use.
     */
    ERRCODE getRecipientList(LinkList<const char *> &list);
    ERRCODE getCCList(LinkList<const char *> &list);
    ERRCODE getBCCList(LinkList<const char *> &list);
    ERRCODE getSubject(const char **);
    ERRCODE attachFile(const char *);
    ERRCODE getBody(const char **);

    ERRCODE populateMailMessage(MailMessage **);
    void    showAddressPicker(STRINGID, MyMultiLineEdit *);

    /**
     * random private routines to help us do our job.
     */
    bool askForSave();
    ERRCODE checkNames();
    ERRCODE initialize();
    ERRCODE insertSignature();
    ERRCODE writeHeader(MailMessage *, const char *prefix = NULL);
    ERRCODE copyOverAttachments(MailMessage *);
    ERRCODE addAttachmentToDisplay(Attachment *);
    ERRCODE formatAndIncludeBody(MailMessage *, const char *prefix);
    ERRCODE parseAddressEdit(const char *, LinkList<const char *> &list);


  private slots:
    /**
     * Menu Item slos
     */
    void fileNew();
    void fileSave();
    void fileSend();
    void filePrint();
    void fileClose();

    void editCut();
    void editCopy();
    void editPaste();
    void editSelectAll();
    void editFindPerson();
    void editFindMessage();
    void editFindText();

    void insertAttachment();
    void insertText();

    void messageCheckNames();
    void messageRecipients();
    void messageAddressBook();
    void messageHigh();
    void messageMedium();
    void messageLow();

    void helpHelpMe();
    void helpHelp();
    void helpAbout();

    void addressBook();

    void textChanged();
    void subjectChanged();

    /**
     * Address book interfaces.
     */
    void toButtonClicked();
    void ccButtonClicked();
    void bccButtonClicked();

    /**
     * Menu popups
     */
    void aboutToShowFileMenu();
    void aboutToShowEditMenu();
    void aboutToShowInsertMenu();
    void aboutToShowMessageMenu();
};





#endif // _MAILEDITWINDOW_H_


