/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <qdragobject.h>
#include <qmessagebox.h>
#include <qpopupmenu.h>

#include <unistd.h>


#include "folderdb.h"
#include "foldersort.h"
#include <intl.h>
#include "kiltdown.h"
#include "localized.h"
#include "maileditwindow.h"
#include "mailfolder.h"
#include "mailviewer.h"
#include "main.h"
#include "uiicons.h"


/**
 * This is the implementation of our mail folder class.  
 */
#define TOOMANY_MESSAGES 10

/**
 * these are the property names for the column info we need for this folder.
 */
const char *folderColumnCountProp = "columnCount";    // int
const char *folderWhichColumnProp = "whichColumn";    // int (enum MailColumn)
const char *folderColumnWidthProp = "columnWidth";    // int    
const char *folderColumnPercentProp = "columnPercent";// double
const char *folderColumnFixedProp = "columnFixed";    // bool
const char *folderSortedUpProp = "folderSortedUp";    // bool
const char *folderSortedProp = "folderSorted";        // int (enum MailColumn)
const char *folderColumnSortingProp = "columnSorting";// int
const char *folderTypeProp = "type";                  // int (MailFolderType)
const char *folderUnreadCountProp = "unreadCount";    // int

/**
 * Random local prototypes
 */
int tooManyMessages();


static const char *messageMimeType = "message/x-kiltdown-message";



/**
 *=------------------------------------------------------------------------=
 *                         MailDragObject
 *=------------------------------------------------------------------------=
 * this is a drag object for mail messages.
 */
class MailDragObject : public QStoredDrag {

  public:
    MailDragObject(QWidget *source) : QStoredDrag(messageMimeType, source) { }
};



/**
 *=------------------------------------------------------------------------=
 * MailFolder::MailFolder
 *=------------------------------------------------------------------------=
 * Private constructor.  Please use createFolder or openFolder to get a
 * new one.
 */
MailFolder::MailFolder()
{
    displayName = NULL;
    baseName = NULL;
    site = NULL;

    haveFocus = false;

    database = NULL;

    type = NormalFolder;

    /**
     * headers in the mail list.
     */
    columns = NULL;
    ccolumns = 0;
    sortedUp = false;
    unreadCount = -1;
    totalCount = 0;
    sortedColumn = MC_SUBJECT;
    columnInfoChanged = false;


    __mailListPopup = NULL;

    unopened = opened = replied = forwarded = lowpri = highpri = attach = NULL;
}
 
/**
 *=------------------------------------------------------------------------=
 * MailFolder::~MailFolder
 *=------------------------------------------------------------------------=
 * time to take out the trash!
 */
MailFolder::~MailFolder()
{

    if (columns
        && columns != defaultToColumnInfo
        && columns != defaultFromColumnInfo) {

        localFree(columns);
    }

    if (database) delete database;
    if (baseName != NULL) localFree(baseName);
    if (displayName != NULL) localFree(displayName);

    if (unopened) delete unopened;
    if (opened) delete opened;
    if (replied) delete replied;
    if (forwarded) delete forwarded;
    if (lowpri) delete lowpri;
    if (highpri) delete highpri;
    if (attach) delete attach;

    if (__mailListPopup != NULL) delete __mailListPopup;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::acceptData                                         [IFolder]
 *=------------------------------------------------------------------------=
 * somebody has dropped/pasted some data.  Make sure it's a mail message,
 * and then go and put it in our folder.
 *
 * Parameters:
 *      const char *                - [in] data
 *      unsigned int                - [in] sizeof(data)
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::acceptData
(
    const char *data,
    unsigned int sizeofData
)
{
    MailMessage **msgs;
    ERRCODE er;
    int c;

    /**
     * Get the array of messages.
     */
    msgs = (MailMessage **)data;
    c = sizeofData / sizeof(MailMessage *);

    /**
     * Great, now add the messages to this folder.  We're going to reset
     * the timeval so that duplicates aren't a problem.
     */
    for (int x = 0; x < c; x++) {
        er = addMessage(msgs[x]);
        RETURN_ON_FAILURE(er);
    }
    
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::acceptsFormat                                      [IFolder]
 *=------------------------------------------------------------------------=
 * indicates whether we understand/accept the given format.
 *
 * Parameters:
 *      const char *                - [in]  MIME format.
 *
 * Output:
 *      ERRCODE                     - S_OK -- yes, otherwise, no.
 */
ERRCODE MailFolder::acceptsFormat
(
    const char *format
)
{
    if (!format) return E_PARAM;

    if (!strcasecmp(format, messageMimeType))
        return S_OK;

    return S_FALSE;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::addMessage
 *=------------------------------------------------------------------------=
 * Adds the given message to this folder.
 *
 * Parameters:
 *      MailMessage *               - [in]  add me!
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::addMessage
(
    MailMessage *message
)
{
    enum ReadStatus stat;
    KMailListItem item;
    bool unread = false;
    ERRCODE er;

    if (!message) return E_PARAM;
    if (!database) return E_INTERNAL;

    /**
     * First, add the message to the database.
     */
    er = database->addMessage(message);
    RETURN_ON_FAILURE(er);
//    database->flush();

    /**
     * Figure out if we need to update the unread count too !!!
     */
    message->getStatus(stat);
    switch (stat) {
        case MSG_UNOPENED:
        case MSG_SENDING:
        case MSG_EDITING:
        case MSG_UNSENT:
            unread = true;
            break;
        default:
            break;
    }

    /**
     * If we have the focus, then we need to add this message to the mail
     * list.
     */
    if (haveFocus) {

        memset(&item, 0, sizeof(KMailListItem));

        item.boldEntry = unread;
        item.itemData = (unsigned int)(message);
        er = populateColumnsFromInfo(&item.columns, ccolumns, columns, message);
        RETURN_ON_FAILURE(er);
        item.ccolumns = ccolumns;

        /**
         * add that mama!
         */
        er = messageView->addRow(&item);
        RETURN_ON_FAILURE(er);
    }

    if (unread) {
        this->unreadCount++;
        updateMessageCounts();
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::clearColumns
 *=------------------------------------------------------------------------=
 * handy little clean up routine we've got here. Cleans up the items in the
 * array, and then cleans up the array itself.
 *
 * Parameters:
 *      KMailListColumn *           - [in]  array of stuff to clean up.
 *      int                         - [in]  number of elements.
 */
void MailFolder::clearColumns
(
    KMailListColumn *cols,
    int ccols
)
{
    K_ASSERT(cols && ccols, "D'uH!  parameters dumbass!");

    for (int x = 0; x < ccols; x++) {
        if (cols[x].pixmap) delete cols[x].pixmap;
        if (cols[x].text) localFree(cols[x].text);
    }

    localFree(cols);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::close
 *=------------------------------------------------------------------------=
 * makes sure that our DB files are closed and there is cosmic harmony and
 * happiness.
 */
void MailFolder::close() 
{
    if (database) database->close();
}
 

/**
 *=------------------------------------------------------------------------=
 * MailFolder::columnHeadersChanged
 *=------------------------------------------------------------------------=
 * the headers a folder have changed.  tell it about them now.
 *
 * Parameters:
 *      int                     - [in]  # cols
 *      KMailListHeader *       - [in]  array of headers with new info.
 */
void MailFolder::columnHeadersChanged
(
    int ccols,
    KMailListHeader *hdr
)
{
    MailColumnInfo *info;

    /**
     * We need to translate these headers into MailColumnInfo structs.
     */
    info = (MailColumnInfo *)localAlloc(sizeof(MailColumnInfo) * ccols);
    if (!info) return;

    /**
     * Copy over the needed data.
     */
    for (int x = 0; x < ccols; x++) {
        
        K_ASSERT((int)hdr[x].itemData >= 0 && (int)hdr[x].itemData <= MC_LAST, "Bogus MC_ Code!!!");

        info[x].which = (enum MailColumn)hdr[x].itemData;
        info[x].fixed = hdr[x].type == CW_FIXED;
        info[x].width = hdr[x].width;
        info[x].pct = hdr[x].pct;
        info[x].sort = (MailColumnSorting) hdr[x].sort;
        if ((MailColumnSorting)info[x].sort == MCS_PRIMARY_UP) 
            sortedUp = true;
        else if ((MailColumnSorting)info[x].sort == MCS_PRIMARY_DOWN)
            sortedUp = false;
    }

    if (columns 
        && columns != defaultToColumnInfo
        && columns != defaultFromColumnInfo) {
        localFree(columns);
    }

    columns = info;
    columnInfoChanged = true;
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::compact                                            [IFolder]
 *=------------------------------------------------------------------------=
 * compacts this folder.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::compact()
{
    K_ASSERT(database != NULL, "I must have a database to be a folder!");

    return database->compact();
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::countUnreadMessages
 *=------------------------------------------------------------------------=
 * counts the number of unread message in this folder.  It will set the
 * member variable unreadCount.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::countUnreadMessages()
{
    MailMessage **messages;
    ReadStatus rs;
    int cmessages;
    ERRCODE er;
    int unread = 0;

    if (!database) return E_INTERNAL;

    /**
     * first -- get an array with all the messages in it from the Folder.
     */
    er = database->getMessageHeaders(&messages, &cmessages);
    RETURN_ON_FAILURE(er);

    /**
     * now whiz through getting the status and counting them !!!
     */
    for (int x = 0; x < cmessages; x++) {
        messages[x]->getStatus(rs);
        switch (rs) {
            case MSG_UNOPENED:
            case MSG_SENDING:
            case MSG_EDITING:
            case MSG_UNSENT:
                unread++;
                break;
            default:
                break;
        }
    }

    if (cmessages) localFree(messages);
    this->unreadCount = unread;
    this->totalCount = cmessages;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::createFolder
 *=------------------------------------------------------------------------=
 * Creates a new mail folder object with underlying files and all that jazz.
 * Returns the new object.
 *
 * Parameters:
 *      const char *                - [in]  the display name for this folder.
 *      const char *                - [in]  the base of the filenames that 
 *                                      will be used.  i.e. inbox_hdr.db
 *      const char *                - [in]  directory in which files will be
 *      MailFolderType              - [in]  what type of folder are we??
 *      bool                        - [in]  is the default view as an inbox
 *                                      (from:) or outbox (to:)
 *      MailFolder **               - [out] the new object.
 * 
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::createFolder
(
    const char *name,
    const char *fileName,
    const char *dir,
    MailFolderType type,
    MailFolder **pfolder
)
{
    return openFolderImpl(NULL, name, fileName, dir, type,
                          true, NULL, pfolder);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::deleteMessages
 *=------------------------------------------------------------------------=
 * given an array of messages, delete them from this folder.
 *
 * Parameters:
 *      int             - [in]  number of messages to remove.
 *      MailMessage **  - [in]  array of messages to delete.
 *      bool            - [in]  indicates if we should confirm if we're the
 *                          DeletedItems folder ...
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::deleteMessages
(
    int count,
    MailMessage **deleteme,
    bool confirmIfDI
)
{
    int fewerUnread = 0;
    unsigned int *rg;
    ReadStatus rs;
    bool ignoreAsk = false;

    if (!count || !deleteme) return E_PARAM;

    parent->getIgnoreAskOnDelete(ignoreAsk);

    /**
     * If there are messages to be deleted, and this is the deleted items
     * folder, then ask them if this what they REALLY want to do??
     */
    if (!ignoreAsk && confirmIfDI && this == parent->getDeletedItems()) {
        const char *title, *text;
        title = intlLoadString(ID_CONFIRM);
        text = intlLoadString(ID_DELETE_CONFIRM);
        if (!title || !text) {
            if (title) localFree(title);
            if (text) localFree(text);
            return E_OUTOFMEMORY;
        }

        int ret = QMessageBox::warning((QWidget *)localMainWidget, QString(title), QString(text),
                                       QMessageBox::Yes | QMessageBox::Default,
                                       QMessageBox::No | QMessageBox::Escape);
        localFree(title);
        localFree(text);
        if (ret == QMessageBox::No) return S_FALSE;
    }

    /**
     * First, try and remove them from the UI
     */
    rg = (unsigned int *)localAlloc(count * sizeof(unsigned int));
    if (!rg) return E_OUTOFMEMORY;
    for (int x = 0; x < count; x++)
        rg[x] = (unsigned int)deleteme[x];
    messageView->removeRows(count, rg);
    localFree(rg);

    /**
     * now figure out how many of the messages are unread, and then remove
     * them from the real db ....
     */
    for (int x = 0; x < count; x++) {
        deleteme[x]->getStatus(rs);
        switch (rs) {
            case MSG_UNOPENED:
            case MSG_SENDING:
            case MSG_EDITING:
            case MSG_UNSENT:
                fewerUnread++;
                break;
            default:
                break;
        }
        database->removeMessage(deleteme[x]);
    }

    /**
     * Now tell the site that some more messages are gone.
     */
    this->totalCount -= count;
    this->unreadCount -= fewerUnread;
    updateMessageCounts();

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::deleteSelectedMessages
 *=------------------------------------------------------------------------=
 * deletes the selected messages, asking for confirmation first.
 */
void MailFolder::deleteSelectedMessages()
{
    ERRCODE er = parent->deleteSelectedInFolder(this);

    if (FAILED(er)) {
        errorShowError(er);
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::dieDieDIE
 *=------------------------------------------------------------------------=
 * we're going away.  prepare ourselves for death.
 */
void MailFolder::dieDieDIE()
{
    // UNDONE: marcw, 2.21.2001 -- we need to orphan messages that are
    // opened against this folder !!!!
    if (site) site->closing(this);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::dragItemStarted
 *=------------------------------------------------------------------------=
 * the user has started dragging items in the mail list.  Go and figure out
 * what the current folder wants to do with that.
 */
void MailFolder::dragItemStarted()
{
    MailMessage **msgs;
    bool completedMove;
    ERRCODE er;
    int c;

    /**
     * See if we need to start dragging at all.
     */
    getSelectedMessages(&c, &msgs);
    if (!c) return;

    MailDragObject *m = new MailDragObject(messageView);
    if (!m) {
        errorShowError(E_OUTOFMEMORY);
        return;
    }

    QByteArray b;
    b.duplicate((const char *)msgs, sizeof(MailMessage *) * c);
    m->setEncodedData(b);

    completedMove = m->dragMove();

    /**
     * Now, if the drag succeeded, go and remove the messages from this
     * folder.
     */
    if (completedMove) {

        er = this->deleteMessages(c, msgs, false);
        if (FAILED(er)) {
            errorShowError(er);
            return;
        }
    }

    localFree(msgs);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::enableMailListPopupItems
 *=------------------------------------------------------------------------=
 * enables and disables items on the maillist popup depending on what's 
 * selected right now, etc ...
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::enableMailListPopupItems()
{
    LinkList<unsigned int> selectedItemDataList;
    bool enableUnread = false;
    bool enableRead = false;
    unsigned int ui;
    ReadStatus rs;
    ERRCODE er;
    int cItems;

    /**
     * First, get a count of selected rows.
     */
    er = messageView->getSelectedRows(selectedItemDataList);
    RETURN_ON_FAILURE(er);

    /**
     * Now, enable and disable items as appropriate.
     */
    cItems = selectedItemDataList.countElements();
    if (cItems == 0) {

        /**
         * No items selected, disable everything.
         */
        __mailListPopup->setItemEnabled(ID_MESSAGE_OPEN, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_PRINT, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_REPLY_SENDER, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_REPLY_ALL, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_FORWARD, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_FORWARD_ATTACHMENT, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_MARKASREAD, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_MARKASUNREAD, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_DELETE_TEXT, false);
        __mailListPopup->setItemEnabled(ID_MESSAGE_ADDSENDER_TOAB, false);

    } else {

        /**
         * These are always enabled when there are items selected.
         */
        __mailListPopup->setItemEnabled(ID_MESSAGE_OPEN, true);
        __mailListPopup->setItemEnabled(ID_MESSAGE_PRINT, true);
        __mailListPopup->setItemEnabled(ID_MESSAGE_FORWARD, true);
        __mailListPopup->setItemEnabled(ID_MESSAGE_FORWARD_ATTACHMENT, true);
        __mailListPopup->setItemEnabled(ID_MESSAGE_DELETE_TEXT, true);

        /**
         * These are only enabled for one item selected.
         */
        __mailListPopup->setItemEnabled(ID_MESSAGE_REPLY_SENDER, cItems == 1);
        __mailListPopup->setItemEnabled(ID_MESSAGE_REPLY_ALL, cItems == 1);
        __mailListPopup->setItemEnabled(ID_MESSAGE_ADDSENDER_TOAB, cItems == 1);

        /**
         * These depend on what's selected or not.
         */
        er = selectedItemDataList.firstElement(&ui);
        while (er == S_OK) {

            K_ASSERT(ui != 0, "Very Bogus message pointer !!");
            
            ((MailMessage *)ui)->getStatus(rs);

            if (rs == MSG_UNOPENED)
                enableRead = true;
            else 
                enableUnread = true;
            if (enableRead && enableUnread) break;

            er = selectedItemDataList.nextElement(&ui);
        }

        __mailListPopup->setItemEnabled(ID_MESSAGE_MARKASREAD, enableRead);
        __mailListPopup->setItemEnabled(ID_MESSAGE_MARKASUNREAD, enableUnread);

    }

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::ensureIconsLoaded
 *=------------------------------------------------------------------------=
 * makes sure all our icons are loaded ...
 */
void MailFolder::ensureIconsLoaded()
{
    if (!unopened) {
        unopened = new QPixmap(unopenedIcon);
        opened = new QPixmap(openedIcon);
        replied = new QPixmap(repliedIcon);
        forwarded = new QPixmap(forwardedIcon);
        lowpri = new QPixmap(lowpriIcon);
        highpri = new QPixmap(priorityhdrIcon);
        attach = new QPixmap(attachhdrIcon);
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::findMessage
 *=------------------------------------------------------------------------=
 * indicates whether this message is in this folder or not.
 *
 * Parameters:
 *      MailMessage *           - [in]  find me please.
 *
 * Output:
 *      ERRCODE                 - S_OK -yes
 *                                S_FALSE nope.
 */
ERRCODE MailFolder::findMessage
(
    MailMessage *message
)
{
    return database->findMessage(message);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::flush
 *=------------------------------------------------------------------------=
 * makes sure that any changes we've made have been written to the disk.
 */
void MailFolder::flush()
{
    if (database) database->flush();
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::forwardSelected                                       [IFolder]
 *=------------------------------------------------------------------------=
 * replies to the selected messages.
 */
void MailFolder::forwardSelected()
{
    MailMessage **rg;
    ERRCODE er;
    int ct;

    /**
     * See how many messages we have to deal with !!!
     */
    getSelectedMessages(&ct, &rg);
    if (ct > TOOMANY_MESSAGES) {
        int ret = tooManyMessages();
        if (ret == QMessageBox::No) {
            localFree(rg);
            return;
        }
    }

    /**
     * whiz through the list and open repliers!!
     */
    for (int x = 0; x < ct; x++) {

        MailEditWindow *wnd;
        er = MailEditWindow::forward(rg[x], this->parent, &wnd);
        if (FAILED(er)) {
            errorShowError(er);
            localFree(rg);
            return;
        }

        wnd->show();
    }

    localFree(rg);
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::gainFocus                                          [IFolder]
 *=------------------------------------------------------------------------=
 * Indicates that we've been given focus.  We need to go and populate the 
 * Mail list now ....
 */
void MailFolder::gainFocus()
{
    ERRCODE er;

    haveFocus = true;
    ensureIconsLoaded();

    /**
     * Clear out the message view!!
     */
    K_ASSERT(messageView, "I really need something to draw with !!!");
    messageView->setUpdatesEnabled(false);
    messageView->nukeAll();

    /**
     * set up the header information.
     */
    er = setHeadersInFolder();
    GOTO_ON_FAILURE(er, Error);

    /**
     * set up the message list in the folder.
     */
    er = setMessagesInFolder();
    GOTO_ON_FAILURE(er, Error);

    /**
     * hook up some events -- do it BEFORE we go and select some items.
     */
    QObject::connect(messageView, SIGNAL(columnHeadersChanged(int, KMailListHeader *)),
                     this, SLOT(columnHeadersChanged(int, KMailListHeader *)));
    QObject::connect(messageView, SIGNAL(itemLaunched(KMailListEvent *)),
                     this, SLOT(itemLaunched(KMailListEvent *)));
    QObject::connect(messageView, SIGNAL(itemSelected(KMailListEvent *)),
                     this, SLOT(itemSelected(KMailListEvent *)));
    QObject::connect(messageView, SIGNAL(rightButtonPressed(QMouseEvent *)),
                     this, SLOT(rightButtonPressed(QMouseEvent *)));
    QObject::connect(messageView, SIGNAL(keyPressed(QKeyEvent *)),
                     this, SLOT(keyPressed(QKeyEvent *)));
    QObject::connect(messageView, SIGNAL(dragItemStarted()),
                     this, SLOT(dragItemStarted()));

    /**
     * Finally, Select either the first or last element, depending on our sort
     * order.
     */
    if (sortedUp)
        messageView->selectFirst();
    else
        messageView->selectLast();

    updateMessageCounts();

  Error:
    if (FAILED(er)) errorShowError(er);
    messageView->setUpdatesEnabled(true);
    messageView->repaint();
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::getContextMenu                                     [IFolder]
 *=------------------------------------------------------------------------=
 * returns a context menu for this folder, or S_NODATA if there isn't one
 * available.  CALLER has full lifetime control over the popup menu 
 * returned. (i.e. we allocate and return, they free)
 *
 * Parameters:
 *      QPopupMenu **             - [out] the popup to show, or NULL
 *
 * Output:
 *      ERRCODE                   - S_NODATA if there isn't a menu.
 */
ERRCODE MailFolder::getContextMenu
(
    QPopupMenu **ppopup
)
{
    if (!ppopup) return E_PARAM;

    *ppopup = NULL;
    return S_NODATA;
}



/**
 *=------------------------------------------------------------------------=
 * MailFolder::getDisplayName                                     [IFolder]
 *=------------------------------------------------------------------------=
 * returns the user friendly name for this folder.
 *
 * Parameters:
 *      const char **           - [out] the name.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::getDisplayName
(
    const char **pname
)
{
    if (!pname) return E_PARAM;
    
    K_ASSERT(displayName, "I'm a folder without a name.  This isn't good.");
    if (!displayName) return E_INTERNAL;

    *pname = copyString(this->displayName);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::getIcon                                            [IFolder]
 *=------------------------------------------------------------------------=
 * returns the icon that is to be displayed with this folder in the folder
 * view. CALLER frees resulting pixmap with "delete"
 *
 * Parameters:
 *      QPixmap **              - [out] our pixmap.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::getIcon
(
    QPixmap **picon
)
{
    if (!picon) return E_PARAM;

    switch (this->type) {
        case InboxFolder:
            *picon = new QPixmap(inboxIcon);
            break;
        case OutboxFolder:
            *picon = new QPixmap(outboxIcon);
            break;
        case SentItemsFolder:
            *picon = new QPixmap(sentitemsIcon);
            break;
        case DeletedItemsFolder:
            *picon = new QPixmap(deleteditemsIcon);
            break;
        default:
            *picon = new QPixmap(mailfolderIcon);
            break;
    }

    return (*picon) ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::getMailListPopup
 *=------------------------------------------------------------------------=
 * returns the popup menu we'll display in the mail list.
 *
 * Output:
 *      QPopupMenu *            - NULL means OOM
 */
QPopupMenu *MailFolder::getMailListPopup()
{
    if (__mailListPopup == NULL) {

        __mailListPopup = new QPopupMenu();
        if (!__mailListPopup) 
            return NULL;

        /**
         * Now go and load in the items and add then now.
         */
        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_OPEN),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_OPEN);
        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_PRINT),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_PRINT);

        __mailListPopup->insertSeparator();

        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_REPLY_SENDER),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_REPLY_SENDER);
        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_REPLY_ALL),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_REPLY_ALL);
        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_FORWARD),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_FORWARD);
        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_FORWARD_ATTACHMENT),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_FORWARD_ATTACHMENT);

        __mailListPopup->insertSeparator();

        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_MARKASREAD),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_MARKASREAD);
        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_MARKASUNREAD),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_MARKASUNREAD);

        __mailListPopup->insertSeparator();

        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_DELETE_TEXT),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_DELETE_TEXT);

        __mailListPopup->insertSeparator();

        __mailListPopup->insertItem(intlLoadStringX(ID_MESSAGE_ADDSENDER_TOAB),
                                    this, SLOT(mailListMenuActivated(int)),
                                    0, ID_MESSAGE_ADDSENDER_TOAB);
    }

    enableMailListPopupItems();
    return __mailListPopup;
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::getMessageBody
 *=------------------------------------------------------------------------=
 * returns the body for this message.
 *
 * Parameters:
 *      MailMessage *               - [in]  aaaiiie!!! i've got no legs.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::getMessageBody
(
    MailMessage *imMissingSomething
)
{
    K_ASSERT(database, "Who's asking for a message body when we've got no db?");
    return database ? database->getMessageBody(imMissingSomething) : E_INTERNAL;
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::getMessages
 *=------------------------------------------------------------------------=
 * returns a copy of the list of messages in this folder.
 *
 * Parameters:
 *      LinkList<MailMessage *> & msgs
 *
 * Output:
 *      ERRCODE
 *
 * NOTE:  THE MESSAGES RETURNED HAVE THEIR REF COUNT INCREASED BY ONE, SO
 *      IN CASE THEY ARE DELETED OR REMOVED, THERE WON'T BE A BAD/DANGLING
 *      POINTER.  CALLEE *MUST* release() THESE POINTERS!!!!
 */
ERRCODE MailFolder::getMessages
(
    LinkList<MailMessage *> &msgs
)
{
    return database->getMessages(msgs);    
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::getSelectedMessages
 *=------------------------------------------------------------------------=
 * returns an array with the currently selected messages in this folder.
 *
 * Parameters:
 *      int *                   - [out] number of selected messages.
 *      MailMessage ***         - [out] where to put pointer to an array.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::getSelectedMessages
(
    int *count,
    MailMessage ***parray
)
{
    LinkList<unsigned int> list;
    unsigned int a;
    ERRCODE er;
    int x;

    if (!haveFocus) {
        K_FAIL("Who's asking me to delete items when I don't have focus??");
        return E_INTERNAL;
    }

    if (!count || !parray) return E_PARAM;
    *count = 0;
    *parray = NULL;

    /**
     * Get the selected rows.
     */
    er = messageView->getSelectedRows(list);
    RETURN_ON_FAILURE(er);
    *count = list.countElements();
    if (!(*count)) return S_FALSE;

    *parray = (MailMessage **)localAlloc(sizeof(MailMessage *) * (*count));
    if (!(*parray)) return E_OUTOFMEMORY;

    x = 0;
    er = list.firstElement(&a);
    while (er == S_OK) {
        (*parray)[x++] = (MailMessage *)a;
        er = list.nextElement(&a);
    }
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::getUnreadCount                                     [IFolder]
 *=------------------------------------------------------------------------=
 * returns the number of unread items in this folder.
 *
 * Parameters:
 *      int *                   - [out] the number of unread messages in
 *                                  this folder.
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::getUnreadCount
(
    int *unread
)
{
    if (!unread) return E_PARAM;

    /**
     * Go and get the count if we don't have it and return the number !!!
     */
    if (unreadCount < 0) countUnreadMessages();
    *unread = this->unreadCount;
    return S_NODATA;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::hasMessages
 *=------------------------------------------------------------------------=
 * does we gots any messages???? 
 *
 * Output:
 *      ERRCODE                     - S_OK      AYE!
 *                                    S_FALSE   NAY!
 */
ERRCODE MailFolder::hasMessages()
{ 
    K_ASSERT(database, "I don't have a DB yet???");
    if (!database) return E_INTERNAL;
    return database->hasMessages(); 
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::itemLaunched
 *=------------------------------------------------------------------------=
 * An item in the mail list has been double clicked or had ENTER pressed on
 * it.  that means we need to do  something!
 *
 * Parameters:
 *      KMailListEvent *            - [in]  details!
 */
void MailFolder::itemLaunched
(
    KMailListEvent *
)
{
    LinkList<unsigned int> selectedItemDataList;
    unsigned int itemData;
    MailMessage *message;
    ReadStatus rs;
    ERRCODE er;
    int count;

    K_ASSERT(parent, "I should always have a parent by now");

    /**
     * First, get the number of selected messages.
     */
    messageView->getSelectedRows(selectedItemDataList);
    
    count = selectedItemDataList.countElements();

    if (count > TOOMANY_MESSAGES) {

        int ret = tooManyMessages();
        if (ret == QMessageBox::No) return;
    }

    /**
     * Whiz through creating the viewers as we need.
     */
    er = selectedItemDataList.firstElement(&itemData);
    while (er == S_OK) {

        message = (MailMessage *)itemData;

        K_ASSERT(message, "Bogus Message pointer selected!!!");
        if (message == NULL) break;

        message->getStatus(rs);

        /**
         * mark the message as read.
         */
        er = updateMessageStatus(message, MSG_READ);
        GOTO_ON_FAILURE(er, Error);

        /**
         * Now go and display it !!!
         */
        if (rs != MSG_UNSENT) {
            MailViewer *viewer = new MailViewer(this);
            er = viewer->initialize((MailMessage *)itemData);
            if (FAILED(er)) {
                delete viewer;
                goto Error;
            }
            viewer->show();
        } else {
            MailEditWindow *edit;
             
            er = MailEditWindow::recomposeNew(message, this->parent, &edit);
            GOTO_ON_FAILURE(er, Error);
            edit->show();
        }

        /**
         * Next element in the list.
         */
        er = selectedItemDataList.nextElement(&itemData);
    }

    er = S_OK;

  Error:
    if (FAILED(er)) errorShowError(er);
    return;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::itemSelected
 *=------------------------------------------------------------------------=
 * an item has been selected in the mail list.  if there is a message for 
 * this item, then show it in the preview pane.
 *
 * Parameters:
 *      KMailListEvent *            - [in]  details, details, details ...
 */
void MailFolder::itemSelected
(
    KMailListEvent *event
)
{
    KMailListItem **items = NULL;
    const char *body = NULL;
    MailMessage *message;
    const char *pstr;
    MIMETYPE type;
    ERRCODE er;
    int count;
    int i;

    K_ASSERT(event, "Who's giving me a NULL event object?");

    /**
     * If there are no items selected, then we just clear out the preview ...
     */
    count = event->itemCount();
    if (!count) {
        messagePreview->clear();
        return;
    }

    /**
     * Otherwise, we need to get the current message, get it's body, and
     * put it in the preview window ...
     */
    items = event->items();
    if (!items) return;

    message = (MailMessage *)items[0]->itemData;
    K_ASSERT(message, "KMailListItem with no MailMessage in the item data !!!!");

    type.setType(Text, Plain);
    er = message->getBodyByType(type, &body);
    if (er == E_NOTFOUND) {
        type.setType(Text, Html);
        er = message->getBodyByType(type, &body);
        if (er == E_NOTFOUND)
            return;
    }
    GOTO_ON_FAILURE(er, Error);

    parent->getReadFont(pstr, i);
    if (pstr) {
        messagePreview->setFont(QFont(pstr, i));
        localFree(pstr);
    }
    messagePreview->setUndoEnabled(false);
    messagePreview->setText(body);
    messagePreview->setCursorPosition(0, 0);
    messagePreview->setUndoEnabled(true);

    er = S_OK;

  Error:
    if (items) localFree(items);
    if (body) localFree(body);
    
    if (FAILED(er)) errorShowError(er);
    return;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::keyPressed
 *=------------------------------------------------------------------------=
 * A key in the message view has been pressed.  We're interested in a couple
 * of them.   Go and figure out what to do ...
 *
 * Parameters:
 *      QKeyEvent *             - [in]  details!
 */
void MailFolder::keyPressed
(
    QKeyEvent *event
)
{
    ERRCODE er;

    switch (event->key()) {
        case Key_Delete:
            er = parent->deleteSelectedInFolder(this);
            if (FAILED(er)) errorShowError(er);
            break;
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::launchSelectedMessages
 *=------------------------------------------------------------------------=
 * open windows for the selected messages.  Warn if they're opening too
 * many.
 */
void MailFolder::launchSelectedMessages()
{
    itemLaunched(NULL);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::loadFolderInfo
 *=------------------------------------------------------------------------=
 * Loads in the information about the folder, including which columns are
 * currently being displayed for this folder.
 *
 * Parameters:
 *      const char *            - [in]  the base name to look for in the props
 *      bool                    - [in]  are we an outbox/sent items type f?
 *      PropertiesFile *        - [in]  which file holds the info.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::loadFolderInfo
(
    const char *base,
    MailFolderType type,
    PropertiesFile *props
)
{
    char buf[MAX_NAMEVALUE_LEN];
    MailColumnInfo *cols;
    int ival, ccols = 0;
    double dval;
    ERRCODE er;
    bool bval;

    /**
     * If no columns have been specified in the file, then we just use the
     * default ones.  Very Eeee-Zed.
     */
    if (props) {
        /**
         * Start getting the info.
         */
        sprintf(buf, "%s.%s", base, folderColumnCountProp);
        er = props->getIntProperty(buf, &ccols);
        RETURN_ON_FAILURE(er);

        sprintf(buf, "%s.%s", base, folderSortedProp);
        er = props->getIntProperty(buf, &ival);
        RETURN_ON_FAILURE(er);
        this->sortedColumn = (enum MailColumn)ival;

        sprintf(buf, "%s.%s", base, folderSortedUpProp);
        er = props->getBooleanProperty(buf, &bval);
        RETURN_ON_FAILURE(er);
        this->sortedUp = bval;

        /**
         * Count of unread messages.
         */
        sprintf(buf, "%s.%s", base, folderUnreadCountProp);
        er = props->getIntProperty(buf, &ival);
        RETURN_ON_FAILURE(er);
        if (er == S_NODATA)
            this->unreadCount = -1;
        else
            this->unreadCount = ival;

        /**
         * Don't forget to get the type of folder.
         */
        snprintf(buf, sizeof(buf), "%s.%s", base, folderTypeProp);
        er = props->getIntProperty(buf, &ival);
        RETURN_ON_FAILURE(er);
        type = (MailFolderType)ival;
    }


    /** 
     * Save out the type!
     */
    this->type = type;
    K_ASSERT(this->type != InvalidFolderType, "I Should definitely have a type by now !!!\n");

    if (!props || ccols == 0) {
        if (type == OutboxFolder || type == SentItemsFolder) {
            this->columns = defaultToColumnInfo;
            this->sortedColumn = MC_SENT;
            this->sortedUp = false;
            this->ccolumns = DEFAULT_TO_COUNT;
        } else {
            this->columns = defaultFromColumnInfo;
            this->sortedColumn = MC_RECEIVED;
            this->sortedUp = false;
            this->ccolumns = DEFAULT_FROM_COUNT;
        }
        return S_OK;
    }

    /**
     * there are folders in here.  Go and load them in now!!!
     */
    cols = (MailColumnInfo *)localAlloc(sizeof(MailColumnInfo) * ccols);
    if (!cols) return E_OUTOFMEMORY;

    for (int x = 0; x < ccols; x++) {
        sprintf(buf, "%s.column%d.%s", base, x, folderWhichColumnProp);
        er = props->getIntProperty(buf, &ival);
        GOTO_ON_FAILURE(er, Error);
        if (ival >= MC_LAST) {
            er = E_BAD_CONFIG;
            goto Error;
        }
        cols[x].which = (enum MailColumn)ival;

        sprintf(buf, "%s.column%d.%s", base, x, folderColumnWidthProp);
        er = props->getIntProperty(buf, &ival);
        GOTO_ON_FAILURE(er, Error);
        cols[x].width = ival;

        sprintf(buf, "%s.column%d.%s", base, x, folderColumnPercentProp);
        er = props->getDoubleProperty(buf, &dval);
        GOTO_ON_FAILURE(er, Error);
        cols[x].pct = dval;

        sprintf(buf, "%s.column%d.%s", base, x, folderColumnFixedProp);
        er = props->getBooleanProperty(buf, &bval);
        GOTO_ON_FAILURE(er, Error);
        cols[x].fixed = bval;

        sprintf(buf, "%s.column%d.%s", base, x, folderColumnSortingProp);
        er = props->getIntProperty(buf, &ival);
        GOTO_ON_FAILURE(er, Error);
        if (ival > MCS_LAST) {
            er = E_BAD_CONFIG;
            goto Error;
        }
        cols[x].sort = (MailColumnSorting)ival;
    }

    this->columns = cols;
    this->ccolumns = ccols;

    er = S_OK;

  Error:
    if (FAILED(er)) {
        localFree(cols);
    }
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::lockForOutput
 *=------------------------------------------------------------------------=
 * takes any UNSENT messages in this folder, and marks them as being in the
 * process of being sent.  This way, the user cannot open them for rediting
 * or any such deeley-bop.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::lockForOutput()
{
    return database->lockForOutput();
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::loseFocus
 *=------------------------------------------------------------------------=
 * we've lost the focus.  we are obliged to release any events we've got
 * hooked up on the messageView now.
 */
void MailFolder::loseFocus()
{
    haveFocus = false;

    QObject::disconnect(messageView,
                        SIGNAL(columnHeadersChanged(int, KMailListHeader *)),
                        this,
                        SLOT(columnHeadersChanged(int, KMailListHeader *)));
    QObject::disconnect(messageView, SIGNAL(itemLaunched(KMailListEvent *)),
                        this, SLOT(itemLaunched(KMailListEvent *)));
    QObject::disconnect(messageView, SIGNAL(itemSelected(KMailListEvent *)),
                        this, SLOT(itemSelected(KMailListEvent *)));
    QObject::disconnect(messageView, SIGNAL(rightButtonPressed(QMouseEvent *)),
                        this, SLOT(rightButtonPressed(QMouseEvent *)));
    QObject::disconnect(messageView, SIGNAL(keyPressed(QKeyEvent *)),
                        this, SLOT(keyPressed(QKeyEvent *)));
    QObject::disconnect(messageView, SIGNAL(dragItemStarted()),
                        this, SLOT(dragItemStarted()));

    messagePreview->clear();
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::mailListMenuActivated
 *=------------------------------------------------------------------------=
 * the user has clicked on an item in the mail list popup menu.  deal with
 * it now.
 *
 * Parameters:
 *      int                     - [in]  ID of item activated
 */
void MailFolder::mailListMenuActivated
(
    int id
)
{
    /**
     * The user activated an id.  Go and take appropriate action.
     */
    switch (id) {
        case ID_MESSAGE_OPEN:
            launchSelectedMessages();
            break;

        case ID_MESSAGE_ADDSENDER_TOAB:
        case ID_MESSAGE_FORWARD_ATTACHMENT:
        case ID_MESSAGE_PRINT:
            QMessageBox::warning(messageView, "Not Implemented", 
                                 "This functionality isn't implemented yet, sorry!", 
                                 QMessageBox::Ok, 0, 0);
            break;

        case ID_MESSAGE_REPLY_SENDER:
            replySelected(false);
            break;
        case ID_MESSAGE_REPLY_ALL:
            replySelected(true);
            break;

        case ID_MESSAGE_FORWARD:
            forwardSelected();
            break;

        case ID_MESSAGE_MARKASREAD:
            markSelectedAs(MSG_READ);
            break;

        case ID_MESSAGE_MARKASUNREAD:
            markSelectedAs(MSG_UNOPENED);
            break;

        case ID_MESSAGE_DELETE_TEXT:
            deleteSelectedMessages();
            break;
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::markMessageAsForwarded
 *=------------------------------------------------------------------------=
 * indicate in the UI that the message has been forwarded.
 *
 * Parameters:
 *      MailMessage *           - [in]  i'm the dude to update.
 */
void MailFolder::markMessageAsForwarded
(
    MailMessage *message
)
{
    if (!message) return;

    updateMessageStatus(message, MSG_FORWARDED);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::markMessageAsReplied
 *=------------------------------------------------------------------------=
 * indicate in the UI that the message has been replied to.
 *
 * Parameters:
 *      MailMessage *           - [in]  i'm the dude to update.
 */
void MailFolder::markMessageAsReplied
(
    MailMessage *message
)
{
    if (!message) return;

    updateMessageStatus(message, MSG_REPLIED);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::markSelectedAs
 *=------------------------------------------------------------------------=
 * marks the selected messages as either read or unread.  Please note that
 * if we're being told to mark the messsages read, and they're already
 * read, replied to, or forwarded, we will not touch them.
 *
 * Parameters:
 *      ReadStatus                  - [in]  new status
 */
void MailFolder::markSelectedAs
(
    ReadStatus rsNew
)
{
    LinkList<unsigned int> selectedItemDataList;
    unsigned int itemData;
    ERRCODE er;

    /**
     * Get the list of selected rows.
     */
    messageView->getSelectedRows(selectedItemDataList);

    /**
     * Now, for each message selected, go and update its status.
     */
    er = selectedItemDataList.firstElement(&itemData);
    while (er == S_OK) {
        
        ERRCODE er2 = updateMessageStatus((MailMessage *)itemData, rsNew);
        if (FAILED(er2)) {
            errorShowError(er2);
            break;
        }

        /**
         * Get the next message in the list ..
         */
        er = selectedItemDataList.nextElement(&itemData);
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::nameChanged
 *=------------------------------------------------------------------------=
 * the name of the folder has changed.  update things now.
 *
 * Parameters:
 *      const char *                - [in]  new name
 */
void MailFolder::nameChanged
(
    const char *newName
)
{
    if (displayName) localFree(displayName);
    displayName = copyString(newName);

    if (site) site->folderNameChanged(this, newName);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::nextMessage
 *=------------------------------------------------------------------------=
 * returns the "next" message in the list of messages, or NULL if there
 * isn't a next.
 *
 * Parameters:
 *      MailMessage *               - [in]  item to index off of.
 *
 * Output:
 *      MailMessage *               - next, or NULL if none.
 */
MailMessage *MailFolder::nextMessage
(
    MailMessage *index
)
{
    KMailListItem *item;

    item = messageView->nextItem((unsigned int)index);
    
    return item ? (MailMessage *)item->itemData : NULL;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::previousMessage
 *=------------------------------------------------------------------------=
 * returns the "previous" message in the list of messages, or NULL if there
 * isn't a previous.
 *
 * Parameters:
 *      MailMessage *               - [in]  item to index off of.
 *
 * Output:
 *      MailMessage *               - previous, or NULL if none.
 */
MailMessage *MailFolder::previousMessage
(
    MailMessage *index
)
{
    KMailListItem *item;

    item = messageView->previousItem((unsigned int)index);

    return item ? (MailMessage *)item->itemData : NULL;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::openFolder
 *=------------------------------------------------------------------------=
 * Opens an existing set of folders for our program.  We will fail if they
 * do not already exist.
 *
 * Parameters:
 *      MailAccount *               - [in]  the account in which we work.
 *      const char *                - [in]  our display name.
 *      const char *                - [in]  our base filename.
 *      const char *                - [in]  our directory.
 *      PropertiesFile *            - [in]  where we load our preferences from
 *      MailFolder **               - [out] the new folder.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::openFolder
(
    MailAccount *parent,
    const char *displayName,
    const char *fileName,
    const char *dir,
    PropertiesFile *props,
    MailFolder **pp
)
{
    return openFolderImpl(parent, displayName, fileName, dir,
                          InvalidFolderType, false, props, pp);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::openFolderImpl
 *=------------------------------------------------------------------------=
 * goes and actually performs the open on the mail folder.  opens up the db
 * files using the FolderDB class, and loadsin preferences while we're at
 * it.
 *
 * Parameters:
 *      MailAccount *               - [in]  the account in which we work.
 *      const char *                - [in]  our display name.
 *      const char *                - [in]  our base filename.
 *      const char *                - [in]  our directory.
 *      MailFolderType              - [in]  what type of folder are we?
 *      bool                        - [in]  do we create it if it doesn't
 *                                      exist yet?
 *      PropertiesFile *            - [in]  where we load our preferences from
 *      MailFolder **               - [out] the new folder.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::openFolderImpl
(
    MailAccount *parent,
    const char *displayName,
    const char *baseName,
    const char *directory,
    MailFolderType type,
    bool createIfNecessary,
    PropertiesFile *props,
    MailFolder **pfolder
)
{
    FolderDB *db = NULL;
    ERRCODE er;

    if (!displayName || !baseName || !directory || !pfolder) return E_PARAM;

    *pfolder = new MailFolder();
    if (!*pfolder) return E_OUTOFMEMORY;

    (*pfolder)->displayName = copyString(displayName);
    if (!(*pfolder)->displayName) {
        delete *pfolder;
        return E_OUTOFMEMORY;
    }

    (*pfolder)->baseName = copyString(baseName);
    if (!(*pfolder)->baseName) {
        delete *pfolder;
        return E_OUTOFMEMORY;
    }

    (*pfolder)->parent = parent;
    (*pfolder)->type = type;

    /**
     * Go and create the FolderDB object.
     */
    er = FolderDB::openFolder(*pfolder, baseName, directory, createIfNecessary, &db);
    GOTO_ON_FAILURE(er, Error);
    (*pfolder)->database = db;

    /**
     * Now load in the user's column preferences for this folder.
     */
    er = (*pfolder)->loadFolderInfo(baseName, type, props);
    GOTO_ON_FAILURE(er, Error);

    er = S_OK;

  Error:
    if (FAILED(er)) {
        if (*pfolder) delete *pfolder;
        if (db) delete db;
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::populateColumnsFromInfo
 *=------------------------------------------------------------------------=
 * Populates the KMailListHeader array given a MailMessage object and a set
 * of MailColumnInfo structs describing which columns are needed!
 *
 * Parameters:
 *      KMailListColumn **          - [out] allocated column of structs.  You
 *                                      must call localFree to clean up!
 *      int                         - [in]  number of columns.
 *      MailColumnInfo *            - [in]  columns to conform to.
 *      MailMessage *               - [in]  message to get info from.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::populateColumnsFromInfo
(
    KMailListColumn **pcolumns,
    int ccolumns,
    MailColumnInfo *schema,
    MailMessage *data
)
{
    LinkList<const char *> strList;
    enum MessagePriority p;
    enum ReadStatus rs;
    const char *pstr;
    ERRCODE er;
    bool bval;
    int ival;

    if (!pcolumns || !ccolumns || !schema || !data) return E_PARAM;

    *pcolumns = (KMailListColumn *)localAlloc(sizeof(KMailListColumn) * ccolumns);
    if (!*pcolumns) return E_OUTOFMEMORY;

    memset(*pcolumns, 0, sizeof(KMailListColumn) * ccolumns);

    /**
     * Whiz through the schema and extract the appropriate information.
     */
    for (int x = 0; x < ccolumns; x++) {

        strList.emptyList();

        switch (schema[x].which) {
            case MC_ATTACHMENTS:
                er = data->getHasAttachments(&bval);
                GOTO_ON_FAILURE(er, Error);
                if (bval) {
                    (*pcolumns)[x].pixmap = attach;
                    if (!(*pcolumns)[x].pixmap) {
                        er = E_OUTOFMEMORY;
                        goto Error;
                    }
                }
                (*pcolumns)[x].alignment = CA_CENTRE;
                break;

            case MC_STATUS:
                data->getStatus(rs);
                switch (rs) {
                    case MSG_UNSENT:
                    case MSG_EDITING:
                    case MSG_SENDING:
                    case MSG_UNOPENED:
                        (*pcolumns)[x].pixmap = unopened;
                        break;
                    case MSG_READ:
                        (*pcolumns)[x].pixmap = opened;
                        break;
                    case MSG_REPLIED:
                        (*pcolumns)[x].pixmap = replied;
                        break;
                    case MSG_FORWARDED:
                        (*pcolumns)[x].pixmap = forwarded;
                        break;
                    default: 
                        K_FAIL("Bogus message status !!!!");
                        break;
                }
                if (!(*pcolumns)[x].pixmap) {
                    er = E_OUTOFMEMORY;
                    goto Error;
                }
                (*pcolumns)[x].alignment = CA_CENTRE;
                break;

            case MC_PRIORITY:
                data->getPriority(p);
                switch (p) {
                    case NormalPriority:
                        break;
                    case LowPriority:
                        (*pcolumns)[x].pixmap = lowpri;
                        if (!(*pcolumns)[x].pixmap) {
                            er = E_OUTOFMEMORY;
                            goto Error;
                        }
                        break;
                    case HighPriority:
                        (*pcolumns)[x].pixmap = /*new QPixmap(priorityhdrIcon);*/ highpri;
                        if (!(*pcolumns)[x].pixmap) {
                            er = E_OUTOFMEMORY;
                            goto Error;
                        }
                        break;
                    default:
                        er = E_INTERNAL;
                        K_FAIL("Bogus message priority!!!!");
                        goto Error;
                        break;
                }
                (*pcolumns)[x].alignment = CA_CENTRE;
                break;

            case MC_FROM:
            case MC_CC:
            case MC_BCC:
            case MC_TO:
                {
                char buf[1024];
                int lensofar = 0;

                switch (schema[x].which) {
                    case MC_CC:
                        er = data->getCCs(strList);
                        break;
                    case MC_BCC:
                        er = data->getBCCs(strList);
                        break;
                    case MC_TO:
                        er = data->getRecipients(strList);
                        break;
                    case MC_FROM:
                        er = data->getSenders(strList);
                        break;
                    default:
                        K_FAIL("Bogus code here!!!");
                        er = E_INTERNAL;
                        break;
                }
                GOTO_ON_FAILURE(er, Error);

                er = strList.firstElement(&pstr);
                while (er == S_OK) {
                    if (pstr) {
                        strncpy(buf + lensofar, pstr, sizeof(buf) - lensofar - 1);
                        lensofar += strlen(pstr);
                        if ((unsigned int)lensofar > sizeof(buf)) break;
                    }
                    er = strList.nextElement(&pstr);
                }
                (*pcolumns)[x].text = copyString(buf);
                if (!(*pcolumns)[x].text) {
                    er = E_OUTOFMEMORY;
                    goto Error;
                }
                (*pcolumns)[x].alignment = CA_LEFT;
                }
                break;

            case MC_SUBJECT:
                er = data->getSubject(&pstr);
                GOTO_ON_FAILURE(er, Error);
                (*pcolumns)[x].text = pstr;
                (*pcolumns)[x].alignment = CA_LEFT;
                break;

            case MC_RECEIVED:
                er = data->getReceived(&pstr);
                GOTO_ON_FAILURE(er, Error);
                (*pcolumns)[x].text = pstr;
                (*pcolumns)[x].alignment = CA_LEFT;
                break;

            case MC_SENT:
                er = data->getSent(&pstr);
                GOTO_ON_FAILURE(er, Error);
                (*pcolumns)[x].text = pstr;
                (*pcolumns)[x].alignment = CA_LEFT;
                break;

            case MC_SIZE:
                {
                char buf[64];
                er = data->getSize(&ival);
                GOTO_ON_FAILURE(er, Error);

                printByteCount(buf, sizeof(buf), ival);
                (*pcolumns)[x].text = copyString(buf);
                if (!(*pcolumns)[x].text) {
                    er = E_OUTOFMEMORY;
                    goto Error;
                }
                (*pcolumns)[x].alignment = CA_LEFT;
                }
                break;

            default:
                K_FAIL("Unknown MailColumn code.  VERY BAD !!!!!");
                er = E_INTERNAL;
                goto Error;
                break;

        } // switch (which)

    } // for(x < ccolumns)

    /**
     * Phew!  we're still alive!
     */
    er = S_OK;

  Error:
    if (FAILED(er)) {
        if (*pcolumns) {
            for (int x = 0; x < ccolumns; x++) {
                if ((*pcolumns)[x].text) localFree((*pcolumns)[x].text);
                if ((*pcolumns)[x].pixmap) delete (*pcolumns)[x].pixmap;
            }
            delete *pcolumns;
        }
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::releaseCompactDisasterLock
 *=------------------------------------------------------------------------=
 * releases the disaster lock for compacting dbs.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::releaseCompactDisasterLock()
{
    return parent->releaseDisasterLock(CompactDatabaseDisaster);
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::removeMessage
 *=------------------------------------------------------------------------=
 * removes the given message from this folder, assuming it's actually IN
 * this folder.
 *
 * Parameters:
 *      MailMessage *               - [in]  remove me please.
 *
 * Output:
 *      ERRCODE                     - E_NOTFOUND if it's not here!!!
 */
ERRCODE MailFolder::removeMessage
(
    MailMessage *message
)
{
    return deleteMessages(1, &message, false);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::removeAllMessages
 *=------------------------------------------------------------------------=
 * removes all the messages in this folder.  VERY DANGEROUS!!!
 *
 * Parameters:
 *      bool                        - [in]  ask before doing so.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::removeAllMessages
(
    bool confirm
)
{
    MailMessage **messages;
    ERRCODE er;
    int c;

    er = database->getMessageHeaders(&messages, &c);
    RETURN_ON_FAILURE(er);

    if (c) {
        er = deleteMessages(c, messages, confirm);
        localFree(messages);
    }
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::replySelected                                       [IFolder]
 *=------------------------------------------------------------------------=
 * replies to the selected messages.
 *
 * Parameters:
 *      bool                        - [in]  reply all ???
 */
void MailFolder::replySelected
(
    bool all
)
{
    MailMessage **rg;
    ERRCODE er;
    int ct;

    /**
     * See how many messages we have to deal with !!!
     */
    getSelectedMessages(&ct, &rg);
    if (ct > TOOMANY_MESSAGES) {
        int ret = tooManyMessages();
        if (ret == QMessageBox::No) {
            localFree(rg);
            return;
        }
    }

    /**
     * whiz through the list and open repliers!!
     */
    for (int x = 0; x < ct; x++) {

        /**
         * Show the message in a new editor window.
         */
        MailEditWindow *wnd;
        er = MailEditWindow::reply(rg[x], this->parent, all, &wnd);
        if (FAILED(er)) {
            errorShowError(er);
            localFree(rg);
            return;
        }

        wnd->show();
    }

    localFree(rg);
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::requestCompactDisasterLock
 *=------------------------------------------------------------------------=
 * requests a disaster lock for the folderdb compact operation.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::requestCompactDisasterLock()
{
    return parent->requestDisasterLock(CompactDatabaseDisaster, this->baseName);
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::rightButtonPressed
 *=------------------------------------------------------------------------=
 * the user has pressed the right button in the mail list.  go and figure
 * out what sort of popup menu we need to apply now.
 *
 * Parameters:
 *      QMouseEvent *                - [in]  more info
 */
void MailFolder::rightButtonPressed
(
    QMouseEvent *event
)
{
    QPopupMenu *m = getMailListPopup();

    if (!m) {
        errorShowError(E_OUTOFMEMORY);
        return;
    }

    m->exec(messageView->mapToGlobal(event->pos()));
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::saveConfiguration
 *=------------------------------------------------------------------------=
 * Saves out our current configuration to the given properties file.  Notably,
 * we save out the column header information, as well as sorting.
 *
 * Parameters:
 *      PropertiesFile *            - [in]  where to save the configuration to
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::saveConfiguration
(
    PropertiesFile *props
)
{
    char buf[MAX_NAMEVALUE_LEN];
    ERRCODE er;

    if (!props) return E_PARAM;

    // unread count
    if (this->unreadCount != -1) {
        sprintf(buf, "%s.%s", this->baseName, folderUnreadCountProp);
        er = props->writeIntProperty(buf, this->unreadCount);
        RETURN_ON_FAILURE(er);
    }

    // type of folder
    snprintf(buf, sizeof(buf), "%s.%s", this->baseName, folderTypeProp);
    er = props->writeIntProperty(buf, (int)this->type);
    RETURN_ON_FAILURE(er);

    /**
     * If nothing has changed, this is very easy.
     */
    if (!columnInfoChanged) return S_OK;

    // # columns
    sprintf(buf, "%s.%s", this->baseName, folderColumnCountProp);
    er = props->writeIntProperty(buf, (int)this->ccolumns);
    RETURN_ON_FAILURE(er);

    // # which column is sorted.
    sprintf(buf, "%s.%s", this->baseName, folderSortedProp);
    er = props->writeIntProperty(buf, (int)this->sortedColumn);
    RETURN_ON_FAILURE(er);

    // direction
    sprintf(buf, "%s.%s", this->baseName, folderSortedUpProp);
    er = props->writeBooleanProperty(buf, sortedUp);
    RETURN_ON_FAILURE(er);


    /**
     * Now whiz through the individual columns and write them.
     */
    for (int x = 0; x < ccolumns; x++) {

        sprintf(buf, "%s.column%d.%s", this->baseName, x, folderWhichColumnProp);
        er = props->writeIntProperty(buf, this->columns[x].which);
        RETURN_ON_FAILURE(er);

        sprintf(buf, "%s.column%d.%s", this->baseName, x, folderColumnWidthProp);
        er = props->writeIntProperty(buf, this->columns[x].width);
        RETURN_ON_FAILURE(er);

        sprintf(buf, "%s.column%d.%s", this->baseName, x, folderColumnPercentProp);
        er = props->writeDoubleProperty(buf, this->columns[x].pct);
        RETURN_ON_FAILURE(er);

        sprintf(buf, "%s.column%d.%s", this->baseName, x, folderColumnFixedProp);
        er = props->writeBooleanProperty(buf, this->columns[x].fixed);
        RETURN_ON_FAILURE(er);

        sprintf(buf, "%s.column%d.%s", this->baseName, x, folderColumnSortingProp);
        er = props->writeIntProperty(buf, this->columns[x].sort);
        RETURN_ON_FAILURE(er);
    }

    /**
     * Pimp Daddy!
     */
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::selectMessage
 *=------------------------------------------------------------------------=
 * Indicates that the given message should be selected in the UI, if 
 * appropriate.
 *
 * Parameters:
 *      MailMessage *               - [in]  select me.
 */
void MailFolder::selectMessage
(
    MailMessage *m
)
{
//    updateMessageStatus(m);
    messageView->selectRowFromItemData((unsigned int)m);
}
 
/**
 *=------------------------------------------------------------------------=
 * MailFolder::setFolderSite                                      [IFolder]
 *=------------------------------------------------------------------------=
 * Sites our site pointer.  we will keep this and get our message view 
 * pointer from it.
 *
 * Parameters:
 *      IFolderSite *           - [in]  our new site.
 */
void MailFolder::setFolderSite
(
    IFolderSite *newSite
)
{
    site = newSite;
    if (!site) return;

    site->getHostInfo(&messageView, &messagePreview);
    K_ASSERT(messageView, "BOGUS Mail list from Site Pointer!!!!");
    K_ASSERT(messagePreview, "BOGUS edit window from site");
    return;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::setHeadersInFolder
 *=------------------------------------------------------------------------=
 * the folder has been selected -- go and set it's headers now.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::setHeadersInFolder()
{
    struct KMailListHeader *headers;
    ERRCODE er;

    K_ASSERT(columns, "I should so totally have some columns by now !!!");

    headers = (KMailListHeader *)localAlloc(sizeof(KMailListHeader) * ccolumns);
    if (!headers) {
        er = E_OUTOFMEMORY;
        goto Error;
    }

    memset(headers, 0, sizeof(KMailListHeader) * ccolumns);

    /**
     * Now whiz through the columns and write out their values.
     */
    for (int x = 0; x < ccolumns; x++) {
        
        headers[x].itemData = columns[x].which;
        headers[x].width = columns[x].width;
        headers[x].type = columns[x].fixed ? CW_FIXED : CW_PERCENTAGE;
        headers[x].sort = (ColumnSorting)columns[x].sort;
        headers[x].pct = columns[x].pct;

        switch (columns[x].which) {

            case MC_ATTACHMENTS:
                headers[x].pixmap = new QPixmap(attachhdrIcon);
                headers[x].sortfn = messageSortPixmap;
                break;

            case MC_STATUS:
                headers[x].pixmap = new QPixmap(mailheaderIcon);
                headers[x].sortfn = messageSortPixmap;
                break;

            case MC_PRIORITY:
                headers[x].pixmap = new QPixmap(priorityhdrIcon);
                headers[x].sortfn = messageSortPixmap;
                break;

            case MC_FROM:
                headers[x].text = intlLoadString(ID_COLUMN_FROM);
                headers[x].sortfn = messageSortString;
                break;
                
            case MC_TO:
                headers[x].text = intlLoadString(ID_COLUMN_TO);
                headers[x].sortfn = messageSortString;
                break;

            case MC_CC:
                headers[x].text = intlLoadString(ID_COLUMN_CC);
                headers[x].sortfn = messageSortString;
                break;

            case MC_BCC:
                headers[x].text = intlLoadString(ID_COLUMN_BCC);
                headers[x].sortfn = messageSortString;
                break;

            case MC_SUBJECT:
                headers[x].text = intlLoadString(ID_COLUMN_SUBJECT);
                headers[x].sortfn = messageSortString;
                break;

            case MC_RECEIVED:
                headers[x].text = intlLoadString(ID_COLUMN_RECEIVED);
                headers[x].sortfn = messageSortDate;
                break;

            case MC_SENT:
                headers[x].text = intlLoadString(ID_COLUMN_SENT);
                headers[x].sortfn = messageSortDate;
                break;

            case MC_SIZE:
                headers[x].text = intlLoadString(ID_COLUMN_SIZE);
                headers[x].sortfn = messageSortString;
                break;

            default:
                K_FAIL("Unknown header in selectFolder !!!!!");
                er = E_INTERNAL;
                goto Error;
                break;
        }

    }

    /**
     * Finally, set up the mail list.
     */
    er = messageView->setColumnHeaders(headers, ccolumns);
    GOTO_ON_FAILURE(er, Error);

    er = S_OK;

  Error:
    if (headers) {
        for (int x = 0; x < ccolumns; x++) {
            if (headers[x].text) localFree(headers[x].text);
            if (headers[x].pixmap) delete headers[x].pixmap;
        }
        localFree(headers);
    }

    return er;
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::setMessagesInFolder
 *=------------------------------------------------------------------------=
 * Set up the message list with the appropriate details about all our 
 * messages.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailFolder::setMessagesInFolder()
{
    MailMessage **messages;
    enum ReadStatus stat;
    KMailListItem item;
    int cmessages;
    ERRCODE er;

    int unread = 0;

    if (!database) return E_INTERNAL;

    /**
     * first -- get an array with all the messages in it from the Folder.
     */
    er = database->getMessageHeaders(&messages, &cmessages);
    RETURN_ON_FAILURE(er);

    /**
     * Now, start dumping them in the message view, one at a time.
     */
    for (int x = 0; x < cmessages; x++) {

        memset(&item, 0, sizeof(KMailListItem));
        messages[x]->getStatus(stat);
        switch (stat) {
            case MSG_UNOPENED:
            case MSG_UNSENT:
            case MSG_SENDING:
            case MSG_EDITING:
                unread++;
                item.boldEntry = true;
                break;
            default:
                item.boldEntry = false;
                break;
        }
        item.itemData = (unsigned int)(messages[x]);

        er = populateColumnsFromInfo(&item.columns, ccolumns, columns, messages[x]);
        GOTO_ON_FAILURE(er, Error);
        item.ccolumns = ccolumns;

        /**
         * add that mama!
         */
        er = messageView->addRow(&item);
        GOTO_ON_FAILURE(er, Error);
    }


    /**
     * Update the count of messages.
     */
    totalCount = cmessages;
    this->unreadCount = unread;
    site->unreadMessageCountChange((IFolder *)this, unread);

    updateMessageCounts();

    er = S_OK;

  Error:
    if (messages) localFree(messages);
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailFolder::supportsReplyForward                                [IFolder]
 *=------------------------------------------------------------------------=
 * mail accounts always support this type of stuff, excepting for when we
 * add more PIM type folders (calendars, etc ...)
 *
 * Output:
 *      true
 */
bool MailFolder::supportsReplyForward()
{
    return true;
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::updateMessageCounts
 *=------------------------------------------------------------------------=
 * tells the host our latest message counts in all the various ways ...
 */
void MailFolder::updateMessageCounts()
{
    const char *pstr;
    char buf[128];

    /**
     * if we're the current folder, update the status bar
     */
    if (haveFocus) {
        pstr = intlLoadString(ID_SB_MSG_COUNT);
        if (!pstr) {
            K_FAIL("Couldn't get message count string !!!");
            return;
        }

        snprintf(buf, sizeof(buf), pstr, totalCount, unreadCount);
        localFree(pstr);

        parent->host->setStatusText(buf);
    }

    /**
     *  always update the folder view 
     */
    site->unreadMessageCountChange(this, this->unreadCount);
}


/**
 *=------------------------------------------------------------------------=
 * MailFolder::updateMessageStatus
 *=------------------------------------------------------------------------=
 * update the message status to be the given value, or return an error if
 * this message is in a state that can't be updated to the desired value for
 * some reason (most common reason -- message is locked for sending).
 *
 * Parameters:
 *      MailMessage *               - [in]  try to update me
 *      ReadStatus                  - [in]  new value.
 *
 * Output:
 *      ERRCODE                     - E_MSG_LOCKED if this message is being
 *                                      sent !!!
 */
ERRCODE MailFolder::updateMessageStatus
(
    MailMessage *message,
    ReadStatus rsNew
)
{
    KMailListItem item;
    ReadStatus rs;
    ERRCODE er;

    /** 
     * first, get the existing read status, so we can figure out what sort
     * of action needs to be taken.
     */
    message->getStatus(rs);

    /**
     * Now, if the message isn't unsent or the message isn't locked for
     * sending, go and update the status with the new value.
     */
    switch (rs) {
        case MSG_REPLIED:
        case MSG_FORWARDED:
            if (rsNew == MSG_READ) rsNew = rs;
            // fall through
        case MSG_READ:
        case MSG_UNOPENED:

            /**
             * update the db with the new status.
             */
            message->setStatus(rsNew);
            er = database->updateHeader(message);
            GOTO_ON_FAILURE(er, Error);
            forceRecount();

            /**
             * Now we have to update the UI for the message, if we have focus
             * in the UI.
             */
            if (!this->haveFocus) break;

            memset(&item, 0, sizeof(KMailListItem));
            item.boldEntry = (rsNew == MSG_UNOPENED);
            item.itemData = (unsigned int)(message);
            er = populateColumnsFromInfo(&item.columns, ccolumns, columns, message);
            GOTO_ON_FAILURE(er, Error);
            item.ccolumns = ccolumns;

            er = messageView->updateRowFromItemData(item.itemData, &item);
            if (FAILED(er)) clearColumns(item.columns, ccolumns);
            GOTO_ON_FAILURE(er, Error);
            break;

        case MSG_UNSENT:
            break;

        case MSG_SENDING:
            er = E_MSG_LOCKED;
            goto Error;
            break;

        default:
            break;
    }

    er = S_OK;

  Error:
    return er;
}





/**
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 *                      Misc local routines
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 */ 






/**
 *=------------------------------------------------------------------------=
 * tooManyMessages
 *=------------------------------------------------------------------------=
 * shows a dialog indicating that there might be a lot of messages coming
 * up.
 *
 * Output:
 *      int                         - QMessageBox::yes or ::no.
 */
int tooManyMessages()
{
    const char *title, *text;
    title = intlLoadString(ID_TOOMANY_TITLE);
    if (!title) {
        errorShowError(E_OUTOFMEMORY);
        return QMessageBox::No;
    }
    text = intlLoadString(ID_TOOMANY_TEXT);
    if (!text) {
        localFree(title);
        errorShowError(E_OUTOFMEMORY);
        return QMessageBox::No;
    }
    int ret = QMessageBox::warning((QWidget *)localMainWidget, title, text,
                                   QMessageBox::Yes | QMessageBox::Default,
                                   QMessageBox::No | QMessageBox::Escape);
    localFree(title);
    localFree(text);
    return ret;
}



