/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _MAILFOLDER_H_
#define _MAILFOLDER_H_

#include <sanity.h>

#include <qobject.h>

#include <kmaillist.h>
#include <ifolder.h>
#include <intl.h>
#include "mailcolumns.h"
#include "mailmessage.h"
#include <properties.h>


class FolderDB;
class QPopupMenu;

/**
 * The current types of folders we have:
 */
enum MailFolderType {

    InvalidFolderType = -1,
    NormalFolder = 0,
    InboxFolder,
    OutboxFolder,
    SentItemsFolder,
    DeletedItemsFolder
};


/**
 * This defines the portion of a mail folder that interacts with the IDE and 
 * participates in UI events, etc ...
 */
class MailFolder : public QObject, public IFolder {

    Q_OBJECT

  public:
    /**
     * Creates a folder with the given name in the given directory.  Two files
     * are actually created, one for headers and one for the message bodies.
     * The key will be a timestamp which individual messages will keep handy.
     * Returnst he created folder.
     */
    static ERRCODE createFolder(const char *name,
                                const char *fileName,
                                const char *dir,
                                MailFolderType type,
                                MailFolder **);

    /**
     * Opens the folder with the given name.  Returns failure on error -- i.e.
     * we do not create it if it exists -- the only way this would happen is
     * if the user deleted the file.
     * Returns the opened folder object.
     */
    static ERRCODE openFolder(MailAccount *parent,
                              const char *name,
                              const char *fileName,
                              const char *dir,
                              PropertiesFile *props,
                              MailFolder **);

    /**
     * Adds the given message to the given folder.
     */
    ERRCODE addMessage(MailMessage *);

    /**
     * Cleans out any data we might have.
     */
    void close();

    /**
     * Ssyncs up the database on the disk.  We'll be sure to do this when we
     * are adding things to the inbox ...
     */
    void flush();

    /**
     * Removes the given messages from this folder.
     */
    ERRCODE deleteMessages(int, MailMessage **, bool confirmDestroy);

    /**
     * Removes the given message from this folder.
     */
    ERRCODE removeMessage(MailMessage *);

    /**
     * BE VERY CAREFUL WITH MEEEEEE !!!!
     */
    ERRCODE removeAllMessages(bool confirm);

    /**
     * returns an array with the currently selected messages in this folder.
     */
    ERRCODE getSelectedMessages(int *count, MailMessage ***messages);

    /**
     * Indicates whether or not we have any messages ...
     */
    ERRCODE hasMessages();

    /**
     * Locks the folder for output -- basically, you can't open any of the 
     * messages in this folder for further editing/etc ...
     */
    ERRCODE lockForOutput();

    /**
     * Use this method to save out the configuration for this folder to
     * the given properties file.  We save out column header information and
     * whatever else is necessary.
     */
    ERRCODE saveConfiguration(PropertiesFile *props);

    /**
     * Returns the messages in this folder.
     */
    ERRCODE getMessages(LinkList<MailMessage *> &msgs);

    /**
     * Indicates whether this message is in this folder or not.
     */
    ERRCODE findMessage(MailMessage *);

    /**
     * These two methods are used to iterate through the messages in this
     * folder.  They return NULL if there isn't a next/previous.
     */
    MailMessage *nextMessage(MailMessage *);
    MailMessage *previousMessage(MailMessage *);

    /**
     * Selects the given message in the folder.
     */
    void selectMessage(MailMessage *);

    /**
     * The name of the folder has been changed.  Update everything to
     * reflect this.
     */
    void nameChanged(const char *newName);

    /**
     * forces teh folder to recount the messages and update any status
     */
    void forceRecount() { countUnreadMessages(); updateMessageCounts(); }


    /**
     * Returns the type of folder that we are !!
     */
    inline MailFolderType getType() { return this->type; }

    /**
     * Returns the account in which we are hosted.
     */
    inline MailAccount *getAccount() { return this->parent; }

    /**
     * returns the 'base' name of this folder for file purposes
     */
    inline const char *getBaseName() { return copyString(this->baseName); }

    /**
     * gets us to prepare ourselves for going away ...
     */
    void dieDieDIE();

    /**
     * Retrieves the body of the given message.
     */
    ERRCODE getMessageBody(MailMessage *imMissingSomething);

    /**
     * Marks the message as replied.
     */
    void markMessageAsForwarded(MailMessage *);
    void markMessageAsReplied(MailMessage *);

    /**
     * Try to update the message's status with the new status.  If the
     * message is in a locked or unmodifiable state, an error message will
     * be returned!
     */
    ERRCODE updateMessageStatus(MailMessage *msg, ReadStatus rs);

    /**
     * These methods are used to go and call the parent accout and
     * ask for some disaster recovery locks.
     */
    ERRCODE requestCompactDisasterLock();
    ERRCODE releaseCompactDisasterLock();

    /**
     * IFolder methods
     */
    virtual ERRCODE acceptData(const char *data, uint sizeofData);
    virtual ERRCODE acceptsFormat(const char *);
    virtual ERRCODE compact();
    virtual void    gainFocus();
    virtual void    loseFocus();
    virtual void    deleteSelectedMessages();
    virtual ERRCODE getContextMenu(QPopupMenu **);
    virtual ERRCODE getDisplayName(const char **);
    virtual ERRCODE getIcon(QPixmap **);
    virtual ERRCODE getUnreadCount(int *);
    virtual void    setFolderSite(IFolderSite *);
    virtual bool    supportsReplyForward();
    virtual void    forwardSelected();
    virtual void    replySelected(bool all);

    virtual ~MailFolder();



  protected slots:

    void columnHeadersChanged(int, KMailListHeader *);
    void rightButtonPressed(QMouseEvent *);
    void itemLaunched(KMailListEvent *);
    void itemSelected(KMailListEvent *);
    void mailListMenuActivated(int);
    void keyPressed(QKeyEvent *);
    void dragItemStarted();



  private:
    /**
     * Don't use me.  Use createFolder() or openFolder() instead.
     */
    MailFolder();

    /**
     * Do we have the focus or not?  if we do, then we need to add this
     * message to the list of messages.
     */
    bool haveFocus;

    /**
     * this is our site pointer.
     */
    IFolderSite *site;

    /**
     * This is our database.
     */
    FolderDB *database;

    /** 
     * This is the account in which we are operating.
     */
    MailAccount *parent;

    /** 
     * This is our base name used for file manipulation, etc.
     */
    const char *baseName;

    /**
     * The type of folder that we are.
     */
    MailFolderType type;

    /**
     * Actually does the opening or creation of a mail folder.
     */
    static ERRCODE openFolderImpl(MailAccount *, const char *, const char *,
                                  const char *, MailFolderType, bool,
                                  PropertiesFile *, MailFolder **);

    /** 
     * the method that goes and loads in the column set for this folder and
     * saves it to the object.
     */
    ERRCODE loadFolderInfo(const char *base, MailFolderType type,
                           PropertiesFile *props);

    /**
     * We've just selected this folder -- set up the headers in the 
     * messageview.
     */
    ERRCODE setHeadersInFolder();

    /**
     * We've just selected this folder -- add the messages to the list and 
     * show them.
     */
    ERRCODE setMessagesInFolder();

    /**
     * Returns the current columns associated with this folder.  You can pass
     * in NULL as the second parameter to just get the count of columns ...
     */
    ERRCODE getColumnInfo(int *count, MailColumnInfo **columns = NULL);

    /**
     * Counts the number of unread messages in this folder.
     */
    ERRCODE countUnreadMessages();

    /**
     * Enable and disable items on the MailList popup as appropriate.
     */
    ERRCODE enableMailListPopupItems();

    /**
     * PopupMenu actions
     */
    void launchSelectedMessages();
    void markSelectedAs(ReadStatus);


    /**
     * This is our printable name.  Used by various UI widgets.
     */
    const char *displayName;

    /**
     * This is where we list messages and preview messages
     */
    KMailList *messageView;
    QMultiLineEdit *messagePreview;

    /**
     * This is our current set of columns and relevant information.
     */
    int ccolumns;
    enum MailColumn sortedColumn;
    MailColumnInfo *columns;
    bool columnInfoChanged;
    bool sortedUp;

    /**
     * message info
     */
    int unreadCount, totalCount;

    /**
     * Pixmaps we'll keep handy
     */
    QPixmap *unopened, *opened, *replied, *forwarded;
    QPixmap *lowpri, *highpri;
    QPixmap *attach;

    /**
     * Popup to show in the UI.  Never call the member directly -- always
     * call getMailListPopup() !!!!
     */
    QPopupMenu *getMailListPopup();
    QPopupMenu *__mailListPopup;

    /**
     * misc crap
     */
    void ensureIconsLoaded();
    void updateMessageCounts();
    void clearColumns(KMailListColumn *, int);
    ERRCODE populateColumnsFromInfo(KMailListColumn **, int,
                                    MailColumnInfo *, MailMessage *);
};

#endif // _MAILFOLDERUI_H_


