/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <sanity.h>

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "attachment.h"
#include "mailfolder.h"
#include "mailmessage.h"
#include "main.h"

/**
 * This class defines an RFC822 message, with MIME extensions supported for 
 * things like file attachments, etc
 */



/**
 *=------------------------------------------------------------------------=
 * MailMessage::createNew
 *=------------------------------------------------------------------------=
 * creates a new mail message object.  Very complex.  use me instead of a
 * constructor.
 *
 * Parameters:
 *      MailFolder *            - [in]  our host folder.
 *      MailMessage **          - [out] the new object.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::createNew
(
    MailFolder *parent,
    MailMessage **pmessage
)
{
    if (!pmessage) return E_PARAM;

    *pmessage = new MailMessage();
    if (*pmessage) (*pmessage)->hostFolder = parent;
    return (*pmessage) ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::MailMessage
 *=------------------------------------------------------------------------=
 * private constructor to create an set up a blank message.
 */
MailMessage::MailMessage()
{
    DEBUG_TEST_START;

    m_cref = 1;


    bodySize = 0;
    subject = NULL;
    messageID = NULL;
    replyTo = NULL;
    returnPath = NULL;
    sent = (time_t)-1;
    realSender = NULL;
    hostFolder = NULL;
    priority = NormalPriority;
    received = (time_t)-1;
    status = MSG_UNOPENED;
    hasAttachments = false;
    mimeType.setType(Text, Plain);
    memset(&tv, 0, sizeof(tv));
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::~MailMessage
 *=------------------------------------------------------------------------=
 * cleans up anything that needs cleaning up!!!
 */
MailMessage::~MailMessage()
{
    InternalMimeContent *mimeBody;
    Attachment *a;
    ERRCODE er;

    DEBUG_TEST_STOP;

    K_ASSERT(m_cref == 0, "Being deleted with dangling refs!!!");


    if (subject) localFree(subject);
    if (replyTo) localFree(replyTo);
    if (returnPath) localFree(returnPath);
    if (messageID) localFree(messageID);
    if (realSender) localFree(realSender);

    nukeStringList(extraHeaders);
    nukeStringList(recipients);
    nukeStringList(senders);
    nukeStringList(bccs);
    nukeStringList(ccs);

    er = attachments.firstElement(&a);
    while (er == S_OK) {
        a->release();
        er = attachments.nextElement(&a);
    }

    er = mimeBodies.firstElement(&mimeBody);
    while (er == S_OK) {
        if (mimeBody->body) localFree(mimeBody->body);
        if (mimeBody->charset) localFree(mimeBody->charset);
        localFree(mimeBody);
        er = mimeBodies.nextElement(&mimeBody);
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::addAttachment
 *=------------------------------------------------------------------------=
 * Adds the given attachment to this message.
 *
 * Parameters:
 *      Attachment *                - [in]  object to attach.
 *
 * Output:
 *      ERRCODE                     - look for E_FILE_ERR to mean the
 *                                      attachment is bad.
 */
ERRCODE MailMessage::addAttachment
(
    Attachment *attachMe
)
{
    if (!attachMe) return E_PARAM;

    /** 
     * Delightful.  Now we'll keep a copy of it.
     */
    attachMe->addRef();
    attachments.addElement(attachMe);
    hasAttachments = true;
    mimeType.setType(Multipart, Mixed);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::addRef
 *=------------------------------------------------------------------------=
 * add a ref count to us.
 *
 * Output:
 *      int
 */
int MailMessage::addRef()
{
    return ++m_cref;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::ensureBodyLoaded
 *=------------------------------------------------------------------------=
 * makes sure that we've loaded in our body.
 */
ERRCODE MailMessage::ensureBodyLoaded()
{
    MIMETYPE mt;
    ERRCODE er;

    if (mimeBodies.isEmpty()) {

        /**
         * There are rare cases where a message won't have a body -- all of 
         * its content will be marked as attachment by the headers.  we'll
         * create a blank body for them now.
         */
        if (!hostFolder) {
            mt.setType(Text, Plain);
            setBodyByType(mt, "", NULL);
            return S_OK;
        }

        er = hostFolder->getMessageBody(this);
        RETURN_ON_FAILURE(er);
    }
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::getAttachments
 *=------------------------------------------------------------------------=
 * copies over a list of the attachments to the list given.
 *
 * Parameters:
 *      LinkList<Attachment *> &        - [in]  where to put new list.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getAttachments
(
    LinkList<Attachment *> &list
)
{
    list = this->attachments;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getBCCs
 *=------------------------------------------------------------------------=
 * returns a list of people to whom this message was/is being blind carbon
 * copied.  This is a linked list class.  The object should be passed by 
 * reference.
 *
 * Parameters:
 *      LinkList<const char *> &  - [in] a reference to an object in which to
 *                                       put all the data.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getBCCs
(
    LinkList<const char *> &bccs
)
{
    /**
     * use the copy operator on the linklist class to copy over this list.
     */
    bccs = this->bccs;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getBodyByType
 *=------------------------------------------------------------------------=
 * Returns the body of this message for the given type.  If none is 
 * available for that type, return an error.
 *
 * Parameters:
 *      MIMETYPE                    - [in]  type of body we want.
 *      const char **               - [out] the body for the given format.
 *                                      CALLER must free.
 *      const char ** [optional]    - [out] the charset for this section.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getBodyByType
(
    MIMETYPE type,
    const char **pbody,
    const char **pcharset
)
{
    InternalMimeContent *body;
    ERRCODE er;

    if (!pbody) return E_PARAM;
    *pbody = NULL;

    /**
     * Make sure that the message is fully loaded.
     */
    er = this->ensureBodyLoaded();
    RETURN_ON_FAILURE(er);

    /**
     * Go and find this body
     */
    er = mimeBodies.firstElement(&body);
    while (er == S_OK) {
        if (body->type == type) {
            *pbody = copyString(body->body ? body->body : "");
            if (!*pbody) return E_OUTOFMEMORY;
            if (pcharset && body->charset) {
                *pcharset = copyString(body->charset);
                if (!*pcharset) {
                    localFree(*pbody);
                    *pbody = NULL;
                    return E_OUTOFMEMORY;
                }
            }
            return S_OK;
        }
        er = mimeBodies.nextElement(&body);
    }

    return E_NOTFOUND;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getBodyFormats
 *=------------------------------------------------------------------------=
 * returns the formats supported by this message.
 *
 * Parameters:
 *      MIMETYPE *              - [out] array of types.
 *      int *                   - [out] count.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getBodyFormats
(
    MIMETYPE **formats,
    int *count
)
{
    InternalMimeContent *body;
    ERRCODE er;
    int c, x;

    if (!formats || !count) return E_PARAM;

    er = this->ensureBodyLoaded();
    RETURN_ON_FAILURE(er);

    c = mimeBodies.countElements();

    if (!c) {
        *formats = NULL;
        *count = 0;
        return S_OK;
    }

    *formats = (MIMETYPE *)localAlloc(sizeof(MIMETYPE) * c);
    if (!(*formats)) return E_OUTOFMEMORY;

    x = 0;
    er = mimeBodies.firstElement(&body);
    while (er == S_OK) {
        (*formats)[x++] = body->type;
        er = mimeBodies.nextElement(&body);
    }

    *count = c;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getBodyRefByType
 *=------------------------------------------------------------------------=
 * This simply returns a pointer to our version of the BODY.  THIS IS VERY
 * DANGEROUS !!!!  do not use this for more than the length of a function
 * call.
 *
 * Parameters:
 *      MIMETYPE                    - [in]  the type of body being requested.
 *      const char **               - [out] the body in text format.  CALLER
 *                                      MUST NOT FREE.
 *      const char ** [optional]    - [out] charset for this section of the
 *                                      body.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getBodyRefByType
(
    MIMETYPE type,
    const char **pbody,
    const char **pcharset
)
{
    InternalMimeContent *body;
    ERRCODE er;

    if (!pbody) return E_PARAM;
    *pbody = NULL;

    er = ensureBodyLoaded();
    RETURN_ON_FAILURE(er);

    er = mimeBodies.firstElement(&body);
    while (er == S_OK) {
        if (body->type == type) {
            *pbody = body->body ? body->body : "";
            if (pcharset) *pcharset = body->charset;
            return S_OK;
        }
        er = mimeBodies.nextElement(&body);
    }

    return E_NOTFOUND;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getCCs
 *=------------------------------------------------------------------------=
 * returns a list of people to whom this message was/is being blind carbon
 * copied.  This is a linked list class.  The object should be passed by 
 * reference.
 *
 * Parameters:
 *      LinkList<const char *> &  - [in] a reference to an object in which to
 *                                       put all the data.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getCCs
(
    LinkList<const char *> &ccs
)
{
    /**
     * use the copy operator on the linklist class to copy over this list.
     */
    ccs = this->ccs;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::getExtraHeaders
 *=------------------------------------------------------------------------=
 * returns any headers we aren't using associated with this message.
 *
 * Parameters:
 *      LinkList<const char *> &     - [in] where to put them.  CALLER must
 *                                      NOT FREE any of these strings.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getExtraHeaders
(
    LinkList<const char *> &hdrs
)
{
    hdrs = this->extraHeaders;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getMessageID
 *=------------------------------------------------------------------------=
 * returns the message ID for this message.  This is only used in replying
 * when one needs to indicate "In-Reply-To:" in the header.
 *
 * Parameters:
 *      const char **       - [out] the messageid.  CALLER frees.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getMessageID
(
    const char **messageid
)
{
    if (!messageid) return E_PARAM;

    if (this->messageID) {
        *messageid = copyString(this->messageID);
        return *messageid ? S_OK : E_OUTOFMEMORY;
    }

    *messageid = NULL;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getRealSender
 *=------------------------------------------------------------------------=
 * returns the REAL sender of this message, in cases where a secretary or
 * some such person has sent a message on behalf of somebody else.
 *
 * Parameters:
 *  const char **           - [out] the real sender's path.  CALLER FREEs
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getRealSender
(
    const char **prealsender
)
{
    if (!prealsender) return E_PARAM;
    *prealsender = NULL;

    if (this->realSender) {
        *prealsender = copyString(this->realSender);
        if (!(*prealsender)) return E_OUTOFMEMORY;
    }
    
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getReceived
 *=------------------------------------------------------------------------=
 * returns the date this message was received, in convenient text format
 * for those who wish to print it out.
 *
 * Parameters:
 *      const char **               - [out] the date in string format.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getReceived
(
    const char **pdate
)
{
    struct tm *tt;
    char buf[128];
    int i;

    if (!pdate) return E_PARAM;

    tt = localtime(&received);

    i = strftime(buf, sizeof(buf), "%Ex ", tt);
    strftime(buf + i, sizeof(buf) - i, "%R", tt);

    *pdate = copyString(buf);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getReceived
 *=------------------------------------------------------------------------=
 * returns the time at which this message was received, in time_t format.  If
 * this hasn't been set yet, (time_t)-1 is returned.
 *
 * Parameters:
 *      time_t *                - [out] the time at which this msg was rcvd
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getReceived
(
    time_t *t
)
{
    if (!t) return E_PARAM;
    *t = received;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::getRecipients
 *=------------------------------------------------------------------------=
 * returns a list of people to whom this message was/is being sent.  This
 * is a linked list class.  The object should be passed by reference.
 *
 * Parameters:
 *      LinkList<const char *> &      - [in] a reference to an object in which to
 *                                  put all the data.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getRecipients
(
    LinkList<const char *> &recips
)
{
    /**
     * use the copy operator on the linklist class to copy over this list.
     */
    recips = this->recipients;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getReplyPath
 *=------------------------------------------------------------------------=
 * returns the people to whom replies should be directed!!
 *
 * Parameters:
 *      LinkList<const char *>      - [in]  where to put return addresses
 *                                      NOTE THAT THESE SHOULD NOT BE FREED!
 *      bool                        - [in] optional -- should we return just
 *                                      the base address, or ALL, in case of
 *                                      a reply-all situation?
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getReplyPath
(
    LinkList<const char *> &list,
    bool all
)

{
    if (all) {
        list = this->senders;
        if (this->replyTo) list.addElement(this->replyTo);
    } else {
        if (this->replyTo) {
            list.addElement(this->replyTo);
        } else {
            list = this->senders;
        }
    }

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::getReplyTo
 *=------------------------------------------------------------------------=
 * returns the value of the Replyto header.
 *
 * Parameters:
 *      const char **       - [out]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getReplyTo
(
    const char **path
)
{
    if (!path) return E_PARAM;

    if (this->replyTo) {
        *path = copyString(this->replyTo);
        return *path ? S_OK : E_OUTOFMEMORY;
    } 

    *path = NULL;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::getReturnPath
 *=------------------------------------------------------------------------=
 * returns the "return-path" header for this message.
 *
 * Parameters:
 *      const char **      - [out] localAlloc() email address.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getReturnPath
(
    const char **path
)
{
    if (!path) return E_PARAM;

    if (this->returnPath) {
        *path = copyString(this->returnPath);
        return *path ? S_OK : E_OUTOFMEMORY;
    } 

    *path = NULL;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getSenders
 *=------------------------------------------------------------------------=
 * returns the senders of this message.  Note that this is an allocated 
 * string and the caller is responsible for freeing it.
 *
 * Parameters:
 *  LinkList<const char *> &        - [in]  where to put strings.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getSenders
(
    LinkList<const char *> &list
)
{
    list = this->senders;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getSent
 *=------------------------------------------------------------------------=
 * returns the date this message was sent, in convenient text format
 * for those who wish to print it out.  We do our darndest to be sensitive
 * to locale here ...
 *
 * Parameters:
 *      const char **               - [out] the date in string format.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getSent
(
    const char **pdate
)
{
    struct tm *tt;
    char buf[128];
    int i;

    if (!pdate) return E_PARAM;

    tt = localtime(&sent);
    i = strftime(buf, sizeof(buf), "%Ex ", tt);
    strftime(buf + i, sizeof(buf) - i, "%R", tt);

    *pdate = copyString(buf);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getSent
 *=------------------------------------------------------------------------=
 * returns the time at which this message was sent, in time_t format.  If
 * this hasn't been set yet, (time_t)-1 is returned.
 *
 * Parameters:
 *      time_t *                - [out] the time at which this msg was sent.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::getSent
(
    time_t *t
)
{
    if (!t) return E_PARAM;
    *t = sent;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::getSubject
 *=------------------------------------------------------------------------=
 * returns the current subject for this message.
 *
 * Parameters:
 *      char **             - [out] the subject.
 *
 * Output:
 *      ERRCODE
 */  
ERRCODE MailMessage::getSubject
(
    const char **psubject
)
{
    if (!psubject) return E_PARAM;

    if (this->subject) {
        *psubject = copyString(this->subject);
        if (!(*psubject)) return E_OUTOFMEMORY;
    } else {
        *psubject = NULL;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::release
 *=------------------------------------------------------------------------=
 * release a ref and delete if appropriate.
 */
int MailMessage::release()
{
    int c = --m_cref;

    if (c <= 0) {
        delete this;
    }
    return c;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::saveAttachment
 *=------------------------------------------------------------------------=
 * saves the given attachment to disk in the given directory
 *
 * Parameters:
 *      Attachment *                    - [in]  which one to save.
 *      const char *                    - [in]  WHERE to save it.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::saveAttachment
(
    Attachment *attach,
    const char *dir
)
{
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setCCs
 *=------------------------------------------------------------------------=
 * sets the list of carbon copies for this message.
 *
 * Parameters:
 *      LinkList<const char *> &      - [in]  the list of carbon copies.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setCCs
(
    LinkList<const char *> &list
)
{
    nukeStringList(this->ccs);
    ccs = list;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setBCCs
 *=------------------------------------------------------------------------=
 * sets the list of blind carbon copies for this message.
 *
 * Parameters:
 *      LinkList<const char *> &      - [in]  the list of blind carbon copies.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setBCCs
(
    LinkList<const char *> &list
)
{
    nukeStringList(this->bccs);
    bccs = list;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::setBodyByType
 *=------------------------------------------------------------------------=
 * sets the body for this message for the given MIME Type.  We will make a 
 * copy of the body and save it in our class.
 *
 * Parameters:
 *      MIMETYPE                - [in]  the MIME type for this body.
 *      const char *            - [in]  message body
 *      const char *            - [in]  charset for this section
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setBodyByType
(
    MIMETYPE type,
    const char *newBody,
    const char *charset
)
{
    InternalMimeContent *body;
    const char *saveMe;
    ERRCODE er;

    /**
     * Go and look and see if we've already got a body for this type.
     */
    er = mimeBodies.firstElement(&body);
    while (er == S_OK) {
        if (body->type == type) break;
        er = mimeBodies.nextElement(&body);
    }

    /**
     * If not, go and allocate new data for this type.
     */
    if (er == S_NODATA) {
        body = (InternalMimeContent *)localAlloc(sizeof(InternalMimeContent));
        memset(body, 0, sizeof(InternalMimeContent));
        body->type = type;
    }

    /**
     * Copy over the data.
     */
    saveMe = body->body;
    body->body = newBody ? copyString(newBody) : NULL;
    if (newBody && !body->body) {
        if (er == S_NODATA) 
            localFree(body);
        else
            body->body = saveMe;
        return E_OUTOFMEMORY;
    } else {
        if (saveMe) localFree(saveMe);
    }

    saveMe = body->charset;
    body->charset = charset ? copyString(charset) : NULL;
    if (charset && !body->charset) {
        if (er == S_NODATA)
            localFree(body);
        else
            body->body = saveMe;
        return E_OUTOFMEMORY;
    } else {
        if (saveMe) localFree(saveMe);
    }

    /**
     * Put it in the list if appropriate.
     */
    if (er == S_NODATA) mimeBodies.addElement(body);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setBodyRefByType
 *=------------------------------------------------------------------------=
 * sets the body for this message for the given MIME Type.  WE KEEP THE 
 * POINTER WE'RE GIVEN AND ASSUME OWNERSHIP OF ITS LIFETIME.
 *
 * Parameters:
 *      MIMETYPE                - [in]  the MIME type for this body.
 *      const char *            - [in]  message body
 *      const char *            - [in]  charset for this section
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setBodyRefByType
(
    MIMETYPE type,
    const char *newBody,
    const char *charset
)
{
    InternalMimeContent *body;
    const char *saveMe;
    ERRCODE er;

    /**
     * Go and look and see if we've already got a body for this type.
     */
    er = mimeBodies.firstElement(&body);
    while (er == S_OK) {
        if (body->type == type) break;
        er = mimeBodies.nextElement(&body);
    }

    /**
     * If not, go and allocate new data for this type.
     */
    if (er == S_NODATA) {
        body = (InternalMimeContent *)localAlloc(sizeof(InternalMimeContent));
        memset(body, 0, sizeof(InternalMimeContent));
        body->type = type;
    }

    /**
     * Copy over the data.
     */
    saveMe = body->body;
    body->body = newBody;
    if (saveMe) localFree(saveMe);

    saveMe = body->charset;
    body->charset = charset;
    if (saveMe) localFree(saveMe);

    /**
     * Put it in the list if appropriate.
     */
    if (er == S_NODATA) mimeBodies.addElement(body);
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::setExtraHeaders
 *=------------------------------------------------------------------------=
 * sets the headers that we're not using for this message.  We assume life
 * time control over the strings in the list!!!
 *
 * Parameters:
 *      LinkList<const char *> &    - [in]  list we will take ownership of.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setExtraHeaders
(
    LinkList<const char *> &hdrs
)
{
    this->extraHeaders = hdrs;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setRealSender
 *=------------------------------------------------------------------------=
 * Sets the REAL sender of this message, as in cases where a secretary or
 * such person has sent a message on somebody else's behalf
 *
 * Parameters:
 *      const char *        - [in]  the new real sender.  we copy it.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setRealSender
(
    const char *realsender
)
{
    if (this->realSender) localFree(this->realSender);
    this->realSender = NULL;

    if (realsender) {
        this->realSender = copyString(realsender);
        if (!this->realSender) return E_OUTOFMEMORY;
    }
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setReceived
 *=------------------------------------------------------------------------=
 * sets the date/time at which this message was received
 *
 * Parameters:
 *      time_t                  - [in]  time at which this message was 
 *                                  received
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setReceived
(
    time_t rcvd
)
{
    this->received = rcvd;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::setRecipients
 *=------------------------------------------------------------------------=
 * Sets the recipients for this message.  Any previous recipients are
 * overwritten
 *
 * Parameters:
 *      LinkList<const char *> &      - [in]  where to get the recipients.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setRecipients
(
    LinkList<const char *> &recips
)
{
    nukeStringList(this->recipients);

    /**
     * Use the copy operator on the LinkList class to do this for us.
     */
    this->recipients = recips;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setReplyTo
 *=------------------------------------------------------------------------=
 * Sets the reply to for this message.
 *
 * Parameters:
 *      const char *        - [in]  the new reply path.  we copy it.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setReplyTo
(
    const char *replyTo
)
{
    const char *pstr;

    if (replyTo) {
        pstr = copyString(replyTo);
        if (!pstr) return E_OUTOFMEMORY;
        if (this->replyTo) localFree(this->replyTo);
        this->replyTo = pstr;
    } else {
        if (this->replyTo) localFree(this->replyTo);
        this->replyTo = NULL;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setReturnPath
 *=------------------------------------------------------------------------=
 * sets the return path for this message.
 * 
 * Parameters:
 *      const char *                - [in]  Return-Path:
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setReturnPath
(
    const char *returnPath
)
{
    const char *pstr;

    if (returnPath) {
        pstr = copyString(returnPath);
        if (!pstr) return E_OUTOFMEMORY;
        if (this->returnPath) localFree(this->returnPath);
        this->returnPath = pstr;
    } else {
        if (this->returnPath) localFree(this->returnPath);
        this->returnPath = NULL;
    }

    return S_OK;
}

  

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setSenders
 *=------------------------------------------------------------------------=
 * Sets the senders of this message.  We make a copy of the incoming string
 * so we are not responsible for lifetime issues on it.
 *
 * Parameters:
 *      LinkList<const char *> &        - [in]  senders of this message.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setSenders
(
    LinkList<const char *> &list
)
{
    this->senders.emptyList();
    this->senders = list;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * MailMessage::setSent
 *=------------------------------------------------------------------------=
 * sets the date/time at which this message was sent.
 *
 * Parameters:
 *      time_t                  - [in]  time at which this message was sent.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setSent
(
    time_t sent
)
{
    this->sent = sent;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::setSubject
 *=------------------------------------------------------------------------=
 * Sets the subject for this message.
 *
 * Parameters:
 *      const char *            - [in]  the new subject.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::setSubject
(
    const char *newSubject
)
{
    if (!newSubject) {
        localFree(this->subject);
        this->subject = NULL;
    } else {
        if (this->subject) localFree(this->subject);
        this->subject = copyString(newSubject);
        if (!this->subject) return E_OUTOFMEMORY;
    }

    return S_OK;
}


