/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _MAILMESSAGE_H_

#include <llist.h>
#include <sys/time.h>


#include "mailaccount.h"
#include "mime.h"

/**
 * forward decls
 */
class  MailFolder;
struct MimeContentInfo;
class  IReader;
struct MimeBody;

/**
 * Max length of an email address
 */
#define MAX_ADDRESSLEN 1024


/** 
 * forward decls
 */
class FolderDB;
class Attachment;

enum ReadStatus {

    MSG_UNOPENED = 0,
    MSG_READ,
    MSG_REPLIED,
    MSG_FORWARDED,
    MSG_EDITING,
    MSG_UNSENT,
    MSG_SENDING
};


/**
 * This class defines an RFC 822 message, with MIME extensions supported.
 */
class MailMessage {

    friend class FolderDB;

    DEBUG_TEST_DECL;

  public:

    /** 
     * creates a new message object.  Use this instead of a public constructor
     * (which doesn't exist for this class).
     */
    static ERRCODE createNew(MailFolder *host, MailMessage **pnewmsg);

    /**
     * Given a stream of text coming in off the network, create the message 
     * and parse out all the data we can get from it and return the 
     * corresponding message.
     * Implementation is in mailmessagenet.cc, since it's so comprehensive.
     */
    static ERRCODE createFromNetwork(IReader *, MailMessage **pnewmsg);

    /**
     * Adds an attachment to this message.
     */
    ERRCODE addAttachment(Attachment *a);

    /**
     * Returns a list of attachments.
     */
    ERRCODE getAttachments(LinkList<Attachment *> &list);

    /**
     * Returns the list of BCCs in this message.
     */
    ERRCODE getBCCs(LinkList<const char *> &recips);

    /**
     * Returns the body of this message in the given format, or an error
     * if the given format isn't available.  Also returns the charset.
     */
    ERRCODE getBodyByType(MIMETYPE, const char **, const char **charset = NULL);

    /**
     * returns a count of the types of bodies available in this message.
     */
    ERRCODE getBodyFormats(MIMETYPE **, int *);

    /**
    * This simply returns a pointer to our version of the BODY for the
    * given type.  THIS IS VERY DANGEROUS !!!!  do not use this for more
    * than the length of a function call.  Also returns a ref to 
     */
    ERRCODE getBodyRefByType(MIMETYPE, const char **, const char ** charset = NULL);

    /**
     * Returns the list of CCs in this message.
     */
    ERRCODE getCCs(LinkList<const char *> &recips);

    /**
     * Returns any extra headers that we aren't using associated with this
     * message.  Do NOT free the strings that we return in this list.
     */
    ERRCODE getExtraHeaders(LinkList<const char *> &headers);

    /**
     * indicates whether or not this message has attachments.  S_OK if yes,
     * S_FALSE if no.
     */
    ERRCODE getHasAttachments(bool *p) { if (!p) return E_PARAM; *p = hasAttachments; return S_OK; }

    /**
     * Returns the message ID for this message.
     */
    ERRCODE getMessageID(const char **);

    /**
     * Returns the priority of this message.
     */
    ERRCODE getPriority( MessagePriority &prior) {
        prior = this->priority;
        return S_OK;
    }

    /**
     * returns the REAL person who sent this message, as in cases where a 
     * secretary has sent the message.
     */
    ERRCODE getRealSender(const char **);

    /**
     * returns the date this message was received, in text format, overloaded
     * from the date version.
     */
    ERRCODE getReceived(const char **);

    /**
     * Returns the date this message was recieved, in time_t format.
     */
    ERRCODE getReceived(time_t *);

    /**
     * Returns a list of all people to whom this message was/is being sent.
     */
    ERRCODE getRecipients(LinkList<const char *> &recips);

    /**
     * Returns the address(es) to which replies should be directed.
     */
    ERRCODE getReplyPath(LinkList<const char *> &list, bool all = false);

    /**
     * Returns the value of the Reply-To: field.
     */
    ERRCODE getReplyTo(const char **replyTo);

    /**
     * Returns the return path for this document.
     */
    ERRCODE getReturnPath(const char ** returnPath);

    /**
     * Gets the sender of this message. CALLER frees allocated string.
     */
    ERRCODE getSenders(LinkList<const char *> &senders);

    /**
     * Returns the date/time this message was sent, in text format.  Overloaded
     * from the date version.
     */
    ERRCODE getSent(const char **);

    /**
     * Returns the date/time this message was sent, in time_t format.
     */
    ERRCODE getSent(time_t *);

    /**
     * returns the size of this message in bytes.
     */
    ERRCODE getSize(int *p) {if (!p) return E_PARAM; *p = bodySize; return S_OK; }

    /**
     * returns the current status of the message -- whether it's been read,
     * replied to, etc ..
     */
    ERRCODE getStatus( ReadStatus &stat) {
        stat = this->status;
        return S_OK;
    }

    /**
     * Gets the subject associated with this message. CALLER frees allocated
     * string.
     */
    ERRCODE getSubject(const char **subject);

    /**
     * Saves the attachment with the given name to disk in the given directory
     */
    ERRCODE saveAttachment(Attachment *attach, const char *dir);

    /**
     * Sets the body for this message to be the given data.  Please note that
     * this is the text body for the message.  MIME data such as added
     * files and the like should be set with addAttachment.  This text body 
     * can include RTF or HTML tags.
     */
    ERRCODE setBodyByType(MIMETYPE, const char *body, const char *charset);

    /**
     * Sets the list of blind carbon copies
     */
    ERRCODE setBCCs(LinkList<const char *> &bccs);

    /**
     * Sets the list of carbon copies.
     */
    ERRCODE setCCs(LinkList<const char *> &ccs);

    /**
     * Sets the extra headers that we aren't using for this message.
     */
    ERRCODE setExtraHeaders(LinkList<const char *> &headers);

    /**
     * Sets a flag noting that this message has attachments.  We need this 
     * because sometimes we won't have a body set yet, but we would like
     * to know that we really do have these or not.
     */
    ERRCODE setHasAttachments(bool f) { this->hasAttachments = f; return S_OK; }

    /**
     * Sets the priority of the message.
     */
    ERRCODE setPriority(MessagePriority pri) { priority = pri; return S_OK; }

    /**
     * Sets the REAL sender of this message, as in cases where a secretary
     * has sent a message on somebody's behalf, etc ...
     */
    ERRCODE setRealSender(const char *);

    /**
     * Sets the reply path for this message.
     */
    ERRCODE setReplyTo(const char *);

    /**
     * Sets the return path for this message.
     */
    ERRCODE setReturnPath(const char *);

    /**
     * Sets the date/time at which this message was received.
     */
    ERRCODE setReceived(time_t received);

    /**
     * Sets the list of recipients.
     */
    ERRCODE setRecipients(LinkList<const char *> &recips);

    /**
     * Sets the senders for this message.
     */
    ERRCODE setSenders(LinkList<const char *> &list);

    /**
     * Sets the date/time at which this message was sent.
     */
    ERRCODE setSent(time_t);

    /**
     * Sets the size of the message's body.
     */
    ERRCODE setSize(int s) { bodySize = s; return S_OK; }

    /**
     * sets the status of this message.
     */
    ERRCODE setStatus(ReadStatus rs) { this->status = rs; return S_OK; }

    /**
     * Sets this message's subject.
     */
    ERRCODE setSubject(const char *subject);

    /**
     * Returns our parent folder.
     */
    inline MailFolder *getParent() { return this->hostFolder; }

    /**
     * Resets the timeval for this message so that next time we're written
     * out, we have to regenerate it.
     */
    inline void resetTimeVal() { memset(&this->tv, 0, sizeof(struct timeval)); }

    /**
     * Lifetime control -- we pass these around to various objects all the 
     * time, and they can be potentially big, so we'll refcount them instead
     * of copy htem all the time.
     */
    int addRef();
    int release();

  private:
    MailMessage();
    ~MailMessage();

    /**
     * A very simple struct for supporting multi-typed MIME bodies
     * (such as text/plain, text/rich, text/html, etc ....)  Our class
     * maintains a list of these so we can support eventually viewing of
     * all the different types.
     */
    struct InternalMimeContent {

        MIMETYPE type;
        const char *body;
        const char *charset;
    };

    /**
     * random private routines.
     */
    ERRCODE ensureBodyLoaded();
    ERRCODE setBodyRefByType(MIMETYPE, const char *body, const char *charset);

    /**
     * Routines for generating message from SMTP mail.
     */
    static ERRCODE processHeader(MailMessage *, MimeContentInfo *);
    static ERRCODE processBody(MimeContentInfo *,
                               const char *, LinkList<MimeBody *> &);
    static ERRCODE processMimeBody(MimeContentInfo *,
                                   const char *, LinkList<MimeBody *> &);

    /**
     * Data member variables
     */
    LinkList<InternalMimeContent *> mimeBodies;
    LinkList<const char *> extraHeaders;
    LinkList<const char *> recipients;
    LinkList<const char *> senders;
    LinkList<const char *> bccs;
    LinkList<const char *> ccs;
    MessagePriority priority;
    const char *realSender;
    const char *returnPath;
    const char *replyTo;
    const char *messageID;
    bool hasAttachments;
    const char *subject;
    ReadStatus status;
    MIMETYPE mimeType;
    time_t received;
    int bodySize;
    time_t sent;
    int m_cref;

    /**
     * Attachments!!!
     */
    LinkList<Attachment *> attachments;

    /** 
     * These help us figure out where we are.
     */
    MailFolder *hostFolder;
    struct timeval tv;
};


#define _MAILMESSAGE_H_
#endif // _MAILMESSAGE_H_


