/**
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

#include "attachment.h"
#include "deencode.h"
#include "intl.h"
#include "localized.h"
#include "mailmessage.h"
#include "main.h"
#include "mime.h"
#include "netapis.h"
#include "userpref.h"

/**
 * This is the set of methods to load in one of these objects from a stream 
 * of data from a server.  It knows how to parse out all the appropriate
 * information.
 */

//
//
//      CRAP CRAP CRAP CRAP CRAP.
// UNDONE: marcw, 12.12.2000 -- ARGH!!! I need message/rfc822 MIME support. 
//      right now, they are treated as a plain text attachment.  
//      CRAP CRAP CRAP CRAP CRAP.
//
//
//

/**
 * CONSIDER: marcw, 1.22.2001 -- should this be TLS??
 */
static char localNetworkBuf[NET_MAXLINELEN];

/**
 *=------------------------------------------------------------------------=
 *                      Local Prototypes
 *=------------------------------------------------------------------------=
 */
const char *stripNewLines(const char *);



/**
 *=------------------------------------------------------------------------=
 * MailMessage::createFromNetwork
 *=------------------------------------------------------------------------=
 * given a stream of data coming in off the network, goes and figures out all
 * the information that is needed to make a mail message object and returns
 * said object.  If there are any errors, returns said error and NO mail
 * message.
 *
 * Parameters:
 *      IReader *                   - [in]  where to read data from.
 *      MailMessage **              - [out] new mail message.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::createFromNetwork
(
    IReader *reader,
    MailMessage **message
)
{
    LinkList<MimeBody *> components;
    MimeContentInfo info;
    ERRCODE er, er2;
    MimeBody *body;
    int cread;

    if (!reader || !message) return E_PARAM;
    
    mimeReader.setReader(reader);

    /**
     * Set up the default information for this message.  These can be set to
     * valid values (including invalid markers) in the header parsing process.
     */
    memset(&info, 0, sizeof(info));
    info.type.setType(Text, Plain);
    info.encoding = SevenBit;
    info.disposition = InvalidDisposition;
    info.separator = copyString(".\r\n");

    *message = new MailMessage();
    if (!(*message)) return E_OUTOFMEMORY;

    /**
     * 1. Read in the header.
     */
    er = processHeader(*message, &info);
    GOTO_ON_FAILURE(er, Error);

    /** 
     * 2. Read in the body !!!
     */
    er = processBody(&info, info.separator, components);
    GOTO_ON_FAILURE(er, Error);

    /**
     * We need to read until we get to the end of the message -- 
     * otherwise the input stream is blocked still getting content
     * from this message.  That's bad.  This is guaranteed to be done
     * if we haven't read a multi-part message (as that's the only way
     * they have of determining the end of the section!!!)
     */
    if (info.separator && strncasecmp(info.separator, ".\r\n", 3)) {
        do {
            /**
             * This should return '\0' if there is no more data to read !!
             */
            mimeReader.readLine(localNetworkBuf, sizeof(localNetworkBuf), cread);
        } while (!(localNetworkBuf[0] == '.' 
                   && (localNetworkBuf[1] == '\r' || localNetworkBuf[1] == '\n'))
                 && localNetworkBuf[0] != '\0');
    }

    /**
     * 3. okay, now go and add the parts to the message.
     */
    er = components.firstElement(&body);
    while (er == S_OK) {
        if (body->isBody) {
            er = (*message)->setBodyRefByType(body->bodyType, body->body, body->charset);
            if (SUCCEEDED(er)) {
                body->body = NULL;
                body->charset = NULL;
            }
        } else {
            er = (*message)->addAttachment(body->attachment);
        }
        GOTO_ON_FAILURE(er, Error);
        er = components.nextElement(&body);
    }

    er = S_OK;

  Error:
    /**
     * Clean up the list of section structures we've put together.  Cleaning
     * up the right stuff is a bit wacky, but not too bad.
     */
    er2 = components.firstElement(&body);
    while (er2 == S_OK) {
        if (body->attachment) body->attachment->release();
        if (body->charset) localFree(body->charset);
        if (body->body) localFree(body->body);
        components.removeElement(body);
        localFree(body);
        er2 = components.firstElement(&body);
    }

    /**
     * Clean up the content info
     */
    if (info.separator) localFree(info.separator);
    if (info.charset) localFree(info.charset);
    if (info.name) localFree(info.name);

    /**
     * And if we failed, we aren't creating any messages ...
     */
    if (FAILED(er)) {
        (*message)->release();
        (*message) = NULL;
    }
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::processBody
 *=------------------------------------------------------------------------=
 * processes the body of the message.  We need to know if we're a multi-part
 * message, and if so, what our separator is.
 *
 * Parameters:
 *      MimeContentInfo *       - [in]  MIME Information about this message.
 *      const char *            - [in] the terminating separator.
 *      LinkList<MimeBody *> &  - [in]  where to add everything we find.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::processBody
(
    MimeContentInfo *info,
    const char *terminatingSeparator,
    LinkList<MimeBody *> &components
)
{
    int cread;

    /**
     * okay, this is a bit wacky, but our entire body parsing scheme always
     * assumes we're at the beginning of a section.  Therefore, if we're a 
     * multipart message, we first need to get to the first separator,
     * and then begin parsing.
     */
    if (info->type.getMediaType() == Multipart) {

        switch (info->type.getSubType()) {
            default:
                fprintf(stderr, "Unexpected Multipart sub-type !!! -- treating as 'Mixed'\n");
            case Signed:
                info->type.setSubType(Mixed);
                // fall through
            case Mixed:
            case Alternative:
            case Related:
                while (1) {
                    mimeReader.readLine(localNetworkBuf, sizeof(localNetworkBuf), cread);
                    if (localNetworkBuf[0] == '\0') { return E_NET_UNEXPECTED; }
                    if (localNetworkBuf[0] == '.') { return E_NET_UNEXPECTED; }
                    if (!strncasecmp(localNetworkBuf, info->separator, strlen(info->separator)))
                        break;
                }
                break;
        }
    }

    /** 
     * okay, now go and parse the rest.
     */
    return processMimeBody(info, terminatingSeparator, components);
}

    
/**
 *=------------------------------------------------------------------------=
 * MailMessage::processHeader
 *=------------------------------------------------------------------------=
 * processes the incoming SMTP header.  This is actually quite an involved
 * process ...
 *
 * Parameters:
 *      MailMessage *           - [in]  message we're building.
 *      MimeContentInfo *       - [in]  where to put the MIME information.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::processHeader
(
    MailMessage *message,
    MimeContentInfo *info
)
{
    bool haveReceived = false;
    SMTPHeader header;
    const char *pstr;
    const char *parse;
    ERRCODE er;

    /**
     * OKay, this is pretty easy -- figure out what the next header is and 
     * process it appropriately.  Once we reach the end of the header, then
     * we're done.
     */
    er = mimeReader.firstHeader(header, pstr);
    while (SUCCEEDED(er)) {
        switch (header) {
            /**
             * We don't know what it is -- but we'll keep it just in case the
             * user wants to look at it.
             */
            case UnknownHeader:
                message->extraHeaders.addElement(pstr);
                break;

            /**
             * We be done!  We be done!!!
             */
            case EndOfHeader:
                if (info->disposition == InvalidDisposition)
                    info->disposition = Inline;
                return S_OK;

            case From:
                er = parseAddressList(pstr, message->senders);
                break;

            case To:
                er = parseAddressList(pstr, message->recipients);
                break;

            case CC:
                er = parseAddressList(pstr, message->ccs);
                break;

            case MessageID:
                if (message->messageID) localFree(message->messageID);
                message->messageID = stripNewLines(pstr);
                break;

            case Sender:
                er = parseSingleAddress(pstr, &parse);
                message->setRealSender(parse);
                localFree(parse);
                break;

            case Received:
                if (!haveReceived) {
                    er = parseReceivedLine(pstr, &message->received);
                    if (SUCCEEDED(er)) haveReceived = true;
                }
                break;

            case Date:
                er = parseDate(pstr, &message->sent);
                break;

            case ReplyTo:
                er = parseSingleAddress(pstr, &parse);
                message->setReplyTo(parse);
                localFree(parse);
                break;

            case ReturnPath:
                er = parseSingleAddress(pstr, &parse);
                message->setReturnPath(parse);
                localFree(parse);
                break;

            case XMSMailPriority:
                er = parsePriority(pstr, &message->priority);
                break;

            case ContentType:
                er = parseContentType(pstr, info);
                break;

            case ContentDisposition:
                er = parseContentDisposition(pstr, info);
                break;

            case ContentTransferEncoding:
                er = parseContentEncoding(pstr, info);
                break;

            case Subject:
                message->subject = stripNewLines(pstr);
                break;

            default:
                K_FAIL("Bogus header returned from nextHeader.  Very bad !!!\n");
                return E_INTERNAL;
        }

        if (header != UnknownHeader && pstr) localFree(pstr);
        er = mimeReader.nextHeader(header, pstr);
    }

    /**
     * Basically, we are done!  First make sure that some sort of disposition
     * has been declared so that we give the message a body.
     */
    if (info->disposition == InvalidDisposition) info->disposition = Inline;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailMessage::processMimeBody
 *=------------------------------------------------------------------------=
 * processes the body of the message.  We need to know if we're a multi-part
 * message.
 *
 * Parameters:
 *      MimeContentInfo *       - [in]  MIME Information about this message.
 *      const char *            - [in]  terminator for this MIME section.
 *                                      [NULL implies EOM is the terminator]
 *      LinkList<MimeBody *> &  - [in]  where to add everything we find.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailMessage::processMimeBody
(
    MimeContentInfo *info,
    const char *terminatingSeparator,
    LinkList<MimeBody *> &components
)
{
    MimeBody *mimeBody = NULL;
    MimeContentInfo localInfo;
    const char *value = NULL;
    const char *body = NULL;
    const char *newsep;
    SMTPHeader header;
    int size, cread;
    ERRCODE er;

    memset(&localInfo, 0, sizeof(localInfo));

    /**
     * Depending on what type of body we've got, deal with it.
     */
    switch (info->type.getMediaType()) {

        /**
         * Download and decode the given section, and then figure out if it's
         * a body or an attachment.
         */
        case Application:
        case Video:
            info->disposition = MailAttachment;
            // fall through
        case Image:
        case Message:
        case UnknownMediaType:
        case Text:
            /**
             * Decode the current section given the encoding from the MIME
             * header, and just tell it to search until the end of the message
             * since this isn't a boundary-separated section.
             */
            er = decodeSection(info->encoding, terminatingSeparator, &size, &body);
            RETURN_ON_FAILURE(er);

            mimeBody = (MimeBody *)localAlloc(sizeof(MimeBody));
            if (!mimeBody) {
                localFree(body);
                return E_OUTOFMEMORY;
            }
            memset(mimeBody, 0, sizeof(MimeBody));
            mimeBody->bodyType = info->type;

            /**
             * figure out if it's a body or an attachment
             */
            if (info->disposition == Inline) {
                mimeBody->body = body;
                mimeBody->isBody = true;
                mimeBody->charset = info->charset;
                info->charset = NULL;
            } else {
                if (!info->name) {
                    info->name = intlLoadString(ID_GENERIC_ATTACHMENT);
                    if (!info->name) { er = E_OUTOFMEMORY; goto Error; }
                }
                mimeBody->attachment = new Attachment(info->name, (BYTE *)body,
                                                      size, info->encoding, info->type);
                if (!mimeBody->attachment) {
                    er = E_OUTOFMEMORY;
                    localFree(body);
                    goto Error;
                }
            }

            er = components.addElement(mimeBody);
            GOTO_ON_FAILURE(er, Error);
            break;


        /**
         * these are multi-part types, and require parsing more MIME headers.
         */
        case Multipart:
            K_ASSERT(terminatingSeparator, "multipart message with no separator\n");

            /**
             * Loop through the sections in the message.  There are a bunch
             * of different ways that this will stop.
             */
            while (1) {

                /**
                 * For each section, process a header.
                 */
                memset(&localInfo, 0, sizeof(localInfo));
                localInfo.type.setType(Text, Plain);
                localInfo.encoding = SevenBit;
                localInfo.disposition = InvalidDisposition;

                er = mimeReader.firstHeader(header, value);
                while (1) {
                    bool exitloop = false;
                    GOTO_ON_FAILURE(er, Error);

                    switch (header) {
                        case EndOfHeader:
                            exitloop = true;
                            break;
                        case ContentType:
                            er = parseContentType(value, &localInfo);
                            break;
                        case ContentTransferEncoding:
                            er = parseContentEncoding(value, &localInfo);
                            break;
                        case ContentDisposition:
                            er = parseContentDisposition(value, &localInfo);
                            break;
                        case UnknownHeader:
                            break;
                        default:
                            fprintf(stderr, "Unexpected MIME header with ignored value: %d\n", header);
                            break;
                    }

                    GOTO_ON_FAILURE(er, Error);
                    if (exitloop) break;
                    if (value) localFree(value);
                    value = NULL;
                    er = mimeReader.nextHeader(header, value);
                }

                /**
                 * Figure out what separator we're looking for next.
                 */
                if (localInfo.type.getMediaType() == Multipart) {
                    newsep = localInfo.separator;
                } else {
                    newsep = terminatingSeparator;
                }

                /**
                 * Make sure it has at least SOME disposition.  The only
                 * case you should see this is the section has no contenttype,
                 * content-encoding, or content-disposition headers.  Not
                 * allowed, but we'll protect against it just in case.
                 */
                if (localInfo.disposition == InvalidDisposition)
                    localInfo.disposition = Inline;

                /** 
                 * go and process the body, recursively.
                 */
                er = processBody(&localInfo, newsep, components);
                GOTO_ON_FAILURE(er, Error);

                /**
                 * Okay, we need to scoot ahead until we find a separator.
                 * if the next separator indicates the end of the multipart
                 * section, then exit.
                 */
                size = strlen(info->separator);
                do {
                    const char *p = stripNewLines(mimeReader.currentLine());
                    if (!strncasecmp(p, info->separator, size) &&
                        (p[size] == '-' || isspace(p[size]) || p[size] == '\0')) {

                        /**
                         * Consume this line, and either continue or exit (if
                         * it's a closing boundary marker).
                         */
                        if (p[size] == '-' && p[size + 1] == '-') {
                            mimeReader.readLine(localNetworkBuf, sizeof(localNetworkBuf), cread);
                            localFree(p);
                            er = S_OK;
                            goto Error;
                        }

                        localFree(p);
                        break;
                    } else if (!strncasecmp(p, info->separator, size)) {
                        localFree(p);
                        break;
                    } else {
                        localFree(p);
                    }

                    mimeReader.readLine(localNetworkBuf, sizeof(localNetworkBuf), cread);

                } while (localNetworkBuf[0] != '.' && localNetworkBuf[0] != '\0');

                if (localInfo.name) localFree(localInfo.name);
                if (localInfo.charset) localFree(localInfo.charset);
                if (localInfo.separator) localFree(localInfo.separator);
            }

            break;

        default:
            K_FAIL("There is a MIME media type we pretend to know about but don't\n");
            break;
    }

    er = S_OK;

  Error:
    if (FAILED(er)) {
        if (mimeBody) {
            if (mimeBody->attachment) mimeBody->attachment->release();
            if (mimeBody->charset) localFree(mimeBody->charset);
            if (mimeBody->body) localFree(mimeBody->body);
        }
        if (value) localFree(value);
    }

    if (localInfo.name) localFree(localInfo.name);
    if (localInfo.charset) localFree(localInfo.charset);
    if (localInfo.separator) localFree(localInfo.separator);

    return er;
}





/**
 *=------------------------------------------------------------------------=
 * stripNewLines
 *=------------------------------------------------------------------------=
 * given a string that likely has new line characters at the end, go and
 * return a copy of that string, but without the new lines ...
 *
 * Parameters:
 *      const char *            - [in]  strip me baby
 *
 * Output:
 *      const char *            - NULL ==> OOM
 */
const char *stripNewLines
(
    const char *stripMe
)
{
    char *buf;
    int x, y;

    x = y = 0;

    buf = (char *)localAlloc(strlen(stripMe) + 1);
    if (!buf) return NULL;

    while (stripMe[x] != '\0') {

        if (stripMe[x] != '\r' && stripMe[x] != '\n')
            buf[y++] = stripMe[x];

        x++;
    }

    buf[y] = '\0';
    return buf;
}


