/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <qbuttongroup.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qfiledialog.h>
#include <qgroupbox.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <sys/param.h>


#include "fileutil.h"
#include <intl.h>
#include "localized.h"
#include "mailprops.h"


/**
 *=------------------------------------------------------------------------=
 *                             Local Prototypes
 *=------------------------------------------------------------------------=
 */
void _generateFileName(const char *, const char *, char *, unsigned int);


/**
 *=------------------------------------------------------------------------=
 * MailProperties::MailProperties
 *=------------------------------------------------------------------------=
 * initialize the dialog.  There's actually a lot of stuff that needs to
 * happen here.
 *
 * Parameters:
 *      const char *                - [in]  account name.
 *      const char *                - [in]  account dir
 *      MailAccountInfo &           - [in]  information to populate with.
 *      - rest as per QDialog.
 */
MailProperties::MailProperties
(
    const char *accountName,
    const char *accountDirectory,
    MailAccountInfo &info,
    QWidget* parent,
    const char* name, 
    bool modal, 
    WFlags fl
)
 : MailPropertiesDialog(parent, name, modal, fl)
{
    this->accountName = copyString(accountName);
    this->accountDirectory = copyString(accountDirectory);

    /**
     * General Page.
     */
    accountNameE->setText(accountName);
    userNameE->setText(info.userName);
    userOrganizationE->setText(info.userOrganization);
    userEmailE->setText(info.userEmail);
    signatureFileE->setText(info.signatureFile);

    /**
     * Incoming page.   The combo box is filled up at localization time.
     * The order is very important, of course.
     */
    selectServerType(info.spoolLocation == NULL ? 0 : 1);
    sourceLocationE->setText(info.spoolLocation != NULL 
                                 ? info.spoolLocation
                                 : info.incomingPop);
    popUserNameE->setText(info.popUserName);
    popPasswordE->setText(info.popPassword);
    rememberPopPasswordC->setChecked(info.rememberPopPassword);
    leaveMessagesC->setChecked(!info.popRemoveFromServer);
    dontDuplicateC->setChecked(info.popNoDuplicates);
    thisManyDaysE->setText(QString::number(info.popDeleteAfterDays));
    dontDuplicateClicked();
    leaveMessagesClicked();

    /**
     * Outgoing Page
     */
    smtpServerE->setText(info.outgoingSMTP);
    smtpNeedsAuthC->setChecked(info.smtpNeedsAuth);
    setSmtpAuthToPopAuth(info.usePopForSmtp);
    setSmtpAuthEnabled(info.smtpNeedsAuth);
    smtpUserNameE->setText(info.smtpUserName);
    smtpPasswordE->setText(info.smtpPassword);
    smtpRememberPasswordC->setChecked(info.rememberSmtpPassword);


    /**
     * Advanced page
     */
    smtpPortE->setText(QString::number(info.smtpPort));
    popNormalPortE->setText(QString::number(info.pop3Port));
    popSSLPortE->setText(QString::number(info.securePop3Port));
    setPopSecurity(info.popSecurityType);
    serverTimeoutE->setText(QString::number(info.serverTimeout));
    popApopAuthB->setEnabled(false);

    /**
     * Hook 'em up baby!
     */
    QObject::connect(sourceTypeM, SIGNAL(activated(int)), this, SLOT(selectServerType(int)));
    QObject::connect(signatureFileP, SIGNAL(clicked()), this, SLOT(signatureButtonClicked()));
    QObject::connect(leaveMessagesC, SIGNAL(clicked()), this, SLOT(leaveMessagesClicked()));
    QObject::connect(dontDuplicateC, SIGNAL(clicked()), this, SLOT(dontDuplicateClicked()));
    QObject::connect(smtpNeedsAuthC, SIGNAL(toggled(bool)), this, SLOT(setSmtpAuthEnabled(bool)));
    QObject::connect(usePop3AuthB, SIGNAL(clicked()), this, SLOT(smtpAuthTypeClicked()));
    QObject::connect(useSmtpAuthB, SIGNAL(clicked()), this, SLOT(smtpAuthTypeClicked()));
    QObject::connect(popNormalAuthB, SIGNAL(clicked()), this, SLOT(popSecurityClicked()));
    QObject::connect(popSSLAuthB, SIGNAL(clicked()), this, SLOT(popSecurityClicked()));
    QObject::connect(popApopAuthB, SIGNAL(clicked()), this, SLOT(popSecurityClicked()));
}

/**
 *=------------------------------------------------------------------------=
 * MailProperties::~MailProperties
 *=------------------------------------------------------------------------=
 * Clean up
 */
MailProperties::~MailProperties()
{
    if (this->accountName) localFree(accountName);
    if (this->accountDirectory) localFree(accountDirectory);
}


/**
 *=------------------------------------------------------------------------=
 * MailProperties::dontDuplicateClicked
 *=------------------------------------------------------------------------=
 */
void MailProperties::dontDuplicateClicked()
{
    if (dontDuplicateC->isChecked()) {
        leaveOnL->setEnabled(true);
        thisManyDaysE->setEnabled(true);
        daysL->setEnabled(true);
    } else {
        leaveOnL->setEnabled(false);
        thisManyDaysE->setEnabled(false);
        daysL->setEnabled(false);
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailProperties::getAccountInformation
 *=------------------------------------------------------------------------=
 * returns the information we gathered about the mailaccount.
 *
 * Parameters:
 *      const char * &                  - [out] new account name.
 *      MailAccountInfo &               - [in]  where to deposit the info.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailProperties::getAccountInformation
(
    const char * & newName,
    MailAccountInfo &info
)
{
    PopSecurity val;

    /**
     * Only return a new account name if it's not the same as the existing
     * one.
     */
    if (strcasecmp((const char *)accountNameE->text(), this->accountName)) {
        newName = copyString((const char *)accountNameE->text());
        if (!newName) return E_OUTOFMEMORY;
    }

    /**
     * General props.
     */
    info.userName = copyString((const char *)userNameE->text());
    if (!info.userName) return E_OUTOFMEMORY;

    if (!userOrganizationE->text().isEmpty()) {
        info.userOrganization = copyString((const char *)userOrganizationE->text());
        if (!info.userOrganization) return E_OUTOFMEMORY;
    }

    info.userEmail = copyString((const char *)userEmailE->text());
    if (!info.userEmail) return E_OUTOFMEMORY;
    if (!signatureFileE->text().isEmpty()) {
        info.signatureFile = copyString(signatureFileE->text());
        if (!info.signatureFile) return E_OUTOFMEMORY;
    }

    /**
     * INcoming props
     */
    if (sourceTypeM->currentItem() == 0) {
        info.incomingPop = copyString((const char *)sourceLocationE->text());
        if (!info.incomingPop) return E_OUTOFMEMORY;
    } else {
        info.spoolLocation = copyString((const char *)sourceLocationE->text());
        if (!info.spoolLocation) return E_OUTOFMEMORY;
    }

    info.popUserName = copyString((const char *)popUserNameE->text());
    if (!info.popUserName) return E_OUTOFMEMORY;
    if (!popPasswordE->text().isEmpty()) {
        info.popPassword = copyString((const char *)popPasswordE->text());
        if (!info.popPassword) return E_OUTOFMEMORY;
    }
    info.rememberPopPassword = rememberPopPasswordC->isChecked();
    info.popRemoveFromServer = !leaveMessagesC->isChecked();
    info.popNoDuplicates = dontDuplicateC->isChecked();
    info.popDeleteAfterDays = thisManyDaysE->text().toInt();
    if (popNormalAuthB->isChecked())
        val = PopNormal;
    else if (popSSLAuthB->isChecked())
        val = PopOverSSL;
    else
        val = PopApop;
    info.popSecurityType = val;
    info.pop3Port = popNormalPortE->text().toInt();
    info.securePop3Port = popSSLPortE->text().toInt();

    /**
     * Outgoing props
     */
    info.outgoingSMTP = copyString((const char *)smtpServerE->text());
    if (!info.outgoingSMTP) return E_OUTOFMEMORY;

    info.smtpNeedsAuth = smtpNeedsAuthC->isChecked();
    info.usePopForSmtp = usePop3AuthB->isChecked();
    info.useSecureSmtp = smtpUseSSLC->isChecked();
    info.smtpPort = smtpPortE->text().toInt();
    info.serverTimeout = serverTimeoutE->text().toInt();
    if (!smtpUserNameE->text().isEmpty()) {
        info.smtpUserName = copyString((const char *)smtpUserNameE->text());
        if (!info.smtpUserName) return E_OUTOFMEMORY;
    }
    if (!smtpPasswordE->text().isEmpty()) {
        info.smtpPassword = copyString((const char *)smtpPasswordE->text());
        if (!info.smtpPassword) return E_OUTOFMEMORY;
    }
    info.rememberSmtpPassword = smtpRememberPasswordC->isChecked();

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailProperties::leaveMessagesClicked
 *=------------------------------------------------------------------------=
 * they have chosen to chagne whether messagse are left ont he server.
 */
void MailProperties::leaveMessagesClicked()
{
    if (leaveMessagesC->isChecked()) {
        dontDuplicateC->setEnabled(true);
        if (dontDuplicateC->isChecked()) {
            leaveOnL->setEnabled(true);
            thisManyDaysE->setEnabled(true);
            daysL->setEnabled(true);
        } else {
            leaveOnL->setEnabled(false);
            thisManyDaysE->setEnabled(false);
            daysL->setEnabled(false);
        }
    } else {
        dontDuplicateC->setEnabled(false);
        leaveOnL->setEnabled(false);
        thisManyDaysE->setEnabled(false);
        daysL->setEnabled(false);
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailProperties::popSecurityClicked
 *=------------------------------------------------------------------------=
 * change in the type of pop security
 */
void MailProperties::popSecurityClicked()
{
    PopSecurity val;

    if (popNormalAuthB->isChecked()) 
        val = PopNormal;
    else if (popSSLAuthB->isChecked())
        val = PopOverSSL;
    else
        val = PopApop;

    setPopSecurity(val);
}

/**
 *=------------------------------------------------------------------------=
 * MailProperties::selectServerType
 *=------------------------------------------------------------------------=
 * selects the server type we're going to work with.
 *
 * Parameters:
 *      int                 - [in]  index in the list of items.
 */
void MailProperties::selectServerType
(
    int index
)
{
    switch (index) {
        case 0:
            sourceTypeM->setCurrentItem(0);
            sourceLocationL->setText(intlLoadStringX(ID_POP_SERVER));
            setPopControlsEnabled(true);
            break;
        case 1:
            sourceTypeM->setCurrentItem(1);
            sourceLocationL->setText(intlLoadStringX(ID_SPOOL_LOCATION));
            setPopControlsEnabled(false);
            break;
        default:
            K_FAIL("Bogus Server Type value!!! \n");
            break;
    }

    sourceLocationE->setText(NULL);
}


/**
 *=------------------------------------------------------------------------=
 * MailProperties::setPopControlsEnabled
 *=------------------------------------------------------------------------=
 * enables or disables the pop server controls.
 *
 * Parameters:
 *      bool                    - [in]  true == enabled.
 */
void MailProperties::setPopControlsEnabled
(
    bool enable
)
{
    popUserNameL->setEnabled(enable);
    popPasswordL->setEnabled(enable);
    popUserNameE->setEnabled(enable);
    popPasswordE->setEnabled(enable);
    rememberPopPasswordC->setEnabled(enable);
}

/**
 *=------------------------------------------------------------------------=
 * MailProperties::setPopSecurity
 *=------------------------------------------------------------------------=
 * sets up the pop security type in the checkboxes.
 *
 * Parameters:
 *      PopSecurity               - [in]  a value from teh PopSecurity enum.
 */
void MailProperties::setPopSecurity
(
    PopSecurity type
)
{
    switch (type) {
        case PopNormal:
            popNormalAuthB->setChecked(true);
            unusedOtherPortL->setEnabled(true);
            unusedPortL->setEnabled(false);
            popNormalPortE->setEnabled(true);
            popSSLPortE->setEnabled(false);
            break;
        case PopOverSSL:
            popSSLAuthB->setChecked(true);
            unusedOtherPortL->setEnabled(false);
            unusedPortL->setEnabled(true);
            popNormalPortE->setEnabled(false);
            popSSLPortE->setEnabled(true);
            break;
        case PopApop:
            popApopAuthB->setChecked(true);
            unusedOtherPortL->setEnabled(false);
            unusedPortL->setEnabled(false);
            popNormalPortE->setEnabled(false);
            popSSLPortE->setEnabled(false);
            break;
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailProperties::setSmtpAuthEnabled
 *=------------------------------------------------------------------------=
 * enables or disables the various controls for SMTP Authentication based
 * on the value of the checkbox
 *
 * Parameters:
 *      bool                    - [in]  they should be enabled
 */
void MailProperties::setSmtpAuthEnabled
(
    bool enabled
)
{
    smtpAuthBG->setEnabled(enabled);
    smtpUserNameL->setEnabled(enabled && useSmtpAuthB->isChecked());
    smtpUserNameE->setEnabled(enabled && useSmtpAuthB->isChecked());
    smtpPasswordL->setEnabled(enabled && useSmtpAuthB->isChecked());
    smtpPasswordE->setEnabled(enabled && useSmtpAuthB->isChecked());
    smtpRememberPasswordC->setEnabled(enabled && useSmtpAuthB->isChecked());
}

/**
 *=------------------------------------------------------------------------=
 * MailProperties::setSmtpAuthToPopAuth
 *=------------------------------------------------------------------------=
 * should secure SMTP channels use the same login info as pop connections, or
 * get their own?
 *
 * Parameters:
 *      bool                    - [in]  true == yes
 */
void MailProperties::setSmtpAuthToPopAuth
(
    bool yes
)
{
    if (yes) {
        usePop3AuthB->setChecked(true);
        smtpUserNameL->setEnabled(false);
        smtpUserNameE->setEnabled(false);
        smtpPasswordL->setEnabled(false);
        smtpPasswordE->setEnabled(false);
        smtpRememberPasswordC->setEnabled(false);
    } else {
        useSmtpAuthB->setChecked(true);
        smtpUserNameL->setEnabled(true);
        smtpUserNameE->setEnabled(true);
        smtpPasswordL->setEnabled(true);
        smtpPasswordE->setEnabled(true);
        smtpRememberPasswordC->setEnabled(true);
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailProperties::signatureButtonClicked
 *=------------------------------------------------------------------------=
 * show a dialog letting the user select a signature file.
 */
void MailProperties::signatureButtonClicked()
{
    const char *title, *text;
    char buf[MAXPATHLEN];
    QString file;
    ERRCODE er;
    int i;

    file = QFileDialog::getOpenFileName(signatureFileE->text());

    /**
     * IF they chose a file, we will offer to copy it to the .kiltdown 
     * directory for them to help them reduce filesystem clutter.
     */
    if (!file.isEmpty()) {

        /**
         * Show a dialog asking.
         */
        title = intlLoadString(ID_KILTDOWN);
        if (!title) return;
        text = intlLoadString(ID_ASKFORMOVE);
        if (!text) { localFree(title); return; }
        snprintf(buf, sizeof(buf), text, accountName);
        i = QMessageBox::information(this, title, buf, QMessageBox::Yes, QMessageBox::No);
        localFree(text); localFree(title);

        /**
         * If they asked to move, it do so now.
         */
        if (i == QMessageBox::Yes) {
            _generateFileName((const char *)file, accountDirectory, buf, sizeof(buf));

            er = fileCopyFile((const char *)file, buf);
            if (FAILED(er)) {
                errorShowError(er);
                return;
            }

            signatureFileE->setText(buf);

        } else {

            /**
             * No, just let them use the regular old .signature file.
             */
            signatureFileE->setText(file);
        }
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailProperties::smtpAuthTypeClicked
 *=------------------------------------------------------------------------=
 */
void MailProperties::smtpAuthTypeClicked()
{
    setSmtpAuthToPopAuth(usePop3AuthB->isChecked());
}



/**
 *=------------------------------------------------------------------------=
 *                             Local Routines
 *=------------------------------------------------------------------------=
 */




/**
 *=------------------------------------------------------------------------=
 * _generateFileName
 *=------------------------------------------------------------------------=
 * given a source full path and a dest dir, put the filename on the dest
 * dir.
 *
 * Parameters:
 *      const char *            - [in]  full source path.
 *      const char *            - [in]  dest dir
 *      char *                  - [in]  buf to put result.
 *      unsigned int            - [in]  sizeof(buf)
 */
void _generateFileName
(
    const char *sourcePath,
    const char *destDir,
    char *buf,
    unsigned int sizeofBuf
)
{
    char *ptr;
    int len;

    K_ASSERT(sourcePath && destDir, "Bogus Parameters!!!");

    /**    
     * First, if they didn't give us a filename, then we need to get that
     * now.
     */
    ptr = strrchr(sourcePath, '/');
    strncpy(buf, destDir, sizeofBuf);
    len = strlen(destDir);
    if (buf[len] != '/') {
        buf[len++] = '/';
        buf[len++] = '\0';
    }
    if (ptr) 
        strncat(buf, ptr + 1, sizeofBuf - len);
    else
        strncat(buf, sourcePath, sizeofBuf - len);
}

