/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include "mailpropsold.h"


#include <qcheckbox.h>
#include <qcombobox.h>
#include <qfiledialog.h>
#include <qgroupbox.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qvalidator.h>
#include <sys/param.h>

#include "fileutil.h"
#include <intl.h>
#include "localized.h"
#include "mailaccount.h"


static const int LEFT_MARGIN = 10;
static const int TOP_MARGIN = 15;
static const int CELL_PAD = 3;

#define CK(ptr)                        if (!(ptr)) return E_OUTOFMEMORY
#define CKN(ptr)                       if (!(ptr)) return NULL

/**
 *=------------------------------------------------------------------------=
 *                            local prototypes
 *=------------------------------------------------------------------------=
 */

QLabel *_createLabel(QWidget *parent, STRINGID id);
QLineEdit *_createLineEdit(QWidget *parent);
QCheckBox *_createCheckBox(QWidget *parent, STRINGID id);
QRadioButton *_createRadioButton(QWidget *parent, STRINGID id);


/**
 *=------------------------------------------------------------------------=
 * AccountInfoPage::AccountInfoPage
 *=------------------------------------------------------------------------=
 * intialize some goo.
 */
AccountInfoPage::AccountInfoPage()
{
    accountNameL = userNameL = userOrganizationL = emailAddressL =
        signatureFileL = NULL;
    accountNameE = userNameE = userOrganizationE = emailAddressE =
        signatureFileE = NULL;
    signatureFileP = NULL;

    accountDirectory = NULL;
    accountName = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * AccountInfoPage::~AccountInfoPage
 *=------------------------------------------------------------------------=
 * clean up some goo.
 */
AccountInfoPage::~AccountInfoPage()
{
    if (accountNameL) delete accountNameL;
    if (userNameL) delete userNameL;
    if (userOrganizationL) delete userOrganizationL;
    if (emailAddressL) delete emailAddressL;
    if (signatureFileL) delete signatureFileL;

    if (accountNameE) delete accountNameE;
    if (userNameE) delete userNameE;
    if (userOrganizationE) delete userOrganizationE;
    if (emailAddressE) delete emailAddressE;
    if (signatureFileE) delete signatureFileE;

    if (signatureFileP) delete signatureFileP;

    if (accountDirectory) localFree(accountDirectory);
    if (accountName) localFree(accountName);
}

/**
 *=------------------------------------------------------------------------=
 * AccountInfoPage::buttonClicked
 *=------------------------------------------------------------------------=
 * shows a dialog letting the user select where to get their signature file,
 * and gives them the option of moving it to their .kiltdown directory.
 */
void AccountInfoPage::buttonClicked()
{
    const char *title, *text;
    char buf[MAXPATHLEN];
    QString file;
    ERRCODE er;
    int i;

    file = QFileDialog::getOpenFileName(signatureFileE->text());

    /**
     * IF they chose a file, we will offer to copy it to the .kiltdown 
     * directory for them to help them reduce filesystem clutter.
     */
    if (!file.isEmpty()) {

        /**
         * Show a dialog asking.
         */
        title = intlLoadString(ID_KILTDOWN);
        if (!title) return;
        text = intlLoadString(ID_ASKFORMOVE);
        if (!text) { localFree(title); return; }
        snprintf(buf, sizeof(buf), text, accountName);
        i = QMessageBox::information(this, title, buf, QMessageBox::Yes, QMessageBox::No);
        localFree(text); localFree(title);

        /**
         * If they asked to move, it do so now.
         */
        if (i == QMessageBox::Yes) {
            _generateFileName((const char *)file, accountDirectory, buf, sizeof(buf));

            er = fileCopyFile((const char *)file, buf);
            if (FAILED(er)) {
                errorShowError(er);
                return;
            }

            signatureFileE->setText(buf);

        } else {

            /**
             * No, just let them use the regular old .signature file.
             */
            signatureFileE->setText(file);
        }
    }
}

/**
 *=------------------------------------------------------------------------=
 * AccountInfoPage::getResults
 *=------------------------------------------------------------------------=
 * the user is done with the dialog.  Get the changes.
 *
 * Parameters:
 *      const char * &              - [out] the new account name, or NULL
 *      MailAccountInfo &           - [in]  where to put other changes.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountInfoPage::getResults
(
    const char * &accountName,
    MailAccountInfo &info
)
{
    if (accountNameE->text().isEmpty()) {
        accountName = NULL;
    } else {
        accountName = copyString((const char *)accountNameE->text());
        CK(accountName);
    }

    if (!userNameE->text().isEmpty()) {
        info.userName = copyString((const char *)userNameE->text());
        CK(info.userName);
    }

    if (!userOrganizationE->text().isEmpty()) {
        info.userOrganization = copyString((const char *)userOrganizationE->text());
        CK(info.userOrganization);
    } else {
        info.userOrganization = NULL;
    }

    if (!emailAddressE->text().isEmpty()) {
        info.userEmail = copyString((const char *)emailAddressE->text());
        CK(info.userEmail);
    }

    if (!signatureFileE->text().isEmpty()) {
        info.signatureFile = copyString((const char *)signatureFileE->text());
        CK(info.signatureFile);
    } else {
        info.signatureFile = NULL;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * AccountInfoPage::initialize
 *=------------------------------------------------------------------------=
 * initializes our account information page.
 *
 * Parameters:
 *      const char *                - [in]  account name.
 *      const char *                - [in]  account directory.
 *      MailAccountInfo &           - [in]  info to grab.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountInfoPage::initialize
(
    const char *accountName,
    const char *accountDirectory,
    MailAccountInfo &info
)
{
    if (!accountName) return E_PARAM;

    this->accountDirectory = copyString(accountDirectory);
    this->accountName = copyString(accountName);

    /**
     * Account Name
     */
    accountNameL = _createLabel(this, ID_ACCOUNTNAME);
    CK(accountNameL);
    accountNameL->setAutoResize(true);

    accountNameE = new QLineEdit(this);
    if (!accountNameE) return E_OUTOFMEMORY;
    accountNameE->setText(accountName);
    accountNameE->setBaseSize(100, 30);
    accountNameL->setBuddy(accountNameE);

    /**
     * User Name.
     */
    userNameL = _createLabel(this, ID_USERNAME);
    CK(userNameL);

    userNameE = new QLineEdit(this);
    if (!userNameE) return E_OUTOFMEMORY;
    userNameE->setText(info.userName);
    userNameL->setBuddy(userNameE);

    /**
     * User Organization.
     */
    userOrganizationL = _createLabel(this, ID_USERORGANIZATION);
    CK(userOrganizationL);

    userOrganizationE = new QLineEdit(this);
    if (!userOrganizationE) return E_OUTOFMEMORY;
    userOrganizationE->setText(info.userOrganization);
    userOrganizationL->setBuddy(userOrganizationE);

    /**
     * Email address
     */
    emailAddressL = _createLabel(this, ID_EMAILADDRESS);
    CK(emailAddressL);

    emailAddressE = new QLineEdit(this);
    if (!emailAddressE) return E_OUTOFMEMORY;
    emailAddressE->setText(info.userEmail);
    emailAddressL->setBuddy(emailAddressE);

    /**
     * Signature File
     */
    signatureFileL = _createLabel(this, ID_SIGNATUREFILE);
    CK(signatureFileL);


    signatureFileE = new QLineEdit(this);
    CK(signatureFileE);
    signatureFileE->setText(info.signatureFile);
    signatureFileL->setBuddy(signatureFileE);

    signatureFileP = new QPushButton(this);
    CK(signatureFileP);
    signatureFileP->setText(" ... ");

    QObject::connect(signatureFileP, SIGNAL(clicked()), this, SLOT(buttonClicked()));

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * AccountInfoPage::paintEvent
 *=------------------------------------------------------------------------=
 * repaint.
 *
 * Parameters:
 *      QPaintEvent *               - [in]  info
 */
void AccountInfoPage::paintEvent
(
    QPaintEvent *paint
)
{
    QWidget::paintEvent(paint);

    QPainter p;

    p.begin(this);

    int t = TOP_MARGIN + accountNameL->height() + accountNameE->height() + 2 * CELL_PAD + 10;

    qDrawShadeLine(&p, LEFT_MARGIN, t, this->width() - LEFT_MARGIN, t, colorGroup(), true, 1, 1);

    p.end();
}

/**
 *=------------------------------------------------------------------------=
 * AccountInfoPage::resizeEvent
 *=------------------------------------------------------------------------=
 * lays all the goo out!
 *
 * Parameters:
 *      QResizeEvent *              - [in]  details.
 */
void AccountInfoPage::resizeEvent
(
    QResizeEvent *event
)
{
    QSize size = event->size();
    int wider, b;
    int top;

    top = TOP_MARGIN;
    accountNameL->move(LEFT_MARGIN, top);
    top += accountNameL->height() + CELL_PAD;
    accountNameE->setGeometry(LEFT_MARGIN, top, size.width() - (2 * LEFT_MARGIN), accountNameE->height());
    top += accountNameE->height() + CELL_PAD;

    /**
     * Figure out which is wider, name or org.
     */
    top += 20;
    b = userOrganizationL->sizeHint().width();
    wider = userNameL->sizeHint().width();
    wider = (wider > b) ? wider : b;
    userNameL->move(LEFT_MARGIN, top);
    b = LEFT_MARGIN + CELL_PAD + wider;
    userNameE->setGeometry(b, top, size.width() - b - LEFT_MARGIN, accountNameE->height());
    top += userNameE->height() + CELL_PAD;

    userOrganizationL->move(LEFT_MARGIN, top);
    userOrganizationE->setGeometry(b, top, size.width() - b - LEFT_MARGIN, userOrganizationE->height());
    top += userOrganizationE->height() + CELL_PAD;
    top += 10;

    emailAddressL->move(LEFT_MARGIN, top);
    top += emailAddressL->height() + CELL_PAD;
    emailAddressE->setGeometry(LEFT_MARGIN, top, size.width() - (2 * LEFT_MARGIN), emailAddressE->height());
    top += emailAddressE->height() + CELL_PAD * 4;

    signatureFileL->move(LEFT_MARGIN, top);
    top += signatureFileL->height() + CELL_PAD;
    signatureFileE->setGeometry(LEFT_MARGIN,
                                top, 
                                size.width() - (2 * LEFT_MARGIN) - (2 * CELL_PAD) - 20,
                                signatureFileE->height());
    signatureFileP->setGeometry(size.width() - LEFT_MARGIN - 20, top,
                                20, signatureFileE->height());
}


/**
 *=------------------------------------------------------------------------=
 * AccountIncomingPage::AccountIncomingPage
 *=------------------------------------------------------------------------=
 * initialize thingies
 */
AccountIncomingPage::AccountIncomingPage()
{
    typeOfServerL = serverNameL = accountInformationL = popUserNameL = 
        popPasswordL = deliveryL = NULL;
    serverNameE = popUserNameE = popPasswordE = NULL;
    rememberPasswordC = leaveOnServerC = dontDuplicateC = NULL;
    typeOfServerB = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * AccountIncomingPage::~AccountIncomingPage
 *=------------------------------------------------------------------------=
 * clean up!
 */
AccountIncomingPage::~AccountIncomingPage()
{
    if (typeOfServerL) delete typeOfServerL;
    if (deliveryL) delete deliveryL;
    if (serverNameL) delete serverNameL;
    if (accountInformationL) delete accountInformationL;
    if (popUserNameL) delete popUserNameL;
    if (popPasswordL) delete popPasswordL;

    if (serverNameE) delete serverNameE;
    if (popUserNameE) delete popUserNameE;
    if (popPasswordE) delete popPasswordE;

    if (rememberPasswordC) delete rememberPasswordC;
    if (leaveOnServerC) delete leaveOnServerC;
    if (dontDuplicateC) delete dontDuplicateC;

    if (typeOfServerB) delete typeOfServerB;
}


/**
 *=------------------------------------------------------------------------=
 * AccountIncomingPage::getResults
 *=------------------------------------------------------------------------=
 * the user is done with the dialog.  Get the changes.
 *
 * Parameters:
 *      MailAccountInfo &           - [in]  where to put other changes.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountIncomingPage::getResults
(
    MailAccountInfo &info
)
{
    if (!serverNameE->text().isEmpty()) {
        info.incomingPop = copyString((const char *)serverNameE->text());
        if (!info.incomingPop) return E_OUTOFMEMORY;
    }
    if (!popUserNameE->text().isEmpty()) {
        info.popUserName = copyString((const char *)popUserNameE->text());
        if (!info.popUserName) return E_OUTOFMEMORY;
    }
    if (!popPasswordE->text().isEmpty()) {
        info.popPassword = copyString((const char *)popPasswordE->text());
        if (!info.popPassword) return E_OUTOFMEMORY;
    }
    
    info.rememberPopPassword = rememberPasswordC->isChecked();
    info.popRemoveFromServer = !leaveOnServerC->isChecked();
    info.popNoDuplicates = dontDuplicateC->isChecked();

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * AccountIncomingPage::initialize
 *=------------------------------------------------------------------------=
 * inits our UI.
 *
 * Parameters:
 *      MailAccountInfo &               - [in]  info 
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountIncomingPage::initialize
(
    MailAccountInfo &info
)
{
    STRINGID acctTypes [] = {
        ID_POP3_SERVER,
        ID_SPOOL_SERVER
    };
    const char *pstr;
    
    /**
     * Type of server
     */
    typeOfServerL = _createLabel(this, ID_SERVER_TYPE);
    CK(typeOfServerL);

    typeOfServerB = new QComboBox(this);
    if (!typeOfServerB) return E_OUTOFMEMORY;

    for (unsigned int x = 0; x < sizeof(acctTypes) / sizeof(STRINGID); x++) {
        pstr = intlLoadString(acctTypes[x]);
        if (!pstr) return E_OUTOFMEMORY;
        typeOfServerB->insertItem(pstr, x);
        localFree(pstr);
    }
    typeOfServerB->setCurrentItem(0);
    typeOfServerL->setBuddy(typeOfServerB);

    /**
     * Server name
     */
    serverNameL = _createLabel(this, ID_SERVER_NAME);
    CK(serverNameL);

    serverNameE = new QLineEdit(this);
    if (!serverNameE) return E_OUTOFMEMORY;
    serverNameE->setText(info.incomingPop);
    serverNameL->setBuddy(serverNameE);

    /**
     * Account info
     */
    accountInformationL = _createLabel(this, ID_ACCOUNTINFORMATION);
    CK(accountInformationL);

    popUserNameL = _createLabel(this, ID_POPNAME);
    CK(popUserNameL);

    popUserNameE = new QLineEdit(this);
    if (!popUserNameE) return E_OUTOFMEMORY;
    popUserNameE->setText(info.popUserName);
    popUserNameL->setBuddy(popUserNameE);

    popPasswordL = _createLabel(this, ID_POPPASSWD);
    CK(popPasswordL);

    popPasswordE = new QLineEdit(this);
    if (!popPasswordE) return E_OUTOFMEMORY;
    popPasswordE->setEchoMode(QLineEdit::Password);
    popPasswordE->setText(info.popPassword);
    popPasswordL->setBuddy(popPasswordE);

    /**
     * remember password
     */
    rememberPasswordC = _createCheckBox(this, ID_REMEMBER_PASSWORD);
    CK(rememberPasswordC);
    rememberPasswordC->setChecked(info.rememberPopPassword);

    /**
     * Delivery info
     */
    deliveryL = _createLabel(this, ID_DELIVERY);
    CK(deliveryL);

    leaveOnServerC = _createCheckBox(this, ID_LEAVEONSERVER);
    CK(leaveOnServerC);
    leaveOnServerC->setChecked(!info.popRemoveFromServer);

    dontDuplicateC = _createCheckBox(this, ID_DONTDUPLICATE);
    CK(dontDuplicateC);
    dontDuplicateC->setChecked(info.popNoDuplicates);
    dontDuplicateC->setEnabled(!info.popRemoveFromServer);

    QObject::connect(leaveOnServerC, SIGNAL(clicked()), this, SLOT(leaveOnServerClicked()));

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * AccountIncomingPage::leaveOnServerClicked
 *=------------------------------------------------------------------------=
 * checkbox has been clicked.
 */
void AccountIncomingPage::leaveOnServerClicked()
{
    dontDuplicateC->setEnabled(leaveOnServerC->isChecked());
}

/**
 *=------------------------------------------------------------------------=
 * AccountIncomingPage::paintEvent
 *=------------------------------------------------------------------------=
 * repaint.
 *
 * Parameters:
 *      QPaintEvent *               - [in]  Incoming
 */
void AccountIncomingPage::paintEvent
(
    QPaintEvent *paint
)
{
    QWidget::paintEvent(paint);

    QPainter p;

    p.begin(this);

    int t = TOP_MARGIN + typeOfServerL->height() + serverNameL->height() 
            + 2 * CELL_PAD + (accountInformationL->height() / 2);
    int left = accountInformationL->width() + LEFT_MARGIN + CELL_PAD;

    qDrawShadeLine(&p, left, t, this->width() - LEFT_MARGIN, t, colorGroup(), true, 1, 1);

    t -= accountInformationL->height() / 2;
    t += accountInformationL->height() + popUserNameL->height() 
         + popPasswordL->height() + rememberPasswordC->height() + 3 * CELL_PAD
         + (deliveryL->height() / 2) + 2 ;
    left = deliveryL->width() + LEFT_MARGIN + CELL_PAD;
    qDrawShadeLine(&p, left, t, this->width() - LEFT_MARGIN, t, colorGroup(), true, 1, 1);


    p.end();
}

/**
 *=------------------------------------------------------------------------=
 * AccountIncomingPage::resizeEvent
 *=------------------------------------------------------------------------=
 * lays all the goo out!
 *
 * Parameters:
 *      QResizeEvent *              - [in]  details.
 */
void AccountIncomingPage::resizeEvent
(
    QResizeEvent *event
)
{
    QSize size = event->size();
    int wider, b;
    int top;

    top = TOP_MARGIN;

    typeOfServerL->setGeometry(LEFT_MARGIN, top, typeOfServerL->sizeHint().width(), typeOfServerL->height());
    b = typeOfServerL->sizeHint().width() + LEFT_MARGIN + CELL_PAD;
    typeOfServerB->setGeometry(b, top, size.width() - b - LEFT_MARGIN, typeOfServerB->height());
    top += typeOfServerB->height() + CELL_PAD;

    serverNameL->setGeometry(LEFT_MARGIN, top, serverNameL->sizeHint().width(), serverNameL->height());
    b = typeOfServerL->sizeHint().width() + LEFT_MARGIN + CELL_PAD;
    serverNameE->setGeometry(b, top, size.width() - b - LEFT_MARGIN, serverNameE->height());
    top += serverNameE->height() + CELL_PAD;

    accountInformationL->setGeometry(LEFT_MARGIN, top, accountInformationL->sizeHint().width(), accountInformationL->height());
    top += accountInformationL->height() + CELL_PAD;

    /**
     * get the wider of the user name/password label.
     */
    b = popUserNameL->sizeHint().width();
    wider = popPasswordL->sizeHint().width();
    wider = (wider > b ) ? wider : b;

    popUserNameL->setGeometry(LEFT_MARGIN, top, popUserNameL->sizeHint().width(), popUserNameL->height());
    b = wider + LEFT_MARGIN + CELL_PAD;
    popUserNameE->setGeometry(b, top, size.width() - b - LEFT_MARGIN, popUserNameE->height());
    top += popUserNameE->height() + CELL_PAD;

    popPasswordL->setGeometry(LEFT_MARGIN, top, popPasswordL->sizeHint().width(), popPasswordL->height());
    popPasswordE->setGeometry(b, top, size.width() - b - LEFT_MARGIN, popPasswordE->height());
    top += popPasswordE->height() + CELL_PAD;

    rememberPasswordC->setGeometry(3 * LEFT_MARGIN, top, rememberPasswordC->sizeHint().width(), rememberPasswordC->height());
    top += rememberPasswordC->height() + CELL_PAD;

    deliveryL->setGeometry(LEFT_MARGIN, top, deliveryL->sizeHint().width(), deliveryL->height());
    top += deliveryL->height() + CELL_PAD;

    leaveOnServerC->setGeometry(LEFT_MARGIN + 3, top, leaveOnServerC->sizeHint().width(), leaveOnServerC->height());
    top += leaveOnServerC->height() + CELL_PAD;
    dontDuplicateC->setGeometry(LEFT_MARGIN * 3, top, dontDuplicateC->sizeHint().width(), dontDuplicateC->height());
    top += dontDuplicateC->height() + CELL_PAD;
}


/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::AccountOutgoingPage
 *=------------------------------------------------------------------------=
 * init
 */
AccountOutgoingPage::AccountOutgoingPage()
{
    serverNameL = smtpUserNameL = smtpPasswordL = NULL;
    serverNameE = smtpUserNameE = smtpPasswordE = NULL;
    serverNeedsAuthC = rememberSmtpPasswordC = NULL;
    frame = NULL;
    usePopB = useSpecifiedB = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::~AccountOutgoingPage
 *=------------------------------------------------------------------------=
 * clean up
 */
AccountOutgoingPage::~AccountOutgoingPage()
{
    if (serverNameL) delete serverNameL;
    if (smtpUserNameL) delete smtpUserNameL;
    if (smtpPasswordL) delete smtpPasswordL;
    if (serverNeedsAuthC) delete serverNeedsAuthC;
    if (rememberSmtpPasswordC) delete rememberSmtpPasswordC;
    if (usePopB) delete usePopB;
    if (useSpecifiedB) delete useSpecifiedB;
    if (frame) delete frame;
}


/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::getResults
 *=------------------------------------------------------------------------=
 * the user is done with the dialog.  Get the changes.
 *
 * Parameters:
 *      MailAccountInfo &           - [in]  where to put other changes.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountOutgoingPage::getResults
(
    MailAccountInfo &info
)
{
    if (!serverNameE->text().isEmpty()) {
        info.outgoingSMTP = copyString((const char *)serverNameE->text());
        if (!info.outgoingSMTP) return E_OUTOFMEMORY;
    }
    info.smtpNeedsAuth = serverNeedsAuthC->isChecked();
    
    if (!smtpUserNameE->text().isEmpty()) {
        info.smtpUserName = copyString((const char *)smtpUserNameE->text());
        if (!info.smtpUserName) return E_OUTOFMEMORY;
    }
    if (!smtpPasswordE->text().isEmpty()) {
        info.smtpPassword = copyString((const char *)smtpPasswordE->text());
        if (!info.smtpPassword) return E_OUTOFMEMORY;
    }
    
    info.usePopForSmtp = usePopB->isChecked();
    info.rememberSmtpPassword = rememberSmtpPasswordC->isChecked();

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::initialize
 *=------------------------------------------------------------------------=
 * init.
 *
 * Parameters:
 *      MailAccountInfo &               - [in]
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountOutgoingPage::initialize
(
    MailAccountInfo &info
)
{
    const char *pstr;

    /**
     * server name
     */
    serverNameL = new QLabel(this);
    CK(serverNameL);
    pstr = intlLoadString(ID_SMTPSERVER_NAME);
    CK(pstr);
    serverNameL->setText(pstr);
    localFree(pstr);
    serverNameE = new QLineEdit(this);
    CK(serverNameE);
    serverNameE->setText(info.outgoingSMTP);
    serverNameL->setBuddy(serverNameE);

    frame = new QGroupBox(this);
    CK(frame);
    frame->setTitle(" ");

    serverNeedsAuthC = new QCheckBox(this);
    CK(serverNeedsAuthC);
    pstr = intlLoadString(ID_SERVERNEEDSAUTH);
    CK(pstr);
    serverNeedsAuthC->setText(pstr);
    localFree(pstr);
    serverNeedsAuthC->setChecked(info.smtpNeedsAuth);

    usePopB = new QRadioButton(frame);
    CK(usePopB);
    pstr = intlLoadString(ID_USEPOPCRED);
    CK(pstr);
    usePopB->setText(pstr);
    localFree(pstr);
    usePopB->setChecked(info.usePopForSmtp);

    useSpecifiedB = new QRadioButton(frame);
    CK(useSpecifiedB);
    pstr = intlLoadString(ID_USESPECIFIED);
    CK(pstr);
    useSpecifiedB->setText(pstr);
    localFree(pstr);
    usePopB->setChecked(!info.usePopForSmtp);

    smtpUserNameL = new QLabel(frame);
    CK(smtpUserNameL);
    pstr = intlLoadString(ID_POPNAME);
    CK(pstr);
    smtpUserNameL->setText(pstr);
    localFree(pstr);
    smtpUserNameE = new QLineEdit(frame);
    CK(smtpUserNameE);
    smtpUserNameE->setText(info.smtpUserName);
    smtpUserNameL->setBuddy(smtpUserNameE);

    smtpPasswordL = new QLabel(frame);
    CK(smtpPasswordL);
    pstr = intlLoadString(ID_POPPASSWD);
    CK(pstr);
    smtpPasswordL->setText(pstr);
    localFree(pstr);
    smtpPasswordE = new QLineEdit(frame);
    CK(smtpPasswordE);
    smtpPasswordE->setText(info.smtpPassword);
    smtpPasswordL->setBuddy(smtpPasswordE);

    rememberSmtpPasswordC = new QCheckBox(frame);
    CK(rememberSmtpPasswordC);
    pstr = intlLoadString(ID_REMEMBER_PASSWORD);
    CK(pstr);
    rememberSmtpPasswordC->setText(pstr);
    localFree(pstr);
    rememberSmtpPasswordC->setChecked(info.rememberSmtpPassword);


    frame->setEnabled(info.smtpNeedsAuth);
    QObject::connect(serverNeedsAuthC, SIGNAL(clicked()), this, SLOT(serverNeedsAuthClicked()));
    QObject::connect(usePopB, SIGNAL(clicked()), this, SLOT(usePopButtonClicked()));
    QObject::connect(useSpecifiedB, SIGNAL(clicked()), this, SLOT(useSpecifiedButtonClicked()));

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::paintEvent
 *=------------------------------------------------------------------------=
 * repaint.
 *
 * Parameters:
 *      QPaintEvent *               - [in]  info
 */
void AccountOutgoingPage::paintEvent
(
    QPaintEvent *paint
)
{
    QWidget::paintEvent(paint);

    QPainter p;

    p.begin(this);

    p.end();
}

/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::resizeEvent
 *=------------------------------------------------------------------------=
 * lays all the goo out!
 *
 * Parameters:
 *      QResizeEvent *              - [in]  details.
 */
void AccountOutgoingPage::resizeEvent
(
    QResizeEvent *event
)
{
    QSize size = event->size();
    int wider, b;
    int top;

    top = TOP_MARGIN;

    b = serverNameL->sizeHint().width();
    serverNameL->setGeometry(LEFT_MARGIN, top, b, serverNameL->height());
    serverNameE->setGeometry(b + LEFT_MARGIN + CELL_PAD, top, size.width() - b - 2 * LEFT_MARGIN - CELL_PAD, serverNameE->height());
    top += serverNameL->height() + 2 * CELL_PAD;

    frame->setGeometry(LEFT_MARGIN, top, size.width() - 2 * LEFT_MARGIN, size.height() - top - LEFT_MARGIN);

    serverNeedsAuthC->setGeometry(LEFT_MARGIN * 2, top - 5, serverNeedsAuthC->sizeHint().width(), serverNeedsAuthC->height());
    top += serverNeedsAuthC->height() + CELL_PAD;

    top = 25;

    usePopB->setGeometry(LEFT_MARGIN , top, usePopB->sizeHint().width(), usePopB->height());
    top += usePopB->height();
    useSpecifiedB->setGeometry(LEFT_MARGIN , top, useSpecifiedB->sizeHint().width(), useSpecifiedB->height());
    top += useSpecifiedB->height();

    b = smtpUserNameL->sizeHint().width();
    wider = smtpPasswordL->sizeHint().width();
    wider = (wider > b) ? wider : b;
    smtpUserNameL->setGeometry(LEFT_MARGIN * 2, top, b, smtpUserNameL->height());
    smtpUserNameE->setGeometry(LEFT_MARGIN * 2 + wider + CELL_PAD, top, 
                               frame->width() - LEFT_MARGIN * 4 - wider - CELL_PAD,
                               smtpUserNameE->height());
    top += smtpUserNameL->height() + CELL_PAD;
    smtpPasswordL->setGeometry(LEFT_MARGIN * 2, top, wider, smtpPasswordL->height());
    smtpPasswordE->setGeometry(LEFT_MARGIN * 2 + wider + CELL_PAD, top,
                               frame->width() - LEFT_MARGIN * 4 - wider - CELL_PAD,
                               smtpPasswordE->height());
    top += smtpPasswordL->height();

    rememberSmtpPasswordC->setGeometry(LEFT_MARGIN * 2, top, rememberSmtpPasswordC->sizeHint().width(), rememberSmtpPasswordC->height());
}

/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::serverNeedsAuthClicked
 *=------------------------------------------------------------------------=
 */
void AccountOutgoingPage::serverNeedsAuthClicked()
{
    frame->setEnabled(serverNeedsAuthC->isChecked());
}

/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::usePopButtonClicked
 *=------------------------------------------------------------------------=
 * the user will use the incoming POP authentication info.
 */
void AccountOutgoingPage::usePopButtonClicked()
{
    usePopB->setChecked(true);
    useSpecifiedB->setChecked(false);

    smtpUserNameL->setEnabled(false);
    smtpUserNameE->setEnabled(false);
    smtpPasswordL->setEnabled(false);
    smtpPasswordE->setEnabled(false);
    rememberSmtpPasswordC->setEnabled(false);
}

/**
 *=------------------------------------------------------------------------=
 * AccountOutgoingPage::useSpecifiedButtonClicked
 *=------------------------------------------------------------------------=
 * the user will use the incoming POP authentication info.
 */
void AccountOutgoingPage::useSpecifiedButtonClicked()
{
    usePopB->setChecked(false);
    useSpecifiedB->setChecked(true);

    smtpUserNameL->setEnabled(true);
    smtpUserNameE->setEnabled(true);
    smtpPasswordL->setEnabled(true);
    smtpPasswordE->setEnabled(true);
    rememberSmtpPasswordC->setEnabled(true);
}

/**
 *=------------------------------------------------------------------------=
 * AccountAdvancedPage::AccountAdvancedPage
 *=------------------------------------------------------------------------=
 * advanced information page for accounts.
 */
AccountAdvancedPage::AccountAdvancedPage()
{
    smtpPortL = pop3L = pop3PortL = pop3SSLPortL = serverTimeoutL = secsL = NULL;
    smtpPortE = pop3PortE = pop3SSLPortE = timeoutE = NULL;
    smtpUseSSLC = NULL;
    regularPopB = securePopB = NULL;

    frame = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * AccountAdvancedPage::~AccountAdvancedPage
 *=------------------------------------------------------------------------=
 * clean up
 */
AccountAdvancedPage::~AccountAdvancedPage()
{
    if (smtpPortL) delete smtpPortL;
    if (pop3L) delete pop3L;
    if (pop3PortL) delete pop3PortL;
    if (pop3SSLPortL) delete pop3SSLPortL;
    if (serverTimeoutL) delete serverTimeoutL;
    if (secsL) delete secsL;

    if (smtpPortE) delete smtpPortE;
    if (pop3PortE) delete pop3PortE;
    if (pop3SSLPortE) delete pop3SSLPortE;
    if (timeoutE) delete timeoutE;

    if (smtpUseSSLC) delete smtpUseSSLC;

    if (regularPopB) delete regularPopB;
    if (securePopB) delete securePopB;

    if (frame) delete frame;
}


/**
 *=------------------------------------------------------------------------=
 * AccountAdvancedPage::getResults
 *=------------------------------------------------------------------------=
 * the user is done with the dialog.  Get the changes.
 *
 * Parameters:
 *      MailAccountInfo &           - [in]  where to put other changes.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountAdvancedPage::getResults
(
    MailAccountInfo &info
)
{
    info.smtpPort = smtpPortE->text().toInt();
    info.pop3Port = pop3PortE->text().toInt();
    info.securePop3Port = pop3SSLPortE->text().toInt();
    info.useSecureSmtp = smtpUseSSLC->isChecked();
    info.useSecurePop3 = securePopB->isChecked();
    info.serverTimeout = timeoutE->text().toInt();

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * AccountAdvanced::initialize
 *=------------------------------------------------------------------------=
 * initializes our account advanced page.
 *
 * Parameters:
 *      MailAccountInfo &           - [in]  info to grab.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE AccountAdvancedPage::initialize
(
    MailAccountInfo &info
)
{
    const char *pstr;
    char buf[10];

    frame = new QGroupBox(this);
    CK(frame);
    pstr = intlLoadString(ID_PORTNUMBERS);
    CK(pstr);
    frame->setTitle(pstr);
    localFree(pstr);

    /**
     * SMTP Port and SSL Checkbox
     */
    smtpPortL = new QLabel(frame);
    CK(smtpPortL);
    pstr = intlLoadString(ID_SMTP);
    CK(pstr);
    smtpPortL->setText(pstr);
    localFree(pstr);

    smtpPortE = new QLineEdit(frame);
    CK(smtpPortE);
    snprintf(buf, sizeof(buf), "%d", info.smtpPort);
    smtpPortE->setText(buf);
    smtpPortE->setValidator(new QIntValidator(smtpPortE));
    smtpPortL->setBuddy(smtpPortE);

    smtpUseSSLC = new QCheckBox(frame);
    CK(smtpUseSSLC);
    pstr = intlLoadString(ID_USESSL);
    CK(pstr);
    smtpUseSSLC->setText(pstr);
    localFree(pstr);
    smtpUseSSLC->setChecked(info.useSecureSmtp);

    /**
     * Pop3 port and radio buttons.
     */
    pop3L = new QLabel(frame);
    CK(pop3L);
    pstr = intlLoadString(ID_POP);
    CK(pstr);
    pop3L->setText(pstr);
    localFree(pstr);

    regularPopB = new QRadioButton(frame);
    CK(regularPopB);
    pstr = intlLoadString(ID_REGULARPOP);
    CK(pstr);
    regularPopB->setText(pstr);
    localFree(pstr);
    regularPopB->setChecked(info.useSecurePop3);

    pop3PortL = new QLabel(frame);
    CK(pop3PortL);
    pstr = intlLoadString(ID_PORT);
    CK(pstr);
    pop3PortL->setText(pstr);
    localFree(pstr);

    pop3PortE = new QLineEdit(frame);
    CK(pop3PortE);
    snprintf(buf, sizeof(buf), "%d", info.pop3Port);
    pop3PortE->setText(buf);
    pop3PortE->setValidator(new QIntValidator(pop3PortE));
    pop3PortL->setBuddy(pop3PortE);

    securePopB = new QRadioButton(frame);
    CK(securePopB);
    pstr = intlLoadString(ID_SSLPOP);
    CK(pstr);
    securePopB->setText(pstr);
    localFree(pstr);
    securePopB->setChecked(!regularPopB->isChecked());

    pop3SSLPortL = new QLabel(frame);
    CK(pop3SSLPortL);
    pstr = intlLoadString(ID_PORT);
    CK(pstr);
    pop3SSLPortL->setText(pstr);
    localFree(pstr);

    pop3SSLPortE = new QLineEdit(frame);
    CK(pop3SSLPortE);
    snprintf(buf, sizeof(buf), "%d", info.securePop3Port);
    pop3SSLPortE->setText(buf);
    pop3SSLPortE->setValidator(new QIntValidator(pop3SSLPortE));
    pop3SSLPortL->setBuddy(pop3SSLPortE);
    
    /**
     * Server Timeout stuff.
     */
    serverTimeoutL = new QLabel(this);
    CK(serverTimeoutL);
    pstr = intlLoadString(ID_SERVERTIMEOUT);
    CK(pstr);
    serverTimeoutL->setText(pstr);
    localFree(pstr);

    timeoutE = new QLineEdit(this);
    CK(timeoutE);
    snprintf(buf, sizeof(buf), "%d", info.serverTimeout);
    timeoutE->setText(buf);
    timeoutE->setValidator(new QIntValidator(timeoutE));
    serverTimeoutL->setBuddy(timeoutE);

    secsL = new QLabel(this);
    CK(secsL);
    pstr = intlLoadString(ID_SECONDS);
    CK(pstr);
    secsL->setText(pstr);
    localFree(pstr);

    QObject::connect(regularPopB, SIGNAL(clicked()), this, SLOT(regularPopButtonClicked()));
    QObject::connect(securePopB, SIGNAL(clicked()), this, SLOT(securePopButtonClicked()));

    if (info.useSecurePop3)
        securePopButtonClicked();
    else
        regularPopButtonClicked();
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * AccountAdvancedPage::paintEvent
 *=------------------------------------------------------------------------=
 * repaint.
 *
 * Parameters:
 *      QPaintEvent *               - [in]  info
 */
void AccountAdvancedPage::paintEvent
(
    QPaintEvent *paint
)
{
    QWidget::paintEvent(paint);

    QPainter p;

    p.begin(this);
    p.end();
}

/**
 *=------------------------------------------------------------------------=
 * AccountAdvancedPage::regularPopButtonClicked
 *=------------------------------------------------------------------------=
 * the user will use the incoming POP authentication info.
 */
void AccountAdvancedPage::regularPopButtonClicked()
{
    regularPopB->setChecked(true);
    securePopB->setChecked(false);

    pop3PortE->setEnabled(true);
    pop3SSLPortE->setEnabled(false);
}


/**
 *=------------------------------------------------------------------------=
 * AccountAdvancedPage::resizeEvent
 *=------------------------------------------------------------------------=
 * lays all the goo out!
 *
 * Parameters:
 *      QResizeEvent *              - [in]  details.
 */
void AccountAdvancedPage::resizeEvent
(
    QResizeEvent *event
)
{
    int b;
    int top;

    top = 15;

    b = smtpPortL->sizeHint().width();
    smtpPortL->setGeometry(LEFT_MARGIN, top, b, smtpPortL->height());
    smtpPortE->setGeometry(b + LEFT_MARGIN + CELL_PAD, top + 2, 100, smtpPortE->font().pixelSize() + 10);
    top += smtpPortL->height() + CELL_PAD;

    smtpUseSSLC->setGeometry(LEFT_MARGIN * 3, top, smtpUseSSLC->sizeHint().width(), smtpUseSSLC->height());
    top += smtpUseSSLC->height() + CELL_PAD;

    pop3L->setGeometry(LEFT_MARGIN * 2, top, pop3L->sizeHint().width(), pop3L->height());
    top += pop3L->height() + CELL_PAD;

    regularPopB->setGeometry(LEFT_MARGIN * 2, top, regularPopB->sizeHint().width(), regularPopB->height());
    top += regularPopB->height() + CELL_PAD;
    b = pop3PortL->sizeHint().width();
    pop3PortL->setGeometry(LEFT_MARGIN * 4, top, b, pop3PortL->height());
    pop3PortE->setGeometry(LEFT_MARGIN * 4 + CELL_PAD + b, top, 100, pop3PortE->height());
    top += pop3PortL->height() + CELL_PAD;

    securePopB->setGeometry(LEFT_MARGIN * 2, top, securePopB->sizeHint().width(), securePopB->height());
    top += securePopB->height() + CELL_PAD;
    b = pop3SSLPortL->sizeHint().width();
    pop3SSLPortL->setGeometry(LEFT_MARGIN * 4, top, b, pop3SSLPortL->height());
    pop3SSLPortE->setGeometry(LEFT_MARGIN * 4 + CELL_PAD + b, top, 100, pop3SSLPortE->height());
    top += pop3SSLPortL->height() + 3 * CELL_PAD;

    frame->setGeometry(LEFT_MARGIN, 5, event->size().width() - 2 * LEFT_MARGIN, top);

    top += 10;
    b = serverTimeoutL->sizeHint().width();
    serverTimeoutL->setGeometry(LEFT_MARGIN, top, b, serverTimeoutL->height());
    b += LEFT_MARGIN + CELL_PAD;
    timeoutE->setGeometry(b, top, 50, timeoutE->height());
    b += 50 + CELL_PAD;
    secsL->setGeometry(b, top, secsL->sizeHint().width(), secsL->height());
    top += secsL->height();
}

/**
 *=------------------------------------------------------------------------=
 * AccountAdvancedPage::securePopButtonClicked
 *=------------------------------------------------------------------------=
 * advanced information page for accounts.
 */
void AccountAdvancedPage::securePopButtonClicked()
{
    regularPopB->setChecked(false);
    securePopB->setChecked(true);

    pop3PortE->setEnabled(false);
    pop3SSLPortE->setEnabled(true);
}




/**
 *=------------------------------------------------------------------------=
 * _createLabel
 *=------------------------------------------------------------------------=
 * creates a label control.
 *
 * Parameters:
 *      QWidget *               - [in]  parent.
 *      STRINGID                - [in]  text
 *
 * Output:
 *      QLabel *
 */
QLabel *_createLabel
(
    QWidget *parent,
    STRINGID id
)
{
    const char *pstr;
    QLabel *l;

    l = new QLabel(parent);
    CKN(l);
    pstr = intlLoadString(id);
    if (!pstr) {
        delete l;
        return NULL;
    }
    l->setText(pstr);
    localFree(pstr);

    return l;
}

/**
 *=------------------------------------------------------------------------=
 * _createCheckBox
 *=------------------------------------------------------------------------=
 * creates a check box control.
 *
 * Parameters:
 *      QWidget *               - [in]  parent.
 *      STRINGID                - [in]  text
 *
 * Output:
 *      QCheckBox *
 */
QCheckBox *_createCheckBox
(
    QWidget *parent,
    STRINGID id
)
{
    const char *pstr;
    QCheckBox *l;

    l = new QCheckBox(parent);
    CKN(l);
    pstr = intlLoadString(id);
    if (!pstr) {
        delete l;
        return NULL;
    }
    l->setText(pstr);
    localFree(pstr);

    return l;
}

/**
 *=------------------------------------------------------------------------=
 * _createRadioButton
 *=------------------------------------------------------------------------=
 * creates a radio button control.
 *
 * Parameters:
 *      QWidget *               - [in]  parent.
 *      STRINGID                - [in]  text
 *
 * Output:
 *      QRadioButton *
 */
QRadioButton *_createRadioButton
(
    QWidget *parent,
    STRINGID id
)
{
    const char *pstr;
    QRadioButton *l;

    l = new QRadioButton(parent);
    CKN(l);
    pstr = intlLoadString(id);
    if (!pstr) {
        delete l;
        return NULL;
    }
    l->setText(pstr);
    localFree(pstr);

    return l;
}




