/**
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <errno.h>
#include <sys/utsname.h>
#include <time.h>
#include <unistd.h>

#include "address.h"
#include "attachment.h"
#include "deencode.h"
#include "kdispatch.h"
#include "mailmessage.h"
#include "mailsmtp.h"
#include "mime.h"
#include "netapis.h"




static const char *mailerID = "Kiltdown 0.7";
static const char *highPri = "High";
static const char *lowPri = "Low";

static const char *mimeWarning = "This message is in MIME format. Since your \
mail reader does not understand\r\nthis format, some or all of this message may \
not be legible.\r\n\r\n";


/**
 *=------------------------------------------------------------------------=
 *                           Local Prototypes
 *=------------------------------------------------------------------------=
 */
ERRCODE establishBasicConnection(SMTPServerInfo *, MailMessage *, int *);
ERRCODE sendRcptsForList(int socketfd, LinkList<const char *> &);
ERRCODE sendSmtpBody(int socketfd, int, MailMessage *, const char *);
ERRCODE sendSmtpHeader(int socketfd, MailMessage *, const char **);
ERRCODE writeAddressListHeader(char *buf, int max, LinkList<const char *> &);
ERRCODE writeMimeHeader(int, MIMETYPE, MimeEncoding, MimeDisposition, const char *, const char *, const char *);
ERRCODE writeMultipartAlternative(int socketfd, int, MailMessage *, const char *);
ERRCODE writeMultipartMixed(int socketfd, int, MailMessage *, const char *);
ERRCODE encodeSection(int socketfd, MimeEncoding, int lineSize, const char *, int size);



/**
 *=------------------------------------------------------------------------=
 * encodeSection
 *=------------------------------------------------------------------------=
 * given a chunk of goo, put it in the given type in the given network
 * stream.
 *
 * Parameters:
 *      int                         - [in]  socket to which we will write.
 *      MimeEncoding                - [in]  encoding we will use.
 *      int                         - [in]  don't let lines be longer than 
 *                                      this (7bit/q-p only)
 *      const char *                - [in]  pile of goo to encode.
 *      int                         - [in]  size of the section.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE encodeSection
(
    int socketfd,
    MimeEncoding encoding,
    int lineSize,
    const char *goo,
    int size
)
{

    switch (encoding) {
        case Base64:
            return base64Encode(socketfd, goo, size);

        case SevenBit:
            return sevenBitEncode(socketfd, lineSize, goo, size);

        case QuotedPrintable:
            return quotedPrintableEncode(socketfd, lineSize, goo, size);

        case Uuencode:
            return uuencodeEncode(socketfd, goo, size);

        default:
            K_FAIL("Message ENcoding I don't know about !!\n");
            return sevenBitEncode(socketfd, lineSize, goo, size);
    }

    // dead code.
}

/**
 *=------------------------------------------------------------------------=
 * establishBasicConnection
 *=------------------------------------------------------------------------=
 * establishes the basic SMTP connection to the server.  sends the HELO,
 * does the MAIL FROM:, and sends a RCPT TO: for each person to whom this
 * message is being sent.
 *
 * Parameters:
 *      SMTPServerInfo *            - [in]  how to do the connection.
 *      MailMessage *               - [in]  message to transmit
 *      int *                       - [out] the socket we're sending.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE establishBasicConnection
(
    SMTPServerInfo *info,
    MailMessage *message,
    int *socketfd
)
{
    LinkList<const char *> strList;
    char buf[NET_MAXLINELEN];
    struct utsname un;
    const char *pstr;
    ERRCODE er;
    int result;

    /**
     * We need the local host name for the HELO
     */
    result = uname(&un);
    if (result != 0) return E_NOLOCALHOSTNAME;

    /**
     * Establish the connection
     */
    er = networkConnect(info->serverName, info->portnum, socketfd);
    RETURN_ON_FAILURE(er);

    /**
     * Try to read the greeting from them.
     */
    result = networkReadLine(*socketfd, buf, sizeof(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    if (strncasecmp(buf, "220", 3)) {
        fprintf(stderr, "Server returned unexpected code!!!\n");
        er = E_NET_UNEXPECTED;
        goto Error;
    }

    /**
     * Send them the HELO
     */
    snprintf(buf, sizeof(buf), "HELO %s\r\n", un.nodename);
    result = networkWrite(*socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    result = networkReadLine(*socketfd, buf, sizeof(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    if (strncasecmp(buf, "250", 3)) {
        fprintf(stderr, "Server returned unexpected code!!!\n");
        er = E_NET_UNEXPECTED;
        goto Error;
    }

    /**
     * Now write out the MAIL FROM: entry.
     */
    er = message->getSenders(strList);
    GOTO_ON_FAILURE(er, Error);
    er = strList.firstElement(&pstr);
    GOTO_ON_FAILURE(er, Error);
    snprintf(buf, sizeof(buf), "MAIL FROM: %s\n", pstr ? pstr : "<unknown-sender>");

    result = networkWrite(*socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    result = networkReadLine(*socketfd, buf, sizeof(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    if (strncasecmp(buf, "250", 3)) {
        fprintf(stderr, "Server returned unexpected code!!!\n");
        er = E_NET_UNEXPECTED;
        goto Error;
    }

    /**
     * Now, for each person to whom the message is being sent, write out a 
     * RCPT TO: line ...
     */
    strList.emptyList();
    er = message->getRecipients(strList);
    GOTO_ON_FAILURE(er, Error);
    er = sendRcptsForList(*socketfd, strList);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Same for CCs
     */
    strList.emptyList();
    er = message->getCCs(strList);
    GOTO_ON_FAILURE(er, Error);
    er = sendRcptsForList(*socketfd, strList);
    GOTO_ON_FAILURE(er, Error);

    /**
     * And BCCs
     */
    strList.emptyList();
    er = message->getBCCs(strList);
    GOTO_ON_FAILURE(er, Error);
    er = sendRcptsForList(*socketfd, strList);
    GOTO_ON_FAILURE(er, Error);

    return S_OK;

  Error:
    close(*socketfd);
    *socketfd = -1;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * sendRcptsForList
 *=------------------------------------------------------------------------=
 * for each element in the list, send a RCPT TO: <list element> to the 
 * server, and look for a response.
 *
 * Parameters:
 *      int                         - [in]  socket file descriptor
 *      LinkList<const char *> &    - [in]  list to transmit.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE sendRcptsForList
(
    int socketfd,
    LinkList<const char *> &list
)
{
    char buf[NET_MAXLINELEN];
    const char *pstr, *addr;
    ERRCODE er;
    int result;

    /**
     * Whiz through and write
     */
    er = list.firstElement(&pstr);
    while (er == S_OK) {

        addressExtractEmailAddress(pstr, addr);
        K_ASSERT(addr, "I should have SOMETHING here !!!\n");

        snprintf(buf, sizeof(buf), "RCPT TO: %s\n", addr);
        localFree(addr);

        result = networkWrite(socketfd, buf, strlen(buf));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_NET_ERR;
        }

        result = networkReadLine(socketfd, buf, sizeof(buf));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_NET_ERR;
        }

        if (strncasecmp(buf, "250", 3)) {
            fprintf(stderr, "server didn't like RCPT TO: line !!!\n");
            return E_NET_UNEXPECTED;
        }

        er = list.nextElement(&pstr);
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * sendSmtpBody
 *=------------------------------------------------------------------------=
 * sends the body for this SMTP message.
 *
 * Parameters:
 *      int                     - [in]  socket to which we will write.
 *      int                     - [in]  max line length for text sections.
 *      MailMessage *           - [in]  message we are writing out.
 *      const char *            - [in]  MIME separator if we're multipart.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE sendSmtpBody
(
    int socketfd,
    int lineLength,
    MailMessage *message,
    const char *boundary
)
{
    char buf[NET_MAXLINELEN];
    bool hasAttach = false;
    MIMETYPE messageType;
    MimeMediaSubType st;
    const char *body;
    MIMETYPE *types;
    ERRCODE er;
    int result;
    int ct;

    /**
     * first, determine our MIME type
     */
    ct = 0; types = NULL;
    message->getBodyFormats(&types, &ct);
    message->getHasAttachments(&hasAttach);

    if (hasAttach)
        messageType.setType(Multipart, Mixed);
    else if (ct > 1) 
        messageType.setType(Multipart, Alternative);
    else if (ct == 1)
        messageType = types[0];
    else {
        messageType.setType(Text, Plain);
    }

    localFree(types);

    if (messageType.getMediaType() == Multipart) {

        result = networkWrite(socketfd, mimeWarning, strlen(mimeWarning));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_NET_ERR;
        }

        st = messageType.getSubType();
        if (st == Mixed)
            er = writeMultipartMixed(socketfd, lineLength, message, boundary);
        else if (st == Alternative)
            er = writeMultipartAlternative(socketfd, lineLength, message, boundary);
        else {
            K_FAIL("Unexpected Multipart type !!!!\n");
            er = E_INTERNAL;
        }

    } else {
        /** 
         * UNDONE: marcw, 12.12.2000 -- how do i determine whether to use 
         * QuotedPrintable??
         */
        message->getBodyRefByType(messageType, &body);
        er = encodeSection(socketfd, SevenBit, lineLength, body, -1);
    }

    if (FAILED(er)) {
        fprintf(stderr, "Unable to write all the of the message out -- this is bad!\n");
        return er;
    }

    /**
     * Finally, write out the message terminator.
     */
    buf[0] = '.';
    buf[1] = '\r';
    buf[2] = '\n';
    buf[3] = '\0';
    result = networkWrite(socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    result = networkReadLine(socketfd, buf, sizeof(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * Phew!  Done
     */
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * sendSmtpHeader
 *=------------------------------------------------------------------------=
 * writes out the header portion of the message.
 *
 * Parameters:
 *      int                     - [in]  socket to which we will write.
 *      MailMessage *           - [in]  message we are writing out.
 *      const char **           - [out] boundary separator, if we generated
 *                                  one.
 *
 * Output:
 *      ERRCODE
 *
 * Notes:
 *      We are going to do error handling a little different in this function,
 *      because we don't want to stop this once we've started unless we 
 *      absolutely have to ...
 */
ERRCODE sendSmtpHeader
(
    int socketfd,
    MailMessage *message,
    const char **boundary
)
{
    LinkList<const char *> strList;
    char buf[NET_MAXLINELEN];
    MIMETYPE messageType;
    MessagePriority mp;
    const char *pstr;
    bool hasAttach;
    int x, ct, ct1;
    MIMETYPE *rg;
    int result;

    *boundary = NULL;
    
    //
    //
    // UNDONE: marcw, 12.12.2000 -- investigate:
    //      Date
    //      ReplyPath
    //      Sender
    //
    //

    /**
     * From Whom is this message?
     */
    message->getSenders(strList);
    if (strList.countElements()) {
        snprintf(buf, sizeof(buf), "From: ");
        x = 6;
        writeAddressListHeader(buf + x, sizeof(buf) - x, strList);
        result = networkWrite(socketfd, buf, strlen(buf));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_NET_ERR;
        }
    }

    /**
     * To Whom is this message?
     */
    strList.emptyList();
    message->getRecipients(strList);
    if(strList.countElements()) {
        snprintf(buf, sizeof(buf), "To: ");
        x = 4;
        writeAddressListHeader(buf + x, sizeof(buf) - x, strList);
        result = networkWrite(socketfd, buf, strlen(buf));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_NET_ERR;
        }
    }

    /**
     * Cc'd to whom?
     */
    strList.emptyList();
    message->getCCs(strList);
    if(strList.countElements()) {
        snprintf(buf, sizeof(buf), "Cc: ");
        x = 4;
        writeAddressListHeader(buf + x, sizeof(buf) - x, strList);
        result = networkWrite(socketfd, buf, strlen(buf));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_NET_ERR;
        }
    }

    /**
     * Subject.
     */
    message->getSubject(&pstr);
    snprintf(buf, sizeof(buf), "Subject: %s\r\n", pstr ? pstr : "");
    if (pstr) localFree(pstr);
    result = networkWrite(socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * MIME Version
     */
    result = networkWrite(socketfd, "MIME-Version: 1.0\r\n", sizeof("MIME-Version: 1.0\r\n") -1);
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * Mime Content Type, etc ...  First, we need to see what the message has
     * to determine the exact type and sub-type.
     */
    rg = NULL; ct = 0;
    message->getBodyFormats(&rg, &ct);
    hasAttach = false;
    message->getHasAttachments(&hasAttach);

    if (hasAttach)
        messageType.setType(Multipart, Mixed);
    else if (ct > 1)
        messageType.setType(Multipart, Alternative);
    else if (ct == 1)
        messageType = rg[0];
    else {
        fprintf(stderr, "message has no bodies !!!!!!!! \n");
        messageType.setType(Text, Plain);
    }
    if (rg) localFree(rg);

    ct = 0;
    while (mimeMediaTypes[ct].type != UnknownMediaType) {
        if (mimeMediaTypes[ct].type == messageType.getMediaType()) break;
        ct++;
    }
    if (mimeMediaTypes[ct].type == UnknownMediaType) {
        fprintf(stderr, "How the heck did I get a bad type ???\n");
        ct = 0;
    }

    ct1 = 0;
    while (mimeMediaSubTypes[ct1].type != UnknownSubType) {
        if (mimeMediaSubTypes[ct1].type == messageType.getSubType()) break;
        ct1++;
    }
    if (mimeMediaSubTypes[ct1].type == UnknownSubType) {
        fprintf(stderr, "How the heck did I get a bad sub type ???\n");
        ct1 = 0;
    }

    /**
     * finally generate the string and write it out !!! phew!
     */
    if (messageType.getMediaType() == Multipart) {

        char bound[80];
        time_t t = time(NULL);
        snprintf(bound, sizeof(bound), "----=NextPart_000_%8ld=", t);
        *boundary = copyString(bound);
        if (!*boundary) return E_OUTOFMEMORY;
        snprintf(buf, sizeof(buf),
                 "Content-Type: %s/%s;\r\n\tboundary=\"%s\"\r\n",
                 mimeMediaTypes[ct].name,
                 mimeMediaSubTypes[ct1].name,
                 bound);
    } else {
        snprintf(buf, sizeof(buf),
                 "Content-Type: %s/%s;\r\n\tcharset=\"iso-8859-1\"\r\n",
                 mimeMediaTypes[ct].name, mimeMediaSubTypes[ct1].name);

    }

    result = networkWrite(socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * Priority
     */
    message->getPriority(mp);
    if (mp == LowPriority || mp == HighPriority) {
        snprintf(buf, sizeof(buf), "X-MSMail-Priority: %s\r\n", mp == LowPriority ? lowPri : highPri);
        result = networkWrite(socketfd, buf, strlen(buf));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_NET_ERR;
        }
    }

    /**
     * Mailer
     */
    snprintf(buf, sizeof(buf), "X-Mailer: %s\r\n\r\n", mailerID);
    result = networkWrite(socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * NOTE:  THE LAST ENTRY ABOVE HAS AN EXTRA CR/LF PAIR AT THE END TO
     * CORRECTLY TERMINATE THE HEADER !!!!!
     */
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * sendSmtpMessage
 *=------------------------------------------------------------------------=
 * sends a message to a server via SMTP.  does all the mime processing goo
 * and all that jazz.  Very exciting.  
 *
 * Parameters:
 *      MailMessage *               - [in]  message to send.
 *      int                         - [in]  max line length for text sections.
 *      SMTPServerInfo *            - [in]  info on how to connect to server.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE sendSmtpMessage
(
    MailMessage *message,
    int lineLength,
    SMTPServerInfo *info
)
{
    char buf[NET_MAXLINELEN];
    const char *boundary;
    int socketfd;
    ERRCODE er;
    int result;

    /**
     * Set up the basic connection, and do the HELO, MAIL and RCPT commands.
     */
    er = establishBasicConnection(info, message, &socketfd);
    RETURN_ON_FAILURE(er);

    /**
     * Write out the DATA command
     */
    result = networkWrite(socketfd, "DATA\r\n", sizeof("DATA\r\n") - 1);
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    result = networkReadLine(socketfd, buf, sizeof(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    if (strncasecmp(buf, "354", 3)) {
        fprintf(stderr, "Server returned unexpected code!!!\n");
        er = E_NET_UNEXPECTED;
        goto Error;
    }

    /**
     * Okay now send out the header portion of the message.  Note that once
     * we've started writing out the message, we want to write as much as
     * we can without stopping, since there is no real good way in SMTP to
     * abort message sending.  Thus, we'll only abort if there's an error
     * trying a network write or read.
     */
    er = sendSmtpHeader(socketfd, message, &boundary);
    if (er == E_NET_ERR) { goto Error; }

    /**
     * finally, send the body.
     */
    er = sendSmtpBody(socketfd, lineLength, message, boundary);
    if (boundary) localFree(boundary);

  Error:
    close(socketfd);
    return er;
}


/**
 *=------------------------------------------------------------------------=
 * writeAddressListHeader
 *=------------------------------------------------------------------------=
 * writes out the given list as an address list in SMTP header format, doing
 * things like line breaks and all that jazz correctly.
 *
 * Parameters:
 *      char *                      - [in]  buf to write to.
 *      int                         - [in]  amount of space in said buf
 *      LinkList<const char *> &    - [in]  list to write out.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE writeAddressListHeader
(
    char *buf,
    int sizeofbuf,
    LinkList<const char *> &list
)
{
    const char *pstr;
    ERRCODE er;
    bool first = true;
    int left = 8;
    int x = 0;
    int len;

    er = list.firstElement(&pstr);
    while (er == S_OK) {
        
        /**
         * If it's too big for this line, put it on the next.
         */
        len = strlen(pstr);
        if ((left + len > 78) && left != 8) {
            if (x + len + 5 >= sizeofbuf) return E_NOSPACE;
            buf[x++] = ',';
            buf[x++] = ' ';
            buf[x++] = '\r';
            buf[x++] = '\n';
            buf[x++] = '\t';
            buf[x] = '\0';
            left = 8;
            strncat(buf + x, pstr, sizeofbuf - x);
        } else {
            if (x + len + 2 >= sizeofbuf) return E_NOSPACE;
            if (!first) {
                buf[x++] = ','; 
                buf[x++] = ' ';
                buf[x] = '\0';
            }
            first = false;
            strncat(buf + x, pstr, sizeofbuf - x);
        }
        left += len;
        x += len;

        er = list.nextElement(&pstr);
    }

    buf[x++] = '\r';
    buf[x++] = '\n';
    buf[x] = '\0';

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * writeMimeHeader
 *=------------------------------------------------------------------------=
 * writes out the MIME header for this section.
 *
 * Parameters:
 *      int                         - [in]  socket to write to.
 *      MIMETYPE                    - [in]  the type for this section.
 *      MimeEncoding                - [in]  how this section is to be encoded
 *      MimeDisposition             - [in]  disposition
 *      const char *                - [in]  filename, if it's an attachment
 *      const char *                - [in]  charset if necessary
 *      const char *                - [in]  boundary seaprator if it's mp.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE writeMimeHeader
(
    int socketfd,
    MIMETYPE type,
    MimeEncoding encoding,
    MimeDisposition disposition,
    const char *name,
    const char *charset,
    const char *boundary
)
{
    char buf[NET_MAXLINELEN];
    int result;
    int ct, ct1;
    int x;

    if (disposition == MailAttachment && name == NULL) {
        K_FAIL("Attachment with bogus name !!!\n");
        name = "";
    }

    /**
     * First, figure out the top level media type
     */
    ct = 0;
    while (mimeMediaTypes[ct].type != UnknownMediaType) {
        if (mimeMediaTypes[ct].type == type.getMediaType()) break;
        ct++;
    }
    if (mimeMediaTypes[ct].type == UnknownMediaType) {
        K_FAIL("How does one of our messages have an unknown type?\n");
        ct = 0;
    }

    /**
     * then figure out the MIME sub type
     */
    ct1 = 0;
    while (mimeMediaSubTypes[ct1].type != UnknownSubType) {
        if (mimeMediaSubTypes[ct1].type == type.getSubType()) break;
        ct1++;
    }
    if (mimeMediaSubTypes[ct1].type == UnknownSubType) {
        K_FAIL("How does one of our messages have an unknown subtype???\n");
        ct1 = 0;
    }

    x = 0;
    snprintf(buf + x, sizeof(buf) - x, "Content-Type: %s/%s",
             mimeMediaTypes[ct].name,
             mimeMediaSubTypes[ct1].name);
    x = strlen(buf);

    if (charset) {
        snprintf(buf + x, sizeof(buf) - x, "; charset=\"%s\"", charset);
        x = strlen(buf);
    }

    if (name) {
        snprintf(buf + x, sizeof(buf) - x, ";\r\n\tname=\"%s\"", name);
        x = strlen(buf);
    }
    if (boundary) {
        snprintf(buf + x, sizeof(buf) - x, ";\r\n\tboundary=\"%s\"", boundary);
        x = strlen(buf);
    }

    if (encoding != InvalidEncoding) {
        ct = 0;
        while (mimeEncodings[ct].type != InvalidEncoding) {
            if (mimeEncodings[ct].type == encoding) {
                snprintf(buf + x, sizeof(buf) - x, "\r\nContent-Transfer-Encoding: %s", mimeEncodings[ct].name);
                x = strlen(buf);
                break;
            }
            ct++;
        }
    }

    switch (disposition) {
        case Inline:
            snprintf(buf + x, sizeof(buf) - x, "\r\nContent-Disposition: inline");
            x = strlen(buf);
            break;
        case MailAttachment:
            snprintf(buf + x, sizeof(buf) - x, "\r\nContent-Disposition: attachment;\r\n\tname=\"%s\"", name);
            x = strlen(buf);
            break;
        default:
            K_FAIL("Bogus disposition!!!\n");
            break;
    }

    snprintf(buf + x, sizeof(buf) - x, "\r\n\r\n");

    result = networkWrite(socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * writeMultipartAlternative
 *=------------------------------------------------------------------------=
 * the message has more than one body-type.  Go and write out the bodies
 * as a MIME multipart/alternative section.
 *
 * Parameters:
 *      int                         - [in]  socket file descriptor
 *      int                         - [in]  max line length for text section.
 *      MailMessage *               - [in]  message we're writing out.
 *      const char *                - [in]  boundary separator
 *
 * Output:
 *      ERRCODE
 */
ERRCODE writeMultipartAlternative
(
    int socketfd,
    int lineLength,
    MailMessage *message,
    const char *boundary
)
{
    char buf[NET_MAXLINELEN];
    MIMETYPE *types = NULL;
    const char *body;
    ERRCODE er;
    int result;
    int ct;

    /**
     * For each body, go and write out a MIME header and the encode the
     * contents.
     */
    message->getBodyFormats(&types, &ct);

    for (int x = 0; x < ct; x++) {
        
        /**
         * Write out the starting boundary
         */
        snprintf(buf, sizeof(buf), "--%s\r\n", boundary);
        result = networkWrite(socketfd, buf, strlen(buf));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            er = E_NET_ERR;
            goto Error;
        }

        /**
         * Write out a header for this section
         */
        er = writeMimeHeader(socketfd, types[x], SevenBit, Inline, NULL, NULL, NULL);
        GOTO_ON_FAILURE(er, Error);

        /**
         * Write out the body
         * UNDONE: marcw, 12.10.2000 -- we should probably support more 
         * persistence types here -- how?
         */
        body = NULL;
        message->getBodyRefByType(types[x], &body);
        er = encodeSection(socketfd, SevenBit, lineLength, body, -1);
        GOTO_ON_FAILURE(er, Error);
    }

    /**
     * Finally, write a terminating boundary marker.
     */
    snprintf(buf, sizeof(buf), "--%s--\r\n\r\n", boundary);
    result = networkWrite(socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    /**
     * And that's it!
     */
    er = S_OK;
    
  Error:
    if (types) localFree(types);
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * writeMultipartMixed
 *=------------------------------------------------------------------------=
 * writes out a message in multipart mixed format out.  Very complex.
 *
 * Parameters:
 *      int                         - [in]  socket fd.
 *      int                         - [in]  max line length
 *      MailMessage *               - [in]  write me out.
 *      const char *                - [in]  boundary separator
 *
 * Output:
 *      ERRCODE
 */
ERRCODE writeMultipartMixed
(
    int socketfd,
    int lineLength,
    MailMessage *message,
    const char *boundary
)
{
    LinkList<Attachment *> attachments;
    MIMETYPE *types = NULL, type;
    char *newBoundary = NULL;
    char buf[NET_MAXLINELEN];
    const char *body;
    Attachment *a;
    ERRCODE er;
    int result;
    int ct;

    /**
     * First, deal with the body(ies).;
     */
    message->getBodyFormats(&types, &ct);

    snprintf(buf, sizeof(buf), "--%s\r\n", boundary);
    result = networkWrite(socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    /**
     * If we've got more than one, it's multipart/alternative
     */
    if (ct > 1) {
        char bound[80];
        time_t t = time(NULL);
        snprintf(bound, sizeof(bound), "----=NextPart_001_%8ld=", t);
        newBoundary = copyString(bound);
        if (!newBoundary) { er = E_OUTOFMEMORY; goto Error; }
        type.setType(Multipart, Alternative);
        er = writeMimeHeader(socketfd, type, InvalidEncoding, Inline, NULL, NULL, newBoundary);
        GOTO_ON_FAILURE(er, Error);

        er = writeMultipartAlternative(socketfd, lineLength, message, newBoundary);
        GOTO_ON_FAILURE(er, Error);
    } else if (ct == 1) {

        /**
         * Otherwise, just write out a simple little part.
         */
        er = writeMimeHeader(socketfd, types[0], SevenBit, Inline, NULL, NULL, NULL);
        GOTO_ON_FAILURE(er, Error);

        body = NULL;
        message->getBodyRefByType(types[0], &body);
        er = encodeSection(socketfd, SevenBit, lineLength, body, -1);
        GOTO_ON_FAILURE(er, Error);
    }

    /**
     * great, now for each attachment, we need to go and write out another 
     * section.
     */
    message->getAttachments(attachments);
    er = attachments.firstElement(&a);
    while (er == S_OK) {

        const char *pp = NULL;
        MimeEncoding e;

        /**
         * write out the separator
         */
        snprintf(buf, sizeof(buf), "--%s\r\n", boundary);
        result = networkWrite(socketfd, buf, strlen(buf));
        if (result == -1) {
            errorSaveRichErrorInfo(strerror(errno));
            return E_NET_ERR;
        }
        
        /**
         * Then the MIME section header.
         */
        a->getFileName(&pp);
        a->getMimeType(type);
        a->getEncoding(e);
        er = writeMimeHeader(socketfd, type, e, MailAttachment, pp, NULL, NULL);
        if (pp) localFree(pp);
        GOTO_ON_FAILURE(er, Error);

        /**
         * Now write out the body
         */
        a->getEncoding(e);
        a->getBinaryDataRef((BYTE **)&pp);
        a->getBinarySize(ct);
        er = encodeSection(socketfd, e, lineLength, pp, ct);
        GOTO_ON_FAILURE(er, Error);

        /**
         * get the next element and continue
         */
        er = attachments.nextElement(&a);
    }

    /**
     * Finally, write a terminating boundary marker.
     */
    snprintf(buf, sizeof(buf), "--%s--\r\n\r\n", boundary);
    result = networkWrite(socketfd, buf, strlen(buf));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        er = E_NET_ERR;
        goto Error;
    }

    er = S_OK;

  Error:
    if (types) localFree(types);
    if (newBoundary) localFree(newBoundary);
    return er;
}


