/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

/**
 * This class is for viewing existing mail messages.  It is, annoying enough,
 * sufficiently different from the MailEditWindow that we split it out into
 * its own class.  d'oh!
 */
#include <sanity.h>

#include <ctype.h>
#include <qcheckbox.h>
#include <qdialog.h>
#include <qfiledialog.h>
#include <qiconview.h>
#include <qmessagebox.h>
#include <qobjectlist.h>
#include <qpopupmenu.h>
#include <qsplitter.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <stdlib.h>
#include <sys/param.h>


#include "about.dlg.h"
#include "findtext.h"
#include <intl.h>
#include "kwindowbase.h"
#include "localized.h"
#include "mailaccount.h"
#include "maileditwindow.h"
#include "mailfolder.h"
#include "mailmessage.h"
#include "mailviewer.h"
#include "main.h"
#include "selectme.h"
#include "superlabel.h"
#include "uiicons.h"
#include "uimenu.h"
#include "uitoolbar.h"
#include "userpref.h"


/**
 *=------------------------------------------------------------------------=
 * This is what we pass to KWindowBase, which does a lot of the tedious
 * stuff for us.
 *=------------------------------------------------------------------------=
 */
KWBVariables viewervars = {

    &prefViewerWindowLeft,
    &prefViewerWindowTop,
    &prefViewerWindowWidth,
    &prefViewerWindowHeight,
    &prefViewerToolBarVisible,
    &prefViewerToolBarUsesBig,
    &prefViewerToolBarShowLabels,
    &prefViewerToolBarLocation
};

/**
 *=------------------------------------------------------------------------=
 * This class is our sub-class of the QIconViewItem.  It basically just can 
 * hold an additional piece of data.
 *=------------------------------------------------------------------------=
 */
class MVIconViewItem : public QIconViewItem {

  public:
    MVIconViewItem(QIconView *parent, const char *text, QPixmap &pixmap, Attachment *a)
     : QIconViewItem(parent, text, pixmap) {
        // UNDONE: marcw, 11.14.2000 -- i don't want to addref this here, but it might
        // be dangerous.  need to think about it more.
        this->attachment = a;
    }

    virtual ~MVIconViewItem() { }
    
    Attachment *getAttachment() { return attachment; }

  private:
    Attachment *attachment;
};



/**
 *=------------------------------------------------------------------------=
 * Stolen from maileditwindow.cc
 */
void stripIntoBuf(char *dest, const char *source, int max);    



/**
 *=------------------------------------------------------------------------=
 * local prototypes
 *=------------------------------------------------------------------------=
 */
QLabel     *createLabel(QWidget *parent, STRINGID id, QFont *font = NULL);
SuperLabel *createAddressLabel(QWidget *parent);
ERRCODE     dumpAddressesIntoLabel(LinkList<const char *> &, SuperLabel *);








/**
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 *                      MailViewer Class
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 */ 






/**
 *=------------------------------------------------------------------------=
 * MailViewer::MailViewer
 *=------------------------------------------------------------------------=
 * Creates a new blank instance of a viewer.  You still need to call 
 * initialize() on it once you've created it.
 *
 * Parameters:
 *      MailFolder *           - [in]  folder in which we are operating.
 */
MailViewer::MailViewer
(
    MailFolder *fldr
)
: KWindowBase(viewervars, WDestructiveClose)
{
    K_ASSERT(fldr, "You're supposed to pass a folder to this winodw!!!");

    folder = fldr;
    message = NULL;

    editBox = NULL;
    attachmentPopup = NULL;
    toLabel = fromLabel = ccLabel = subjectLabel = NULL;
    fromValue = ccValue = toValue = NULL;
    dateValue = subjectValue = NULL;

    viewHtmlButton = NULL;
    
    findText = NULL;

    DEBUG_TEST_START;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::~MailViewer
 *=------------------------------------------------------------------------=
 * Destroys an instance of this class.
 */
MailViewer::~MailViewer()
{
    Attachment *a;
    ERRCODE er;

    if (editBox) delete editBox;
    if (toLabel) delete toLabel;
    if (fromLabel) delete fromLabel;
    if (ccLabel) delete ccLabel;
    if (subjectLabel) delete subjectLabel;
    if (toValue) delete toValue;
    if (fromValue) delete fromValue;
    if (ccValue) delete ccValue;

    if (message) message->release();

    er = attachments.firstElement(&a);
    while (er == S_OK) {
        attachments.removeElement(a);
        a->release();
        er = attachments.firstElement(&a);
    }

    if (attachmentPopup) delete attachmentPopup;

    if (findText) delete findText;

    DEBUG_TEST_STOP;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::aboutToShowFileMenu
 *=------------------------------------------------------------------------=
 * we're about to show the file menu.  We need to disable the attachments
 * options if the message has no attachments.
 */
void MailViewer::aboutToShowFileMenu()
{
    bool f = false;
    
    if (message) message->getHasAttachments(&f);
    menuBar()->setItemEnabled(ID_FILE_SAVEATTACH_TEXT, f);
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::aboutToShowEditMenu
 *=------------------------------------------------------------------------=
 * we're about to show the edit menu.  We need to make sure that cut and
 * paste are disabled.
 */
void MailViewer::aboutToShowEditMenu()
{
    menuBar()->setItemEnabled(ID_EDIT_CUT_TEXT, false);
    menuBar()->setItemEnabled(ID_EDIT_PASTE_TEXT, false);
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::aboutToShow
 *=------------------------------------------------------------------------=
 */
void MailViewer::aboutToShowViewMenu()
{
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::addAttachmentToDisplay
 *=------------------------------------------------------------------------=
 * given an attachment, adds it to our UI.
 *
 * Parameters:
 *      Attachment *                - [in]  add me.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailViewer::addAttachmentToDisplay
(
    Attachment *attach
)
{
    const char *pstr;
    QPixmap *pixmap;
    char buf[128];
    ERRCODE er;

    /**
     * First, pick a pixmap for the attachment.
     */
    pixmap = attach->chooseIcon();
    if (!pixmap) return E_OUTOFMEMORY;

    /**
     * Next, extract the basic file name.
     */
    er = attach->getFileName(&pstr);
    RETURN_ON_FAILURE(er);

    strncpy(buf, pstr, sizeof(buf));
    unsigned int i = strlen(pstr);
    localFree(pstr);
    i = (i >= sizeof(buf)) ? sizeof(buf) - 1 : i;
    buf[i] = '\0';
    strncat(buf, attach->getPrintableSize(), sizeof(buf) - (i + 1));
    
    attachmentView->show();
    new MVIconViewItem(attachmentView, buf, *pixmap, attach);
    delete pixmap;

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::createdPopup
 *=------------------------------------------------------------------------=
 * our parent class created a popup menu.  We override this so we can 
 * correctly ssync up on it's aboutToShow() signal.
 *
 * Parameters:
 *      int                 - [in]  id of the popup created.
 *      QPopupMenu *        - [in]  menu created.
 */
void MailViewer::createdPopup
(
    int id,
    QPopupMenu *popup
)
{
    switch (id) {
        case ID_FILE_MENU_TEXT:
            QObject::connect(popup, SIGNAL(aboutToShow()),
                             this, SLOT(aboutToShowFileMenu()));
            break;
        case ID_EDIT_MENU_TEXT:
            QObject::connect(popup, SIGNAL(aboutToShow()),
                             this, SLOT(aboutToShowEditMenu()));
            break;
        case ID_VIEW_MENU_TEXT:
            QObject::connect(popup, SIGNAL(aboutToShow()),
                             this, SLOT(aboutToShowViewMenu()));
            break;
        default:
            break;
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::doubleClickedIV
 *=------------------------------------------------------------------------=
 * the user has double clicked in the icon view.  see if we need to do
 * something with an item.
 *
 * Parameters:
 *      QIconViewItem *             - [in]  item clicked, or NULL.
 */
void MailViewer::doubleClickedIV
(
    QIconViewItem *item
)
{
    if (!item) return;

    /**
     * First, figure out what program to use to view this particular item.
     */
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::editCopy
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::editCopy()
{

}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::editCut
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::editCut()
{

}



/**
 *=------------------------------------------------------------------------=
 * MailViewer::editFindMessage
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::editFindMessage()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::editFindPerson
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::editFindPerson()
{

}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::editFindText
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::editFindText()
{
    K_ASSERT(!this->findText, "Somebody is leaving turds around!!!!");

    /**
     * Create the dialog!!
     */
    this->findText = new FindText(this, "findText", true, 0);
    if (!this->findText) {
        errorShowError(E_OUTOFMEMORY);
        return;
    }

    /**
     * Hook up our slot and show it.
     */
    findText->hookUpFindButton(this, SLOT(searchForText()));
    findText->exec();

    /**
     * Unhook and clean up!
     */
    findText->unhookFindButton(this, SLOT(searchForText()));
    delete findText;
    findText = NULL;
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::editPaste
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::editPaste()
{

}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::editSelectAll
 *=------------------------------------------------------------------------=
 * The user has asked to select all the text in the message.  Oblige them.
 */
void MailViewer::editSelectAll()
{
    editBox->selectAll();
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::fileClose
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::fileClose()
{
    this->close(true);
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::fileNew
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::fileNew()
{
    ERRCODE er;

    MailEditWindow *wnd;
     
    er = MailEditWindow::composeNew(this->folder->getAccount(), &wnd);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }

    wnd->show();
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::filePrint
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::filePrint()
{

}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::fileSave
 *=------------------------------------------------------------------------=
 * Lets the user save the message to disk.  Asks them for the name first.
 */
void MailViewer::fileSave()
{
    char buf[MAXPATHLEN];
    const char *pstr;
    QString file;
    ERRCODE er;
    char *ptr;
    int len;

    /**
     * We'll just use their home directory as the starting point.
     */
    ptr = getenv("HOME");
    if (ptr)
        strcpy(buf, ptr);
    else
        buf[0] = '\0';

    len = strlen(buf);
    buf[len++] = '/';
    buf[len] = '\0';

    /**
     * Figure out if we can just use the subject, or if we need to go and
     * use a generic file name.
     */
    if (!subjectValue->text().isEmpty()) {
        stripIntoBuf(buf + len, (const char *)subjectValue->text(), 
                     sizeof(buf) - (len + 1));
    } else {
        pstr = intlLoadString(ID_SAVE_MESSAGE);
        if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
        strcat(buf, pstr);
        localFree(pstr);
    }

    /**
     * Load in the filter.
     */
    pstr = intlLoadString(ID_MSG_SAVE_FILTER);
    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }

    /**
     * Now get the filename
     */
    file = QFileDialog::getSaveFileName(buf, pstr);
    localFree(pstr);

    if (!file.isEmpty()) {
        K_FAIL("UNDONE: 11.1.2000 -- save me now dumbass!!!!");
        fprintf(stderr, "saving file: %s\n", (const char *)file);
    }

    return;

  Error:
    errorShowError(er);
    return;   
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::fileSaveAttach
 *=------------------------------------------------------------------------=
 * provides some UI to save all the attachments in the message.
 *
 */
void MailViewer::fileSaveAttach()
{
    LinkList<Attachment *> whichOnes;
    const char *pstr = NULL;
    Attachment *a;
    ERRCODE er;

    /** 
     * First, show a dialog seeing which ones they'd like to save.  if we get
     * S_FALSE back, they changed their minds.  Then we'll ask WHERE they'd
     * like to save them.
     */
    er = selectAttachments(whichOnes);
    GOTO_ON_FAILURE(er, Error);
    if (er == S_FALSE) return;

    er = selectAttachDirectory(&pstr);
    GOTO_ON_FAILURE(er, Error);
    if (er == S_FALSE) return;

    /**
     * now save 'em out one at a time.
     */
    er = whichOnes.firstElement(&a);
    while (er == S_OK) {
        ERRCODE er2 = message->saveAttachment(a, pstr);
        GOTO_ON_FAILURE(er2, Error);

        er = whichOnes.nextElement(&a);
    }
    
    return;

  Error:
    if (pstr) localFree(pstr);
    errorShowError(er);
    return;
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::fileSend
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::fileSend()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::getHtmlButton
 *=------------------------------------------------------------------------=
 * gets the toolbarbutton corresponding to the view HTML button.  It's a bit
 * of a pain to get, so we'll cache it as much as we can.
 *
 * Output:
 *      QToolButton *
 */
QToolButton *MailViewer::getHtmlButton()
{
    QObjectList *list = (QObjectList *)localToolBar->children();
    const char *pstr = NULL, *pstrH = NULL;
    QToolButton *button;
    QObject *item;

    if (viewHtmlButton) return viewHtmlButton;

    pstrH = intlLoadString(ID_VIEW_HTML_TEXT);

    item = list->first();
    while (item) {

        if ((button = dynamic_cast<QToolButton *>(item)) != NULL) {
            pstr = button->textLabel();
            if (!strcasecmp(pstr, pstrH)) {
                viewHtmlButton = button;
                localFree(pstrH);
                return button;
            }
        }
        item = list->next();
    }

    localFree(pstrH);
    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::getIconViewSelectionCount
 *=------------------------------------------------------------------------=
 * returns the number of items selected in the icon view.
 *
 * Output:
 *      int                         - 0 or more.
 */
int MailViewer::getIconViewSelectionCount()
{
    QIconViewItem *item;
    int count = 0;

    K_ASSERT(attachmentView, "I should have this by nowww!");

    item = attachmentView->firstItem();

    while (item) {
        if (item->isSelected()) count++;

        item = item->nextItem();
    }

    return count;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::helpAbout
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::helpAbout()
{
    AboutDialog ad(this, "ad", true, 0);
    ad.exec();
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::helpHelp
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::helpHelp()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::helpHelpMe
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::helpHelpMe()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::initialize
 *=------------------------------------------------------------------------=
 * Initializes our UI and starts pulling things from the MailMessage object
 * that we've got as a victim .
 *
 * Parameters:
 *      MailMessage *           - [in]  where to get data from.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailViewer::initialize
(
    MailMessage *msg
)
{
    const char *pstr;
    ERRCODE er;
    QFont font;
    int i;

    if (!msg) return E_PARAM;

    /**
     * Figure out how to scale the font.
     */
    font = this->font();
    if (prefFontScale > 0) {
        font.setPointSize((int)((double)font.pointSize() * 1.2) + 2);
        this->setFont(font);
    } else if (prefFontScale < 0) {
        font.setPointSize((int)((double)font.pointSize() * 0.75) + 2);
        this->setFont(font);
    }

    /**
     * Initialize our toolbars and menus
     */
    er = this->initializeUI(this, mailViewWindowMenu, viewerToolBarData);
    RETURN_ON_FAILURE(er);

    this->setIcon(QPixmap(editwindowIcon));

    font = this->font();
    font.setBold(true);

    /**
     * great, now start setting up our own specific stuff!!!
     */
    fromLabel = createLabel(this, ID_FROM_TEXT, &font);
    if (!fromLabel) return E_OUTOFMEMORY;
    fromValue = createAddressLabel(this);
    if (!fromValue) return E_OUTOFMEMORY;

    dateLabel = createLabel(this, ID_SENT_TEXT, &font);
    if (!dateLabel) return E_OUTOFMEMORY;
    dateValue = new QLabel(this);
    if (!dateValue) return E_OUTOFMEMORY;

    toLabel = createLabel(this, ID_TO_TEXT, &font);
    if (!toLabel) return E_OUTOFMEMORY;

    toValue = new SuperLabel(this);
    if (!toValue) return E_OUTOFMEMORY;
    

    ccLabel = createLabel(this, ID_CC_TEXT, &font);
    if (!ccLabel) return E_OUTOFMEMORY;
    ccValue = createAddressLabel(this);
    if (!ccValue) return E_OUTOFMEMORY;

    subjectLabel = createLabel(this, ID_SUBJECT_TEXT);
    if (!subjectLabel) return E_OUTOFMEMORY;
    subjectValue = new QLabel(this);
    if (!subjectValue) return E_OUTOFMEMORY;

    /**
     * Priority label
     */
    priorityLabel = new QLabel(this);
    if (!priorityLabel) return E_OUTOFMEMORY;
    priorityLabel->setGeometry(200, 200, 1, 1);
    priorityLabel->setBackgroundColor(QColor(224, 228,177));
    priorityPixLabel = new QLabel(this);
    if (!priorityPixLabel) return E_OUTOFMEMORY;
    priorityPixLabel->setBackgroundColor(QColor(224, 228,177));
    priorityLabel->hide();
    priorityPixLabel->hide();

    /**
     * the edit box will share a splitter with the attachment view.
     */
    this->contentSplitter = new QSplitter(Qt::Vertical, this);
    if (!contentSplitter) return E_OUTOFMEMORY;

    /**
     * the edit box to display all the content.
     */
    this->editBox = new MyMultiLineEdit(contentSplitter);
    editBox->setWordWrap(MyMultiLineEdit::WidgetWidth);
    editBox->setReadOnly(true);
    editBox->show();
    this->folder->getAccount()->getReadFont(pstr, i);
    if (pstr) {
        editBox->setFont(QFont(pstr, i));
        localFree(pstr);
    }

    /**
     * create our attachment viewer
     */
    attachmentView = new QIconView(contentSplitter);
    if (!attachmentView) return E_OUTOFMEMORY;
    QObject::connect(attachmentView, SIGNAL(rightButtonPressed(QIconViewItem *, const QPoint &)),
                     this, SLOT(rightIVButtonPressed(QIconViewItem *, const QPoint &)));
    QObject::connect(attachmentView, SIGNAL(doubleClicked(QIconViewItem *)),
                     this, SLOT(doubleClickedIV(QIconViewItem *)));
    QObject::connect(attachmentView, SIGNAL(returnPressed(QIconViewItem *)),
                     this, SLOT(doubleClickedIV(QIconViewItem *)));
    attachmentView->setItemTextPos(QIconView::Right);
    attachmentView->setAutoArrange(true);
    attachmentView->setResizeMode(QIconView::Adjust);
    attachmentView->setSelectionMode(QIconView::Extended);
    attachmentView->setMinimumSize(20, 50);
    attachmentView->hide();

    contentSplitter->setResizeMode(attachmentView, QSplitter::KeepSize);

    /**
     * Set up our popup menus for the attachment view.
     */
    attachmentPopup = new QPopupMenu();
    if (!attachmentPopup) return E_OUTOFMEMORY;

    /**
     * We've got to add some items to the menu now.
     */
    pstr = intlLoadString(ID_OPEN);
    if (!pstr) return E_OUTOFMEMORY;
    attachmentPopup->insertItem(pstr, this, SLOT(openAttachment()), 0, ID_OPEN);
    localFree(pstr);

    pstr = intlLoadString(ID_SAVE_AS);
    if (!pstr) return E_OUTOFMEMORY;
    attachmentPopup->insertItem(QIconSet(*(new QPixmap(filesaveIcon))), 
                                pstr, this, SLOT(saveAttachment()), 0, ID_SAVE_AS);
    localFree(pstr);

    pstr = intlLoadString(ID_SELECT_ALL);
    if (!pstr) return E_OUTOFMEMORY;
    attachmentPopup->insertItem(pstr, this, SLOT(selectAllAttachments()), 0, ID_SELECT_ALL);
    localFree(pstr);

    /**
     * And finally, populate the data from the mail message.
     */
    er = populateFromMailMessage(msg);
    RETURN_ON_FAILURE(er);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::keyPressEvent
 *=------------------------------------------------------------------------=
 * a key has been pressed.  Wahooo!
 *
 * Parameters:
 *      QKeyEvent *                 - [in]  the event.
 */
void MailViewer::keyPressEvent
(
    QKeyEvent *event
)
{
    return KWindowBase::keyPressEvent(event);
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::layoutComponents
 *=------------------------------------------------------------------------=
 * positions everything correctly in the viewer window.
 */
void MailViewer::layoutComponents()
{
    QWidget *widgetArray[] = {
        fromLabel,
        toLabel,
        ccLabel,
        dateLabel,
        subjectLabel
    };
    int ctlHeight, ctlWidth, ctlHeight2;
    MessagePriority p;
    const char *pstr;
    QRect rect;
    int top;
    int z;

    /**
     * Get/compute some sizes
     */
    ctlHeight = fromLabel->font().pixelSize() + 6;  // 2 x 4 for over/under-hang
    ctlWidth = getMaxWidth(5, widgetArray);
    getClientRect(rect);

    top = rect.top() + 3;
    z = rect.width() - (ctlWidth + 19);

    message->getPriority(p);
    switch (p) {
        case NormalPriority:
            priorityPixLabel->hide();
            priorityLabel->hide();
            break;

        case HighPriority:
        case LowPriority:
            pstr = intlLoadString(p == LowPriority 
                                  ? ID_LOW_PRIORITY_TEXT 
                                  : ID_HIGH_PRIORITY_TEXT);
            if (pstr) {
                QPixmap p(p == LowPriority ? lowpriIcon : priorityhdrIcon);
                priorityPixLabel->setPixmap(p);
                priorityPixLabel->setGeometry(4, top, p.width() + 1, ctlHeight);
                priorityLabel->setText(pstr);
                localFree(pstr);
                priorityLabel->setGeometry(p.width() + 5, top,
                                           this->width() - p.width() - 13,
                                           ctlHeight);
                top += ctlHeight + 5;
                priorityPixLabel->show();
                priorityLabel->show();
            } else {
                priorityPixLabel->hide();
                priorityLabel->hide();
            }
            break;
    }

    fromLabel->setGeometry(8, top, ctlWidth, ctlHeight);
    ctlHeight2 = fromValue->getHeightHintForWidth(z);
    if (ctlHeight2 < ctlHeight) ctlHeight2 = ctlHeight;
    fromValue->setGeometry(ctlWidth + 11, top, z, ctlHeight2);
    top += ctlHeight2 + 3;

    dateLabel->setGeometry(8, top, ctlWidth, ctlHeight);
    ctlHeight2 = dateValue->sizeHint().height();
    if (ctlHeight2 < ctlHeight) ctlHeight2 = ctlHeight;
    dateValue->setGeometry(ctlWidth + 11, top, z, ctlHeight2);
    top += ctlHeight + 3;

    if (!toLabel->isHidden()) {
        toLabel->setGeometry(8, top, ctlWidth, ctlHeight);
        ctlHeight2 = toValue->getHeightHintForWidth(z);
        if (ctlHeight2 < ctlHeight) ctlHeight2 = ctlHeight;
        toValue->setGeometry(ctlWidth + 11, top, z, ctlHeight2);
        top += ctlHeight2 + 3;
    }

    if (!ccLabel->isHidden()) {
        ccLabel->setGeometry(8, top, ctlWidth, ctlHeight);
        ctlHeight2 = ccValue->getHeightHintForWidth(z);
        if (ctlHeight2 < ctlHeight) ctlHeight2 = ctlHeight;
        ccValue->setGeometry(ctlWidth + 11, top, z, ctlHeight2);
        top += ctlHeight2 + 3;
    }

    subjectLabel->setGeometry(8, top, ctlWidth, ctlHeight);
    ctlHeight2 = subjectValue->sizeHint().height();
    if (ctlHeight2 < ctlHeight) ctlHeight2 = ctlHeight;
    subjectValue->setGeometry(ctlWidth + 11, top, z, ctlHeight2);
    top += ctlHeight2 + 5;

    contentSplitter->setGeometry(3, top, rect.width() - 2, this->height() - top - 2);
}
 

/**
 *=------------------------------------------------------------------------=
 * MailViewer::messageDelete
 *=------------------------------------------------------------------------=
 * The user wishes to delete this message.  If we can we go an select the 
 * next or previous one (user preference).
 */
void MailViewer::messageDelete()
{
    MailMessage *next;
    MailFolder *ob;
    bool f;

    this->folder->getAccount()->getSelectPreviousOnDelete(f);

    if (f)
        next = this->folder->previousMessage(this->message);
    else
        next = this->folder->nextMessage(this->message);

    /** 
     * Make sure put this message in the deleted items first.
     */
    ob = this->folder->getAccount()->getDeletedItems();
    if (this->folder != ob) {
        ERRCODE er = ob->addMessage(this->message);
        if (FAILED(er)) {
            errorShowError(er);
            return;
        }
    }

    /**
     * Now remove the message and select the appropriate next message.
     */
    if (this->folder->deleteMessages(1, &this->message, true) == S_FALSE) 
        return;

    if (next) {
        this->message->release();
        this->message = NULL;
        this->populateFromMailMessage(next);
        folder->selectMessage(next);
    } else {
        this->close(true);
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::messageForward
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::messageForward()
{
    MailEditWindow *wnd;
    ERRCODE er;
    bool b;

    er = MailEditWindow::forward(this->message, this->folder->getAccount(), &wnd);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }
    wnd->show();

    /**
     * tell the folder that this message has been replied to.
     */
    this->folder->markMessageAsForwarded(this->message);

    /**
     * Figure out if we need to close this or not.
     */
    this->folder->getAccount()->getCloseOnReply(b);
    if (b) fileClose();
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::messageReply
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::messageReply()
{
    MailEditWindow *wnd;
    ERRCODE er;
    bool b;

    er = MailEditWindow::reply(this->message, this->folder->getAccount(), false, &wnd);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }
    wnd->show();

    /**
     * Now figure out if we need to close this window or not.
     */
    this->folder->getAccount()->getCloseOnReply(b);
    if (b) fileClose();
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::messageReplyAll
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::messageReplyAll()
{
    MailEditWindow *wnd;
    ERRCODE er;
    bool b;

    er = MailEditWindow::reply(this->message, this->folder->getAccount(), true, &wnd);
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }
    wnd->show();

    /**
     * Now figure out if we need to close this window or not.
     */
    this->folder->getAccount()->getCloseOnReply(b);
    if (b) fileClose();
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::openAttachment
 *=------------------------------------------------------------------------=
 * The user wishes to open the currently selected attachment.
 */
void MailViewer::openAttachment()
{
    doubleClickedIV(attachmentView->currentItem());
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::populateFromMailMessage
 *=------------------------------------------------------------------------=
 * given a mail message, go and set up all our data for this.
 *
 * Parameters:
 *      MailMessage *               - [in]  the message to work with.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailViewer::populateFromMailMessage
(
    MailMessage *msg
)
{
    LinkList<const char *> strList;
    LinkList<Attachment *> list;
    QToolButton *button;
    const char *pstr;
    MIMETYPE type;
    Attachment *a;
    ERRCODE er;

    /**
     * Keep it around in case the host deletes it and loses a ref.
     */
    this->message = msg;
    msg->addRef();

    /**
     * from whom is the message?
     */
    er = message->getSenders(strList);
    RETURN_ON_FAILURE(er);
    dumpAddressesIntoLabel(strList, fromValue);

    /**
     * When did we get the message?
     */
    er = message->getSent(&pstr);
    RETURN_ON_FAILURE(er);
    if (pstr) {
        dateValue->setText(pstr);
        localFree(pstr);
    } else {
        dateValue->setText("");
    }

    /**
     * To whom was this message sent?
     */
    er = message->getRecipients(strList);
    RETURN_ON_FAILURE(er);
    dumpAddressesIntoLabel(strList, toValue);

    if (!strList.isEmpty()) {
        toLabel->show();
        toValue->show();
    } else {
        toLabel->hide();
        toValue->hide();
    }
    strList.emptyList();

    /**
     * CCs;
     */
    er = message->getCCs(strList);
    RETURN_ON_FAILURE(er);
    dumpAddressesIntoLabel(strList, ccValue);

    if (!strList.isEmpty()) {
        ccLabel->show();
        ccValue->show();
    } else {
        ccLabel->hide();
        ccValue->hide();
    }
    strList.emptyList();

    /**
     * Message Subject??
     */
    er = message->getSubject(&pstr);
    RETURN_ON_FAILURE(er);
    if (pstr) {
        subjectValue->setText(pstr);
        this->setCaption(pstr);
        localFree(pstr);
    } else {
        subjectValue->setText("");
    }

    /**
     * get the body and show it.  If they don't have a plain text body, return
     * the HTML body.  If there are no bodies -- that's weird.
     */
    type.setType(Text, Plain);
    er = message->getBodyByType(type, &pstr, NULL);
    if (er == E_NOTFOUND) {
        type.setType(Text, Html);
        er = message->getBodyByType(type, &pstr, NULL);
        if (er == E_NOTFOUND) {
            pstr = intlLoadString(ID_NO_BODY);
            if (!pstr) return E_OUTOFMEMORY;
        } else {
            RETURN_ON_FAILURE(er);
        }
    } else {
        RETURN_ON_FAILURE(er);
    }
    this->editBox->setText(pstr);
    localFree(pstr);

    /**
     * Finally, set up any attachments.
     */
    attachmentView->clear();
    attachmentView->hide();
    er = message->getAttachments(list);
    RETURN_ON_FAILURE(er);

    er = list.firstElement(&a);
    while (er == S_OK) {
        er = addAttachmentToDisplay(a);
        RETURN_ON_FAILURE(er);

        er = list.nextElement(&a);
    }

    layoutComponents();

    /**
     * IF the message has an HTML format, then we need to go and set the
     * toolbar button up so that people can view it!!!
     */
    type.setType(Text, Html);
    er = message->getBodyRefByType(type, &pstr, NULL);
    button = getHtmlButton();
    if (button) button->setEnabled(!FAILED(er));

    editBox->setFocus();
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::rightIVButtonPressed
 *=------------------------------------------------------------------------=
 * the user has right clicked inside the iconview for attachments.  show
 * a popupmenu.
 *
 * Parameters:
 *      QIconViewItem *         - [in]  item clicked on or NULL.
 *      const QPoint &          - [in]  WHERE they clicked.
 */
void MailViewer::rightIVButtonPressed
(
    QIconViewItem *item,
    const QPoint &loc
)
{
    if (!attachmentPopup) { K_FAIL("I should have a popup by now !!!!"); return; }

    /**
     * Enable or disable items, depending on what's selected.
     */
    if (item) {
        attachmentPopup->setItemEnabled(ID_OPEN, getIconViewSelectionCount() == 1);
        attachmentPopup->setItemEnabled(ID_SAVE_AS, getIconViewSelectionCount() == 1);
        attachmentPopup->setItemEnabled(ID_SAVE_ALL, false);
    } else {
        attachmentPopup->setItemEnabled(ID_OPEN, false);
        attachmentPopup->setItemEnabled(ID_SAVE_AS, false);
        attachmentPopup->setItemEnabled(ID_SAVE_ALL, true);
    }

    attachmentPopup->popup(loc);
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::saveAttachment
 *=------------------------------------------------------------------------=
 * the user wishes to save the selected Attachment.  We can handle that.
 */
void MailViewer::saveAttachment()
{
    QString file, filter, caption;
    MVIconViewItem *item;
    Attachment *a = NULL;
    const char *pstr;
    ERRCODE er;

    /**
     * Load in some strings.
     */
    pstr = intlLoadString(ID_SAVE_ATTACHMENT);
    if (!pstr) return;
    caption = pstr;
    localFree(pstr);

    /**
     * if there is no directory for this yet, set it to the users's home
     * dir.
     */
    if (!prefAttachmentDir) {
        char *ptr;
        ptr = getenv("HOME");
        if (ptr) {
            prefAttachmentDir = copyString(ptr);
        }
    }
    
    file = prefAttachmentDir;
    item = (MVIconViewItem *)attachmentView->currentItem();
    if (item) {
        a = item->getAttachment();
        K_ASSERT(a, "Item with no attachment !!!!!");
        a->getFileName(&pstr);
        if (pstr)  {
            file += "/";
            file += pstr;
            localFree(pstr);
        }
    }

    /**
     * filter will just be "All Files"
     */
    pstr = intlLoadString(ID_INSERT_FILTER);
    if (pstr) {
        filter = pstr;
        localFree(pstr);
    }

    /**
     * Show the dialog
     */
    file = QFileDialog::getSaveFileName(file, filter, this, "save", caption);

    /**
     * If they selected a file, save it
     */
    if (!file.isEmpty()) {
        er = a->saveToFile((const char *)file);
        if (FAILED(er))
            errorShowError(er);

        /**
         * extract the directory they were last in so we can remember
         * it for them.
         */
        int index = file.findRev('/', -1);
        if (index > 0) {
            file.truncate(index);
            if (prefAttachmentDir) localFree(prefAttachmentDir);
            prefAttachmentDir = copyString((const char *)file);
        }
    }
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::searchForText
 *=------------------------------------------------------------------------=
 * somebody is using the findText dialog and is looking for text.  do the
 * search now.
 */
void MailViewer::searchForText()
{
    bool forwards, caseSensitive;
    const char *text, *title;
    QString line, findMe;
    int row = -1, index;
    int curRow, curCol;

    K_ASSERT(findText, "How on earth is this getting called??");

    /**
     * whiz through the document looking for this string.  we do not go 
     * through the document more than once.
     */
    caseSensitive = findText->caseSensitiveC->isChecked();
    forwards = !findText->searchBackwardsC->isChecked();
    this->editBox->getCursorPosition(&curRow, &curCol);
    findMe = findText->searchForE->text();

    /**
     * If there's another match in this line, it's a bit easier.
     */
    line = editBox->textLine(curRow);
    if ((index = line.find(findMe, curCol, caseSensitive)) != -1
        && (forwards || (index < curCol))) {
        editBox->deselect();
        editBox->setCursorPosition(curRow, index, false);
        editBox->setCursorPosition(curRow, index + findMe.length(), true);
    } else {
        /**
         * Otherwise, just go whizzing through all the other lines. We will
         * not go through the whole edit more than once.
         */
        if (forwards) {
            row = curRow + 1;
            if (row >= editBox->numLines()) row = 0;
        } else {
            row = curRow - 1;
            if (row < 0) row = editBox->numLines() - 1;
        }

        while (row != curRow) {
            line = editBox->textLine(row);
            if ((index = line.find(findMe, 0, caseSensitive)) != -1) {
                editBox->deselect();
                editBox->setCursorPosition(row, index, false);
                editBox->setCursorPosition(row, index + findMe.length(), true);
                break;
            }

            if (forwards) {
                row++;
                if (row >= editBox->numLines()) row = 0;
            } else {
                row--;
                if (row < 0) row = editBox->numLines() - 1;
            }
        }

        /**
         * Didn't find it -- show a dialog.
         */
        if (row == curRow) {
            text = intlLoadStringX(ID_TEXT_NOT_FOUND);
            title = intlLoadStringX(ID_KILTDOWN);
            QMessageBox::warning((QWidget *)this, QString(title), QString(text), QMessageBox::Ok, 0, 0);
        }
    }

}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::selectAttachDirectory
 *=------------------------------------------------------------------------=
 * Asks the user what directory they'd like to save their attachments in.
 *
 * Parameters:
 *      const char **           - [out] allocated directory.  localFree() me.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailViewer::selectAttachDirectory
(
    const char **pstr
)
{
    const char *p;
    QString dir;

    if (!pstr) return E_PARAM;

    p = intlLoadString(ID_SELECT_DIRECTORY);
    if (!p) return E_OUTOFMEMORY;

    dir = QFileDialog::getExistingDirectory(getenv("HOME"), 0, 0, p, true);
    localFree(p);
    if (!dir.isEmpty()) {
        *pstr = copyString((const char *)dir);
        return *pstr ? S_OK : E_OUTOFMEMORY;
    } else {
        *pstr = NULL;
        return S_FALSE;
    }

    // deed code
    return E_INTERNAL;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::selectAllAttachments
 *=------------------------------------------------------------------------=
 * selects all the attachments in the icon view.
 */
void MailViewer::selectAllAttachments()
{
    attachmentView->selectAll(true);
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::selectAttachments
 *=------------------------------------------------------------------------=
 * pops up a little dialog asking the user which attachments they'd like to
 * save out.  
 *
 * Parameters:
 *      LinkList<Attachment *> &        - [in]  we'll put a list of the ones
 *                                          they've selected in here.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailViewer::selectAttachments
(
    LinkList<Attachment *> &whichOnes
)
{
    LinkList<unsigned int> list;
    const char *pstr, *pstr2;
    SelectMeDialog *smd;
    unsigned int ui;
    Attachment *a;
    ERRCODE er;
    int ret;

    smd = new SelectMeDialog();
    if (!smd) return E_OUTOFMEMORY;

    pstr = intlLoadString(ID_SMD_TITLE);
    if (!pstr) return E_OUTOFMEMORY;
    pstr2 = intlLoadString(ID_SMD_INTRO);
    if (!pstr2) {
        localFree(pstr);
        return E_OUTOFMEMORY;
    }

    er = smd->initializeUI(pstr, pstr2);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Now add everything to the dialog.
     */
    er = this->attachments.firstElement(&a);
    while (er == S_OK) {
        a->getFileName(&pstr);
        if (pstr) {
            smd->addItem(pstr, (unsigned int)a);
            localFree(pstr);
        }
        er = this->attachments.nextElement(&a);
    }

    ret = smd->exec();

    /**
     * If they selected OK, then get the list of files.
     */
    if (ret == QDialog::Accepted) {
        er = smd->getSelectedItems(list);
        GOTO_ON_FAILURE(er, Error);

        er = list.firstElement(&ui);
        while (er == S_OK) {
            whichOnes.addElement((Attachment *)ui);
            er = list.nextElement(&ui);
        }

        er = S_OK;
    } else {
        er = S_FALSE;
    }

  Error:
    if (smd) delete smd;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::toolsAddressBook
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::toolsAddressBook()
{

}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::toolsAddToAddressBook
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::toolsAddToAddressBook()
{

}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::viewAllHeaders
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::viewAllHeaders()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::viewHtmlContent
 *=------------------------------------------------------------------------=
 * if this message has HTML content, offer to bring up a browser to view
 * it.
 */
void MailViewer::viewHtmlContent()
{

}

/**
 *=------------------------------------------------------------------------=
 * MailViewer::viewNext
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::viewNext()
{
    MailMessage *next;

    next = this->folder->nextMessage(this->message);

    if (next) {
        this->message->release();
        this->message = NULL;
        this->populateFromMailMessage(next);
        folder->selectMessage(next);
        folder->updateMessageStatus(next, MSG_READ);
    }
}


/**
 *=------------------------------------------------------------------------=
 * MailViewer::viewPrevious
 *=------------------------------------------------------------------------=
 * responds to the  menu/toolbar event.
 */
void MailViewer::viewPrevious()
{
    MailMessage *prev;

    prev = this->folder->previousMessage(this->message);
    if (prev) {
        this->message->release();
        this->message = NULL;
        this->populateFromMailMessage(prev);
        folder->selectMessage(prev);
        folder->updateMessageStatus(prev, MSG_READ);
    }
}





/**
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 *                          Local Routines
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 *=------------------------------------------------------------------------=
 */

/**
 *=------------------------------------------------------------------------=
 * createAddressLabel
 *=------------------------------------------------------------------------=
 * goes and creates an address label with the given parent.
 *
 * Parameters: 
 *      QWidget *               - [in]  parent
 *
 * Output:
 *      SuperLabel *            - the new dude.
 */
SuperLabel *createAddressLabel
(
    QWidget *parent
)
{
    return new SuperLabel(parent);
}


/**
 *=------------------------------------------------------------------------=
 * createLabel
 *=------------------------------------------------------------------------=
 * given a string id, create the label for it with that text.
 *
 * Parameters:
 *      QWidget *           - [in]  parent widget
 *      STRINGID            - [in]  string id to load in
 *      QFont *             - [in]  OPTIONAL -- font to set
 *
 * Output:
 *      QLabel *            - new label created, or NULL if OOM
 */
QLabel *createLabel
(
    QWidget *parent,
    STRINGID id,
    QFont *font
)
{
    const char *pstr;
    QLabel *lbl;

    /**
     * Load in the string first.
     */
    pstr = intlLoadStringX(id);
    if (!pstr) return NULL;

    /**
     * create the label!
     */
    lbl = new QLabel(parent);
    if (!lbl) return NULL;

    lbl->setText(pstr);
    lbl->setGeometry(200, 200, 1, 1);
    lbl->setAutoResize(true);
    if (font) lbl->setFont(*font);

    return lbl;
}

/**
 *=------------------------------------------------------------------------=
 * dumpAddressesIntoLabel
 *=------------------------------------------------------------------------=
 * dump the given list of addresses into a super label
 *
 * Parameters:
 *      LinkList<const char *> &    - [in]  list to dump
 *      SuperLabel *                - [in]  where to dump.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE dumpAddressesIntoLabel
(
    LinkList<const char *> &list,
    SuperLabel *widget
)
{
    const char *pstr;
    ERRCODE er;

    widget->setUpdatesEnabled(false);
    widget->emptyList();

    /**
     * whip through the list adding the addreses one by one.
     */
    er = list.firstElement(&pstr);
    while (er == S_OK) {
        widget->addAddress(pstr);
        er = list.nextElement(&pstr);
    }

    widget->setUpdatesEnabled(true);
    widget->repaint();
    return S_OK;
}


