/**
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <qapplication.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "instance.h"
#include "kdispatch.h"
#include "kiltdown.h"
#include "main.h"
#include "sighandler.h"
#include "userpref.h"

#ifdef _KILTDOWN_USEKDE
#include <kapp.h>
#endif // _KILTDOWN_USEKDE


/**
 *=------------------------------------------------------------------------=
 *                      Local Prototypes
 *=------------------------------------------------------------------------=
 */ 
void processArguments(int argc, char **argv, const char * &);


const char *mailtoArg = "mailto:";

/**
 * This is our global application object.  We save it here so that everybody 
 * can add things to it.
 */
#ifdef _KILTDOWN_USEKDE
KApplication *localApplication = NULL;
#else 
QApplication *localApplication = NULL;
#endif 

/**
 *=------------------------------------------------------------------------=
 * main
 *=------------------------------------------------------------------------=
 * This is our main function.
 */
int main (int argc, char **argv)
{
    const char *mailto = NULL;
    char *noSignals;
    ERRCODE er;

    /**
     * intialize our local subsystems
     */
    lallocInit();
    lerrorInit();

    /**
     * Only install signal handlers if we're not debugging them!!!
     */
    noSignals = getenv("KD_NOSIGNALS");
    if (noSignals == NULL || *noSignals == '\0')
        signalInstallHandlers();

    /**
     * We want everything we write out to be safe/protected from prying eyes.
     * People can TEMPORARILY change this elsewhere in the program, but
     * should definitely change it back.
     */
    umask(077);

    /**
     * Create the main QT Application object and feed it the command line
     * parameters
     */
#ifdef _KILTDOWN_USEKDE
    KApplication mainApp(argc, argv, "kiltdown");
#else 
    QApplication mainApp(argc, argv);
#endif // _KILTDOWN_USEKDE

    localApplication = &mainApp;

//    mainApp.setFont(QFont("Helvetica", 12));

    /**
     * Next, look and see if there are any args we're interested in.
     */
    processArguments(argc, argv, mailto);

    /**
     * Make sure there's only one instance of this program running.  It
     * will just call exit(3) if we're not supposed to run any more.
     */
    instanceCheckInstance(mailto);

    /**
     * Load in the user preferences and initialize our UI portion.
     */
    er = prefInit();
    GOTO_ON_FAILURE(er, Error);

    er = Kiltdown::initializeSystem(mailto);
    GOTO_ON_FAILURE(er, Error);

    /**
     * Run the Application.  NOTE: because we hook up the signal here, we
     * can't delete this object ourselves, because Qt wants it around for
     * eventual application cleanup.  Yikes.
     */
    mainApp.connect(&mainApp, SIGNAL(lastWindowClosed()), &mainApp, SLOT(quit()) ); 
    mainApp.exec();

    er = S_OK;

  Error:

    instanceClearInstance();

    if (FAILED(er)) errorShowError(er);

    /**
     * Cleanup and destroy any objects that we have lying around still.
     */
    if (mailto) localFree(mailto);
    Kiltdown::terminateSystem();
    prefTerm();
    lerrorTerm();

    /**
     *=------------------------------------------------------------------=
     * THIS MUST BE DONE LAST !!!!
     *=------------------------------------------------------------------=
     * Finally, terminate our memory allocator and die.
     */
    downloadThread.terminate();
    lallocTerm();

    return 0;
}


/**
 *=------------------------------------------------------------------------=
 * nukeStringList
 *=------------------------------------------------------------------------=
 * blows away the contents of the given string list.
 *
 * Parameters:
 *      LinkList<const char *> &         - [in]  list to nuke.
 */
void nukeStringList
(
    LinkList<const char *> &list
)
{
    int x = list.countElements();
    const char *pstr;

    for (int y = 0; y < x; y++) {
        list.elementAt(0, &pstr);
        list.removeElement(pstr);
        localFree(pstr);
    }
}



/**
 *=------------------------------------------------------------------------=
 * processArguments
 *=------------------------------------------------------------------------=
 * processes the command line arguments to this program.  We only support
 * one right now:  "-mailto"
 *
 * Parameters:
 *      int                         - [in]  argc
 *      char **                     - [in]  argv
 *      const char * &              - [out] mailto string, or NULL;
 */
void processArguments
(
    int argc,
    char **argv,
    const char * & mailto
)
{
    int len;

    /**
     * look for an argv:  mailto:bobo@theclown.com
     */
    len = strlen(mailtoArg);
    for (int x = 0; x < argc; x++) {
        if (!strncasecmp(argv[x], mailtoArg, len)) {
            mailto = copyString(argv[x] + len);
            return;
        }
    }
}


