/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _MIME_H_
#define _MIME_H_

#include <llist.h>

#include "netapis.h"

/**
 * Forward decls
 */
class IReader;
class Attachment;

enum MessagePriority {

    NormalPriority = 0,
    LowPriority,
    HighPriority
};


/**
 *=------------------------------------------------------------------------=
 * These describes a MIME section.  There are properties: type, and
 * possibly:
 *      - an encoding
 *      - a name
 *      - a boundary separator
 *      - a charset
 * etc ...
 *=------------------------------------------------------------------------=
 */

/**
 * These are top level MIME media types.  NEW MEDIA TYPES SHOULD BE ADDED TO 
 * THE END, OR ELSE YOU RISK BREAKING PERSISTENCE CODE !!!!
 */
enum MimeMediaType {

    UnknownMediaType = -1,
    Text = 0,
    Image,
    Audio,
    Video,
    Application,
    Multipart,
    Message
};


/**
 * These are MIME media subtypes.   NEW MEDIA SUBTYPES SHOULD BE ADDED TO 
 * THE END, OR ELSE YOU RISK BREAKING PERSISTENCE CODE !!!!
 */
enum MimeMediaSubType {

    UnknownSubType = -1,
    Plain = 0,
    Html,
    Mixed,
    Alternative,
    OctetStream,
    Jpeg,
    Gif,
    Png,
    Mpeg,
    Related,
    XMSDownload,
    RFC822,
    XMSAsf,
    MsTnef,
    Signed,
    XVCard
};

/**
 * This is a MIME type.  Basically, we split a 32bit unsigned int up into
 * 16bits for the media type, and 16bits for the subtype.  Included are a
 * couple of inline functions to manipulate them.
 */
class MIMETYPE {
  public:
    inline MimeMediaType getMediaType() { return mediaType; }
    inline MimeMediaSubType getSubType() { return subType; }
    inline void setMediaType(MimeMediaType m) { mediaType = m; }
    inline void setSubType(MimeMediaSubType st) { subType = st; }
    inline void setType(MimeMediaType t, MimeMediaSubType st) {
        mediaType = t; subType = st;
    }

    inline unsigned int getWholeThing() { return (mediaType << 16) | subType; }
    inline void setWholeThing(unsigned int i) {
        mediaType = (MimeMediaType)((i & 0xffff0000) >> 16);
        subType = (MimeMediaSubType)(i & 0xffff);
    }

    MIMETYPE();

    inline bool operator== (MIMETYPE &other) {
        return other.mediaType == mediaType && other.subType == subType;
    }

    inline void operator= (MIMETYPE &other) {
        mediaType = other.mediaType;
        subType = other.subType;
    }

 private:
    MimeMediaType mediaType;
    MimeMediaSubType subType;
};

/**
 * These are the MIME encodings that we understand/support
 */
enum MimeEncoding {

    InvalidEncoding = -1,
    Base64 = 0,
    Uuencode,
    QuotedPrintable,
    SevenBit,
    EightBit
};

/**
 * These are the dispositions we know about
 */
enum MimeDisposition {

    InvalidDisposition = -1,
    Inline = 0,
    MailAttachment

};

/**
 * This struct is used to hold the information about a given mime section.
 */
struct MimeContentInfo {

    MIMETYPE type;
    MimeEncoding encoding;
    MimeDisposition disposition;
    const char *separator;
    const char *charset;
    const char *name;
};


/**
 * this struct defines something we need to add to a mail message.  It's
 * either going to be a MIME-typed body or an attachment.
 */
struct MimeBody {
    
    Attachment *attachment;
    MIMETYPE bodyType;
    const char *body;
    const char *charset;
    bool isBody;

};




/**
 *=------------------------------------------------------------------------=
 * These are all the SMTP Headers we know about.   For those we don't know,
 * we will just return Unknown
 *=------------------------------------------------------------------------=
 */
enum SMTPHeader {
    EndOfHeader = -2,
    UnknownHeader = -1,
    From = 0,
    To,
    MessageID,
    Sender,
    Received,
    CC,
    Date,
    ReturnPath,
    XMSMailPriority,
    ContentType,
    ContentTransferEncoding,
    Subject,
    ContentDisposition,
    ReplyTo
};

/**
 * These are structures and tables that map between encodings or types and
 * strings, etc...
 */
struct MimeMediaTypeMap {
    MimeMediaType type;
    const char *name;
    int size;
};

struct MimeMediaSubTypeMap {
    MimeMediaSubType type;
    const char *name;
    int size;
    MimeDisposition preferredDisp;
};


struct MimeEncodingMap {
    MimeEncoding type;
    const char *name;
    int size;
};

extern const MimeEncodingMap mimeEncodings [];
extern const MimeMediaTypeMap mimeMediaTypes[];
extern const MimeMediaSubTypeMap mimeMediaSubTypes[];

/**
 * This is the maximum length of a MIME header that we support.  Hopefully
 * it's not too small.
 * UNDONE: marcw, 2001-1-21 -- determine max SMTP unsplit header length
 */
#define MAX_HEADER_LEN         8192

/**
 *=------------------------------------------------------------------------=
 * This is a class that does reading on top of an IReader object.  It's a 
 * bit more complex than the simple IReader case just because we need to be
 * able to unsplit split up header lines, and also buffer/re-read the current
 * line to help find end of section boundary markers in MIME.
 *=------------------------------------------------------------------------=
 */
class MimeReader {

  public:
    MimeReader();

    /**
     * sets the reader to work with.
     */
    inline void setReader(IReader *reader) { this->reader = reader; }

    /**
     * Reads in a line.  Note that if we're processing headers, then we should
     * be using firstHeader/nextHeader, as these will unsplit multi-line 
     * headers.  You can always see this result again by calling currentLine.
     */
    ERRCODE readLine(void *p, unsigned int s, int &cread);

    /**
     * this function gets the first header.  It's not strictly 100% necessary,
     * but will make sure everything's kosher ...
     */
    ERRCODE firstHeader(SMTPHeader &, const char * & value);

    /**
     * This function is used to iterate over headers.  When you're done 
     * reading in headers, then you should just use readLine() until you get
     * to another place where you would be reading in headers again.
     */
    ERRCODE nextHeader(SMTPHeader &, const char * & value);

    /**
     * returns the current line.
     */
    const char *currentLine();

  private:
    IReader *reader;
    bool haveNext;
    char curHeader[MAX_HEADER_LEN];
    char curLine[NET_MAXLINELEN];

    /**
     * this just unsplits a header line into the curHeader buffer.
     */
    ERRCODE readHeaderLine();
};

extern MimeReader mimeReader;


/**
 *=------------------------------------------------------------------------=
 *                      MIME Header Function Prototypes
 *=------------------------------------------------------------------------=
 */
ERRCODE parseAddressList(const char *, LinkList<const char *> &);
ERRCODE parseContentType(const char *, MimeContentInfo *);
ERRCODE parseContentEncoding(const char *, MimeContentInfo *);
ERRCODE parseContentDisposition(const char *, MimeContentInfo *);
ERRCODE parseDate(const char *, time_t *);
ERRCODE parsePriority(const char *, MessagePriority *);
ERRCODE parseSingleAddress(const char *, const char **);
ERRCODE parseReceivedLine(const char *, time_t *);

/**
 *=------------------------------------------------------------------------=
 *                      MIME Body Function Prototypes
 *=------------------------------------------------------------------------=
 */
ERRCODE decodeSection(MimeEncoding, const char *, int *, const char **);



#endif // _MIME_H_



