/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>
#include <string.h>
#include <time.h>

#include <llist.h>
#include "main.h"
#include "mime.h"
#include "userpref.h"


/**
 *=------------------------------------------------------------------------=
 * These are some simple arrays to map MIME types or encodings into
 * enumerations that MailMessages will understand
 *=------------------------------------------------------------------------=
 */

struct MimeDispositionMap {
    MimeDisposition disposition;
    const char *name;
    int size;
};

const MimeMediaTypeMap mimeMediaTypes[] = {
    { Text, "text", sizeof("text") - 1},
    { Multipart, "multipart", sizeof("multipart") - 1},
    { Application, "application", sizeof("application") -1},
    { Image, "image", sizeof("image") - 1},
    { Video, "video", sizeof("video") -1},
    { Message, "message", sizeof("message") - 1},
    { UnknownMediaType, NULL, -1 }
};


const MimeMediaSubTypeMap mimeMediaSubTypes [] = {
    { Plain, "plain", sizeof("plain") - 1, Inline },
    { Html, "html", sizeof("html") - 1, Inline },
    { Mixed, "mixed", sizeof("mixed") - 1, InvalidDisposition },
    { Alternative, "alternative", sizeof("alternative") - 1, InvalidDisposition },
    { OctetStream, "octet-stream", sizeof("octet-stream") - 1, MailAttachment },
    { Jpeg, "jpeg", sizeof("jpeg") - 1, MailAttachment },
    { Gif, "gif", sizeof("gif") - 1, MailAttachment },
    { Png, "png", sizeof("png") - 1, MailAttachment },
    { Mpeg, "mpeg", sizeof("mpeg") - 1, MailAttachment },
    { Related, "related", sizeof("related") - 1, InvalidDisposition },
    { XMSDownload, "x-msdownload", sizeof("x-msdownload") -1, MailAttachment },
    { MsTnef, "ms-tnef", sizeof("ms-tnef") - 1, MailAttachment },
    { RFC822, "rfc822", sizeof("rfc822") - 1, MailAttachment },
    { XMSAsf, "x-ms-asf", sizeof("x-ms-asf") - 1, MailAttachment },
    { Signed, "signed", sizeof("signed") - 1, InvalidDisposition },
    { XVCard, "x-vcard", sizeof("x-vcard") - 1, MailAttachment },
    { UnknownSubType, NULL, -1, InvalidDisposition }
};

const MimeEncodingMap mimeEncodings [] = {
    { Base64, "base64", sizeof("base64") - 1 },
    { Uuencode, "uuencode", sizeof("uuencode") - 1 },
    { QuotedPrintable, "quoted-printable", sizeof("quoted-printable") - 1 },
    { SevenBit, "7bit", sizeof("7bit") - 1 },
    { EightBit, "8bit", sizeof("8bit") - 1 },
    { EightBit, "binary", sizeof("8bit") - 1},
    { InvalidEncoding, NULL, 0 }
};

const MimeDispositionMap mimeDispositions [] = {
    { MailAttachment, "attachment", sizeof("attachment") - 1 },
    { Inline, "inline", sizeof("inline") - 1 }
};


/**
 * Maximum EMail Address length we support.
 */
const int MAX_ADDRESSLEN = 512;

/**
 * Maximum length of a boundary separator we'll support.
 */
const int SEPARATOR_MAXLEN = 512;



/**
 *=------------------------------------------------------------------------=
 *                           Local Prototypes
 *=------------------------------------------------------------------------=
 */
const char *prepareAndCopySeparator(const char *);


/**
 *=------------------------------------------------------------------------=
 *            Prototypes Shamelessly stolen from Other Files
 *=------------------------------------------------------------------------=
 */
const char *stripNewLines(const char *);


/**
 *=------------------------------------------------------------------------=
 * MIMETYPE::MIMETYPE
 *=------------------------------------------------------------------------=
 * constructor for the MIMETYPE class -- I've just put it here to make
 * sure the compiler doesn't duplicate them.
 */
MIMETYPE::MIMETYPE() {
    mediaType = UnknownMediaType;
    subType = UnknownSubType;
}


/**
 *=------------------------------------------------------------------------=
 * parseAddressList
 *=------------------------------------------------------------------------=
 * the given structured field is a full list of addresses.  Add the items in
 * the field to the given string list.
 *
 * Parameters:
 *      const char *                - [in]  where the addresses are all at.
 *      LinkList<const char *> &    - [in]  where to add strings.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE parseAddressList
(
    const char *stripMe,
    LinkList<const char *> &list
)
{
    char addr[MAX_ADDRESSLEN];
    bool inquotes, inangles;
    const char *pstr;
    const char *buf;
    bool lastwasnl;
    int x = 0,  y;
    ERRCODE er;

    /**
     * basically, we're going to go whizzing through the list, looking for
     * commas that are not nested inside of ""s or <>s.
     */
    if (!stripMe) return E_PARAM;
    buf = stripNewLines(stripMe);
    if (!buf) return E_OUTOFMEMORY;

    /**
     * Okay, now start whizzing through, noting whether we're in ""s or
     * <>s or whatever.
     */
    inquotes = false;
    inangles = false;
    y = 0;

    /**
     * get rid of any leading whitespace first.
     */
    while (isspace(buf[x])) x++;

    /**
     * Now see if it's a quote or an angle bracket or some such thing.
     */
    lastwasnl = false;
    while (buf[x] != '\0') {
        if (buf[x] == '"' && !inangles) {
            inquotes = !inquotes;
        } else if (buf[x] == '<' && !inquotes) {
            inangles = true;
        } else if (buf[x] == '>' && inangles){
            inangles = false;
        }

        /**
         * IF it's a comma, and we're not in angles or quotes, then we've found
         * an item.
         */
        if (buf[x] == ',' && !inquotes && !inangles) {
            addr[y] = '\0';
            pstr = stripNewLines(addr);
            if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
            er = list.addElement(pstr);
            if (FAILED(er)) {
                localFree(pstr);
                goto Error;
            }
            y = 0;

            /**
             * Get rid of any space after this one.
             */
            x++;
            while (isspace(buf[x])) x++;

        } else {
            if (buf[x] == '\r' || buf[x] == '\n') {
                if (!lastwasnl) {
                    addr[y++] = ' ';
                }
                lastwasnl = true;
                x++;
            } else {
                lastwasnl = false;
                addr[y++] = buf[x++];
            }
        }
    }

    if (buf[x] == '\0' && y) {
        addr[y] = '\0';
        pstr = stripNewLines(addr);
        if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
        er = list.addElement(pstr);
        if (FAILED(er)) {
            localFree(pstr);
            goto Error;
        }
    }

    if (inquotes || inangles) fprintf(stderr, "warning: malformed email address on input -- ignorning.\n******** %s\n", buf);
    er = S_OK;

  Error:
    localFree(buf);
    if (FAILED(er)) nukeStringList(list);
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * parseContentEncoding
 *=------------------------------------------------------------------------=
 * the current line starts with Content-Transfer-Encoding and will tell us
 * more about the encoding format.  Should be pretty easy.
 *
 * Parameters:
 *      const char *                - [in]  the current line, minus the
 *                                      Content-Transfer-Encoding.
 *      MimeContentInfo *           - [in]  where to put the encoding
 *                                      information.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE parseContentEncoding
(
    const char *buf,
    MimeContentInfo *info
)
{
    char enc[256];
    int x, y, z;

    /**
     * Params
     */
    if (!buf || !info) return E_PARAM;
    info->encoding = InvalidEncoding;

    /**
     * Begin parsing.
     * first, go and skip over any whitespace
     */
    x = 0;
    while (isspace(buf[x])) x++;
    if (buf[x] == '\0') {
        fprintf(stderr, "Malformed Content-Transfer-Encoding line ...\n");
        return S_OK;
    }

    /**
     * Great, now see what we've got:
     */
    y = 0;
    while (buf[x] != '\r' && buf[x] != '\n' && buf[x] != '\0' && buf[x] != ';')
        enc[y++] = buf[x++];
    enc[y] = '\0';

    /**
     * See if it's a content type we know about.
     */
    z = 0;
    while (mimeEncodings[z].type != InvalidEncoding) {
        if (toupper(mimeEncodings[z].name[0]) == toupper((enc)[0])
            && !strncasecmp(enc, mimeEncodings[z].name, mimeEncodings[z].size)) {

            info->encoding = mimeEncodings[z].type;
            break;
        }
        z++;
    }
     
    return S_OK;
}
 
/**
 *=------------------------------------------------------------------------=
 * parseContentDisposition
 *=------------------------------------------------------------------------=
 * we've got a "Content-Disposition" header.  We want to see if the object
 * is trying to be an attachment or not.
 *
 * Parameters:
 *      const char *                    - [in]  the info
 *      MimeContentInfo *               - [in]  where we'll say its an 
 *                                          attachment or not.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE parseContentDisposition
(
    const char *buf,
    MimeContentInfo *info
)
{
    char keyword[128];
    char content[256];
    int x = 0, y, z;

    if (!buf || !info) return E_PARAM;

    /**
     * Begin parsing.
     * first, go and skip  any whitespace
     */
    x = 0;
    while (isspace(buf[x])) x++;
    if (buf[x] == '\0') {
        fprintf(stderr, "malformed Content-Disposition\n");
        info->type.setMediaType(UnknownMediaType);
        return S_OK;
    }

    /**
     * See if it's a content disposition we know about.
     */
    for (z = 0; z < (int)(sizeof(mimeDispositions) / sizeof(MimeDispositionMap)); z++) {
        if (toupper(mimeDispositions[z].name[0]) == toupper((buf + x)[0])
            && !strncasecmp(buf + x, mimeDispositions[z].name, mimeDispositions[z].size)) {

            info->disposition = mimeDispositions[z].disposition;
            break;
        }
    }

    if (z == (int)(sizeof(mimeDispositions) / sizeof(MimeDispositionMap))) {
        fprintf(stderr, "Unknown disposition -- treating as attachment: %s\n", buf + x);
        info->disposition = MailAttachment;
    } else {
        /**
         * Position the pointer after the keyword ...
         */
        x += strlen(mimeDispositions[z].name) + 1;
        while (buf[x] == ';' || isspace(buf[x])) x++;
    }

    /**
     * Now start looking for keywords we might recognize.
     */
    while (buf[x] != '\0') {

        /**
         * Skip over whitespace first.
         */
        while (isspace(buf[x])) x++;
        if (buf[x] == '\0') break;

        /**
         * Copy over the keyword.
         */
        y = 0;
        while (buf[x] != '\0' && buf[x] != '=') {
            if (!isspace(buf[x])) 
                keyword[y++] = buf[x];
            x++;
        }
        keyword[y] = '\0';
        x++;

        /**
         * now the value
         */
        while (isspace(buf[x])) x++;
        if (buf[x] == '"') x++;
        y = 0;
        while (buf[x] != '\0' && buf[x] != '"' && buf[x] != ';') {
            if (!isspace(buf[x]))
                content[y++] = buf[x];
            x++;
        }
        content[y] = '\0';
        if (buf[x] == '"') x++;
        while (isspace(buf[x])) x++;
        if (buf[x] == ';') x++;
        while (isspace(buf[x])) x++;

        /** 
         * See what we've got.  We currently recognize:
         * - name
         */
        switch (toupper(keyword[0])) {
            case 'N':
            case 'F':
                if (!strcasecmp(keyword, "name") || !strcasecmp(keyword, "filename")) {
                    if (info->name) localFree(info->name);
                    info->name =  copyString(content);
                    if (!(info->name))
                        return E_OUTOFMEMORY;
                } else {
                    fprintf(stderr, "unknown keyword in Content-Type: %s\n", keyword);
                }
                break;

            default:
                fprintf(stderr, "unknown keyword in Content-Disposition: %s\n", keyword);
                break;
        }

    } // while

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * parseContentType
 *=------------------------------------------------------------------------=
 * get the MIME information from the given structured header.
 *
 * Parmeters:
 *      const char *                    - [in]  the content type
 *      MimeContentInfo *               - [in]  where to put what we learn
 *                                          about the section.
 * Output:
 *      ERRCODE
 */
ERRCODE parseContentType
(
    const char *buf,
    MimeContentInfo *info
)
{
    char keyword[128];
    char content[256];
    bool inquote;
    ERRCODE er;
    int x = 0;
    int z, y;

    if (!buf || !info) return E_PARAM;

    /**
     * Begin parsing.
     * first, go and skip  any whitespace
     */
    x = 0;
    while (isspace(buf[x])) x++;
    if (buf[x] == '\0') {
        fprintf(stderr, "malformed Content-Type\n");
        info->type.setMediaType(UnknownMediaType);
        return S_OK;
    }

    /**
     * The very next thing in the buffer will be the actual content-type, 
     * followed by attributes.  extract the top-level media type first.
     */
    z = 0;
    while (mimeMediaTypes[z].type != UnknownMediaType) {
        if (toupper(mimeMediaTypes[z].name[0]) == toupper((buf + x)[0])
            && !strncasecmp(buf + x, mimeMediaTypes[z].name, mimeMediaTypes[z].size)) {
            info->type.setMediaType(mimeMediaTypes[z].type);
            break;
        }
        z++;
    }

    /**
     * IF we don't know the type -- the rest is ignored -- we'll just return
     * the section as an unknown section.
     */
    if (mimeMediaTypes[z].type == UnknownMediaType) {
        fprintf(stderr, "This is a MIME media type I don't know about -- \n'%s'\n", buf);
        info->type.setMediaType(UnknownMediaType);
        return S_OK;
    } else {
        /**
         * Position the pointer after the keyword ...
         */
        x += mimeMediaTypes[z].size;
    }

    /**
     * Now, we should have a / character and the subtype.  Get that now.
     */
    if (*(buf + x) != '/') {
        info->type.setSubType(UnknownSubType);
    } else {

        x++;
        z = 0;
        while (mimeMediaSubTypes[z].type != UnknownSubType) {
            if (toupper(mimeMediaSubTypes[z].name[0]) == toupper((buf + x)[0])
                && !strncasecmp(buf + x, mimeMediaSubTypes[z].name, mimeMediaSubTypes[z].size)) {
                info->type.setSubType(mimeMediaSubTypes[z].type);
                if (info->disposition == InvalidDisposition)
                    info->disposition = mimeMediaSubTypes[z].preferredDisp;
                break;
            }
            z++;
        }

        if (mimeMediaSubTypes[z].type == UnknownSubType) {
            fprintf(stderr, "This is a MIME media subtype I don't know about -- \n'%s'\n", buf);
            info->type.setSubType(UnknownSubType);
            if (info->disposition == InvalidDisposition)
                info->disposition = MailAttachment;
            
            /**
             * Get the cursor past the unknown sub-type!
             */
            while (!isspace(*(buf + x)) && *(buf + x) != ';') x++;
            x++;
        } else {
            /**
             * Position the pointer after the keyword ...
             */
            x += mimeMediaSubTypes[z].size;
        }
    }

    /**
     * now skip over the ; that comes after the type, and also get rid of any
     * white space.
     */
    while (buf[x] == ';' || isspace(buf[x])) x++;

    /**
     * Now start looking for keywords we might recognize.
     */
    while (buf[x] != '\0') {

        /**
         * Skip over whitespace first.
         */
        while (isspace(buf[x])) x++;
        if (buf[x] == '\0') break;


        /**
         * Copy over the keyword.
         */
        y = 0;
        while (buf[x] != '\0' && buf[x] != '=') {
            if (!isspace(buf[x])) 
                keyword[y++] = buf[x];
            x++;
        }
        keyword[y] = '\0';
        x++;

        /**
         * now the value
         */
        while (isspace(buf[x])) x++;
        if (buf[x] == '"') {
            inquote = true;
            x++;
        } else
            inquote = false;
        y = 0;
        while (buf[x] != '\0' && buf[x] != '"' && buf[x] != ';') {
            if (inquote || !isspace(buf[x]))
                content[y++] = buf[x];
            x++;
        }
        content[y] = '\0';
        if (buf[x] == '"') {
            inquote = false; 
            x++;
        }
        while (isspace(buf[x])) x++;
        if (buf[x] == ';') x++;
        while (isspace(buf[x])) x++;

        /** 
         * See what we've got.  We currently recognize:
         * - charset
         * - boundary
         * - name
         * - format     (although we have no idea what this means)
         */
        switch (toupper(keyword[0])) {
            case 'B':
                if (!strcasecmp(keyword, "boundary")) {
                    if (info->separator) localFree(info->separator);
                    info->separator = prepareAndCopySeparator(content);
                    if (!(info->separator)) {
                        er = E_OUTOFMEMORY;
                        goto Error;
                    }
                }
                break;

            case 'C':
                if (!strcasecmp(keyword, "charset")) {
                    if (info->charset) localFree(info->charset);
                    info->charset = copyString(content);
                    if (!(info->charset)) {
                        er = E_OUTOFMEMORY;
                        goto  Error;
                    }
                }
                break;

            case 'F':
                if (!strcasecmp(keyword, "format")) {
                    /**
                     * we have no idea what to do with this just yet.  ignore 
                     * it for now.
                     */
                    ;
                }
                break;

            case 'N':
                if (!strcasecmp(keyword, "name")) {
                    if (info->name) localFree(info->name);
                    info->name =  copyString(content);
                    if (!(info->name)) {
                        er = E_OUTOFMEMORY;
                        goto Error;
                    }
                }
                break;

            default:
                fprintf(stderr, "unknown keyword in Content-Type: %s\n", keyword);
                break;
        }

    } // while

    /**
     * UNDONE: marcw, 12.8.2000 --  learn more about ms-tnef and figure it out.
     */
    if (info->type.getSubType() == MsTnef && info->name == NULL) {
        info->name = copyString("Kooky Microsoft Section");
        if (!info->name) { er = E_OUTOFMEMORY; goto Error; }
    }

    er = S_OK;

  Error:
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * parseDate
 *=------------------------------------------------------------------------=
 * given a string that is a  structured field (date).  parse it out and
 * return it in a time_t.  yuck.  Fortunately, SMTP dates are extremely
 * structured.
 *
 * Parameters:
 *      const char *                - [in]  structured date
 *      time_t *                    - [out] date we read in.
 *
 * Output:
 *      ERRCODE
 *
 * Notes:
 *      There may be some stuff AFTER the date that isn't date related. Be
 *      ready for that.
 */
ERRCODE parseDate
(
    const char *buf,
    time_t *pdate
)
{
    static const char *months[] = {
        "Jan", "Feb", "Mar", "Apr", "May", "Jun",
        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
    };
    int x, day, year, mindex;
    struct tm timeInfo;
    int hr, min, sec;
    char month[4];
    int cread;

    if (!buf || !pdate) return E_PARAM;
    *pdate = 0;

    /**
     * Begin parsing.
     * first, go and skip over any whitespace
     */
    x = 0;
    while (isspace(buf[x])) x++;
    if (buf[x] == '\0') {
        fprintf(stderr, "Malformed date on input: %s\n", buf);
        return S_OK;
    }

    /**
     * great, now see if it starts with a day of the week.
     */
    if (isalpha(buf[x])) {

        /**
         * Skip over the day of the week.
         */
        while (isalpha(buf[x])) x++;
        
        /**
         * Now skip over any whitespace or commas.
         */
        while (isspace(buf[x]) || buf[x] == ',') x++;
    }

    /**
     * We should now have something of the form:  XX mon (YY)YY
     * UNDONE: marcw, 11.27.2000 -- is sscanf too restrictive?  Just
     *      how much variation in SMTP dates is there out there?
     */
    cread = sscanf(buf + x, "%d %3s %d", &day, month, &year);
    if (cread != 3) {
        NET_LOG("Malformed date on message input.  Using current time instead\n");
        *pdate = time(NULL);
        return S_OK;
    }

    /**
     * CONSIDER: marcw, 11.26.2000 -- Y2K issues here !!!!
     * Do a little y2k massaging here ..  The timeInfo struct expects the
     * number of years since 1900 ...
     */
    if (year < 15)  
        year += 100;
    else if (year > 1900)
        year -= 1900;

    /**
     * Figure out what month it is.
     */
    for (mindex = 0; mindex < 12; mindex++) {
        if (!strncasecmp(month, months[mindex], 3))
            break;
    }


    if (mindex == 12) {
        NET_LOG("Malformed date on message input.  Using current time instead\n");
        *pdate = time(NULL);
        return S_OK;
    }

    timeInfo.tm_year = year;
    timeInfo.tm_mday = day;
    timeInfo.tm_mon = mindex;

    /**
     * Great, now go and get the time, first, we need to skip over all the
     * goo we just read in.
     */
    while (isdigit(buf[x])) x++;
    while (isspace(buf[x])) x++;
    while (isalpha(buf[x])) x++;
    while (isspace(buf[x])) x++;
    while (isdigit(buf[x])) x++;
    while (isspace(buf[x])) x++;

    /**
     * Now we should have something of the form:
     *  hh:mm[:ss] TIMEZONE
     */
    sec = 0;
    cread = sscanf(buf + x, "%d : %d : %d", &hr, &min, &sec);
    if (cread < 2) {
        NET_LOG("Malformed date on message input.  Using current time instead\n");
        *pdate = time(NULL);
        return S_OK;
    }

    timeInfo.tm_sec = sec;
    timeInfo.tm_min = min;
    timeInfo.tm_hour = hr;

    /**
     * Now this is really gnarly -- There are all sorts of time zone values
     * that seem to get thrown around the net.  We will attempt to parse them
     * out here.
     */
    while (isdigit(buf[x])) x++;
    while (isspace(buf[x])) x++;
    if (buf[x] == ':') x++;    
    while (isspace(buf[x])) x++;
    if (buf[x] == ':') {
        while (isspace(buf[x])) x++;
        while (isdigit(buf[x])) x++;
        while (isspace(buf[x])) x++;
    }

    if (buf[x] == '\0') goto Done;

/**
 * UNDONE: marcw, 12.9.2000 -- timezone info most definitely needed !!!!!
 */
//    K_FAIL("Timezone information processing required\n");

  Done:
    *pdate = mktime(&timeInfo);
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * parsePriority
 *=------------------------------------------------------------------------=
 * parses out the priority for this message.  This is an extension put in
 * by the Evil Empire, but we'll support it, because it's cool.
 *
 * Parameters:
 *      const char *                - [in]  the priority
 *      MessagePriority *           - [out] the priority
 *
 * Output:
 *      ERRCODE
 */
ERRCODE parsePriority
(
    const char *buf,
    MessagePriority *pri
)
{
    int x = 0;

    if (!buf || !pri) return E_PARAM;

    /**
     * skip over starting whitespace
     */
    while (isspace(buf[x])) x++;

    switch (toupper(buf[x])) {
        case 'H':
            *pri = HighPriority;
            break;

        case 'L':
            *pri = LowPriority;
            break;

        case 'N':
            *pri = NormalPriority;
            break;

        default:
            fprintf(stderr, "unknown Priority: '%s'\n", buf);
            *pri = NormalPriority;
            break;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * parseSingleAddress
 *=------------------------------------------------------------------------=
 * the given structured field is a single e-mail address.  Extract said 
 * address and return it in an allocated string.
 *
 * Parameters:
 *      const char *                - [in]  basically, the address.
 *      const char **               - [out] the network address.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE parseSingleAddress
(
    const char *stripMe,
    const char **pstr
)
{
    char addr[MAX_ADDRESSLEN];
    bool lastwasnl = false;
    const char *buf;
    int x = 0;
    int y = 0;

    if (!stripMe || !pstr) return E_PARAM;

    buf = stripNewLines(stripMe);
    if (!buf) return E_OUTOFMEMORY;

    /**
     * this is pretty easy -- we just need to strip out any newlines at the
     * end of the string.
     */
    while (buf[x] != '\0') {
        if (buf[x] == '\r' || buf[x] == '\n') {
            if (!lastwasnl) addr[y] = ' ';
            lastwasnl = true;
        } else {
            addr[y++] = buf[x];
            lastwasnl = false;
        }
        x++;
    }

    /**
     * tie it off, copy it over, and return.
     */
    addr[y] = '\0';
    *pstr = copyString(addr);
    localFree(buf);
    return *pstr ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * parseReceivedLine
 *=------------------------------------------------------------------------=
 * parses a Received: line in the message.  We will only parse the first
 * one we've seen, and then go from there.
 *
 * Parameters:
 *      const char *                - [in]  the whole line (minus the
 *                                      "Received".
 *      time_t *                    - [out] we want the time this message
 *                                      was received.
 *
 * Output:
 */
ERRCODE parseReceivedLine
(
    const char *buf,
    time_t *received
)
{
    int x = 0;

    /**
     * Okay, right now, we're basically just going to go look for the date
     * this message was received.  It's after the first semi-colon.
     */
    while (buf[x] != ';' && buf[x] != '\0') x++;

    /**
     * great, we're either at the end (bad) or we're at the date
     */
    if (buf[x] == '\0') {
        fprintf(stderr, "Message has no received date??\n");
        *received = time(NULL);
        return S_OK;
    }

    /**
     * return the date.  We'll ignore the rest of this field for now.
     */
    return parseDate(buf + x + 1, received);
}

/**
 *=------------------------------------------------------------------------=
 * prepareAndCopySeparator
 *=------------------------------------------------------------------------=
 * removes any CRLF from a mime separator and prepends the --
 *
 * Parameters:
 *      const char *                - [in]  header to fix up.
 */
const char *prepareAndCopySeparator
(
    const char *sep
)
{
    char buf[SEPARATOR_MAXLEN];
    int index;

    buf[0] = '-';
    buf[1] = '-';
    index = 0;
    while (sep[index] != '\n' && sep[index] != '\r'
           && sep[index] != '\0' && (index + 2 < (int)sizeof(buf))) {
        buf[index + 2] = sep[index++];
    }

    if (index + 2 < (int)sizeof(buf))
        buf[index + 2] = '\0';
    else
        buf[sizeof(buf)] = '\0';

    return copyString(buf);
}



