/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>
#include <string.h>

#include "mime.h"
#include "netapis.h"


/**
 * this is our global mime reader -- it's just a lot easier to create it here
 * than have to create a 9k object on the heap for EACH message loading ...
 */
MimeReader mimeReader;

/**
 *=------------------------------------------------------------------------=
 * MimeReader::MimeReader
 *=------------------------------------------------------------------------=
 * creates a reader.  Very complex.
 */
MimeReader::MimeReader()
{
    curHeader[0] = '\0';
    curLine[0] = '\0';
    haveNext = false;
}

/**
 *=------------------------------------------------------------------------=
 * MimeReader::currentLine
 *=------------------------------------------------------------------------=
 * returns the contents of the current line
 * 
 * Output:
 *      const char *                - [out] DON'T ATTEMPT TO MODIFY OR 
 *                                      DELETE ME !!!!
 */
const char *MimeReader::currentLine()
{
    return curLine;
}

/**
 *=------------------------------------------------------------------------=
 * MimeReader::firstHeader
 *=------------------------------------------------------------------------=
 * primes the network pump and gets the first line.
 *
 * Parameters:
 *      SMTPHeader &                - [out] what header value was returned.
 *      const char *&               - [out] the value of the header, or
 *                                      the ENTIRE LINE if Unknown is returned
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MimeReader::firstHeader
(
    SMTPHeader &header,
    const char *&value
)
{
    haveNext = false;
    return nextHeader(header, value);
}

/*
 *=------------------------------------------------------------------------=
 * MimeReader::nextHeader
 *=------------------------------------------------------------------------=
 * given an input stream that is still a stream of headers, go and get the
 * next header, extract the name from the value, and return these to the
 * caller.
 *
 * Parameters:
 *      SMTPHeader &                - [out] what header value was returned.
 *      const char *&               - [out] the value of the header, or
 *                                      the ENTIRE LINE if Unknown is returned
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MimeReader::nextHeader
(
    SMTPHeader &header,
    const char *&value
)
{
    ERRCODE er;
    int x = 0;

    value = NULL;

    /**
     * Merge the entire header into one line.
     */
    er = readHeaderLine();
    RETURN_ON_FAILURE(er);
                   
    /**
     * since RFC 822 doesn't allow for any whitespace before the field name,
     * this is pretty E-Zed.
     */
    switch (toupper(curHeader[0])) {
        case 'F':
            if (!strncasecmp(curHeader, "from", 4)) {
                header = From;
            } else {
                header = UnknownHeader;
            }
            break;
        case 'T':
            if (!strncasecmp(curHeader, "to", 2)) {
                header = To;
            } else {
                header = UnknownHeader;
            }
            break;
        case 'M':
            if (!strncasecmp(curHeader, "message-id", sizeof("message-id") - 1)) {
                header = MessageID;
            } else {
                header = UnknownHeader;
            }
            break;
        case 'S':
            if (!strncasecmp(curHeader, "sender", sizeof("sender") - 1)) {
                header = Sender;
            } else if (!strncasecmp(curHeader, "subject", sizeof("subject") - 1)) {
                header = Subject;
            } else {
                header = UnknownHeader;
            }
            break;
        case 'R':
            if (!strncasecmp(curHeader, "received", sizeof("received") - 1)) {
                header = Received;
            } else if (!strncasecmp(curHeader, "return-path", sizeof("return-path") - 1)) {
                header = ReturnPath;
            } else if (!strncasecmp(curHeader, "reply-to", sizeof("reply-to") - 1)) {
                header = ReplyTo;
            } else {
                header = UnknownHeader;
            }
            break;
        case 'C':
            if (!strncasecmp(curHeader, "cc", sizeof("cc") - 1)) {
                header = CC;
            } else if (!strncasecmp(curHeader, "content-type", sizeof("content-type") - 1)) {
                header = ContentType;
            } else if (!strncasecmp(curHeader, "content-transfer-encoding", sizeof("content-transfer-encoding") - 1)) {
                header = ContentTransferEncoding;
            } else if (!strncasecmp(curHeader, "content-disposition", sizeof("content-disposition") - 1)) {
                header = ContentDisposition;
            } else {
                header = UnknownHeader;
            }
            break;
        case 'D':
            if (!strncasecmp(curHeader, "date", sizeof("date") - 1)) {
                header = Date;
            } else {
                header = UnknownHeader;
            }
            break;
        case 'X':
            if (!strncasecmp(curHeader, "x-msmail-priority", sizeof("x-msmail-priority") - 1)) {
                header = XMSMailPriority;
            } else {
                header = UnknownHeader;
            }
            break;
        case '\r':
        case '\n':
            header = EndOfHeader;
            value = NULL;
            return S_OK;

        case '\0':
            return E_NET_UNEXPECTED;

        default:
            header = UnknownHeader;
            break;
    }

    /**
     * Great -- we've got a header.  Copy over the value of the line, or the 
     * whole thing in the case of an unknown header.
     */
    x = 0;
    if (header != UnknownHeader) {
        while (curHeader[x] != ':' && curHeader[x] != '\0')
            x++;
        x++;
        while (isspace(curHeader[x])) x++;
    }

    /**
     * copy out the resulting string and return it.
     */
    value = copyString(curHeader + x);
    return value ? S_OK : E_OUTOFMEMORY;
}


/**
 *=------------------------------------------------------------------------=
 * MimeReader::readHeaderLine
 *=------------------------------------------------------------------------=
 * assuming that we are reading headers, we want to read in the entire 
 * contents of a header 'line', which may actually be split over numerous
 * actual lines on the input.
 *
 * Parameters:
 *      const char * &                  - [out]  a pointer to a buf holding
 *                                          what we just read.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MimeReader::readHeaderLine()
{
    char buf[NET_MAXLINELEN];
    int cread, len;
    ERRCODE er;

    len = 0;

    /**
     * First figure out if we've looked ahead a line.  If so, just copy it
     * over.
     */
    er = readLine(curHeader, sizeof(curHeader), cread);
    RETURN_ON_FAILURE(er);
    len += strlen(curHeader);

    /** 
     * IF we're at the end of the header, then we abort and we're done!
     */
    if (curHeader[0] == '\r' || curHeader[0] == '\n')
        return S_OK;

    /**
     * now whiz through until we've completed unsplitting this entire
     * line.
     */
    while (1) {
        er = readLine(buf, sizeof(buf), cread);
        RETURN_ON_FAILURE(er);
        
        if (isspace(curLine[0]) && curLine[0] != '\r' && curLine[0] != '\n') {
            strncpy(curHeader + len, curLine, sizeof(curHeader) - len);
            len += (cread);
        } else {
            break;
        }
    }

    haveNext = true;

    /**
     * Okay, since we have to look ahead to read these headers, note that 
     * we've already got the next line.
     */
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MimeReader::readLine
 *=------------------------------------------------------------------------=
 * reads in the next line, buffering it as we go.
 *
 * Parameters:
 *      void *                      - [in]  buffer to read into.
 *      unsigned int                - [in]  max num bytes to read.
 *      int &                       - [out] number of bytes actually read
 *                                      - 1 on error.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MimeReader::readLine
(
    void *buf,
    unsigned int maxSize,
    int &cread
)
{
    ERRCODE er;

    /**
     * If we've already got the next line, then just go ahead and copy it .
     */
    if (haveNext) {
        haveNext = false;
        cread = strlen(curLine);
    } else {
        er = reader->readLine(curLine, sizeof(curLine), cread);
        RETURN_ON_FAILURE(er);
    }

    strncpy((char *)buf, curLine, maxSize);
    cread = (int)maxSize < cread ? (int)maxSize : cread;
    return S_OK;
}


