/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>

#include "netapis.h"
#include "userpref.h"


// CONSIDER: marcw, 11.26.2000 -- to make this routine really thread safe,
//           you need to make the static buffer in _readLineHelper TLS.



/**
 *=------------------------------------------------------------------------=
 *                          Local Prototypes
 *=------------------------------------------------------------------------=
 */
static ssize_t _readLineHelper(int, char *);

/**
 * These are used for buffering.
 */
static int read_count = 0;
static char *read_ptr;
static char read_buf[NET_MAXLINELEN];



/**
 *=------------------------------------------------------------------------=
 * networkConnect
 *=------------------------------------------------------------------------=
 * Given a Host name and a port, attempts to connect to said port on said
 * host and returns a file descripter for the connection.  
 *
 * Parameters:
 *      const char *                - [in]  hostname
 *      int                         - [in]  port
 *      int *                       - [out] socket file descriptor
 *
 * Output:
 *      ERRCODE
 * 
 * Notes:
 *      UNDONE: marcw, 11.21.2000 --  make this IPv6 compatible.
 */
ERRCODE networkConnect
(
    const char *host,
    int port,
    int *fd
)
{
    char buf[NET_MAXLINELEN];
    struct sockaddr_in serv;
    struct hostent *he;
    int socketfd;
    int result;

    if (!host || !fd) return E_PARAM;

    if (prefNetworkLog) {
        snprintf(buf, sizeof(buf), "Requesting connect to: %s, port %d\n", host, port);
        NET_LOG(buf);
    }

    /**
     * Get the IP address of this host,
     */
    he = gethostbyname(host);
    if (he == NULL) {
        errorSaveRichErrorInfo(host, strerror(errno));
        return E_DNS_LOOKUP;
    }

    if (prefNetworkLog) {
        char buf2[128];
        inet_ntop(AF_INET, he->h_addr_list[0], buf, sizeof(buf));
        snprintf(buf2, sizeof(buf2), "Got IP address %s for host %s\n", buf, host);
        NET_LOG(buf2);
    }

    /**
     * Create the network socket.
     */
    socketfd = socket(AF_INET, SOCK_STREAM, 0);
    if (socketfd == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * Now go and actually try and connect to the server.
     */
    memset(&serv, 0, sizeof(serv));
    serv.sin_family = AF_INET;
    serv.sin_port = htons(port);
    memcpy(&serv.sin_addr, he->h_addr_list[0], sizeof(struct in_addr));

    result = connect(socketfd, (struct sockaddr *)&serv, sizeof(serv));
    if (result == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }


    NET_LOG("Successfully Connected to Host\n");
    *fd = socketfd;
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * networkRead
 *=------------------------------------------------------------------------=
 * this is a wrapper on the regular UNIX read() to support reading a line
 * over the network -- it does a bit of extra work to make sure that it
 * deals with interruptions and such correctly.
 *
 * Parameters:
 *      int                         - [in]  file descriptor
 *      void *                      - [in]  where to put read data.
 *      size_t                      - [in]  how much to read.
 *
 * Output:
 *      ssize_t                     - amount read
 */
ssize_t networkRead
(
    int fd,
    void *buf,
    size_t thisMany
)
{
    size_t remaining;
    ssize_t nread;
    char *ptr;

    ptr = (char *)buf;
    remaining = thisMany;
    while (remaining > 0) {

        nread = read(fd, ptr, remaining);
        if (nread < 0) {
            /**
             * If the read was interrupted, start it again.
             */
            if (errno == EINTR)
                nread = 0;
            else
                return -1;
        } else if (nread == 0) {
            // done!
            break;
        }

        remaining -= nread;
        ptr += nread;
    }

    return thisMany - remaining;
}
 

/**
 *=------------------------------------------------------------------------=
 * networkReadLine
 *=------------------------------------------------------------------------=
 * this routine reads a line of input from the given network file descriptor,
 * making sure it can handle things such as interrupted reads.
 *
 * Parameters:
 *      int                     - [in]  file descriptor
 *      void *                  - [in]  where to put the output.
 *      size_t                  - [in]  max size of a line.
 *
 * Output:
 *      ssize_t                 - amount read.
 */
ssize_t networkReadLine
(
    int fd,
    void *buf,
    size_t maxlen
)
{
    ssize_t x, res;
    char *ptr;
    char c;

    ptr = (char *)buf;

    /**
     * Now read in each char, one at a time from our helper routine (which
     * does some buffering) until we see a new line or end of input.
     */
    for (x = 0; x < (int)maxlen; x++) {
        res = _readLineHelper(fd, &c);
        if (res == 1) {
            /**
             * We read a character -- this is good. If we've seen a newline, 
             * then we can return!
             */
            *ptr++ = c;
            if (c == '\n') {
                x++;
                break;
            }
        } else if (res == 0) {
            /**
             * EOF.  Indicate if some data was read.
             */
            if (x == 1)
                return 0;
            else
                break;
        } else {
            /**
             * error -- errno is set
             */
            return -1;
        }
    }

    *ptr = '\0';
    return x;
}
 
/**
 *=------------------------------------------------------------------------=
 * networkSocketHasData
 *=------------------------------------------------------------------------=
 * indicates whether the given socket has any data to read.  We will wait
 * the specified amount of time.
 *
 * Parameters:
 *      int                         - [in]  network file descriptor
 *      int                         - [in]  ms to wait.
 *
 * Output:
 *      ERRCODE                     - S_OK - yes
 *                                    S_FALSE no
 *                                    E_* failure.
 */
ERRCODE networkSocketHasData
(
    int socketfd,
    int mswait
)
{
    struct timeval tv;
    int s, us, ret;
    fd_set fdset;

    FD_ZERO(&fdset);
    FD_SET(socketfd, &fdset);

    /**
     * Set up the time vals ...
     */
    s = mswait / 1000;
    us = (mswait % 1000) * 1000;
    tv.tv_sec = s;
    tv.tv_usec = us;

    /**
     * Finally, call select.
     */
    NET_LOG("Select()ing on Socket\n");
    ret = select(socketfd + 1, &fdset, NULL, NULL, &tv);
    NET_LOG("select() returned on socket\n");
    if (ret == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    return (ret == 0) ? S_FALSE : S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * networkRLSocketHasMoreData
 *=------------------------------------------------------------------------=
 * you need to call this to see if a socket has more data if you've been
 * using readLine, as it buffers some data.
 *
 * Parameters:
 *      int                         - [in]  network file descriptor
 *      int                         - [in]  ms to wait.
 *
 * Output:
 *      ERRCODE                     - S_OK - yes
 *                                    S_FALSE no
 *                                    E_* failure.
 */
ERRCODE networkRLSocketHasData
(
    int socketfd,
    int mswait
)
{
    if (read_count > 0) return S_OK;
    return networkSocketHasData(socketfd, mswait);
}


/**
 *=------------------------------------------------------------------------=
 * networkWrite
 *=------------------------------------------------------------------------=
 * writes the given data to the given file descriptor, making sure to handle
 * EINTR.
 *
 * Parameters:
 *      int                         - [in]  file descriptor
 *      const void *                - [in]  data to write.
 *      size_t                      - [in]  how much there is to write.
 *
 * Output:
 *      ssize_t                     - amount we wrote
 */
ssize_t networkWrite
(
    int fd,
    const void *writeMe,
    size_t writeThisMany
)
{
    size_t remaining;
    size_t nwritten;
    const char *ptr;

    ptr = (const char *)writeMe;
    remaining = writeThisMany;
    while (remaining > 0) {
        nwritten = write(fd, ptr, remaining);
        /**
         * Was it interrupted by a signal, or was there a bona-fide error.
         */
        if (nwritten <= 0) {
            if (errno == EINTR)
                nwritten = 0;
            else
                return -1;
        }

        remaining -= nwritten;
        ptr += nwritten;
    }

    return writeThisMany;
}


/**
 *=------------------------------------------------------------------------=
 * _readLineHelper
 *=------------------------------------------------------------------------=
 * this is a little helper for readline to speed it up a bit.  Again, this 
 * code is all snarfed from W. Richard Stevens's book.
 *
 * Parameters:
 *      int                     - [in]  fd.
 *      char *                  - [in]  character read
 *
 * Output:
 *      ssize_t
 */
static ssize_t _readLineHelper
(
    int fd,
    char *pchar
)
{
    /**
     * If there are no chars left, we need to go get a few more.
     */
    if (read_count <= 0) {
        
        /**
         * read until we get all the uninterrupted data, there is
         * no data left, or there is an error.
         */
        while (1) {
            read_count = read(fd, read_buf, sizeof(read_buf));
            if (read_count == 0)
                return 0;
            else if (read_count > 0) 
                break;

            if (errno == EINTR) continue;
            return -1;
        }

        read_ptr = read_buf;
    }

    read_count--;
    *pchar = *read_ptr++;
    return 1;
}


/**
 *=------------------------------------------------------------------------=
 * NetworkReader::NetworkReader
 *=------------------------------------------------------------------------=
 * this is a little helper class we've got to read data in over the network.
 * it's not terribly robust, but does return 0 when a socket would block 
 * i.e. -- there is no more data on it.
 *
 * Parameters:
 *      int                             - [in]  socket fd
 */
NetworkReader::NetworkReader
(
    int socketfd,
    int timeoutinMS
) : connectionfd(socketfd), timeout(timeoutinMS)
{
}

/**
 *=------------------------------------------------------------------------=
 * NetworkReader::readLine                                        [IReader]
 *=------------------------------------------------------------------------=
 * reads in a line from the network, returning 0 if there is no data available
 * on the socket.
 *
 * Parameters:
 *      void *                      - [in]  buffer to read into.
 *      unsigned int                - [in]  max num bytes to read.
 *      int &                       - [out] number of bytes actually read
 *                                      - 1 on error.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE NetworkReader::readLine
(
    void *buf,
    unsigned int maxRead,
    int &cread
)
{
    ssize_t x;

    if (!buf) return E_PARAM;
    ((char *)buf)[0] = '\0';
    cread = 0;

    if (networkRLSocketHasData(connectionfd, timeout) != S_OK) return S_OK;
    x = networkReadLine(connectionfd, buf, maxRead);
    if (x == -1) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    cread = x;
    return S_OK;
}

#if 0
/**
 *=------------------------------------------------------------------------=
 * NetworkReader::readLine                                        [IReader]
 *=------------------------------------------------------------------------=
 * reads in a line from the network, returning 0 if there is no data available
 * on the socket.  This version uses the timeout that the user specifies
 * not the default value on the class.
 *
 * Parameters:
 *      void *                      - [in]  buffer to read into.
 *      size_t                      - [in]  max num bytes to read.
 *      int                         - [in]  timeout in milliseconds.
 *
 * Output:
 *      ssize_t                     - num bytes READ
 */
ssize_t NetworkReader::readLine
(
    void *buf,
    size_t maxRead,
    int timeOut
)
{
    if (!buf) return 0;
    ((char *)buf)[0] = '\0';
    if (networkRLSocketHasData(connectionfd, timeOut) != S_OK) return 0;
    return networkReadLine(connectionfd, buf, maxRead);
}

#endif // 0


