/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <qcheckbox.h>
#include <qfiledialog.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qpushbutton.h>
#include <qradiobutton.h>

#include <intl.h>
#include "kiltdown.h"
#include "localized.h"
#include "mailaccount.h"
#include "main.h"
#include "newacctwiz.h"


#define SMTP_PORT                    25
#define POP3_PORT                   110
#define POP3_SECURE_PORT            995
#define DEFAULT_DELETE_AFTER          5
#define DEFAULT_SERVER_TIMEOUT       60


/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::NewAccountWizard
 *=------------------------------------------------------------------------=
 * initialize the dialog.  There's actually a lot of stuff that needs to
 * happen here.
 *
 * Parameters:
 *      - as per QDialog.
 */
NewAccountWizard::NewAccountWizard
(
    QWidget* parent,
    const char* name, 
    bool modal, 
    WFlags fl
) : AccountWizardDialog(parent, name, modal, fl)
{
    setHelpEnabled(welcomePage, false);
    setHelpEnabled(userNamePage, false);
    setHelpEnabled(emailAddressPage, false);
    setHelpEnabled(serverTypePage, false);
    setHelpEnabled(spoolServerPage, false);
    setHelpEnabled(popServerPage, false);
    setHelpEnabled(outgoingServerPage, false);
    setHelpEnabled(accountNamePage, false);

    setNextEnabled(welcomePage, true);
    setAppropriate(spoolServerPage, false);

    /**
     * personal data events.
     */
    QObject::connect(userNameE, SIGNAL(textChanged(const QString &)), this, SLOT(personalDataChanged(const QString &)));

    /**
     * email data events.
     */
    QObject::connect(userEmailE, SIGNAL(textChanged(const QString &)), this, SLOT(emailDataChanged(const QString &)));

    /**
     * Server type events.
     */
    QObject::connect(popServerB, SIGNAL(clicked()), this, SLOT(serverTypeChanged()));
    QObject::connect(spoolServerB, SIGNAL(clicked()), this, SLOT(serverTypeChanged()));

    /**
     * Spool data events.
     */
    QObject::connect(spoolPathE, SIGNAL(textChanged(const QString &)), this, SLOT(spoolDataChanged(const QString &)));
    QObject::connect(spoolPathP, SIGNAL(clicked()), this, SLOT(spoolPathClicked()));

    /**
     * Pop data changed.
     */
    QObject::connect(incomingPopE, SIGNAL(textChanged(const QString &)), this, SLOT(popDataChanged(const QString &)));
    QObject::connect(popUserNameE, SIGNAL(textChanged(const QString &)), this, SLOT(popDataChanged(const QString &)));

    /**
     * outgoing data changed.
     */
    QObject::connect(outgoingSmtpE, SIGNAL(textChanged(const QString &)), this, SLOT(outgoingDataChanged(const QString &)));

    /**
     * account name data.
     */
    QObject::connect(accountNameE, SIGNAL(textChanged(const QString &)), this, SLOT(nameDataChanged(const QString &)));

    /**
     * Finally, set it up so they have to enter some data.
     */
    setNextEnabled(userNamePage, false);
    setNextEnabled(emailAddressPage, false);
    setNextEnabled(spoolServerPage, false);
    setNextEnabled(popServerPage, false);
    setNextEnabled(outgoingServerPage, false);
    setNextEnabled(accountNamePage, false);

}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::~NewAccountWizard
 *=------------------------------------------------------------------------=
 * wake up -- time to die.
 */
NewAccountWizard::~NewAccountWizard()
{

}


/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::areYouSure
 *=------------------------------------------------------------------------=
 * make sure the user wants to terminate the wizard.
 *
 * Output:
 *      bool                    - true if yes, false if nope.
 */
bool NewAccountWizard::areYouSure()
{
    const char *title, *text;

    title = intlLoadString(ID_NAW_TERMTITLE_TEXT);
    if (!title) return true;
    text = intlLoadString(ID_NAW_TERMINATE_TEXT);
    if (!text) {
        localFree(title);
        return true;
    }

    int ret = QMessageBox::warning(this, title, text, 
                                   QMessageBox::Yes | QMessageBox::Default,
                                   QMessageBox::No | QMessageBox::Escape);

    localFree(title);
    localFree(text);
    if (ret == QMessageBox::Yes) return true;
    return false;
}



/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::closeEvent
 *=------------------------------------------------------------------------=
 * The user is closing the dialog using the sys menu/button .
 * We want to ask them and be sure they want to.
 *
 * Parameters:
 *      QCloseEvent *           - [in]  the event object describing this.
 */
void NewAccountWizard::closeEvent
(
    QCloseEvent *event
)
{
    if (!event) return;

    if (areYouSure()) {
        event->accept();
    } else 
        event->ignore();

}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::emailDataChanged
 *=------------------------------------------------------------------------=
 * data on the email page has changed.  Enable or disable the Next button 
 * as appropriate.
 *
 * Parameters:
 *      const QString &
 */
void NewAccountWizard::emailDataChanged(const QString &)
{
    if (userEmailE->text().isEmpty())
        setNextEnabled(emailAddressPage, false);
    else
        setNextEnabled(emailAddressPage, true);
}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::getAccountInformation
 *=------------------------------------------------------------------------=
 * given that we've run the dialog and are now done, we must go and figure
 * out what values the user input.  please note that all the strings placed
 * in the structure must be freed by the CALLER.
 *
 * Parameters:
 *      const char **               - [out] this account's name.
 *      MailAccountInfo &           - [in]  a struct in which to put info.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE NewAccountWizard::getAccountInformation
(
    const char *&accountName,
    MailAccountInfo &info
)
{
    memset(&info, 0, sizeof (MailAccountInfo));

    accountName = copyString(accountNameE->text());
    info.userName = copyString(userNameE->text());
    if (!userOrganizationE->text().isEmpty())
        info.userOrganization = copyString(userOrganizationE->text());
    info.userEmail = copyString(userEmailE->text());
    info.incomingPop = copyString(incomingPopE->text());
    info.popUserName = copyString(popUserNameE->text());
    if (!incomingPasswdE->text().isEmpty())
        info.popPassword = copyString(incomingPasswdE->text());
    info.outgoingSMTP = copyString(outgoingSmtpE->text());
    info.rememberPopPassword = rememberPasswordC->isChecked();

    info.popRemoveFromServer = true;
    info.popNoDuplicates = true;
    info.popDeleteAfterDays = DEFAULT_DELETE_AFTER;  // 5 days
    info.smtpPort = SMTP_PORT;                       // 25
    info.pop3Port = POP3_PORT;                       // 110
    info.securePop3Port = POP3_SECURE_PORT;          // 995
    info.serverTimeout = DEFAULT_SERVER_TIMEOUT;     // 60s

    /**
     * finally, do a quick error check.
     */
    if (!(accountName)
        || !info.userName
        || !info.userEmail
        || !info.incomingPop
        || !info.popUserName
        || !info.outgoingSMTP) {

        if (accountName) localFree(accountName); accountName = NULL;
        if (info.userName) localFree(info.userName);
        if (info.userOrganization) localFree(info.userOrganization);
        if (info.userEmail) localFree(info.userEmail);
        if (info.incomingPop) localFree(info.incomingPop);
        if (info.popUserName) localFree(info.popUserName);
        if (info.popPassword) localFree(info.popPassword);
        if (info.outgoingSMTP) localFree(info.outgoingSMTP);
        memset(&info, 0, sizeof(MailAccountInfo));
        return E_OUTOFMEMORY;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::nameDataChanged
 *=------------------------------------------------------------------------=
 * data on the account name   page has changed.  Enable or disable the Next 
 * button as appropriate.
 *
 * Parameters:
 *      const QString &
 */
void NewAccountWizard::nameDataChanged(const QString &)
{
    if (accountNameE->text().isEmpty())
        setFinishEnabled(accountNamePage, false);
    else
        setFinishEnabled(accountNamePage, true);
}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::outgoingDataChanged
 *=------------------------------------------------------------------------=
 * data on the  smtp  page has changed.  Enable or disable the Next button as
 * appropriate.
 *
 * Parameters:
 *      const QString &
 */
void NewAccountWizard::outgoingDataChanged(const QString &)
{
    if (outgoingSmtpE->text().isEmpty()) 
        setNextEnabled(outgoingServerPage, false);
    else
        setNextEnabled(outgoingServerPage, true);
}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::personalDataChanged
 *=------------------------------------------------------------------------=
 * data on the personal page has changed.  Enable or disable the Next button 
 * as appropriate.
 *
 * Parameters:
 *      const QString &
 */
void NewAccountWizard::personalDataChanged(const QString &)
{
    if (userNameE->text().isEmpty()) {
        setNextEnabled(userNamePage, false);
    } else {
        setNextEnabled(userNamePage, true);
    }
}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::popDataChanged
 *=------------------------------------------------------------------------=
 * data on the pop page has changed.  Enable or disable the Next button as
 * appropriate.
 *
 * Parameters:
 *      const QString &
 */
void NewAccountWizard::popDataChanged(const QString &)
{
    if (incomingPopE->text().isEmpty() || popUserNameE->text().isEmpty())
        setNextEnabled(popServerPage, false);
    else
        setNextEnabled(popServerPage, true);
}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::reject
 *=------------------------------------------------------------------------=
 * The user is aborting the dialog.  If they are closing it because they are
 * cancelling, then we want to ask them and be sure they want to.
 *
 * Parameters:
 *      QCloseEvent *           - [in]  the event object describing this.
 */
void NewAccountWizard::reject()
{
    if (areYouSure()) 
        AccountWizardDialog::reject();
}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::serverTypeChanged
 *=------------------------------------------------------------------------=
 * the server type has changed.  Enable and disable pages as appropriate.
 */
void NewAccountWizard::serverTypeChanged()
{
    if (popServerB->isChecked()) {
        setAppropriate(spoolServerPage, false);
        setAppropriate(popServerPage, true);
    } else {
        setAppropriate(popServerPage, false);
        setAppropriate(spoolServerPage, true);
    }
}

/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::spoolDataChanged
 *=------------------------------------------------------------------------=
 * data on the spool page has changed.  Enable or disable the Next button as
 * appropriate.
 *
 * Parameters:
 *      const QString &
 */
void NewAccountWizard::spoolDataChanged(const QString &)
{
    if (spoolPathE->text().isEmpty())
        setNextEnabled(spoolServerPage, false);
    else
        setNextEnabled(spoolServerPage, true);
}


/**
 *=------------------------------------------------------------------------=
 * NewAccountWizard::spoolPathClicked
 *=------------------------------------------------------------------------=
 * the user wants to use a dialog to choose the spool path.
 */
void NewAccountWizard::spoolPathClicked()
{
    QString file;

    file = QFileDialog::getOpenFileName(spoolPathE->text());

    if (file) spoolPathE->setText(file);
}


