/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <sanity.h>

#include <qcheckbox.h>
#include <qlineedit.h>
#include <qhbox.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qvbox.h>

#include "passworddlg.h"
#include <intl.h>
#include "localized.h"


/**
 *=------------------------------------------------------------------------=
 *                               PasswordDialog                            *
 *=------------------------------------------------------------------------=
 */



/**
 *=------------------------------------------------------------------------=
 * PasswordDialog::PasswordDialog
 *=------------------------------------------------------------------------=
 * initializes everything but the UI stuff.
 */
PasswordDialog::PasswordDialog()
: QDialog(0,0, true, WStyle_NoBorder)
{
    name = password = NULL;
    nameLabel = passwordLabel = NULL;
    vbox = NULL;
    hbox = NULL;
}


/**
 *=------------------------------------------------------------------------=
 * PasswordDialog::~PasswordDialog
 *=------------------------------------------------------------------------=
 * cleans it all up baby.
 */
PasswordDialog::~PasswordDialog()
{
    if (name) delete name;
    if (password) delete password;
    if (nameLabel) delete nameLabel;
    if (passwordLabel) delete passwordLabel;
    if (ok) delete ok;
    if (cancel) delete cancel;
    if (hbox) delete hbox;
    if (checkBox) delete checkBox;
}

/**
 *=------------------------------------------------------------------------=
 * PasswordDialog::getInput
 *=------------------------------------------------------------------------=
 * returns the information that the user entered.
 *
 * Parameters:
 *      const char * &          - [out] user name
 *      const char * &          - [out] password
 *      bool &                  - [out] remember password?
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PasswordDialog::getInput
(
    const char * &newName,
    const char * &newPassword,
    bool & remember
)
{
    newName = newPassword = NULL;

    if (!this->name->text().isEmpty()) {
        newName = copyString((const char *)this->name->text());
        if (!newName) return E_OUTOFMEMORY;
    }
    if (!this->password->text().isEmpty()) {
        newPassword = copyString((const char *)this->password->text());
        if (!newPassword) {
            if (newName) localFree(newName);
            return E_OUTOFMEMORY;
        }
    }
    remember = this->checkBox->isChecked();

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * PasswordDialog::initialize
 *=------------------------------------------------------------------------=
 * initializes our UI
 *
 * Parameters:
 *      const char *            - [in]  account
 *      const char *            - [in]  default user name
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PasswordDialog::initialize
(
    const char *account,
    const char *uname
)
{
    const char *pstr = NULL;
    ERRCODE er = S_OK;
    char buf[128];

    /**
     * Create a vbox to flow things vertically.
     */
    vbox = new QVBox(this);
    if (!vbox) return E_OUTOFMEMORY;
    vbox->setSpacing(5);
    vbox->setMargin(5);


    pstr = intlLoadString(ID_ENTER_NAME_PWD);
    if (!pstr) return E_OUTOFMEMORY;
    infoLabel = new QLabel(vbox);
    if (!infoLabel) { er = E_OUTOFMEMORY; goto Error; }
    infoLabel->setText(pstr);
    localFree(pstr);
    infoLabel->setAlignment(AlignCenter);
    infoLabel->adjustSize();

    pstr = intlLoadString(ID_FOR_ACCOUNT);
    if (!pstr) return E_OUTOFMEMORY;
    snprintf(buf, sizeof(buf), pstr, account);
    localFree(pstr); pstr = NULL;
    accountLabel = new QLabel(vbox);
    accountLabel->setText(buf);
    accountLabel->setAlignment(AlignCenter);

    /**
     * Now the two labels with their associated edit boxes.
     */    
    pstr = intlLoadString(ID_USER_NAME);
    if (!pstr) goto Error;
    nameLabel = new QLabel(vbox);
    if (!nameLabel) { er = E_OUTOFMEMORY; goto Error; }
    nameLabel->setText(pstr);
    localFree(pstr); pstr = NULL;

    name = new QLineEdit(vbox);
    if (!name) { er = E_OUTOFMEMORY; goto Error; }
    name->setText(uname);

    pstr = intlLoadString(ID_USER_PASSWD);
    if (!pstr) goto Error;
    passwordLabel = new QLabel(vbox);
    if (!passwordLabel) { er = E_OUTOFMEMORY; goto Error; }
    passwordLabel->setText(pstr);
    localFree(pstr); pstr = NULL;

    password = new QLineEdit(vbox);
    if (!password) { er = E_OUTOFMEMORY; goto Error; }
    password->setEchoMode(QLineEdit::Password);

    pstr = intlLoadString(ID_REMEMBER_PASSWORD);
    if (!pstr) { er = E_OUTOFMEMORY; goto Error; }
    checkBox = new QCheckBox(vbox);
    if (!checkBox) { er = E_OUTOFMEMORY; goto Error; }
    checkBox->setText(pstr);
    localFree(pstr); pstr = NULL;

    /**
     * Now create a horizontally flowed box to hold the two command buttons.
     */
    hbox = new QHBox(vbox);
    if (!hbox) { er = E_OUTOFMEMORY; goto Error; }
    hbox->setSpacing(15);
    hbox->setMargin(5);

    ok = new QPushButton(hbox);
    if (!ok) { er = E_OUTOFMEMORY; goto Error; }
    pstr = intlLoadString(ID_OK_TEXT);
    if (!pstr) return E_OUTOFMEMORY;
    ok->setText(pstr);
    ok->setDefault(true);
    localFree(pstr); pstr = NULL;

    cancel = new QPushButton(hbox);
    if (!cancel) { er = E_OUTOFMEMORY; goto Error; }
    pstr = intlLoadString(ID_CANCEL_TEXT);
    if (!pstr) return E_OUTOFMEMORY;
    cancel->setText(pstr);
    localFree(pstr); pstr = NULL;

    /**
     * Hook up signals
     */
    QObject::connect(ok, SIGNAL(clicked()), this, SLOT(accept()));
    QObject::connect(cancel, SIGNAL(clicked()), this, SLOT(reject()));

    /**
     * finally, set up some of our own properties, some size, and focus
     */
    vbox->setGeometry(0, 0, infoLabel->width() + 10, 230);
    this->setBaseSize(QSize(300, 100));
    this->password->setFocus();

    er = S_OK;

  Error:
    if (pstr) localFree(pstr);
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * PasswordDialog::resizeEvent
 *=------------------------------------------------------------------------=
 * we need to know when our size has changed so we can resize the components
 * as appropriate
 *
 * Parameters:
 *      QResizeEvent *              - [in] our new size.
 */
void PasswordDialog::resizeEvent
(
    QResizeEvent *event
)
{
    if (!event) return;
    QSize size = event->size();

    vbox->setGeometry(0, 0, size.width(), size.height());
}



