/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <sanity.h>

#include <ctype.h>
#include <errno.h>
#include <unistd.h>


#include "kdispatch.h"
#include "mailmessage.h"
#include "netapis.h"
#include "pop.h"
#include "userpref.h"

//
// UNDONE: marcw, 11.26.2000 -- all of the current network reads will block if
//      the server crashes.  put a select() in front of all of them!!!!
//


/**
 * Defines the routines for getting the mail messages via POP3.
 */

/**
 *=------------------------------------------------------------------------=
 * PopConnection::PopConnection
 *=------------------------------------------------------------------------= 
 * initializes our object, using the given network timeout.
 *
 * Parameters:
 *      int                     - [in]  timeout for operations, in ms.
 */
PopConnection::PopConnection
(
    int networkTimeoutInMS
)
{
    networkTimeout = networkTimeoutInMS;
    connectionfd = -1;
    messageMax = -1;
}

/**
 *=------------------------------------------------------------------------=
 * PopConnection::~PopConnection
 *=------------------------------------------------------------------------=
 * cleans up as necessary.
 */
PopConnection::~PopConnection()
{
    disconnect();
}

/**
 *=------------------------------------------------------------------------=
 * PopConnection::connect
 *=------------------------------------------------------------------------=
 * establishes a connection with the POP3 server on the given port.  Will 
 * return the network file descripter.
 *
 * Parameters:
 *      const char *            - [in]  server host name/IP address
 *      int                     - [in]  port, typically 110.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PopConnection::connect
(
    const char *hostName,
    int port
)
{
    ERRCODE er;

    /**
     * First, try to make the connection.
     */
    er = networkConnect(hostName, port, &connectionfd);
    RETURN_ON_FAILURE(er);

    /**
     * Make sure the server sent us back a +OK.
     */
    er = readResult();
    RETURN_ON_FAILURE(er);

    if (er == S_FALSE) return E_POP_CONNECTION;
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * PopConnection::deleteMessage
 *=------------------------------------------------------------------------=
 * removes the given message from the server.
 *
 * Parameters:
 *      int                 - [in]  index (1 based) of message to delete.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PopConnection::deleteMessage
(
    int index
)
{
    char buf[POP_MAXLINELENGTH];
    ERRCODE er;
    int wrote;

    snprintf(buf, sizeof(buf), "DELE %d\r\n", index);
    wrote = networkWrite(connectionfd, buf, strlen(buf));
    if (wrote < 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * Make sure the server sent us back a +OK.
     */
    er = readResult();
    RETURN_ON_FAILURE(er);
    if (er == S_FALSE) return E_POP_DELETE;

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * PopConnection::disconnect
 *=------------------------------------------------------------------------=
 * hangs-up if necessary.
/ */
void PopConnection::disconnect()
{
    /**
     * If we have an active connection, then we need to terminate it.
     */
    if (connectionfd != -1) {
        networkWrite(connectionfd, "QUIT\r\n", sizeof("QUIT\r\n") - 1);
        close(connectionfd);
    }
    connectionfd = -1;
}

/**
 *=------------------------------------------------------------------------=
 * PopConnection::downloadMessage
 *=------------------------------------------------------------------------=
 * downloads the given message, and creates a MailMessage object for it.
 * very exciting.
 *
 * Parameters:
 *      int                     - [in]  which message?
 *      bool                    - [in] delete me???
 *      MailMessage **          - [out] the new message.
 *
 * Output:
 *      ERRCODE                 - all sorts of stuff can go wrong!!!
 */
ERRCODE PopConnection::downloadMessage
(
    int cmessage,
    bool deleteMe,
    MailMessage **message
)
{
    char buf[POP_MAXLINELENGTH];
    ssize_t wrote;
    ERRCODE er;

    if (!message || (cmessage <= 0) || (cmessage > messageMax)) return E_PARAM;

    snprintf(buf, sizeof(buf), "RETR %d\r\n", cmessage);
    wrote = networkWrite(connectionfd, buf, strlen(buf));
    if (wrote <= 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    er = readResult();
    RETURN_ON_FAILURE(er);
    if (er == S_FALSE) {
        fprintf(stderr, "Unexpected reply from RETR command.\n");
        return E_NET_UNEXPECTED;
    }

    /** 
     * Create a network streaming reading object, and pass it off to the
     * MailMessage class to try and instantiate the message.
     */
    NetworkReader nr(this->connectionfd, networkTimeout);
    er = MailMessage::createFromNetwork(&nr, message);
    RETURN_ON_FAILURE(er);

    /**
     * Finally, Delete the message on the server.  If there are errors
     * with this account, we will abort and not delete any of them.
     */
    if (deleteMe) {
        er = deleteMessage(cmessage);
        RETURN_ON_FAILURE(er);
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * PopConnection::getMessageInfo
 *=------------------------------------------------------------------------=
 * returns the number of messages waiting for this user as well as the total
 * size of all the messages in bytes.
 *
 * Parameters:
 *      int *                   - [out] number of messages waiting.
 *      int *                   - [out] total size of messages.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PopConnection::getMessageInfo
(
    int *cmessages,
    int *totalsize
)
{
    char buf[POP_MAXLINELENGTH];
    ssize_t wrote;
    ERRCODE er;
    int ret, i;

    if (!cmessages || !totalsize) return E_PARAM;
    if (connectionfd == -1) return E_NOT_CONNECTED;

    /**
     * Send out a STAT to the server to get the details.
     */
    snprintf(buf, sizeof(buf), "STAT\r\n");
    wrote = networkWrite(connectionfd, buf, strlen(buf));
    if (wrote <= 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }


    /** 
     * See what they've got.  Is there any data waiting for us?
     */
    er = networkSocketHasData(connectionfd, networkTimeout);
    RETURN_ON_FAILURE(er);
    if (er == S_FALSE) return E_NET_UNEXPECTED;

    /**
     * Yes?  download it now.
     */
    wrote = networkReadLine(connectionfd, buf, sizeof(buf));
    if (wrote <= 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * We are expecting a result in the form of +OK <count> <size>.  get
     * that, skipping over whitespace at the begining.
     */
    i = 0;
    while (buf[i] != '\0' && isspace(buf[i])) i++;
    ret = sscanf(buf + i, "+OK %d %d", cmessages, totalsize);
    if (ret != 2) return E_NET_UNEXPECTED;
    messageMax = *cmessages;

    /**
     * Otherwise, we're done.
     */
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * PopConnection::login
 *=------------------------------------------------------------------------=
 * performs a login to the server to which we are connected.  Will return an
 * error if we are not connected to a server.
 *
 * Parameters:
 *      const char *            - [in]  user name string.
 *      const char *            - [in]  password string (encrypted or not)
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PopConnection::login
(
    const char *userName,
    const char *password,
    enum PopLoginType type
)
{
    char buf[POP_MAXLINELENGTH];
    ssize_t wrote;
    ERRCODE er;

    if (!userName) return E_USERNAME;
    if (type != PLAIN_TEXT) return E_PARAM;
    if (connectionfd == -1) return E_NOT_CONNECTED;

    /**
     * First thing we need to do is send the user name.
     */
    snprintf(buf, sizeof(buf), "USER %s\r\n", userName);
    wrote = networkWrite(connectionfd, buf, strlen(buf));
    if (wrote <= 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * Make sure the server liked the name.
     */
    er = readResult();
    RETURN_ON_FAILURE(er);
    if (er == S_FALSE) {
        disconnect();
        return E_USERNAME;
    }

    /**
     * Now set up the password.
     */
    snprintf(buf, sizeof(buf), "PASS %s\r\n", password ? password : "");
    wrote = networkWrite(connectionfd, buf, strlen(buf));
    if (wrote <= 0) {
        errorSaveRichErrorInfo(strerror(errno));
        return E_NET_ERR;
    }

    /**
     * Make sure the server liked the password.
     */
    er = readResult();
    RETURN_ON_FAILURE(er);

    if (er == S_FALSE) {
        disconnect();
        return E_PASSWORD;
    }

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * PopConnection::readResult
 *=------------------------------------------------------------------------=
 * reads in the result from the POP server and figures out whether it starts
 * with +OK or -ERR.
 *
 * Output:
 *      ERRCODE                 - S_OK -- all went well.  +OK
 *                                S_FALSE -- -ERR was returned.
 *                                E_* -- couldn't read from the network.
 */
ERRCODE PopConnection::readResult()
{
    char buf[POP_MAXLINELENGTH];
    ERRCODE er;
    int i;

    /** 
     * See what they've got -- waiting up to a second.
     */
    er = networkSocketHasData(connectionfd, networkTimeout);
    RETURN_ON_FAILURE(er);

    /**
     * No?, then this is easy.
     */
    if (er == S_FALSE) return E_NET_UNEXPECTED;

    /**
     * Yes?  download it now.
     */
    networkReadLine(connectionfd, buf, sizeof(buf));

    /**
     * Figure out if it starts with +OK or -ERR
     */
    i = 0;
    while (buf[i] && isspace(buf[i]))
        i++;

    if (buf[i] == '+') {
        if (buf[i + 1] != '\0' && (buf[i + 1] == 'O' || buf[i + 1] == 'o'))
            if (buf[i + 2] != '\0' && (buf[i + 2] == 'K' || buf[i + 2] == 'k'))
                return S_OK;
    } else if (buf[i] == '-') {
        if (buf[i + 1] != '\0' && (buf[i + 1] == 'E' || buf[i + 1] == 'e'))
            if (buf[i + 2] != '\0' && (buf[i + 2] == 'R' || buf[i + 2] == 'r'))
                if (buf[i + 3] != '\0' && (buf[i + 3] == 'R' || buf[i + 3] == 'r'))
                    return S_FALSE;
    }

    return E_NET_UNEXPECTED;
}


