/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <properties.h>

#include <ctype.h>
#include <errno.h>
#include <lerror.h>
#include <stdlib.h>
#include <string.h>


/** 
 * Implementation of our properties file class.  Nothing terribly shocking here.
 */
const char *trueString = "true";
const char *falseString = "false";

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::openPropertiesFile
 *=------------------------------------------------------------------------=
 * Opens the given properties file.  If necessary, creates it.
 *
 * Parameters:
 *      const char *            - [in]  Filename
 *      PropertiesFile **       - [out] the new object with which to work.
 *      bool                    - [in]  indicates whether the file should be
 *                                  created if it doesn't exist.
 *      int                     - [in]  umask to create file with.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PropertiesFile::openPropertiesFile
(
    const char *filename,
    PropertiesFile **pproperties,
    bool createIfNecessary
)
{
    char buf[2 * MAX_NAMEVALUE_LEN + 2];
    NameValuePair *nvp = NULL;
    ERRCODE er;
    FILE *file = NULL;
    int cread, bufIndex;

    if (!filename || !pproperties) return E_PARAM;
    *pproperties = new PropertiesFile();
    if (!(*pproperties)) return E_OUTOFMEMORY;

    /**
     * Save out the filename for when we write out the file again.
     */
    (*pproperties)->fileName = copyString(filename);
    if (!(*pproperties)->fileName) return E_OUTOFMEMORY;

    /**
     * try to open the file for reading.  If we can, then there is cosmic
     * harmony and happiness.
     */
    file = fopen(filename, "r");
    if (!file && !createIfNecessary) {
        er = E_FILEERR;
        goto Error;
    }


    /**
     * if there is no file/data yet, this is easy.
     */
    if (!file) {
        er = S_CREATED;
        goto Error;
    }

    /**
     * Otherwise, loop through the file one line at a time, parsing out the data
     */
    while (true) {

        if (!fgets(buf, sizeof(buf), file)) break;

        /** 
         * If it's a comment, then just save it out.
         */
        if (buf[0] == '#') {
            nvp = new NameValuePair();
            if (!nvp) {
                er = E_OUTOFMEMORY;
                goto Error;
            }
            nvp->comment = copyString(buf);
            if (!nvp->comment) {
                er = E_OUTOFMEMORY;
                goto Error;
            }
            (*pproperties)->keyList.addElement(nvp);        
            nvp = NULL;

        } else if (buf[0] == '\0' || buf[0] == '\n') {
            /**
             * If it's an empty line, just ignore it.
             */
            ;
        } else {
            nvp = new NameValuePair();
            if (!nvp) {
                er = E_OUTOFMEMORY;
                goto Error;
            }
            memset(nvp, 0, sizeof(NameValuePair));

            /**
             * whip through the line looking for the equals sign.
             */
            bufIndex = 0;
            while (buf[bufIndex] != '\n' && buf[bufIndex] != '\0') {

                if (buf[bufIndex] == '=') {
                    buf[bufIndex] = '\0';

                    /**
                     * Skip over whitespace before the =
                     */
                    cread = bufIndex - 1;
                    while (isspace(buf[cread]) && cread >= 0) {
                        cread--;
                    }
                    buf[cread + 1] = '\0';

                    /**
                     * Copy the string out and save it.
                     */
                    nvp->name = copyString(buf);
                    if (!nvp->name) {
                        er = E_OUTOFMEMORY;
                        goto Error;
                    }
                    break;
                }
                bufIndex++;
            }
            
            if (nvp->name == NULL) {
                er = E_INVALIDFORMAT;
                goto Error;
            }

            /**
             * now get the value, skipping over whitespace first and removing the
             * newline next.
             */
            cread = bufIndex + 1;
            while (isspace(buf[cread]) && buf[cread] != '\0')
                cread++;

            bufIndex = cread;
            while (buf[bufIndex] != '\0') {
                if (buf[bufIndex] == '\n')
                    buf[bufIndex] = '\0';
                bufIndex++;
            }
            nvp->value = copyString(buf + cread);
            if (!nvp->value) {
                localFree(nvp->name);
                er = E_OUTOFMEMORY;
                goto Error;
            }
            (*pproperties)->keyList.addElement(nvp);
            nvp = NULL;
        }
    }


    if (feof(file)) 
        er = S_OK;
    else
        er = E_FILEERR;

  Error:
    /**
     * Clean up and go home!
     */
    if (file) fclose(file);
    if (FAILED(er)) {
        if (*pproperties) delete *pproperties;
        *pproperties = NULL;
        if (nvp) delete nvp;
    }

    return er;
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::PropertiesFile
 *=------------------------------------------------------------------------=
 * initializes everything so that there is cosmic harmony and happiness.
 */
PropertiesFile::PropertiesFile()
{

}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::~PropertiesFile
 *=------------------------------------------------------------------------=
 * Time to take out the trash!!!
 */
PropertiesFile::~PropertiesFile()
{
    NameValuePair *nvp;
    ERRCODE er;

    er = keyList.firstElement(&nvp);
    while (er == S_OK) {
        if (nvp->name) localFree(nvp->name);
        if (nvp->value) localFree(nvp->value);
        if (nvp->comment) localFree(nvp->comment);
        delete nvp;

        er = keyList.nextElement(&nvp);
    }

    if (fileName) localFree(fileName);

}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::findProperty
 *=------------------------------------------------------------------------=
 * finds the namevaluepair associated with the given property name, or
 * returns NULL.
 *
 * Parameters:
 *      const char *            - [in]  name to look for.
 *
 * Output:
 *      NameValuePair *         - resulting pair or NULL if not found.
*/
NameValuePair *PropertiesFile::findProperty
(
    const char *name
)
{
    NameValuePair *nvp;
    ERRCODE er;

    /**
     * Find the property name.
     */
    er = keyList.firstElement(&nvp);
    while (er == S_OK) {

        if (nvp->name && nvp->name[0] && !strcasecmp(name, nvp->name)) {
            return nvp;
        }

        er = keyList.nextElement(&nvp);
    }

    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::close
 *=------------------------------------------------------------------------=
 * Writes out the file and wipes out our internal list.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PropertiesFile::close()
{
    // UNDONE: marcw, 10.12.2000 -- i suppose i should nuke the list here
    //  somewhere...
    //
    return sync();
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::getBooleanProperty
 *=------------------------------------------------------------------------=
 * Gets the value of the given boolean property.
 *
 * Parameters:
 *      const char *            - [in]  name of property
 *      bool *                  - [out] value of the property
 *
 * Output:
 *      ERRCODE                 - S_NODATA if not there, S_OK on success, E_
 *                                  on fatal problem.
 */
ERRCODE PropertiesFile::getBooleanProperty
(
    const char *name,
    bool *pvalue
)
{
    NameValuePair *nvp;

    if (!name || !pvalue) return E_PARAM;

    /**
     * If the property isn't there, this is easy!
     */
    *pvalue = false;
    nvp = findProperty(name);
    if (!nvp) return S_NODATA;

    if (nvp->value) {
        if (!strcasecmp(nvp->value, trueString))
            *pvalue = true;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::getDoubleProperty
 *=------------------------------------------------------------------------=
 * Gets the value of the given double property.
 *
 * Parameters:
 *      const char *            - [in]  name of property
 *      double *                - [out] the double.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PropertiesFile::getDoubleProperty
(
    const char *name,
    double *pvalue
)
{
    NameValuePair *nvp;

    if (!name || !pvalue) return E_PARAM;

    /**
     * If the property isn't there, this is easy!
     */
    *pvalue = 0.0;
    nvp = findProperty(name);
    if (!nvp) return S_NODATA;

    if (nvp->value)
        *pvalue = strtod(nvp->value, NULL);

    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::getIntProperty
 *=------------------------------------------------------------------------=
 * Gets the value of the given integer property.
 *
 * Parameters:
 *      const char *            - [in]  name of property
 *      int *                   - [out] value of the property
 *
 * Output:
 *      ERRCODE                 - S_NODATA if not there, S_OK on success, E_
 *                                  on fatal problem.
 */
ERRCODE PropertiesFile::getIntProperty
(
    const char *name,
    int *pvalue
)
{
    NameValuePair *nvp;

    if (!name || !pvalue) return E_PARAM;

    /**
     * If the property isn't there, this is easy!
     */
    *pvalue = 0;
    nvp = findProperty(name);
    if (!nvp) return S_NODATA;

    if (nvp->value)
        *pvalue = atoi(nvp->value);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::getStringProperty
 *=------------------------------------------------------------------------=
 * Gets the value of the given string property.  The value returned is
 * newly allocated when non-null, and the CALLER is responsible for calling
 * localFree() on it.
 *
 * Parameters:
 *      const char *            - [in]  name of property
 *      const char **           - [out] value of the property, or NULL if none
 *
 * Output:
 *      ERRCODE                 - S_NODATA if not there, S_OK on success, E_
 *                                  on fatal problem.
 */
ERRCODE PropertiesFile::getStringProperty
(
    const char *name,
    const char **pvalue
)
{
    NameValuePair *nvp;

    if (!name || !pvalue) return E_PARAM;

    /**
     * If the property isn't there, this is easy!
     */
    *pvalue = NULL;
    nvp = findProperty(name);
    if (!nvp) return S_NODATA;

    /**
     * If it IS there, then return the data to the user.
     */
    if (nvp->value) {
        *pvalue = copyString(nvp->value);
        if (!(*pvalue)) return E_OUTOFMEMORY;
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::sync
 *=------------------------------------------------------------------------=
 * Makes sure that the file is up to date with our data.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PropertiesFile::sync()
{
    char buf[2 * MAX_NAMEVALUE_LEN + 2];
    NameValuePair *nvp;
    ERRCODE er;
    FILE *file;

    if (!fileName) return E_INTERNAL;

    /**
     * Open the file for writing, truncating it to zero if it already exists.
     */
    file = fopen(fileName, "w");
    if (!file) return E_FILEERR;

    /**
     * Okay now go and construct the strings and write them out one at a time.
     */
    er = keyList.firstElement(&nvp);
    while (er == S_OK) {
        if (nvp->value) {
            sprintf(buf, "%s=%s\n", nvp->name, nvp->value);
            fputs(buf, file);
        }

        er = keyList.nextElement(&nvp);
    }

    fclose(file);
    return S_OK;
}


/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::writeBooleanProperty
 *=------------------------------------------------------------------------=
 * Adds the given boolean property to the list.  If the property name already
 * exists, the value is replaced and S_WARNING is returned.
 *
 * Parameters:
 *      const char *            - [in]  name of the property
 *      bool                    - [in]  value of the property
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PropertiesFile::writeBooleanProperty
(
    const char *name,
    bool value
)
{

    return writeStringProperty(name, value ? trueString : falseString);
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::writeDoubleProperty
 *=------------------------------------------------------------------------=
 * Adds the given double property to the list.  If the property name already
 * exists, the value is replaced and S_WARNING is returned.
 *
 * Parameters:
 *      const char *            - [in] property name.
 *      double                  - [in] it's value.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PropertiesFile::writeDoubleProperty
(
    const char *name,
    double value
)
{
    char buf[64];
    sprintf(buf, "%f", value);

    return writeStringProperty(name, buf);
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::writeIntProperty
 *=------------------------------------------------------------------------=
 * Adds the given integer property to the list.  If the property name already
 * exists, the value is replaced and S_WARNING is returned.
 *
 * Parameters:
 *      const char *            - [in]  name of the property
 *      int                     - [in]  value of the property
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PropertiesFile::writeIntProperty
(
    const char *name,
    int value
)
{
    char buf[32];

    sprintf(buf, "%d", value);

    return writeStringProperty(name, buf);
}

/**
 *=------------------------------------------------------------------------=
 * PropertiesFile::writeStringProperty
 *=------------------------------------------------------------------------=
 * Adds the given string property to the list.  If the property name already
 * exists, the value is replaced and S_WARNING is returned.  A NULL value
 * removes the key from the list.
 *
 * Parameters:
 *      const char *            - [in]  name of the property
 *      const char *            - [in]  value of the property
 *
 * Output:
 *      ERRCODE
 */
ERRCODE PropertiesFile::writeStringProperty
(
    const char *name,
    const char *value
)
{
    NameValuePair *nvp;

    if (!name) return E_PARAM;

    /**
     * if the name already exists, replace it's value.
     */
    nvp = findProperty(name);
    if (nvp) {
        if (value) {
            if (nvp->value)
                localFree(nvp->value);
            nvp->value = copyString(value);
            if (!nvp->value) return E_OUTOFMEMORY;
        } else {
            if (nvp->value)
                localFree(nvp->value);
            nvp->value = NULL;
        }

        return S_WARNING;
    }

    /**
     * okay, value wasn't found.  add it now:
     */
    nvp = new NameValuePair();
    if (!nvp) return E_OUTOFMEMORY;
    memset(nvp, 0, sizeof(NameValuePair));

    nvp->name = copyString(name);
    if (!nvp->name) return E_OUTOFMEMORY;
    if (value) {
        nvp->value = copyString(value);
        if (!nvp->value) return E_OUTOFMEMORY;
    }

    return keyList.addElement(nvp);
}


