/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>

#include "deencode.h"
#include "mailmessage.h"
#include "mime.h"
#include "netapis.h"

/**
 * This simple table is used for quoted printable stuff.
 */
char quotedPrintableValues [] = {
    '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
};

#define IS_QP_HEXA(c)        (((c) >= '0' && (c) <= '9') || (toupper((c)) >= 'A' && toupper((c)) <= 'F'))

/**
 * Local Routines
 */
char qpHexaDecode(const char *);



/**
 *=------------------------------------------------------------------------=
 * quotedPrintableDecode
 *=------------------------------------------------------------------------=
 * given a stream of characters that is in quoted printable format, get
 * all the text, decoding the quoted printable stuff as we go along.
 *
 * Parameters:
 *      const char *                - [in]  terminating separator.
 *      int *                       - [out] the size of the attachment.
 *      const char **               - [out] the decoded section.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE quotedPrintableDecode
(
    const char *separator,
    int *size,
    const char **body
)
{
    char inputBuf[NET_MAXLINELEN];
    int currentSize, cread;
    bool lastWasEqual;
    BYTE *buf = NULL;
    int bufSize = 0;
    int seplen = 0;
    ERRCODE er;
    int len;
    int x;

    if (!body || !separator) return E_PARAM;
    seplen = strlen(separator);

    /**
     * this is easy -- while we're not at the separating line, add in the
     * text that we're seeing.  E-Zed.
     */
    currentSize = 0;
    while (1) {
        mimeReader.readLine(inputBuf, sizeof(inputBuf), cread);

        /**
         * First see if this line is the terminating separator.  If
         * so, then break out of the loop and call ourselves done !!!
         */
        if (separator && !strncasecmp(inputBuf, separator, seplen)) {
            break;
        } else if (inputBuf[0] == '\0') {
            break;
        }

        /**
         * Otherwise, process the text.  First, make sure there is enough room
         * to hold the new string.
         */
        len = strlen(inputBuf);
        if (!ensureBufferSize((char **)&buf, &bufSize, currentSize + len, BUFFER_INCREMENT)) {
            er = E_OUTOFMEMORY;
            goto Error;
        }

        /**
         * Whiz through the input looking for = characters.
         */
        x = 0;
        lastWasEqual = false;       
        while (inputBuf[x] != '\0') {

            /**
             * If it's an equals sign, then everything gets a little wacky,
             * as we need to figure out if:
             *
             * a. it's got two hexadecimal characters afterwards
             * b. it's got only whitespace and then CRLF
             * c. it's got neither, in which case it's bogus, and we'll just
             *    treat it as poorly formed input
             */
            if (inputBuf[x] == '=') {

                if (IS_QP_HEXA(inputBuf[x + 1])
                    && IS_QP_HEXA(inputBuf[x + 2])) {

                    // a.
                    buf[currentSize++] = qpHexaDecode(inputBuf + x + 1);
                    x += 3;

                } else if (isspace(inputBuf[x + 1])) {

                    // b.
                    int z = x + 1;
                    while (inputBuf[z] && isspace(inputBuf[z]))
                        z++;
                    /**
                     * if it was a soft line break, then don't copy over
                     * the '=' or the newline, and we're done with this
                     * line !!!
                     */
                    if (inputBuf[z] == '\0') {
                        x = z;
                        continue;
                    } else {
                        // c.
                        buf[currentSize++] = inputBuf[x++];
                    }

                } else {
                    // c.
                    buf[currentSize++] = inputBuf[x++];
                }
            } else {
                buf[currentSize++] = inputBuf[x++];
            }

        } // while inputBuf[x] != 0

        buf[currentSize] = '\0';

    }  // while 1


    *body = (const char *)buf;
    *size = currentSize;
    er = S_OK;

  Error:
    if (FAILED(er)) { if (buf) localFree(buf); }
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * quotedPrintableEncode
 *=------------------------------------------------------------------------=
 * given some text, write it out as seven bit text.  Very complex.
 *
 * Parameters:
 *      int                         - [in]  network stream to write to.
 *      int                         - [in]  don't let lines be longer than 
 *                                      this (7bit/q-p only)
 *      const char *                - [in]  data to write out
 *      int                         - [in]  size of said data.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE quotedPrintableEncode
(
    int socketfd,
    int lineLength,
    const char *body,
    int size
)
{
    int result;

    // UNDONE: marcw, 12.12.2000 implement me.
    K_FAIL("Quoted Printable Encoding needs to be done!!!!");

    result = networkWrite(socketfd, body, size == -1 ? strlen(body) : size);
    if (result == -1) {
        perror("networkWrite: ");
        return E_NET_ERR;
    }

    result = networkWrite(socketfd, "\r\n\r\n\r\n", 6);
    if (result == -1) {
        perror("networkWrite: ");
        return E_NET_ERR;
    }

    return S_OK;
}





/**
 *=------------------------------------------------------------------------=
 * qpHexaDecode
 *=------------------------------------------------------------------------=
 * given a quoted printable =9D kinda sequence, return the 8bit character
 * for it.
 *
 * Parameters:
 *      const char *                - [in]  look at me at the next char
 *
 * Output:
 *      char
 */
char qpHexaDecode
(
    const char *p
)
{
    char c = 0;

    for (unsigned int x = 0; x < sizeof(quotedPrintableValues) / sizeof(char); x++) {
        if (quotedPrintableValues[x] == *p) {
            c = 16 * x;
            break;
        }
    }

    for (unsigned int x = 0; x < sizeof(quotedPrintableValues) / sizeof(char); x++) {
        if (quotedPrintableValues[x] == *(p + 1)) {
            c += x;
            break;
        }
    }

    return c;
}


