/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include "selectme.h"

#include <qlabel.h>
#include <qlistbox.h>
#include <qpushbutton.h>

#include "attachment.h"
#include <intl.h>
#include "localized.h"

/**
 * This is a simple little dialog to show the user all available attachments
 * and let them select the ones they want to save.
 */


class ListBoxTextEx : public QListBoxText {

  public:
    ListBoxTextEx(QListBox *a, const QString &b, unsigned int c)
        : QListBoxText(a, b, 0), itemData(c) { }
    
    unsigned int getItemData() { return this->itemData; }

  private:
    const unsigned int itemData;
};



/**
 *=------------------------------------------------------------------------=
 * SelectMeDialog::SelectMeDialog
 *=------------------------------------------------------------------------=
 * creates a new one.
 */
SelectMeDialog::SelectMeDialog()
: QDialog(0,0, true, WStyle_NoBorder)
{
    DEBUG_TEST_START;

    listBox = NULL;
    ok = cancel = NULL;
    label = NULL;
}

/**
 *=------------------------------------------------------------------------=
 * SelectMeDialog::~SelectMeDialog
 *=------------------------------------------------------------------------=
 * time to take out the trash.
 */
SelectMeDialog::~SelectMeDialog()
{
    DEBUG_TEST_STOP;

    if (listBox) delete listBox;
    if (ok) delete ok;
    if (cancel) delete cancel;
}

/**
 *=------------------------------------------------------------------------=
 * SelectMeDialog::addItem
 *=------------------------------------------------------------------------=
 * adds an item to the list box.  Very exciting.
 *
 * Parameters:
 *      const char *            - [in]  item.
 *      unsigned int            - [in]  item data.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE SelectMeDialog::addItem
(
    const char *item,
    unsigned int itemData
)
{
    if (!this->listBox) return E_INTERNAL;

    ListBoxTextEx *i = new ListBoxTextEx(listBox, item, itemData);
    return i ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * SelectMeDialog::initializeUI
 *=------------------------------------------------------------------------=
 * sets up our UI
 *
 * Parameters:
 *      const char *                - [in]  title
 *      const char *                - [in]  text.
 *      bool                        - [in]  OPTIONAL: do we allow multi-select
 *
 * Output:
 *      ERRCODE
 */
ERRCODE SelectMeDialog::initializeUI
(
    const char *title,
    const char *text,
    bool multi
)
{
    const char *pstr;

    this->setCaption(title);

    label = new QLabel(this);
    if (!label) return E_OUTOFMEMORY;
    label->setTextFormat(Qt::RichText);
    label->setText(text);
    label->setIndent(5);
    label->show();

    listBox = new QListBox(this);
    if (!listBox) return E_OUTOFMEMORY;
    if (multi) listBox->setSelectionMode(QListBox::Extended);
    listBox->show();

    pstr = intlLoadString(ID_OK_TEXT);
    if (!pstr) return E_OUTOFMEMORY;
    ok = new QPushButton(pstr, this);
    localFree(pstr);
    if (!ok) return E_OUTOFMEMORY;
    ok->show();

    pstr = intlLoadString(ID_CANCEL_TEXT);
    if (!pstr) return E_OUTOFMEMORY;
    cancel = new QPushButton(pstr, this);
    localFree(pstr);
    if (!cancel) return E_OUTOFMEMORY;
    cancel->show();

    QObject::connect(ok, SIGNAL(clicked()), this, SLOT(accept()));
    QObject::connect(cancel, SIGNAL(clicked()), this, SLOT(reject()));

    this->resize(225, 200);

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * SelectMeDialog::getSelectedItems
 *=------------------------------------------------------------------------=
 * returns a list of the items that were selected in the dialog.
 *
 * Parameters:
 *      LinkList<const char *> &        - [in]  the list to dump results into.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE SelectMeDialog::getSelectedItems
(
    LinkList<unsigned int> & results
)
{
    int c = listBox->count();

    /**
     * whiz through them finding the selected ones.
     */
    for (int x = 0; x < c; x++) {
        if (listBox->isSelected(x)) {
            ListBoxTextEx *item = dynamic_cast<ListBoxTextEx *>(listBox->item(x));
            if (!item) {
                K_FAIL("I have an item in the list that i didn't create!!");
                return E_INTERNAL;
            }
            results.addElement(item->getItemData());
        }
    }

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * SelectMeDialog::resizeEvent
 *=------------------------------------------------------------------------=
 * resize ourselves!!
 *
 * Parameters:
 *      QResizeEvent *          - [in]
 */
void SelectMeDialog::resizeEvent
(
    QResizeEvent *event
)
{
    int width = event->size().width();
    int height = event->size().height();
    int top = 0;

    label->setGeometry(0, top, width, label->sizeHint().height());
    top += label->sizeHint().height();
    listBox->setGeometry(0, top + 4, width, height - top - 40);

    ok->setGeometry(width - 160, height - 30, 75, 30);
    cancel->setGeometry(width - 80, height - 30, 75, 30);
}

/**
 *=------------------------------------------------------------------------=
 * SelectMeDialog::selectAll
 *=------------------------------------------------------------------------=
 * selects everything in the list.
 */
void SelectMeDialog::selectAll()
{
    if (listBox) listBox->selectAll(true);
}


