/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <ctype.h>

#include "deencode.h"
#include "mailmessage.h"
#include "mime.h"
#include "netapis.h"


/**
 *=------------------------------------------------------------------------=
 * sevenBitDecode
 *=------------------------------------------------------------------------=
 * given a stream of characters that is basically plain text, extract all the
 * text and return it in one big chunk.
 *
 * Parameters:
 *      const char *                - [in]  terminating separator.
 *      int *                       - [out] the size of the attachment.
 *      const char **               - [out] the decoded section.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE sevenBitDecode
(
    const char *separator,
    int *size,
    const char **body
)
{
    char inputBuf[NET_MAXLINELEN];
    int currentSize = 0;
    BYTE *buf = NULL;
    int bufSize = 0;
    int len, cread;
    int seplen = 0;
    ERRCODE er;

    if (!body || !separator) return E_PARAM;
    seplen = strlen(separator);

    /**
     * this is easy -- while we're not at the separating line, add in the
     * text that we're seeing.  E-Zed.
     */
    while (1) {
        mimeReader.readLine(inputBuf, sizeof(inputBuf), cread);

        if (!strncasecmp(inputBuf, separator, seplen)) {
            break;
        } else if (inputBuf[0] == '\0') {
            break;
        }

        len = strlen(inputBuf);
        if (!ensureBufferSize((char **)&buf, &bufSize, currentSize + len, BUFFER_INCREMENT)) {
            er = E_OUTOFMEMORY;
            goto Error;
        }
        strncpy((char *)buf + currentSize, inputBuf, bufSize - currentSize - 1);
        currentSize += len;
    }

 
    *body = (const char *)buf;
    *size = currentSize;
    er = S_OK;

  Error:
    if (FAILED(er)) { if (buf) localFree(buf); }
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * sevenBitEncode
 *=------------------------------------------------------------------------=
 * given some text, write it out as seven bit text.  We have to make sure 
 * that we wrap on appropriate boundaries.
 *
 * Parameters:
 *      int                         - [in]  network stream to write to.
 *      int                         - [in]  don't let lines be longer than 
 *                                      this (7bit/q-p only)
 *      const char *                - [in]  data to write out
 *      int                         - [in]  size of said data.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE sevenBitEncode
(
    int socketfd,
    int lineLength,
    const char *body,
    int size
)
{
    int x = 0 , lastSpaceX = 0, lastSpaceY = 0, y = 0;
    char buf[NET_MAXLINELEN];
    int result;

    /**
     * Now start copying things into the buffer one word at a time, making
     * sure we wrap on appropriate boundaries.  If the line length is greater
     * than the buffer size, we'll wrap anyhoo, as it's the MAX SMTP length.
     */
    while (body[x]) {

        if (body[x] == ' ' || body[x] == '\t') {
            lastSpaceX = x;
            lastSpaceY = y;
        } else if (body[x] == '\r' || body[x] == '\n') {

            buf[y++] = '\r';
            buf[y++] = '\n';
            buf[y] = '\0';

            result = networkWrite(socketfd, buf, y);
            if (result == -1) {
                perror("networkWrite: ");
                return E_NET_ERR;
            }

            x++;
            if (body[x - 1] == '\r' && body[x] == '\n') x++;
            y = 0;
            lastSpaceX = 0;
            lastSpaceY = 0;
            continue;
        } 


        buf[y++] = body[x++];
        if (y >= lineLength) {
            if (lastSpaceY == 0) {
                lastSpaceY = lineLength;
                lastSpaceX = x - 1;
            }
            buf[lastSpaceY] = '\r';
            buf[lastSpaceY + 1] = '\n';
            buf[lastSpaceY + 2] = '\0';
            x = lastSpaceX + 1;

            result = networkWrite(socketfd, buf, strlen(buf));
            if (result == -1) {
                perror("networkWrite: ");
                return E_NET_ERR;
            }
            y = 0;
            lastSpaceY = 0;
        }
                
    }


    result = networkWrite(socketfd, "\r\n", 2);
    if (result == -1) {
        perror("networkWrite: ");
        return E_NET_ERR;
    }

    return S_OK;
}


