/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include <sanity.h>

#include <errno.h>
#include <qthread.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "dbmgr.h"
#include "instance.h"
#include <lerror.h>
#include "sighandler.h"


/**
 *=------------------------------------------------------------------------=
 *                              Prototypes
 *=------------------------------------------------------------------------=
 */
void signalHandler(int);


QMutex signalMutex;

/**
 *=------------------------------------------------------------------------=
 * This is the list of fatal signals that will kill us.  We will replace it
 * with our own handler that will go and call abort().
 */
struct FatalSignal {

    int signo;
    void (*sigFunc)(int);

} fatalSignals [] = {

    { SIGBUS, NULL },
#ifdef SIGEMT
    { SIGEMT, NULL },
#endif 
    { SIGFPE, NULL },
    { SIGHUP, NULL },
    { SIGILL, NULL },
    { SIGINT, NULL },
    { SIGIO, NULL },
    { SIGQUIT, NULL },
    { SIGSEGV, NULL },
    { SIGSYS, NULL },
    { SIGTERM, NULL },
    { SIGTRAP, NULL },
    { SIGXCPU, NULL },
    { SIGXFSZ, NULL },
};


/**
 *=------------------------------------------------------------------------=
 * signalInstallHandlers
 *=------------------------------------------------------------------------=
 * installs signal handlers for the signals that we're going to worry about
 * when running this program.  We want to be sure that, in cases where we
 * crash or are aborted, we ssync and save out the DBs, and remove any of
 * our instance locking information.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE signalInstallHandlers()
{
    struct sigaction act, oact;
    int res;

    act.sa_handler = signalHandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;

    /**
     * Start replacing the signals we want to replace!!!
     */
    for (int x = 0; x < (int)(sizeof(fatalSignals) / sizeof(FatalSignal)); x++) {

        res = sigaction(fatalSignals[x].signo, &act, &oact);
        if (res) goto Error;
        fatalSignals[x].sigFunc = oact.sa_handler;
    }

    return S_OK;

  Error:
    errorSaveRichErrorInfo(strerror(errno));
    return E_SYSTEM_ERR;
}


/**
 *=------------------------------------------------------------------------=
 * signalHandler
 *=------------------------------------------------------------------------=
 * this is our signal handling function.
 *
 * Parameters:
 *      int                             - [in]  signal received, i believe.
 */
void signalHandler
(
    int a
) 
{
    
    signalMutex.lock();

    fprintf(stderr, "Received fatal signal: %d.  Terminating process.\n", a);

    /**
     * Make sure that the instance fifo is gone, otherwise other processes
     * will not be able to start up properly !!!
     */
    instanceNukeFifo();

    /**
     * Sync up the databases so there are no problems !!!!
     */
    dbmgrSyncDatabasesOnDeath();

    signalMutex.unlock();

#if 0
    /**
     * Now call the old handler for this signal, if it exists.
     */
    for (int x = 0; x < (int)(sizeof(fatalSignals) / sizeof(FatalSignal)); x++) {

        if (fatalSignals[x].signo == a) {
            fatalSignals[x].sigFunc(a);
            break; // in case that function actually returns.
        }
    }
#endif // 0

    /**
     * in case we fell through to this place, just go and die now.
     */
    abort();
    
}


