/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */

#include <sanity.h>

#include <stdlib.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "fileutil.h"
#include <intl.h>
#include "kiltdown.h"
#include "localized.h"
#include "main.h"
#include "properties.h"
#include "submail.h"


/** 
 *=------------------------------------------------------------------------=
 * This is the implementation of the main object for the subsystem that
 * supports basic internet mail functionality. 
 *=------------------------------------------------------------------------=
 */
const char *mailConfigFile = "mail-config";

/**
 *=------------------------------------------------------------------------=
 * Global Mail Properties
 *=------------------------------------------------------------------------=
 */
const char *accountCountProp = "AccountCount";

/**
 *=------------------------------------------------------------------------=
 * Global Mail Preferences.
 *=------------------------------------------------------------------------=
 */
const char *readFontFamilyProp = "readFontFamily";
const char *readFontSizeProp = "readFontSize";
const char *composeFontFamilyProp = "composeFontFamily";
const char *composeFontSizeProp = "composeFontSize";
const char *closeOnReplyProp = "closeOnReply";
const char *forwardPrefixProp = "forwardPrefix";
const char *indentForwardProp = "indentForward";
const char *replyPrefixProp = "replyPrefix";
const char *indentReplyProp = "indentReply";
const char *selectPrevDeleteProp = "selectPreviousOnDelete";
const char *lineWrapCountProp = "lineWrapCount";
const char *dontConfirmDIDeleteProp = "dontConfirmDeletedItemsDelete";
const char *ignoreAskOnDeleteProp = "ignoreAskOnDelete";

/**
 *=------------------------------------------------------------------------=
 * Per account properties.  Persisted in the form:
 * XX.<<propname>>
 * where XX is the account number.
 *=------------------------------------------------------------------------=
 */
const char *accountNameProp = "accountName";
const char *accountDirectoryProp = "accountDirectory";


/**
 *=------------------------------------------------------------------------=
 * We'll use this to hold the names and directories of the accounts we're
 * going to load.  NOTE:  this is kinda lazy -- i could just use an
 * array of char *s, but this seems a little cleaner.
 */
struct NameDirPair {
    const char *name;
    const char *dir;
};

/**
 *=------------------------------------------------------------------------=
 * Default font family and size
 *=------------------------------------------------------------------------=
 */
static const char *defaultFontFamily = "Helvetica";
static const int   defaultFontSize = 12;

static const int DEFAULT_LINE_WRAP = 72;


/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::createNew
 *=------------------------------------------------------------------------=
 * Creates a new instance of the mail subsystem, and returns it as an 
 * ISubsystem interface.  the destroy() method is called when the program
 * no longer requires this subsystem.
 *
 * Parameters:
 *      MailSubsystem **       - [out]  the new object as a subsystem.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::createNew
(
    ISubsystem **pnew
)
{
    if (!pnew) return E_PARAM;

    MailSubsystem *mailSubsystem = new MailSubsystem();
    if (!mailSubsystem) return E_OUTOFMEMORY;

    *pnew = (ISubsystem *)mailSubsystem;
    return S_OK;
}
 
/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::MailSubsystem
 *=------------------------------------------------------------------------=
 * Constructor that doesn't do very much
 */
MailSubsystem::MailSubsystem()
{
    site = NULL;

    preferencesChanged = false;

    composeFontFamily = copyString(defaultFontFamily);
    readFontFamily = copyString(defaultFontFamily);
    composeFontSize = 12;
    readFontSize = 12;
    closeOnReply = true;
    ignoreAskOnDelete = false;

    indentReply = false;
    replyPrefix = NULL;
    indentForward = false;
    forwardPrefix = NULL;
    dontConfirmDIDelete = false;
    selectPreviousOnDelete = false;
    lineWrapCount = DEFAULT_LINE_WRAP;
}
 
/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::~MailSubsystem
 *=------------------------------------------------------------------------=
 * Our destructor.  Goes and kills a few other objects, making sure they 
 * clean up after themselves.
 */
MailSubsystem::~MailSubsystem()
{
    MailAccount *acct;
    ERRCODE er;
    
    er = accounts.firstElement(&acct);
    while (er == S_OK) {
        delete acct;
        er = accounts.nextElement(&acct);
    }

    if (composeFontFamily) localFree(composeFontFamily);
    if (readFontFamily) localFree(readFontFamily);
}
 
/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::accountNameChanged
 *=------------------------------------------------------------------------=
 * the given account now has the given name.
 *
 * Parameters:
 *      MailAccount *               - [in]
 *      const char *                - [in]
 */
void MailSubsystem::accountNameChanged
(
    MailAccount *acct,
    const char *newName
)
{
    char buf[MAX_NAMEVALUE_LEN];
    PropertiesFile *mailConfig;
    MailAccount *a;
    ERRCODE er;
    int x = 0;
    
    /**
     * First, find the account.
     */
    er = accounts.firstElement(&a);
    while (er == S_OK) {

        if (a == acct) break;

        er = accounts.nextElement(&a);
        x++;
    }

    if (er != S_OK) {
        K_FAIL("Asked to change the name of an account I don't know about !!\n");
        return;
    }

    er = getSubsystemConfig(&mailConfig);
    if (FAILED(er)) {
        K_FAIL("Couldn't get subsystem configuration!\n");
        return;
    }

    /**
     * Now update our config file with the new account name.
     */
    sprintf(buf, "%d.%s", x, accountNameProp);
    mailConfig->writeStringProperty(buf, newName);
    mailConfig->close();
    delete mailConfig;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::createNewAccount                             [ISubsystem]
 *=------------------------------------------------------------------------=
 * creates a new mail account and adds it to the UI ...
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::createNewAccount()
{
    PropertiesFile *props;
    IAccount *acct;
    ERRCODE er;

    /**
     * get the config file for this subsystem.
     */
    er = getSubsystemConfig(&props);
    RETURN_ON_FAILURE(er);

    /**
     * Show the wizard to create a new account!
     */
    er = createNewAccount(props, &acct);
    if (FAILED(er) || er == S_FALSE) goto Error;

    /**
     * Finally, tell the host that there was a new account created!!!
     */
    er = site->addAccount(acct);
    GOTO_ON_FAILURE(er, Error);

    props->close();

    er = S_OK;

  Error:
    delete props;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::getAccount                                   [ISubsystem]
 *=------------------------------------------------------------------------=
 * Gets the account with the given number (zero-based index.)
 * 
 * Parameters:
 *      int             - [in]  the account at the given index.
 *      IAccount **     - [out] the specified account.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::getAccount
(
    int index,
    IAccount **paccount
)
{
    MailAccount *acct;
    ERRCODE er;

    /**
     * Check params
     */
    if (!paccount) return E_PARAM;

    er = accounts.elementAt(index, &acct);
    *paccount = (acct) ? (IAccount *)acct : NULL;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::getAccountCount                              [ISubsystem]
 *=------------------------------------------------------------------------=
 * returns the number of active accounts within this subsystem.
 * 
 * Parameters:
 *      int *          - [out]  the number of accounts currently active.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::getAccountCount
(
    int *pcount
)
{
    if (!pcount) return E_PARAM;

    *pcount = accounts.countElements();
    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::getComposeFont
 *=------------------------------------------------------------------------=
 * returns the font used to compose plain text mail, including it's font 
 * size
 *
 * Parameters:
 *      const char * &              - [out] the family name.
 *      int &                       - [out] the point size.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::getComposeFont
(
    const char *&family,
    int &point
)
{
    family = copyString(composeFontFamily);
    point = composeFontSize;
    return S_OK;
}
 
/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::getDisplayName                               [ISubsystem]
 *=------------------------------------------------------------------------=
 * returns the display  name of this subsystem.
 * 
 * Parameters:
 *      const char **                   - [out]  CALLER Frees
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::getDisplayName
(
    const char **name
)
{
    if (!name) return E_PARAM;

    *name = intlLoadString(ID_SUB_MAIL);
    return *name ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::getForwardPrefix
 *=------------------------------------------------------------------------=
 * returns the prefix to use for the included text when forwarding mail.
 *
 * Parameters:
 *      const char *&               - [out] prefix each line with this.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::getForwardPrefix
(
    const char *&prefix
)
{
    char buf[32];

    if (indentForward) {
        snprintf(buf, sizeof(buf), "    %s", forwardPrefix ? forwardPrefix : "");
    } else {
        snprintf(buf, sizeof(buf), "%s", forwardPrefix ? forwardPrefix : "");
    }

    prefix = copyString(buf);
    return prefix ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::getReadFont
 *=------------------------------------------------------------------------=
 * returns the font used to read plain text mail, including it's font size.
 *
 * Parameters:
 *      const char * &              - [out] the family name.
 *      int &                       - [out] the point size.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::getReadFont
(
    const char *&family,
    int &point
)
{
    family = copyString(readFontFamily);
    point = readFontSize;
    return S_OK;
}
 
/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::getReplyPrefix
 *=------------------------------------------------------------------------=
 * returns the prefix to use for the included text when replying to mail.
 *
 * Parameters:
 *      const char *&               - [out] prefix each line with this.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::getReplyPrefix
(
    const char *&prefix
)
{
    char buf[32];

    if (indentReply) {
        snprintf(buf, sizeof(buf), "    %s", replyPrefix ? replyPrefix : "> ");
    } else {
        snprintf(buf, sizeof(buf), "%s", replyPrefix ? replyPrefix : "> ");
    }

    prefix = copyString(buf);
    return prefix ? S_OK : E_OUTOFMEMORY;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::getSubsystemConfig
 *=------------------------------------------------------------------------=
 * returns the config file for this subsystem.
 *
 * Parameters:
 *      PropertiesFile **           - [out] our config file.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::getSubsystemConfig
(
    PropertiesFile **cfg
)
{
    char buf[MAXPATHLEN];
    ERRCODE er;

    /**
     * 1. Get the kiltdown dir.
     */
    er = Kiltdown::getKiltdownDirectory(buf, sizeof(buf));
    RETURN_ON_FAILURE(er);

    /**
     * 2. open the mail configuration file.
     */
    strcat(buf, mailConfigFile);
    er = PropertiesFile::openPropertiesFile(buf, cfg, true);
    RETURN_ON_FAILURE(er);
    K_ASSERT(*cfg, "Aiieieee!  No es bueno!!!  propfile ptr is NULL");

    return S_OK;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::initialize
 *=------------------------------------------------------------------------=
 * Reads in the configuration file and figures out how many accounts we have
 * active on this system, etc ...
 *
 * Parameters:
 *      ISubsystemHost *                - [in]  our site pointer.
 * 
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::initialize
(
    ISubsystemHost *site
)
{
    PropertiesFile *mailConfig;
    NameDirPair *infoArray = NULL;
    MailAccount *account;
    int accountCount = 0;
    const char *pstr;
    ERRCODE er;
    bool b;
    int i;

    this->site = site;

    /**
     * Get the config file for this account.
     */
    er = getSubsystemConfig(&mailConfig);
    RETURN_ON_FAILURE(er);

    /**
     * Load in misc properties!!!
     */
    mailConfig->getStringProperty(readFontFamilyProp, &pstr);
    if (pstr != NULL) {
        if (readFontFamily) localFree(readFontFamily);
        readFontFamily = pstr;
    }
    mailConfig->getIntProperty(readFontSizeProp, &i);
    if (i != 0) readFontSize = i;

    mailConfig->getStringProperty(composeFontFamilyProp, &pstr);
    if (pstr != NULL) {
        if (composeFontFamily) localFree(composeFontFamily);
        composeFontFamily = pstr;
    }
    mailConfig->getIntProperty(composeFontSizeProp, &i);
    if (i != 0) composeFontSize = i;

    mailConfig->getBooleanProperty(closeOnReplyProp, &b);
    this->closeOnReply = b;

    mailConfig->getStringProperty(forwardPrefixProp, &this->forwardPrefix);
    mailConfig->getStringProperty(replyPrefixProp, &this->replyPrefix);
    mailConfig->getBooleanProperty(indentForwardProp, &this->indentForward);
    mailConfig->getBooleanProperty(indentReplyProp, &this->indentReply);
    mailConfig->getBooleanProperty(selectPrevDeleteProp, &this->selectPreviousOnDelete);
    mailConfig->getIntProperty(lineWrapCountProp, &i);
    mailConfig->getBooleanProperty(dontConfirmDIDeleteProp, &this->dontConfirmDIDelete);
    mailConfig->getBooleanProperty(ignoreAskOnDeleteProp, &this->ignoreAskOnDelete);

    if (i != 0) this->lineWrapCount = i;

    /**
     * Now, find out how many accounts we have in our system.
     */
    er = mailConfig->getIntProperty(accountCountProp, &accountCount);
    GOTO_ON_FAILURE(er, Error);

    if (accountCount == 0) {
        /**
         * Create a new account (if the user wants to), and add it to the list
         * of accounts.
         */
        er = createNewAccount(mailConfig);
        GOTO_ON_FAILURE(er, Error);

    } else {

        /**
         * Load in the account information, put it in a list of
         * NameDirPair structs, and then load in the given accounts.
         */
        infoArray = new NameDirPair[accountCount];
        if (!infoArray) {
            er = E_OUTOFMEMORY;
            goto Error;
        }
        memset(infoArray, 0, accountCount * sizeof(NameDirPair));

        er = loadAccountInfo(mailConfig, infoArray, accountCount);
        GOTO_ON_FAILURE(er, Error);

        /**
         * Now go and load in the accounts.
         */
        for (int x = 0; x < accountCount; x++) {
            er = MailAccount::openExistingAccount(this,
                                                  infoArray[x].name,
                                                  infoArray[x].dir,
                                                  &account);
            if (FAILED(er)) {
                delete infoArray;
                goto Error;
            }

            /**
             * Okay, add the account to the list of valid accounts.
             */
            er = accounts.addElement(account);
            GOTO_ON_FAILURE(er, Error);
        }
    }

    er = S_OK;

  Error:
    if (infoArray != NULL) {
        for (int x = 0; x < accountCount; x++) {
            if (infoArray[x].name) localFree(infoArray[x].name);
            if (infoArray[x].dir) localFree(infoArray[x].dir);
        }
        delete infoArray;
    }

    mailConfig->close();
    delete mailConfig;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::loadAccountInfo
 *=------------------------------------------------------------------------=
 * Loads in all the information about the accounts in the given properties
 * file.  these are returned in an array of NameDirPair structs which
 * can then be used to create MailAccounts.
 *
 * Parameters:
 *      PropertiesFile *        - [in]  properties file from which to load.
 *      NameDirPair *           - [in]  array in which to place values.
 *      int                     - [in]  count of accounts.
 *
 * Output:
 *      ERRCODE
 */
ERRCODE MailSubsystem::loadAccountInfo
(
    PropertiesFile *configFile,
    NameDirPair *infoArray,
    int accountCount
)
{
    char name[MAX_NAMEVALUE_LEN];
    const char *svalue;
    ERRCODE er;
    int x;

    /**
     * Whiz through the accounts & load in their values.
     */
    for (x = 0; x < accountCount; x++) {

        /**
         * Start loading in the properties.
         */
        sprintf(name, "%d.%s", x, accountNameProp);
        er = configFile->getStringProperty(name, &svalue);
        if (FAILED(er) || er == S_NODATA) {
            er = E_BAD_CONFIG;
            goto Error;
        }
        if (svalue && svalue[0] == '\0') {
            localFree(svalue); svalue = NULL;
        }
        infoArray[x].name = svalue;

        sprintf(name, "%d.%s", x, accountDirectoryProp);
        er = configFile->getStringProperty(name, &svalue);
        if (FAILED(er) || er == S_NODATA) {
            er = E_BAD_CONFIG;
            goto Error;
        }
        GOTO_ON_FAILURE(er, Error);
        if (svalue && svalue[0] == '\0') {
            localFree(svalue); svalue = NULL;
        }
        infoArray[x].dir = svalue;
    }

    er = S_OK;

  Error:
    if (FAILED(er)) {
        for (int y = 0; y < accountCount; y++) {
            if (infoArray[y].name) localFree(infoArray[y].name);
            if (infoArray[y].dir) localFree(infoArray[y].dir);
            if (y == x) break;
        }
    }

    return er;
}


/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::createNewAccount
 *=------------------------------------------------------------------------=
 * shows the wizard to create a new account.  returns the information about
 * said new account.
 *
 * Parameters:
 *      PropertiesFile *            - [in]  where to write out new account
 *                                      informaton.
 *      IAccount **                 - [out] Optional:  the account that was
 *                                      created.
 * Output:
 *      ERRCODE                     - S_OK if a new account was created,
 *                                      S_FALSE if not.
 */
ERRCODE MailSubsystem::createNewAccount
(
    PropertiesFile *mailConfig, 
    IAccount **paccount
)
{
    const char *accountName = NULL;
    const char *accountDir = NULL;
    MailAccount *account = NULL;
    char buf[MAX_NAMEVALUE_LEN];
    int result;
    ERRCODE er;

    /**
     * Run the wizard and create the account.
     */
    er = MailAccount::createNewAccount(this, &accountName, &accountDir, &account);
    RETURN_ON_FAILURE(er);
    if (er == S_FALSE)
        return er;
    K_ASSERT(account, "D'oH!  Was given a NULL account back on success!");

    /**
     * Add it to our list of accounts.
     */
    er = accounts.addElement(account);
    GOTO_ON_FAILURE(er, Error);
    if (paccount) *paccount = account;

    /**
     * Set up a few things in the config file.
     */
    mailConfig->getIntProperty(accountCountProp, &result);
    sprintf(buf, "%d.%s", result, accountNameProp);
    mailConfig->writeStringProperty(buf, accountName);
    sprintf(buf, "%d.%s", result, accountDirectoryProp);
    mailConfig->writeStringProperty(buf, accountDir);
    result++;
    mailConfig->writeIntProperty(accountCountProp, result);

    er = S_OK;

  Error:
    if (accountName) localFree(accountName);
    if (accountDir) localFree(accountDir);

    if (FAILED(er) && account) delete account;
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::removeAccount
 *=------------------------------------------------------------------------=
 * removes the given account from our list of accounts, udpates the 
 * configuration files, and, if appropriate, deletes all files associated
 * with the account.
 *
 * Parameters:
 *      IAccount *              - [in]
 *      bool                    - [in]  should underlying files be destroyed.
 */
void MailSubsystem::removeAccount
(
    IAccount *removeMe,
    bool destroy
)
{
    MailAccount *acct;
    ERRCODE er;

    acct = dynamic_cast<MailAccount *>(removeMe);
    if (!acct) {
        K_FAIL("This isn't a MailAccount!!! D'oh!\n");
        return;
    }

    /**
     * Remove it from our list of goo.
     */
    er = accounts.removeElement(acct);
    if (er != S_OK) {
        K_FAIL("For some reason, I don't have this acct in my list!!\n");
        return;
    }

    /**
     * Tell the account to clean up its UI.
     */
    if (site) site->removeAccount(acct);
    acct->dieDieDIE();

    /**
     * Now go and save out the new list of accounts.
     */
    er = writeAccountList();
    if (FAILED(er)) {
        errorShowError(er);
        return;
    }

    /**
     * Now, if we're supposed, go and NUKE THAT MUTHAH!!!
     */
    if (destroy) fileRmfr(acct->getAccountDirRef());
    delete acct;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::terminate
 *=------------------------------------------------------------------------=
 * Destroys an instance of the MailSubsystem object that somebody has 
 * created with the createNew() method.
 */
void MailSubsystem::terminate()
{
    PropertiesFile *mailConfig;
    MailAccount *acct;
    ERRCODE er;

    /**
     * Write out our configration.  We write out any properties that have
     * changed!!
     */
    if (this->preferencesChanged) {
        er = getSubsystemConfig(&mailConfig);
        if (SUCCEEDED(er)) {
            
            mailConfig->writeStringProperty(readFontFamilyProp, this->readFontFamily);
            mailConfig->writeStringProperty(composeFontFamilyProp, this->composeFontFamily);
            mailConfig->writeIntProperty(readFontSizeProp, this->readFontSize);
            mailConfig->writeIntProperty(composeFontSizeProp, this->composeFontSize);
            mailConfig->writeBooleanProperty(closeOnReplyProp, this->closeOnReply);
            mailConfig->writeStringProperty(forwardPrefixProp, this->forwardPrefix);
            mailConfig->writeStringProperty(replyPrefixProp, this->replyPrefix);
            mailConfig->writeBooleanProperty(indentForwardProp, this->indentForward);
            mailConfig->writeBooleanProperty(indentReplyProp, this->indentReply);
            mailConfig->writeBooleanProperty(selectPrevDeleteProp, this->selectPreviousOnDelete);
            mailConfig->writeIntProperty(lineWrapCountProp, this->lineWrapCount);
            mailConfig->writeBooleanProperty(dontConfirmDIDeleteProp, this->dontConfirmDIDelete);
            mailConfig->writeBooleanProperty(ignoreAskOnDeleteProp, this->ignoreAskOnDelete);
        }
    }

    /**
     * Whiz through the accounts and make sure everything is properly saved
     * out.
     */    
    er = accounts.firstElement(&acct);
    while (er == S_OK) {

        acct->saveConfiguration();
        er = accounts.nextElement(&acct);
    }

    delete this;
}

/**
 *=------------------------------------------------------------------------=
 * MailSubsystem::writeAccountList
 *=------------------------------------------------------------------------=
 * writes out the list of accounts to our configuration file.
 *
 * Output:
 *      ERRCODE
 *
 * Notes:
 *      Please note that in addition to writing out accounts.countElements()
 *  entries, we also write out one more, to make sure any extras are cleared
 *  out properly.  It should result in them being removed from the mail-config
 *  file.
 */
ERRCODE MailSubsystem::writeAccountList()
{
    char buf[MAX_NAMEVALUE_LEN];
    PropertiesFile *mailConfig;
    MailAccount *acct;
    ERRCODE er;
    int x;

    /**
     * Get our config file.
     */
    er = getSubsystemConfig(&mailConfig);
    RETURN_ON_FAILURE(er);

    /**
     * Write out the new count of accounts.
     */
    x = accounts.countElements();
    mailConfig->writeIntProperty(accountCountProp, x);

    /**
     * For each account, we've got two properties.
     */
    x = 0;
    er = accounts.firstElement(&acct);
    while (er == S_OK) {
        sprintf(buf, "%d.%s", x, accountNameProp);
        mailConfig->writeStringProperty(buf, acct->getAccountNameRef());
        sprintf(buf, "%d.%s", x, accountDirectoryProp);
        mailConfig->writeStringProperty(buf, acct->getAccountDirRef());
        x++;

        er = accounts.nextElement(&acct);
    }

    RETURN_ON_FAILURE(er);

    /**
     * Now write out a bit of extra goo to nuke any extras.
     */
    sprintf(buf, "%d.%s", x, accountNameProp);
    mailConfig->writeStringProperty(buf, NULL);
    sprintf(buf, "%d.%s", x, accountDirectoryProp);
    mailConfig->writeStringProperty(buf, NULL);

    mailConfig->close();
    delete mailConfig;

    return S_OK;
}


