/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _SUBMAIL_H_
#define _SUBMAIL_H_

#include <llist.h>

#include <isubsystem.h>
#include "mailaccount.h"

/**
 * Forward decls
 */
class PropertiesFile;
struct NameDirPair;

/** 
 * This is the main object for the subsystem that supports basic internet mail
 * functionality.   It loads, manages, and saves the various mail accounts 
 * that the user might be working with.
 */
class MailSubsystem : public ISubsystem {

  public:
    /**
     * Creates a new subsystem
     */
    static ERRCODE createNew(ISubsystem **);

    /**
     * ISubsystem methods!!!
     */
    virtual ERRCODE createNewAccount();
    virtual ERRCODE getAccount(int, IAccount **);
    virtual ERRCODE getAccountCount(int *);
    virtual ERRCODE initialize(ISubsystemHost *);
    virtual void    removeAccount(IAccount *, bool);
    virtual void    terminate();

    /**
     * These are preferences global to all mail accounts.
     */
    ERRCODE getComposeFont(const char *&family, int &size);
    ERRCODE getReadFont(const char * &family, int &size);
    inline  void getCloseOnReply(bool &b) { b = this->closeOnReply; }
    ERRCODE getForwardPrefix(const char *&);
    ERRCODE getReplyPrefix(const char *&);
    inline  void getSelectPreviousOnDelete(bool &f)  { f = this->selectPreviousOnDelete; }
    inline  void getLineWrapCount(int &l) { l = lineWrapCount; }
    inline  void getDontConfirmDIDelete(bool &f) { f = this->dontConfirmDIDelete; }
    inline  void getIgnoreAskOnDelete(bool &f) { f = this->ignoreAskOnDelete; }

    /**
     * The given account now has the given name.
     */
    void accountNameChanged(MailAccount *, const char *);

    /**
     * Clean up!
     */
    virtual ~MailSubsystem();

  private:
    /**
     * Use createNew() instead!!
     */
    MailSubsystem();

    /**
     * Loads in the information about the various accounts the user has
     * installed, and gets everything going.
     */
    ERRCODE initialize();

    /**
     * Brings up a wizard to create a new mail account.  Very complex. Returns
     * S_FALSE if the user elects not to create an account.
     */
    ERRCODE createNewAccount(PropertiesFile *, IAccount **newA = NULL);

    /**
     *  returns the name of the subsystem.
     */
    ERRCODE getDisplayName(const char **);

    /**
     * Returns the account config file for this account.
     */
    ERRCODE getSubsystemConfig(PropertiesFile **);

    /** 
     * Load in the properties for the given accounts from the given
     * properties file.
     */
    ERRCODE loadAccountInfo(PropertiesFile *,  NameDirPair *, int);

    /**
     * Write out a new list of all accounts.
     */
    ERRCODE writeAccountList();

    /**
     * This is the list of mail accounts that we are managing.
     */
    LinkList<MailAccount *> accounts;

    /**
     * Our site pointer.
     */
    ISubsystemHost *site;

    /**
     * Preferences that are used about the UI.
     */
    bool  preferencesChanged;        // do we have to persist them??
    const char *readFontFamily;
    int   readFontSize;
    const char *composeFontFamily;
    int   composeFontSize;
    bool  closeOnReply;
    bool  ignoreAskOnDelete;

    bool indentReply;
    const char *replyPrefix;
    bool indentForward;
    const char *forwardPrefix;
    bool selectPreviousOnDelete;
    int  lineWrapCount;
    bool dontConfirmDIDelete;
};

#endif // _SUBMAIL_H_



