/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#include "superlabel.h"

#include <qfontmetrics.h>
#include <qpainter.h>
#include <qscrollbar.h>
#include <qpixmap.h>
#include <qpopupmenu.h>

#include "address.h"
#include "addressbook.h"
#include "intl.h"
#include "kiltdown.h"
#include "localized.h"
#include "main.h"


static const int PAD = 2;
static const int SCROLLBAR_WIDTH = 12;

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::SuperLabel
 *=------------------------------------------------------------------------=
 * initializes our data structures.  Very complex.
 *
 * Parameters:
 *      QWidget *               - [in]  parent as per QWidget
 *      const char *            - [in]  name as per QWidget
 */
SuperLabel::SuperLabel
(
    QWidget *parent,
    const char *name
)
: QWidget(parent, name)
{
    this->_doubleBuffer = new QPixmap(2, 2);
    this->mustUpdateBuffer = false;

    this->scrollBar = new QScrollBar(this);

    QObject::connect(this->scrollBar, SIGNAL(valueChanged(int)),
                     this, SLOT(scrollBarChanged(int)));

    this->crowTop = 0;
    this->rowHeight = 0;
    this->totalRows = 0;
    this->visibleRows = 0;
    this->separatorWidth = 0;

    this->separatorText = "; ";

    this->popupMenu = NULL;
    this->showingPopup = false;

    setFocusPolicy(ClickFocus);

}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::~SuperLabel
 *=------------------------------------------------------------------------=
 * clean up.
 */
SuperLabel::~SuperLabel()
{
    if (_doubleBuffer) delete _doubleBuffer;
    if (scrollBar) delete scrollBar;
    if (popupMenu) delete popupMenu;

    /**
     * Clean up the list.
     */
    emptyList();
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::addAddress
 *=------------------------------------------------------------------------=
 * adds an address to the list.
 *
 * Parameters:
 *      const char *                - [in]  add me please.
 */
ERRCODE SuperLabel::addAddress
(
    const char *addr
)
{
    LabelItem *item;
    ERRCODE er;

    if (!addr) return E_PARAM;

    /**
     * Create the new item and initialize it.
     */
    item = (LabelItem *)localAlloc(sizeof(LabelItem));
    if (!item) return E_OUTOFMEMORY;
    memset(item, 0, sizeof(LabelItem));
    item->objectRect.setCoords(-1, -1, -1, -1);

    /**
     * Get the email address and user name for this address passed in.  Please
     * note that if there is no friendly name, the email address is just used
     * instead.  all extra SMTP characters, namely <> and "" are stripped.
     */
    er = addressExtractEmailAddress(addr, item->address);
    RETURN_ON_FAILURE(er);
    er = addressExtractUserNameCopy(addr, item->name);
    if (FAILED(er)) {
        localFree(item->address);
        return er;
    }

    /**
     * Add it to our list.
     */
    er = addressList.addElement(item);
    RETURN_ON_FAILURE(er);

    /**
     * This is theoretically quite inefficient, but we want to make sure that
     * adding a new item causes proper updates .... not sure what the best 
     * thing to do here is ...  fortunately, most calls to addAddress are
     * wrapped in a setUpdatesEnabled block ....
     */
    computeVisibleRowCount(this->getClientWidth());
    this->forceUpdate();
    return er;
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::addSelectedToAddressBook
 *=------------------------------------------------------------------------=
 * adds the selected items to the address book.
 */
void SuperLabel::addSelectedToAddressBook()
{
    LabelItem *item;
    const char *pstr;
    ERRCODE er ;

    /**
     * whiz through the list of selected items and add them one by one.
     */
    item = firstSelectedItem();
    while (item != NULL) {
        
        AddressBook *b = localMainWidget->getAddressBook();

        addressGenerateSmtpCopy(item->name, item->address, pstr);

        if (pstr) {
            er = b->addNewEntryFromAddress(pstr);
            localFree(pstr);
            if (FAILED(er)) errorShowError(er);
        }

        item = nextSelectedItem(item);
    }
}


/**
 *=------------------------------------------------------------------------=
 * SuperLabel::computeVisibleRowCount
 *=------------------------------------------------------------------------=
 * figure out how many rows to display.  This is actually kind of a pain in
 * the butt, because we need to go through all our items and figure out 
 * wrapping info, etc ... blech.
 *
 * The net result is that this->totalRows is correctly updated, and a 
 * scrollbar is shown if necessary.
 *
 * Parameters:
 *      int                         - [in]  width to compute for.
 */
void SuperLabel::computeVisibleRowCount
(
    int myWidth
)
{
    QFontMetrics metrics(this->font());
    LabelItem *item;
    int lineWidth;
    ERRCODE er;
    QRect rect;
    int crows;

    lineWidth = 0;

    /**
     * Compute the row height and width of a separator
     */
    this->rowHeight = metrics.ascent() + metrics.descent() + 2 * PAD;
    this->separatorWidth = metrics.boundingRect(QString(this->separatorText)).width() + PAD;

    /**
     * Next, compute the number of rows by whizzing through the list until
     * there are no more items.  If an item causes us to go over a line
     * boundary, increment totalRows.
     */
    crows = 1;
    er = this->addressList.firstElement(&item);
    while (er == S_OK) {

        /**
         * Get the bounding rect.
         */
        rect = metrics.boundingRect(QString(item->name), -1);
        lineWidth += rect.width() + 2 * PAD;

        /**
         * Now try and get the next item -- if there is one, then we also
         * need to add in the width of a separator.  Otherwise, we don't
         */
        er = this->addressList.nextElement(&item);
        if (er == S_OK) {

            lineWidth += this->separatorWidth;

            /**
             * Is the line too wide?
             */
            if (lineWidth > myWidth && lineWidth != rect.width() && lineWidth != 0) {
                lineWidth = rect.width();
                crows++;
            }
        }
    }

    this->totalRows = crows;

    /**
     * If we've got too many rows, then set up the scrollbar!!!
     */
    if (this->totalRows > MAX_VISIBLE_ROWS) {
        this->scrollBar->show();
        this->scrollBar->setRange(0, totalRows - MAX_VISIBLE_ROWS);
        this->scrollBar->setValue(0);
    } else {
        this->scrollBar->hide();
    }

    /**
     * Now computer the DESIRED number of visible rows
     */
    this->visibleRows = (this->totalRows <= MAX_VISIBLE_ROWS)
                        ? this->totalRows
                        : MAX_VISIBLE_ROWS;

}
 
/**
 *=------------------------------------------------------------------------=
 * SuperLabel::countSelectedItems
 *=------------------------------------------------------------------------=
 * returns a count of selected items in this label.
 *
 * Output:
 *      int
 */
int SuperLabel::countSelectedItems()
{
    LabelItem *item;
    ERRCODE er;
    int c = 0;

    er = addressList.firstElement(&item);
    while (er == S_OK) {
        if (item->selected) c++;

        er = addressList.nextElement(&item);
    }

    return c;
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::drawContentsInBuffer
 *=------------------------------------------------------------------------=
 * draws the control.
 */
void SuperLabel::drawContentsInBuffer()
{
    QFontMetrics metrics(this->font());
    QColorGroup cg = colorGroup();
    LabelItem *item, *nextItem;
    QPainter bufferPainter;
    int lineWidth;
    int myWidth;
    ERRCODE er;
    int curRow;
    int w, top;

    /**
     * Open the painter for painting and fill in the background.
     */
    bufferPainter.begin(_doubleBuffer);
    _doubleBuffer->fill(cg.background());

    myWidth = this->getClientWidth();

    /**
     * now go and draw rows, depending on what our top row is.
     */
    curRow = 0;
    lineWidth = 0;
    top = PAD;
    er = addressList.firstElement(&item);
    while (er == S_OK) {

        /**
         * Get the width of the item we're about to draw, and wrap to the
         * next line if necessary. (but only wrap if there's another item
         * already on this line ... if not, then this item is just too dang
         * big)
         */
        w = metrics.boundingRect(QString(item->name)).width() + 2 * PAD;
        if (lineWidth + w > myWidth && lineWidth != 0) {
            curRow++;
            if (curRow > crowTop) top += this->rowHeight;
            lineWidth = 0;
        }

        /**
         * Before we do any painting, go and see if we have another item.
         */
        er = addressList.nextElement(&nextItem);

        /**
         * If this is a row that needs to be painted, do so now.  We also
         * cache the item's rectangle here in case we need to be able to find
         * it later on in a mouse event or some such thing.
         */
        if (curRow >= crowTop && curRow < crowTop + visibleRows) {

            drawItem(bufferPainter, item,
                     QRect(lineWidth + PAD, top, w, this->rowHeight),
                     er == S_OK);

            item->objectRect = QRect(lineWidth + PAD, top, w, this->rowHeight);

        } else {

            item->objectRect.setCoords(-1, -1, -1, -1);
        }

        item = nextItem;
        lineWidth += w + this->separatorWidth;
    }
    
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::drawItem
 *=------------------------------------------------------------------------=
 * draw the actual item in the list.  We ned the item, a bounding rect,
 * and whether or not there's another item after this.  We will handle
 * things like highlighting for selection, etc...
 *
 * Parameters:
 *      QPainter &                      - [in]  where to paint to.
 *      LabelItem *                     - [in]  item to paint.
 *      const QRect &                   - [in]  drawing bounding rect
 *      bool                            - [in]  is there an item after this?
 */
void SuperLabel::drawItem
(
    QPainter &painter,
    LabelItem *item,
    const QRect &rect,
    bool hasNext
)
{
    QBrush brushSave;
    QPen penSave;

    /**
     * If we're drawing a selected item, then we want to set up another
     * pen and brush, and fill in the background.
     */
    if (item->selected && this->hasFocus()) {
        brushSave = painter.brush();
        penSave = painter.pen();
        
        painter.setPen(this->colorGroup().highlightedText());
        painter.setBrush(this->colorGroup().highlight());

        painter.fillRect(rect, painter.brush());
    }

    /**
     * Draw the text!!
     */
    painter.drawText(rect, Qt::AlignLeft, QString(item->name));

    /**
     * Restore the pen and brush
     */
    if (item->selected && this->hasFocus()) {
        painter.setBrush(brushSave);
        painter.setPen(penSave);
    }

    /**
     * if there's another item, draw the separator now
     */
    if (hasNext) {
        painter.drawText(rect.left() + rect.width() + PAD,
                         rect.top(),
                         this->separatorWidth,
                         this->rowHeight,
                         Qt::AlignLeft,
                         QString(this->separatorText));

    }
}

 

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::emptyList
 *=------------------------------------------------------------------------=
 * nuke da list , baby.
 */
void SuperLabel::emptyList()
{
    LabelItem *item;
    ERRCODE er;

    er = addressList.firstElement(&item);
    while (er == S_OK) {

        if (item->name) localFree(item->name);
        if (item->address) localFree(item->address);

        addressList.removeElement(item);
        localFree(item);

        er = addressList.firstElement(&item);
    }

    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::findItemAtPoint
 *=------------------------------------------------------------------------=
 * finds the item at the given coordinates, or returns NULL if there isn't
 * one at the given coordinates.
 *
 * Parameters:
 *      const QPoint &              - [in]  find the item here.
 *
 * Output:
 *      LabelItem *                 - NULL means no item.
 */
SuperLabel::LabelItem *SuperLabel::findItemAtPoint
(
    const QPoint &point
)
{
    LabelItem *item;
    ERRCODE er;
    QRect nullRect(-1, -1, -1, -1);

    /**
     * Go whipping through the list, looking to see if the given point is
     * within any of the rects, provided they're not a null rect.
     */
    er = addressList.firstElement(&item);
    while (er == S_OK) {

        if (item->objectRect != nullRect) {
            if (item->objectRect.contains(point))
                return item;
        }

        er = addressList.nextElement(&item);
    }

    /**
     * IF we're here, we didn't find it.  return empty-handed.
     */
    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::firstSelectedItem
 *=------------------------------------------------------------------------=
 * returns the first selected item we've got.
 *
 * Output:
 *      LabelItem *
 */
SuperLabel::LabelItem *SuperLabel::firstSelectedItem()
{
    LabelItem *item;
    ERRCODE er;

    er = addressList.firstElement(&item);
    while (er == S_OK) {
    
        if (item->selected) return item;

        er = addressList.nextElement(&item);
    }

    return NULL;
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::focusInEvent
 *=------------------------------------------------------------------------=
 * gained focus.
 *
 * Parameters:
 *      QFocusEvent *                       - [in]  details.
 */
void SuperLabel::focusInEvent
(
    QFocusEvent *event
)
{
    if (showingPopup) return;

    QWidget::focusInEvent(event);
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::focusOutEvent
 *=------------------------------------------------------------------------=
 * lost focus.
 *
 * Parameters:
 *      QFocusEvent *                       - [in]  details.
 */
void SuperLabel::focusOutEvent
(
    QFocusEvent *event
)
{
    if (showingPopup) return;

    QWidget::focusOutEvent(event);
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::getClientWidth
 *=------------------------------------------------------------------------=
 * returns our width minus the width of the scrollbar.
 *
 * Output:
 *      int
 */
int SuperLabel::getClientWidth()
{
    int w = this->width();

    if (this->scrollBar->isVisible())
        w -= this->scrollBar->width();

    return w;
}


/**
 *=------------------------------------------------------------------------=
 * SuperLabel::getHeightHintForWidth
 *=------------------------------------------------------------------------=
 * given a width, return the height we want for this.
 *
 * Parameters:
 *      int                         - [in]  proposed height.
 *
 * Output:
 *      int
 */
int SuperLabel::getHeightHintForWidth
(
    int proposedW
)
{
    int w = this->getClientWidth();
    int proposedH;

    computeVisibleRowCount(proposedW);
    proposedH = visibleRows * this->rowHeight;

    computeVisibleRowCount(w);
    return proposedH;
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::getPopupMenu
 *=------------------------------------------------------------------------=
 * returns the popup menu associated with this control.  also is sure to do
 * things like disable items as appropriate, etc ...
 *
 * Output:
 *      QPopupMenu *                - you must delete after using !!!!
 */
QPopupMenu *SuperLabel::getPopupMenu()
{
    int c;

    /**
     * IF we don't have the menu yet, go and create it now.
     */
    if (!popupMenu) {
        popupMenu = new QPopupMenu(this);
        if (!popupMenu) return NULL;

        popupMenu->insertItem(intlLoadStringX(ID_ADD_TO_ADDR_BOOK), this,
                              SLOT(addSelectedToAddressBook()), 0, ID_ADD_TO_ADDR_BOOK);
        popupMenu->insertSeparator();
        popupMenu->insertItem(intlLoadStringX(ID_PROPERTIES), this,
                              SLOT(showProperties()), 0, ID_PROPERTIES);
    }

    /**
     * Now figure out what ones we need to enable/disable.
     */
    c = countSelectedItems();

    popupMenu->setItemEnabled(ID_PROPERTIES, c == 1);
    popupMenu->setItemEnabled(ID_ADD_TO_ADDR_BOOK, c > 0);

    return popupMenu;
}


/**
 *=------------------------------------------------------------------------=
 * SuperLabel::mouseDoubleClickEvent
 *=------------------------------------------------------------------------=
 * the user double clicked in the control.  Figure out what item is underneath
 * the mouse (if any) and show something for it if there is.
 *
 * Parameters:
 *      QMouseEvent *                   - [in]  details, details, details...
 */
void SuperLabel::mouseDoubleClickEvent
(
    QMouseEvent *event
)
{
    LabelItem *item = findItemAtPoint(event->pos());

    /**
     * If it's no the left button, then we do nothing.
     */
    if (event->button() != Qt::LeftButton) return;

    /**
     * If we found one, then we'll go tell the address book to display 
     * whatever info we've got for it.
     */
    if (item != NULL) {

        showPropertiesForItem(item);
    }

}


/**
 *=------------------------------------------------------------------------=
 * SuperLabel::mousePressEvent
 *=------------------------------------------------------------------------=
 * the mouse has been pressed.  Find out if there is an item beneath it, and
 * if so, select it.
 *
 * Parameters:
 *      QMouseEvent *                   - [in]  details
 */
void SuperLabel::mousePressEvent
(
    QMouseEvent *event
)
{
    QPopupMenu *m;

    LabelItem *item = findItemAtPoint(event->pos());

    /**
     * This will clear the selection if item == NULL
     */
    selectSingleItem(item);

    /**
     * If it's the right button, then show the popup menu as well!
     */
    if (event->button() == Qt::RightButton) {

        m = getPopupMenu();

        showingPopup = true;
        if (m)  m->exec(mapToGlobal(event->pos()));
        showingPopup = false;
    }

}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::nextSelectedItem
 *=------------------------------------------------------------------------=
 * given an item, return the next item after this that is selected.
 *
 * Parameters:
 *      LabelItem *                     - [in]  return first after me.
 *
 * Output:
 *      LabelItem *
 */
SuperLabel::LabelItem *SuperLabel::nextSelectedItem
(
    LabelItem *findNext
)
{
    LabelItem *item = NULL;
    ERRCODE er;

    /**
     * First, find this item.
     */
    er = addressList.firstElement(&item);
    while (er == S_OK) {

        if (item == findNext) break;

        er = addressList.nextElement(&item);
    }

    if (er != S_OK) return NULL;

    /** 
     * Now find the first item after this that is selected.
     */
    er = addressList.nextElement(&item);
    while (er == S_OK) {

        if (item->selected) return item;

        er = addressList.nextElement(&item);
    }

    return NULL;
}


/**
 *=------------------------------------------------------------------------=
 * SuperLabel::paintEvent
 *=------------------------------------------------------------------------=
 * now we need to actually go and paint our control.  We will attempt to 
 * buffer as much as we can in the pixmap buffer, and just blt that as often
 * as possible.
 *
 * Parameters:
 *      QPaintEvent *                   - [in]
 */
void SuperLabel::paintEvent
(
    QPaintEvent *
)
{
    if (mustUpdateBuffer)
        drawContentsInBuffer();

    bitBlt(this, 0, 0, _doubleBuffer);
}


/**
 *=------------------------------------------------------------------------=
 * SuperLabel::resizeEvent
 *=------------------------------------------------------------------------=
 * resizes our control as appropriate.
 * 
 * Parameters:
 *      QResizeEvent *                  - [in]
 */
void SuperLabel::resizeEvent
(
    QResizeEvent *event
)
{
    /**
     * first, we need to resize and invalidate our double buffer.
     */
    mustUpdateBuffer = true;
    if (event)
        _doubleBuffer->resize(event->size());
    else
        _doubleBuffer->resize(this->size());

    /**
     * Resize the scrollbar
     */
    this->scrollBar->setGeometry(this->width() - SCROLLBAR_WIDTH,
                                 0, SCROLLBAR_WIDTH, this->height());

    /**
     * Next, we need to update the number of visible rows in the control.
     */
    computeVisibleRowCount(this->getClientWidth());
    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::scrollBarChanged
 *=------------------------------------------------------------------------=
 * the user has moved the scrollbar -- go and update our status based on
 * that now.
 *
 * Parameters:
 *      int                         - [in]  new scroll value.
 */
void SuperLabel::scrollBarChanged
(
    int newValue
)
{
    this->crowTop = newValue;
    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::selectSingleItem
 *=------------------------------------------------------------------------=
 * selects a single item in the list.
 *
 * Parameters:
 *      LabelItem *                 - [in]  select me, and only me. NULL
 *                                          means select nobody.
 */
void SuperLabel::selectSingleItem
(
    LabelItem *item
)
{
    LabelItem *iter;
    ERRCODE er;

    er = addressList.firstElement(&iter);
    while (er == S_OK) {

        /**
         * Select iff they're the same item.
         */
        iter->selected = (item == iter);

        er = addressList.nextElement(&iter);
    }

    this->forceUpdate();
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::showProperties
 *=------------------------------------------------------------------------=
 * shows the properties for the currently selected item.
 */
void SuperLabel::showProperties()
{
    LabelItem *item;

    item = firstSelectedItem();

    if (item) 
        showPropertiesForItem(item);
}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::showPropertiesForItem
 *=------------------------------------------------------------------------=
 * shows the properties for the given item.
 *
 * Parameters:
 *      LabelItem *                 - [in]  show my props.
 */
void SuperLabel::showPropertiesForItem
(
    LabelItem *item
)
{
    const char *pstr;

    AddressBook *b = localMainWidget->getAddressBook();

    addressGenerateSmtpCopy(item->name, item->address, pstr);

    if (pstr && b)
        b->showAddressEntry(pstr, this);

    if (pstr) localFree(pstr);

}

/**
 *=------------------------------------------------------------------------=
 * SuperLabel::sizeHint
 *=------------------------------------------------------------------------=
 * helps people figure out how big to make us. the width is irrelevant, but
 * the height is of particular use.
 *
 * Output:
 *      QSize const                 - size we would like to be.
 */
QSize SuperLabel::sizeHint()
{
    computeVisibleRowCount(this->getClientWidth());
    return QSize(width(), this->rowHeight * visibleRows);
}


