/*
 *  Copyright (C)  2000-2001 Marc Wandschneider <mw@kiltdown.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation.
 *
 *  For more information look at the file COPYRIGHT in this package
 *
 */
#ifndef _SUPERLABEL_H_
#define _SUPERLABEL_H_

#include <sanity.h>
#include <llist.h>
#include <qrect.h>
#include <qwidget.h>

class QPixmap;
class QScrollBar;
class QPopupMenu;

const int MAX_VISIBLE_ROWS = 4;

/**
 *=------------------------------------------------------------------------=
 * The super label is a label-like widget for email addresses only.  It
 * supports things like tool-tips showing you the current address, and also
 * supports proper wrapping, etc ...
 *=------------------------------------------------------------------------=
 */
class SuperLabel : public QWidget {

    Q_OBJECT


  public:
    SuperLabel(QWidget * parent, const char * name = 0);
    virtual ~SuperLabel();

    /**
     * Adds a name to the list of addresses we're displaying.
     */
    ERRCODE addAddress(const char *);

    /**
     * Helps people figure out how big to make us.
     */
    virtual QSize sizeHint();

    /**
     * General public routines.
     */
    void emptyList();

    /**
     * Get the height of the control for the given width
     */
    int getHeightHintForWidth(int myWidth);



  signals:
    /**
     * The user double clicked on an item.  This tells them which one
     * it is.
     */
    void itemDoubleClicked(const char *);



  private slots:
    /**
     * We need to know where our scrollbar is!
     */
    void scrollBarChanged(int);

    void addSelectedToAddressBook();
    void showProperties();

  protected:
    /**
     * Event routines that we need to override.
     */
    virtual void paintEvent(QPaintEvent *);
    virtual void resizeEvent(QResizeEvent *);
    virtual void focusInEvent(QFocusEvent *);
    virtual void focusOutEvent(QFocusEvent *);
    virtual void mousePressEvent(QMouseEvent *);
    virtual void mouseDoubleClickEvent(QMouseEvent *);



  private:
    /**
     * This is the structure that holds the items in our list.
     */
    struct LabelItem {
        const char *name;
        const char *address;
        bool selected;
        QRect objectRect;
    };

    /**
     * This is the list of items in our control.
     */
    LinkList<LabelItem *> addressList;

    /**
     * What items are separated by.
     */
    const char *separatorText;

    /**
     * This is the number of  rows and the size of a given row.  There
     * is a max value for the former (MAX_VISIBLE_ROWS)
     */
    int totalRows;
    int rowHeight;
    int crowTop;
    int visibleRows;
    int separatorWidth;

    /**
     * For fast painting.
     */
    bool mustUpdateBuffer;
    QPixmap *_doubleBuffer;

    /**
     * If there are too many items, then this scrollbar shows up on the right.
     */
    QScrollBar *scrollBar;

    /**
     * This is the popup menu we will show and a flag to know when we're 
     * showing it so we can take care of some focus artifacts.
     */
    QPopupMenu *popupMenu;
    bool showingPopup;

    /**
     * invalidates our buffer and forces an update.
     */
    void forceUpdate() {
        mustUpdateBuffer = true;
        this->repaint();
    }

    /**
     *=--------------------------------------------------------------------=
     * Private routines
     */
    void computeVisibleRowCount(int myWidth);
    int  countSelectedItems();
    void drawContentsInBuffer();
    void drawItem(QPainter &, LabelItem *item, const QRect &, bool hasNext);
    int  getClientWidth();
    void selectSingleItem(LabelItem *);
    void showPropertiesForItem(LabelItem *);

    LabelItem *findItemAtPoint(const QPoint &);
    QPopupMenu *getPopupMenu();

    /**
     * Use these to iterate over selected items.
     */
    LabelItem *firstSelectedItem();
    LabelItem *nextSelectedItem(LabelItem *);


};



#endif // _SUPERLABEL_H_


